/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#define LOG_DOMAIN "vortex-channel"

#include <vortex_channel.h>

/** 
 * @internal
 *
 * Here is the VortexChannel data structure. It is an opaque structure
 * handled by a set of functions to ensure data hiding principle and
 * to be easily bindable.
 *
 * If you modify this structure you are likely to be interested on \ref
 * vortex_channel_new_emtpy.
 */
struct _VortexChannel {
	gint                    channel_num;

	gint                    last_message_sent;
	gint                    last_message_received;

	gint                    last_reply_sent;
	gint                    last_reply_expected;

	gint                    last_seq_no;
	gint                    last_seq_no_expected;

	gint                    max_seq_no_accepted;
	gint                    max_seq_no_remote_accepted;
	GList                 * pending_messages;

	gint                    last_ansno_sent;
	gint                    last_ansno_expected;

	/* the profile:
	 * 
	 * Defines the profile under the channel is working. */
	gchar                 * profile;
	
	/** 
	 * mime type:
	 *
	 * This variable is used to control which is the default mime
	 * type used on the given channel. If the mime type is the
	 * default one established by the BEEP RFC the mime type is
	 * then omitted. In the case it is different from the default,
	 * it is used automatically while creating frames.
	 */
	gchar                 * mime_type;
	
	/** 
	 * transfer encoding:
	 * ~~~~~~~~~~~~~~~~~~
	 *
	 * This variable is used to control which is the default
	 * transfer encoding value for the given channel. Again, if
	 * the transfer encoding used is different from the default,
	 * then it is placed a notification for the value used while
	 * building frames.
	 */
	gchar                 * transfer_encoding;

	VortexHash            * data;

	/* is_opened
	 *
	 * This value allows to detect if a channel is closed and
	 * cannot be used anymore for any type of transmission. This
	 * value is actually used by vortex_channel_send_msg,
	 * __vortex_channel_common_rpy. */
	gboolean                is_opened;

	
	/* being_closed
	 * 
	 * These values is similar to is_opened but allows a different
	 * error detection. While a channel is closed (a process that
	 * can take a while) BEEP RFC 3080 says no msg can be sent but
	 * more msg can be received and that msg must be reply so,
	 * once a channel is requested to be closed, it can still
	 * receive more msg and must be replied. This value is actually
	 * used by vortex_channel_send_msg. */
	gboolean                being_closed;


	/* window_size
	 *
	 * This value is used to control which is the size to
	 * advertise once a frame is received and a SEQ frame should
	 * be generated. The default value is 4096, which means that
	 * the window size advertised will 4096 plus the seq_no
	 * value. */
	gint                    window_size;


	gboolean                complete_flag;
	GList                 * previous_frames;
	VortexConnection      * connection;
	VortexOnCloseChannel    close;
	gpointer                close_user_data;
	VortexOnFrameReceived   received;
	gpointer                received_user_data;

	/* the waiting_msgno
	 * 
	 * This thread safe queue is mainly used by
	 * vortex_channel_wait_reply and
	 * vortex_channel_invoke_received_handler. A detailed
	 * explanation can be found on at
	 * vortex_channel_invoke_received_handler. 
	 *
	 * As a brief, this waiting queue contains all waiting threads
	 * that are waiting to receive a reply no matter what type:
	 * ERR, RPY, ANS and NUL. */
	VortexQueue           * waiting_msgno;

	/* the send mutex:	
	 *
	 * This mutex is used to make the function
	 * vortex_channel_send_msg to be a critical section. This
	 * avoid race conditions on getting actual message number to
	 * use on other frame header while building and sending a
	 * message.
	 *
	 * This mutex it is also used by __vortex_channel_common_rpy
	 * which is a common implementation for err and rpy
	 * replies. Due to seqno BEEP specification, this value is
	 * shared among all message sent, no matter the type of the
	 * frame actually send, ERR, RPY MSG, ANS or NUL.
	 * 
	 * So this mutex allows to make critical section and mutually
	 * exclusive each function that send one of the previous
	 * message types. */
	GMutex                * send_mutex;


	/* the receive_mutex:
	 *
	 * This mutex is used to make mutually exclusive some parts of
	 * the function vortex_channel_invoke_receive_handler and
	 * vortex_channel_wait_reply. 
	 * 
	 * The wait_reply function it is used to wait blocked for a
	 * specific reply to a msgno so it creates a async queue and
	 * insert it into hash_table that can be used by vortex reader
	 * thread to be able to figure out if there are a waiting
	 * thread or can directly execute the frame received handler.
	 *
	 * The vortex reader also use this mutex through the function
	 * vortex_channel_lock_to_receive and
	 * vortex_channel_unlock_to_receive so it can
	 */
	GMutex                * receive_mutex;

	
	/* the reply_cond:
	 *
	 *  This conditional mutex is used to ensure all replies on the
	 * same channel are sent in order.
	 *
	 * If client application calls to vortex_channel_send_reply
	 * trying to reply a message having previous message without
	 * been replied then application is blocked. */
	GCond                 * reply_cond;

	/* the queue_msgs:
	 *
	 * The queue_msgs is used by the vortex sequencer process as a
	 * buffer of frames to be sent.  The sequencer receive request
	 * of messages to be sent. Then, them are splitted if the
	 * message is larger than channel window size into one or
	 * several frames ready to sent.
	 *
	 * Once the sequencer have queue some frames to be sent it
	 * signal the vortex writer to sent pending frames. */
	VortexQueue           * queue_msgs;


	/* the queue_seqs:
	 *
	 * The queue_seqs is used to organize all SEQ frames to be
	 * sent having a greater priority. This queue and the
	 * queue_msgs are internally handled by the vortex library.
	 */
	VortexQueue           * queue_seqs;


	/* the close_cond:
	 *
	 * This conditional is used to get blocked close message
	 * process and is response. As described in RFC 3080 a channel
	 * can receive the channel close message at any time but, must
	 * still await to receive its MSG replies.
	 * 
	 * Once all replies have been received the channel can be
	 * closed and the ok or error message will be sent. */
	GCond                 * close_cond;

	
	/* the close_mutex:
	 *
	 * This mutex is used in conjunction with close_cond to ensure
	 * all message replies have been received during the close
	 * channel process. */
	GMutex                * close_mutex;


	/* the waiting_replies
	 *
	 * This variable is used by the
	 * vortex_channel_0_frame_received_close_msg to flag a channel
	 * that have been blocked on that function because it didn't
	 * still receive all replies to message already sent. */
	gboolean               waiting_replies;


	/* the reply_processed:
	 *
	 * This variable is used by the vortex reader and close
	 * channel process to synchronize channel frame replies
	 * reception and its process and the possible on going channel
	 * closing. 
	 * 
	 * A channel can be closed until all replies to all message
	 * are receive and they are delivered to application level. */
	gboolean               reply_processed;


	/* the being_sending */
	gboolean               being_sending;

	/* the pending_mutex 
	 *
	 * This mutex is used to ensure we have sent all replies to
	 * all message receive during the close process before getting
	 * blocked waiting for the ok message.  This mutex is used at
	 * __vortex_channel_block_until_replies_are_sent and
	 * vortex_channel_signal_reply_sent_on_close_blocked */
	GMutex                * pending_mutex;

	/* the pending_cond
	 *
	 * The conditional variable is used in conjunction with
	 * previous mutex pending_mutex. */
	GCond                 * pending_cond;

	/* the pool
	 *
	 * If the channel was created inside a pool this variable will
	 * be set to point to the channel it belongs so on close
	 * operation this module will be able to notify vortex channel
	 * pool module to remove the channel reference. */
	VortexChannelPool     * pool;
};

typedef struct _VortexChannelData {
	VortexConnection        * connection;
	gint                      channel_num;
	VortexOnChannelCreated    on_channel_created;
	gchar                   * serverName;
	gchar                   * profile;
	VortexEncoding            encoding;
	gchar                   * profile_content;
	gint                      profile_content_size;
	gpointer                  user_data;
	VortexOnCloseChannel      close;
	gpointer                  close_user_data;
	VortexOnFrameReceived     received;
	gpointer                  received_user_data;
	gboolean                  threaded;
}VortexChannelData;

/** 
 * @internal
 *
 * @brief Free resources allocated by the structure used to create new
 * channels.
 * 
 * @param data The VortexChannelData to be destroyed.
 */
void vortex_channel_data_free (VortexChannelData * data)
{
	if (data == NULL)
		return;
	if (data->serverName)
		g_free (data->serverName);
	if (data->profile_content)
		g_free (data->profile_content);
	g_free (data);
	
	return;
}

/** 
 * @internal
 *
 * @brief Allows to calculate which is the current size of the mime
 * headers that are placed for every message sent inside Vortex
 * Library.
 *
 * Because mime is handled per profile and them per channel, next
 * sequence number to be calculated must consider the size of the mime
 * headers plus the size of the message being sent.
 * 
 * @param channel The channel where the mime header size will be
 * reported.
 */
gint __vortex_channel_get_mime_headers_size (VortexChannel * channel)
{
	gchar     * mime_type         = vortex_channel_get_mime_type (channel);
	gchar     * transfer_encoding = vortex_channel_get_transfer_encoding (channel);
	gboolean    headers_found     = FALSE;
	gint        size              = 0;

	/* check for Content-Type header configuration */
	if (mime_type != NULL && g_strcasecmp (mime_type, "application/octet-stream")) {
		/* Seems that the mime entity header content-type is
		 * defined so, count 16 bytes for the "Content-Type: "
		 * plus 2 bytes more due to ending \x0D\x0A. */
		size           = 16 + strlen (mime_type);
		headers_found  = TRUE;
	}

	/* check for the Content-Transfer-Encoding header configuration */
	if (transfer_encoding != NULL && g_strcasecmp (transfer_encoding, "binary")) {
		size          += 29 + strlen (transfer_encoding);
		headers_found  = TRUE;
	}
	
	/* if some of the previous headers are defined, add 2 more
	 * bytes for the ending \x0D\x0A */
	if (headers_found)
		size += 2;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "mime headers size calculated for channel %d was: %d",
	       channel->channel_num, size);
	
	/* return the result calculated */
	return size;
}

/** 
 * @internal
 *
 * Returns current mime headers that should be added to the message
 * being sent according to the channel that is goint to carry the
 * data.
 * 
 * @param channel The channel where the mime header configuration will
 * be consulted to get current mime headers.
 *
 * @param buffer The buffer to use to report current mime header
 * configuration.
 * 
 */
void __vortex_channel_get_mime_headers (VortexChannel * channel, gchar * buffer)
{
	gchar     * mime_type         = vortex_channel_get_mime_type (channel);
	gchar     * transfer_encoding = vortex_channel_get_transfer_encoding (channel);
	gboolean    headers_found     = FALSE;
	gint        size              = 0;

	/* check for Content-Type header configuration */
	if (mime_type != NULL && g_strcasecmp (mime_type, "application/octet-stream")) {
		/* add mime Content-Type: entity header */
		memcpy (buffer, "Content-Type: ", 14);
		
		/* add mime type value */
		memcpy (buffer + 14, mime_type, strlen (mime_type));

		/* add trailing \x0D\x0A value */
		memcpy (buffer + 14 + strlen (mime_type), "\x0D\x0A", 2);

		/* get current size dumped */
		size           = 16 + strlen (mime_type);
		headers_found  = TRUE;
	}

	/* check for the Content-Transfer-Encoding header configuration */
	if (transfer_encoding != NULL && g_strcasecmp (transfer_encoding, "binary")) {
		/* add mime Content-Type: entity header */
		memcpy (buffer + size, "Content-Transfer-Encoding: ", 27);
		
		/* add mime type value */
		memcpy (buffer + size + 27, mime_type, strlen (mime_type));

		/* add trailing \x0D\x0A value */
		memcpy (buffer + size + 27 + strlen (mime_type), "\x0D\x0A", 2);

		size          += 29 + strlen (transfer_encoding);
		headers_found  = TRUE;
	}
	
	/* if some of the previous headers are defined, add 2 more
	 * bytes for the ending \x0D\x0A */
	if (headers_found)
		memcpy (buffer + size, "\x0D\x0A", 2);

	/* mime header built */
	return;
	
}

/**
 * \defgroup vortex_channel Vortex Channel: Related function to create and manage Vortex Channels.
 */

/**
 * \addtogroup vortex_channel
 * @{
 */

/** 
 * @internal
 * @brief Set channel status to be connected.
 *
 * Private function to set channel connected status to TRUE. The
 * channel connection status is used to know if a channel is actually
 * usable to send or receive data. This avoids using a channel that is
 * being created or closed but that process didn't finished yet.
 * 
 * @param channel the channel to operate.
 */
void __vortex_channel_set_connected (VortexChannel * channel)
{
	g_return_if_fail (channel);

	channel->is_opened = TRUE;

	return;
}

gboolean __vortex_channel_validate_start_reply (VortexFrame * frame, gchar * _profile, VortexChannel * channel)
{

	/* xml document variable declaration */
	axlDoc      * doc;
	axlNode     * node;
	axlError    * error;
	
	/* application variables */
	gchar       * profile;
	gchar       * profile_content;
	gboolean      result;

	/* check frame type response */
	if (vortex_frame_get_type (frame) != VORTEX_FRAME_TYPE_RPY) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "received a negative reply to the start channel num=%d, profile=%s",
		       channel->channel_num, channel->profile);
		/* free resources */
		vortex_support_free (2, channel, vortex_channel_free, frame, vortex_frame_free);
		return FALSE;
	}

	/* parse xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		/* drop a log */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "Unable to parse XML start channel reply document, unable to create the channel: %s",
		       axl_error_get (error));
		axl_error_free (error);
		       
		/* free resources */
		vortex_support_free (2, channel, vortex_channel_free, frame, vortex_frame_free);
		return FALSE;
	}
	
	/* get the root element ( profile element ) */
	node = axl_doc_get_root (doc);
	if (! NODE_CMP_NAME (node, "profile")) {
		
		/* free resources */
		vortex_support_free (3, 
				     channel, vortex_channel_free, 
				     frame, vortex_frame_free,
				     doc, axl_doc_free);
		return FALSE;
	}
	/* check profile requested */
	profile = axl_node_get_attribute_value (node, "uri");
	if (g_strcasecmp (profile, _profile)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "received a profile confirmation which is different from the requested=%s", profile);
		result = FALSE;
	} else
		result = TRUE;
	
	/* check if profile received is what we were expecting */
	if (result) {
		
		/* check for piggyback received */
		profile_content      = axl_node_get_content (node, NULL);
		if (profile_content != NULL) {
			/* log the profile received */
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received profile content: '%s'", profile_content);
			
			/* set piggyback frame */
			vortex_channel_set_piggyback (channel, profile_content);
			
			/* it is not required to make a deallocation
			 * for the profile_content, because the
			 * reference received is not a copy */
		}
	}
	
	/* free resources (the profile allocated and the xml doc in memory) */
	vortex_support_free (2, doc, axl_doc_free, frame, vortex_frame_free);

	if (!result) {
		/* free the channel and the frame in case something went wrong */
		vortex_channel_free (channel);
	}
	return result;
}


/** 
 * @internal
 *
 * Helper function for \ref vortex_channel_new.
 */
gpointer __vortex_channel_new (VortexChannelData * data) 
{
	VortexChannel    * channel    = NULL;
	VortexChannel    * channel0   = NULL;
	VortexFrame      * frame      = NULL;
	WaitReplyData    * wait_reply = NULL;
	gchar            * start_msg  = NULL;
	gint               msg_no;

	/* check if remote peer support actual profile */
	if (!vortex_connection_is_profile_supported (data->connection, data->profile)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "channel profile not supported by remote peer, channel can't be created");
		goto __vortex_channel_new_invoke_caller;
	}

	/* check if channel to be created already exists */
	if ((data->channel_num != 0) && 
	    vortex_connection_channel_exists (data->connection, data->channel_num)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "creating a channel identified by %d that already exists",
		       data->channel_num);
		goto __vortex_channel_new_invoke_caller;
	}

	/* check if user wants to create a new channel and
	 * automatically allocate a channel number */
	if (data->channel_num == 0) {
		data->channel_num      = vortex_connection_get_next_channel (data->connection);
		if (data->channel_num == -1) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "vortex connection get next channel have failed, this could be bad, I mean, really bad..");
			goto __vortex_channel_new_invoke_caller;
		}

	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "channel num returned=%d", data->channel_num);

	/* creates the channel */
	channel = vortex_channel_empty_new (data->channel_num, data->profile, data->connection);
	
	/* set default handlers */
	vortex_channel_set_close_handler    (channel, data->close, data->close_user_data);
	vortex_channel_set_received_handler (channel, data->received, data->received_user_data);
	
	/* get the channel 0 */
	channel0 = vortex_connection_get_channel (data->connection, 0);
	if (channel0 == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "internal vortex error: unable to get channel 0 for a session");

		/* free allocated channel and exist */
		vortex_channel_free (channel);
		channel = NULL;
		goto __vortex_channel_new_invoke_caller;
	}

	/* create wait reply object. */
	wait_reply = vortex_channel_create_wait_reply ();

	/* build up the frame to send */
	start_msg  = vortex_frame_get_start_message (vortex_channel_get_number (channel),
						     data->serverName,
						     data->profile,
						     data->encoding,
						     data->profile_content,
						     data->profile_content_size);
	/* send start message */
	if (!vortex_channel_send_msg_and_wait (channel0, 
					       start_msg,
					       strlen (start_msg),
					       &msg_no, wait_reply)) {
		g_free (start_msg);
		channel = NULL;
		vortex_channel_free_wait_reply (wait_reply);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to send start message for channel %d and profile %s",
		       vortex_channel_get_number (channel),
		       vortex_channel_get_profile (channel));
		goto __vortex_channel_new_invoke_caller;
	}
	g_free (start_msg);

	/* now, remote peer must accept channel creation or deny it
	 * this operation will block us until frame is received */
	frame      = vortex_channel_wait_reply (channel0, msg_no, wait_reply);
	if (frame == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "something have failed while received start message response for channel %d under profile %s",
		       vortex_channel_get_number (channel), 
		       vortex_channel_get_profile (channel));

		vortex_channel_free      (channel);
		channel = NULL;
		goto __vortex_channel_new_invoke_caller;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received reply to start message");

	/* check start reply data */
	if (!__vortex_channel_validate_start_reply (frame, channel->profile, channel)) {
		channel = NULL;
		goto __vortex_channel_new_invoke_caller;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "channel start reply validation stage finished");

	/* register channel on connection */
	if (channel != NULL) {
		/* it seems the channel have been created, know we
		 * have to check if user have forgot to register
		 * profile used. If does, register it for the user. */
		if (!vortex_profiles_is_registered (channel->profile)) {
			vortex_profiles_register (channel->profile,
						  /* use default start handler */
						  NULL, NULL, 
						  /* use default close handler */
						  NULL, NULL,
						  /* use no frame received handler */
						  NULL, NULL);
		}

		/* add the channel created to the connection */
		vortex_connection_add_channel (data->connection, channel);
	}

	__vortex_channel_new_invoke_caller:
	/* log a message if */
	if (channel == NULL) 
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "channel=%d creation have failed", data->channel_num);
	
	/* finally, invoke caller with channel result */
	if (data->threaded) {
		/* invoke the channel created handler */
		data->on_channel_created (channel ? channel->channel_num : -1, 
					  channel ? channel : NULL, 
					  data->user_data);

		/* if piggyback is defined, invoke the frame received
		 * using the piggyback */
		if (vortex_channel_have_piggyback (channel)) {
			/* get the piggyback */
			frame = vortex_channel_get_piggyback (channel);
			
			/* invoke the frame received */
			vortex_channel_invoke_received_handler (data->connection, channel, frame);
		}

		/* free no longer needed data */
		vortex_channel_data_free (data);
		return NULL;
	}

	/* free no longer needed data */
	vortex_channel_data_free (data);
	return channel;
}

/** 
 * @brief Creates a new channel over the given connection.
 *
 * Creates a new channel over the session <i>connection</i> with the
 * channel num <i>channel_num</i> specified. If channel num is 0, the
 * next free channel number is used. The new channel will be created
 * under the terms of the profile defined. Both peers must support the
 * given profile, otherwise the channel will not be created.
 *
 * Because the channel is a <i>MSG/RPY</i> transaction between the
 * client and the server, it will block caller until channel is
 * created.
 * 
 * If you want to avoid getting blocked, you can use
 * \ref VortexOnChannelCreated "on_channel_created" handler. If you define this handler, the
 * function will unblock the caller and create the channel on a
 * separated thread. Once the channel is created, caller will be
 * notified through the callback defined by \ref VortexOnChannelCreated "on_channel_created".
 *
 * There are some events that happens during the channel life. They
 * are \ref VortexOnCloseChannel "close handler" and \ref VortexOnFrameReceived "received handler". They are
 * executed when the event they represent happens. A description for
 * each handler is the following:
 *
 * <ul>
 *
 * <li>
 * <i>close handler</i>: (see more info on vortex_handlers:
 * \ref VortexOnCloseChannel) This handler is executed when a channel close
 * petition have arrived from remote peer. 
 *
 * If you signal to close channel through \ref vortex_channel_close,
 * this handler is NOT executed, but it is executed when the given
 * channel receive a remote channel close request. This handler should
 * return TRUE if channel can be closed or FALSE if not.
 * 
 * This can be convenient if there are more data to be sent and to be
 * able to notify remote peer this is actually happening. 
 * 
 * This handler is optional. You can use NULL and the first level
 * handler for close event will be used. If the first level handler is
 * also not defined, a default handler will be used, which returns
 * always TRUE. This means to agree to close the channel always.
 * 
 * You can also define user data to be passed into <i>close
 * handler</i> on its execution. This user data is
 * <i>close_user_data</i>.</li>
 * 
 * <li><i>received handler</i>: (see more info on vortex_handlers:
 * \ref VortexOnFrameReceived) This handler is executed when a channel
 * receive a frame. 
 * 
 * As the same as <i>close</i> handler definition, if you
 * provide a NULL value for this handler the first level handler will
 * be executed. If this first level handler is not defined, frame is
 * dropped.</li>
 *
 * </ul>
 * 
 * As you can see, there are 2 level of handlers to be executed when
 * events happens on channel life.  You can also see
 * \ref vortex_profiles_register documentation. Also check out
 * \ref vortex_manual_dispatch_schema "this section where is explained in more detail how frames are received". 
 * 
 * Later, to be able to close and free the channels created you must
 * use \ref vortex_channel_close. DO NOT USE \ref vortex_channel_free to close
 * or free channel resources. This function is actually called by
 * Vortex Library when the channel resources can be deallocated.
 *
 * On channel creation, it could happen an application do not register
 * the profile that is going to be used for the new channel. This is a
 * problem because if you don't define the close, start or frame
 * received many problems will appear. 
 * 
 * By default, if \ref vortex_channel_new detects you didn't register
 * a profile, then the function will do it for you. On this automatic profile
 * registration, vortex library will assign the default close and
 * start handler which always replies TRUE to close and start
 * channels, but frame received handler will still not defined.
 * 
 * As a consequence you must ensure to register a <i>frame
 * received</i> handler at first level using vortex_profiles function
 * or ensure to register a frame received handler for this function.
 *
 * Again, you may be asking your self why not simply deny channel
 * creation for those petition which didn't define frame received at
 * any level. This is done because there \ref vortex_manual_dispatch_schema "are still more method" 
 * to retrieve frames from remote peers, bypassing
 * the frame received handlers (for the second and first levels). 
 *
 * This method can be used with \ref vortex_channel_wait_reply which
 * wait for a specific reply. Check out \ref vortex_manual_wait_reply "this section to know more about Wait Reply method".
 *
 * <i><b>NOTE:</b> on threaded mode, channel creation process can fail, as the
 * same as non-threaded mode, so you can also get a null value for
 * both invocation models. 
 * 
 * This executing model is slightly different from
 * \ref vortex_connection_new where the value returned must be checked with
 * \ref vortex_connection_is_ok.
 * 
 * Because connection creation is made only once, and channel creation
 * many, the model to check values for new channel created is to check if
 * returned values is not NULL.</i>
 * 
 *
 * @param connection           session where channel will be created.
 * @param channel_num          the channel number. Using 0 automatically assigns the next channel number free.
 * @param profile              the profile under the channel will be created.
 * @param close                handler to manage channel closing.
 * @param close_user_data      user data to be passed in to close handler.
 * @param received             handler to manage frame reception on channel.
 * @param received_user_data   data to be passed in to <i>received</i> handler.
 * @param on_channel_created   a callback to be able to make channel process to be async.
 * @param user_data            user data to be passed in to \ref VortexOnChannelCreated "on_channel_created".
 * 
 * @return the newly created channel or NULL if fails under
 * non-threaded model.  Under threaded model returned value will
 * always be NULL and newly channel created will be notified on
 * \ref VortexOnChannelCreated "on_channel_created".
 */
VortexChannel * vortex_channel_new (VortexConnection      * connection, 
				    gint                    channel_num, 
				    gchar                 * profile,
				    VortexOnCloseChannel    close,
				    gpointer                close_user_data,
				    VortexOnFrameReceived   received,
				    gpointer                received_user_data,
				    VortexOnChannelCreated  on_channel_created, gpointer user_data)
{
	return vortex_channel_new_full (connection, channel_num,        
					NULL,                           /* serverName */ 
					profile,
					EncodingUnknown,                /* encoding value */
					NULL,                           /* profile content data */
					0,                              /* profile content size */
					close, close_user_data,         /* close handlers stuff */
					received, received_user_data,   /* frame received stuff */
					on_channel_created, user_data); /* on channel created   */
					
					
}

/** 
 * @brief Extended version for \ref vortex_channel_new, supporting all options available while creating new channels.
 *
 * This function allows to create a new channel like \ref
 * vortex_channel_new does, but providing support for all options that
 * are available described by the BEEP definition. Options provided by this function not provided by
 * \ref vortex_channel_new are:
 *
 *   - This function allows to set the <b>serverName</b> parameter for
 *   the start element. This parameter is used by some profiles
 *   definitions, as <b>TLS</b>, to establish default certificate to
 *   be used inside the session tuning.
 *
 *   - This function allows to configure the <b>profile element content</b> and
 *   its <b>encoding</b>. Again, some profiles definition needs to send
 *   additional data inside the profile element during the channel
 *   creation exchange. 
 *
 * You can check the full documentation on how to use the function at
 * \ref vortex_channel_new. Apart from previous parameters, this
 * function does the same operations.
 *
 * While sending profile content data you must provide its
 * encoding. Allowed encoding you have to use are none and
 * base64. Vortex Library allows you to report which encoding is used
 * for the profile content with \ref VortexEncoding. By default,
 * unless explicit use by the profile definition or by your
 * requirements, you should use \ref EncodingNone. 
 *
 * This function doesn't encode data received into the encoding
 * selected. This is a task reserved to the profile implementation or
 * the application level. Encoding value is used to report to the
 * remote peer which encoding is being used.
 *
 * For the content profile data and the serverName this function will
 * perform a local copy of the given data. If one of both is NULL it
 * will ignored.
 *
 * 
 * @param connection           Session where channel will be created.
 * @param channel_num          The channel number. Using 0 automatically assigns the next channel number free.
 * @param serverName           Data sent to the remote site as an initial data exchanged.
 * @param profile              The profile under the channel will be created.
 * @param encoding             Profile content encoding used.
 * @param profile_content      The profile content data to be sent at the channel creation. This data is also referred as piggyback. The function will perform a copy of the content profile passed in.
 * @param profile_content_size How huge is the content profile data.
 * @param close                Handler to manage channel closing.
 * @param close_user_data      User data to be passed in to close handler.
 * @param received             Handler to manage frame reception on channel.
 * @param received_user_data   Data to be passed in to <i>received</i> handler.
 * @param on_channel_created   A callback to be able to make channel process to be async.
 * @param user_data            User data to be passed in to \ref VortexOnChannelCreated "on_channel_created".
 * 
 * @return the newly created channel or NULL if fails under
 * non-threaded model.  Under threaded model returned value will
 * always be NULL and newly channel created will be notified on
 * \ref VortexOnChannelCreated "on_channel_created".
 */
VortexChannel * vortex_channel_new_full (VortexConnection      * connection,
					 gint                    channel_num, 
					 gchar                 * serverName,
					 gchar                 * profile,
					 VortexEncoding          encoding,
					 gchar                 * profile_content,
					 gint                    profile_content_size,
					 VortexOnCloseChannel    close,
					 gpointer                close_user_data,
					 VortexOnFrameReceived   received,
					 gpointer                received_user_data,
					 VortexOnChannelCreated  on_channel_created, gpointer user_data)
{
	VortexChannelData * data;

	g_return_val_if_fail (connection, NULL);
	g_return_val_if_fail (vortex_connection_is_ok (connection, FALSE), NULL);
	g_return_val_if_fail (profile, NULL);
	

	/* creates data to be passed in to __vortex_channel_new */
	data                        = g_new0 (VortexChannelData, 1);
	data->connection            = connection;
	data->channel_num           = channel_num;
	data->close                 = close;
	data->close_user_data       = close_user_data;
	data->received              = received;
	data->received_user_data    = received_user_data;

	/* server name data */
	data->serverName            = g_strdup (serverName);

	/* profile stuff */
	data->profile               = profile;
	data->encoding              = encoding;
	data->profile_content_size  = profile_content_size;
	
	/* profile content  */
	if (profile_content != NULL && profile_content_size > 0) {
		data->profile_content       = g_new0 (gchar, profile_content_size + 1);
		memcpy (data->profile_content, profile_content, profile_content_size);
	}

	data->on_channel_created    = on_channel_created;
	data->user_data             = user_data;
	data->threaded              = (on_channel_created != NULL);

	/* launch threaded version if on_channel_created is defined */
	if (data->threaded) {
		vortex_thread_pool_new_task ((GThreadFunc) __vortex_channel_new, data);
		return NULL;
	}

	return __vortex_channel_new (data);
}

/** 
 * @brief Allows to create a new channel using all possible option and
 * making possible to define profile content using a printf-like
 * syntax.
 * 
 * This function behaves the same way than \ref
 * vortex_channel_new_full but allowing to define the profile content
 * using the printf-like format. This could be handy while sending
 * profile content that is more elaborated than just a simple string. 
 *
 * For more information see \ref vortex_channel_new_full and \ref
 * vortex_channel_new.
 * 
 * 
 * @param connection             The connection where the channel will be created.
 * @param channel_num            The channel num requested to create.
 * @param serverName             The serverName value for this channel request.
 * @param profile                The profile requested.
 * @param encoding               The content profile encoding.
 * @param close                  The close channel handler.
 * @param close_user_data        User space data to be passed in to the close channel handler.
 * @param received               The frame received handler.
 * @param received_user_data     User space data to be passed in to the frame received handler.
 * @param on_channel_created     Async notification callback to be executed once the channel is created or not.
 * @param user_data              User space data to be passed in to <b>on_channel_created</b>
 * @param profile_content_format The profile content using a printf-like format: "<blob>%s</blob>"
 * 
 * @return                       A new channel created or NULL if the channel wasn't possible to be created.
 */
VortexChannel     * vortex_channel_new_fullv                   (VortexConnection      * connection,
								gint                    channel_num, 
								gchar                 * serverName,
								gchar                 * profile,
								VortexEncoding          encoding,
								VortexOnCloseChannel    close,
								gpointer                close_user_data,
								VortexOnFrameReceived   received,
								gpointer                received_user_data,
								VortexOnChannelCreated  on_channel_created, 
								gpointer user_data,
								gchar                 * profile_content_format, 
								...)
{
	va_list         args;
	gchar         * profile_content;
	VortexChannel * channel;

	/* initialize the args value */
	va_start (args, profile_content_format);

	/* build the message */
	profile_content = g_strdup_vprintf (profile_content_format, args);

	/* end args values */
	va_end (args);

	/* create the channel */
	channel = vortex_channel_new_full (connection, 
					   channel_num,
					   serverName,
					   profile,
					   encoding,
					   profile_content,
					   strlen (profile_content),
					   close,
					   close_user_data,
					   received,
					   received_user_data,
					   on_channel_created,
					   user_data);
	/* free no longer needed profile content */
	g_free (profile_content);

	/* return value returned by this function */
	return channel;
	
	
}

/** 
 * @brief Returns the next message number to be used on this channel.
 *
 * This message number is the one used to message sent. This have
 * nothing to do with next message number expected.
 *
 * This function is used by the Vortex Library for its internal
 * process to figure out what message number to use on next sending request.
 * 
 * @param channel the channel to work on
 * 
 * @return the message number or -1 if fail
 */
gint            vortex_channel_get_next_msg_no (VortexChannel * channel)
{	
	g_return_val_if_fail (channel, -1);

	return channel->last_message_sent + 1;
}

/**
 * vortex_channel_get_next_expected_msg_no:
 * @channel: 
 * 
 * Returns actual message number expected to be received on this channel. This function must
 * be not confused with vortex_channel_get_next_msg_no. Explanation:
 * Once a channel is created, both peer connected by a session where the channel is created, are
 * expected message from each other and will send messages to each other. In this context, 
 * message that peer A sends to peer B are controlled by vortex_get_next_expected_msg_no on B peer
 * side. This means, B is expecting message 12 to arrive and this '12' will be returned by previous
 * function. Once a message is properly received and processed, this previous number will updated to
 * 13. 
 * Now, when A sends to peer B a message, A needs to now what message number to use on its next
 * message. That why vortex_channel_get_next_msg_no stands for. This function returns actual message
 * number to use or actual message number is waiting other peer to receive.
 *
 * This explanation applies to all frame indication found on this module. As a short path to remember
 * how this function works think as follow:
 * "If actual peer is going to send a message I will use vortex_channel_get_next_* functions to build up
 * it. If I need to validate that incoming message is right I will use vortex_channel_get_next_expected_*
 * functions.
 * 
 * Return value: 
 **/

/** 
 * @brief Returns actual message number expected to be received on this channel.
 *
 *
 * This function must be not confused with
 * \ref vortex_channel_get_next_msg_no. 
 * 
 * Explanation: Once a channel is created, both peer connected by a
 * session where the channel is created, are expecting messages from
 * each other and will send messages to each other. 
 * 
 * On this context, message that peer A sends to peer B are controlled
 * by vortex_get_next_expected_msg_no on B peer side. This means, B is
 * expecting message 12 to arrive and this '12' will be returned by
 * previous function. 
 * 
 * Once a message is properly received and processed, this previous
 * number will updated to 13.  Now, when A sends to peer B a message,
 * A needs to now what message number to use on its next message. That
 * why vortex_channel_get_next_msg_no stands for. 
 * 
 * This function returns actual message number to use or actual
 * message number is waiting other peer to receive.
 *
 * This explanation applies to all frame indication found on this
 * module. As a short path to remember how this function works think
 * as follow: 
 * 
 * <i>"If actual peer is going to send a message I will use
 * vortex_channel_get_next_* functions to build up it. If I need to
 * validate that incoming message is right I will use
 * vortex_channel_get_next_expected_* functions."</i>
 *
 * 
 * @param channel the channel to operate on.
 * 
 * @return expected message number to receive or -1 if fails
 */
gint            vortex_channel_get_next_expected_msg_no (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->last_message_received + 1;
}


/** 
 * @internal
 * @brief Creates a empty channel structure
 *
 * 
 * Creates a new empty channel struct, filling it with passed in
 * data. This function should not be used from outside of vortex
 * library. It is used for internal vortex library purposes.
 * 
 * Its basic behavior is to create a ready to use channel initialized
 * with the correct internal state to be used inside vortex library.
 * So, it can't be useful for vortex consumers.
 * 
 * @param channel_num the channel number to use.
 * @param profile the profile that will manage this channel.
 * @param connection the connection where channel is associated.
 * 
 * @return a new allocated channel.
 */
VortexChannel * vortex_channel_empty_new (gint channel_num,
					  gchar * profile,
					  VortexConnection * connection)
{
	VortexChannel * channel = NULL;

	g_return_val_if_fail (profile, NULL);
	g_return_val_if_fail (vortex_connection_is_ok (connection, FALSE), NULL);

	channel                                 = g_new0 (VortexChannel, 1);
	channel->channel_num                    = channel_num;
	channel->last_message_sent              = -1;
	channel->last_message_received          = -1;
	channel->last_reply_sent                = 0;
	channel->last_reply_expected            = 0;
	channel->last_seq_no                    = 0;
	channel->last_seq_no_expected           = 0;
	channel->is_opened                      = FALSE;
	channel->being_closed                   = FALSE;
	channel->complete_flag                  = TRUE;
	channel->profile                        = g_strdup (profile);
	channel->connection                     = connection;
	channel->window_size                    = 4096;
	/* Attention commander: the following two values *must* be
	 * 4095 because they represent the maximum seq no value
	 * accepted at the channel start up, which is the same to say,
	 * I'm willing to accept 4096 bytes (counting from 0 up to
	 * 4095 yields 4096) */
	channel->max_seq_no_accepted            = 4095;
	channel->max_seq_no_remote_accepted     = 4095;
	channel->pending_messages               = NULL;
	channel->waiting_msgno                  = vortex_queue_new ();
	channel->queue_msgs                     = vortex_queue_new ();
	channel->queue_seqs                     = vortex_queue_new ();
	channel->send_mutex                     = g_mutex_new ();
	channel->receive_mutex                  = g_mutex_new ();
	channel->reply_cond                     = g_cond_new  ();
	channel->close_cond                     = g_cond_new  ();
	channel->close_mutex                    = g_mutex_new ();
	channel->pending_mutex                  = g_mutex_new ();
	channel->pending_cond                   = g_cond_new  ();
	channel->waiting_replies                = FALSE;
	channel->reply_processed                = TRUE;
	channel->data                           = vortex_hash_new_full (g_str_hash, g_str_equal, NULL, 
									__vortex_connection_data_free_value);

	/* The following are a set of values implicitly accepted as
	 * default for the channel 0.  It includes:
	 * 
	 *  - setting the frame receive function so all message
	 *  received on channel 0 are handled by *
	 *  vortex_channel_0_frame_received.
	 *
	 *  - setting default mime content, letting transfer encoding
	 *  to be the default. */ 
	if (channel_num == 0) {
		channel->received              = vortex_channel_0_frame_received;
		channel->mime_type             = "application/beep+xml";
		channel->transfer_encoding     = "binary";

		/* some message number exceptions */
		channel->last_message_received = 0;
		channel->last_message_sent     = 0;
	}else {
		channel->mime_type         = vortex_profiles_get_mime_type (profile);
		channel->transfer_encoding = vortex_profiles_get_transfer_encoding (profile);
	}

	return channel;
}

/** 
 * @brief Allows to set close channel to the given channel.
 *
 * Set the handler to be executed on this channel when a close channel
 * event arrives.  If you call several times to this function, the
 * handler and user data will be changed to passed in data. So you can
 * use this function to reset handlers definition on this channel.
 *
 * @param channel the channel to configure
 * @param close the close handler to execute
 * @param user_data the user data to pass to <i>close</i> execution
 */
void            vortex_channel_set_close_handler (VortexChannel * channel,
						  VortexOnCloseChannel close,
						  gpointer user_data)
{
	g_return_if_fail (channel);
	
	channel->close           = close;
	channel->close_user_data = user_data;
	
	return;
}

/** 
 * @brief Allows to set the frame received handler.
 *
 * Set the handler to be executed on this channel when a <i>received</i>
 * channel event arrives.  If you call several times to this function,
 * the handler and user data will be changed to passed in data. So you
 * can use this function to reset handlers definition on this channel.
 *
 * See also the \ref vortex_manual_implementing_request_response_pattern "following document to get useful information" about
 * using this function.
 *
 * 
 * @param channel the channel to configure.
 * @param received the start handler to execute.
 * @param user_data the user data to pass to <i>received</i> execution.
 */
void            vortex_channel_set_received_handler (VortexChannel * channel, 
						     VortexOnFrameReceived received,
						     gpointer user_data)
{
	g_return_if_fail (channel);

	channel->received           = received;
	channel->received_user_data = user_data;

	return;
}

/** 
 * @brief Allows to set complete frames flag
 *
 * BEEP protocol defines that an application level message must be
 * splitted into pieces equal to or less than the channel window
 * size. Those pieces are called frames.
 * 
 * This enable BEEP peers to avoid receiving large messages and other
 * problems relationed with channel starvation. As a consequence,
 * while receiving a message reply, represented by a RPY, ANS or NUL
 * frame type, it could be splitted into several frames.
 * 
 * If you don't want to handle frames segments and then join them back
 * into a single message you can set the complete flag. Once
 * activated, Vortex will join all frames into a single message and
 * deliver it to the application level using the default application
 * delivery defined (second and first level handlers or wait reply
 * method).
 *
 * By default, every channel created have the complete flag activated. 
 *
 * To clarify:
 * \code
 *      
 *       -----------------
 *      |    App level    | Sends a message
 *       -----------------
 *      |    A message    | Which is splitted into
 *       -----------------
 *      | Several frames  | that frames are sent..
 *       -----------------
 *      |  Over a channel |
 *       -----------------
 *              ||
 *              ||
 *       -----------------
 *      |  Over a channel | 
 *       -----------------
 *      | Several Frames  | are received from remote peer side..
 *       -----------------
 *      |    A message    | joined into a single message
 *       -----------------
 *      |   App level     | App level receive a complete message
 *       -----------------
 *    
 * \endcode
 *
 * @param channel the channel to configure.
 * @param value TRUE activate complete flag, FALSE deactivates it.
 */
void               vortex_channel_set_complete_flag            (VortexChannel * channel,
								gboolean value)
{
	g_return_if_fail (channel);

	channel->complete_flag = value;

	return;
}

/** 
 * @internal
 * @brief Returns if the given channel have stored a previous channel.
 *
 * This function is used by the vortex reader to be able to join frame
 * fragments which are bigger than the channel window size.
 *
 * This function should not be useful for Vortex Library consumers.
 *
 * @param channel the channel to operate on.
 * @param frame   New frame received to check if previous frame for this have been already received.
 * 
 * @return TRUE if previous frame is defined or FALSE if not
 */
gboolean           vortex_channel_have_previous_frame          (VortexChannel * channel, 
								VortexFrame   * frame)
{
	GList       * list_aux;
	VortexFrame * previous_frame = NULL;

	g_return_val_if_fail (channel, FALSE);
	g_return_val_if_fail (frame,   FALSE);

	list_aux = g_list_first (channel->previous_frames);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		previous_frame    = list_aux->data;

		if (vortex_frame_are_joinable (previous_frame, frame))
			return TRUE;
	}
	return FALSE;	
}


/** 
 * @internal
 * @brief Allows to get previous frame stored for the given channel.
 * 
 * @param channel the channel to operate on.
 * 
 * @return the previous frame or NULL if fails
 */
VortexFrame      * vortex_channel_get_previous_frame           (VortexChannel * channel, 
								VortexFrame   * frame)
{
	GList       * list_aux;
	VortexFrame * previous_frame = NULL;

	g_return_val_if_fail (channel, FALSE);
	g_return_val_if_fail (frame,   FALSE);

	list_aux = g_list_first (channel->previous_frames);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		previous_frame    = list_aux->data;

		if (vortex_frame_are_joinable (previous_frame, frame)) {
			channel->previous_frames = g_list_remove (channel->previous_frames,
								  previous_frame);
			return previous_frame;
		}
	}

	return NULL;	
}

/** 
 * @internal
 * @brief Allows to set previous frame for the given channel.
 * 
 * @param channel        The channel to operate on.
 * @param previous_frame The previous frame to replace
 * @param new_frame      The frame to be used as a replacement.
 */
void               vortex_channel_set_previous_frame           (VortexChannel * channel, 
								VortexFrame   * previous_frame,
								VortexFrame   * new_frame)
{
	GList       * list_aux;
	VortexFrame * aux_frame;

	g_return_if_fail (channel);
	
	/* previous frame parameter is defined, we have to replace
	 * existing value */
	if (previous_frame != NULL) {
		list_aux = g_list_first (channel->previous_frames);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "previous frames stored: %d", g_list_length (list_aux));
		for (; list_aux ; list_aux = g_list_next (list_aux)) {
			aux_frame = list_aux->data;
			if (vortex_frame_are_equal (previous_frame, aux_frame)) {
				/* frame found, just replace */
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "found previous stored frame, replacing..");
				list_aux->data = new_frame;
				return;
			}
		}
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to find previous stored frame, adding the frame without joining");
	} else
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "previous frame received is null, adding new frame");
	/* add the frame to the list */
	channel->previous_frames = g_list_append (channel->previous_frames, new_frame);

	return;
}

/** 
 * @brief Allows to get actual complete flag status for the given channel.
 *
 * This function allows to get actual complete flag for the given
 * channel. This function is actually used by the vortex reader to be
 * able to join fragment frames. 
 *
 * @param channel the channel to operate on.
 * 
 * @return TRUE if complete flag is defined or FALSE if not.
 */
gboolean           vortex_channel_have_complete_flag           (VortexChannel * channel)
{
	g_return_val_if_fail (channel, FALSE);

	return (channel->complete_flag); 
}

/** 
 * @brief Returns the channel number for selected channel.
 * 
 * @param channel the channel to get the number
 * 
 * @return the channel number or -1 if fail
 */
gint            vortex_channel_get_number (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->channel_num;
}

/** 
 * @internal
 * @brief Update channel status for sending messages.
 *
 * This function is for internal vortex library support. It update
 * actual channel status for messages to be sent.  Function also
 * allows to control what values are updated. This is necessary to
 * update only seqno value for a channel and not the msgno while
 * sending RPY, ANS and NUL frame types messages.
 *
 * @param channel    The channel to update internal status for next message to be sent
 * @param frame_size The frame size of the frame sent.
 * @param update     What values to update
 */
void vortex_channel_update_status (VortexChannel * channel, gint frame_size, WhatUpdate update)
{
	g_return_if_fail (channel);

	/* update msgno */
	if ((update & UPDATE_MSG_NO) == UPDATE_MSG_NO) {
		channel->last_message_sent +=  1;
	}

	/* update rpy no */
	if ((update & UPDATE_RPY_NO) == UPDATE_RPY_NO) {
		channel->last_reply_sent   += 1;
	}

	/* update seqno */
	if ((update & UPDATE_SEQ_NO) == UPDATE_SEQ_NO) {
		channel->last_seq_no        = ((channel->last_seq_no + frame_size) % (MAX_SEQ_NO));
	}

	/* update ansno */
	if ((update & UPDATE_ANS_NO) == UPDATE_ANS_NO) {
		channel->last_ansno_sent   += 1;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "updating channel sending status to: msgno=%d, rpyno=%d, seqno=%d, ansno=%d..",
	       channel->last_message_sent, channel->last_reply_sent, channel->last_seq_no, channel->last_ansno_sent);

	return;
}

/** 
 * @internal
 * @brief Updates channel status for received messages.
 *
 *
 * This function is for internal vortex library support. It update
 * actual channel status for messages to be received. This is actually
 * used by vortex_reader thread to check incoming messages.
 *
 * 
 * @param channel The channel to update internal status for next message to be received 
 * @param frame_size the frame size of the frame received
 * @param update what parts of the channel status to update.
 */
void vortex_channel_update_status_received (VortexChannel * channel, gint frame_size,
					    WhatUpdate update)
{
	g_return_if_fail (channel);
	
	/* update expected msgno */
	if ((update & UPDATE_MSG_NO) == UPDATE_MSG_NO) {
		channel->last_message_received += 1;
	}

	/* update expected rpy no */
	if ((update & UPDATE_RPY_NO) == UPDATE_RPY_NO) {
		channel->last_reply_expected   += 1;
	}

	/* update expected seqno */
	if ((update & UPDATE_SEQ_NO) == UPDATE_SEQ_NO) {
		channel->last_seq_no_expected   = ((channel->last_seq_no_expected + frame_size) % (MAX_SEQ_NO));
	}
	
	/* update expected ansno */
	if ((update & UPDATE_ANS_NO) == UPDATE_ANS_NO) {
		channel->last_ansno_expected    += 1;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "updating channel receiving status to: msgno=%d, rpyno=%d, seqno=%d, ansno=%d..",
	       channel->last_message_received, 
	       channel->last_reply_expected, 
	       channel->last_seq_no_expected, 
	       channel->last_ansno_expected);

	return;
}


/** 
 * @internal
 * @brief Common function support other function to send message.
 * 
 * @param channel the channel where the message will be sent.
 * @param message the message to be sent.
 * @param message_size the message size.
 * @param msg_no message number reference used for this message sending attempt.
 * @param wait_reply optional wait reply to implement Wait Reply method.
 * 
 * @return TRUE if channel was sent or FALSE if not.
 */
gboolean    __vortex_channel_send_msg_common (VortexChannel   * channel,
					      gchar           * message,
					      gint              message_size,
					      gint            * msg_no,
					      WaitReplyData   * wait_reply)
{
	VortexSequencerData * data;
	gint                  mime_header_size;

	g_return_val_if_fail (channel,                           FALSE);
	g_return_val_if_fail (message,                           FALSE);
	g_return_val_if_fail (channel->is_opened,                FALSE);

	if (channel->channel_num != 0)
		g_return_val_if_fail (! channel->being_closed,   FALSE);

	/* lock send mutex */
	g_mutex_lock (channel->send_mutex);

	/* flag the channel to be awaiting to process message */
	channel->being_sending = TRUE;

	/* prepare data to be signaled to the vortex sequencer */
	data                   = g_new0 (VortexSequencerData, 1);
	data->channel          = channel;
	data->type             = VORTEX_FRAME_TYPE_MSG;
	data->channel_num      = vortex_channel_get_number (channel);
	data->msg_no           = vortex_channel_get_next_msg_no (channel);
	data->first_seq_no     = vortex_channel_get_next_seq_no (channel);

	/* get current mime header configuration */
	mime_header_size       = __vortex_channel_get_mime_headers_size (channel);
	data->message_size     = message_size + mime_header_size;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "new message to sent, type=%d channel=%d msgno=%d size (%d) = msg size (%d) + mime size (%d)",
		    data->type, data->channel_num, data->msg_no,
		    data->message_size, message_size, mime_header_size);

	/* copy mime headers according to channel configuration, that
	 * comes from profile configuration. */
	data->message          = g_new0 (gchar, data->message_size + 1);
	if (mime_header_size > 0)
		__vortex_channel_get_mime_headers (channel, data->message);
	
	/* copy message content */
	memcpy (data->message + mime_header_size, message, message_size);

	/* return back message no used  */
	if (msg_no != NULL) 
		(* msg_no) = vortex_channel_get_next_msg_no (channel);

	/* is there a wait reply? */
	if (wait_reply != NULL) {
		
		/* install queue  */
		wait_reply->msg_no_reply = vortex_channel_get_next_msg_no (channel);
		
		/* enqueue reply data on this */
		g_mutex_lock      (channel->receive_mutex);
		vortex_queue_push (channel->waiting_msgno, wait_reply);
		g_mutex_unlock    (channel->receive_mutex);
	}

	/* update channel status */
	vortex_channel_update_status (channel, data->message_size, UPDATE_SEQ_NO | UPDATE_MSG_NO);

	/* send data to sequencer */
	vortex_sequencer_queue_data (data);

	/* unlock send mutex */
	g_mutex_unlock (channel->send_mutex);

	/* flag the channel to be ready (no message being_sending) */
	channel->being_sending = FALSE;

	return TRUE;
}

/** 
 * @brief Allows to send a message.
 *
 * Sends the <i>message</i> over the selected <i>channel</i>. It will
 * create the needed frames to be sent to remote peer connected by
 * session where <i>channel</i> is created.
 *
 * On <i>msg_no</i> will be returned the message number used for this
 * delivery. This can be useful for waiting response for this message
 * on this channel. 
 *
 * A usual question you may have is: could I remove the message chunk
 * I've used to call this function. The answer is yes. This function
 * makes a local copy you can directly free used message as follows:
 *
 * \code 
 *
 *      // we create some dynamically allocated message
 *      gchar * message = g_strdup_printf ("%d - %s ..\n", 10, 
 *                                         "some kind of very important message");
 *
 *      if (!vortex_channel_send_msg (channel, message, strlen (message), NULL)) {
 *             // some kind of error process and free your resources
 *             g_free (message);
 *
 *      }
 *
 *      // you should free dynamically allocated resources no longer
 *      // used just after running vortex_channel_send_msg.
 *
 *      g_free (message);
 *
 * \endcode
 * 
 * @param channel where message will be sent
 * @param message the message to send
 * @param message_size the message size (user space calculated, that is actual <i>message</i> size)
 * @param msg_no returns the message number used for this deliver. If NULL the message number will not be returned.
 * 
 * @return TRUE if message could be sent, FALSE if fail
 */
gboolean        vortex_channel_send_msg   (VortexChannel * channel,
					   gchar         * message,
					   gint            message_size,
					   gint          * msg_no)
{
	return __vortex_channel_send_msg_common (channel, message, message_size, msg_no, NULL);
}

/** 
 * @brief Allows to send message using a printf-like format.
 *
 * This function works and does the same like \ref
 * vortex_channel_send_msg but allowing you the define a message in a
 * printf like format.
 *
 * A simple example can be:
 * \code
 *    vortex_channel_send_msgv (channel, &msg_no,
 *                              "this a value %s", 
 *                              "an string");
 * \endcode
 *
 * It uses stdargs to build up the message to send and uses strlen to
 * figure out the message size to send.  See also \ref
 * vortex_channel_send_msg.
 *
 * 
 * @param channel the channel used to send the message
 * @param msg_no the msgno assigned to message sent
 * @param format the message format (printf-style) 
 * 
 * @return the same as \ref vortex_channel_send_msg.
 */
gboolean        vortex_channel_send_msgv       (VortexChannel * channel,
						gint          * msg_no,
						gchar         * format,
						...)
{
	gchar      * msg;
	gboolean     result;
	va_list      args;

	/* initialize the args value */
	va_start (args, format);

	/* build the message */
	msg = g_strdup_vprintf (format, args);

	/* end args values */
	va_end (args);

	/* execute send_msg */
	result = vortex_channel_send_msg (channel, msg, strlen (msg), msg_no);
	
	/* free value and return */
	g_free (msg);
	return result;
	
}

/** 
 * @brief Allows to send a message and start a \ref vortex_manual_wait_reply "wait reply".
 *
 * This is especial case of \ref vortex_channel_send_msg. This function is
 * mainly used to be able to make a synchronous reply waiting for the
 * message this function is going to sent. 
 *
 * Because in some cases the thread sending the message can run
 * faster, or simply the thread planner have chosen to give greeter
 * priority, than the thread actually does the wait, this function
 * allows to avoid this race condition.
 *
 * Application designer which are planning to call \ref vortex_channel_wait_reply
 * should use this function to send message.
 *
 * You can also read more about Wait Reply Method \ref vortex_manual_wait_reply "here".
 * 
 * @param channel the channel where message will be sent.
 * @param message the message to be sent.
 * @param message_size the message size to be sent.
 * @param msg_no a required integer reference to store the message number used
 * @param wait_reply a Wait Reply object created using \ref vortex_channel_create_wait_reply 
 * 
 * @return the same as \ref vortex_channel_send_msg
 */
gboolean           vortex_channel_send_msg_and_wait               (VortexChannel   * channel,
								   gchar           * message,
								   gint              message_size,
								   gint            * msg_no,
								   WaitReplyData   * wait_reply)
{
	return __vortex_channel_send_msg_common (channel, message, message_size, msg_no, wait_reply);
}

/** 
 * @brief Printf-like version for the \ref
 * vortex_channel_send_msg_and_wait function.
 *
 * This function works the same way \ref
 * vortex_channel_send_msg_and_wait does but supporting printf-like
 * message definition.
 * 
 * @param channel the channel where message will be sent.
 * @param msg_no a required integer reference to store message number used.
 * @param wait_reply a wait reply object created with \ref vortex_channel_create_wait_reply
 * @param format a printf-like string format defining the message to be sent.
 * 
 * @return 
 */
gboolean           vortex_channel_send_msg_and_waitv              (VortexChannel   * channel,
								   gint            * msg_no,
								   WaitReplyData   * wait_reply,
								   gchar           * format,
								   ...)
{
	gchar      * msg;
	gboolean     result;
	va_list      args;

	/* initialize the args value */
	va_start (args, format);

	/* build the message */
	msg = g_strdup_vprintf (format, args);

	/* end args values */
	va_end (args);

	/* execute send_msg */
	result = vortex_channel_send_msg_and_wait (channel, msg, strlen (msg), msg_no, wait_reply);
	
	/* free value and return */
	g_free (msg);
	return result;
}


/** 
 * @internal
 * @brief Common function to perform message replies.
 * 
 * @param channel the channel where the reply will be sent.
 * @param type the reply type for the message.
 * @param message the message included on the reply.
 * @param message_size the message reply size
 * @param msg_no_rpy the message number this function is going to reply.
 * 
 * @return TRUE if reply was sent, FALSE if not.
 */
gboolean  __vortex_channel_common_rpy (VortexChannel    * channel,
				       VortexFrameType    type,
				       gchar            * message,
				       gint               message_size,
				       gint               msg_no_rpy)
{
	VortexSequencerData * data;
	gint                  mime_header_size;

	g_return_val_if_fail (channel,            FALSE);
	/* check message content only if the frame type is not NUL */
	if (type != VORTEX_FRAME_TYPE_NUL)
		g_return_val_if_fail (message,    FALSE);
	g_return_val_if_fail (channel->is_opened, FALSE);
	
	/* lock send mutex */
	g_mutex_lock (channel->send_mutex);
	channel->being_sending = TRUE;

	/* check we are going to reply the message is waiting to be
	 * replied */
	while (vortex_channel_get_next_reply_no (channel) != msg_no_rpy) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "Received a rpy request for message %d while already waiting to reply to %d, blocking caller",
		       msg_no_rpy,
		       vortex_channel_get_next_reply_no (channel));
		g_cond_wait (channel->reply_cond, channel->send_mutex);
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "sending reply for message %d (size: %d)\n", 
	       msg_no_rpy, message_size);
	
	/* create data to be sent by the sequencer */
	data                  = g_new0 (VortexSequencerData, 1);
	data->channel         = channel;
	data->type            = type;
	data->channel_num     = vortex_channel_get_number (channel);
	data->msg_no          = msg_no_rpy;
	data->first_seq_no    = vortex_channel_get_next_seq_no (channel);

	/* get current mime header configuration */
	mime_header_size      = __vortex_channel_get_mime_headers_size (channel);
	data->message_size    = message_size + mime_header_size;
	data->ansno           = channel->last_ansno_sent;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "new reply message to sent size (%d) = msg size (%d) + mime size (%d)",
	       data->message_size, message_size, mime_header_size);
	
	/* copy the message to be send using memcpy */
	if (message != NULL && data->message_size > 0) {
		/* copy mime header configuration if defined */
		data->message = g_new0 (gchar, data->message_size + 1);
		if (mime_header_size > 0)
			__vortex_channel_get_mime_headers (channel, data->message);

		/* copy application level message */
		memcpy (data->message + mime_header_size, message, message_size);
	}

	switch (type) {
	case VORTEX_FRAME_TYPE_NUL:
		/* because tne NUL reply ends a series of ANS reply, we
		 * update the seqno and the next expected reply
		 * number. */
		vortex_channel_update_status (channel, data->message_size, UPDATE_SEQ_NO | UPDATE_RPY_NO);

		/* reset last ansno message sent. */
		channel->last_ansno_sent = 0;
		
		break;
	case VORTEX_FRAME_TYPE_ANS:
		/* due to the ANS reply, we only update the seqno number.  */
		vortex_channel_update_status (channel, data->message_size, UPDATE_SEQ_NO | UPDATE_ANS_NO);
		break;
	case VORTEX_FRAME_TYPE_RPY:
	case VORTEX_FRAME_TYPE_ERR:
		/* update channel status. */
		vortex_channel_update_status (channel, data->message_size, UPDATE_SEQ_NO | UPDATE_RPY_NO);
		break;
	default:
		/* nothing to do */
		g_critical ("Reaching a switch case at common_rpy which should not be possible");
		break;
	}

	/* send data to sequencer */
	vortex_sequencer_queue_data (data);

	/* unlock send mutex */
	g_cond_broadcast (channel->reply_cond);
	g_mutex_unlock   (channel->send_mutex);
	channel->being_sending = FALSE;	

	return TRUE;
}

/** 
 * @brief printf-like version for \ref vortex_channel_send_rpy function.
 *
 * This function is a wrapper to \ref vortex_channel_send_rpy but allowing
 * you to use a prinf-style message format definition. See also
 * vortex_channel_send_rpy.
 * 
 * @param channel the channel where the message will be sent.
 * @param msg_no_rpy the message number to reply over this channel.
 * @param format a printf-like string format to build the message reply to be sent.
 * 
 * @return TRUE if reply was sent, FALSE if not.
 */
gboolean        vortex_channel_send_rpyv       (VortexChannel *channel,
						gint    msg_no_rpy,
						gchar * format,
						...)
{
	gchar      * msg;
	gboolean     result;
	va_list      args;

	/* initialize the args value */
	va_start (args, format);

	/* build the message */
	msg = g_strdup_vprintf (format, args);

	/* end args values */
	va_end (args);

	/* execute send_msg */
	result = vortex_channel_send_rpy (channel, msg, strlen (msg), msg_no_rpy);
	
	/* free value and return */
	g_free (msg);
	return result;
}

/** 
 * @brief Allows to send a message reply using RPY type.
 *
 * Sends a RPY message to received message MSG identified by
 * <i>msg_no_rpy</i>. This functions allows you to reply a to a received MSG
 * in a way you only need to provide the <i>msg_no_rpy</i>.  According to
 * RFC3080: the BEEP Core, a message reply can not be done if there
 * are other message awaiting to be replied.
 *
 * This means calling to \ref vortex_channel_send_rpy trying to send a
 * reply having previous message been waiting for a reply yields to
 * block caller until previous replies are sent.
 *
 * These may seems to be an annoying behavior but have some advantages
 * and can be also avoided easily.  First of all, think about the
 * following scenario: a vortex client (or beep enabled client) sends
 * over the same channel 3 request using 3 MSG frame type without
 * waiting to be replied.
 * 
 * Beep RPC 3080 section "2.6.1 Withing a Single Channel" says
 * listener role side have to reply to these 3 MSG in the order they
 * have come in, if they have entered through the same channel. This ensure client
 * application that its request are sent in order to remote server but also
 * its reply are received in the same order. 
 *
 * This feature is a really strong one because application designer
 * don't need to pay attention how message replies are received no
 * matter how long each one have taken to be processed on the server
 * side.
 *
 * As a side effect, it maybe happen you don't want to get hang until
 * reply is received just because it doesn't make sense for the
 * application to receive all replies in the same order msg were sent.
 * 
 * For this case client application must send all msg over different
 * channel so each MSG/RPY pair is independent from each other. This
 * can also be found on RFC 3080 section "2.6.2 Between Different
 * Channels"
 *
 * As a consequence:
 * <ul>
 *
 *  <li>Message sent over the same channel will receive its replies in
 *  the same order the message were sent. This have a <i>"serial behavior"</i></li>
 *  
 *  <li>Message sent over different channels will receive replies as
 *  fast as the remote BEEP node replies to them. This have a <i>"parallel behavior"</i></li>
 *
 * </ul>
 * 
 * 
 * @param channel the channel where  the message will be sent
 * @param message the message to sent
 * @param message_size the message size
 * @param msg_no_rpy the message number this function is going to reply to.
 * 
 * @return TRUE if message could be sent, FALSE if fail
 */
gboolean        vortex_channel_send_rpy        (VortexChannel *channel,  
						gchar * message,
						gint    message_size,
						gint    msg_no_rpy)
{
	return __vortex_channel_common_rpy (channel, VORTEX_FRAME_TYPE_RPY,
					    message, message_size, msg_no_rpy);
}

/** 
 * @brief Allows to perform an ANS/NUL reply to a given MSG frame received.
 *
 * While replying to a message received, a MSG frame type, the server
 * side may chose to reply to it using a RPY frame type. But, it may
 * chose to use a series of ANS frame type ended by a NUL frame
 * type. This message exchange type is termed as one-to-many.
 *
 * This allows to issue a MSG request and to receive several replies
 * to the same message because the server side have to perform several
 * tasks to complete the MSG request received. Instead of keep on
 * waiting to all tasks to be completed, the server side could send an
 * ANS reply for each partial task completed. Once all tasks are
 * completed, the server side will send the last reply frame type
 * using a NUL frame type.
 *
 * Previous description allows client application to get actual
 * responses from the server side as long as each piece of the global
 * task gets completed.
 *
 * We have described that a series of ANS frame types must be used
 * ended by a NUL frame type. This description is handled using the
 * both function \ref vortex_channel_send_ans_rpy and \ref
 * vortex_channel_send_ans_rpyv, to generate ANS message reply and
 * using \ref vortex_channel_finalize_ans_rpy to generate the last
 * one, payload-empty message, using NUL frame type.
 * 
 * Let's see a piece of code:
 * \code
 *    while (some yet not ready condition) {
 *       // perform some task which is returned by a function
 *       response = function_which_returns_the_task (&response_size);
 *
 *       // send the reply
 *       if (!vortex_channel_send_ans_rpy (channel, response,
 *                                        response_size,
 *                                        // the MSG num we are replying to
 *                                        msg_no_rpy, 
 *                                        // we didn't finish yet on replying 
 *                                        // ANS messages
 *                                        FALSE) {
 *           // well, we have found an error while sending a message
 *           manage_my_specific_error ();
 *       }
 *
 *       // update the condition to break the loop
 *       update_my_internal_loop_conditions ();
 *    }
 *  
 *    // we have finished, signal peer client side that no more ANS
 *    // will be received.
 *    if (!vortex_channel_finalize_ans_rpy (channel, msg_no_rpy) {
 *         // well, we have found an error while sending a message
 *         manage_my_especific_error ();
 *    }
 * \endcode
 *
 * This message style exchange is especially suited for those
 * situations where the client peer wants to get access to some huge
 * or unlimited content, such as streaming, making a request such as:
 * "get that huge file" to be replied by the server side with a series
 * of ANS replies ended by a NUL message.
 * 
 * Vortex Library is prepared to send messages no matter how huge they
 * are, but you have to consider the size of that set of messages in
 * memory. Oversizing you memory resources will cause your app to
 * start trashing or, in some OS, to just fail with unpredictable
 * behaviors.
 * 
 * As a conclusion:
 *
 *   - You can always use one-to-one message exchange using MSG/RPY
 *   but, if the message content replied is considerable, i.e. 4
 *   megabytes, you should consider starting to use MSG/ANS message
 *   exchange style.
 *
 *   - Using a MSG/RPY message exchange having considerable sized
 *   message replies, will cause the client peer side to get no
 *   response until all frames from the remote side have been received
 *   and joined into one single message. This could not be a problem,
 *   but, in the context of graphical user interfaces, the overall
 *   effect is that your applications is hanged, without knowing if
 *   the server side is death or is still processing current reply
 *   issued.
 * 
 *   - Using the ANS/NUL reply semantic is an elegant way to reply
 *   several messages to the same message because allows you to
 *   support that process on top of the Vortex Library API rather than
 *   splitting your message into several pieces to be sent to the
 *   remote side. 
 *
 *   - Once started an one-to-many reply using ANS, it must be ended
 *   using the NUL frame type (\ref vortex_channel_finalize_ans_rpy).
 * 
 * 
 * @param channel      The channel where the reply will be performed.
 * @param message      The message replied.
 * @param message_size The message size replied.
 * @param msg_no_rpy   Message number to reply.
 * 
 * @return TRUE if the given reply was sent, otherwise FALSE is
 * returned.
 */
gboolean           vortex_channel_send_ans_rpy                    (VortexChannel * channel,
								   gchar         * message,
								   gint            message_size,
								   gint            msg_no_rpy)
{
	return __vortex_channel_common_rpy (channel, VORTEX_FRAME_TYPE_ANS,
					    message, message_size,
					    msg_no_rpy);
}

/** 
 * @brief Allows to send ANS message reply using stdargs argument.
 * Check documentation for \ref vortex_channel_send_ans_rpy
 * function. This function only differ in the sense that message to be
 * sent is built from the StdArg received. Here is an example:
 * 
 * \code
 *     // sending a simple message using StdArgs 
 *     if (!vortex_channel_send_ans_rpyv (channel, msg_no_rpy, 
 *                                        "My message contains: %s, with the %d value",
 *                                        "This string value", 8)) {
 *            // manage the error found
 *     }
 * \endcode
 * 
 *   
 * 
 * @param channel      The channel where the reply will be performed.
 * @param msg_no_rpy   Message number to reply.
 * @param format       StdArg (printf-like) message format.
 * 
 * @return TRUE if the message was sent properly.
 */
gboolean           vortex_channel_send_ans_rpyv                   (VortexChannel * channel,
								   gint            msg_no_rpy,
								   gchar         * format,
								   ...)
{
	gchar      * msg;
	gboolean     result;
	va_list      args;

	/* initialize the args value */
	va_start (args, format);

	/* build the message */
	msg = g_strdup_vprintf (format, args);

	/* end args values */
	va_end (args);

	/* execute send_msg */
	result = vortex_channel_send_ans_rpy (channel, msg, strlen (msg), msg_no_rpy);
	
	/* free value and return */
	g_free (msg);
	return result;
}

/** 
 * @brief Allows to finalize the series of ANS reply already sent with a NUL reply.
 *
 * Inside the one-to-many exchange, it is defined that several ANS
 * reply messages are ended by a NUL reply. This function allows you
 * to sent that one NUL ending reply.
 *
 * @param channel The channel where previous ANS reply message was sent.
 *
 * @param msg_no_rpy The message we where replying to and the one to
 * be notified that one-to-many reply have ended.
 * 
 * @return TRUE if the message was sent properly.
 */
gboolean           vortex_channel_finalize_ans_rpy                (VortexChannel * channel,
								   gint            msg_no_rpy)
{
	return __vortex_channel_common_rpy (channel, VORTEX_FRAME_TYPE_NUL,
					    NULL, 0, msg_no_rpy);
}

/** 
 * @brief Allows to reply a message using ERR message type.
 *
 * Sends a error reply to message received identified by
 * <i>msg_no_rpy</i>. This function does the same as \ref vortex_channel_send_rpy
 * but sending a ERR frame, so take a look to its documentation.
 *
 * @param channel the channel where error reply is going to be sent
 * @param message the message
 * @param message_size the message size
 * @param msg_no_rpy the message number to reply
 * 
 * @return TRUE if message was sent or FALSE if fails
 */
gboolean   vortex_channel_send_err        (VortexChannel *channel,  
					   gchar * message,
					   gint    message_size,
					   gint    msg_no_rpy)
{
	return __vortex_channel_common_rpy (channel, VORTEX_FRAME_TYPE_ERR,
					    message, message_size, msg_no_rpy);
}

/** 
 * @brief Printf-like version for \ref vortex_channel_send_err.
 *
 * Allows to send a ERR message type using a printf-like message format.
 * 
 * @param channel the channel where message reply will be sent.
 * @param msg_no_err the message number to reply to.
 * @param format a printf-like string format.
 * 
 * @return TRUE if message was sent of FALSE if fails
 */
gboolean        vortex_channel_send_errv       (VortexChannel * channel,
						gint    msg_no_err,
						gchar * format,
						...)
{
	gchar      * msg;
	gboolean     result;
	va_list      args;

	/* initialize the args value */
	va_start (args, format);

	/* build the message */
	msg = g_strdup_vprintf (format, args);

	/* end args values */
	va_end (args);

	/* execute send_msg */
	result = vortex_channel_send_err (channel, msg, strlen (msg), msg_no_err);
	
	/* free value and return */
	g_free (msg);
	return result;
}

/** 
 *
 * @brief Returns next sequence number to be used. 
 *
 * Return the next sequence number to be used on this channel. This function must not be
 * confused with \ref vortex_channel_get_next_expected_seq_no. 
 *
 * This function is actually used to know what message seq number to
 * use on next frame to be sent. If you what to get next frame seq no
 * to expected to be received on this channel you must use \ref vortex_channel_get_next_expected_seq_no.
 * 
 * @param channel the channel to operate on.
 * 
 * @return the next seq_number to use, or -1 if fail
 */
guint32          vortex_channel_get_next_seq_no (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);
	return channel->last_seq_no;
}

/** 
 * @brief Returns actual channel window size.
 *
 * Maximum window size accepted by remote peer. This is the number of
 * bytes that the remote peer is willing to accept until a new SEQ
 * frame is set.
 * 
 * @param channel the channel to operate on.
 * 
 * @return the window size or -1 if fails
 */
gint            vortex_channel_get_window_size (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);
	
	return channel->window_size;
}

/** 
 * @brief Returns current mime type used for messages exchange perform
 * on the given channel.
 * 
 * @param channel The channel to operate on.
 * 
 * @return Current mime type or NULL if it fails. Value returned must
 * not be deallocated.
 */
gchar            * vortex_channel_get_mime_type                (VortexChannel * channel)
{
	g_return_val_if_fail (channel, NULL);

	return channel->mime_type;
}

/** 
 * @brief Returns current content transfer encoding used for messages exchange perform
 * on the given channel.
 * 
 * @param channel The channel to operate on.
 * 
 * @return Current content transfer encoding or NULL if it fails. Value returned must
 * not be deallocated.
 */
gchar            * vortex_channel_get_transfer_encoding        (VortexChannel * channel)
{
	g_return_val_if_fail (channel, NULL);
	
	return channel->transfer_encoding;
}

/** 
 * @internal
 *
 * @brief Updates current notion about how much bytes is willing to
 * accept the remote peer on the given channel.
 * 
 * While processing SEQ frames, it is needed to update current notion
 * about how much bytes to be accepted by the remote side while
 * sending new frames. This is done because the flow control needed to
 * allow several channel to share the connection bandwidth.
 *
 * Once the vortex reader receives a SEQ frame to a remote peer, this
 * function updates current status for the selected channel.
 * 
 * @param channel The channel where the update operation will be performed.
 *
 * @param frame The SEQ frame received containing information about
 * the current buffer size to be accepted.
 */
void vortex_channel_update_remote_incoming_buffer (VortexChannel * channel, 
						   VortexFrame   * frame)
{
	int new_size;

	g_return_if_fail (channel);
	g_return_if_fail (frame);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Received SEQ frame update operation: channel=%d ackno=%d window=%d",
	       vortex_frame_get_channel (frame), vortex_frame_get_seqno (frame), vortex_frame_get_content_size (frame));

	/* The following two function gets the ackno value and the
	 * window value from the SEQ frame received:
	 *
	 * What we have to keep in mind that SEQ frames represent
	 * first, the channel they applies to, second the last octet
	 * count received from the remote side at the time that the
	 * SEQ frame was generated and finally, the window value,
	 * which is a value representing how many bytes is prepare to
	 * accept remote side on the given channel at the time the SEQ
	 * frame was generated.
	 *
	 * buffer: #####0000000000000000000000
	 *             ^                     ^
	 *             | <-- window size --> |
	 *     ack no -+                     +
	 *    
	 *  # -> octet received
	 *  0 -> free space for an incoming octet.
	 *
	 * So, the maximum seqno value to be accepted for the remote
	 * side is the sum of the ackno value plus the window size
	 * minus one octet. */
	new_size = vortex_frame_get_seqno (frame) + vortex_frame_get_content_size (frame) -1;

	if (new_size <= channel->max_seq_no_remote_accepted) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "Received a SEQ frame specifying a new seq no maximum value (%d = %d + %d) that is smaller than the max seq no stored (%d), protocol violation",
		       new_size, vortex_frame_get_seqno (frame), 
		       vortex_frame_get_content_size (frame),
		       channel->max_seq_no_remote_accepted);
		__vortex_connection_set_not_connected (vortex_channel_get_connection (channel), 
						       "Received a SEQ frame specifying a new seq no maximum value that is smaller than the max seq no stored, protocol violation");
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received SEQ frame, updated maximum seq no allowed from %d to %d",
	       channel->max_seq_no_remote_accepted, new_size);
	channel->max_seq_no_remote_accepted = new_size % (MAX_SEQ_NO);
	return;
}

/** 
 * @internal
 *
 * @brief Allows to get current maximum seq number accepted on the
 * given channel by the remote side.
 *
 * While creating BEEP frames, vortex sequencer uses this information
 * to track which is the maximum number of octets that the channel is
 * willing to accept. This allows to not oversize that value, which
 * would mean a protocol violation.
 *
 * Value returned from this function is the maximum sequence number to
 * be accepted by the remote side including the max seq no value.
 * 
 * @param channel The channel where the value is checked.
 * 
 * @return Current max sequence number to be accepted.
 */
gint vortex_channel_get_max_seq_no_remote_accepted (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->max_seq_no_remote_accepted;
}

/** 
 * @internal
 * 
 * @brief Allows to update current buffer status for the maximum seq
 * no value to be accepted.
 *
 * Maximum sequence number to be accepted is a mechanism to perform
 * flow control so remote side do not send more bytes that the
 * advertised by this function.
 *
 * When this function is executed, vortex reader process have already
 * checked that the frame received is inside the buffer allowed (that
 * is, inside the maximum sequence number allowed).
 * 
 * @param channel The channel to update current status
 *
 * @param frame The containing data to be accepted, with a particular
 * size and sequence number.
 *
 * @return The function returns TRUE if the channel buffer have
 * changed, so a SEQ frame must be notified or FALSE to not notify any
 * change.
 */
gboolean vortex_channel_update_incoming_buffer (VortexChannel * channel, 
						VortexFrame   * frame,
						gint          * ackno,
						gint          * window)
{
	g_return_val_if_fail (channel, FALSE);
	g_return_val_if_fail (frame, FALSE);
	int new_max_seq_no_accepted;
	int consumed_seqno;
	int window_size;

	/* check if the channel is being close, is that is right, do
	 * not generate more SEQ frames for the given channel. This is
	 * to avoid generating a SEQ frame replaying to the <ok /> *
	 * message that is received once the channel is accepted to be
	 * closed.
	 *
	 * Howerver, if the channel is flaged to be "being closed", we
	 * must keep on replying SEQ frames until the <ok /> is
	 * recevied.  This only applies to channel 0 */
	if (channel->being_closed && channel->channel_num == 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "avoiding generate SEQ frame for a channel that is being closed");
		return FALSE;
	}

	/* Next sentence allows to get the next sequence that should
	 * be used for the next frame received on the given
	 * channel. With this value, a check for the maximum sequence
	 * to be accepted have been met and to update current window
	 * size. */
	consumed_seqno = vortex_frame_get_seqno (frame) + vortex_frame_get_content_size (frame);

	/* record current window size so we use the same value here
	 * and below. */
	window_size             = channel->window_size;

	/* generate a new value for the seqno accepted */
	new_max_seq_no_accepted = (consumed_seqno + window_size - 1) % (MAX_SEQ_NO);
	
	/* check if we have to update current maximum sequence number
	 * to be accepted, according to the new max seq no accepted
	 * value. That is, if the value calculated is smaller than the
	 * already advertised, we just do nothing, no update and no
	 * SEQ frame advertisement. This will allow to control flow
	 * data received is the already not advertised window is
	 * configured to be smaller. */
	if (new_max_seq_no_accepted > channel->max_seq_no_accepted) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "updating allowed max seq no to be received from %d to %d (delta: %d)",
		       channel->max_seq_no_accepted, new_max_seq_no_accepted, 
		       (new_max_seq_no_accepted - channel->max_seq_no_accepted));
		/* update new max seq no accepted value */
		channel->max_seq_no_accepted = new_max_seq_no_accepted;
		
		/* notify caller new ackno and window value */
		(* ackno  ) = consumed_seqno;
		(* window ) = window_size;
		return TRUE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "not updating current max seq values: new max seq no: %d < max seq no: %d",
	       new_max_seq_no_accepted, channel->max_seq_no_accepted);
	/* it seems that the application level is slowing down
	 * incoming channel buffer */
	(* ackno  )  = -1;
	(* window ) = -1;
	return FALSE;
}

/** 
 * @internal
 *
 * @brief Allows to get maximum max sequence value to be accepted for
 * incoming data.
 * 
 * @param channel  The channel where this data has been requested.
 * 
 * @return The maximum sequence number.
 */
gint vortex_channel_get_max_seq_no_accepted (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->max_seq_no_accepted;
}

/** 
 * @internal
 *
 * @brief Allows to queue an new message pending to be sequenced
 * because the channel is stalled.
 * 
 * @param channel The channel where the message is going to be queue.
 * @param message The message to be queued.
 */
void               vortex_channel_queue_pending_message         (VortexChannel * channel,
								 gpointer        message)
{
	g_return_if_fail (channel);
	g_return_if_fail (message);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "queueing a new message pending to be sent: %d messages already queued on channel=%d",
	       g_list_length (channel->pending_messages),
	       channel->channel_num);
	channel->pending_messages = g_list_append (channel->pending_messages, message);
	return;
}

/** 
 * @internal
 *
 * @brief Allows to get the next pending message already queued.
 * 
 * @param channel The channel to query for messages to be queued.
 * 
 * @return The next message queued on the channel or NULL if not
 * message queued is pending.
 */
gpointer           vortex_channel_next_pending_message          (VortexChannel * channel)
{
	GList               * list_aux;

	g_return_val_if_fail (channel, NULL);
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "returning the first pending message: current length=%d for channel=%d",
	       g_list_length (channel->pending_messages), channel->channel_num);
	
	/* no pending messages, return NULL */
	if (g_list_length (channel->pending_messages) == 0)
		return NULL;

	/* return the first element */
	list_aux = g_list_first (channel->pending_messages);

	return list_aux->data;
}

/** 
 * @internal
 * 
 * Internal function which allows to check if the provided channel
 * have pending messages to be sequenced.
 */
gboolean         vortex_channel_is_empty_pending_message (VortexChannel * channel)
{
	/* check for the channel reference received */
	g_return_val_if_fail (channel, FALSE);

	/* return if the pending messages are 0 */
	return  (g_list_length (channel->pending_messages) == 0);
}

/** 
 * @internal
 *
 * @brief Allows to remove the next message queued on the pending
 * message queue.
 * 
 * @param channel The channel where the removal operation will take
 * place.
 */
void               vortex_channel_remove_pending_message        (VortexChannel * channel)
{
	GList * list_aux;

	g_return_if_fail (channel);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "removing next pending message from channel=%d, length=%d", 
	       channel->channel_num, g_list_length (channel->pending_messages));

	/* check that the pending messages queue is not empty */
	if (channel->pending_messages == NULL)
		return;

	/* get the first pending message to be removed and check it is not NULL */
	list_aux = g_list_first (channel->pending_messages);
	if (list_aux == NULL)
		return;

	channel->pending_messages = g_list_delete_link (channel->pending_messages, list_aux);
	return;
}

/** 
 * @internal
 * @brief Queue a new frame to be sent by the Vortex Writer.
 *
 * This function is for vortex library internals. This function is not
 * useful to vortex consumer.  It allows vortex sequencer process to
 * queue frames to be sent on this channel.  Later the vortex writer
 * will take all queue message for each channel on each connection a
 * sent them in a round robin manner.
 * 
 * @param channel the channel where frame will be queued
 * @param data the frame to be queued.
 * @return If the packet was queued or not.
 */
gboolean    vortex_channel_queue_frame     (VortexChannel * channel, VortexWriterData * data)
{
	g_return_val_if_fail (channel,             FALSE);
	g_return_val_if_fail (channel->queue_msgs, FALSE);
	g_return_val_if_fail (channel->queue_seqs, FALSE);
	g_return_val_if_fail (data,                FALSE);
	g_return_val_if_fail (vortex_connection_is_ok (channel->connection, FALSE), FALSE);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "pushing a new frame for channel=%d", 
	       channel->channel_num);

	if (data->type == VORTEX_FRAME_TYPE_SEQ) {
		/* queue the seq frame to be the very first to be
		 * sent */
		return vortex_queue_push (channel->queue_seqs, data);
	}

	/* queue a frame to be sent by vortex writer */
	return vortex_queue_push (channel->queue_msgs, data);
}

/** 
 * @internal
 * @brief Returns true is actual frame queue is empty or not. 
 *
 * 
 * @param channel the channel to check.
 * 
 * @return TRUE if frame queue is empty, otherwise FALSE.
 */
gboolean    vortex_channel_queue_is_empty  (VortexChannel * channel)
{
	g_return_val_if_fail (channel, FALSE);

	return (vortex_queue_is_empty (channel->queue_msgs) && vortex_queue_is_empty (channel->queue_seqs));
}

/** 
 * @internal
 * @brief Return the next queued frame to be sent for the given channel.
 * 
 * Get the next frame to be sent over this channel. A description of
 * the result returned by this function can be found at \ref
 * vortex_types
 *
 * This function is for internal vortex purposes so it should not be
 * useful to vortex library consumers.
 * 
 * @param channel the channel to operate on.
 * 
 * @return the frame already prepare to be sent of NULL if fail
 */
VortexWriterData    * vortex_channel_queue_next_msg (VortexChannel * channel)
{
	g_return_val_if_fail (channel, NULL);

	/* if the SEQ frame pending queue is empty, then send more
	 * data */
	if (! vortex_queue_is_empty (channel->queue_seqs))
		return vortex_queue_pop (channel->queue_seqs);

	/* if the SEQ frame queue is not empty, give priority to SEQ
	 * frames. */
	if (! vortex_queue_is_empty (channel->queue_msgs))
		return vortex_queue_pop (channel->queue_msgs);

	/* there are nothing in both queues */
	return NULL;
}

/** 
 * @internal
 *
 * @brief Returns the current status of pending frames to be sent for
 * the given channel.
 *
 * This function will return how many frames are pending to be sent
 * for the given channel.
 * 
 * @param channel the channel where messages pending will be counted.
 * 
 * @return pending message for the given channel
 */
gint               vortex_channel_queue_length                    (VortexChannel * channel)
{
	g_return_val_if_fail (channel, 0);

	/* return current queue length */
	return vortex_queue_get_length (channel->queue_msgs) + vortex_queue_get_length (channel->queue_seqs);
}

/** 
 * @brief Allows to store a pair key/value into the channel
 * 
 * This allows application designers to use the channel as a way to
 * transport data among function or handler execution.
 *
 * To remove store item you will have to store again the key used
 * passing in a NULL value. Example:
 * \code
 *          vortex_channel_set_data (my_channel, "my_key", NULL);
 *          // previous will remove "my_key" pair from the channel
 * \endcode
 *
 * If you are looking to store at value that is have to be deallocated
 * once the channel is released, you can check \ref
 * vortex_channel_set_data_full. That function allows to store
 * functions to be called to release the key and the value stored.
 *
 * @param channel channel where data will be stored.
 * @param key the key index to look up the data store
 * @param value the data to be stored.
 */
void               vortex_channel_set_data                        (VortexChannel * channel,
								   gpointer key,
								   gpointer value)
{
	/* store the value without providing a pointer to destroy the
	 * key and the value */
	vortex_channel_set_data_full (channel, key, value, NULL, NULL);
	return;
}

/** 
 * @brief Allows to store user space that, indexed by the provided
 * key, providing handlers to be called to release them, once the
 * channel is closed.
 * 
 * This function works like: \ref vortex_channel_set_data but
 * extending the API to support two handler that are called to
 * deallocate memory used by the key and the value stored.
 * 
 * @param channel The channel where the data will be stored.
 *
 * @param key The key to index the data stored.
 *
 * @param value The data to store.
 *
 * @param key_destroy The optional key destroy handler to be called to
 * release the particular key provided.
 *
 * @param value_destroy The optional value destroy handler to be
 * called to release the particular value provided.
 */
void             vortex_channel_set_data_full (VortexChannel    * channel, 
					       gpointer           key, 
					       gpointer           value, 
					       GDestroyNotify     key_destroy,
					       GDestroyNotify     value_destroy)
{
	/* perform some environment checks */
	g_return_if_fail (channel);
	g_return_if_fail (key);

	/* check if the value is not null. It it is null, remove the
	 * value. */
	if (value == NULL) {
		vortex_hash_remove (channel->data, key);
		return;
	}

	/* store the data selected replacing previous one */
	vortex_hash_replace (channel->data, key,
			     __vortex_connection_data_create (key, key_destroy, value, value_destroy));
	return;
}

/** 
 * @brief Returns the value indexed by the given key inside the given
 * channel.
 * 
 * @param channel the channel where data will be looked up.
 * @param key the index key to use.
 * 
 * @return the value or NULL if fails. 
 */
gpointer           vortex_channel_get_data                        (VortexChannel * channel,
								   gpointer key)
{
	g_return_val_if_fail (channel, NULL);
	g_return_val_if_fail (key,     NULL);

	/* get the data stored */
	return __vortex_connection_data_get (vortex_hash_lookup (channel->data, key));
	
}

/** 
 * @brief Returns actual reply number to be used.
 * 
 * This reply number must be used to now if a message reply can be
 * done. This represent the first message identifier to be replied.
 * 
 * @param channel the channel to query
 * 
 * @return the next reply number to use or -1 if fails
 */
gint            vortex_channel_get_next_reply_no          (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);
	
	return channel->last_reply_sent;
}

/**
 * @brief Returns actual reply number expected to be received from
 * remote peer.
 * 
 * @param channel the channel to query
 * 
 * @return the reply number expected or -1 if fails
 */
gint            vortex_channel_get_next_expected_reply_no (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->last_reply_expected;
}

/** 
 * @brief Returns actual seq no expected to be received on this channel. 
 *
 * This function is used to check incoming message seq no.
 * 
 * @param channel the channel to operate on.
 * 
 * @return -1 if fails or expected seq no for this channel.
 */
guint32        vortex_channel_get_next_expected_seq_no (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);
	
	return channel->last_seq_no_expected;
}

/** 
 * @brief Gets next ansno number to be used for the next message to be sent.
 * 
 * @param channel The channel where the ANS message will be sent.
 * 
 * @return The ansno number.
 */
guint32            vortex_channel_get_next_ans_no              (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->last_ansno_sent;
}

/** 
 * @brief Returns the next expected ansno number message to be received.
 * 
 * @param channel The channel where the ANS was received.
 * 
 * @return The ansno number.
 */
guint32            vortex_channel_get_next_expected_ans_no     (VortexChannel * channel)
{
	g_return_val_if_fail (channel, -1);

	return channel->last_ansno_expected;
}

typedef struct {
	VortexChannel              * channel;
	VortexOnClosedNotification   on_closed;
	gboolean                     threaded;
} VortexChannelCloseData;


/** 
 * @brief Validates err message received from remote peer.
 * 
 * @param frame The frame where the error message was received.
 * @param code  The error code inside the frame payload.
 * @param msg   The message code inside the frame payload.
 *
 * @return TRUE if the a error message as found, FALSE if not.
 */
gboolean vortex_channel_validate_err (VortexFrame * frame, 
				      gchar ** code, gchar **msg)
{
	axlDtd   * channel_dtd;
	axlDoc   * doc;
	axlNode  * node;
	axlError * error;

	/* first clear received variables */
	if (code != NULL)
		(* code) = NULL;
	if (msg != NULL)
		(* msg)  = NULL;

	/* get channel management DTD */
	if ((channel_dtd = vortex_dtds_get_channel_dtd ()) == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to load dtd file, cannot validate incoming message, returning error frame");
		return FALSE;
	}

	/* parser xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to parse err reply, we have a buggy remote peer: %s",
		       axl_error_get (error));
		axl_error_free (error);
		return FALSE;
	}

	/* validate de document */
	if (! axl_dtd_validate (doc, channel_dtd, &error)) {
		/* Validation failed */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reply received doesn't contain an error element: %s",
		       axl_error_get (error));
		
		/* free error reported */
		axl_error_free (error);

		/* free document read */
		axl_doc_free (doc);
		return FALSE;
	}
	
	/* get a reference to document root (the error) */
	node  = axl_doc_get_root (doc);
	if (! NODE_CMP_NAME (node, "error")) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "not found error element on err reply, we have a buggy remote peer");
		
		/* free document load */
		axl_doc_free (doc);
		return FALSE;
	}
	
	/* get error code returned */
	if (code != NULL)
		(* code ) = axl_node_get_attribute_value_copy (node, "code");

	/* get the message value returned */
	if (msg != NULL) 
		(* msg  ) = axl_node_get_content_copy (node, NULL);

	/* free not needed data */
	axl_doc_free (doc);
	return TRUE;
}

/** 
 * @internal
 * @brief Validates rpy message received from remote peer for close
 * channel request sent.
 * 
 * @param channel 
 * @param frame 
 */
void __vortex_channel_close_validate_rpy (VortexChannel * channel, 
					  VortexFrame   * frame)
{
	
	axlDtd       * channel_dtd;
	axlDoc       * doc;
	axlNode      * node;
	axlError     * error;
	

	/* get channel management DTD */
	if ((channel_dtd = vortex_dtds_get_channel_dtd ()) == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to load dtd file, cannot validate incoming message, returning error frame");
		return;
	}

	/* parser xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to parse rpy reply, we have a buggy remote peer: %s",
		       axl_error_get (error));
		
		/* release memory used by the error object */
		axl_error_free (error);
		return;
	}

	/* validate the document */
	if (! axl_dtd_validate (doc, channel_dtd, &error)) {
		/* Validation failed */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "validation failed for rpy reply, we have a buggy remote peer: %s",
		       axl_error_get (error));

		/* free error reported */
		axl_error_free (error);
		
		/* free document reported */
		axl_doc_free (doc);
		return;		 
	}
	
	/* get a reference to document root (the error) */
	node  = axl_doc_get_root (doc);
	if (! NODE_CMP_NAME (node, "ok"))
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "not found ok element on rpy reply, we have a buggy remote peer");
	
	/* free not needed data */
	axl_doc_free (doc);
	return;
}

/** 
 * @internal
 * @brief Blocks a channel until all replies to message sent are
 * received.
 *
 * This function is synchronized with vortex reader so that process
 * notifies when a reply have been received.
 *
 * This function uses a mutex (close_mutex) and a conditional var
 * (close_cond) that are signaled from
 * vortex_channel_signal_on_close_blocked.
 * 
 * 
 * @param channel the channel to operate on.
 */
void __vortex_channel_block_until_replies_are_received (VortexChannel * channel)
{
	VortexConnection * connection;
	g_return_if_fail (channel);


	channel->being_closed = TRUE;
	g_mutex_lock (channel->close_mutex);

	connection = vortex_channel_get_connection (channel);

	while (! vortex_channel_is_ready (channel)) {

		channel->waiting_replies = TRUE;
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, 
		       "we still didn't receive all replies for connection=%d channel=%d (MSG %d != RPY %d), %s",
		       vortex_connection_get_id (channel->connection),
		       channel->channel_num,
		       channel->last_message_sent, channel->last_reply_expected - 1,
		       vortex_connection_is_ok (connection, FALSE) ? "block until all replies are received" :
		       "leaving remaining replies because connection is broken..");
		
		if (vortex_connection_is_ok (connection, FALSE))
			g_cond_wait (channel->close_cond, channel->close_mutex);
	}
	g_mutex_unlock (channel->close_mutex);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "we have received the last reply we can now close the channel %d (MSG %d = RPY %d)", 
	       channel->channel_num, channel->last_message_sent, channel->last_reply_expected - 1);

	return;
}



/** 
 * @internal
 * @brief Blocks a channel until all replies to message receive are
 * sent.
 *
 * This function is used by __vortex_channel_close to ensure we send
 * all replies awaited by the remote peer until getting blocked for
 * the ok message.
 *
 * 
 * @param channel the channel to block
 */
void __vortex_channel_block_until_replies_are_sent (VortexChannel * channel)
{
	g_return_if_fail (channel);

	g_mutex_lock (channel->pending_mutex);

	while (channel->last_message_received != (channel->last_reply_sent - 1)) {
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, 
		       "we still didn't sent replies for connection=%d channel=%d, block until all replies are sent (RPY %d != MSG %d)",
		       vortex_connection_get_id (channel->connection),
		       channel->channel_num,
		       (channel->last_reply_sent - 1), channel->last_message_received);
		
		g_cond_wait (channel->pending_cond, channel->pending_mutex);
	}
	g_mutex_unlock (channel->pending_mutex);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "we have sent the last reply we can now close the channel %d (RPY %d = MSG %d)", 
	       channel->channel_num, 
	       (channel->last_reply_sent - 1), 
	       channel->last_message_received);

	return;
}

gpointer __vortex_channel_close (VortexChannelCloseData * data)
{
	VortexChannel              * channel     = NULL;
	VortexChannel              * channel0    = NULL;
	VortexFrame                * frame       = NULL;
	VortexConnection           * connection  = NULL;
	WaitReplyData              * wait_reply  = NULL;
	VortexOnClosedNotification   on_closed   = NULL;
	
	/* TRUE channel is closed, FALSE not. */
	gboolean                     result      = FALSE;
	gboolean                     threaded;
	gchar                      * code        = NULL;
	gchar                      * msg         = NULL;
	gchar                      * close_msg   = NULL;
	gint                         msg_no      = -1;
	
	
	g_return_val_if_fail (data, NULL);

	/* get a reference to received data (comes from
	 * vortex_channel_close) */
	channel    = data->channel;
	on_closed  = data->on_closed;
	threaded   = data->threaded;

	/* free data */
	g_free (data);

	/* According to section 2.3.1.3 "The close message" from RFC
	 * 3080.
	 * 
	 * We are going to close a channel, so the process is as
	 * follows:
	 *
	 * 1) Send a close channel msg indication to remote peer.
	 * But, if there are pending message replies to be received,
	 * wait for them until send close message.
	 *
	 * 2) Once close message have been sent, we have to avoid
	 * sending futher message and only be prepared to still
	 * receive more msg from remote peer. Of course we are allowed
	 * to (and we must) reply to this received MSG frame type.
	 *
	 * 3) Once received, we have to check there are not pending
	 * RPY to be sent and then close the channel.
	 *
	 * 4) At the same time we keep on waiting remote peer response
	 * to close indication.
	 *
	 * step 1) */
	 connection = channel->connection;

	 /* check for a connection already closed */
	if (!vortex_connection_is_ok (connection, FALSE)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "trying to close a channel over a terminated session, this is a bug on client side code, freeing vortex channel resources");
		result = TRUE;
		goto __vortex_channel_close_invoke_caller;
	}


	/* before sending close message we have to ensure all replies
	 * to message sent have been received if not, we can't send
	 * the close message (2.3.1.3 The Close Message BEEP Core) */
	__vortex_channel_block_until_replies_are_received (channel);

	/* get a reference to administrative channel */
	channel0 = vortex_connection_get_channel (connection, 0);

	/* check channel 0 being closed */
	if (channel->channel_num == 0) {
		/* flag the connection to be at close processing. We
		 * set this flag so vortex reader won't complain about
		 * not properly connection close.  To close the
		 * channel 0 means to close the beep session. */
		vortex_connection_set_data (connection, "being_closed", GINT_TO_POINTER (TRUE));
	}

	/* send the close channel message */
	wait_reply = vortex_channel_create_wait_reply ();

	/* create the close message */
	close_msg  = vortex_frame_get_close_message (channel->channel_num, "200",
						     NULL, NULL);

	if (!vortex_channel_send_msg_and_wait (channel0, close_msg, strlen (close_msg), 
					       &msg_no, wait_reply)) {
		/* free close message */
		g_free (close_msg);

		/* because be could close the channel unflag the
		 * "being_closed".  This should be only applied when
		 * closing the channel 0 and an error was get, but the
		 * same sentence cause the same effect on other
		 * channels. */
		vortex_connection_set_data (connection, "being_closed", NULL);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to send close channel message");
		result = FALSE;
		vortex_channel_free_wait_reply (wait_reply);
		goto __vortex_channel_close_invoke_caller;
	}
	/* free close message */
	g_free (close_msg);

	/* step 2)
	 *
	 * This step is actually been fulfilled due to being_closed
	 * attribute. This allows to keep on sending rpy to incoming
	 * msg but not to send more msg. */

	/* step 3)
	 *
	 * ensure we have sent all message replies to remote peer
	 * MSG */
	__vortex_channel_block_until_replies_are_sent (channel);

	frame = vortex_channel_wait_reply (channel0, msg_no, wait_reply);
	switch (vortex_frame_get_type (frame)) {
	case VORTEX_FRAME_TYPE_ERR:
		/* remote peer do not agree to close the channel, get
		 * values returned */
		vortex_channel_validate_err (frame, &code, &msg);
		result = FALSE;
		break;
	case VORTEX_FRAME_TYPE_RPY:
		/* remote peer agree to close the channel */
		__vortex_channel_close_validate_rpy (channel, frame);
		result = TRUE;
		break;
	default:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "unexpected reply found for a close channel indication, remote peer have a buggy implementation");
		/* __vortex_connection_set_not_connected (connection,
		   "unexpected reply found for a close channel
		   indication, remote peer have a buggy
		   implementation"); */
		result = TRUE;
		goto __vortex_channel_close_invoke_caller;
	}

	/* set channel to be closed. It result = TRUE channel will be
	 * marked as closed so vortex reader can check if a message
	 * can be actually delivered */
	channel->is_opened = !result;

 __vortex_channel_close_invoke_caller:
	if (threaded) {
		/* invoke handler */
		on_closed (channel->channel_num, result, code, msg);
	}

	/* free not needed values */
	vortex_support_free (3, code, g_free, msg, g_free, frame, vortex_frame_free);

	/* remove the channel from the connection */
	if (!result && (channel->channel_num == 0)) {
		/* unflag the connection to be at close processing. We
		 * unset this flag because the connection were not
		 * closed */
		vortex_connection_set_data (connection, "being_closed", NULL);
	}

	if (result) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "because channel=%d over connection id=%d have been accepted to be removed, remove it from its session",
		       channel->channel_num,
		       vortex_connection_get_id (connection));
		vortex_connection_remove_channel (connection, channel);
		/* actually, vortex_connection_remove_channel already
		 * does a vortex_channel_free */
	}
	
	/* return the value */
	if (threaded)
		return NULL; /* on threaded mode always returns
			      * NULL */
	return result ? GINT_TO_POINTER (result) : NULL;
}

/** 
 * @brief Close the given channel. 
 *
 * Free all resource allocated by channel and tries to close it if is
 * opened.  Once you have called this function, the channel will no
 * longer be available until it is created again. Once the channel is
 * closed, this function will notify vortex session (the connection
 * associated with this channel) to remove this channel reference from
 * it.
 *
 * In some cases the channel close request that is function issue
 * could be denied. This is because the channel close process may be
 * denied from other side so channel close couldn't be done. This may
 * happen if other side have something remaining to be sent while you
 * are trying to close. This means the function will block you until
 * channel close process reply is receive and then returned to you.
 *
 * Of course you can avoid get blocked on channel close by setting
 * the <i>on_close_notification</i> which is called at the end of the process,
 * notifying if channel was closed or not.
 * 
 * On both models, making the blocking close operation or the
 * non-blocking one using the <i>on_close_notification</i>, if channel
 * is closed, channel resources will be freed so channel pointer
 * provided will no longer pointing to a valid channel.
 * 
 * Another thing about channel close process to keep in mind it that
 * close handler defined at \ref vortex_channel_new or at
 * \ref vortex_profiles_register is NOT executed when \ref vortex_channel_close
 * is called.  Those handlers are actually executed when your channel
 * receive a close indication. This also means this function sends a
 * close indication to remote peer so, if remote peer is also a vortex
 * enabled and have a close handler defined, then it will be executed.
 * 
 * There are an exception to keep in mind. Channel 0 can not be closed
 * using this function. This channel is hard wired to actual session
 * so, there is no reason to support closing channel 0. Channel 0 is
 * closed on session close. If you still want to close channel 0 call
 * \ref vortex_connection_close instead.
 * 
 * During the <i>on_closed</i> handler execution, if defined, you can
 * still get access to the channel being closed. This is provided
 * because you may be interesting on getting the channel status, etc.
 * After <i>on_closed</i> is executed and finished, the channel
 * resources are deallocated.
 * 
 * @param channel the channel to free
 * @param on_closed a optional handler notification
 * 
 * @return TRUE if the channel was closed or FALSE if not. On
 * threaded mode, activated by defining the <i>on_closed</i> handler,
 * always return TRUE. Actual close process result is notified on
 * <i>on_closed</i>
 */
gboolean vortex_channel_close (VortexChannel * channel, 
			       VortexOnClosedNotification on_closed)
{
	
	VortexChannelCloseData * data;
	
	g_return_val_if_fail (channel,                   FALSE);
	g_return_val_if_fail (channel->channel_num >= 0, FALSE);

	/* check the channel have not received another close petition
	 * from another thread or is being close due to a remote peer
	 * petition */
	g_return_val_if_fail (! channel->being_closed,   FALSE);

	/* set this channel to be in closed channel process */
	channel->being_closed = TRUE;
	
	/* creates data to be passed in to __vortex_channel_close  */
	data            = g_new0 (VortexChannelCloseData, 1);
	data->channel   = channel;
	data->on_closed = on_closed;
	data->threaded  = (on_closed != NULL);

	/* launch threaded mode */
	if (data->threaded) {
		vortex_thread_pool_new_task ((GThreadFunc) __vortex_channel_close, data);
		return TRUE;
	}

	/* launch blocking mode */
	return (__vortex_channel_close (data) != NULL);
}


/** 
 * @brief Returns the current profile the channel is running.
 * 
 * You must not free returned value. You can also check \ref
 * vortex_channel_is_running_profile if you are looking to check for
 * some profile to be run under the given channel.
 * 
 * 
 * @param channel the channel to operate on
 * 
 * @return the profile the channel have or NULL if fails
 */
gchar         * vortex_channel_get_profile (VortexChannel * channel)
{
	g_return_val_if_fail (channel, NULL);

	return channel->profile;
}

/** 
 * @brief Allows to check if the given channel is running a particular profile.
 *
 * \code
 * // check if my channel is running XML-RPC profile
 * if (vortex_channel_is_running_profile (myChannel, VORTEX_XML_RPC_PROFILE)) {
 *     // sure, it is running the XML-RPC profile
 * }
 * \endcode
 * 
 * @param channel The channel to check for running some particular profile.
 * @param profile The profile to be checked.
 * 
 * @return TRUE if the channel is running the given profile, otherwise FALSE is returned.
 */
gboolean           vortex_channel_is_running_profile           (VortexChannel * channel,
								gchar         * profile)
{
	g_return_val_if_fail (channel, FALSE);
	g_return_val_if_fail (profile, FALSE);

	return (!g_strcasecmp (profile, vortex_channel_get_profile (channel)));
}

/** 
 * @brief Returns actual channel session (or Vortex Connection). 
 * 
 * @param channel the channel to operate on.
 * 
 * @return the connection where channel was created or NULL if
 * fails.
 */
VortexConnection * vortex_channel_get_connection               (VortexChannel * channel)
{
	g_return_val_if_fail (channel, NULL);

	return channel->connection;
}

/** 
 * @brief Return whenever the received handler have been defined for this
 * channel.
 * 
 * @param channel the channel to check
 * 
 * @return If received handler have been defined returns TRUE if
 * not returns FALSE.
 */
gboolean        vortex_channel_is_defined_received_handler (VortexChannel * channel)
{
	g_return_val_if_fail (channel, FALSE);
	return (channel->received != NULL);
}


typedef struct _ReceivedInvokeData {
	VortexChannel * channel;
	VortexFrame   * frame;
	gint            channel_num;
}ReceivedInvokeData;

/** 
 * @internal
 *
 * @brief Support function for vortex_channel_invoke_received_handler.
 *
 * This is the function that finally perform the frame deliverance
 * inside the user space application.
 * 
 * @param data Data to deliver the frame received onto the channel.
 * 
 * @return 
 */
gpointer __vortex_channel_invoke_received_handler (ReceivedInvokeData * data)
{
	VortexChannel    * channel      = data->channel;
	VortexConnection * connection   = vortex_channel_get_connection (channel);
	VortexFrame      * frame        = data->frame;
	gchar            * raw_frame    = NULL;
	gboolean           is_connected;

	/* get a reference to channel number so we can check after
	 * frame received handler if the channel have been closed.
	 * Once the frame received have finished this will help us to
	 * know if application space have issued a close channel. */
	gint              channel_num   = data->channel_num;
	
 	g_return_val_if_fail (data,             NULL);
	g_return_val_if_fail (data->channel,    NULL);
	g_return_val_if_fail (data->frame,      NULL);
	g_return_val_if_fail (channel_num >= 0, NULL);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "inside the frame received handler (new task)");

	/* show received frame */
	if (vortex_log_is_enabled ()) {
		raw_frame = vortex_frame_get_raw_frame (frame);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "second level frame receive invocation for connection=%d channel=%d channel_received=%d (seqno: %d size: %d)\n%s",
		       vortex_connection_get_id (connection),
		       channel_num, channel->channel_num, 
		       vortex_frame_get_seqno (frame),
		       vortex_frame_get_content_size (frame),
		       raw_frame);
		g_free (raw_frame);
	}

	/* record actual connection state */
	is_connected = vortex_connection_is_ok (connection, FALSE);
	if (! is_connected) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "invoking frame receive on a non-connected session");
		goto free_resources;
	}

	/* invoke handler */
	if (channel->received) {
		channel->received (channel, channel->connection, frame, 
				   channel->received_user_data);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame received invocation for second level finished");
	}else {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "invoking frame received on channel %d with not handler defined",
		       vortex_channel_get_number (channel));
	}

	/* before frame receive handler we have to check if client
	 * have closed its connection */
	is_connected = vortex_connection_is_ok (connection, FALSE);
	if (! is_connected ) {
		/* the connection have been closed inside frame receive */
		goto free_resources;
	}

	/* The function __vortex_channel_0_frame_received_close_msg
	 * can be blocked awaiting to receive all replies
	 * expected. The following signal tries to wake up a possible
	 * thread blocked until last_reply_expected change. */
	if (vortex_connection_channel_exists (connection, channel_num))
		vortex_channel_signal_on_close_blocked (channel);
	else
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "channel=%d was closed during frame receive execution, unable to signal on close",
		       channel_num);

	/* log a message */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "invocation frame received handler for channel %d finished (second level: channel)", channel_num);

	/* free data and frame */
 free_resources:

	/* decrease connection reference counting */
	vortex_connection_unref (connection, "second level handler (frame received)");

	vortex_frame_free (frame);
	g_free            (data);

	return NULL;
}



/** 
 * @internal
 * @brief Invokes received handler on this channel. 
 * 
 * This handler is mainly used by vortex reader process at function
 * vortex_reader_process_socket.
 *
 * This enable vortex reader to notify that a frame have been received
 * on a particular channel. This function also check if there are
 * waiting thread blocked at vortex_channel_wait_reply. This means
 * there are threads waiting to a RPY frame type or ERR frame type to
 * be received.
 *
 * If no wait reply is found for a particular RPY or ERR, then the
 * default channel frame received handler is invoked.
 *
 * Frame comes to this function in a ordered manner. Vortex reader
 * thread filter all message received on registered connection to
 * detect channel synchronization failures. So, MSG received are
 * already ordered when they comes to this function and RPY, ERR
 * frames are also ordered.
 *
 * @param channel  The channel where frame received invocation will happen
 * @param frame    The frame actually producing the invocation
 * 
 * @return TRUE if frame was delivered or FALSE if not.
 */
gboolean      vortex_channel_invoke_received_handler (VortexConnection * connection, VortexChannel * channel, VortexFrame * frame)
{
	ReceivedInvokeData * data;
	WaitReplyData      * wait_reply;
	
	/* check data to avoid self-encrintation ;-) */
	g_return_val_if_fail (channel, FALSE);
	g_return_val_if_fail (frame,   FALSE);

	/* query if we have a waiting queue */
	switch (vortex_frame_get_type (frame)) {
	case VORTEX_FRAME_TYPE_RPY:
	case VORTEX_FRAME_TYPE_ERR:
		/* lock the channel during channel operation */
		g_mutex_lock (channel->receive_mutex);

		/* we have a reply message: rpy or err. this means we
		   can check if there are waiting thread */
		if ((wait_reply = vortex_queue_peek (channel->waiting_msgno)) == NULL) {
			g_mutex_unlock (channel->receive_mutex);
			break;
		}
		/* it seems we have waiting threads, so check if
		   received frame have a msgno for the next waiting
		   queue */
		if (vortex_frame_get_msgno (frame) != wait_reply->msg_no_reply) {
			g_mutex_unlock (channel->receive_mutex);
			break;
		}

		/* ok, we have found there are a thread waiting, so
		 * push data */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "found a frame that were waited by a thread: reply no='%d' over channel=%d",
		       vortex_frame_get_msgno (frame),
		       channel->channel_num);

		/* remove waiting reply data */
		vortex_queue_pop (channel->waiting_msgno);

		/* queue frame received */
		g_async_queue_push (wait_reply->queue, frame);

		/* unlock the channel */
		g_mutex_unlock (channel->receive_mutex);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame push and receive (channel=%d) mutex unlocked",
		       channel->channel_num);

		return TRUE;
	default:
		/* do nothing */
		break;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "no waiting queue for this message: %d", 
	       vortex_frame_get_msgno (frame));
	
	/* no waiting queue, invoke frame received handler so create
	 * the thread on this newly created thread, the frame will be
	 * deleted */
	if (channel->received == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "it seems you didn't define a second level received handler");
		return FALSE;
	}

	/* prepare data to be passed in to thread */
	data              = g_new0 (ReceivedInvokeData, 1);
	data->channel     = channel;
	data->frame       = frame;
	data->channel_num = channel->channel_num;

	/* increase connection reference counting to avoid reference
	 * loosing if connection is closed */
	vortex_connection_ref (connection, "second level invocation (frame received)");
	
	/* create the thread to invoke frame received handler */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "about to invoke the frame received under a newly created handler");
	vortex_thread_pool_new_task ((GThreadFunc)__vortex_channel_invoke_received_handler, data);
	return TRUE;

}

/** 
 * @brief Returns if the given channel have defined its close handler.
 * 
 * @param channel the channel to operate on
 * 
 * @return TRUE if close handler is defined or FALSE if not.
 */
gboolean        vortex_channel_is_defined_close_handler (VortexChannel * channel)
{
	g_return_val_if_fail (channel, FALSE);
	
	return (channel->close != NULL);
}

/** 
 * @internal
 * @brief Invoke the close handler for the given channel if were defined. 
 * 
 * @param channel the channel to operate on
 * 
 * @return what have returned close handler invocation
 */
gboolean            vortex_channel_invoke_close_handler     (VortexChannel  * channel)
{
	g_return_val_if_fail (channel,        FALSE);
	g_return_val_if_fail (channel->close, FALSE);

	return channel->close (channel->channel_num, channel->connection, 
			       channel->close_user_data);
}

/** 
 * @internal
 * @brief Internal Vortex Library function to validate received frames on channel 0.
 * 
 * @param channel0 the channel 0 itself
 * @param frame the frame received
 * 
 * @return TRUE if validated was ok or FALSE if it fails.
 */
gboolean __vortex_channel_0_frame_received_validate (VortexChannel * channel0, VortexFrame * frame)
{
	axlDtd    * channel_dtd;
	axlDoc    * doc;
	axlError  * error;
	gchar     * error_msg;

	/* get channel management DTD */
	if ((channel_dtd = vortex_dtds_get_channel_dtd ()) == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to load dtd file, cannot validate incoming message, returning error frame");
		error_msg = vortex_frame_get_error_message ("421", 
							    "service not available, unable to load dtd file, cannot validate incoming message",
							    NULL);
		vortex_channel_send_err (channel0, 
					 error_msg,
					 strlen (error_msg),
					 vortex_frame_get_msgno (frame));
		g_free (error_msg);
		return FALSE;
	}

	/* parser xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		/* report an error to the remote peer */
		error_msg = vortex_frame_get_error_message ("500", "general syntax error: xml parse error", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));
		g_free (error_msg);

		/* flag an error */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "received message on channel 0 with xml parse error, closing connection: %s",
		       axl_error_get (error));
		axl_error_free (error);

		/* drop the connection */
		__vortex_connection_set_not_connected (vortex_channel_get_connection (channel0),
						       "general syntax error: xml parse error");
		return FALSE;		 
	}
	
	/* validate document */
	if (! axl_dtd_validate (doc, channel_dtd, &error)) {
		/* Validation failed */
		error_msg = vortex_frame_get_error_message ("501", "syntax error in parameters: non-valid XML", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));
		g_free (error_msg);

		/* set the channel to be not connected */
		__vortex_connection_set_not_connected (vortex_channel_get_connection (channel0),
						       "501 syntax error in parameters: non-valid XML");

		/* report a log and free the error reported */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "received message on channel 0 with syntax error in parameters: non-valid XML: %s",
		       axl_error_get (error));
		axl_error_free (error);

		/* release the document read */
		axl_doc_free (doc);
		return FALSE;		 
	}
	
	/* release the document read */
	axl_doc_free (doc);
	return TRUE;
}

enum {START_MSG, CLOSE_MSG, ERROR_MSG, OK_MSG};


/** 
 * @internal
 * @brief Identifies the frame type received over channel 0.
 * 
 * @param channel0 the channel 0 itself
 * @param frame the frame received
 * 
 * @return the frame type received.
 */
gint __vortex_channel_0_frame_received_identify_type (VortexChannel * channel0, VortexFrame * frame)
{
	axlDoc     * doc;
	gchar      * error_msg;
	axlNode    * node;
	axlError   * error;
	gint         result = -1;

	/* parser xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		/* report a error to the remote peer */
		error_msg = vortex_frame_get_error_message ("500", "general syntax error: xml parse error", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg),
					 vortex_frame_get_msgno (frame));
		g_free (error_msg);

		/* free the error reported */
		axl_error_free (error);
		return FALSE;		 
	}

	
	/* Get the root element ( ? element) */
	node = axl_doc_get_root (doc);
	if (NODE_CMP_NAME (node, "start"))
		result = START_MSG;
	if (NODE_CMP_NAME (node, "close"))
		result = CLOSE_MSG;
	if (NODE_CMP_NAME (node, "error"))
		result = ERROR_MSG;
	if (NODE_CMP_NAME (node, "ok"))
		result = OK_MSG;

	/* return value found */
	axl_doc_free (doc);
	return result;
}


/** 
 * @internal
 * @brief Parses and validates the start message received.
 * 
 * @param frame The frame received
 * @param channel_num 
 * @param profile 
 * 
 * @return 
 */
gboolean __vortex_channel_0_frame_received_get_start_param (VortexFrame    * frame,
							    gint           * channel_num,
							    gchar         ** profile,
							    gchar         ** profile_content,
							    gchar         ** serverName,
							    VortexEncoding * encoding)
{
	axlDoc  * doc;
	axlNode * start;
	axlNode * profile_node;
	gchar   * channel;
	gchar   * enc;

	/* parser xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), NULL);
	if (!doc) {
		return FALSE;
	}

	/* Get the root element (start element): this is always the
	 * start element. */
	start           = axl_doc_get_root (doc);

	/* get channel number from the start element property: the
	 * function return a reference so it is not necessary to
	 * deallocate it */
	channel         = axl_node_get_attribute_value (start, "number");
	(* channel_num) = atoi (channel);

	/* get serverName value from the start element property */
	(* serverName)  = axl_node_get_attribute_value_copy (start, "serverName");

	/* Get the position of the first child the start element has
	 * and check if it is the profile element. This is done
	 * because libxml handle the as a node all characters found
	 * between the <start> and the next <profile> element.
	 * Because the message could be built using either
	 * <start><profile.. or <start>\x0D\x0Da<profile>, this must
	 * be checked. */
	profile_node    = axl_node_get_child_nth (start, 0);
	
	/* get profiles */
	(* profile)     = axl_node_get_attribute_value_copy (profile_node, "uri");

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "profile received %s", (* profile));

	/* get current encoding */
	enc        = axl_node_get_attribute_value (start, "encoding");
	if (enc == NULL) 
		(* encoding) = EncodingNone;
	else if (!g_strcasecmp (enc, "none"))
		(* encoding) = EncodingNone;
	else if (!g_strcasecmp (enc, "base64"))
		(* encoding) = EncodingBase64;
	else
		(* encoding) = EncodingUnknown;

	/* get profile content, the piggyback, setting as default a
	 * NULL value and updating it to the value found if defined */
	(* profile_content ) = axl_node_get_content_copy (profile_node, NULL);
	
	/* free the axl document */
	axl_doc_free (doc);
	return TRUE;

}

/** 
 * @internal
 * @brief Support function for \ref __vortex_channel_0_frame_received_start_msg. 
 * 
 * It allows to invoke current start channel handler providing
 * additional data needed for the process, that is, the serverName, the
 * profile content and its encoding.
 * 
 * @param profile         The profile notified for its creation
 * @param channel_num     The channel number requested to be created.
 * @param connection      The connection where the channel will be created.
 * @param serverName      The optional serverName start message attribute content
 * @param profile_content The optional profile content.
 * @param encoding        The optional profile content encoding.
 * 
 * @return TRUE if the channel is accepted to be created, or FALSE if channel couldn't be created.
 */
gboolean __vortex_channel_0_frame_received_start_msg_support_invoke (gchar            * profile, 
								     gint               channel_num, 
								     VortexConnection * connection,
								     gchar            * serverName,
								     gchar            * profile_content,
								     gchar           ** profile_content_reply,
								     VortexEncoding     encoding)
{
	gboolean     result;

	/* invoke the start channel handler */
	result = vortex_profiles_invoke_start (profile, channel_num, connection,
					       serverName, profile_content, 
					       profile_content_reply, encoding);

	/* deallocate unused memory (do not remove profile, it is used on
	 * the next steps) */
	vortex_support_free (2, serverName, g_free, profile_content, g_free); 

	return result;
	
}

/** 
 * @internal
 *
 * @brief Dispatch an start message received on the given channel
 * invoking start handlers defined.
 * 
 * @param channel0 The channel where the start message was received.
 * @param frame    The frame containing the frame received.
 */
void __vortex_channel_0_frame_received_start_msg (VortexChannel * channel0, VortexFrame * frame)
{
	gint               channel_num           = -1;
	gchar            * profile               = NULL;
	gchar            * profile_content       = NULL;
	gchar            * profile_content_reply = NULL;
	gchar            * serverName            = NULL;
	gchar            * error_msg             = NULL;
	gchar            * start_rpy             = NULL;
	VortexEncoding     encoding;       
	VortexChannel    * new_channel           = NULL;
	VortexConnection * connection            = channel0->connection;
	
	/* get and check channel num and profile */
	if (!__vortex_channel_0_frame_received_get_start_param (frame, &channel_num, 
								&profile, &profile_content, 
								&serverName, &encoding)) {
		error_msg = vortex_frame_get_error_message ("500", "unable to get channel start param", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));
		g_free (error_msg);
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "start message received: channel='%d' profile='%s' serverName='%s' profile_content='%s' encoding='%s'", 
	       channel_num, profile,
	       (serverName != NULL) ? serverName : "",
	       (profile_content != NULL) ? profile_content : "",
	       (encoding == EncodingNone) ? "none" : "base64");

	/* check if channel exists */
	if (vortex_connection_channel_exists (connection, channel_num)) {
		/* send an error reply */
		error_msg = vortex_frame_get_error_message ("554", "transaction failed: channel requested already exists", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));

		/* deallocate unused memory */
		vortex_support_free (4, error_msg,    g_free, 
				     profile,         g_free, 
				     serverName,      g_free, 
				     profile_content, g_free);
		return;
	}

	/* check if profile already exists */
	if (!vortex_profiles_is_registered (profile)) {

		/* send an error reply */
		error_msg = vortex_frame_get_error_message ("554", "transaction failed: channel profile requested not supported", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));

		/* deallocate unused memory */
		vortex_support_free (4, error_msg,    g_free, 
				     profile,         g_free, 
				     serverName,      g_free, 
				     profile_content, g_free);
		return;
	}

	/* ask if we have an start handler defined */
	if (!vortex_profiles_is_defined_start (profile)) {
		/* send an error reply */
		error_msg = vortex_frame_get_error_message ("421", "service not available: channel can not be created, no start handler defined", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));

		/* deallocate unused memory */
		vortex_support_free (4, error_msg,    g_free, 
				     profile,         g_free, 
				     serverName,      g_free, 
				     profile_content, g_free);
		return;
	}

	/* create the channel reference and set it to the connection
	 * to allow the start handler implementation to handle channel
	 * reference. If start handler fails, the channel is
	 * deallocated. Flag the channel to be not connected.*/
	new_channel = vortex_channel_empty_new (channel_num, profile, connection);
	vortex_connection_add_channel (connection, new_channel);
	new_channel->is_opened = FALSE;
	

	/* ask if channel can be created, (before this function it is
	 * not needed to call any deallocation code for profile,
	 * profile_content and serverName. This is already done by the
	 * following function. profile variable is still needed to be
	 * deallocated. */
	if (!__vortex_channel_0_frame_received_start_msg_support_invoke (profile,    channel_num, connection,
									 serverName, profile_content, 
									 &profile_content_reply, encoding)) {
		/* build and report error message */
		error_msg = vortex_frame_get_error_message ("421", 
							    "service not available: channel can not be created because application level have denied it, unable to start channel requested", 
							    NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));

		/* deallocate memory requested */
		vortex_support_free (2, error_msg, g_free, profile, g_free);

		/* deallocate the channel created */
		vortex_connection_remove_channel (connection, new_channel);

		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "channel %d with profile %s successfully created", channel_num, profile);
	/* flag the channel to be connected */
	new_channel->is_opened = TRUE;

	/* build start reply here using the optional content profile
	 * reply */
	start_rpy = vortex_frame_get_start_rpy_message (profile, profile_content_reply);

	/* send rpy and then create the channel */
	if (!vortex_channel_send_rpy (channel0, 
				      start_rpy, strlen (start_rpy),
				      vortex_frame_get_msgno (frame))) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "an error have happen while sending start channel reply, closing channel");
	}

	/* deallocate memory deallocated */
	vortex_support_free (3, start_rpy, g_free, profile, g_free, profile_content_reply, g_free);
	return;
}


/** 
 * @internal
 * 
 * @param frame 
 * @param channel_num 
 * @param code 
 * 
 * @return 
 */
gboolean __vortex_channel_0_frame_received_get_close_param (VortexFrame * frame,
							    gint * channel_num,
							    gchar ** code)
{
	axlDoc      * doc;
	axlNode     * close;
	gchar       * channel;

	/* parser xml document */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), NULL);
	if (!doc) 
		return FALSE;

	/* Get the root element (close element) */
	close           = axl_doc_get_root (doc);

	channel         = axl_node_get_attribute_value (close, "number");
	(* channel_num) = atoi (channel);

	(* code )       = axl_node_get_attribute_value_copy (close, "code");
	
	/* release the document */
	axl_doc_free (doc);

	return TRUE;
}


/** 
 * @internal
 * 
 * Looking for the code which actually process close messages received
 * on the channel 0. You have just hit it!
 *
 * This function process the close messages received for the channel
 * 0.
 * 
 * @param channel0 The channel where the message was received.
 * @param frame    The frame actually received.
 */
void __vortex_channel_0_frame_received_close_msg (VortexChannel * channel0, 
						  VortexFrame   * frame)
{
	
	gint               channel_num     = -1;
	gchar            * code            = NULL;
	gchar            * channel_num_str = NULL;
	VortexChannel    * channel;
	VortexConnection * connection      = vortex_channel_get_connection (channel0);
	gchar            * error_msg       = NULL;
	gboolean           close_value     = FALSE;
	
	/* get and check channel num and profile */
	if (!__vortex_channel_0_frame_received_get_close_param (frame, &channel_num, &code)) {
		/* build error reply */
		error_msg = vortex_frame_get_error_message ("500", "unable to get channel close param", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));
		g_free (error_msg);
		return;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "close message received: channel='%d' code='%s'",  channel_num, code);

	/* code actually not used */
	g_free (code);

	/* check if channel exists */
	if (!vortex_connection_channel_exists (connection, channel_num)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received a request to close a channel=%d which doesn't exist",
		       channel_num);
		error_msg = vortex_frame_get_error_message ("554", "transaction failed: channel requested to close doesn't exists", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));
		g_free (error_msg);
		return;
	}
	
	/* get channel to remove */
	channel = vortex_connection_get_channel (connection, channel_num);

	/* ask second level handler */
	if (vortex_channel_is_defined_close_handler (channel)) {
		close_value = vortex_channel_invoke_close_handler (channel);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "close second level handler response=%d", close_value);
		goto end;
	}

	/* ask first level handler */
	if (vortex_profiles_is_defined_close (channel->profile)) {
		close_value = vortex_profiles_invoke_close (channel->profile, 
							    channel->channel_num, 
							    channel->connection);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "close first level handler response=%d", close_value);
		goto end;
	}

	/* if channel0 have no defined its close handler we assume
	 * close = TRUE */
	if (channel->channel_num == 0) 
		close_value = TRUE;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "no close handler defined for any level, assuming=%d", close_value);
 end:
	/* now we have executed close handler for all levels and we
	 * know what to do we need to follow the next indications:
	 * 
	 * 1) If close_value is false, simple reply an error message
	 * to close avoiding to get blocked until all replies have 
	 * been received.
	 * 
	 * 2) If close_value is true, we have to wait to all replies
	 * to be received and tag the channel as being closed because
	 * the application have accepted its closing. This is equal to
	 * call vortex_channel_close. The channel gets into a state
	 * where no new MSG can be sent and only replies will be
	 
	 * check 1)
	 * ~~~~~~~~
	 * flag the channel to be usable again */
	channel->being_closed = close_value;
	if (close_value == FALSE) {
		error_msg = vortex_frame_get_error_message ("550", "still working", NULL);
		vortex_channel_send_err (channel0, error_msg, strlen (error_msg), vortex_frame_get_msgno (frame));
		g_free (error_msg);
		return;
	}

	/* check 2)
	 * ~~~~~~~~
	 * Ok, close we have to make sure have already receive all
	 * replies for all message, that is, to be equal channel MSG
	 * sent to RPY or ERR received */
	__vortex_channel_block_until_replies_are_received (channel);
	
	/* So, anyone have something to say about closing this
	 * channel?, no?, then close the channel.  We remove the
	 * channel before sending replies so we avoid channel closing
	 * collision. */
	if (channel->channel_num == 0) {
		/* install a waiter into the channel 0 for the message
		 * to be sent */
		channel_num_str = g_strdup_printf ("%d", vortex_frame_get_msgno (frame));
		vortex_channel_install_waiter  (channel0, channel_num_str);

		/* flag the connection to be at close processing. We
		 * set this flag so vortex reader won't complain about
		 * not properly connection close. */
		vortex_connection_set_data (connection, "being_closed", GINT_TO_POINTER (TRUE));

		/* send ok message */
		vortex_channel_send_rpyv       (channel0, vortex_frame_get_msgno (frame), 
						vortex_frame_get_ok_message ());
		
		/* block until the message is sent */
		vortex_channel_wait_until_sent (channel0, channel_num_str);
		g_free (channel_num_str);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "message <ok />, sent, closing the connection");

		/* flag the connection to be closed and non usable */ 
		__vortex_connection_set_not_connected (connection, "channel closed properly");

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "connection closed..");
		
	}else {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "channel %d with profile %s successfully closed", 
		       channel->channel_num, channel->profile);

		/* send the positive reply (ok message) */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "sending reply to close channel=%d", channel_num);
		vortex_channel_send_rpyv (channel0, vortex_frame_get_msgno (frame),
					  vortex_frame_get_ok_message ());
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reply sent to close=%d", channel_num);

		/* before closing the channel ensure there are not
		 * message being_sending */
		if (channel->being_sending) {
			g_mutex_lock   (channel->send_mutex);
			g_mutex_unlock (channel->send_mutex);
		}

		/* remove channel from connection */
		vortex_connection_remove_channel (connection, channel);
	}

	return;
}

/** 
 * @internal
 * 
 * @param channel0 
 * @param frame 
 */
void __vortex_channel_0_frame_received_ok_msg (VortexChannel * channel0, VortexFrame * frame)
{
	
	return;
}

/** 
 * @internal
 * @brief Frame received for channel 0 inside Vortex Library.
 *
 * Internal vortex library function. This function handle all incoming
 * frames that are received on channel 0 for a given connection.
 * 
 * Channel 0 is a really important channel for each session on BEEP
 * standard. Over this channel is negotiated new channel creation and
 * profiles used for these new channels.
 *
 * If you are looking for the handler which process: start and close
 * message defined on RFC 3080 this is the place.
 *
 * @param channel0    The channel 0 where a message was received.
 * @param connection  The connection where the channel 0 belongs to.
 * @param frame       The frame actually received
 * @param user_data   User defined data
 */
void vortex_channel_0_frame_received (VortexChannel    * channel0,
				      VortexConnection * connection,
				      VortexFrame      * frame,
				      gpointer           user_data)
{
	/* check we are handling the channel 0 */
	if (vortex_channel_get_number (channel0) != 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invoked channel 0 frame received over a different channel (%d)\n",
		       vortex_channel_get_number (channel0));
		__vortex_connection_set_not_connected (connection, 
						       "invoked channel 0 frame received over a different channel (%d)\n");
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame received over channel 0 (size: %d):\n%s\n",
	       vortex_frame_get_content_size (frame),
	       vortex_frame_get_payload (frame));

	/* validate message */
	if (!__vortex_channel_0_frame_received_validate (channel0, frame))
		return;
	
	/* dispatch the frame received over the channel 0 to the
	 * appropriate place */
	switch (__vortex_channel_0_frame_received_identify_type (channel0, frame)) {
	case START_MSG:
		/* received a start message */
		__vortex_channel_0_frame_received_start_msg (channel0, frame);
		break;
	case CLOSE_MSG:
		/* received a close message */
		__vortex_channel_0_frame_received_close_msg (channel0, frame);
		break;
	case OK_MSG:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received <ok /> reply");
		/* do nothing, Vortex Library uses wait reply method
		 * to get remote ok message.  */
		break;
	case ERROR_MSG:
		/* should not happen */
		break;
	}
	
	return;
}



/** 
 * @brief Return channel opened state. 
 * 
 * This is actually used by vortex_reader thread to detect if an
 * incoming message received can be delivered. Vortex reader use this
 * function to detect if a channel is being closed but waiting for
 * some received msg to be replied after sending close message.
 *
 * Before issuing a close message to close a channel
 * (\ref vortex_channel_close) the channel must fulfill some steps. So,
 * until reply to a close message is receive a channel is considered
 * to be valid to keep on sending replies to received messages.  
 * 
 * Note that this situation can take some time because remote peer, at
 * the moment of receiving a close message, could have tons of message
 * waiting to be sent.
 * 
 * Vortex library consumer can use this function to avoid sending
 * message over a channel that can be on an ongoing closing process.
 * 
 * @param channel the channel to operate on.
 * 
 * @return TRUE if a channel is opened or FALSE if not.
 */
gboolean        vortex_channel_is_opened       (VortexChannel * channel)
{
	return channel->is_opened;
}

/** 
 * @brief Return if a channel is on the process of being closed. 
 * 
 * If a channel is being closed means no more message MSG can be sent
 * over it. But, actually more message can still be received (and
 * these message must be replied). On this context a channel being
 * closed can be used to send rpy and err message.
 *
 * This function can be used by vortex library consumer to know if a
 * channel can be used to send more replies to received messages.
 * 
 * @param channel the channel to operate on
 * 
 * @return if channel is being closed.
 */
gboolean        vortex_channel_is_being_closed (VortexChannel * channel)
{
	return channel->being_closed;
}

/** 
 * @brief free channel allocated resources.
 *
 * Free channel resources. Vortex Library consumers must not call this
 * function directly.
 *
 * In order to close and free the channel properly
 * \ref vortex_channel_close must be used.
 * 
 * @param channel the channel to free.
 */
void vortex_channel_free (VortexChannel * channel)
{
	VortexWriterData * writer_data;
	WaitReplyData    * wait_reply;
	VortexFrame      * frame;
	GList            * list_aux;

	g_return_if_fail (channel);
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing resources for channel %d", channel->channel_num);

	/* check if this channel belongs to a pool to remove it */
	if (channel->pool) {
		/* it seems this channel belongs to a pool remove it
		 * from the pool */
		vortex_channel_pool_deattach (channel->pool, channel);
	}


	/* ensure no wait reply is running remaining to unlock the
	 * channel receive_mutex. Lock on it and unlock it
	 * immediately. */
	g_mutex_lock        (channel->receive_mutex);
	g_mutex_unlock      (channel->receive_mutex);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing receive_mutex");
	g_mutex_free        (channel->receive_mutex);
	channel->receive_mutex = NULL;

	g_mutex_lock        (channel->send_mutex);
	g_mutex_unlock      (channel->send_mutex);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing send_mutex channel=%d", channel->channel_num);
	g_mutex_free        (channel->send_mutex);
	channel->send_mutex = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing close_mutex");
	g_mutex_free        (channel->close_mutex);
	channel->close_mutex = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing pending_mutex");
	g_mutex_free        (channel->pending_mutex);
	channel->pending_mutex = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing reply_cond");
	g_cond_free         (channel->reply_cond);
	channel->reply_cond = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing close_cond");
	g_cond_free         (channel->close_cond);
	channel->close_cond = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing pending_cond");
	g_cond_free         (channel->pending_cond);
	channel->pending_cond = NULL;

	g_free              (channel->profile);
	channel->profile = NULL;

	/* destroy pending piggyback */
	if (vortex_channel_have_piggyback (channel))
		vortex_frame_free (vortex_channel_get_piggyback (channel));

	/* destroy data hash */
	vortex_hash_destroy (channel->data);
	channel->data = NULL;

	/* freeing queue messages */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing messages queue");
	while (!vortex_queue_is_empty (channel->queue_msgs)) {
		writer_data = vortex_queue_pop (channel->queue_msgs);
		vortex_writer_data_free (writer_data);
	}
	vortex_queue_free   (channel->queue_msgs);
	channel->queue_msgs = NULL;

	/* freeing SEQ frames queue */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing SEQ frames queue");
	while (!vortex_queue_is_empty (channel->queue_seqs)) {
		writer_data = vortex_queue_pop (channel->queue_seqs);
		vortex_writer_data_free (writer_data);
	}
	vortex_queue_free   (channel->queue_seqs);
	channel->queue_seqs = NULL;

	/* freeing wait replies */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing waiting reply queue");
	while (!vortex_queue_is_empty (channel->waiting_msgno)) {
		wait_reply = vortex_queue_pop (channel->waiting_msgno);
		vortex_channel_free_wait_reply (wait_reply);
	}
	vortex_queue_free   (channel->waiting_msgno);
	channel->waiting_msgno = NULL;

	/* free pending messages to be completed */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing previous frames");
	list_aux = g_list_first (channel->previous_frames);
	for (;list_aux; list_aux = g_list_next (list_aux)) {
		frame = list_aux->data;
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "found pending frame...");
		vortex_frame_free (frame);
	}
	g_list_free (channel->previous_frames);
	channel->previous_frames = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing channel node");
	/* free channel */
	g_free (channel);

	return;
}


/** 
 * @brief Frees allocated memory by wait reply data
 *
 * See \ref vortex_manual_wait_reply "this section" to know more about this function
 * and how it is used inside the Wait Reply process.
 * 
 * @param wait_reply Wait reply to free.
 */
void     vortex_channel_free_wait_reply (WaitReplyData * wait_reply)
{
	g_return_if_fail (wait_reply);

	g_async_queue_unref (wait_reply->queue);
	g_free (wait_reply);

	return;
}

/** 
 * @brief Returns the actual state for a given channel about pending
 * replies. 
 * 
 * Due to RFC3080/RFC3081 design any message sent will be replied in
 * the order the message was issued. This means that if you use the
 * same channel to send two messages the reply for the second one will
 * not be received until the previous one. This is true even the
 * second message have a task-impact lower than previous on the remote
 * peer.
 *
 * Of course this is a bottleneck problem because you can increase
 * greatly the message/reply processing relation using channels that
 * are ready in the sense they are not waiting for message replies so
 * your message will not be blocked due to previous ones.
 *
 * This function returns if a channel is ready in the previous sense:
 * <i>"No previous message is waiting to be replied over this channel so
 * my message will only wait as long as the remote peer process my
 * message and reply it"</i>.
 *
 * Keep in mind a channel is always ready to accept new message to be
 * sent messages. In fact, eventually any message sent over a channel
 * will have its reply but this "ready sense" is a matter of
 * performance not availability.
 * 
 * @param channel the channel to check for its readyness state.
 * 
 * @return TRUE if ready or FALSE if not.
 */
gboolean           vortex_channel_is_ready                       (VortexChannel * channel)
{
	g_return_val_if_fail (channel, FALSE);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "checking channel to be ready (%d == %d) && reply_processed=%d",
	       channel->last_message_sent, 
	       (channel->last_reply_expected - 1), 
	       channel->reply_processed);
	
	return ((channel->last_message_sent == (channel->last_reply_expected - 1)) && channel->reply_processed);
}

/** 
 * @brief Support function that could be used as a frame received
 * function which queue all frames received inside a GAsyncQueue.
 *
 * See \ref vortex_channel_get_reply for more explanation about using
 * this function.
 * 
 *
 *
 * @param channel    The channel where the frame was received.
 * @param connection The connection where the channel lives.
 * @param frame      The frame received.
 * @param user_data  A queue where the frame received will be pushed.
 */
void               vortex_channel_queue_reply                    (VortexChannel    * channel,
								  VortexConnection * connection,
								  VortexFrame      * frame,
								  gpointer user_data)
{
	GAsyncQueue * queue      = user_data;
	VortexFrame * frame_copy = NULL;
	if (queue == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "NULL reference where a queue was expected, unable to queue frame received");
		return;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received frame, queueing frame over channel=%d at connection=%d",
	       channel->channel_num, vortex_connection_get_id (connection));

	/* because this frame handler is run under the semantic of the
	 * second level handler, once finished this function, the
	 * frame to push will be deallocate. We have to perform a
	 * copy. */
	frame_copy = vortex_frame_copy (frame);
	
	g_async_queue_push (queue, frame_copy);
	return;
}

/** 
 * @brief Allows to get the next frame received on the given channel
 * due to channel start reply piggybacking or due to a frame received
 * while using the function \ref vortex_channel_queue_reply as a \ref VortexOnFrameReceived "frame received handler".
 *
 * This function is used with \ref vortex_channel_queue_reply to
 * receive the message replies (in fact, all messages received). It
 * also support returning the data received due to channel start
 * piggybacking through \ref vortex_channel_get_piggyback.
 *
 * If you would like to know more about the piggyback concept, check
 * the following section to \ref vortex_manual_piggyback_support "know more about how piggyback could improve your protocol startup".
 *
 * The idea is to use \ref vortex_channel_queue_reply as a frame
 * receive handler which queues all frames received on a queue. Then,
 * every time \ref vortex_channel_get_reply is called, a frame from
 * the queue is returned.
 * 
 * This allows to support, using the same code, to receive a message
 * reply either as a piggyback or as a normal reply (and any other
 * frame received).
 *
 * Let's suppose you are implementing a profile which could send an
 * initial start message with the piggyback, so it is saved one round
 * trip: <i>"with one message, we send the start channel request, including
 * the first message exchanged."</i>.
 *  
 * At the remote peer, the channel creation request, with a piggyback
 * content, could be replied, using two, standard allowed, methods:
 *  
 * - A reply to the &lt;start> message accepting the channel to be
 * created and then reply to the implicit content received, in a
 * separated message.
 * 
 * - Or reply to the &lt;start> message accepting the channel to be
 * created, including inside that reply, the reply to the first
 * piggyback message received.
 *
 * This leads to a source code problem at the client side that perform
 * a blocking wait: <i>"It is needed a way to get, not only the
 * piggyback, but also the same reply not using the piggyback
 * method."</i>
 *
 * Because piggyback is closely related to the channel creation, and
 * to avoid writing two different piece of code to manage both cases,
 * you have two methods:
 * 
 * - 1. Set an \ref VortexOnChannelCreated "on created handler" at the
 * \ref vortex_channel_new or \ref vortex_channel_new_full and you'll
 * receive, as the first frame, the piggyback. For this case, the
 * first frame received over the channel will either the piggyback or,
 * if not defined, the first frame received.
 * 
 * - 2. If you need to make synchronous, the code you are doing, set
 * the \ref vortex_channel_queue_reply as the frame received handler
 * and then, once the channel is created call to \ref
 * vortex_channel_get_reply. This method works the same as <b>1.</b>
 * 
 * <i><b>NOTE: </b> The frame returned by this function have to be
 * deallocated using \ref vortex_frame_free when no longer
 * needed. This is an special case because all frames delivered with
 * first and second invocation handlers are automatically managed,
 * that is, deallocated when the frame handler scope has gone.</i>
 *
 * Because examples are more clear, here is one:
 *
 * \code
 * // create a channel making all frames received, including the
 * // piggyback to be received in a blocking way:
 * VortexChannel * channel; 
 * GAsyncQueue   * queue;
 * VortexFrame   * frame;
 *
 * // create the queue to be used to perform the async wait
 * queue   = g_async_queue_new ();
 *
 * // create the channel
 * channel = vortex_channel_new (// the connection where the channel
 *                               // will be created
 *                               connection, 
 *                               // let vortex library to manage
 *                               // the next channel number available
 *                               0, 
 *                               // use default channel close handling
 *                               NULL, NULL,
 *                               // set the frame receive handling
 *                               vortex_channel_queue_reply, queue,
 *                               // no channel creation notification
 *                               NULL, NULL);
 * 
 * // now wait and process all replies and messages received
 * while (TRUE) {
 *     // get the next message, blocking at this call.
 *     frame = vortex_channel_get_reply (channel, queue);
 *
 *     if (frame == NULL) {
 *          // timeout found, do some error reporting
 *          // and default action on timeout received.
 *          //
 *          // for our example, the default action is:
 *          //     keep on reading!.
 *          continue;
 *     }
 *     g_print ("Frame received, content: %s\n",
 *              vortex_frame_get_payload (frame);
 *    
 *     // deallocate the frame received
 *     vortex_frame_free (frame);
 * }
 * \endcode
 *
 * @param channel The channel where is expected to receive a frame
 * using the frame received handler or the initial piggyback.  @param
 * queue The queue where the frame is expected to be received. The
 * queue received, no matter the result will be deallocated.
 * 
 * @return The frame received or NULL if the timeout is expired. 
 */
VortexFrame      * vortex_channel_get_reply                      (VortexChannel    * channel, 
								  GAsyncQueue      * queue)
{
	VortexFrame * frame;

	/* timeout object */
	GTimeVal      timeout;

	
	/* check for piggyback reply */
	frame = vortex_channel_get_piggyback (channel);
	if (frame) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting piggyback from remote peer at get_reply");
		return frame;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting next frame from remote peer at get_reply");

	/* wait for the data to be returned, first prepare timeout
	 * value */
	g_get_current_time (&timeout);
	g_time_val_add (&timeout, vortex_connection_get_timeout ());

	/* get reply from queue */
	frame = g_async_queue_timed_pop (queue, &timeout);
	g_async_queue_unref (queue);
	if (frame == NULL) {
		/* drop a log in case of a timeout is reached. */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "timeout was expired while waiting at get-reply");
	}
	
	/* return whatever is returned */
	return frame;
}

/** 
 * @brief Allows to get initial piggyback received on the channel
 * start reply.
 *
 * Once the functions returns the piggyback, next calls will return
 * NULL.
 *
 * @param channel The channel where the piggyback was received.
 * 
 * @return Piggyback received or NULL if not defined.
 */
VortexFrame      * vortex_channel_get_piggyback                  (VortexChannel    * channel)
{
	
	VortexFrame * frame;

	g_return_val_if_fail (channel, NULL);

	frame = vortex_channel_get_data (channel, "piggyback");

	vortex_channel_set_data (channel, "piggyback", NULL);

	return frame;
}

/** 
 * @brief Allows to check if the given channel have piggyback waiting to be processed.
 * 
 * If the function returns TRUE then \ref vortex_channel_get_piggyback
 * should be used to get current piggyback received.
 * 
 * @param channel The channel to query.
 * 
 * @return TRUE if the channel has piggyback, otherwise FALSE is returned.
 */
gboolean           vortex_channel_have_piggyback                 (VortexChannel   * channel)
{
	g_return_val_if_fail (channel, FALSE);

	return (vortex_channel_get_data (channel, "piggyback") != NULL);
}

/** 
 * @internal
 * @brief Allows to set current piggyback for the given channel. 
 *
 * This function should not be used by Vortex Library API
 * consumers. It is used by Vortex Library to support general
 * piggybacking.
 * 
 * @param channel The channel where the piggyback will be set.
 * @param frame   The frame to be set as piggyback.
 */
void               vortex_channel_set_piggyback                  (VortexChannel    * channel,
								  gchar            * profile_content)
{
	VortexFrame * piggyback_frame;

	g_return_if_fail (channel         != NULL);
	g_return_if_fail (profile_content != NULL);



	/* build the frame to convert the piggyback received, that is
	 * the profile element content, into a frame as it were the
	 * first frame received on this channel. */
	piggyback_frame = vortex_frame_create (VORTEX_FRAME_TYPE_RPY,
					       0,     /* channel number */
					       0,     /* the msgno */
					       FALSE, /* more flag */
					       0,     /* seqno number */
					       strlen (profile_content),
					       0,
					       profile_content);

	/* store the frame */
	vortex_channel_set_data (channel, "piggyback", piggyback_frame);
	
	return;
}

/** 
 * @internal
 * 
 * This function is for internal Vortex library purposes. Actually
 * this function is used by the vortex_channel_pool to check if a
 * message is ready from the point of view that all replies to message
 * sent have been received discarding if that replies have been
 * actually processed.
 *
 * This function should not be useful for vortex library consumer but
 * can be used to check if a channel have received (but may not
 * processed) all replies.
 * 
 * @param channel the channel to operate on
 * 
 * @return 
 */
gboolean           vortex_channel_is_up_to_date (VortexChannel * channel)
{
	g_return_val_if_fail (channel, FALSE);
	
	return (channel->last_message_sent == (channel->last_reply_expected - 1));
}

/** 
 * @brief Creates a new wait reply to be used to wait for a specific
 * reply.
 *
 * @return a new Wait Reply object. 
 */
WaitReplyData * vortex_channel_create_wait_reply ()
{
	WaitReplyData * data;	
	data         = g_new0 (WaitReplyData, 1);
	data->queue  = g_async_queue_new ();

	return data;
}

/** 
 * @brief Allows caller to wait for a particular reply to be received. 
 * 
 * Check \ref vortex_manual_wait_reply "this section" to know more about how using
 * this function inside the Wait Reply Method.
 *
 * Keep in mind that this function could return a NULL frame while
 * performing a wait. This timeout is configured through the following
 * functions:
 *
 *   - \ref vortex_connection_timeout
 *   - \ref vortex_connection_get_timeout
 *
 * 
 * @param channel The channel where \ref vortex_manual_wait_reply "Wait Reply" is done. This value is
 * not optional and should be a valid channel. This function will
 * check if the reference provided is not NULL.
 *
 * @param msg_no The message number we are waiting to be replied. This
 * value is not optional and should be greater than 0. This funtion
 * will check if the value meets this requirement.
 *
 * @param wait_reply The wait reply object used. A valid wait reply
 * object created by \ref vortex_channel_create_wait_reply. This
 * function will check if the provided value is not NULL.
 * 
 * @return The frame reply or NULL if a timeout occurs, having all
 * parameters well specified, or there was a parameter error detected.
 */
VortexFrame   * vortex_channel_wait_reply              (VortexChannel * channel, gint msg_no, 
							WaitReplyData * wait_reply)
{

	VortexFrame * frame;
	/* timeout object */
	GTimeVal      timeout;

	g_return_val_if_fail (channel,     NULL);
	g_return_val_if_fail (msg_no >= 0, NULL);
	g_return_val_if_fail (wait_reply,  NULL);


	/* wait for the data to be returned, first prepare timeout
	 * value */
	g_get_current_time (&timeout);
	g_time_val_add (&timeout, vortex_connection_get_timeout ());
	
	/* wait for the message to be replied */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting reply at wait reply from the queue");
	frame = g_async_queue_timed_pop (wait_reply->queue, &timeout);
	if (frame == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "received a timeout while waiting performing a 'wait reply'");
	}else {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received reply, freeing wait reply object");

		/* free data no longer needed */
		vortex_channel_free_wait_reply (wait_reply);

		/* because we have accept to deliver the frame to a waiting
		 * thread, we understand this frame have being delivered */
		channel->reply_processed = TRUE;
	}

	/* return whatever we have get */
	return frame;
}

/** 
 * @internal
 *
 * This function actually is used by the vortex reader process. It
 * allows vortex reader process to create a critical section while
 * update channel status.
 *
 * <b>NOTE: </b> this function is deprecated and should not be used anymore.
 * 
 * @param channel the channel to operate.
 */
void            vortex_channel_lock_to_update_received   (VortexChannel * channel)
{
	g_return_if_fail (channel);
	return;
}

/** 
 * @internal
 * 
 * This function actually is used by the vortex reader process. It
 * allows vortex reader process to create a critical section while
 * update channel status. 
 *
 * <b>NOTE: </b> this function is deprecated and should not be used
 * anymore.
 * 
 * @param channel the channel to operate on.
 */
void            vortex_channel_unlock_to_update_received (VortexChannel * channel)
{
	g_return_if_fail (channel);
	return;
}

/** 
 * @internal
 *
 * @brief Allows external vortex subsystems to lock the receive mutex.
 * 
 * @param channel The channel to operate.
 */
void               vortex_channel_lock_to_receive                (VortexChannel * channel)
{
	g_return_if_fail (channel);

	g_mutex_lock (channel->receive_mutex);

	return;
}

/** 
 * @internal
 *
 * @brief Allows external vortex subsystems to unlock the receive
 * mutex.
 * 
 * @param channel The channel to operate.
 */
void               vortex_channel_unlock_to_receive              (VortexChannel * channel)
{
	g_return_if_fail (channel);
	
	g_mutex_unlock (channel->receive_mutex);

	return;
}

/** 
 * @internal
 *
 * The function __vortex_channel_0_frame_received_close_msg can be
 * blocked awaiting to receive all replies expected. The following
 * signal tries to wake up a possible thread blocked until
 * last_reply_expected change.
 * 
 * @param channel 
 */
void           	   vortex_channel_signal_on_close_blocked        (VortexChannel    * channel)
{

	g_return_if_fail (channel);
	g_mutex_lock   (channel->close_mutex);

	/* flag channel to be up to date on reply process */
	channel->reply_processed = TRUE;

	/* check if channel is being closed */
	if (!channel->being_closed) {
		g_mutex_unlock (channel->close_mutex);
		return;
	}
	
	/* do not signal if not close waiting */
	if (!channel->waiting_replies) {
		g_mutex_unlock (channel->close_mutex);
		return;
	}

	/* do a close waiting signal */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "signaling blocked thread on close msg process");

	g_cond_signal  (channel->close_cond);
	g_mutex_unlock (channel->close_mutex);

	return;
}

/** 
 * @internal
 *
 * Function used to unblock threads on close process that are waiting
 * all replies to message received are sent. Until done this, the
 * channel can not wait for a ok message.
 * 
 * @param channel 
 */
void   vortex_channel_signal_reply_sent_on_close_blocked (VortexChannel * channel)
{
	g_return_if_fail (channel);

	/* check if channel is being closed */
	if (!channel->being_closed) {
		return;
	}
	

	g_mutex_lock (channel->pending_mutex);
	g_cond_signal (channel->pending_cond);
	g_mutex_unlock (channel->pending_mutex);

	return;
}

/** 
 * @brief Allows to flag that the channel have processed its last
 * reply.
 *
 * Under normal situations, this function is used by the vortex reader
 * to notify the reply processed status for a provided channel. This
 * allows to control race conditions and to keep \ref
 * vortex_channel_is_ready working.
 *
 * The reply processed flag is only activated once the frame was
 * delivered and the frame received handler for the provided channel
 * has finished. This ensures that calls to \ref
 * vortex_channel_is_ready returns proper status for the provided
 * channel and all internal functions for a channel (especially the
 * close operation) keeps working and synchronized.
 *
 * Howerver, it could happen, especially for request-response pattern
 * implemented on top of Vortex Library, like XML-RPC, that a race
 * condition could happen while calling to invoke several services
 * using the same channel. Because the thread that executes the
 * invocation code could be faster than the internal thread, inside
 * the vortex library, that flags the channel to have its reply
 * processed, makes the next invocation to fail because it is detected
 * that other ongoing operation is being taking place.
 *
 * To avoid such situations RPC implementations on top of Vortex
 * Library could use this function to force that a channel have
 * processed its reply. An example could be found inside the
 * implementation of \ref vortex_xml_rpc_invoke.
 *
 * @param channel The channel what is about to be flagged as ready (or
 * not, according to the flag value).
 *
 * @param flag The status to flag.
 */
void vortex_channel_flag_reply_processed (VortexChannel * channel, gboolean flag)
{
	g_return_if_fail (channel);

	g_mutex_lock   (channel->close_mutex);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "setting pending reply processed=%d for channel=%d",
	       flag, channel->channel_num);
	channel->reply_processed = flag;
	g_mutex_unlock (channel->close_mutex);

	return;
}


/** 
 * @internal
 *
 * This function allows to close channel processing while this message
 * refers to channel0 to properly synchronize the ok message response
 * and connection closing. 
 *
 * Used in conjunction with \ref vortex_channel_wait_until_sent enable to
 * get blocked until a reply have been sent.
 *
 * This function is for vortex internal purposes so it should not be
 * useful for vortex consumers.
 * 
 * @param channel 
 * @param rpy 
 */
void               vortex_channel_install_waiter                 (VortexChannel * channel,
								  gchar * rpy)
{
	GAsyncQueue    * queue;

	g_return_if_fail (channel);
	g_return_if_fail (rpy && (* rpy));

	queue = g_async_queue_new ();
	vortex_channel_set_data (channel, rpy, queue);

	return;
}

/** 
 * @internal
 *
 * Block the caller until the message have been sent. This function is
 * for internal vortex purposes so it should not be useful to vortex
 * consumers.
 * 
 * @param channel 
 * @param rpy 
 */
void               vortex_channel_wait_until_sent                (VortexChannel * channel,
								  gchar *rpy)
{
	GAsyncQueue * queue;

	g_return_if_fail (channel);
	g_return_if_fail (rpy && (* rpy));

	/* get the queue */
	queue = vortex_channel_get_data (channel, rpy);

	g_return_if_fail (queue);

	/* wait until message is sent */
	g_async_queue_pop   (queue);
	g_async_queue_unref (queue);
	
	return;
}

/** 
 * @internal
 * 
 * @param channel 
 * @param rpy 
 */
void             vortex_channel_signal_rpy_sent                 (VortexChannel * channel,
								 gchar * rpy)
{
	GAsyncQueue * queue;
	g_return_if_fail (channel);
	g_return_if_fail (rpy && (* rpy));

	/* get the queue */
	queue = vortex_channel_get_data (channel, rpy);

	if (queue == NULL)
		return;

	/* push some data, to wakeup up the waiting thread */
	g_async_queue_push (queue, GINT_TO_POINTER (TRUE));

	return;
}

/** 
 * @internal
 *
 * This function is for internal vortex purposes. This function
 * actually allows vortex channel pool module to notify this channel
 * that belongs to a pool so in case of a channel close (received or
 * invoked) this channel will be able to notify channel pool to remove
 * the reference to the channel.
 * 
 * @param channel the channel to notify it belongs to a pool.
 * @param pool the pool the channel belongs to.
 */
void               vortex_channel_set_pool                       (VortexChannel * channel,
								  VortexChannelPool * pool)
{
	g_return_if_fail (channel);
	g_return_if_fail (pool);

	channel->pool = pool;

	return;
}


/* @} */
