/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>
#define LOG_DOMAIN "vortex-connection"
#define VORTEX_CONNECTION_BUFFER_SIZE 32768

/** 
 * @internal
 * @brief A connection identifier (for internal vortex use).
 */
glong                connection_id                               = 1;
static GStaticMutex  connection_id_mutex                         = G_STATIC_MUTEX_INIT;
gboolean             __vortex_connection_enable_sanity_check     = TRUE;

/** 
 * @internal
 * @brief Auto TLS profile negotiation internal support variables.
 */
gboolean             __vortex_connection_auto_tls                = FALSE;
gboolean 	     __vortex_connection_auto_tls_allow_failures = FALSE;
gchar *              __vortex_connection_auto_tls_server_name    = NULL;


/** 
 * @internal
 * @brief Internal VortexConnection representation.
 */
struct _VortexConnection {
	/** 
	 * @brief Vortex Connection unique identifier.
	 */
	gint         id;

	/** 
	 * @brief Host name this connection is actually connected to.
	 */
	gchar      * host;

	/** 
	 * @brief Port this connection is actually connected to.
	 */
	gchar      * port;

	/** 
	 * @brief Allows to hold and report current connection status.
	 */
	gboolean     is_connected;

	/**
	 * @brief App level message to report current connection
	 * status using a textual message.
	 */
	gchar      * message;

	/** 
	 * @brief Underlying transport descriptor.
	 * This is actually the network descriptor used to send and
	 * received data.
	 */
	VORTEX_SOCKET  session;

	/** 
	 * @brief Allows to configure if the given session (actually
	 * the socket) associated to the given connection should be
	 * closed once __vortex_connection_set_not_connected function.
	 */
	gboolean       close_session;

	/**
	 * Profiles supported by the remote peer.
	 */
	GList      * remote_supported_profiles;
	/** 
	 * Channels already created inside the given VortexConnection.
	 */
	VortexHash * channels;
	/** 
	 * Hash to hold miscellaneous data, sometimes used by the
	 * Vortex Library itself, but also exposed to be used by the
	 * Vortex Library programmers through the functions:
	 *   - \ref vortex_connection_get_data 
	 *   - \ref vortex_connection_set_data
	 */
	VortexHash * data;
	/** 
	 * @brief Keeps track for the last channel created on this
	 * session. The connection role modifies the way channel
	 * number are allocated. In the case the role is initiator,
	 * client role, odd channel numbers are generated. In the case
	 * the role is listener, server role, even channel numbers are
	 * generated.
	 */
	gint         last_channel;

	/** 
	 * @brief Remote features supported/requested.
	 *
	 * This variable holds possible features content set at the
	 * greetings element received from the remote peer.
	 */
	gchar      * features;

	/** 
	 * @brief Remote localize supported/requested.
	 * 
	 * This variable holds possible localization values the remote
	 * peer is requesting.
	 */
	gchar      * localize;

	/**
	 * @brief The channel_mutex
	 *
	 * This mutex is used at vortex_connection_get_next_channel to
	 * avoid race condition of various threads calling in a
	 * reentrant way on this function getting the same value.
	 * 
	 * Although this function use vortex_hash_table, a
	 * thread-safe hash table, this is not enough to ensure only
	 * one thread can execute inside this function.
	 * 
	 * The vortex_connection_add_channel adds a channel into a
	 * existing connection. Inside the implementation, this
	 * function first lookup if channel already exists which means
	 * a program error and the insert the new channel if no
	 * channel was found.
	 *
	 * This mutex is also used on that function to avoid reentrant
	 * conditions on the same function and between function which
	 * access to channel hash table.
	 */
	GMutex     * channel_mutex;

	/**
	 * @brief The ref_count.
	 *
	 * This enable a vortex connection to keep track of the
	 * reference counting.  The reference counting is controlled
	 * thought the vortex_connection_ref and
	 * vortex_connection_unref.
	 * 
	 */
	gint ref_count;
	
	/**
	 * @brief The ref_mutex
	 * This mutex is used at vortex_connection_ref and
	 * vortex_connection_unref to avoid race conditions especially
	 * at vortex_connection_unref. Because several threads can
	 * execute at the same time this unref function this mutex
	 * ensure only one thread will execute the vortex_connection_free.
	 * 
	 */
	GMutex  * ref_mutex;

	/**
	 * @brief The op_mutex
	 * This mutex allows to avoid race condition on operating with
	 * the same connection from different threads.
	 */
	GMutex * op_mutex;
	
	/** 
	 * @brief The handlers mutex
	 *
	 * This mutex is usted to secure code sections destinated to
	 * handle handlers stored associated to the connection.
	 */
	GMutex * handlers_mutex;

	/**
	 * @brief The channel_pool_mutex
	 * This mutex is used by the vortex_channel_pool while making
	 * operations with channel pool over this connection. This is
	 * necessary to make mutually exclusive the function from that
	 * module while using the same connection.
	 */
	GMutex * channel_pool_mutex;

	/**
	 * @brief The next_channel_pool_id
	 * This value is used by the channel pool module to hold the
	 * channel pool identifiers that will be used.
	 */
	gint next_channel_pool_id;

	/**
	 * @brief The channel_pools
	 * This hash actually holds every channel pool created over
	 * this connection. Every channel pool have an id that unique
	 * identifies the channel pool from each over inside this
	 * connection and it is used as key.
	 */
	VortexHash * channel_pools;

	/** 
	 * @brief The peer role
	 * 
	 * This is used to know current connection role.	
	 */
	VortexPeerRole role;

	/** 
	 * @brief Writer function used by the Vortex Library to actually send data.
	 */
	VortexSendHandler    send;

	/** 
	 * @brief Writer function used by the Vortex Library to actually received data
	 */
	VortexReceiveHandler receive;

	/** 
	 * @brief On close handler
	 */
	GList * on_close;

	/** 
	 * @brief On close handler, extended version.
	 */
	GList * on_close_full;

	/** 
	 * @brief Data associated to the connection, to be passed to
	 * the on close full.
	 */
	GList * on_close_full_data;
};



typedef struct __VortexConnectionData {
	gpointer       key;
	GDestroyNotify key_destroy_func;
	gpointer       data;
	GDestroyNotify data_destroy_func;
}VortexConnectionData;

/** 
 * @internal
 *
 * @brief Support function to create nodes of data to be stored into
 * the vortex connection data hash.
 * 
 * @param data The data to be stored
 * @param destroy_func The destroy function to be used.
 * 
 * @return A newly allocated vortex hash data
 */
gpointer  __vortex_connection_data_create (gpointer key,  GDestroyNotify key_destroy_func,
					   gpointer data, GDestroyNotify data_destroy_func)
{
	VortexConnectionData * result = g_new0 (VortexConnectionData, 1);
	result->key                   = key;
	result->key_destroy_func      = key_destroy_func;
	result->data                  = data;
	result->data_destroy_func     = data_destroy_func;

	return result;
}

/** 
 * @internal
 *
 * @brief Implements value deallocation for values stored into the
 * vortex connection data hash.
 * 
 * @param data The vortex hash data node to be unrefered.
 */
void __vortex_connection_data_free_value (gpointer data) 
{
	VortexConnectionData * result  = data;

	/* find and exec data destroy function */
	if (result->data_destroy_func != NULL) 
		result->data_destroy_func (result->data);

	/* find and exec key destroy function */
	if (result->key_destroy_func  != NULL)
		result->key_destroy_func (result->key);
	g_free (result);
	return;
}

/** 
 * @internal
 *
 * @brief Support function used by \ref vortex_connection_get_data to
 * support key/value destroy function per item defined.
 * 
 * @param data The VortexConnectionData node stored
 * 
 * @return The value inside the node.
 */
gpointer __vortex_connection_data_get (gpointer data) 
{
	VortexConnectionData * result  = data;
	if (result == NULL) 
		return NULL;
	return result->data;
}


/** 
 * @internal
 * @brief Perform the connection socket sanity check. 
 * 
 * This prevents from having problems while using socket descriptors
 * which could conflict with reserved file descriptors such as 0,1,2..
 * 
 * @param connection The connection to check.
 * 
 * @return TRUE if the socket sanity check have passed, otherwise
 * FALSE is returned.
 */
gboolean vortex_connection_do_sanity_check (VORTEX_SOCKET session)
{

	/* warn the user if it is used a socket descriptor that could
	 * be used */
	if (__vortex_connection_enable_sanity_check) {
		
		/* check for a valid socket descriptor. */
		switch (session) {
		case 0:
		case 1:
		case 2:
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "created socket descriptor using a reserved socket descriptor (%d), this is likely to cause troubles",
			       session);
			/* return sanity check have failed. */
			return FALSE;
		}
	}

	/* return the sanity check is ok. */
	return TRUE;
}
/** 
 * @internal
 * @brief Default handler used to send data.
 * 
 * See \ref vortex_connection_set_send_handler.
 */
gint vortex_connection_default_send (VortexConnection * connection,
				     gchar            * buffer,
				     gint               buffer_len)
{
	/* send the message */
	return send (connection->session, buffer, buffer_len, 0);
}

/** 
 * @internal
 * @brief Default handler to be used while receiving data
 *
 * See \ref vortex_connection_set_receive_handler.
 */
gint vortex_connection_default_receive (VortexConnection * connection,
					gchar            * buffer,
					gint               buffer_len)
{
	/* receive content */
	return recv (connection->session, buffer, buffer_len, 0);
}

/** 
 * @internal
 *
 * @brief Support function for connection identificators.  
 *
 * This is used to generate and return the next connection identifier.
 *
 * @return Next connection identifier available.
 */
gint __vortex_connection_get_next_id ()
{
	gint result;
	g_static_mutex_lock (&connection_id_mutex);
	
	result = connection_id;
	connection_id++;

	g_static_mutex_unlock (&connection_id_mutex);

	return result;
}


/** 
 * @internal
 * @brief allows to set a channel as connected
 * 
 * Really bad trick to be able to set connected state to a channel
 * this is needed for channel 0 connections and all channels 
 *
 * @param channel The channel to set the channel status.
 */
extern void __vortex_channel_set_connected (VortexChannel * channel);

typedef struct _VortexConnectionNewData {
	VortexConnection    * connection;
	VortexConnectionNew   on_connected;
	gpointer              user_data;
	gboolean              threaded;
}VortexConnectionNewData;


typedef struct _VortexConnectionCloseData {
	GList     * list;
}VortexConnectionCloseData;

void __vortex_connection_close_list_channels (gpointer key, gpointer value, gpointer user_data)
{
	
	VortexChannel             * channel = value;
	VortexConnectionCloseData * data    = user_data;
	
	/* save channel */
	data->list = g_list_append (data->list, channel);

	return;
}

/** 
 * @brief Closes all channel opened on this connection, including the channel 0.
 *
 * This function allows to close all channels opened under the given
 * connection without setting the connection to be unusable. This
 * function could be used to support those profiles which reset all
 * channels status due to a tuning reset. An example of this type of
 * tuning reset is TLS negotiation.
 * 
 * @param connection The connection where all channels will be closed
 * and released
 *
 * @param also_channel_0 Allows to control if the channel 0 have to be closed.
 *
 * @return TRUE if all channels were close, FALSE if not. Keep in mind
 * that this function returns FALSE if only one channel have rejected
 * to be closed.
 */
gboolean vortex_connection_close_all_channels (VortexConnection * connection, gboolean also_channel_0)
{
	VortexConnectionCloseData * result;	
	VortexChannel             * channel0 = NULL;
	VortexChannel             * a_channel;
	GList                     * list_aux;
	gboolean                    error = FALSE;

	/* block connection operations during channel closing for this
	 * session */
	g_mutex_lock   (connection->op_mutex);
	switch (vortex_hash_size (connection->channels)) {
	case 0:
		/* no channels to close, going to close connection */
		goto close_connection;
	case 1:
		/* only the channel 0 is available, close channel
		 * 0. */
		goto close_channel0;
	default:
		/* many channels to close. */
		break;
	}

	/* create a list of opened channels */
	result       = g_new0 (VortexConnectionCloseData, 1);
	vortex_hash_foreach (connection->channels, __vortex_connection_close_list_channels, result);
	
	/* for each channel found, try to close, letting channel 0
	 * closing to the end. */
	list_aux = g_list_first (result->list);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		
		/* get a reference to a channel found */
		a_channel = list_aux->data;
		
		/* check this channel is not the administrative one. */
		if (vortex_channel_get_number (a_channel) == 0)
			continue;

		/* close the channel, */
		if (!vortex_channel_close (a_channel, NULL)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to close a channel during vortex connection closing id=%d",
			       connection->id);
			error = TRUE;
		}
	} /* end for */

	g_list_free (result->list);
	g_free (result);
	if (error) {

		/* release lock and free connection resources */
		g_mutex_unlock (connection->op_mutex);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to close connection, there are channels still working");
		return FALSE;
	}

 close_channel0:
	/* check if it is requested to also close the channel 0 */
	if (also_channel_0) {
		/* now we have closed all channels we have to close
		 * channel 0 */
		channel0 = vortex_connection_get_channel (connection, 0);
		if (channel0 && !vortex_channel_close (channel0, NULL)) {
			/* release lock and free connection
			 * resources */
			g_mutex_unlock (connection->op_mutex);
			
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to close the channel 0..");
			return FALSE;
		}
	}
	
 close_connection:
	/* release lock and free connection resources */
	g_mutex_unlock (connection->op_mutex);

	return TRUE;

}

void __vortex_connection_parse_greetings_get_profiles (VortexConnection * connection,  axlDoc * document)
{
	axlNode     * node;
	axlNode     * child;
	axlList     * childs;
	gchar       * uri;
	int           iterator;
	int           length;

	/* Get the root element (greetings element) */
	node = axl_doc_get_root (document);
	
	/* before getting current profiles sent by the listener peer,
	 * check for localize and features */
	connection->features = axl_node_get_attribute_value_copy (node, "features");
	connection->localize = axl_node_get_attribute_value_copy (node, "localize");
	
	/* Get the position of the first profile element */
	childs   = axl_node_get_childs (node);
	iterator = 0;
	length   = axl_node_get_child_num (node);
	while (iterator < length) {

		/* get the uri value from the profile node */
		child = axl_node_get_child_nth (node, iterator);

		/* get profiles */
		uri   = axl_node_get_attribute_value_copy (child, "uri");
		connection->remote_supported_profiles = g_list_prepend (connection->remote_supported_profiles,
									uri);
		/* step to the next element next element */
		iterator++;
	}
	
	/* profiles read */
	return;
}

gboolean __vortex_connection_parse_greetings (VortexConnection * connection, VortexFrame * frame)
{
	axlDoc     * doc;
	axlDtd     * channel_dtd;
	axlError   * error;

	g_return_val_if_fail (connection, FALSE);
	g_return_val_if_fail (frame,      FALSE);

	if ((channel_dtd = vortex_dtds_get_channel_dtd ()) == NULL) {
		__vortex_connection_set_not_connected (connection, "Cannot find DTD definition for channel management");
		return FALSE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "About to parse the following message: (size: %d) '%s'",
	       vortex_frame_get_payload_size (frame), vortex_frame_get_payload (frame));

	/* dtd correct */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "found an error while parsing greetings message: %s",
		       axl_error_get (error));
		axl_error_free (error);
		__vortex_connection_set_not_connected (connection, "Cannot parse xml message to validate it");
		return FALSE;
	}

	/* doc correct */
	if (axl_dtd_validate (doc, channel_dtd, &error)) {
		__vortex_connection_parse_greetings_get_profiles (connection, doc);
		
		/* Validation successful */
		axl_doc_free (doc);
		return TRUE;
	}

	/* Validation failed */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "validation failed: %s", axl_error_get (error));
	axl_error_free (error);

	/* release memory used by the xml document */
	axl_doc_free   (doc);

	/* flag the connection to be not connected */
	__vortex_connection_set_not_connected (connection, "Incoming greetings validation failed");
	return FALSE;
}

/**
 * \defgroup vortex_connection Vortex Connection: Related function to create and manage Vortex Connections.
 */

/**
 * \addtogroup vortex_connection
 * @{
 */

/** 
 * @brief Allows to create a new \ref VortexConnection from a socket
 * that is already connected.
 *
 * This function creates the basics structures for a new \ref
 * VortexConnection instance is it were created with \ref
 * vortex_connection_new but without performing any additional BEEP
 * operation such us sending initial greetings. 
 *
 * This function only takes the given underlying transport descriptor
 * (that is, the socket) and gets the remote peer name in order to set
 * which is the remote host peer.
 *
 * This function also sets the default handler for sending and
 * receiving data using the send and recv operations. If you are
 * developing a profile which needs to notify the way data is actually
 * sent or received you should change that behavior after calling
 * this function.
 *
 *
 * @param session An already connected socket.  
 * @param role    The role to be set to the connection being created.
 * 
 * @return a newly allocated \ref VortexConnection. 
 */
VortexConnection * vortex_connection_new_empty            (gint session, VortexPeerRole role)
{
	/* creates a new connection */
	return vortex_connection_new_empty_from_connection (session, NULL, role);
}

/** 
 * @internal
 *
 * Internal function used to create new connection starting from a
 * socket, and optional from internal data stored on the provided
 * connection. It is supposed that the socket provided belongs to the
 * connection provided.
 *
 * The function tries to creates a new connection, using the socket
 * provided. The function also keeps all internal connection that for
 * the new connection creates extracting that data from the connection
 * reference provided. This is useful while performing tuning
 * operations that requires to reset the connection but retain data
 * stored on the connection.
 * 
 * @param socket The socket to be used for the new connection.
 *
 * @param connection The connection where the user space data will be
 * extracted. This reference is optional.
 *
 * @param role The connection role to be set to this function.
 * 
 * @return A newly created connection, using the provided data, that
 * must be deallocated using \ref vortex_connection_close.
 */
VortexConnection * vortex_connection_new_empty_from_connection (VORTEX_SOCKET session,
								VortexConnection * __connection,
								VortexPeerRole role)
{
	VortexConnection   * connection;
	VortexChannel      * channel;
	struct sockaddr_in   sin;
#ifdef G_OS_WIN32
	/* windows flavors */
	int                  sin_size = sizeof (sin);
#else
	/* unix flavors */
	socklen_t            sin_size = sizeof (sin);
#endif

	/* perform connection sanity check */
	if (!vortex_connection_do_sanity_check (session)) 
		return NULL;
	
	/* get remote peer name */
	if (getpeername (session, (struct sockaddr *) &sin, &sin_size) < -1) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to get remote hostname and port");
		return NULL;
	}

	connection                     = g_new0   (VortexConnection, 1);
	connection->id                 = __vortex_connection_get_next_id ();
	connection->host               = g_strdup (inet_ntoa (sin.sin_addr));
	connection->port               = g_strdup_printf ("%d", ntohs (sin.sin_port));
	connection->message            = g_strdup ("session established and ready");
	connection->session            = session;
	connection->is_connected       = TRUE;
	connection->ref_count          = 1;
	connection->ref_mutex          = g_mutex_new ();
	connection->op_mutex           = g_mutex_new ();
	connection->handlers_mutex     = g_mutex_new ();
	connection->channel_pool_mutex = g_mutex_new ();
	connection->channels           = vortex_hash_new_full (NULL, NULL, NULL,
							       (GDestroyNotify) vortex_channel_free);
	/* creates the user space data */
	if (__connection != NULL) {
		connection->data       = __connection->data;
		__connection->data     = NULL;
	}else 
		connection->data       = vortex_hash_new_full (g_str_hash, g_str_equal,
							       NULL, 
							       __vortex_connection_data_free_value);

	connection->channel_pools      = vortex_hash_new_full (NULL, NULL,
							       NULL, 
							       (GDestroyNotify) vortex_channel_pool_close);
	connection->channel_mutex      = g_mutex_new ();
	
	/* establish the connection role and its initial next channel
	 * number available. */
	connection->role               = role;
	connection->last_channel       = (role == VortexRoleListener) ? 2 : 1;

	/* set default send and receive handlers */
	connection->send               = vortex_connection_default_send;
	connection->receive            = vortex_connection_default_receive;

	/* set by default to close the underlying connection when the
	 * connection is closed */
	connection->close_session      = TRUE;
	
	/* create channel 0 (virtually always is created but, is
	 * necessary to have a representation for channel 0, in order
	 * to make channel management function to be consistent). */
	channel = vortex_channel_empty_new (0, "not applicable", connection);
	
	/* associate channel 0 with actual connection */
	vortex_connection_add_channel (connection, channel);


	return connection;	
}


/** 
 * \brief Allows to change connection semantic to blocking.
 *
 * This function should not be useful for Vortex Library consumers
 * because the internal Vortex Implementation requires connections to
 * be non-blocking semantic.
 * 
 * @param connection the connection to set as blocking
 * 
 * @return TRUE if blocking state was set or FALSE if not.
 */
gboolean vortex_connection_set_blocking_socket (VortexConnection    * connection)
{
#ifdef G_OS_UNIX
	gint flags;
#endif

	g_return_val_if_fail (connection, FALSE);
	g_return_val_if_fail (connection->is_connected, FALSE);
	
#ifdef G_OS_WIN32
	if (!vortex_win32_blocking_enable (connection->session)) {
		__vortex_connection_set_not_connected (connection, "unable to set blocking I/O");
		return FALSE;
	}
#else
	if ((flags = fcntl (connection->session, F_GETFL, 0)) < -1) {
		__vortex_connection_set_not_connected (connection, 
						       "unable to get socket flags to set non-blocking I/O");
		return FALSE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "actual flags state before setting blocking: %d", flags);
	flags &= ~O_NONBLOCK;
	if (fcntl (connection->session, F_SETFL, flags) < -1) {
		__vortex_connection_set_not_connected (connection, "unable to set non-blocking I/O");
		return FALSE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "actual flags state after setting blocking: %d", flags);
#endif
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "setting connection as blocking");
	return TRUE;
}

/** 
 * \brief Allows to change connection semantic to nonblocking.
 *
 * Sets a connection to be non-blocking while sending and receiving
 * data. This function should not be useful for Vortex Library
 * consumers.
 * 
 * @param connection the connection to set as nonblocking.
 * 
 * @return TRUE if nonblocking state was set or FALSE if not.
 */
gboolean vortex_connection_set_nonblocking_socket (VortexConnection * connection)
{
#ifdef G_OS_UNIX
	gint flags;
#endif
	g_return_val_if_fail (connection, FALSE);
	g_return_val_if_fail (connection->is_connected, FALSE);

#ifdef G_OS_WIN32
	if (!vortex_win32_nonblocking_enable (connection->session)) {
		__vortex_connection_set_not_connected (connection, "unable to set non-blocking I/O");
		return FALSE;
	}
#else
	if ((flags = fcntl (connection->session, F_GETFL, 0)) < -1) {
		__vortex_connection_set_not_connected (connection, "unable to get socket flags to set non-blocking I/O");
		return FALSE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "actual flags state before setting nonblocking: %d", flags);
	flags |= O_NONBLOCK;
	if (fcntl (connection->session, F_SETFL, flags) < -1) {
		__vortex_connection_set_not_connected (connection, "unable to set non-blocking I/O");
		return FALSE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "actual flags state after setting nonblocking: %d", flags);
#endif
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "setting connection as non-blocking");
	return TRUE;
}


/** 
 * @internal
 * @brief Support function to vortex_connection_new. 
 *
 * This function actually does the work for the vortex_connection_new.
 * 
 * @param data To perform vortex connection creation process
 * 
 * @return on thread model NULL on non-thread model the connection
 * created (connected or not connected).
 */
gpointer __vortex_connection_new (VortexConnectionNewData * data)
{
	VortexConnection   * connection = data->connection;
	struct hostent     * he;
        struct in_addr     * haddr;
        struct sockaddr_in   saddr;
	VortexFrame        * frame;
	VortexChannel      * channel;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "executing connection new in %s mode to %s:%s id=%d",
	       (data->threaded == TRUE) ? "thread" : "blocking", 
	       connection->host, connection->port,
	       connection->id);

	/*
	 * standard tcp socket voodo connection (I would like to know
	 * who was the great mind designer of this api)
	 */
        he = gethostbyname(connection->host);
        if (he == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to get host name by using gethostbyname");
		connection->message      = g_strdup ("unable to get host name by using gethostbyname");
		goto __vortex_connection_new_finalize;
	}

	/* create the socket and check if it */
        haddr                    = ((struct in_addr *) (he->h_addr_list)[0]);
        connection->session      = socket(AF_INET, SOCK_STREAM, 0);
	if (connection->session == VORTEX_INVALID_SOCKET) {

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to create socket");
		connection->message = g_strdup ("unable to create socket");
		goto __vortex_connection_new_finalize;
	}


	/* do a sanity check on socket created */
	if (!vortex_connection_do_sanity_check (connection->session)) {
		connection->message = g_strdup_printf ("created socket descriptor using a reserved socket descriptor (%d), this is likely to cause troubles",
						       connection->session);
		goto __vortex_connection_new_finalize;
	}
	
	/* prepare socket configuration to operate using TCP/IP
	 * socket */
        memset(&saddr, 0, sizeof(saddr));
        memcpy(&saddr.sin_addr, haddr, sizeof(struct in_addr));
        saddr.sin_family    = AF_INET;
        saddr.sin_port      = htons((uint16_t) g_ascii_strtod (connection->port, NULL));

	/* do a tcp connect */
        if (connect(connection->session, 
		    (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
		shutdown (connection->session, SHUT_RDWR);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to connect to remote host");
		connection->message = g_strdup ("unable to connect to remote host");
		goto __vortex_connection_new_finalize;
	}

	connection->is_connected = TRUE;

	/* activate vortex writer for this connection because all
	 * sending process is done through vortex sequencer plus
	 * vortex writer. */
	vortex_writer_watch_connection      (connection);
	goto __vortex_connection_new_invoke;

 __vortex_connection_new_finalize:
	/* common case when everything goes wrong */
	connection->is_connected = FALSE;
	vortex_close_socket (connection->session);
	connection->session = -1;
		

 __vortex_connection_new_invoke:
	/* according to the connection status (is_connected attribute)
	 * perform the final operations so the connection becomes
	 * usable. Later, the user app level is notified. */
	if (connection->is_connected) {

		/* create channel 0 (virtually always is created but,
		 * is necessary to have a representation for channel
		 * 0, in order to make channel management function to
		 * be consistent). */
		channel = vortex_channel_empty_new (0, "not applicable", connection);
		vortex_connection_add_channel  (connection, channel);

		/* now we have to send greetings and process them */
		if (! vortex_greetings_client_send (connection)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, vortex_connection_get_message (connection));
			goto __vortex_connection_new_invoke_caller;
		}
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "greetings sent, waiting for reply");

		/* block thread until received remote greetings */
		frame = vortex_greetings_client_process (connection);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "greetings received, process reply frame");

		/* process frame response */
		if (!vortex_connection_parse_greetings_and_enable (connection, frame))
			goto __vortex_connection_new_finalize;

		/* check for auto TLS negotiation */
		if (__vortex_connection_auto_tls) {
			/* seems automatic TLS profile negotiation is
			 * activated, check for TLS support */
			if  (! vortex_tls_is_enabled () && !__vortex_connection_auto_tls_allow_failures) {
				__vortex_connection_set_not_connected (connection, "Unable to create a new connection, auto TLS activated and current Vortex Library doesn't have support for TLS profile");
				goto __vortex_connection_new_invoke_caller;
			}

			/* seems that current Vortex Library have TLS
			 * profile built-in support and auto TLS is
			 * activated */
			connection = vortex_tls_start_negociation_sync (connection,
									__vortex_connection_auto_tls_server_name,
									NULL, NULL);
			if (!vortex_connection_is_ok (connection, FALSE)) {
				goto __vortex_connection_new_invoke_caller;
			}
			
			/* the connection is ok, check if the TLS
			 * profile was activated, but only if auto tls
			 * allow failures is not set */
			if (! __vortex_connection_auto_tls_allow_failures) {
				if (! vortex_connection_is_tlsficated (connection)) {
					__vortex_connection_set_not_connected (connection,
									       "Automatic TLS-fication have failed, the connection have been flagged to be closed due to not allowing TLS failure settings");
					goto __vortex_connection_new_invoke_caller;
				} /* end if */
			} /* end if */
			
			/* lets continue calling the user space code */
		}
	}
 __vortex_connection_new_invoke_caller: 
	/* notify on callback or simply return */
	if (data->threaded) {
		data->on_connected (connection, data->user_data);
		g_free (data);			
		return NULL;
	}
	g_free (data);
	return connection;
}

/** 
 * @brief Allows to create a new BEEP session (connection) to the given <i>host:port</i>.
 *
 * While working with <b>BEEP</b> frameworks, you need to create a new
 * connection (\ref VortexConnection), and over it, create channels (\ref
 * VortexChannel), which will allow you to send and receive data. Keep
 * in mind connection doesn't send and receive data, this is actually
 * done by channels.
 *
 * Once you have created an connection, using this function, you can
 * use \ref vortex_channel_new function to create new channels, using
 * the profile you have selected. 
 * 
 * <b>Profiles</b> are semantic definitions, representing an agreement
 * between peers for those message to be exchanged inside a channel,
 * and what they mean. You can read the following \ref profile_example
 * "tutorial" to know more \ref profile_example "about profiles" and
 * \ref profile_example "how they interact" with Vortex Library.
 *
 * This function will block you until the connection is created with
 * remote site or until timeout mechanism is reached. You can define
 * vortex timeout for connections creation by using
 * \ref vortex_connection_timeout.
 *
 * Optionally you can define an \ref VortexConnectionNew "on_connected handler" 
 * to process response, avoiding getting blocked on \ref vortex_connection_new call.
 *
 * If the \ref VortexConnectionNew "on_connected handler" is defined
 * the connection creation will be performed in an independent
 * thread. This will allow the caller to keep on doing other tasks
 * while the connection is being created. This means \ref
 * vortex_connection_new function will never block the caller if the
 * \ref VortexConnectionNew "on_connected handler" is defined.
 * 
 * Inside the connection process, a session negotiation will take
 * place. BEEP RFC defines that remote server peer must send its
 * supported profiles. Due to this, \ref VortexConnection will hold
 * all remote supported profiles. You will need this information to
 * know which profiles supports the remote peer, so you can check if
 * your BEEP client application can actually talk to the remote
 * server. Use \ref vortex_connection_get_remote_profiles or \ref vortex_connection_is_profile_supported for this
 * issue.
 *
 * BEEP framework also allows to define special features and localize
 * options to be notified to the remote peer we are going to connect
 * to. This is actually done by using \ref
 * vortex_greetings_set_features and \ref
 * vortex_greetings_set_localize. Values set on that function will be
 * used for all connections created. 
 *
 * To actually get features and localize requested by the peer we have
 * connected to, you have to use \ref vortex_connection_get_features and
 * \ref vortex_connection_get_localize.
 *
 * Once finished with the connection, at the client side, you have to
 * call to \ref vortex_connection_close function close and dispose the
 * BEEP session.
 *
 * If you are interested on getting notifications once your connection
 * is broken, in unix environments know as "broken pipe", you can use
 * the following handler:
 * 
 *   - \ref vortex_connection_set_on_close
 *
 * At listener side, it could be required to get notifications about
 * new incoming connections accepted, mainly to perform some special
 * initialization task. Then you can use the following:
 * 
 *  - \ref vortex_listener_set_on_connection_accepted 
 *
 * Additionally, you may be interesting on securing the connection
 * created using the TLS profile. This is actually done with \ref
 * vortex_tls_start_negociation. Optionally, you can configure your
 * Vortex Library client peer to automatically negotiate the TLS
 * profile for every connection created, allowing to get from this
 * function a connection that already have TLS profile activated. You
 * can configure this behavior using \ref vortex_connection_set_auto_tls. 
 *
 * Check out the \ref vortex_connection_set_auto_tls "documentation"
 * for \ref vortex_connection_set_auto_tls to know more about using
 * automatic TLS profile negotiation.
 *
 * Finally, while considering how to transport user space data that
 * actually applies to the connection being created consider using
 * \ref vortex_connection_set_data and \ref
 * vortex_connection_set_data_full.  This function will allow you to
 * store user space data in a hash-index manner that could be
 * automatically deallocated while using \ref vortex_connection_set_data_full.
 *
 * 
 * 
 * @param host The remote peer to connect to.
 * @param port The peer's port to connect to.
 * @param on_connected Optional handler to process connection new status.
 * @param user_data Optional handler to process connection new status
 * 
 * @return a new VortexConnection. You should use vortex_connection_is_ok
 * to check if you have already connected.
 */
VortexConnection * vortex_connection_new (gchar * host, gchar * port, 
					  VortexConnectionNew on_connected,
					  gpointer user_data)
{

	VortexConnectionNewData * data;

	g_return_val_if_fail (host, NULL);
	g_return_val_if_fail (port, NULL);

	data                                  = g_new0 (VortexConnectionNewData, 1);
	data->connection                      = g_new0 (VortexConnection, 1);
	data->connection->id                  = __vortex_connection_get_next_id ();
	data->connection->host                = g_strdup (host);
	data->connection->port                = g_strdup (port);
	data->connection->channels            = vortex_hash_new_full (NULL, NULL, NULL,
								      (GDestroyNotify) vortex_channel_free);
	data->connection->channel_mutex       = g_mutex_new ();
	data->connection->ref_count           = 1;
	data->connection->ref_mutex           = g_mutex_new ();
	data->connection->op_mutex            = g_mutex_new ();
	data->connection->handlers_mutex      = g_mutex_new ();
	data->connection->channel_pool_mutex  = g_mutex_new ();
	data->connection->data                = vortex_hash_new_full (g_str_hash, g_str_equal,
								      NULL,
								      __vortex_connection_data_free_value);
	data->connection->channel_pools       = vortex_hash_new_full (NULL, NULL, NULL, 
								      (GDestroyNotify) vortex_channel_pool_close);

	/* establish the connection role and initial next channel
	 * available. */
	data->connection->role                = VortexRoleInitiator;
	data->connection->last_channel        = 1;

	/* set default send and receive handlers */
	data->connection->send                = vortex_connection_default_send;
	data->connection->receive             = vortex_connection_default_receive;

	/* set by default to close the underlying connection when the
	 * connection is closed */
	data->connection->close_session       = TRUE;

	/* establish connection status, connection negotiation method
	 * and user data */
	data->on_connected                    = on_connected;
	data->user_data                       = user_data;
	data->threaded                        = (on_connected != NULL);

	if (data->threaded) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invoking connection_new threaded mode");
		vortex_thread_pool_new_task ((GThreadFunc) __vortex_connection_new, data);
		return NULL;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invoking connection_new non-threaded mode");
	return __vortex_connection_new (data);
}

/** 
 * @brief Allows to reconnect the given connection, using actual connection settings.
 * 
 * Given a connection that is actually broken, this function can be
 * used to reconnect it trying to find, on the same host and port, the
 * peer that the connection was linked to.
 *
 * The connection "reconnected" will keep the connection identifier,
 * any data set on this connection using the \ref
 * vortex_connection_set_data and \ref vortex_connection_set_data_full
 * functions, the connection reference count and remote profiles read.
 *
 * If the function detect that the given connection is currently
 * working and connected then the function does nothing.
 *
 * Here is an example:
 * \code
 * // check if the connection is broken, maybe remote site have
 * // failed.
 * if (! vortex_connection_is_ok (connection, FALSE)) {
 *       // connection is not available, try to reconnect.
 *       if (! vortex_connection_reconnect (connection,
 *                                          // no async notify
 *                                          NULL,
 *                                          // no user data
 *                                          NULL)) {
 *            g_print ("Link failure detected..");
 *       }else {
 *            g_print ("Unable to reconnect to remote site..");
 *       }
 * } // end if
 * \endcode
 * 
 * @param connection    The connection to reconnect to.
 *
 * @param on_connected Optional handler. This function works as \ref
 * vortex_connection_new. This handler can be defined to notify when
 * the connection has been reconnected.
 *
 * @param user_data User defined data to be passed to on_connected.
 * 
 * @return TRUE if the connection was successfully reconnected or FALSE
 * if not. If the <i>on_connected</i> handler is defined, this
 * function will always return TRUE.
 */
gboolean            vortex_connection_reconnect              (VortexConnection * connection,
							      VortexConnectionNew on_connected,
							      gpointer user_data)
{
	VortexConnectionNewData * data;

	/* we have to check a basic case. A null connection. Null
	 * connection no data to reestablish the connection. */
	if (connection == NULL)
		return FALSE;

	if (vortex_connection_is_ok (connection, FALSE))
		return TRUE;


	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "requested a reconnection on a connection which isn't null and is not connected..");

	/* close all pending channels (clear all channels and pools):
	 * the following to lines should be this way and this
	 * order. If not, really bad things will happen. This is
	 * because releasing all channels, due to freeing it from the
	 * hash table, automatically remove it from the pool. Once all
	 * the channels are removed, channels pools are removed with
	 * no problem.  */
	vortex_hash_clear (connection->channels);
	vortex_hash_clear (connection->channel_pools);
	
	/* reset channel pool id counting */
	connection->next_channel_pool_id = 0;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "channels, channels pools and channel pools indications cleared..");

	/* create data needed to invoke the service */
	data               = g_new0 (VortexConnectionNewData, 1);
	data->connection   = connection;
	
	
	data->on_connected = on_connected;
	data->user_data    = user_data;
	data->threaded     = (on_connected != NULL);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reconnection request is prepared, doing so..");
	
	if (data->threaded) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reconnecting connection in threaded mode");
		vortex_thread_pool_new_task ((GThreadFunc) __vortex_connection_new, data);
		return TRUE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reconnecting connection in non-threaded mode");
	return vortex_connection_is_ok (__vortex_connection_new (data), FALSE);

}


/** 
 * @brief Tries to close properly a connection and all channels inside it.
 * 
 * This function close a connection and unref it using \ref vortex_connection_unref.
 *
 * Because there can be channels created and still working this
 * function will try to close them by executing vortex_channel_close
 * on them. If vortex_channel_close fails for one of them, the
 * vortex_connection_close will be stopped and the connection will not
 * be closed.
 *
 * Do not call this function twice because a unref operations is
 * performed over the connection if the function returns TRUE. This
 * means after calling this function the connection reference must not
 * be used until a new connection is created. Otherwise a segfault
 * may happen.
 *
 * If the function receives a null connection reference, it will just
 * return a TRUE value without doing nothing.
 * 
 * @param connection the connection to close properly.
 * 
 * @return TRUE if connection was closed and FALSE if not. If there
 * are channels still working, the connection will not be closed.
 */
gboolean               vortex_connection_close                  (VortexConnection * connection)
{

	/* if the connection is a null reference, just return TRUE */
	if (connection == NULL)
		return TRUE;

	/* close all channel on this connection */
	if (vortex_connection_is_ok (connection, FALSE)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "closing a connection which is already opened with %d channels opened..",
		       vortex_hash_size (connection->channels));
		
		/* close all channels because the connection at this
		 * point is running properly */
		if (!vortex_connection_close_all_channels (connection, TRUE)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "failed while closing all channels");
			return FALSE;
		}
		
		/* set the connection to be not connected */
		__vortex_connection_set_not_connected (connection, "close connection called");
	}else 
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "closing a connection which is not opened, unref resources..");

	/* unref vortex connection resources */
	vortex_connection_unref (connection, "vortex_connection_close");

	return TRUE;
}

/** 
 * @brief Increase internal vortex connection reference counting.
 * 
 * Because Vortex Library design, several on going threads shares
 * references to the same connection for several purposes. 
 * 
 * Connection reference counting allows to every on going thread to
 * notify the system that connection reference is no longer be used
 * so, if the reference counting reach a zero value, connection
 * resources will be deallocated.
 *
 * While using the Vortex Library is not required to use this function
 * especially for those applications which are built on top of a
 * profile which is layered on Vortex Library. 
 *
 * This is because connection handling is through functions such \ref
 * vortex_connection_new and \ref vortex_connection_close.
 *
 * However, while implementing new profiles these function becomes a
 * key concept to ensure the profile implementation don't get lost
 * connection references.
 *
 * Keep in mind that using this function implied to use \ref
 * vortex_connection_unref function in all code path implemented. For
 * each call to \ref vortex_connection_ref it should exist a call to
 * \ref vortex_connection_unref. Failing on doing this will cause
 * either memory leak or memory corruption because improper connection
 * deallocations.
 * 
 *
 * @param connection the connection to operate.
 * @param who who have increased the reference.
 */
void               vortex_connection_ref                    (VortexConnection * connection, gchar * who)
{
	g_return_if_fail (connection);

	/* lock ref/unref operations over this connection */
	g_mutex_lock   (connection->ref_mutex);

	/* increase and log the connection increased */
	connection->ref_count++;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "increased connection id=%d reference to %d by %s",
	       connection->id,
	       connection->ref_count, who ? who : "??" );

	/* unlock ref/unref options over this connection */
	g_mutex_unlock (connection->ref_mutex);

	return;
}

/** 
 * @brief Decrease vortex connection reference counting.
 *
 * Allows to decrease connection reference counting. If this reference
 * counting goes under 0 the connection resources will be deallocated. 
 *
 * See also \ref vortex_connection_ref
 * 
 * @param connection The connection to operate.
 * @param who        Who have decreased the reference. This is a string value used to log which entity have decreased the connection counting.
 */
void               vortex_connection_unref                  (VortexConnection * connection, gchar * who)
{
	g_return_if_fail (connection);

	/* lock the connection being unrefered */
	g_mutex_lock     (connection->ref_mutex);

	/* decrease reference counting */
	connection->ref_count--;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "decreased connection id=%d reference count to %d decreased by %s", 
	       connection->id,
	       connection->ref_count, who ? who : "??");

	/* check if the current thread should unref the connection  */
	if (connection->ref_count == 0) {
		g_mutex_unlock (connection->ref_mutex);
		vortex_connection_free (connection);
		return;
	}
	g_mutex_unlock (connection->ref_mutex);
	return;
}


/** 
 * @brief Allows to get current reference count for the provided connection.
 *
 * See also the following functions:
 *  - \ref vortex_connection_ref
 *  - \ref vortex_connection_unref
 * 
 * @param connection The connection that is requested to return its
 * count.
 *
 * @return The function returns the reference count or -1 if it fails.
 */
gint                vortex_connection_ref_count              (VortexConnection * connection)
{
	g_return_val_if_fail (connection, -1);

	/* return the reference count */
	return connection->ref_count;
}

/** 
 * @brief Allows to tweak vortex internal timeouts
 * 
 * This function allows to set the timeout to use on new
 * connection creation.  Default timeout is defined to 10 seconds (10
 * x 1000000 = 10000000 milliseconds).
 *
 * If you call to create a new connection with \ref vortex_connection_new
 * and remote peer doesn't responds within the period,
 * \ref vortex_connection_new will return with a non-connected vortex
 * connection.
 *
 *
 * @param milliseconds_to_wait timeout value to be used.
 */
void               vortex_connection_timeout (glong milliseconds_to_wait)
{
	gchar * value;
	
	/* clear previous value */
	if (milliseconds_to_wait == 0) {
		g_unsetenv ("VORTEX_SYNC_TIMEOUT");
		return;
	}

	/* set new value */
	value = g_strdup_printf ("%ld", milliseconds_to_wait);
	g_setenv ("VORTEX_SYNC_TIMEOUT", value, 1);
	g_free (value);

	return;
}

static glong   __vortex_connection_std_timeout             = 60000000;

/** 
 * @brief Allows to get current timeout set for \ref VortexConnection
 * synchronous operations.
 *
 * See also \ref vortex_connection_timeout.
 *
 * @return Current timeout set. 
 */
glong             vortex_connection_get_timeout ()
{
	gchar     * c_timeout   = NULL;
	gchar     * str_aux     = NULL;
	glong       d_timeout   = __vortex_connection_std_timeout;
	gboolean    invalid_char;
	gint        i;

	c_timeout = (gchar *) g_getenv ("VORTEX_SYNC_TIMEOUT");
	if (c_timeout == NULL) {
		/* no timeout definition done using default timeout 10 seconds */
		return __vortex_connection_std_timeout;
	}else {
		/* check every char is a digit */
		invalid_char = FALSE;
		for (i = 0; c_timeout[i]; i++) {
			if (!g_ascii_isdigit (c_timeout[i])) {
				invalid_char = TRUE;
				break;
			}
		}
		
		/* get actual timeout to wait request response */
		d_timeout =  g_strtod (c_timeout, &str_aux);
		if (str_aux && (*str_aux)) {
			/* user have misspelled timeout value */
			return __vortex_connection_std_timeout;
		}
	}
	return d_timeout;		
}

/** 
 * @brief Allows to get current connection status
 *
 * This function will allow you to check if your vortex connection is
 * actually connected. You must use this function before calling
 * vortex_connection_new to check what have actually happen.
 *
 * You can also use vortex_connection_get_message to check the message
 * returned by the vortex layer. This may be useful on connection
 * errors.  The free_on_fail parameter can be use to free vortex
 * connection resources if this vortex connection is not
 * connected. This operation will be done by using \ref vortex_connection_close.
 *
 * 
 * @param connection the connection to get current status.
 *
 * @param free_on_fail if TRUE the connection will be closed using
 * vortex_connection_close on not connected status.
 * 
 * @return current connection status for the given connection
 */
gboolean           vortex_connection_is_ok (VortexConnection * connection, gboolean free_on_fail)
{
	/* check connection null referencing. */
	if  (connection == NULL) 
		return FALSE;

	/* implement free_on_fail flag */
	if (free_on_fail && ! connection->is_connected) {
		vortex_connection_close (connection);
		return FALSE;
	}
	
	/* check for the socket this connection has */
	if (connection->session < 0)
		return FALSE;

	/* return current connection status. */
	return connection->is_connected;
}

/** 
 * @brief Returns actual message status for the given connection
 *
 * Return message associated to this vortex connection. The message
 * associated may be useful on connection errors. On an successful
 * connected vortex connection it will return "already connected".
 * 
 * @param connection the connection used to get message status.
 * 
 * @return a message about vortex connection status. You must not free
 * this value. Use vortex_connection_free.
 */
gchar            * vortex_connection_get_message (VortexConnection * connection)
{
	g_return_val_if_fail (connection, NULL);
	return connection->message;
}


/** 
 * @brief Frees vortex connection resources
 * 
 * Free all resources allocated by the vortex connection. You have to
 * keep in mind if <i>connection</i> is already connected,
 * vortex_connection_free will close all channels running on this and
 * also close the connection.
 *
 * Generally is not a good a idea to call this function. This is
 * because every connection created using the vortex API is registered
 * at some internal process (the vortex reader, sequencer and writer)
 * so they have references to created connection to do its job. 
 *
 * To close a connection properly call \ref vortex_connection_close.
 * 
 * @param connection the connection to free
 */
void               vortex_connection_free (VortexConnection * connection)
{
	GList * list;
	
	g_return_if_fail (connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection id=%d", connection->id);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection message id=%d", connection->id);

        /* free all resources */
	if (connection->message != NULL) {
		g_free (connection->message);
		connection->message = NULL;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection host id=%d", connection->id);

	/* free host and port */
	if (connection->host != NULL) {
		g_free (connection->host);
		connection->host    = NULL;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection port id=%d", connection->id);

	if (connection->port != NULL) {
		g_free (connection->port);
		connection->port    = NULL;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection profiles id=%d", connection->id);

	/* free profiles */
	if (connection->remote_supported_profiles) {
		list = g_list_first (connection->remote_supported_profiles);
		for (;list; list = g_list_next (list)) {
			axl_free (list->data);
		}
		g_list_free (list);
		connection->remote_supported_profiles = NULL;
	}

	/* free features and localize */
	if (connection->features != NULL)
		axl_free (connection->features);
	if (connection->localize != NULL)
		axl_free (connection->localize);
	

	/*
	 * NOTE: The order in which the channels and the channel pools
	 * are closed must be this way: first channels and the channel
	 * pools. Doing it other way will produce funny dead-locks.
	 */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection channels id=%d", connection->id);

	/* free channel resources */
	if (connection->channels) {
		vortex_hash_destroy (connection->channels);
		connection->channels = NULL;
	}
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection channel pools id=%d", connection->id);
	/* free channel pools */
	if (connection->channel_pools) {
		vortex_hash_destroy (connection->channel_pools);
		connection->channel_pools = NULL;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection custom data holder id=%d", connection->id);
	/* free connection data hash */
	if (connection->data) {
		vortex_hash_destroy (connection->data);
		connection->data = NULL;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection channel mutex id=%d", connection->id);

	/* free mutex */
	g_mutex_free (connection->channel_mutex);
	connection->channel_mutex       = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection reference counting mutex id=%d", connection->id);

	/* release ref mutex */
	g_mutex_free   (connection->ref_mutex);
	connection->ref_mutex           = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection operational mutex id=%d", connection->id);

	g_mutex_free (connection->op_mutex);
	connection->op_mutex            = NULL;

	g_mutex_free (connection->handlers_mutex);
	connection->handlers_mutex      = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing connection channel pools mutex id=%d", connection->id);

	g_mutex_free (connection->channel_pool_mutex);
	connection->channel_pool_mutex  = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "freeing/terminating connection id=%d", connection->id);

	/* close connection */
	if (( connection->close_session) && (connection->session != -1)) {
		/* it seems that this connection is  */
		shutdown (connection->session, SHUT_RDWR);
		vortex_close_socket (connection->session);
		connection->session = -1;
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "session socket closed");
	}

	/* release on close handler list */
	if (connection->on_close)
		g_list_free (connection->on_close);
	if (connection->on_close_full)
		g_list_free (connection->on_close_full);
	if (connection->on_close_full_data)
		g_list_free (connection->on_close_full_data);

	/* free connection */
	g_free (connection);

	return;
}

/** 
 * @brief Returns the remote peer supported profiles
 * 
 * When a Vortex connection is opened, remote server sends a list of
 * BEEP supported profiles.
 *
 * This is necessary to be able to create new channels. The profile
 * selected for the channel to be created must be supported for both
 * sides.
 *
 * This function allows to get remote peer supported
 * profiles. This can be helpful to avoid connection to a remote BEEP
 * peers that actually doesn't support your profile.  
 *
 * You must not free the returned GList. As a example, you can use
 * this function as follows:
 * 
 * \code
 * GList * profiles = NULL;
 * int     iterator = 0;
 *
 * profiles = vortex_connection_get_remote_profiles (connection);
 * g_print ("profiles for this peer: %d\n", g_list_length (profiles));
 * list_aux = g_list_first (profiles);
 * for (; list_aux; list_aux = g_list_next (list_aux)) {
 *     g_print ("  %d) %s\n", iterator, (gchar*) list_aux->data);
 *     iterator++;
 * }
 * \endcode
 * 
 * @param connection the connection to get remote peer profiles.
 * 
 * @return A GList containing each element a uri identifying a remote peer profile.
 */
GList            * vortex_connection_get_remote_profiles (VortexConnection * connection)
{
	g_return_val_if_fail (connection, NULL);
	g_return_val_if_fail (connection->is_connected, NULL);

	/* NOTE IMPLEMENTER: this function is being used by
	 * __vortex_sasl_filter_profiles, inside the vortex_sasl.c
	 * module supposing that the list returned is an internal
	 * reference. Keep in mind this while modifying this
	 * function. */

	return connection->remote_supported_profiles;
}

/** 
 * @internal
 *
 * This function is not exposed as normal API to be consumed because
 * it is used internal by profiles implementors to be able to
 * configure profile list. 
 *
 * Currently this function is used by the sasl module to filter all
 * SASL auth profiles that are listed once a SASL auth process is
 * successfully finished. Because SASL auth process doesn't involves a
 * connection reset, a new greting message is not issue making this
 * profile list to be not updated.
 * 
 * @param connection The connection where the profile list will be
 * listed.
 *
 * @param profiles The new list profile list reference to be set, keep
 * in mind that previous list is not removed, so, you must provide an
 * updted reference to the same list already stored. This list could
 * be retrieved by calling to \ref vortex_connection_get_remote_profiles.
 */
void                vortex_connection_set_remote_profiles    (VortexConnection * connection,
							      GList            * profiles)
{
	g_return_if_fail (connection);
	g_return_if_fail (profiles);

	/* just set the new list */
	connection->remote_supported_profiles = profiles;
	
	return;
}

/** 
 * @brief Allows to check if the given profile is supported by the
 * remote peer.
 *
 * This function allows to check if the profile identified by
 * <i>uri</i> is supported by remote peer connected to by <i>connection</i>.
 *
 * You can use this function to check, before creating a channel, if
 * remote peer will support this profile.
 * 
 * @param connection the connection to check.
 * @param uri the profile identified by the string uri to check.
 * 
 * @return TRUE if the profile is supported by remote peer or FALSE if not.
 */
gboolean           vortex_connection_is_profile_supported (VortexConnection * connection, gchar * uri)
{
	GList * list_aux;

	g_return_val_if_fail (connection, FALSE);
	g_return_val_if_fail (uri, FALSE);

	if (!connection->is_connected) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "trying to get supported profile in a non-connected session");
		return FALSE;
	}

	/* lock the connection while getting the list */
	g_mutex_lock (connection->op_mutex);
	list_aux = g_list_first (connection->remote_supported_profiles);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		if (!g_strcasecmp (list_aux->data, uri)) {
			g_mutex_unlock (connection->op_mutex);
			return TRUE;
		}
	}

	g_mutex_unlock (connection->op_mutex);
	return FALSE;
}

/** 
 * @brief Check if a channel is already created on the given connection. 
 *
 * This function allows to check if a channel is already created
 * over this connection (or session).  Channel number must be a
 * positive integer starting from 0. Channel 0 is always opened and
 * exists until connection (or session) is closed.
 * 
 * @param connection vortex connection to check channel existence
 * @param channel_num  channel number to check
 * 
 * @return TRUE if channel identified by <i>channel_num</i> exists,
 * otherwise FALSE.
 */
gboolean           vortex_connection_channel_exists       (VortexConnection * connection, gint channel_num)
{
	gboolean result;

	g_return_val_if_fail (channel_num >= 0,         FALSE);
	g_return_val_if_fail (connection,               FALSE);
	g_return_val_if_fail (connection->is_connected, FALSE);

	/* channel 0 always exists, and cannot be closed. It's closed
	 * when connection (or session) is closed */
	if (channel_num == 0) 
		return TRUE;
	
	result = (vortex_hash_lookup (GINT_TO_POINTER (connection->channels), 
				     GINT_TO_POINTER (channel_num)) != NULL);
	return result;
}

/** 
 * @brief Allows to perform an iterator over all channels created
 * inside the given connection.
 *
 * Allows an application to iterate over all channel created over this
 * session. This allows to check or do some operation over the session
 * channels.  
 *
 * The channel iteration is made following the channel number
 * order. First low-numbered channel following high-numbered ones.
 *
 * Example:
 *
 * \code
 * void my_function_foreach (gpointer key, gpointer value, gpointer user_data)
 * {
 *      VortexChannel * channel = value;
 *    
 *      // do some operation with the channel.
 * }
 *
 * void my_function (VortexConnection * connection) 
 * {
 *
 *      // do a channel foreach on this connection
 * 	vortex_connection_foreach_channel (connection, my_function_foreach, NULL);
 *
 *      return;
 * }
 * \endcode
 * 
 * @param connection the connection where channels will be iterated.
 * @param func the function to apply for each channel found
 * @param user_data user data to be passed into the foreach function.
 */
void            vortex_connection_foreach_channel        (VortexConnection * connection,
							  GHFunc func,
							  gpointer user_data)
{
	g_return_if_fail (connection);
	g_return_if_fail (func);
	
	vortex_hash_foreach (connection->channels, func, user_data);

	return;
}


/** 
 * 
 * @brief Returns the next channel number free to be used over this
 * session.
 *
 * This function iterates over the allowed channel number range, that
 * is: 1..2147483647 including both limits, stepping over channel
 * number 0, which always exists during the connection live.
 *
 * Channel numbers assigned automatically depends on the connection
 * role. In the case the connection role is "initiator", next channel
 * number returned will be odd-numbered. In the case the connection
 * role is "listener", next channel number will be even-numbered.
 *
 *
 * @param connection the connection where the channel is going to be
 * created
 * 
 * @return the channel number to use or -1 if fail.
 */
gint               vortex_connection_get_next_channel     (VortexConnection * connection)
{
	gint channel_num = -1;
	g_return_val_if_fail (connection, -1);
	g_return_val_if_fail (connection->is_connected, -1);
	
	/* create a critical section until found next channel to
	 * use */
	g_mutex_lock (connection->channel_mutex);
	
	while (TRUE) {
		/* get the next channel number to use */
		connection->last_channel = ((connection->last_channel + 2) % MAX_CHANNELS_NO);

		/* avoid to return 0 as valid channel number to be used */
		if (connection->last_channel == 0) {
			connection->last_channel = (vortex_connection_get_role (connection) == VortexRoleInitiator) ? 1 : 2;
		}

		/* check if channel exists */
		if (!vortex_connection_channel_exists (connection, connection->last_channel)) {
			channel_num = connection->last_channel;
			break;
		}
	}
	/* unblock critical section until found next channel to use */
	g_mutex_unlock (connection->channel_mutex);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "returning as next channel to create: %d",
	       channel_num);
	return channel_num;
}


/** 
 * @brief Returns a reference to the channel identified by <i>channel_num</i> on this
 * connection (or vortex session).
 * 
 * @param connection the connection to look for the channel.
 * @param channel_num the channel_num identifier.
 * 
 * @return the VortexChannel reference or NULL if it doesn't exists.
 */
VortexChannel    * vortex_connection_get_channel          (VortexConnection * connection, gint channel_num)
{
	g_return_val_if_fail (channel_num >= 0,         NULL);
	g_return_val_if_fail (connection,               NULL);
	g_return_val_if_fail (connection->is_connected, NULL);
	

	/* channel 0 always exists, and cannot be closed. It's closed
	 * when connection (or session) is closed */
	return vortex_hash_lookup (GINT_TO_POINTER (connection->channels), GINT_TO_POINTER(channel_num));
}

/** 
 * @brief Returns the socket used by this VortexConnection object.
 * 
 * @param connection the connection to get the socket.
 * 
 * @return the socket used or -1 if fail
 */
VORTEX_SOCKET    vortex_connection_get_socket           (VortexConnection * connection)
{
	g_return_val_if_fail (connection,               -1);

	return connection->session;
}

/** 
 * @brief Allows to configure what to do with the underlying socket
 * connection when the \ref VortexConnection is closed.
 *
 * This function could be useful if it is needed to prevent from
 * vortex library to close the underlying socket.
 * 
 * @param connection The connection to configure.
 *
 * @param action TRUE if the given connection have to be close once a
 * Vortex Connection is closed. FALSE to avoid the socket to be
 * closed.
 */
void                vortex_connection_set_close_socket       (VortexConnection * connection, 
							      gboolean action)
{
	g_return_if_fail (connection);
	connection->close_session = action;
	return;
}

/** 
 * @brief Adds a VortexChannel into an existing VortexConnection. 
 *
 * The channel to be added must not exists inside <i>connection</i>,
 * otherwise an error will be produced and channel will not be added.
 *
 * This function only adds the data structure which represents a
 * channel. It doesn't make any work about beep channel starting or
 * something similar.  
 *
 * Mainly, this function is only useful for internal vortex library
 * purposes.
 * 
 * @param connection the connection where channel will be added.
 * @param channel the channel to add.
 */
void               vortex_connection_add_channel          (VortexConnection * connection, 
							   VortexChannel * channel)
{
	VortexChannel * _channel;
	g_return_if_fail (connection);
	g_return_if_fail (channel);
	g_return_if_fail (connection->channels);

	/* lock channel hash */
	g_mutex_lock (connection->channel_mutex);

	/* check if channel doesn't exists on this connection */
	_channel = vortex_hash_lookup (connection->channels, GINT_TO_POINTER (vortex_channel_get_number (channel)));
	if (_channel != NULL) {
		g_mutex_unlock (connection->channel_mutex);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "trying to add a channel on a connection which already have this channel");
		return;
	}

	/* insert new channel on this connection */
	vortex_hash_replace (connection->channels, 
			      GINT_TO_POINTER (vortex_channel_get_number (channel)),  
			      channel);

	/* make channel to be on state connected */
	__vortex_channel_set_connected (channel);

	/* unlock channel hash */
	g_mutex_unlock (connection->channel_mutex);

	return;
}

/** 
 * @brief Removes the given channel from this connection. 
 * 
 * @param connection the connection where the channel will be removed.
 * @param channel the channel to remove from the connection.
 */
void               vortex_connection_remove_channel       (VortexConnection * connection, 
							   VortexChannel    * channel)
{
	g_return_if_fail (connection);
	g_return_if_fail (channel);
	g_return_if_fail (connection->channels);

	vortex_hash_remove (connection->channels,
			    GINT_TO_POINTER (vortex_channel_get_number (channel)));
	return;
}

/** 
 * @brief Returns the actual host this connection is connected to.
 *
 * You must not free returned value.  If you do so, you will get
 * unexpected behaviors.
 * 
 * 
 * @param connection the connection to get host value.
 * 
 * @return the host the given connection is connected to or NULL if something fail.
 */
gchar            * vortex_connection_get_host             (VortexConnection * connection)
{
	g_return_val_if_fail (connection, NULL);

	return connection->host;
}

/** 
 * @brief  Returns the connection unique identifier.
 *
 * The connection identifier is a unique integer assigned to all
 * connection created under Vortex Library. This allows Vortex programmer to
 * use this identifier for its own application purposes
 *
 * @param connection the unique integer identifier for the given connection.
 * 
 * @return the unique identifier.
 */
gint               vortex_connection_get_id               (VortexConnection * connection)
{
	g_return_val_if_fail (connection, -1);

	return connection->id;
}

/** 
 * @brief Returns the actual port this connection is connected to.
 *
 * You must not free returned value.  If you do so, you will get
 * unexpected behaviors.
 * 
 * @param connection the connection to get the port value.
 * 
 * @return the port or NULL if something fails.
 */
gchar            * vortex_connection_get_port             (VortexConnection * connection)
{
	g_return_val_if_fail (connection, NULL);

	return connection->port;
}

/** 
 * @internal 
 *
 * Function used to perform on close invocation
 * 
 * @param conneciton The connection where all on close handlers will
 * be notified.
 */
void __vortex_connection_invoke_on_close (VortexConnection * connection, gboolean is_full)
{
	GList                       * list_aux;
	GList                       * list_aux2;
	VortexConnectionOnClose       on_close_handler;
	VortexConnectionOnCloseFull   on_close_full_handler;


	/* invoke full */
	if (is_full) {
		/* iterate over all full handlers and invoke them */
		list_aux  = g_list_first (connection->on_close_full);
		list_aux2 = g_list_first (connection->on_close_full_data);
		for (; list_aux; list_aux = g_list_next (list_aux)) {
			/* get a reference to the handler */
			on_close_full_handler = list_aux->data;

			/* invoke */
			on_close_full_handler (connection, list_aux2->data);

			/* update indexes */
			list_aux2 = g_list_next (list_aux2);
		} /* for */
	} else {
		/* iterate over all full handlers and invoke them */
		list_aux  = g_list_first (connection->on_close);
		for (; list_aux; list_aux = g_list_next (list_aux)) {
			/* get a reference to the handler */
			on_close_handler = list_aux->data;

			/* invoke */
			on_close_handler (connection);
		} /* for */
	} /* if */

	return;
} 

/** 
 * @internal
 * @brief Sets the given connection the not connected status.
 *
 * This internal vortex function allows library to set connection
 * status to false for a given connection. This function should not be
 * used by vortex library consumers.
 *
 * This function is callable over and over again on the same
 * connection. The first time the function is called sets to FALSE
 * connection state and its error message. It also close the BEEP
 * session and sets its socket to -1 in order to make it easily to
 * recognize for other functions.
 *
 * It doesn't make a connection unref. This actually done by the
 * Vortex Library internal process once they detect that the given
 * connection is not connected. 
 *
 * NOTE: next calls to function will lose error message.
 * 
 * @param connection the connection to set as.
 * @param message the new message to set.
 */
void           __vortex_connection_set_not_connected (VortexConnection * connection, gchar * message)
{
	g_return_if_fail (connection);
	g_return_if_fail (message);
	
	/* set connection status to false if weren't */
	g_mutex_lock (connection->op_mutex);

	if (connection->is_connected) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "flagging the connection as non-connected");
		connection->is_connected = FALSE;

		/* unlock now the op mutex is not blocked */
		g_mutex_unlock (connection->op_mutex);

		/* renew the message */
		if (connection->message)
			g_free (connection->message);
		connection->message = g_strdup (message);

		/* check to invoke on close handler */
		if (connection->on_close != NULL) {
			/* invoking on close handler */
			__vortex_connection_invoke_on_close (connection, FALSE);
		}

		/* check for the close handler full definition */
		if (connection->on_close_full != NULL) {
			/* invokin on close handler full */ 
			__vortex_connection_invoke_on_close (connection, TRUE);
		}

		/* NOTE: the following code was performing a connection
		 * close, (close call plus shutdown) making the
		 * connection close at once. 
		 * 
		 * This could cause several race conditions because
		 * the BEEP peers, as demonstrated while working with
		 * other BEEP implementations.
		 *
		 * So, the approach taken by the Vortex is to flag the
		 * connection to be not connected without performing a
		 * connection close. Then, the connection close is
		 * performed at vortex_connection_free, which is
		 * called by the last vortex_connection_unref
		 * performed.
		 */
		
		/* close socket connection if weren't  */
		if (( connection->close_session) && (connection->session != -1)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "closing connection id=%d to %s:%s", 
			       connection->id,
			       connection->host, connection->port);
			shutdown (connection->session, SHUT_RDWR); 
			vortex_close_socket (connection->session);  
			connection->session      = -1;
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "closing session id=%d and set to be not connected",
			       connection->id);
 	        } 
		return;
	}

	g_mutex_unlock (connection->op_mutex);

	return;
}


/** 
 * @internal
 *
 * @brief The function which actually send the data
 *
 * This function helps the vortex writer to do a sending round over
 * this connection. Because there are several channels over the same
 * connection sending frames, this function ensure a package for every
 * channel is sent on every round. This allows to avoid channel or
 * session starvation. 
 *
 * This function is executed from \ref
 * vortex_connection_do_a_sending_round as a foreach channel for a
 * given connection.
 * 
 * @param key   
 * @param value a channel to check if packets to be sent are waiting
 * @param user_data 
 */
void  __vortex_connection_one_sending_round (gpointer key,
					     gpointer value,
					     gpointer user_data)
{
	/* the channel we are going to use for this sending round */
	VortexChannel    * channel   = value;
	gint             * messages;
	VortexWriterData * packet;
	VortexConnection * connection;
	gchar            * rpy_num;

	/* if channel msg queue is empty do nothing */
	if (vortex_channel_queue_is_empty (channel)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "no message queue waiting for channel=%d..",
		       vortex_channel_get_number (channel));
		return;
	}

	/* check connection status */
	connection = vortex_channel_get_connection (channel);
	if (!vortex_connection_is_ok (connection, FALSE)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to send message, connection is not ok");
		return;
	}
	
	/* it seems channel queue is not empty send next one
	 * message */
	packet = vortex_channel_queue_next_msg (channel);
	g_return_if_fail (packet);

	/* send the frame */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Sending message, size (%d) over connection id=%d:\n%s",
	       packet->the_size,  
	       vortex_connection_get_id (connection), 
	       packet->the_frame);

	if (! vortex_frame_send_raw (connection, packet->the_frame, packet->the_size)) {
		/* drop a log */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to send the frame: errno=(%d): %s", 
			    errno, g_strerror (errno));
		
		__vortex_connection_set_not_connected (connection, "unable to write frame");
	}

	/* increase message sent */
	messages = user_data;
	(* messages) ++;

	/* signal the message have been sent */
	if (packet->type == VORTEX_FRAME_TYPE_RPY) {

		/* unblock waiting thread for replies sent */
		vortex_channel_signal_reply_sent_on_close_blocked (channel);
		
		rpy_num = g_strdup_printf ("%d", packet->msg_no);
		vortex_channel_signal_rpy_sent (channel, rpy_num);
		g_free (rpy_num);
	}

	/* free packet resources */
	g_free (packet->the_frame);
	packet->the_frame = NULL;
	g_free (packet);
	packet            = NULL;
	

	return;
}

/**
 * @internal
 * @brief Helper function to Vortex Writer process
 * 
 * This function must be not used by vortex library consumer. This
 * function is mainly used by the vortex writer process to signal a
 * connection to do a sending round. 
 *
 * A sending round means to check every channel queue to see if have
 * pending frames to send. If have pending frames to be sent, the next
 * one to is sent. This tries to avoid a channel consuming all the
 * vortex connection bandwidth.
 * 
 * @param connection a connection where a sending round robin will
 * performed.
 * 
 * @return  it returns how message frame have been sent or 0 if
 * no message was sent.
 */
gint               vortex_connection_do_a_sending_round (VortexConnection * connection)
{
	gint messages = 0;
	g_return_val_if_fail (vortex_connection_is_ok (connection, FALSE), 0);

	/* foreach channel, check if we have something to send */
	vortex_hash_foreach (connection->channels, 
			     __vortex_connection_one_sending_round, &messages);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "message sent for this connection id=%d: %d", 
	       vortex_connection_get_id (connection),
	       messages);

	return messages;
}

/** 
 * @brief Sets user defined data associated with the given connection.
 *
 * This function allows to store arbitrary data inside the given
 * connection. This data is store in a hash string indexed. 
 *
 * If value is NULL, this will be considered as a removing petition
 * for the given key.
 * 
 * See also \ref vortex_connection_set_data_full function.
 *
 * @param connection the connection to set data.
 * @param key the string key indexing the data.
 * @param value the value to be stored.
 */
void               vortex_connection_set_data               (VortexConnection * connection,
							     gchar            * key,
							     gpointer           value)
{
	/* use the full version so all source code is supported in one
	 * function. */
	vortex_connection_set_data_full (connection, key, value, NULL, NULL);
	return;
}

/** 
 * @brief Allows to store user space data into the connection like
 * \ref vortex_connection_set_data does but configuring functions to
 * be used while unreferring data stored.
 *
 * While storing user defined data into the connection it could be
 * necessary to also define destroy functions for the value stored and
 * the key stored. This allows to not worry about to free those data
 * (including the key) once the connection is dropped.
 *
 * This function allows to store data into the given connection
 * defining destroy functions for the key and the value stored, per item.
 * 
 * \code
 * [...]
 * void destroy_data_1 (gpointer data) 
 * {
 *     // perform a memory deallocation for data1
 * }
 * 
 * void destroy_data_2 (gpointer data) 
 * {
 *     // perform a memory deallocation for data2
 * }
 * [...]
 * // store data 1 providing a destroy value function
 * vortex_connection_set_data_full (connection, "some:data:1", 
 *                                  data_1, NULL, destroy_data_1);
 *
 * // store data 2 providing a destroy value function
 * vortex_connection_set_data_full (connection, "some:data:2",
 *                                  data_2, NULL, destroy_data_2);
 * [...]
 * \endcode
 *
 * @param connection    The connection where the data will be stored.
 * @param key           The unique string key value.
 * @param value         The value to be stored associated to the given key.
 * @param key_destroy   An optional key destroy function used to destroy (deallocate) memory used by the key.
 * @param value_destroy An optional value destroy function used to destroy (deallocate) memory used by the value.
 */
void                vortex_connection_set_data_full          (VortexConnection * connection,
							      gchar            * key,
							      gpointer           value,
							      GDestroyNotify     key_destroy,
							      GDestroyNotify     value_destroy)
{
	g_return_if_fail (connection);
	g_return_if_fail (key);

	/* check if the value is not null. It it is null, remove the
	 * value. */
	if (value == NULL) {
		vortex_hash_remove (connection->data, key);
		return;
	}

	/* store the data selected replacing previous one */
	vortex_hash_replace (connection->data, key,
			     __vortex_connection_data_create (key, key_destroy, value, value_destroy));
	
	/* return from setting the value */
	return;
}


/** 
 * @brief Allows to activate TLS profile automatic negotiation for every connection created.
 * 
 * Once a user application is developed using Vortex Library it could
 * be interesting to instruct Vortex Library to automatically
 * negotiate the TLS profile for every connection created. This will
 * make that every call to \ref vortex_connection_new will return not
 * only an instance already connected but also with the TLS profile
 * already activated.
 * 
 * This allows to take advantage about source code developed to create
 * and wait for a \ref VortexConnection to be created rather than
 * having two steps at the user space: first create the connection and
 * the TLS-ficate it with \ref vortex_tls_start_negociation.
 *
 * The function allows to specify the optional serverName value to be
 * used when \ref vortex_tls_start_negociation is called. The values
 * set on this function will make effect to all connections created.
 * 
 * Once a \ref VortexConnection "connection" is created, the TLS
 * profile negotiation could fail. This is because the remote peer could be not
 * accepting TLS request, or the serverName request is not accepted or
 * any other issue. 
 * 
 * This could be a security problem because there is no difference from
 * using a \ref VortexConnection with TLS profile activated from other
 * one without it. This could cause user application to start using a
 * connection that is successfully connected but the TLS profile have
 * actually failed, sending and receiving text in plain mode.
 * 
 * The parameter <b>allow_tls_failures</b> allows to configure what is
 * the default action is to be taken on TLS failures. By default, if
 * TLS profile negotiation fails, the connection is closed, returning that the TLS
 * profile have failed. 
 * 
 * Using a TRUE value allows to still keep on working even if the TLS
 * profile negotiation have failed.
 *
 * By default, Vortex Library have auto TLS feature disabled.
 * 
 * <i><b>NOTE:</b> If current Vortex Library doesn't have built-in
 * support for TLS profile, automatic TLS profile negotiation will
 * always fails. This means that setting <b>allow_tls_failures </b>
 * to FALSE will cause Vortex Library client peer to always fail to
 * create new connections.</i>
 * 
 * @param enabled TRUE to activate the automatic TLS profile
 * negotiation for every connection created, FALSE to disable it.
 *
 * @param allow_tls_failures Configure how to handle errors produced
 * while activating automatic TLS negotiation.
 *
 * @param serverName The server name value to be passed in to \ref
 * vortex_tls_start_negociation. If the received is not NULL the
 * function will perform a local copy
 */
void                vortex_connection_set_auto_tls           (gboolean   enabled,
							      gboolean   allow_tls_failures,
							      gchar    * serverName)
{
	/* save boolean values */
	__vortex_connection_auto_tls                = enabled;
	__vortex_connection_auto_tls_allow_failures = allow_tls_failures;

	/* unref previous values */
	if (__vortex_connection_auto_tls_server_name != NULL)
		g_free (__vortex_connection_auto_tls_server_name);

	/* store new value */
	__vortex_connection_auto_tls_server_name    = (serverName != NULL) ? g_strdup (serverName) : NULL;
	
	return;
}

/** 
 * @brief Gets stored value indexed by the given key inside the given connection.
 *
 * Returns value indexed by key inside connection. This function does
 * not retrieve data from remote peer, it just returns stored data
 * using \ref vortex_connection_set_data or \ref vortex_connection_set_data_full.
 * 
 * @param connection the connection where the value will be looked up.
 * @param key the key to look up.
 * 
 * @return the value or NULL if fails.
 */
gpointer           vortex_connection_get_data               (VortexConnection * connection,
							     gchar            * key)
{
	g_return_val_if_fail (connection, NULL);
	g_return_val_if_fail (key,        NULL);

	return __vortex_connection_data_get (vortex_hash_lookup (connection->data, key));
}

/** 
 * @brief Returns the channel pool identified by <i>pool_id</i>.
 * 
 * @param connection the connection where the channel pool is found.
 * @param pool_id the channel pool id to look up.
 * 
 * @return the channel pool or NULL if fails.
 */
VortexChannelPool * vortex_connection_get_channel_pool       (VortexConnection * connection,
							      gint               pool_id)
{
	VortexChannelPool * pool = NULL;

	g_return_val_if_fail (connection,  NULL);
	g_return_val_if_fail (pool_id > 0, NULL);

	pool = vortex_hash_lookup (connection->channel_pools, GINT_TO_POINTER (pool_id));

	return pool;
}


/** 
 * @internal
 * @brief Locks the channel pool mutex for the given connection.
 *
 * This function is for internal Vortex purposes. This function is
 * actually used by the vortex_channel_pool module to lock the
 * channel_pool_mutex this connection have while operating with
 * channel pool over this connection.
 * 
 * @param connection the connection to lock.
 */
void                vortex_connection_lock_channel_pool      (VortexConnection * connection)
{
	g_return_if_fail (connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "locking pool channel..");
	g_mutex_lock (connection->channel_pool_mutex);
	
	return;
}

/** 
 * @internal
 * @brief Unlocks the channel pool mutex for the given connection.
 *
 * This function is for internal vortex purposes. This function is
 * actually used by the vortex_channel_pool module to unlock the
 * channel_pool_mutex this connection have while operating with
 * channel pool over this connection.
 * 
 * @param connection the connection to unlock.
 */
void                vortex_connection_unlock_channel_pool    (VortexConnection * connection)
{
	g_return_if_fail (connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unlocking pool channel..");

	g_mutex_unlock (connection->channel_pool_mutex);

	return;
}

/** 
 * @internal
 * @brief Return the next channel pool identifier.
 *
 * This function is for internal vortex purposes. This function is
 * actually used by the vortex_channel_pool module to get the next
 * channel pool id to be used.
 * 
 * @param connection the connection where the channel pool will be created.
 * 
 * @return the next id to use or -1 if fails
 */
gint                vortex_connection_next_channel_pool_id   (VortexConnection * connection)
{
	g_return_val_if_fail (connection, -1);

	connection->next_channel_pool_id++;
	return connection->next_channel_pool_id;
}

/** 
 * @internal
 * @brief Adds a new channel pool on the given connection.
 * 
 * This function is for internal vortex purposes. This function allows
 * vortex_channel_pool module to add new channel pool created over the
 * given connection.  
 *
 * @param connection the connection to operate.
 * @param pool the channel pool to add.
 */
void                vortex_connection_add_channel_pool       (VortexConnection  * connection,
							      VortexChannelPool * pool)
{
	g_return_if_fail (connection);
	g_return_if_fail (pool);

	vortex_hash_replace (connection->channel_pools, 
			     GINT_TO_POINTER (vortex_channel_pool_get_id (pool)),
			     pool);

	return;
}

/** 
 * @brief
 * @internal Removes the given channel pool on the given connection.
 *
 * This function is for internal vortex purposes. This function allows
 * vortex_channel_pool module to remove channel pools from the given
 * connection. 
 *
 * @param connection the connection where the pool will be removed.
 * @param pool the pool to remove.
 */
void                vortex_connection_remove_channel_pool    (VortexConnection  * connection,
							      VortexChannelPool * pool)
{
	g_return_if_fail (connection);
	g_return_if_fail (pool);

	vortex_hash_remove (connection->channel_pools, 
			    GINT_TO_POINTER (vortex_channel_pool_get_id (pool)));

	return;
}


/** 
 * @internal
 * @brief Support function for \ref vortex_connection_get_pending_msgs
 *
 * This function adds the message pending to be send over a given
 * channel.
 * 
 * @param key 
 * @param value the channel to retrieve how many messages are waiting
 * @param user_data 
 */
void __vortex_connection_get_pending_msgs (gpointer key, gpointer value, gpointer user_data)
{
	VortexChannel * channel  = value;
	gint          * messages = user_data;
	gint            msgs     = vortex_channel_queue_length (channel);
	
	(* messages ) = (* messages) + msgs;

	return;
}

/** 
 * @brief Allows to get current frames waiting to be sent on the given connection.
 *
 * This function will iterate over all channels the connection have
 * checking if there are frames pending to be sent. The cumulative
 * result for this iteration will be returned as the pending message
 * to be sent over this connection or session.
 *
 * This function is actually used by the vortex writer to be able to
 * normalize its pending task to be processed situation while
 * unreferring connections.
 * 
 * @param connection a connection to know how many message are pending
 * to be sent.
 * 
 * @return the number or message pending to be sent.
 */
gint                vortex_connection_get_pending_msgs       (VortexConnection * connection)
{
	gint messages = 0;

	vortex_hash_foreach (connection->channels, 
			     __vortex_connection_get_pending_msgs, &messages);	
	return messages;
}

/** 
 * @brief Allows to get current connection role.
 * 
 * @param connection The VortexConnection to get the current role from.
 * 
 * @return Current role represented by \ref VortexPeerRole. If the
 * given connection is not connected, using \ref
 * vortex_connection_is_ok, the function will return \ref VortexRoleUnknown.
 */
VortexPeerRole      vortex_connection_get_role               (VortexConnection * connection)
{
	if (!vortex_connection_is_ok (connection, FALSE))
		return VortexRoleUnknown;

	return connection->role;
}

/** 
 * @brief Returns current features requested by the remote peer this
 * connection is linked to.
 *
 * Features are a way for the remote peer to advertise some requested
 * feature. This is used/defined for each profile. Not all profiles
 * makes use of the feature optional attribute. Value get from this
 * function is the feature set by the remote peer. In case the remote
 * peer is a Vortex Library one, this value is set using \ref
 * vortex_greetings_set_features.
 * 
 * Features are sent at greetings time, which is the initial handshake
 * done between BEEP peers and it is previous to any channel
 * created. Features could be considered globally to the connection.
 *
 * You should not confuse this function with \ref
 * vortex_greetings_get_features. That function returns current status
 * of features that will be sent to the remote peers while current
 * host creates a new session.
 *
 * @param connection The connection where the greetings were received,
 * signaling the features request.
 * 
 * @return Features requested by the remote peer or NULL if none
 * features was defined. This function also return NULL if the
 * connection is not connected (verified with \ref
 * vortex_connection_is_ok).
 */
gchar             * vortex_connection_get_features           (VortexConnection * connection)
{
	if (!vortex_connection_is_ok (connection, FALSE))
		return NULL;
	return connection->features;
}

/** 
 * @brief Returns current localize requested by the remote peer this
 * connection is linked to.
 *
 * Localize value is a way for the remote peer to advertise preferred
 * language while generating textual diagnostics. If no localize is
 * defined by the remote peer, that is, the value returned by this
 * function is NULL, then you can generate the default language
 * diagnostic.
 *
 * You should not confuse this function with \ref
 * vortex_greetings_get_localize. That function is used to get
 * preferred localize to be sent while connecting to a remote peer.
 * 
 * @param connection The connection where the greetings were received,
 * signaling the localize requested.
 * 
 * @return Localize requested by the remote peer or NULL if no
 * localize was defined. This function also return NULL if the
 * connection is not connected (verified with \ref
 * vortex_connection_is_ok).
 */
gchar             * vortex_connection_get_localize           (VortexConnection * connection)
{
	if (!vortex_connection_is_ok (connection, FALSE))
		return NULL;
	return connection->localize;
}

/** 
 * @brief Allows to configure the send handler used to actually
 * perform sending operations over the underlying connection.
 *
 * 
 * @param connection The connection where the send handler will be set.
 * @param send_handler The send handler to be set.
 * 
 * @return Returns the previous send handler defined or NULL if fails.
 */
VortexSendHandler      vortex_connection_set_send_handler    (VortexConnection * connection,
							      VortexSendHandler  send_handler)
{
	VortexSendHandler previous_handler;

	g_return_val_if_fail (connection,   NULL);
	g_return_val_if_fail (send_handler, NULL);

	/* save previous handler defined */
	previous_handler = connection->send;

	/* set the new send handler to be used. */
	connection->send = send_handler;

	/* returns previous handler */
	return previous_handler;
 
	
}

/** 
 * @brief Allows to configure receive handler use to actually receive
 * data from remote peer. 
 * 
 * @param connection The connection where the receive handler will be set.
 * @param receive_handler The receive handler to be set.
 * 
 * @return Returns current receive handler set or NULL if it fails.
 */
VortexReceiveHandler   vortex_connection_set_receive_handler (VortexConnection     * connection,
							      VortexReceiveHandler   receive_handler)
{
	VortexReceiveHandler previous_handler;

	g_return_val_if_fail (connection,      NULL);
	g_return_val_if_fail (receive_handler, NULL);

	/* save previous handler defined */
	previous_handler    = connection->receive;

	/* set the new send handler to be used. */
	connection->receive = receive_handler;

	/* returns previous handler */
	return previous_handler;
}

/** 
 * @brief Set default IO handlers to be used while sending and
 * receiving data for the given connection.
 * 
 * Vortex Library allows to configure which are the function to be
 * used while sending and receiving data from the underlying
 * transport. This functions are usually not required API consumers.
 *
 * However, while implementing tuning profiles that needs to modify
 * how data is received and sent, the following function are used:
 * 
 *  - \ref vortex_connection_set_receive_handler
 *  - \ref vortex_connection_set_send_handler 
 *
 * This function allows to returns to the default handlers used by the
 * Vortex Library to perform these operations. This useful to
 * implement handler restoring when the tuning profile have failed.
 * 
 * @param connection The connection where the handler restoring will
 * take place.
 *
 */
void                   vortex_connection_set_default_io_handler (VortexConnection * connection)
{
	g_return_if_fail (connection);

	/* set default send and receive handlers */
	connection->send       = vortex_connection_default_send;
	connection->receive    = vortex_connection_default_receive;
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "restoring default IO handlers for connection id=%d", 
	       connection->id);

	return;
}

/** 
 * @brief Allows to set a new on close handler to be executed only
 * once the connection is being closed.
 *
 * The handler provided on this function is called once the connection
 * provided is closed. This is useful to detect connection broken or "broken pipe".
 *
 * If you require to set a user pointer to be received by the handler,
 * you can use \ref vortex_connection_set_on_close_full.
 * 
 * @param connection The connection being closed
 * @param on_close_handler The handler that will be executed.
 * 
 * @return In the past, the function was returning the previous
 * handler configured. However, because the function now support
 * receiving several on close handlers to be invoked, this have
 * changed. The function always return NULL.
 */
VortexConnectionOnClose vortex_connection_set_on_close       (VortexConnection * connection,
							      VortexConnectionOnClose on_close_handler)
{

	g_return_val_if_fail (connection,      NULL);
	g_return_val_if_fail (on_close_handler, NULL);

	/* lock until done */
	g_mutex_lock (connection->handlers_mutex);

	/* save previous handler defined */
	connection->on_close = g_list_append (connection->on_close, on_close_handler);

	/* unlock now the item is removed */
	g_mutex_unlock (connection->handlers_mutex);

	/* returns previous handler */
	return NULL;
}

/** 
 * @brief Extended version for \ref vortex_connection_set_on_close
 * handler which also support receiving a user data pointer.
 *
 * See \ref vortex_connection_set_on_close for more details. This
 * function could be called several times to install several handlers.
 *
 * Once a handler is installed, you can use \ref
 * vortex_connection_remove_on_close_full to uninstall the handler if
 * it is required to avoid getting more notifications.
 * 
 * @param connection The connection that is required to get close
 * notifications.
 *
 * @param on_close_handler The handler to be executed once the event
 * is produced.
 *
 * @param data User defined data to be passed to the handler.
 * 
 * @return In the past, the function was implemented to support only
 * one handler. This was changed to support several handlers to be
 * notified. Now, the function doesn't return the previous
 * function. Always returns NULL.
 */
VortexConnectionOnCloseFull vortex_connection_set_on_close_full  (VortexConnection * connection,
								  VortexConnectionOnCloseFull on_close_handler,
								  gpointer data)
{
	g_return_val_if_fail (connection,      NULL);
	g_return_val_if_fail (on_close_handler, NULL);

	/* lock during the operation */
	g_mutex_lock (connection->handlers_mutex);

	/* save handler defined */
	connection->on_close_full      = g_list_append (connection->on_close_full, on_close_handler);
	
	/* save data associated */
	connection->on_close_full_data = g_list_append (connection->on_close_full_data, data);

	/* unlock now it is done */
	g_mutex_unlock (connection->handlers_mutex);

	/* returns previous handler */
	return NULL;
}

/** 
 * @brief Allows to uninstall a particular handler installed to get
 * notifications about the connection close.
 *
 * If the handler is found, it will be uninstalled and the data
 * associated to the handler will be returned.
 * 
 * @param connection The connection where the uninstall operation will
 * be performed.
 *
 * @param on_close_handler The handler to uninstall.
 * 
 * @return The user data associated to the handler, configured at \ref
 * vortex_connection_set_on_close_full
 */
gpointer vortex_connection_remove_on_close_full (VortexConnection              * connection, 
						 VortexConnectionOnCloseFull     on_close_handler)
{
	GList    * list_aux;
	GList    * list_aux2;
	gpointer   data;

	g_return_val_if_fail (connection, NULL);
	g_return_val_if_fail (on_close_handler, NULL);

	/* look during the operation */
	g_mutex_lock (connection->handlers_mutex);

	/* look up for the handler */
	list_aux  = g_list_first (connection->on_close_full);
	list_aux2 = g_list_first (connection->on_close_full_data);
	
	for (; list_aux; list_aux = g_list_next (list_aux->data)) {
		if (list_aux->data == on_close_handler) {
			/* handler found remove it */
			connection->on_close_full      = g_list_remove (connection->on_close_full, 
									on_close_handler);
			
			/* get a reference to the data before removing it */
			data                           = list_aux2->data;
			connection->on_close_full_data = g_list_remove (connection->on_close_full_data,
									data);

			/* now finished, unlock */
			g_mutex_unlock (connection->handlers_mutex);

			/* return stored data reference */
			return data;
		}

		/* update the index */
		list_aux2 = g_list_next (list_aux2);
	}

	/* unlock */
	g_mutex_unlock (connection->handlers_mutex);

	return NULL;
}

/** 
 * @internal
 * @brief Allows to invoke current receive handler defined by \ref VortexReceiveHandler.
 * 
 * This function is actually no useful for Vortex Library consumer. It
 * is used by the library to actually perform the invocation in a
 * transparent way.
 *
 * @param connection The connection where the receive handler will be invoked.
 * @param buffer     The buffer to be received.
 * @param buffer_len The buffe size to be received.
 * 
 * @return How many data was actually received or -1 if it fails.
 */
gint                vortex_connection_invoke_receive         (VortexConnection * connection,
							      gchar            * buffer,
							      gint               buffer_len)
{
	g_return_val_if_fail (connection,          -1);
	g_return_val_if_fail (buffer,              -1);
	g_return_val_if_fail (connection->receive, -1); 

	return connection->receive (connection, buffer, buffer_len);
}

/** 
 * @internal
 * @brief Allows to invoke current send handler defined by \ref VortexSendHandler.
 * 
 * This function is actually not useful for Vortex Library
 * consumers. It is used by the library to perform the invocation of the send handler.
 * 
 * @param connection The connection where the invocation of the send handler will be performed.
 * @param buffer     The buffer data to be sent.
 * @param buffer_len The buffer data size.
 * 
 * @return How many data was actually sent or -1 if it fails.
 */
gint                vortex_connection_invoke_send            (VortexConnection * connection,
							      gchar            * buffer,
							      gint               buffer_len)
{
	g_return_val_if_fail (connection,       -1);
	g_return_val_if_fail (buffer,           -1);
	g_return_val_if_fail (connection->send, -1);

	return connection->send (connection, buffer, buffer_len);
}

/** 
 * @brief Allows to disable sanity socket check, by default enabled.
 *
 * While allocating underlying socket descriptors, at the connection
 * creation using (<b>socket</b> system call) \ref
 * vortex_connection_new, it could happen that the OS assign a socket
 * descriptor used for standard task like: stdout(0), stdin(1),
 * stderr(2).
 * 
 * This happens when the user app issue a <b>close() </b> call over
 * the previous standard descriptors (<b>0,1,2</b>) causing vortex to
 * allocate a reserved descriptor with some funny behaviors.
 * 
 * Starting from previous context, any user app call issuing a console
 * print will cause to automatically send to the remote site the
 * message printed, bypassing all vortex mechanisms.
 * 
 * Obviously, this is an odd situation and it is not desirable. By
 * default, Vortex Library includes a sanity check to just reject
 * creating a Vortex Connection with an underlying socket descriptor
 * which could cause applications problems.
 *
 * You can disable this sanity check using this function.
 * 
 */
void                vortex_connection_sanity_socket_check (gboolean enable)
{
	__vortex_connection_enable_sanity_check  = enable;
	return;
}


/** 
 * @brief Allows to parse greetings reply received and prepare the
 * connection to be become usable.
 * 
 * This function allows to parse a greetings reply received and set
 * values received on that reply into the given connection. Values
 * received are: profiles supported and localize and features
 * attributes.
 * 
 * If the previous process is finished correctly, the connection is
 * registered into the vortex reader so the connection could start to
 * receive data from remote peer. 
 * 
 * If the function fails while parsing greetings reply, the connection
 * received will be set to be closed (that is, not usable). You should
 * call to \ref vortex_connection_unref if such situation
 * happens. Because this function is to be used while implementing a
 * profile that requires a tuning reset the \ref
 * vortex_connection_close function should not be used for close the
 * connection. 
 * 
 * This function also deallocate the frame received. 
 * 
 * @param connection The connection where the greetings status
 * checking will be performed.
 *
 * @param frame      The frame containing a greeting reply
 * 
 * @return TRUE if the process have finished properly, otherwise FALSE
 * is returned.
 */
gboolean vortex_connection_parse_greetings_and_enable (VortexConnection * connection, 
						       VortexFrame      * frame)
{
	
	/* process frame response */
	if (frame != NULL) {
		/* now, we have to read remote site supported
		 * profiles */
		if (!__vortex_connection_parse_greetings (connection, frame)) {
			/* parse ok, free frame and establish new
			 * message */
			vortex_frame_free (frame);

			/* something wrong have happened while parsing
			 * XML greetings */
			return FALSE;
		}
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "greetings parsed..");
		
		/* parse ok, free frame and establish new message */
		vortex_frame_free (frame);
		connection->message = g_strdup ("session established and ready");
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "new connection created to %s:%s", connection->host, connection->port);
		
		/* now, every initial message have been sent, we need
		 * to send this to the reader manager */
		vortex_reader_watch_connection (connection);
		return TRUE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received a null frame (null reply) from remote side when expected a greetings reply, closing session");
	__vortex_connection_set_not_connected (connection, "received a null frame (null reply) from remote side when expected a greetings reply, closing session");
	return FALSE;
}


/** 
 * @brief Allows to configure a handler to be executed before any
 * operations is applied inside the Vortex Reader process.
 * 
 * @param connection The connection where the pre read will be executed.
 *
 * @param pre_accept_handler The handler to be executed. If NULL is
 * used, the handler is no longer executed.
 */
void                vortex_connection_set_preread_handler        (VortexConnection * connection, 
								  VortexConnectionOnPreRead pre_accept_handler)
{
	g_return_if_fail (connection);

	vortex_connection_set_data (connection, "connection:pre-read-enabled", pre_accept_handler);
	
	return;
}

/** 
 * @internal
 * @brief Allows to set current TLS status for the given connection.
 * 
 * This function is not for Vortex Library API consumers.
 * 
 * @param connection The connection to set TLS-fication status.
 * @param status The status to set.
 */
void                vortex_connection_set_tlsfication_status     (VortexConnection * connection,
								  gboolean           status)
{
	/* flag this connection to be already TLS-ficated */
	vortex_connection_set_data (connection, "tls-fication:status", GINT_TO_POINTER (status));
	return;
}

/** 
 * @brief Allows to get current status for TLS activation on the given connection.
 *
 * Every \ref VortexConnection instance used inside the Vortex Library
 * could be already running under the TLS profile. This function
 * allows to get current TLS activation status. 
 * 
 * @param connection The connection to check for TLS status.
 * 
 * @return TRUE if TLS status is already activated, otherwise FALSE is returned.
 */
gboolean            vortex_connection_is_tlsficated              (VortexConnection * connection)
{
	return (GPOINTER_TO_INT (vortex_connection_get_data (connection, "tls-fication:status")));
}


/** 
 * @brief Allows to check if there are an pre read handler defined on the given connection.
 * 
 * @param connection The connection to check.
 * 
 * @return TRUE if the pre-read handler is defined, otherwise, FALSE is returned
 */
gboolean            vortex_connection_is_defined_preread_handler (VortexConnection * connection)
{
	return (vortex_connection_get_data (connection, "connection:pre-read-enabled") != NULL);
}

/** 
 * @brief Invokes the prer-read handler defined on the given
 * connection.
 * 
 * @param connection The connection where the preread handler will be
 * invoked
 * 
 * @return 
 */
void            vortex_connection_invoke_preread_handler     (VortexConnection * connection)
{
	VortexConnectionOnPreRead pre_accept_handler = vortex_connection_get_data (connection, "connection:pre-read-enabled");

	if (pre_accept_handler != NULL)
		pre_accept_handler (connection);
	return;
}


/* @} */
