/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>

#define LOG_DOMAIN "vortex-frame-factory"

/** 
 * @internal
 * @brief Internal Note
 *
 * While writing \r\n CR+LF you have to use \\x0D\\x0A
 *  \\n is portable mapped to \\x0A
 *  \\r is portable mapped to \\x0D
 * 
 * More information about writing portable CR LF bytes on
 * wikipedia. Fabulous article at:
 *
 *     http://en.wikipedia.org/wiki/CRLF
 */


struct _VortexFrame {
	/** 
	 * Frame unique identifier. Every frame read have a different
	 * frame id. This is used to track down frames that are
	 * allocated and deallocated.
	 */
	gint              id;
	
	/* frame header definition */
	VortexFrameType   type;
	gint32            channel;
	gint32            msgno;
	gboolean          more;
	gchar             more_char;
	guint32           seqno;
	gint32            size;
	gint32            ansno;
	
	/* content type */
	gchar           * content_type;

	/* transfer encoding */
	gchar           * transfer_encoding;

	/* the size of the mime headers, including last \x0D\x0A\x0D\x0A */
	gint              mime_headers_size;
	
	/* the payload message */
	gchar           * payload;
 };

/** 
 * @internal
 *
 * Internal variables (frame_id and frame_id_mutex) to make frame
 * identification for the on going process. This allows to check if
 * two frames are equal or to track which frames are not properly
 * released by the Vortex Library.
 *
 * Frames are generated calling to __vortex_frame_get_next_id. Thus,
 * every frame created while the running process is alive have a
 * different frame unique identifier. 
 */
glong                frame_id                               = 1;
static GStaticMutex  frame_id_mutex                         = G_STATIC_MUTEX_INIT;

/** 
 * @internal
 *
 * @brief Support function for frame identificators.  
 *
 * This is used to generate and return the next frame identifier, an
 * unique integer value to track frames created.
 *
 * @return Next frame identifier available.
 */
gint __vortex_frame_get_next_id (gchar * from)
{
	gint result;
	g_static_mutex_lock (&frame_id_mutex);
	
	result = frame_id;
	frame_id++;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "from='%s' allocating a new frame id=%d", from, result);

	g_static_mutex_unlock (&frame_id_mutex);

	return result;
}

/**
 * \defgroup vortex_frame Vortex Frame Factory: Function to manipulate frames inside Vortex Library
 */

/**
 * \addtogroup vortex_frame
 * @{
 */


/**
 * @brief Returns the frame in plain text form from a VortexFrame object.
 * 
 * Returns actual frame message (the complete string) <i>frame</i> represents. The return value
 * must be freed when no longed is needed.
 *
 * @param frame the frame to get as raw text.
 * 
 * @return the raw frame or NULL if fails
 **/
gchar *       vortex_frame_get_raw_frame         (VortexFrame * frame)
{
	g_return_val_if_fail (frame, NULL);

	return vortex_frame_build_up_from_params_s (frame->type,
						    frame->channel,
						    frame->msgno,
						    frame->more,
						    frame->seqno,
						    frame->size,
						    frame->ansno,
						    frame->content_type,
						    frame->transfer_encoding,
						    frame->payload,
						    NULL);
}

/** 
 * @brief Allows to create a new SEQ frame.
 *
 * SEQ frame are especial to BEEP because they allows to control how
 * flow control is handled based on channel window size. 
 * 
 * @param channel_num The channel number where the channel will 
 *
 * @param ackno The sequence number expected to be received at the
 * time the SEQ frame was generated.
 *
 * @param window_size The window size that the SEQ frame will
 * acknowledge.
 * 
 * @return A newly allocated raw SEQ frame. Returned value should be
 * deallocated using g_free.
 */
gchar * vortex_frame_seq_build_up_from_params (gint channel_num,
					       gint ackno,
					       gint window_size)
{
	/* just build and return the string */
	return axl_stream_strdup_printf ("SEQ %d %d %d\x0D\x0A", channel_num, ackno, window_size);
}

/** 
 * @brief Builds a frame using the given parameters.
 *
 * This function should not be useful for Vortex Library consumer
 * because all frames received and sent are actually builded by Vortex
 * Library.
 *
 * If you need to get the size of the frame returned at the same time
 * it is built, you should call to \ref
 * vortex_frame_build_up_from_params_s. This function implements the
 * same operation like this function but also returning the frame
 * size. 
 * 
 * <b>NOTE: This function is considered to be deprecated because there
 * is no accurate way to get current size for the frame returned.</b>
 * 
 * Using binary protocols that includes 0x0 as a valid value for its
 * charset could make to produce buggy results from using functions
 * such strlen to calculate current frame size returned. This is
 * because this functions looks for the next 0x0 value found on the
 * string to delimit it, but, as stated before this value couldn't be
 * the last value. It is preferred to use \ref vortex_frame_build_up_from_params_s.
 *
 * @param type    Frame type.
 * @param channel The channel number for the frame.
 * @param msgno   The message number for the frame.
 * @param more    More flag status for the frame.
 * @param seqno   Sequence number for the frame
 * @param size    The frame payload size.
 * @param ansno   The answer number for the frame
 * @param payload The payload is going to have the frame.
 * 
 * @return A newly created frame that should be freed using g_free
 */
gchar * vortex_frame_build_up_from_params (VortexFrameType   type,
					   gint32            channel,
					   gint32            msgno,
					   gboolean          more,
					   guint32           seqno,
					   gint32            size,
					   gint32            ansno,
					   gchar  *          payload)
{
	return vortex_frame_build_up_from_params_s (type, channel, msgno, more, seqno, size, ansno, NULL, NULL, payload, NULL);
}

/** 
 * @brief Creates a new frame from using the given data and returning
 * current frame size resulting from the operation.
 *
 * This function replaces \ref vortex_frame_build_up_from_params and
 * it is considered to be more accurate and secure.
 *
 * @param type              Frame type.
 *
 * @param channel           The channel number for the frame.
 *
 * @param msgno             The message number for the frame.
 *
 * @param more              More flag status for the frame.
 *
 * @param seqno             Sequence number for the frame
 *
 * @param size              The frame payload size.
 *
 * @param ansno             The answer number for the frame
 *
 * @param content_type Optional content type to be used for the
 * channel being sent. If no value is provided no content type will be
 * placed. If content type provided is also the default value then no
 * content type will be placed.
 *
 * @param transfer_encoding Optional content transfer encoding. If no
 * value is provided no content transfer type will be placed. If
 * content transfer type provided is also the default value then no
 * content type will be placed.
 *
 * @param payload           The payload is going to have the frame.
 *
 * @param frame_size An optional pointer to an integer value which
 * holds current frame size.
 *
 * @return A newly created frame that should be freed using g_free.
 */
gchar * vortex_frame_build_up_from_params_s (VortexFrameType   type,
					     gint32            channel,
					     gint32            msgno,
					     gboolean          more,
					     guint32           seqno,
					     gint32            size,
					     gint32            ansno,
					     gchar  *          content_type,
					     gchar  *          transfer_encoding,
					     gchar  *          payload,
					     gint   *          frame_size)
{
	gchar    * message_type = NULL;
	gchar    * ansno_string = NULL;
	gchar    * value;
	gint       header_size;
	gint       header_length;
	gboolean   place_content_type;
	gboolean   place_transfer_encoding;

	/*
	 * According to the frame type build the initial frame header
	 * using BEEP prefix.
	 */
	switch (type) {
	case VORTEX_FRAME_TYPE_UNKNOWN:
		return NULL;
	case VORTEX_FRAME_TYPE_MSG:
		message_type = "MSG";
		break;
	case VORTEX_FRAME_TYPE_RPY:
		message_type = "RPY";
		break;
	case VORTEX_FRAME_TYPE_ANS:
		message_type = "ANS";
		break;
	case VORTEX_FRAME_TYPE_ERR:
		message_type = "ERR";
		break;
	case VORTEX_FRAME_TYPE_NUL:
		message_type = "NUL";
		break;
	case VORTEX_FRAME_TYPE_SEQ:
		/* use seq frame building function */
		return vortex_frame_seq_build_up_from_params (channel, seqno, size);
	default:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received a non-handled frame type");
		if (frame_size != NULL)
			(* frame_size ) = -1;
		return NULL;
	}

	/*
	 * Now, for the frame ANS, we have to build the ansno value to
	 * be included into the string to be built.
	 */
	if (type == VORTEX_FRAME_TYPE_ANS) 
		ansno_string = axl_stream_strdup_printf (" %d", ansno);

	/*
	 * check if we have to build content type and content transfer
	 * encoding headers ... */
	place_content_type      = (content_type      != NULL) && ! axl_stream_cmp (content_type, "application/octet-stream", 24);
	place_transfer_encoding = (transfer_encoding != NULL) && ! axl_stream_cmp (transfer_encoding, "binary", 6);

	/* payload size for the BEEP header */
	header_size = size;
	
	/* aggregate the content type, if defined, to the payload
	 * size */
	if (place_content_type)
		header_size += 16 + strlen (content_type);

	/* aggregate the content transfer type, if defined, to the
	 * payload size */
	if (place_transfer_encoding)
		header_size += 29 + strlen (transfer_encoding);

	/* aggregate two bytes more to the payload size of one of the
	 * previous entity headers were defined, representing \r\n */
	if (place_transfer_encoding || place_content_type)
		header_size += 2;

	/* build BEEP header */
	value  = axl_stream_strdup_printf_len (
		/* provide the printf message format */
		"%s %d %d %c %u %d%s\x0D\x0A%s%s%s%s%s%s%s",
		/* provide a reference to get the header length */
		&header_length,
		/* BEEP header values */
		message_type,                      
		channel,                           
		msgno,                             
		more ? '*' : '.',                  
		seqno,                             
		/* payload size, do not use "size" because it specify
		   the payload size without considering content type
		   and content transfer encoding. 
		  
		   "header_size" is not the BEEP frame header size but
		   the "size" value that goes inside the BEEP
		   header. */
		header_size,                       
		(ansno_string != NULL) ? ansno_string : "", /* ansno ? */
		/* content type values */
		place_content_type ? "Content-Type: "   : "",
		place_content_type ? content_type       : "",
		place_content_type ? "\x0D\x0A"         : "",
		/* content transfer encoding */
		place_transfer_encoding ? "Content-Transfer-Encoding: "   : "",
		place_transfer_encoding ? transfer_encoding               : "",
		place_transfer_encoding ? "\x0D\x0A"                      : "",
		/* place entity header trailing header if some of them is defined */
		( place_content_type || place_transfer_encoding) ? "\x0D\x0A" : "");

	/*
	 * Now build the frame according to previous BEEP header and
	 * payload size received.  Next memory allocation stands from
	 * counting header size plus payload size and then 6
	 * additional bytes: END\\x0D\\x0A which are 5 bytes ended by a
	 * 0x0 value meaning 1 byte more.
	 */

	/*
	 * Allocate a memory chunk to hold frame values, including
	 * header length and payload content. On this case we use
	 * "size" rather than "header_size". This is because
	 * header_length includes all size for the BEEP header
	 * (including Content-Type and Content-Transfer-Encoding
	 * entity headers) and "size" is the payload size. 
	 */
	value = g_realloc (value, header_length + size + 6);
	
	/* copy BEEP frame payload */
	memcpy (value + header_length, payload, size);

	/* copy BEEP frame trailing */
	memcpy (value + header_length + size, "END\x0D\x0A", 5);

	/* nullify the last char */
	value [header_length + size + 5] = 0;

	/* deallocate value used to create the ansno value */
	if (ansno_string != NULL)
		g_free (ansno_string);

	/*
	 * return calculated frame size, NOTE: that the returned size
	 * for constant value is 7 rather than 8 as stated on previous
	 * statement allocating memory, this is because we don't want
	 * to write to the socket a 0x0 value. 
	 */
	if (frame_size != NULL)
		(* frame_size) = header_length + 5 + size;

	return value;	
}

/** 
 * @brief Creates a new frame object (\ref VortexFrame) using the given data.
 *
 * Also check \ref vortex_frame_create_full for a full support on
 * creating new frames, also specifying content type and content
 * transfer encoding values.
 *
 * @param type    The frame type to build.
 * @param channel The channel number the frame have.
 * @param msgno   The message number the frame have. 
 * @param more    More flag configuration for the frame.
 * @param seqno   The sequence number for the frame to build.
 * @param size    The frame payload size.
 * @param ansno   The answer number the frame have.
 * @param payload The payload the frame hold.
 * 
 * @return A newly created \ref VortexFrame object that must be
 * unrefered using \ref vortex_frame_free when no longer needed.
 */
VortexFrame * vortex_frame_create               (VortexFrameType   type,
						 gint32            channel,
						 gint32            msgno,
						 gboolean          more,
						 guint32           seqno,
						 gint32            size,
						 gint32            ansno,
						 gchar           * payload)
{
	/* create a new frame using new_full interface and return
	 * it. */
	return vortex_frame_create_full (type, channel, msgno, more, seqno, size, ansno, 
					 NULL,     /* content type value */
					 NULL,     /* content transfer encoding value */
					 payload); /* frame payload */
}


/** 
 * @brief Creates a new frame as \ref vortex_frame_create but also
 * allowing to specify mime header content, that is, Content-Type and
 * Content-Transfer-Encoding values.
 *
 * This function works like \ref vortex_frame_create function but
 * allowing to specify every parameter used (possible) for a frame.
 * 
 * The function also performs a local copy for the content type ("Content-Type:" ) value  and
 * the transfer encoding ("Content-Transfer-Encoding:" )  value provided.
 * 
 * @param type    The frame type to build.
 * @param channel The channel number the frame have.
 * @param msgno   The message number the frame have. 
 * @param more    More flag configuration for the frame.
 * @param seqno   The sequence number for the frame to build.
 * @param size    The frame payload size.
 * @param ansno   The answer number the frame have.
 *
 * @param content_type The MIME Content-Type value to be used for this
 * frame. Optional value.
 *
 * @param transfer_encoding The MIME Content-Transfer-Encoding value
 * to be used for this frame. Optional value.
 *
 * @param payload The payload the frame hold.
 * 
 * @return A newly created \ref VortexFrame object that must be
 * unrefered using \ref vortex_frame_free when no longer needed.
 */
VortexFrame * vortex_frame_create_full          (VortexFrameType   type,
						 gint32            channel,
						 gint32            msgno,
						 gboolean          more,
						 guint32           seqno,
						 gint32            size,
						 gint32            ansno,
						 gchar           * content_type,
						 gchar           * transfer_encoding,
						 gchar           * payload)
{
	VortexFrame * result = g_new0 (VortexFrame, 1);
	
	result->id           = __vortex_frame_get_next_id ("create-full");
	result->type         = type;
	result->channel      = channel;
	result->msgno        = msgno;
	result->seqno        = seqno;
	result->size         = size;
	result->ansno        = ansno;

	/* copy the payload */
	if (size > 0 && payload != NULL) {
		result->payload           = g_new0 (gchar, size + 1);
		memcpy (result->payload, payload, size);
	}

	/* copy content type */
	if (content_type != NULL) {
		result->content_type      = g_strdup (content_type);
		result->mime_headers_size = 16 + strlen (content_type);
	}

	/* copy content transfer encoding */
	if (transfer_encoding != NULL) {
		result->transfer_encoding = g_strdup (transfer_encoding);
		result->mime_headers_size += 29 + strlen (transfer_encoding);
	}

	/* establish a mime header size */
	if (content_type != NULL || transfer_encoding != NULL)
		result->mime_headers_size += 2; /* last \x0D\x0A */
		
	return result;	
}


/** 
 * @brief Perform a copy from the received frame.
 * 
 * @param frame A frame to copy.
 * 
 * @return A newly allocated frame that must be unrefered when no
 * longer needed using \ref vortex_frame_free. NULL will be returned if the frame received is NULL.
 */
VortexFrame * vortex_frame_copy                 (VortexFrame      * frame)
{
	if (frame == NULL)
		return NULL;

	return vortex_frame_create (frame->type, frame->channel, frame->msgno,
				    frame->more, frame->seqno, frame->size, 
				    frame->ansno, frame->payload);
}

/** 
 * @internal
 * @brief reads n bytes from the connection.
 * 
 * @param connection the connection to read data.
 * @param buffer buffer to hold data.
 * @param maxlen 
 * 
 * @return how many bytes were read.
 */
gint         __vortex_frame_readn    (VortexConnection * connection, gchar * buffer, gint maxlen)
{
	gint    nread;
	gchar * error_msg;

 __vortex_frame_readn_keep_reading:
	/* clear buffer */
	memset (buffer, 0, maxlen * sizeof (gchar));
	if ((nread = vortex_connection_invoke_receive (connection, buffer, maxlen)) == VORTEX_SOCKET_ERROR) {
		if (errno == EWOULDBLOCK) {
			return 0;
		}
		if (errno == EINTR)
			goto __vortex_frame_readn_keep_reading;
		
		error_msg = vortex_errno_get_last_error ();
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to readn=%d, error was: %s",
		       maxlen, error_msg ? error_msg : "");
	}

	buffer[nread] = 0;
	return nread;
}

/**
 * @internal

 * 
 * Read the next line, byte by byte until it gets a \n or maxlen is reached. Some code errors
 * are used to manage exceptions (see return values)
 * 
 * @param connection
 * @param buffer
 * @param maxlen
 * 
 * @return  values returned by this function follows:
 *  0 - remote peer have closed the file
 * -1 - an error have happened while reading
 * -2 - could read because this connection is on non-blocking mode and there is no
 *  n - some data was read.
 * 
 **/
gint         __vortex_frame_readline (VortexConnection * connection, gchar * buffer, gint maxlen)
{
	gint  n, rc;
	gchar c, *ptr;
	gchar    * error_msg;

	/* clear the buffer received */
	memset (buffer, 0, maxlen * sizeof (gchar));

	/* read current next line */
	ptr = buffer;
	for (n = 1; n < maxlen; n++) {
	__vortex_frame_readline_again:
		if (( rc = vortex_connection_invoke_receive (connection, &c, 1)) == 1) {
			*ptr++ = c;
			if (c == '\x0A')
				break;
		}else if (rc == 0) {
			if (n == 1)
				return 0;
			else
				break;
		}else {
			if (errno == EINTR) 
				goto __vortex_frame_readline_again;
			if (errno == EWOULDBLOCK) 
				return (-2);
			
			/* if the connection is closed, just return
			 * without logging a message */
			if (vortex_connection_is_ok (connection, FALSE)) {
				error_msg = vortex_errno_get_last_error ();
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to read a line, error was: %s",
					    error_msg ? error_msg : "");
			}
			return (-1);
		}
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "line read from underlying transport: '%s'", buffer);
	*ptr = 0;
	return (n);

}

#ifdef __GNUC__ 
__inline__ 
#endif 
gint get_int_value (gchar * string, gint * position)
{
	int iterator = 0;
	int result;
	while ((string[iterator] != ' ')    &&
	       (string[iterator] != '\x0A') && 
	       (string[iterator] != '\x0D'))
		iterator++;
	/* set the new terminator */
	string[iterator] = 0;

	/* use atoi because we don't want to detect errors at this
	 * place */
	result   = strtol (string, NULL, 10);

	/* return the iterator position */
	*position = (*position) + iterator + 1;
	
	return result;
}


/** 
 * @internal
 *
 * Internal function which reads the frame hacer received, and place
 * that data into the the frame reference received.
 * 
 * @param beep_header An string representing the beep frame header.
 *
 * @param frame The VortexFrame header where the data will be placed.
 * 
 * @return Returns the number of bytes read from the beep header line.
 */
gint vortex_frame_get_header_data (gchar * beep_header, VortexFrame * frame)
{
	gint position = 0;

	/* check for null frame reader */
	if (beep_header == NULL)
		return 0;

	/* according to the frame type, read the content */
	switch (frame->type) {
	case VORTEX_FRAME_TYPE_SEQ:
		/* SEQ frame format: "%d %d %d\x0D\x0A"
		   channel seqno size */
		frame->channel    = get_int_value (beep_header, &position);
		frame->seqno      = get_int_value (beep_header + position, &position);
		frame->size       = get_int_value (beep_header + position, &position);
		break;
	default:
		/* where all cases matches: "%d %d %c %d %d\x0D\x0A"
		   channel msgno more_char seqno size */
		frame->channel    = get_int_value (beep_header, &position);
		frame->msgno      = get_int_value (beep_header + position, &position);

		/* get more char */
		frame->more_char  = beep_header [position];
		position         += 2; /* one space and point to the
					* next item */
		frame->seqno      = get_int_value (beep_header + position, &position);
		frame->size       = get_int_value (beep_header + position, &position);
		break;

	}
	
	/* check ANSNO case */
	if (frame->type == VORTEX_FRAME_TYPE_ANS) {
		/* ANS frame format: "%d %d %c %u %d %d\x0D\x0A"
		   channel msgno more_char seqno size ansno: read one
		   more int */
		frame->ansno      = get_int_value (beep_header + position, &position);
	}

	/* return current position plus one unit (\x0D and \x0A) */
	return position + 1;
	
}

/**
 * @internal
 * 
 * Tries to get the next incomming frame inside the given connection. If
 * received frame is ok, it return a newly-allocated VortexFrame with
 * its data. This function is only useful for vortex library
 * internals. Vortex library consumer should not use it.
 *
 * This function also close you connection if some error happens.
 * 
 * @param connection the connection where frame is going to be
 * received
 * 
 * @return a frame  or NULL if frame is wrong.
 **/
VortexFrame * vortex_frame_get_next     (VortexConnection * connection)
{
	gint          bytes_read;
	gint          remaining;
	VortexFrame * frame;
	gchar         line[100];
	gchar       * buffer = NULL;

	/* before reading anything else, we have to check if previous
	 * read was complete if not, we are in a frame fragment case */
	buffer = vortex_connection_get_data (connection, "buffer");
	if (buffer && (* buffer)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "received more data after a frame fragment, previous read isn't still complete");
		/* get previous frame */
		frame        = vortex_connection_get_data (connection, "frame");
		g_return_val_if_fail (frame, NULL);

		/* get previous remaining */
		remaining    = GPOINTER_TO_INT (vortex_connection_get_data (connection, "remaining_bytes"));
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "remaining bytes to be read: %d", remaining);
		g_return_val_if_fail (remaining > 0, NULL);

		/* get previous bytes read */
		bytes_read = GPOINTER_TO_INT (vortex_connection_get_data (connection, "bytes_read"));
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "bytes already read: %d", bytes_read);

		bytes_read = __vortex_frame_readn (connection, buffer + bytes_read, remaining);
		if (bytes_read == 0) {
			vortex_frame_free (frame);
			g_free (buffer);
			vortex_connection_set_data (connection, "buffer", NULL);
			vortex_connection_set_data (connection, "frame", NULL);
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "remote peer have closed connection while reading the rest of the frame having received part of it");
			__vortex_connection_set_not_connected (connection, "remote peer have closed connection while reading the rest of the frame");
			return NULL;
		}


		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "bytes ready this time: %d", bytes_read);

		/* check data received */
		if (bytes_read != remaining) {
			/* add bytes read to keep on reading */
			bytes_read += GPOINTER_TO_INT (vortex_connection_get_data (connection, "bytes_read"));
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "the frame fragment isn't still complete, total read: %d", bytes_read);
			goto save_buffer;
			
		}
		

		/* We have a complete buffer for the frame, let's
		 * continue the process but, before doing that we have
		 * to restore expected state of bytes_read. */
		bytes_read = (frame->size + 5);
		vortex_connection_set_data (connection, "buffer", NULL);
		vortex_connection_set_data (connection, "frame", NULL);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "this already complete (total size: %d", frame->size);
		goto process_buffer;
	}
	
	/* parse frame header, read the first line */
	bytes_read = __vortex_frame_readline (connection, line, 99);
	if (bytes_read == -2) {
                vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "no data were waiting on this non-blocking connection (EWOULDBLOCK error)");
		return NULL;
	}

	if (bytes_read == 0) {
		/* check if channel is expected to be closed */
		if (vortex_connection_get_data (connection, "being_closed")) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "properly connection close");
			__vortex_connection_set_not_connected (connection, "connection properly closed");
			return NULL;
		}

		/* check if we have a non-blocking connection */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "client have disconnected without closing properly this session id=%d",
		       vortex_connection_get_id (connection));
		__vortex_connection_set_not_connected (connection, "client have disconnected without closing session");
		return NULL;
	}
	if (bytes_read == -1) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "an error have ocurred while reading socket");
		__vortex_connection_set_not_connected (connection, "client have disconnected without closing session");
		return NULL;
	}

	if ((line[bytes_read - 1] != '\x0A') || (line[bytes_read - 2] != '\x0D')) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "no line definition found for frame, over connection id=%d, bytes read: %d, line: \n'%s'\n, closing session",
		       vortex_connection_get_id (connection),
		       bytes_read, line);
		__vortex_connection_set_not_connected (connection, "no line definition found for frame");
		return NULL;
	}

	/* create a frame */
	frame       = g_new0 (VortexFrame, 1);

	/* associate the next frame id available */
	frame-> id  = __vortex_frame_get_next_id ("get-next");

	/* check initial frame spec */
	frame->type = VORTEX_FRAME_TYPE_UNKNOWN;
	if (axl_stream_cmp (line, "MSG", 3))
		frame->type = VORTEX_FRAME_TYPE_MSG;
	if (axl_stream_cmp (line, "RPY", 3))
		frame->type = VORTEX_FRAME_TYPE_RPY;
	if (axl_stream_cmp (line, "ANS", 3))
		frame->type = VORTEX_FRAME_TYPE_ANS;
	if (axl_stream_cmp (line, "ERR", 3))
		frame->type = VORTEX_FRAME_TYPE_ERR;
	if (axl_stream_cmp (line, "NUL", 3))
		frame->type = VORTEX_FRAME_TYPE_NUL;
	if (axl_stream_cmp (line, "SEQ", 3))
		frame->type = VORTEX_FRAME_TYPE_SEQ;

	if (frame->type == VORTEX_FRAME_TYPE_UNKNOWN) {
		/* unref frame value */
		g_free (frame);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "poorly-formed frame: message type not defined");
		__vortex_connection_set_not_connected (connection, "poorly-formed frame: message type not defined");
		return NULL;
	}

	/* get the next frame according to the expected values */
	bytes_read = vortex_frame_get_header_data (&(line[4]), frame);

	/* in the case it is a SEQ frame */
	if (frame->type == VORTEX_FRAME_TYPE_SEQ) 
		return frame;

	/* check bytes read */
	if (bytes_read < 5) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "poorly-formed frame: message values are wrong  (%d < 9)", bytes_read);
		__vortex_connection_set_not_connected (connection, "poorly-formed frame: message values are wrong (%d < 9)");

		/* unref frame node allocated */
		g_free (frame);
		return NULL;
	}

	/* check more flag */
	if (frame->more_char != '.' && frame->more_char != '*') {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "poorly-formed frame: more char is wrong");
		__vortex_connection_set_not_connected (connection, "poorly-formed frame: more char is wrong");

		/* unref frame node allocated */
		g_free (frame);
		return NULL;
	}

	/* set more flag */
	if (frame->more_char == '.')
		frame->more = FALSE;
	else
		frame->more = TRUE;
	
	/* check incoming frame size */
	if (frame->size >= MAX_BUFFER_SIZE) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received an excessive sized frame, closing session");
		__vortex_connection_set_not_connected (connection, "received an excessive sized frame, closing session");

		/* unref frame node allocated */
		g_free (frame);
		return NULL;
	}

	/* allocate exactly frame->size + 5 bytes */
	buffer = g_new0 (gchar, frame->size + 6);
	
	/* read the next frame content */
	bytes_read = __vortex_frame_readn (connection, buffer, frame->size + 5);
	if (bytes_read == 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "remote peer have closed connection while reading the rest of the frame");
		__vortex_connection_set_not_connected (connection, "remote peer have closed connection while reading the rest of the frame");

		/* unref frame node allocated */
		g_free (frame);

		/* unref buffer allocated */
		g_free (buffer);
		return NULL;
	}

	if (bytes_read != (frame->size + 5)) {
		/* ok, we have received few bytes than expected but
		 * this is not wrong. Non-blocking sockets behave this
		 * way. What we have to do is to store the frame chunk
		 * we already read and the rest of bytes expected. 
		 *
		 * Later on, next frame_get_next calls on the same
		 * connection will return the rest of frame to be read. */

	save_buffer:
		/* save current frame */
		vortex_connection_set_data (connection, "frame", frame);
		
		/* save current buffer read */
		vortex_connection_set_data (connection, "buffer", buffer);
		
		/* save remaining bytes */
		vortex_connection_set_data (connection, "remaining_bytes", 
					    GINT_TO_POINTER ((frame->size + 5) - bytes_read));

		/* save read bytes */
		vortex_connection_set_data (connection, "bytes_read", 
					    GINT_TO_POINTER (bytes_read));

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "(ok message) received a frame fragment (expected: %d read: %d remaining: %d), storing into this connection id=%d",
		       (frame->size + 5), bytes_read, (frame->size + 5) - bytes_read, vortex_connection_get_id (connection));
		return NULL;
	}

process_buffer:

	/* check frame have ended */
	if (! axl_stream_cmp (&buffer[bytes_read - 5], "END\x0D\x0A", 5)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "poorly formed frame: frame trailer CR LF not found, discarding content: '%s'",
		       (buffer != NULL) ? buffer : "(null content)");
		__vortex_connection_set_not_connected (connection, "poorly formed frame: frame trailer CR LF not found, discarding content");

		/* unref frame node allocated */
		g_free (frame);

		/* unref buffer allocated */
		g_free (buffer);
		return NULL;
	}
	
	/* get if frame have a valid content type spec, get actual
	 * content type */
	frame->mime_headers_size = 0;
	if (axl_stream_cmp (buffer, "Content-Type: ", 14)) {

		while (buffer[(frame->mime_headers_size + 14)] != '\x0D')
			frame->mime_headers_size++;

		frame->content_type = g_new0 (gchar, frame->mime_headers_size + 1 );
		memcpy (frame->content_type, buffer + 14, frame->mime_headers_size);
		frame->content_type[frame->mime_headers_size] = 0;

		/* update iterator index (skip over the next \n\r\n
		 * values and point to the first value of the frame
		 * stream) */
		frame->mime_headers_size += 18;
	}else
		frame->content_type = NULL;

	/* locate body frame init */
	frame->size                 = frame->size - frame->mime_headers_size;

	/* allocate memory from frame payload and limit it */
	frame->payload              = g_new0 (gchar, frame->size + 1);

	/* copy content */
	memcpy (frame->payload, buffer + frame->mime_headers_size, frame->size);

	/* free allocated buffer */
	g_free (buffer);

	/* log frame on channel received */
	if (vortex_log_is_enabled ())
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Frame received on channel %d, content type=%s, transfer encoding=%s, payload size=%d, mime content size=%d", 
		       frame->channel,
		       (frame->content_type != NULL) ? frame->content_type : "" ,
		       (frame->transfer_encoding != NULL) ? frame->transfer_encoding : "",
		       frame->size, frame->mime_headers_size);

	return frame;

}


/** 
 * @internal
 * 
 * Sends data over the given connection
 * 
 * @param connection 
 * @param a_frame 
 * @param frame_size 
 * 
 * @return 
 */
gboolean          vortex_frame_send_raw     (VortexConnection * connection, gchar * a_frame, gint frame_size)
{
	gboolean result = TRUE;
	gint     bytes  = 0;
	gchar  * error_msg;

	g_return_val_if_fail (connection, FALSE);
	g_return_val_if_fail (vortex_connection_is_ok (connection, FALSE), FALSE);
	g_return_val_if_fail (a_frame && (* a_frame), FALSE);

 again:

	if ((bytes = vortex_connection_invoke_send (connection, a_frame, frame_size)) == VORTEX_SOCKET_ERROR) {
		if (errno == EINTR)
			goto again;
		if (errno == EWOULDBLOCK) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to write data to socket, socket not prepare to write");
			goto end;
		}
		
		/* check if socket have been disconnected (macro
		 * definition at vortex.h) */
		if (vortex_is_disconnected) {
			__vortex_connection_set_not_connected (connection, 
							       "remote peer have closed connection");
			return FALSE;
		}
		error_msg = vortex_errno_get_last_error ();
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to write data to socket: %s",
		       error_msg ? error_msg : "");
		__vortex_connection_set_not_connected (connection, "unable to write data to socket:");
		return FALSE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "bytes written: %d", bytes);

	if (bytes == 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "remote peer have closed before sending proper close connection, closing");
		__vortex_connection_set_not_connected (connection, 
						       "remote peer have closed before sending proper close connection, closing");
		return FALSE;
	}

	if (bytes != frame_size) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "write request mismatch with write done (%d != %d)",
		       bytes, frame_size);
		return FALSE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "write on socket request=%d written=%d", frame_size, bytes);
	
 end:
	return result;
}


/**
 * @brief Deallocate the frame.
 *
 * Frees a allocated \ref VortexFrame. Keep in mind that, unless
 * stated, \ref VortexFrame object received from the Vortex Library
 * API are not required to be unrefered. 
 *
 * In particular, first and second level invocation automatically
 * handle frame disposing. See \ref vortex_manual_dispatch_schema
 * "Vortex Frame Dispatch schema".
 *
 * @param frame The frame to free.
 **/
void          vortex_frame_free (VortexFrame * frame)
{
	if (frame == NULL)
		return;

	/* log a frame deallocated message */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "deallocating frame id=%d", frame->id);

	/* free content type */
	if (frame->content_type != NULL)
		g_free (frame->content_type);

	/* free content transfer encoding */
	if (frame->transfer_encoding != NULL)
		g_free (frame->transfer_encoding);

	/* free frame payload */
	if (frame->payload != NULL)
		g_free (frame->payload);

	/* free the frame node itself */
	g_free (frame);
	return;
}

VortexFrame * __vortex_frame_join_common (VortexFrame * a, VortexFrame * b, gboolean reuse)
{
	VortexFrame * result;
	
	/* only check if frames are joinables if the first one is different from NULL */
	if (!vortex_frame_are_joinable (a, b))
		return NULL;

	/* copy current frame values */
	result                    = g_new0 (VortexFrame, 1);
	/* get next Id for this new frame */
	result->id                = __vortex_frame_get_next_id ("frame-join");
	result->type              = a->type;
	result->channel           = a->channel;
	result->msgno             = a->msgno;
	result->more              = (a->more && b->more);
	result->seqno             = a->seqno;
	result->size              = a->size + b->size;
	result->mime_headers_size = a->mime_headers_size;
	result->ansno             = a->ansno;

	/* join payload for both frames */
	if (reuse) {
		/* allocates memory only for the rest of b */
		result->payload  = g_realloc (a->payload, a->size + b->size + 1);
		a->payload       = NULL;
	} else {
		/* allocates memory to hold both elements a and b */
		result->payload  = g_new0 (gchar, a->size + b->size + 1);

		/* mem copy a over result */
		memcpy (result->payload, a->payload, a->size);
	}
	/* now copy b over result starting from a's ending */
	memcpy (result->payload + a->size, b->payload, b->size);

	/* copy mime content type headers, do not make a memory copy
	 * because it is not necessary */
	if (a->content_type != NULL) {
		if (reuse) {
			/* just use the same memory and nullify source
			 * reference */
			result->content_type = a->content_type;
			a->content_type      = NULL;
		}else 
			result->content_type = g_strdup (a->content_type);
	} /* end if */

	/* copy mime content transfer header, do not make a memory
	 * copy because it is not necessary */
	if (a->transfer_encoding != NULL) {
		if (reuse) {
			/* just reuse the same memory and nullify
			   source reference */
			result->transfer_encoding = a->transfer_encoding;
			a->transfer_encoding      = NULL;
		}else 
			result->transfer_encoding = g_strdup (a->transfer_encoding);
	} /* end if */

	return result;
}

/**
 * @brief Allows to join two frames into a newly allocated one.
 * 
 * Joins the frame a with b. This is done by concatenating frame a
 * payload followed by frame b payload and adding both payload size,
 * both seqno number.  It also checks the more flag for both frames to
 * if it detect some error it will return NULL.  
 *
 * Frame types for both are also checked so if frame type differs a
 * NULL is returned. Channel number for both frames are also checked.
 *
 * As a result, frame joining done by this function is equal to having
 * activated \ref vortex_channel_set_complete_flag "the complete flag".
 *
 * 
 * @param a The frame to join.
 * @param b The frame to join.
 * 
 * @return the frame joined or NULL if fails. Returned value is
 * newly allocated.
 **/
VortexFrame * vortex_frame_join         (VortexFrame * a, VortexFrame * b)
{
	/* join without reusing */
	return __vortex_frame_join_common (a, b, FALSE);
}

/** 
 * @brief Allows to join two frames, without allocating a new one but,
 * reusing memory allocated by the first frame, saving memory an
 * memory dumping operations.
 * 
 * This function is the same as \ref vortex_frame_join. However, it
 * returns a copy to a new frame joined, reusing memory used by the
 * frame <b>a</b>. Once finished this function, the frame <b>a</b>
 * can't be used. 
 * 
 * @param a The frame to use as base for the resulting frame.
 * @param b The frame to add.
 * 
 * @return The new frame reference, contaning the union.
 */
VortexFrame * vortex_frame_join_extending (VortexFrame * a, VortexFrame * b)
{
	/* join without reusing */
	return __vortex_frame_join_common (a, b, TRUE);
}

/** 
 * @internal
 * Check two strings to be equal, including NULL values.
 * 
 * @param value_a The first value to check.
 * @param value_b The second value to check.
 * 
 * @return 
 */
gboolean vortex_frame_common_string_check (gchar * value_a, gchar * value_b)
{
	if (value_a == NULL && value_b == NULL)
		return TRUE;
	if (value_a != NULL && value_b == NULL)
		return FALSE;
	if (value_a == NULL && value_b != NULL)
		return FALSE;
	return !g_strcasecmp (value_a, value_b);
}

/** 
 * @brief Allows to check if the given frames are joinable or the first one follows the next.
 * 
 * @param a The frame to check.
 * @param b The frame to check.
 * 
 * @return TRUE if the frame b comes after a being possible to be both
 * joinable into only one frame.
 */
gboolean vortex_frame_are_joinable (VortexFrame * a, VortexFrame * b) 
{
	/*
	 * Because frames inside vortex with mime type are handled
	 * separating payload from entity-headers (mime headers), the
	 * payload size the frame have is only the payload of the mime
	 * content.
	 *
	 * If no mime content is used the payload size is the same
	 * that the frame payload content. According to BEEP
	 * specification, the payload and its mime content is
	 * considered payload. 
	 * 
	 * However, to simply using frames under Vortex Library, and
	 * to avoid making API consumers to place code that process
	 * payload with mime headers and payload without them, frame
	 * are read in a way that the payload size is the size of the
	 * mime content, without including the mime headers.
	 *
	 * As a consequence, when this tries to check if two frames
	 * are joinable it have to consider if the frame payload size
	 * plus mime type entity headers size, if defined. 
	 *
	 * That's why this function check the seq number with the mime
	 * type headers plus the payload size.
	 */

	g_return_val_if_fail (a, FALSE);
	g_return_val_if_fail (b, FALSE);
	
	if (a->type != b->type) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because type mismatch");
		return FALSE;
	}

	if (!a->more && !b->more)  {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because more flag mismatch");
		return FALSE;
	}

	if (!a->more && b->more) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because more flag mismatch 2");
		return FALSE;
	}

	if (a->channel != b->channel)  {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because channel mismatch");
		return FALSE;
	}

	if (a->msgno != b->msgno)  {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because  msgno mismatch");
		return FALSE;
	}

	if (a->ansno != b->ansno)  {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because ansno mismatch");
		return FALSE;
	}

	if ((a->seqno + a->size + a->mime_headers_size) != b->seqno)  {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not joinable because seqno mismatch (%d + %d + %d != %d)",
		       a->seqno, a->size, a->mime_headers_size, b->seqno);
		return FALSE;
	}

	return TRUE;
}

/** 
 * @brief Allows to check if the given frames are equal.
 *
 * This function will check header specification and payload to
 * report if the given frames are equal.
 * 
 * @param a The frame to compare
 * @param b The frame to compare
 * 
 * @return TRUE if both frames are equal, FALSE if not.
 */
gboolean vortex_frame_are_equal (VortexFrame * a, VortexFrame * b)
{
	g_return_val_if_fail (a, FALSE);
	g_return_val_if_fail (b, FALSE);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "checking for equality");

	/* check frame type */
	if (a->type != b->type) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to type (%d != %d)",
		       a->type, b->type);
		return FALSE;
	}
	
	/* check more flag type */
	if (a->more != b->more) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to more flag (%d != %d)",
		       a->more, b->more);
		return FALSE;
	}
	
	/* check channel number */
	if (a->channel != b->channel) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to channel number (%d != %d)",
		       a->channel, b->channel);
		return FALSE;
	}

	/* check message number */
	if (a->msgno != b->msgno) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to message number (%d != %d)",
		       a->msgno, b->msgno);
		return FALSE;
	}

	/* check sequence number */
	if (a->seqno != b->seqno) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to sequence number (%d != %d)",
		       a->seqno, b->seqno);
		return FALSE;
	}
	
	/* check frame size */
	if (a->size != b->size) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to frame size (%d != %d)",
		       a->size, b->size);
		return FALSE;
	}
	
	/* check ans number */
	if (a->ansno != b->ansno) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to ansno name (%d != %d)",
		       a->ansno, b->ansno);
		return FALSE;
	}

	/* check frame content type */
	if (!vortex_frame_common_string_check (a->content_type, b->content_type)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to content type value");
		return FALSE;
	}

	/* check for frame transfer encoding type */
	if (!vortex_frame_common_string_check (a->transfer_encoding, b->transfer_encoding)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to content transfer encoding value");
		return FALSE;
	}

	/* check payload */
	if (!vortex_frame_common_string_check (a->payload, b->payload)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "frames are not equal due to payload");
		return FALSE;
	}

	return TRUE;
}

/** 
 * @brief Allows to get the unique frame identifier for the given
 * frame.
 *
 * Inside Vortex Library, every frame created is flagged with an
 * unique frame identifier which is not part of the BEEP protocol,
 * transmitted or received from the remote side.
 *
 * This unique identifiers was created for the internal Vortex Library
 * support but, it has become a useful mechanism to have frames
 * identified. 
 *
 * Application level could use this value to implement fast frame
 * recognition rather than using \ref vortex_frame_are_equal, which is
 * more expensive.
 *
 * Vortex Library will ensure, as long as the process is in memory
 * that every unique identifier returned/generated for a frame will be
 * unique.
 * 
 * @param frame The frame where the unique identifier value will be
 * get.
 * 
 * @return Returns the unique frame identifier or -1 if it fails. The
 * only way for this function to fail is to receive a NULL value.
 */
gint          vortex_frame_get_id                (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);
	return frame->id;
}

/**
 * @brief return actual frame type.
 *
 * Returns actual frame type.
 *
 * @param frame The frame where the type is being requested.
 * 
 * @return the actual frame type or -1 if fail. The function only
 * fails if a null frame is received.
 **/
VortexFrameType vortex_frame_get_type   (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);
	
	return frame->type;
}

/**
 * @brief Returns frame content type.
 * 
 * Return actual frame's content type. Return value can be a NULL value
 * which means this frame has no content type defined.  You must not
 * free returned value.
 *
 * @param frame The frame where the content type will be returned.
 *
 * @return the content type if defined or NULL if not. On error, NULL
 * also is returned. Only one error is check, if frame is not NULL. If
 * you don't want to fall in the same case of having a correct frame
 * with no content type and having a wrong frame (NULL value) you
 * should check if frame is not NULL first.
 **/
gchar       * vortex_frame_get_content_type (VortexFrame * frame)
{
	g_return_val_if_fail (frame, NULL);

	return frame->content_type;
}

/** 
 * @brief Allows to get current Content-Transfer-Encoding mime header
 * configuration for the given frame (if defined).
 * 
 * @param frame The frame where the data request will be reported.
 * 
 * @return Current mime header configuration request or NULL if not
 * defined.
 */
gchar       * vortex_frame_get_transfer_encoding (VortexFrame * frame)
{
	g_return_val_if_fail (frame, NULL);

	return frame->transfer_encoding;
}

/** 
 * @brief Allows to get current status of the mime header size for the given frame. 
 *
 * This function is useful to avoid using \ref
 * vortex_frame_get_transfer_encoding and \ref
 * vortex_frame_get_content_type function and then adding mime tags
 * plus trailing characters.
 * 
 * This function return the total sum for the given frame.
 * 
 * @param frame The frame where the mime header size will be reported.
 * 
 * @return 0 or the mime header size count.
 */
gint          vortex_frame_get_mime_header_size  (VortexFrame * frame)
{
	g_return_val_if_fail (frame, 0);
	return frame->mime_headers_size;
}

/**
 * @brief Returns current channel used to sent or receive the given frame.
 * 
 * Returns actual frame's channel number.
 *
 * @param frame The frame where the channel number value is requested.
 *
 * @return the channel number or -1 if fails
 **/
gint          vortex_frame_get_channel  (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);

	return frame->channel;
}

/**
 * @brief Return current message number used for the given frame
 *
 * Returns actual frame's msgno.
 * 
 * @param frame The frame where the msgno value is requested.
 * 
 * @return the msgno or -1 if fail.
 **/
gint          vortex_frame_get_msgno    (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);

	return frame->msgno;
}

/**
 * @brief Returns actual more flag status for the given frame.
 * 
 * Returns actual frame more flag status. If more flag is activated 1
 * will be returned.  In case of been deactivated 0 will be returned.
 *
 * @param frame The frame where the more flag value is requested.
 *
 * Return value: the actual more flag status or -1 if fails
 **/
gint          vortex_frame_get_more_flag (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);

	if (frame->more)
		return 1;
	else
		return 0;
}

/**
 * @brief Returns the current sequence number for the given frame.
 * 
 * Returns the actual frame's seqno.
 * 
 * @param frame The frame where the seqno value is requested.
 *
 * @return the frame seqno or -1 if fails
 **/
gint          vortex_frame_get_seqno    (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);

	return frame->seqno;
}

/**
 * @brief Returns the payload the given frame have.
 * 
 * Return actual frame payload. You must not free returned payload.
 *
 * Keep in mind that this function returns the frame payload without
 * the mime headers. All mime headers read could be accessed through
 * \ref vortex_frame_get_content_type and \ref
 * vortex_frame_get_transfer_encoding.
 *
 * @param frame The frame where the payload is being requested.
 * 
 * @return the actual frame's payload or NULL if fail.
 **/
gchar      *  vortex_frame_get_payload  (VortexFrame * frame)
{
	g_return_val_if_fail (frame, NULL);

	return frame->payload;
}

/**
 * @brief Returns current ans number for the given frame.
 * 
 * Returns actual frame ansno.
 *
 * @param frame The frame where the ansno value is requested.
 *
 * @return the ansno value or -1 if fails
 **/
gint          vortex_frame_get_ansno    (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);

	return frame->ansno;
}


/**
 * @brief Returns the current payload size the given frame have
 * without taking into account mime headers.
 *
 * This function will return current payload size stored on the given
 * frame, skipping mime header size. This is because Vortex Library
 * manages mime headers for every frame received, making a separation
 * between the mime header part and the mime content.
 *
 * However, if you don't use mime headers, this function will return
 * the entire payload size.
 *
 * If you need to get current frame content size, including payload
 * size and mime content size, use \ref vortex_frame_get_content_size.
 *
 * 
 * @param frame The frame where the payload size will be reported.
 *
 * @return the actual payload size or -1 if fail
 **/
gint   vortex_frame_get_payload_size (VortexFrame * frame)
{
	g_return_val_if_fail (frame, -1);

	return frame->size;

}

/** 
 * @brief Allows to get current frame content size, including payload
 * and mime headers.
 *
 * Every frame have a payload that could be separated into the mime
 * headers and the mime content. Only the first one frame generated
 * due to a message splitting contains a mime header specification.
 *
 * You can get current mime header size by using \ref
 * vortex_frame_get_mime_header_size and current payload size by
 * calling to \ref vortex_frame_get_payload_size.
 *
 * @param frame The frame where the current content size will be
 * reported.
 * 
 * @return Current content size or 0 if it fails.
 */
gint          vortex_frame_get_content_size      (VortexFrame * frame)
{
	g_return_val_if_fail (frame, 0);
	
	return frame->size + frame->mime_headers_size;
	
}

/* Mapping values \n -> \x0A \r -> \x0D */
#define _ok_msg  "<ok />"


/** 
 * @brief Returns the ok message.
 *
 * @return A ok message. You must not free the returned message. It is
 * a copy to an internal Vortex Frame module.
 */
gchar       * vortex_frame_get_ok_message        ()
{
	return _ok_msg;
}

/** 
 * @brief Creates and return the error message.
 *
 * Allows to create a new error message that should be unrefered when
 * no longer is needed.
 * 
 * @param code          The error code to report.
 * @param error_content The error content, actually the textual error diagnostic.
 * @param xml_lang      The language used to define the diagnostic message.
 * 
 * @return A newly allocated error message.
 */
gchar       * vortex_frame_get_error_message    (gchar * code, 
						 gchar * error_content,
						 gchar * xml_lang)
{
	return g_strdup_printf ("<error code='%s'%s%s%s%s%s%s",
				code,
				(xml_lang != NULL)      ? " xml:lang='"       : "",
				(xml_lang != NULL)      ? xml_lang            : "",
				(xml_lang != NULL)      ? "' "                : "",
				(error_content != NULL) ? " >"                : " />\x0D\x0A",
				(error_content != NULL) ? error_content       : "",
				(error_content != NULL) ? "</error>\x0D\x0A"  : "");
}

/** 
 * @brief Allows to check if the given frame contains an error message
 * inside the frame payload.
 *
 * The function not only returns if the frame is an error message but
 * also returns the error code and the textual message.
 * 
 * @param frame   The frame to check for error message inside.
 * @param code    The error code the error message have (if defined)
 * @param message The textual error message (if defined)
 * 
 * @return TRUE if the frame contains an error message, FALSE if not.
 */
gboolean      vortex_frame_is_error_message      (VortexFrame * frame,
						  gchar ** code,
						  gchar ** message)
{
	g_return_val_if_fail (frame, FALSE);
	return vortex_channel_validate_err (frame, code, message);
}

/** 
 * @brief Creates and return an new start message using the given data.
 * 
 * 
 * @param channel_num          The channel number to be used for the new start message
 * @param serverName           The serverName attribute to be used for the start element.
 * @param profile              The profile element to be used inside the uri value.
 * @param encoding             How is encoded the profile content.
 * @param content_profile      The content profile data
 * @param profile_content_size Content profile size.
 * 
 * @return new start message that have to be deallocated when the it
 * is no longer need.
 */
gchar       * vortex_frame_get_start_message   (gint             channel_num,
						gchar          * serverName, 
						gchar          * profile,  
						VortexEncoding   encoding,
						gchar          * content_profile,
						gint             profile_content_size)
{
	/* build initial start message using channel num and
	 * serverName value. */
	return g_strdup_printf ("<start number='%d'%s%s%s>\x0D\x0A   <profile uri='%s'%s%s%s%s%s%s</start>\x0D\x0A",
				channel_num, 
				/* put server name definition if defined */
				(serverName != NULL)         ? " serverName='"      : "",
				(serverName != NULL)         ? serverName           : "",
				(serverName != NULL)         ? "'"                  : "",
				profile,
				(encoding == EncodingBase64) ? " encoding='base64'" : "",
				(content_profile == NULL)    ? " />\x0D\x0A"        : ">",
				(content_profile != NULL)    ? "<![CDATA["          : "",
				(content_profile != NULL)    ? content_profile      : "",
				(content_profile != NULL)    ? "]]>"                : "",
				(content_profile != NULL)    ? "</profile>\x0D\x0A" : "");
}

/** 
 * @brief Creates and return the close message.
 * 
 * Once the close message is no longer needed, it must be deallocated.
 * 
 * @param number        The number for the close message
 * @param code          The close code to report.
 * @param xml_lang      The xml lang value.
 * @param close_content The close message content.
 * 
 * @return A newly allocate close message which has to be deallocated
 * when no longer needed.
 */
gchar       * vortex_frame_get_close_message     (gint    number,
						  gchar * code,
						  gchar * xml_lang,
						  gchar * close_content)
{
	return g_strdup_printf ("<close number='%d' code='%s'%s%s%s%s%s%s",
				number,
				code,
				(xml_lang != NULL)      ? " xml:lang='"       : "",
				(xml_lang != NULL)      ? xml_lang            : "",
				(xml_lang != NULL)      ? "' "                : "",
				(close_content != NULL) ? " >"                : " />\x0D\x0A",
				(close_content != NULL) ? close_content       : "",
				(close_content != NULL) ? "</close>\x0D\x0A"  : "");
}

/** 
 * @brief Allows to create and return a start reply message.
 * 
 * @param profile The profile URI.
 * @param profile_content The optional profile content (piggyback value).
 * 
 * @return A newly allocated message that must be deallocated when no
 * longer needed.
 */
gchar       * vortex_frame_get_start_rpy_message (gchar * profile, 
						  gchar * profile_content)
{
	return g_strdup_printf ("<profile uri='%s'%s%s%s%s%s",
				profile,
				(profile_content != NULL) ? " >"                  : " />\x0D\x0A",
				(profile_content != NULL) ? "<![CDATA["           : "",
				(profile_content != NULL) ? profile_content       : "",
				(profile_content != NULL) ? "]]>"                 : "",
				(profile_content != NULL) ? "</profile>\x0D\x0A"  : "");
}

/* @} */
