/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#include <vortex.h>

#define LOG_DOMAIN "vortex-greetings"

gchar * __vortex_greetings_features = NULL;
gchar * __vortex_greetings_localize = NULL;

/**
 * \defgroup vortex_greetings Vortex Greetings: Related function to Initial Greeting messages.
 */

/**
 * \addtogroup vortex_greetings
 * @{
 */

/**
 * @internal
 * 
 * Sends the BEEP init connection greeting over the channel 0 of this @connection
 * This initial greetings message reply is sent by actual listener role to remote beep client peer.
 * 
 * @param connection The connection where the greeting will be sent.
 *
 * @return TRUE if greetings message was sent or FALSE if not
 **/
gboolean vortex_greetings_send (VortexConnection * connection)
{
	gchar         * the_payload         = NULL;
	gchar         * profiles            = NULL;
	gchar         * str_aux             = NULL;
	GList         * registered_profiles = vortex_profiles_get_actual_list ();
	GList         * list_aux            = NULL;
	gchar         * localize            = NULL;
	gchar         * features            = NULL;

	/* build up supported registered profiles */
	if (registered_profiles == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "unable to build and send greetings message: unable to found any profile registered");
		return FALSE;
	}

	list_aux = g_list_first (registered_profiles);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		/* avoid showing again TLS profile if the connection
		 * is already TLS ficated */
		if (!g_strcasecmp (VORTEX_TLS_PROFILE_URI, (gchar *) list_aux->data)) {
			if (vortex_connection_is_tlsficated (connection))
				continue;
		}

		/* don't show SASL profiles the the connection is
		 * already authenticated */
		if (!g_strncasecmp (VORTEX_SASL_FAMILY, (gchar *) list_aux->data, strlen (VORTEX_SASL_FAMILY))) {
			if (vortex_sasl_is_authenticated (connection))
				continue;
		}

		/* save profile */
		if (profiles == NULL)
			profiles = g_strdup_printf ("   <profile uri='%s' />\x0D\x0A",
						    (gchar *) list_aux->data);
		else {
			str_aux  = profiles;
			profiles = g_strdup_printf ("%s   <profile uri='%s' />\x0D\x0A",
						    profiles, 
						    (gchar *) list_aux->data);
			g_free (str_aux);
		}
	}
	g_list_free (registered_profiles);

	/* check for features and localize */
	if (vortex_greetings_get_features () != NULL)
		features = g_strdup_printf (" features='%s'", vortex_greetings_get_features ());
	if (vortex_greetings_get_localize () != NULL)
		localize = g_strdup_printf (" localize='%s'", vortex_greetings_get_localize ());

	/* build greetings message using registered profile info */
	the_payload = g_strdup_printf ("<greeting%s%s>\x0D\x0A%s</greeting>\x0D\x0A",
				       (features != NULL) ? features : "",
				       (localize != NULL) ? localize : "",
				       profiles);
	/* free value */
	g_free (profiles);

	/* free value */
	if (features != NULL)
		g_free (features);
	if (localize != NULL)
		g_free (localize);

	/* send the message */
	if (!vortex_channel_send_rpy (vortex_connection_get_channel (connection, 0),
				      the_payload,
				      strlen (the_payload),
				      0)) {
		__vortex_connection_set_not_connected (connection, "unable to send listener greetings message");
		g_free (the_payload);
		return FALSE;
	}

	g_free (the_payload);
	return TRUE;
}

/**
 * @internal
 * 
 * These functions helps vortex library to parse and process greetings
 * message response.  The greetings message is a important piece of
 * the BEEP protocol. On this message, profiles for listener peer are
 * sent, so actual BEEP client can figure out about is possibility of
 * connection to remote site.
 *
 * This function should not be useful for vortex library consumer. It
 * is used mainly by vortex internal purposes.
 *
 * @param connection connection where a greeting reply is expected.
 * 
 * @return The frame read or NULL if it fails.
 **/
VortexFrame *  vortex_greetings_process (VortexConnection * connection)
{
	VortexFrame        * frame;

	/*
	 * Because this is a really especial case where we need to get
	 * totally blocked until the *WHOLE* frame is receive, we set
	 * the connection as blocking.
	 */
	frame = vortex_frame_get_next (connection);
	if (frame == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "no frame received from remote peer");
		return NULL;
	}

	/* ensure from this point to be frame not NULL  */
	if (vortex_greetings_is_reply_ok (frame, connection))
		return frame;
	return NULL;
}

/**
 * @internal
 * 
 * This function should not be useful for vortex library
 * consumers.This function returns if the greeting reply is ok. The
 * greeting reply is supposed to be held the the received frame.
 *
 * If the frame is wrong this function free the frame.
 * 
 * @param frame the frame which contains the greetings reply.
 *
 * @return TRUE if frame greeting reply is ok or FALSE if not.
 **/
gboolean       vortex_greetings_is_reply_ok    (VortexFrame      * frame, VortexConnection * connection)
{
	VortexChannel * channel;

	/* check greetings reply */
	if (vortex_frame_get_type (frame) != VORTEX_FRAME_TYPE_RPY) {
		__vortex_connection_set_not_connected (connection, "frame error, expected RPY frame type on greetings process");
		vortex_frame_free (frame);
		return FALSE;
	}

	/* check frame header spec */
	if (vortex_frame_get_channel (frame) != 0 || 
	    vortex_frame_get_msgno   (frame) != 0 || 
	    vortex_frame_get_seqno   (frame) != 0) {
		__vortex_connection_set_not_connected (connection, "frame error, expected greetings message on channel 0, message 0 or sequence number 0");
		vortex_frame_free (frame);
		return FALSE;
	}
	
	/* check content-type reply */
	if ((vortex_frame_get_content_type (frame) == NULL) ||
	    (g_strcasecmp (vortex_frame_get_content_type (frame), "application/beep+xml"))) {

		__vortex_connection_set_not_connected (connection, 
						       "frame error, expected content type: application/beep+xml, not found");
		vortex_frame_free (frame);
		return FALSE;
	}

	/* update channel status */
	channel = vortex_connection_get_channel (connection, 0);
	vortex_channel_update_status_received (channel, 
					       vortex_frame_get_content_size (frame),
					       UPDATE_SEQ_NO | UPDATE_RPY_NO);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "greetings frame header specification is ok");
	
	return TRUE;	
}

/**
 * @internal
 * 
 * This function helps vortex library to be able to complete initial
 * session negotiation. It send the session greetings for a client
 * role. You can see vortex_greetings_send documentation for more
 * details.
 *
 * This function is mainly used by vortex library for its internal
 * purposes, so it should not be useful for vortex library consumers.
 * 
 * @return TRUE greetings message was sent or FALSE if not.
 **/
gboolean      vortex_greetings_client_send     (VortexConnection * connection)
{
	gchar * the_payload = NULL;
	gchar * features    = NULL;
	gchar * localize    = NULL;
	
	/* check for features and localize */
	if (vortex_greetings_get_features () != NULL)
		features = g_strdup_printf (" features='%s'", vortex_greetings_get_features ());
	if (vortex_greetings_get_localize () != NULL)
		localize = g_strdup_printf (" localize='%s'", vortex_greetings_get_localize ());

	the_payload = g_strdup_printf ("<greeting %s%s/>\x0D\x0A",
				       (features != NULL) ? features : "",
				       (localize != NULL) ? localize : "");
	/* free value */
	if (features != NULL)
		g_free (features);
	if (localize != NULL)
		g_free (localize);

	/* send the message */
	if (!vortex_channel_send_rpy (vortex_connection_get_channel (connection, 0),
				      the_payload,
				      strlen (the_payload),
				      0)) {
		g_free (the_payload);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,  "unable to send initial client greetings message");
		__vortex_connection_set_not_connected (connection, 
						       "unable to send initial client greetings message");
		return FALSE;
	}
	
	g_free (the_payload);				      
	return TRUE;
}

/**
 * @internal
 * 
 * This function helps vortex library to be able to process client
 * greetings response process.  This function is mainly used by vortex
 * library for its internal purposes. This function should not be
 * useful for vortex library consumers.
 * 
 * @return The vortex frame read or NULL if it fails..
 **/
VortexFrame *  vortex_greetings_client_process (VortexConnection * connection)
{
	VortexFrame * frame = vortex_greetings_process (connection);

	return frame;
}


/** 
 * @brief Allows to set advertised, optional, features attributes
 * while sending greeting element.
 *
 * Inside the 2.3.1.1, RFC 3080, The Greeting Message is defined that
 * the greeting message support an optional attribute called
 * "features". This attribute contains optional features that could be
 * advertised to the remote peer.
 *
 * This function allows to set the content of the "features"
 * attribute. It will be used by any greeting generated. These features
 * is only to be used to modify the channel management BEEP profile. 
 *
 * The function doesn't make any copy of the received value. You
 * should not free passed in value.
 * 
 * @param features The string to be set as features.
 */
void           vortex_greetings_set_features   (gchar * features)
{
	__vortex_greetings_features = features;
}

/** 
 * @brief Allows to get current "features" greetings attribute.
 *
 * Check documentation for: \ref vortex_greetings_set_features.
 * 
 * @return Current "features" status.
 */
gchar       *  vortex_greetings_get_features   ()
{
	return __vortex_greetings_features;
}

/** 
 * @brief Allows to set current "localize" greetings attribute. This
 * attribute allows to advertise to the remote peer the language
 * desired to generate error code responses.
 *
 * For values to identify your preferred language you must check: RFC
 * 3066, January 2001, "Tags for the Identification of Languages".
 *
 * This function doesn't make any copy of the received value. You
 * should not free passed in value.
 * 
 * @param localize The localize value.
 */
void           vortex_greetings_set_localize   (gchar * localize)
{
	__vortex_greetings_localize = localize;
}

/** 
 * @brief Return current "localize" greetings optional attribute status.
 * 
 * @return Current localize value.
 */
gchar       *  vortex_greetings_get_localize   ()
{
	return __vortex_greetings_localize;
}

/* @} */
