/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#ifndef __VORTEX_HANDLERS_H__
#define __VORTEX_HANDLERS_H__

#include <vortex.h>

/**
 * \defgroup vortex_handlers Vortex Handlers: Handlers used across Vortex Library for async notifications.
 */

/**
 * \addtogroup vortex_handlers
 * @{
 */


/** 
 * @brief Async notification for listener creation.
 *
 * Functions using this handler:
 * - \ref vortex_listener_new
 *
 * Optional handler defined to report which host and port have
 * actually allocated a listener peer. If host and port is null means
 * listener have failed to run.
 *
 * You should not free any parameter received, vortex system will do
 * this for you.  If you want to actually keep a copy you should use
 * g_strdup.
 * 
 * @param host the final host binded
 * @param port the final port binded
 * @param status the listener creation status.
 * @param message the message reporting the listener status creation.
 * @param user_data user data passed in to this async notifier.
 */
typedef void (*VortexListenerReady)           (gchar * host, gint port, VortexStatus status, 
					       gchar * message, gpointer user_data);



/** 
 * @brief Async notification for connection creation process.
 *
 * Functions using this handler:
 *  - \ref vortex_connection_new 
 * 
 * Optional handler defined to report the final status for a \ref
 * vortex_connection_new process. This handler allows you to create a
 * new connection to a vortex server (BEEP enabled peer) in a
 * non-blocking (or asynchronously way). 
 * 
 * Once the connection creation process have finished, no matter which
 * is the final result, the handler is invoked.
 *
 * @param connection The connection created (or not, check \ref vortex_connection_is_ok).
 * @param user_data user defined data passed in to this async notifier.
 */
typedef void (*VortexConnectionNew) (VortexConnection * connection, gpointer user_data);


/** 
 * @brief Async notification for start channel message received for a given profile.
 *
 * Functions using this handler:
 *  - \ref vortex_profiles_register
 *
 * This handler is used to notify listener role that a channel is
 * asked to be created. The channel to be created is associated with
 * connection reference. If listener agree to create the channel TRUE
 * must be returned. Otherwise, FALSE must be returned if channel
 * creation is not wanted. 
 * 
 * Vortex library ensure no petition to create new channels will
 * conflict with actual channel created to actual connection passed
 * Keep in mind BEEP protocol defines a connection as a session holder
 * of channels. But, channels on different session (or connection) are
 * independent.
 * 
 * On again, at the moment OnStartChannel is invoked, this channel is
 * not already created. So you can't get a reference to this channel
 * until you return TRUE.
 *
 * You can also provied a user_data pointer to functions with accept
 * this type of handlers to get access to some user defined data.
 * 
 * @param channel_num the channel num for the new channel attempted to be created.
 * @param connection the connection where the channel is going to be created.
 * @param user_data user defined data passed in to this async notifier.
 * 
 * @return TRUE if the new channel can be created or FALSE if not.
 */
typedef  gboolean (*VortexOnStartChannel)      (gint channel_num,
						VortexConnection * connection,
						gpointer user_data);


/** 
 * @brief Channel start received handler with support for extended
 * attributes.
 * 
 * On most situations, while using Vortex Library, you can assume to
 * not implement the start channel because it is used to accept new
 * channel creation or a initial tuning. That's why \ref
 * VortexOnStartChannel should fulfill most cases.
 *
 * However, an start message request could have several attributes an
 * additional parameters that are notified to set up new channels or
 * to perform tuning such as TLS negotiation. This handler allows to
 * get start message notification with all possible data received.
 * 
 * Here is an channel start message example:
 * \code
 *       <start number='1' serverName='my.domain.com'>
 *            <profile uri='http://some-unique-profile-uri/mech' encoding='none'>
 *                <![CDATA[some necessary initial round trip data]]>
 *            </profile>
 *       </start>
 * \endcode
 * 
 * This handler will notify the channel number being requested to be
 * created. In this case the value notified is 1, the serverName
 * notified is 'my.domain.com', the profile requested
 * 'http://some-unique-profile-uri/mech', the encoding 'non' and as a
 * profile content: '<![CDATA[some necessary initial round trip
 * data]]>'. 
 * 
 * Keep in mind that the encoding is an implicit attribute. Having it
 * not defined yields to the default value 'none', which is
 * represented by EncodingNone.
 *
 * Functions using this handler: 
 *   - \ref vortex_profiles_register_extended_start
 * 
 * @param profile The profile being request to create a new channel.
 * @param channel_num Channel being requested for creation.
 * @param connection  Connection where the channel creation request was received.
 * @param serverName  Value for this optional start message attribute. This is used 
 * @param profile_content Optional profile content received inside the channel start message.
 * @param profile_content_reply Optional profile content reply to be used for channel start reply.
 * @param encoding Encoding used for the profile content. 
 * @param user_data User defined data to be passed in to the handler when it is executed.
 * 
 * @return TRUE if the new channel can be created or FALSE if not. 
 */
typedef gboolean (*VortexOnStartChannelExtended) (gchar             * profile,
						  gint                channel_num,
						  VortexConnection  * connection,
						  gchar             * serverName,
						  gchar             * profile_content,
						  gchar            ** profile_content_reply,
						  VortexEncoding      encoding,
						  gpointer            user_data);
/** 
 * @brief Async notification for incoming close channel petition.
 * 
 * Functions using this handler:
 *   - \ref vortex_profiles_register
 *   - \ref vortex_channel_new
 * 
 * This handler allow you to control what to do on close channel
 * notification. Close notification is received when remote peer
 * wants to close some channel. Vortex library will ensure to 
 * handle automatically wrong close notification. 
 * 
 * Once again, the handler must return TRUE if channel can be close or
 * FALSE if not.  Once the handler is executed, Vortex library will
 * respond to remote peer using the BEEP protocol definition.
 * 
 * @param channel_num the channel num requesting to be closed.
 * @param connection the connection where the channel resides.
 * @param user_data user defined data passed in to this Async notifier.
 * 
 * @return TRUE if channel can be closed or FALSE if not.
 */
typedef  gboolean (*VortexOnCloseChannel)      (gint channel_num,
						VortexConnection * connection,
						gpointer user_data);

/** 
 * @brief Async notifier for frame received event.
 *
 * Functions using this handler:
 *   - \ref vortex_profiles_register
 *   - \ref vortex_channel_new
 *   - \ref vortex_channel_new_full
 *   - \ref vortex_channel_new_fullv
 *   - \ref vortex_channel_set_received_handler
 *   - \ref vortex_channel_pool_new
 *   - \ref vortex_channel_queue_reply
 * 
 * This handler allows you to control received frames from remote
 * peers. <b>channel</b> parameter is the channel where the
 * <b>frame</b> was received. <b>connection</b> parameter represent
 * the connection where the channel is running. Finally,
 * <b>user_data</b>, is a user defined pointer passed in to this
 * function and defined when the handler was registered.
 *
 * You must not free received frame. Vortex library will do it for you.
 * Remember that, if you want to have permanent access to frame payload,
 * make a copy. After running this handler, Vortex library will free
 * all VortexFrame resources.
 * 
 * @param channel the channel where the frame have been received.
 * @param connection the connection where the channel is running
 * @param frame the frame received
 * @param user_data user defined data passed in to this async notifier.
 */
typedef  void     (*VortexOnFrameReceived)     (VortexChannel    * channel,
						VortexConnection * connection,
						VortexFrame      * frame,
						gpointer user_data);


/** 
 * @brief Async notifier for channel creation process.
 *
 * Function using this handler:
 *  - \ref vortex_channel_new
 *
 * This handler allows you to define a notify function that will be
 * called once the channel is created or something have failed with
 * errors such as timeout reached or channel creation process
 * failures.
 *
 * If something fails, a -1 will be received on channel_num and
 * channel will be NULL.  Otherwise, you will received the channel
 * number created and a reference to channel created.  It is not
 * needed to hold a reference to channel. You can get it by using
 * \ref vortex_connection_get_channel.
 * 
 * 
 * @param channel_num the channel number for the new channel created
 * @param channel the channel created
 * @param user_data user defined data passed in to this async notifier.
 */
typedef void      (*VortexOnChannelCreated)  (gint            channel_num,
					      VortexChannel * channel,
					      gpointer        user_data);

/** 
 * @brief Async notifier for Vortex Channel Pool creation.
 *
 * Function using this handler:
 *  - \ref vortex_channel_pool_new
 *
 * This handler allows to make an async notification once the channel
 * pool process have finished. Because the channel pool creation
 * process can take a long time this handler is really useful to keep
 * on doing other things such as interface updating.
 * 
 * @param pool the channel pool created
 * @param user_data user defined data passed in to this async notifier.
 */
typedef void      (*VortexOnChannelPoolCreated) (VortexChannelPool * pool,
						 gpointer user_data);

/** 
 * @brief Async notifier for the channel close process.
 *
 * Function using this handler:
 *  - \ref vortex_channel_close
 *
 * This handler allows you to defined a notify function that will be
 * called once the channel close indication is received. This is
 * mainly used by \ref vortex_channel_close to avoid caller get
 * blocked by calling that function.
 *
 * Data received is the channel num this notification applies to, the
 * channel closing status and if was_closed is FALSE (the channel was
 * not closed) code hold error code returned and msg the message
 * returned by remote peer. 
 *
 * @param channel_num the channel num identifying the channel closed
 * @param was_closed status for the channel close process.
 * @param code the code representing the channel close process.
 * @param msg the message representing the channel close process.
 */
typedef void     (*VortexOnClosedNotification) (gint            channel_num,
						gboolean        was_closed,
						gchar         * code,
						gchar         * msg);

/** 
 * @brief Defines the writers handlers used to actually send data
 * through the underlaying socket descriptor.
 * 
 * This handler is used by: 
 *  - \ref vortex_connection_set_send_handler
 * 
 * @param connection Vortex Connection where the data will be sent.
 * @param buffer     The buffer holding data to be sent
 * @param buffer_len The buffer len.
 * 
 * @return     How many data was actually sent.
 */
typedef gint     (*VortexSendHandler)         (VortexConnection * connection,
					       gchar            * buffer,
					       gint               buffer_len);

/** 
 * @brief Defines the readers handlers used to actually received data
 * from the underlying socket descriptor.
 *  
 * This handler is used by: 
 *  - \ref vortex_connection_set_receive_handler
 * 
 * @param connection The Vortex connection where the data will be received.
 * @param buffer     The buffer reference used as container for data received.
 * @param buffer_len The buffer len use to know the limits of the buffer.
 * 
 * @return How many data was actually received.
 */
typedef gint     (*VortexReceiveHandler)         (VortexConnection * connection,
						  gchar            * buffer,
						  gint               buffer_len);

/** 
 * @brief Allows to set a handler that will be called when a
 * connection is about being closed.
 * 
 * This method could be used to establish some especial actions to be
 * taken before the connection is completely unrefered. Keep in mind
 * that this function is executed only once a a connection is broken.
 * 
 * This handler is used by:
 *  - \ref vortex_connection_set_on_close
 * 
 * <i><b>NOTE:</b> You must not free the connection received at the
 * handler. This is actually done by the library. </i>
 * 
 * @param connection The connection that is about to be unrefered.
 * 
 */
typedef void     (*VortexConnectionOnClose)      (VortexConnection * connection);

/** 
 * @brief Extended version for \ref VortexConnectionOnClose, which
 * also supports passing an user defined data.
 *
 * This handler definition is the same as \ref
 * VortexConnectionOnClose, being notified at the same time, but
 * providing an user defined pointer.
 *
 * See also:
 * 
 *  - \ref vortex_connection_set_on_close_full
 *  - \ref vortex_connection_set_on_close
 * 
 * @param The connection that is about to be closed.
 * @param The user defined pointer configured at \ref vortex_connection_set_on_close_full.
 * 
 * @return 
 */
typedef void     (*VortexConnectionOnCloseFull)  (VortexConnection * connection, gpointer data);


/** 
 * @brief Async handler definition to get notification about
 * connections created at the server side.
 *
 * This handler is executed once the connection is accepted and
 * registered in the system. If the function return FALSE, the
 * connection will be dropped, without reporting any error to the
 * remote peer. 
 *
 * This function could be used as a initial configuration for every
 * connection create more than as a filter way. So, if it is only
 * required to make a connection initialization, the handler provided
 * must always return TRUE to avoid dropping the connection.
 *
 * This handler is used by:
 * 
 *  - \ref vortex_listener_set_on_connection_accepted 
 *
 * @param connection The connection that has been accepted to be
 * managed by the listener.
 *
 * @param data The connection data to be passed in to the handler. 
 * 
 * @return TRUE if the connection is finally accepted. FALSE to drop
 * the connection, rejecting the incoming connection.
 */
typedef gboolean (*VortexOnAcceptedConnection)   (VortexConnection * connection, gpointer data);

/** 
 * @brief Async notifications for TLS activation.
 *
 * Once the process for TLS negotiation have started, using \ref
 * vortex_tls_start_negociation function, the status for such process is notified
 * using this handler type definition.
 *
 * The <i>status</i> value have to be checked in order to know if the
 * transport negotiation have finished successfully. Along the previous
 * variable, the <i>status_message</i> have a textual diagnostic about
 * the current status received.
 * 
 * While invoking \ref vortex_tls_start_negociation you could provide an user
 * space pointer, using the <i>user_data</i> parameter. That user data
 * is received on this handler.
 *
 * Functions using this handler:
 *  \ref vortex_tls_start_negociation
 * 
 * 
 * @param connection The connection where the TLS activation status is
 * being notified.
 *
 * @param status The process status.
 *
 * @param status_message A textual message representing the process
 * status.
 *
 * @param user_data A user defined pointer established at function
 * which received this handler.
 */
typedef void     (*VortexTlsActivation)          (VortexConnection * connection,
						  VortexStatus       status,
						  gchar            * status_message,
						  gpointer           user_data);


/** 
 * @brief Handler definition for those function used to configure if a
 * given TLS request should be accepted or denied.
 *
 * Once a TLS request is received this handler will be executed to
 * notify user space application on top of Vortex Library if the request should be denied or not. The
 * handler will receive the connection where the request was received
 * and an optional value serverName which represents a request to act as 
 * the server name the value represent.
 * 
 * This handler definition is used by:
 *   - \ref vortex_tls_accept_negociation
 * 
 * @param connection The connection where the TLS request was received.
 * @param serverName Optional serverName value requesting, if defined, to act as the server defined by this value.
 * 
 * @return TRUE if the TLS request should be accepted. FALSE to deny
 * the session TLS-fication.
 */
typedef gboolean (*VortexTlsAcceptQuery) (VortexConnection * connection,
					  gchar            * serverName);

/** 
 * @brief Handler definition for those function allowing to locate the
 * certificate file to be used while enabling TLS support.
 * 
 * Once an TLS negotiation is started two files are required to
 * enable TLS ciphering: the certificate and the private key. Two
 * handlers are used by the Vortex Library to allow user app level to
 * configure file locations for both files.
 * 
 * This handler is used to configure location for the certificate
 * file. The function will receive the connection where the TLS is
 * being request to be activated and the serverName value which hold a
 * optional host name value requesting to act as the server configured
 * by this value.
 * 
 * The function must return a path to the certificate using a
 * dynamically allocated value. Once finished, the Vortex Library will
 * unref it.
 * 
 * <b>The function should return a basename file avoiding full path file
 * names</b>. This is because the Vortex Library will use \ref
 * vortex_support_find_data_file function to locate the file
 * provided. That function is configured to lookup on the configured
 * search path provided by \ref vortex_support_add_search_path or \ref
 * vortex_support_add_search_path_ref.
 * 
 * As a consequence: 
 * 
 * - If all certificate files are located at
 *  <b>/etc/repository/certificates</b> and the <b>serverName.cert</b> is to
 *   be used <b>DO NOT</b> return on this function <b>/etc/repository/certificates/serverName.cert</b>
 *
 * - Instead, configure <b>/etc/repository/certificates</b> at \ref
 *    vortex_support_add_search_path and return <b>servername.cert</b>.
 * 
 * - Doing previous practices will allow your code to be as
 *   platform/directory-structure independent as possible. The same
 *   function works on every installation, the only question to be
 *   configured are the search paths to lookup.
 *  
 * 
 * @param connection The connection where the TLS negotiation was received.
 *
 * @param serverName An optional value requesting to as the server
 * <b>serverName</b>. This value is supposed to be used to select the
 * right certificate file.
 * 
 * This handler is used by:
 *  - \ref vortex_tls_accept_negociation 
 * 
 * @return A newly allocated value containing the path to the certificate file.
 */
typedef gchar * (* VortexTlsCertificateFileLocator) (VortexConnection * connection,
						     gchar            * serverName);
	
/** 
 * @brief Handler definition for those function allowing to locate the
 * private key file to be used while enabling TLS support.
 * 
 * See \ref VortexTlsCertificateFileLocator handler. This handler
 * allows to define how is located the private key file used for the
 * session TLS-fication.
 * 
 * This handler is used by:
 *  - \ref vortex_tls_accept_negociation 
 * 
 * @return A newly allocated value containing the path to the private key file.
 */
typedef gchar * (* VortexTlsPrivateKeyFileLocator) (VortexConnection * connection,
						    gchar            * serverName);
						  

/** 
 * @brief Pre read handler definition.
 * 
 * This handler definition is mainly used to perform especial
 * operations prior to fully accept a connection into the vortex
 * library. This handler is detected and invoked in an abstract manner
 * from inside the Vortex Reader.
 *
 * Functions using this handler are: 
 *  - \ref vortex_connection_set_preread_handler
 * 
 * @param connection The connection which is receiving the on pre read
 * event.X
 */
typedef void   (* VortexConnectionOnPreRead)       (VortexConnection * connection);

/** 
 * @brief Async notifications for SASL auth process. 
 *
 * This notification handler is used to report user space what is the
 * final status for the SASL profile negotiation selected. 
 *
 * The handler report an <b>status</b> variable and a textual
 * diagnostic error on <b>status_message</b>.
 *
 * This function is used by: 
 *   - \ref vortex_sasl_start_auth
 *
 *
 * @param connection     The connection where the SASL process have ended
 * @param status         Final status for the SASL process
 * @param status_message A textual diagnostic
 * @param user_data      User defined data provided at the function receiving this handler.
 */
typedef void     (*VortexSaslAuthNotify)          (VortexConnection * connection,
						   VortexStatus       status,
						   gchar            * status_message,
						   gpointer           user_data);

/** 
 * @brief Asynchronous notification to enable user space to accept or
 * deny authentication for SASL EXTERNAL profile.
 *
 * Functions using this handler are:
 *   - \ref vortex_sasl_set_external_validation
 * 
 * @param connection The connection where the request was received.
 * @param authorization_id The authorization id to authenticate.
 * 
 * @return TRUE to authenticate and allow request received or FALSE to deny it.
 */
typedef gboolean     (*VortexSaslAuthExternal)        (VortexConnection * connection,
						       const gchar      * authorization_id);

/** 
 * @brief Asynchronous notification to enable user space to accept or
 * deny authentication for SASL EXTERNAL profile, and passes a user defined pointer.
 *
 * Functions using this handler are:
 *   - \ref vortex_sasl_set_external_validation_full
 * 
 * @param connection The connection where the request was received.
 * @param authorization_id The authorization id to authenticate.
 * @param user_data The user defined pointer to be passed to future SASL callbacks
 * 
 * @return TRUE to authenticate and allow request received or FALSE to deny it.
 */
typedef gboolean     (*VortexSaslAuthExternalFull)        (VortexConnection * connection,
						       const gchar      * authorization_id,
							   gpointer           user_data);

/** 
 * @brief Asynchronous notification to enable user space to accept or
 * deny anonymous authentication for SASL ANONYMOUS profile.
 * 
 * Function using this definition:
 * - \ref vortex_sasl_set_anonymous_validation
 * 
 * @param connection The connection where the anonymous request was received.
 *
 * @param anonymous_token The anonymous token provided by the remote
 * client side. You must not modify or deallocate this value.
 * 
 * @return TRUE to validate the incoming anonymous auth request. FALSE to deny it.
 */
typedef gboolean     (*VortexSaslAuthAnonymous)       (VortexConnection * connection,
						       const gchar      * anonymous_token);

							   /** 
 * @brief Asynchronous notification to enable user space to accept or
 * deny anonymous authentication for SASL ANONYMOUS profile.
 * 
 * Function using this definition:
 * - \ref vortex_sasl_set_anonymous_validation_full
 * 
 * @param connection The connection where the anonymous request was received.
 *
 * @param anonymous_token The anonymous token provided by the remote
 * client side. You must not modify or deallocate this value.
 * 
 * @param user_data The user defined pointer to be passed to future SASL callbacks
 *
 * @return TRUE to validate the incoming anonymous auth request. FALSE to deny it.
 */
typedef gboolean     (*VortexSaslAuthAnonymousFull)       (VortexConnection * connection,
						       const gchar      * anonymous_token,
							   gpointer           user_data);

/** 
 * @brief Asynchronous notification to enable user space code to
 * validate SASL PLAIN request received.
 * 
 * Function using this handler are:
 *  - \ref vortex_sasl_set_plain_validation
 * 
 * 
 * @param connection The connection where the SASL notification was received.
 * @param auth_id User identifier
 * @param authorization_id User identifier which is acting on behalf of.
 * @param password Current user password
 * 
 * @return TRUE to accept incoming SASL request or FALSE to deny it.
 */
typedef gboolean     (*VortexSaslAuthPlain)           (VortexConnection * connection,
						       const gchar      * auth_id,
						       const gchar      * authorization_id,
						       const gchar      * password);

/** 
 * @brief Asynchronous notification to enable user space code to
 * validate SASL PLAIN request received.
 * 
 * Function using this handler are:
 *  - \ref vortex_sasl_set_plain_validation_full
 * 
 * 
 * @param connection The connection where the SASL notification was received.
 * @param auth_id User identifier
 * @param authorization_id User identifier which is acting on behalf of.
 * @param password Current user password
 * @param user_data The user defined pointer to be passed to future SASL callbacks.
 * 
 * @return TRUE to accept incoming SASL request or FALSE to deny it.
 */
typedef gboolean     (*VortexSaslAuthPlainFull)           (VortexConnection * connection,
						       const gchar      * auth_id,
						       const gchar      * authorization_id,
						       const gchar      * password,
							   gpointer           user_data);

/** 
 * @brief Asynchronous notification to enable user space code to
 * validate SASL CRAM MD5 request received.
 *
 * Inside this function, programmer should implement password lookup
 * for the given user (<b>auth_id</b>) and return a copy it the clear
 * text password.
 *
 * Functions using this handler are:
 *  - \ref vortex_sasl_set_cram_md5_validation
 * 
 * @param connection The connection where the SASL notification was received
 * @param auth_id User id to authenticate.
 * 
 * @return The password the given auth_id have or NULL if the request
 * should be denied. Password returned should be the plain one. Later
 * SASL engine will perform hashing operations over it to be compared
 * with the hash value received from the client side. <b>Returned
 * value should be dynamically allocated. </b> Later, Vortex will unref it.
 */
typedef gchar *      (*VortexSaslAuthCramMd5)         (VortexConnection * connection,
						       const gchar      * auth_id);

/** 
 * @brief Asynchronous notification to enable user space code to
 * validate SASL CRAM MD5 request received.
 *
 * Inside this function, programmer should implement password lookup
 * for the given user (<b>auth_id</b>) and return a copy it the clear
 * text password.
 *
 * Functions using this handler are:
 *  - \ref vortex_sasl_set_cram_md5_validation_full
 * 
 * @param connection The connection where the SASL notification was received
 * @param auth_id User id to authenticate.
 * @param user_data The user defined pointer to be passed to future SASL callbacks.
 * 
 * @return The password the given auth_id have or NULL if the request
 * should be denied. Password returned should be the plain one. Later
 * SASL engine will perform hashing operations over it to be compared
 * with the hash value received from the client side. <b>Returned
 * value should be dynamically allocated. </b> Later, Vortex will unref it.
 */
typedef gchar *      (*VortexSaslAuthCramMd5Full)         (VortexConnection * connection,
						       const gchar      * auth_id,
							   gpointer           user_data);

/** 
 * @brief Asynchronous notification to enable user space to validate
 * SASL DIGEST MD5 received requests.
 *
 * Inside this function, programmer should implement password lookup
 * for the given user (<b>auth_id</b>) and return a copy it the clear
 * text password.
 *
 * Functions using this handler are:
 *   - \ref vortex_sasl_set_digest_md5_validation
 * 
 * @param connection The connection where the SASL notification was received.
 * @param auth_id User identifier to authenticate
 * @param authorization_id If set, requesting auth_id is asking to get authorized to act as this value.
 * @param realm Optional realm value where the auth_id and the authorization_id will be validated.
 * 
 * @return The password the given auth_id have or NULL if the request
 * should be denied. Password returned should be the plain one. Later
 * SASL engine will perform hashing operations over it to be compared
 * with the hash value received from the client side. <b>Returned
 * value should be dynamically allocated. </b> Later Vortex will unref it.
 */
typedef gchar *      (*VortexSaslAuthDigestMd5)       (VortexConnection * connection,
						       const gchar      * auth_id,
						       const gchar      * authorization_id,
						       const gchar      * realm);

/** 
 * @brief Asynchronous notification to enable user space to validate
 * SASL DIGEST MD5 received requests.
 *
 * Inside this function, programmer should implement password lookup
 * for the given user (<b>auth_id</b>) and return a copy it the clear
 * text password.
 *
 * Functions using this handler are:
 *   - \ref vortex_sasl_set_digest_md5_validation_full
 * 
 * @param connection The connection where the SASL notification was received.
 * @param auth_id User identifier to authenticate
 * @param authorization_id If set, requesting auth_id is asking to get authorized to act as this value.
 * @param realm Optional realm value where the auth_id and the authorization_id will be validated.
 * @param user_data The user defined pointer to be passed to future SASL callbacks.
 * 
 * @return The password the given auth_id have or NULL if the request
 * should be denied. Password returned should be the plain one. Later
 * SASL engine will perform hashing operations over it to be compared
 * with the hash value received from the client side. <b>Returned
 * value should be dynamically allocated. </b> Later Vortex will unref it.
 */
typedef gchar *      (*VortexSaslAuthDigestMd5Full)       (VortexConnection * connection,
						       const gchar      * auth_id,
						       const gchar      * authorization_id,
						       const gchar      * realm,
							   gpointer           user_data);

/** 
 * @brief Async notification handler for the XML-RPC channel boot.
 *
 * This async handler is used by: 
 *  - \ref vortex_xml_rpc_boot_channel
 * 
 * @param booted_channel The channel already booted or NULL if
 * something have happened.
 *
 * @param status The status value for the XML-RPC boot process
 * request, \ref VortexOk if boot was ok, \ref VortexError if something
 * have happened.
 *
 * @param message A textual diagnostic message reporting current status.
 *
 * @param user_data User space data defined at the function requesting
 * this handler.
 */
typedef void (* VortexXmlRpcBootNotify)               (VortexChannel    * booted_channel,
						       VortexStatus       status,
						       gchar            * message,
						       gpointer           user_data);

/** 
 * @brief This async handler allows to control how is accepted XML-RPC
 * initial boot channel based on a particular resource.
 *
 * XML-RPC specificaiton states that the method invocation is
 * considered as a two-phase invocation. 
 *
 * The first phase, the channel boot, is defined as a channel
 * creation, where the XML-RPC profile is prepared and, under the same
 * step, it is asked for a particular resource value to be supported
 * by the server (or the listener which is actually the entity
 * receiving the invocation).
 *
 * The second step is actually the XML-RPC invocation. 
 *
 * This resource value, by default, is "/". This allows to group
 * services under resources like: "/sales", "/sales/dep-a", etc. This
 * also allows to ask to listerners if they support a particular
 * interface. Think about grouping services, which represents a
 * concrete interface, under the same resource. Then, listener that
 * supports the resource are reporting that they, indeed, support a
 * particular interface.
 *
 * However, resource validation is not required at all. You can live
 * without it safely.
 *
 * As a note, you can pass a NULL handler reference to the \ref
 * vortex_xml_rpc_accept_negociation, making the Vortex XML-RPC engine
 * to accept all resources (resource validation always evaluated to
 * TRUE).
 *
 * This function is used by:
 *
 *  - \ref vortex_xml_rpc_accept_negociation
 *
 * Here is an example of a resource validation handler that accept all
 * resources:
 * \code
 * gboolean validate_all_resources (VortexConnection * connection,
 *                                  gint               channel_number,
 *                                  gchar            * serverName,
 *                                  gchar            * resource_path)
 * {
 *     // Ahis is a kind of useless validation handler because 
 *     // it is doing the same like passing a NULL reference.
 *     // 
 *     // On this handler could be programmer any politic about 
 *     // or not a resource, maybe based on the connection source
 *     // the number of channels that the connection have, etc.
 *     return TRUE;
 * }
 * \endcode
 * 
 * @param connection     The connection where the resource bootstrapping request was received.
 * @param channel_number The channel number to create requested.
 * @param serverName     An optional serverName value to act as.
 * @param resource_path  The resource path requested.
 * 
 * @return TRUE to accept resource requested. FALSE if not.
 */
typedef gboolean (*VortexXmlRpcValidateResource) (VortexConnection * connection, 
						  gint               channel_number,
						  gchar            * serverName,
						  gchar            * resource_path);


/** 
 * @brief Async handler to process all incoming service invocation
 * through XML-RPC.
 *
 * This handler is executed to notify user space that a method
 * invocation has been received and should be dispatched, returning a
 * \ref XmlRpcMethodResponse. This reponse (\ref XmlRpcMethodResponse)
 * will be used to generate a reply as fast as possible.
 *
 * The method invocation, represented by this async handler, could
 * also return a NULL value, making it possible to not reply at the
 * moment the reply was received, but once generated the reply, the
 * method call object (\ref XmlRpcMethodCall) received should be used
 * in conjuction with the method response associated (\ref
 * XmlRpcMethodResponse) calling to \ref vortex_xml_rpc_notify_reply.
 *
 * Previous function, \ref vortex_xml_rpc_notify_reply, could allow to
 * marshall the invocation into another language execution, making it
 * easy because it is only required to marshall the method invocator
 * received but not to have a handler inside the runtime for the other
 * language. However, this is mainly provided to make it possible to defer
 * the reply.
 *
 *
 * @param channel The channel where the method call was received.
 *
 * @param method_call The method call object received, representing
 * the method being invoked (\ref XmlRpcMethodCall).
 *
 * @param user_data User space data. 
 *
 * @return The dispatch function should return a \ref
 * XmlRpcMethodResponse containing the value reply to be returned or
 * NULL if the method invocation have been deferred. 
 */
typedef XmlRpcMethodResponse *  (*VortexXmlRpcServiceDispatch)  (VortexChannel    * channel,
								 XmlRpcMethodCall * method_call,
								 gpointer           user_data);

/** 
 * @brief Async method response notifier.
 *
 * This handler is provided to enable Vortex XML-RPC engine to notify
 * the user space that a method response has been received.
 * 
 * This handler is used by:
 *
 *   - \ref vortex_xml_rpc_invoke
 * 
 * @param channel  The channel where the invocation took place.
 * @param response The reply received from the remote invocation.
 * @param user_data User defined data.
 * 
 */
typedef void     (* XmlRpcInvokeNotify)          (VortexChannel        * channel, 
						  XmlRpcMethodResponse * response,
						  gpointer               user_data);


/** 
 * @brief IO handler definition to allow defining the method to be
 * invoked while createing a new fd set.
 * 
 * @return A newly created fd set pointer, opaque to Vortex, to a
 * structure representing the fd set, that will be used to perform IO
 * waiting operation at the \ref vortex_io "Vortex IO module".
 * 
 */
typedef gpointer     (* VortexIoCreateFdGroup)        ();

/** 
 * @brief IO handler definition to allow defining the method to be
 * invoked while destroying a fd set. 
 *
 * The reference that the handler will receive is the one created by
 * the \ref VortexIoCreateFdGroup handler.
 * 
 * @param VortexIoDestroyFdGroup The fd_set, opaque to vortex, pointer
 * to a structure representing the fd set to be destroy.
 * 
 */
typedef void     (* VortexIoDestroyFdGroup)        (gpointer               fd_set);

/** 
 * @brief IO handler definition to allow defining the method to be
 * invoked while clearing a fd set.
 * 
 * @param VortexIoClearFdGroup The fd_set, opaque to vortex, pointer
 * to a structure representing the fd set to be clear.
 * 
 */
typedef void     (* VortexIoClearFdGroup)        (gpointer               fd_set);



/** 
 * @brief IO handler definition to allow defining the method to be
 * used while performing a IO blocking wait, by default implemented by
 * the IO "select" call.
 *
 * @param VortexIoWaitOnFdGroup The handler to set.
 *
 * @param The maximum value for the socket descriptor being watched.
 *
 * @param The requested operation to perform.
 * 
 * @return An error code according to the description found on this
 * function: \ref vortex_io_waiting_set_wait_on_fd_group.
 */
typedef gint     (* VortexIoWaitOnFdGroup)       (gpointer               fd_group,
						  gint                   max_fds,
						  VortexIoWaitingFor     wait_to);

/** 
 * @brief IO handler definition to perform the "add to" the fd set
 * operation.
 * 
 * @param fds The socket descriptor to be added.
 *
 * @param fd_group The socket descriptor group to be used as
 * destination for the socket.
 * 
 */
typedef void     (* VortexIoAddToFdGroup)        (gint                   fds,
						  gpointer               fd_group);

/** 
 * @brief IO handler definition to perform the "is set" the fd set
 * operation.
 * 
 * @param fds The socket descriptor to be added.
 *
 * @param fd_group The socket descriptor group to be used as
 * destination for the socket.
 *
 * @return TRUE if the socket descriptor is active in the given fd
 * group.
 * 
 */
typedef gboolean (* VortexIoIsSetFdGroup)        (gint                   fds,
						  gpointer               fd_group);



#endif

/* @} */
