/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>

#define LOG_DOMAIN "vortex-io"

/**
 * \defgroup vortex_io Vortex IO: Vortex Library IO abstraction layer 
 */

/** 
 * \addtogroup vortex_io
 * @{
 */

/** 
 * @internal
 *
 * @brief Internal vortex implementation to create a compatible
 * "select" IO call fd set reference.
 *
 * @return A newly allocated fd_set reference.
 */
gpointer __vortex_io_waiting_default_create () 
{
	gpointer result = g_new0 (fd_set, 1);

	return result;
}

/** 
 * @internal
 *
 * @brief Internal vortex implementation to destroy the "select" IO
 * call created by the default create.
 * 
 * @param fd_group The fd group to be deallocated.
 */
void    __vortex_io_waiting_default_destroy (gpointer fd_group)
{
	fd_set * __fd_set = fd_group;

	/* release memory allocated */
	g_free (__fd_set);
	
	/* nothing more to do */
	return;
}

/** 
 * @internal
 *
 * @brief Internal vortex implementation to clear the "select" IO
 * call created by the default create.
 * 
 * @param fd_group The fd group to be deallocated.
 */
void    __vortex_io_waiting_default_clear (gpointer __fd_group)
{
	fd_set * fd_group = __fd_group;

	/* clear the fd set */
	FD_ZERO (fd_group);

	/* nothing more to do */
	return;
}

/** 
 * @internal
 *
 * @brief Default internal implementation for the wait operation to
 * change its status at least one socket description inside the fd set
 * provided.
 * 
 * @param __fd_group The fd set having all sockets to be watched.
 * @param wait_to The operation requested.
 * 
 * @return The error code as described by \ref vortex_io_waiting_set_wait_on_fd_group.
 */
gint __vortex_io_waiting_wait_on_default (gpointer __fd_group, gint max_fds, VortexIoWaitingFor wait_to)
{
	gint               result = -1;
	struct timeval     tv;
	fd_set           * fd_set = __fd_group;

	/* perform the select operation according to the
	 * <b>wait_to</b> value. */
	switch (wait_to) {
	case READ_OPERATIONS:
		/* init wait */
		tv.tv_sec    = 0;
		tv.tv_usec   = 500000;
		result       = select (max_fds + 1, fd_set, NULL,   NULL, &tv);
		break;
	case WRITE_OPERATIONS:
		tv.tv_sec    = 2;
		tv.tv_usec   = 0;
		result       = select (max_fds + 1, NULL,   fd_set, NULL, &tv);
		break;
	}
	
	/* check result */
	if ((result == VORTEX_SOCKET_ERROR) && (errno == EINTR))
		return -1;
	
	return result;
}

/** 
 * @internal
 *
 * @brief Default Vortex Library implementation for the "add to" on fd
 * set operation.
 * 
 * @param fds The socket descriptor to be added.
 *
 * @param fd_set The fd set where the socket descriptor will be added.
 */
void __vortex_io_waiting_add_to_default (gint fds, gpointer __fd_set)
{
        fd_set * value = __fd_set;

	FD_SET (fds, value);
	return;
}

/** 
 * @internal
 *
 * @brief Default Vortex Library implementation for the "is set" on fd
 * set operation.
 * 
 * @param fds The socket descriptor to be checked to be active on the
 * given fd group.
 *
 * @param fd_set The fd set where the socket descriptor will be checked.
 */
gboolean __vortex_io_waiting_is_set_default (gint fds, gpointer __fd_set)
{
	fd_set * value = __fd_set;
	
	return FD_ISSET (fds, value);
}


/** 
 * @internal
 * 
 * Following is the set of handlers to be used while performing IO
 * waiting operation, as defined by the API exported to modify
 * function to be used.
 *  
 * There is one handler for each operation that could be performed. No
 * mutex and any other blocking operation should be used while
 * implementing this set of handler.
 *
 * Critical section, resource exclusion and any other race condition
 * is ensure by the Vortex Library to not happen. 
 *
 * See default handles implemented to get an idea about how should be
 * implemented other IO blocking operations using other APIs rather than
 * select.
 */
VortexIoCreateFdGroup  __vortex_io_waiting_create  = __vortex_io_waiting_default_create;
VortexIoDestroyFdGroup __vortex_io_waiting_destroy = __vortex_io_waiting_default_destroy;
VortexIoClearFdGroup   __vortex_io_waiting_clear   = __vortex_io_waiting_default_clear;
VortexIoWaitOnFdGroup  __vortex_io_waiting_wait_on = __vortex_io_waiting_wait_on_default;
VortexIoAddToFdGroup   __vortex_io_waiting_add_to  = __vortex_io_waiting_add_to_default;
VortexIoIsSetFdGroup   __vortex_io_waiting_is_set  = __vortex_io_waiting_is_set_default;

/** 
 * @brief Allows to set current handler that allows to create a fd
 * group that will be used by Vortex internals to perform IO blocking.
 * 
 * @param create The handler to be executed by the Vortex Library to
 * create a new IO fd set. The function will fail to set the default
 * create handler is a NULL handler is provided. No handler will be
 * modified is the create handler is not provided.
 */
void                 vortex_io_waiting_set_create_fd_group (VortexIoCreateFdGroup create)
{
	/* check that the handler provided is not null */
	g_return_if_fail (create);

	/* set default create handler */
	__vortex_io_waiting_create = create;
	
	return;
}

/** 
 * @internal
 *
 * @brief Perform an invocation to create a fd group using current
 * create handler defined.
 * 
 * This function is for internal Vortex Library purposes and should
 * not be used by API consumers.
 * 
 * @return Returns a reference to a new allocated fd set.
 */
gpointer             vortex_io_waiting_invoke_create_fd_group ()
{
	/* check current create IO handler configuration */
	if (__vortex_io_waiting_create == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "critical error, found that create fd set handler is not properly set, this is critical malfunctions");
		return NULL;
	}
	
	/* return a newly allocated fd set */
	return __vortex_io_waiting_create ();
}


/** 
 * @brief Allows to define the destroy method to be used on the fd_set
 * allocated by calling current configured \ref VortexIoCreateFdGroup
 * "create handler".
 *
 * @param destroy The handler to deallocate. The function will fail if
 * the provided destroy function is NULL. No operation will be
 * performed in such case.
 */
void                 vortex_io_waiting_set_destroy_fd_group    (VortexIoDestroyFdGroup destroy)
{
	/* check if the destroy handler is not a NULL reference */
	g_return_if_fail (destroy);

	/* set default destroy handler */
	__vortex_io_waiting_destroy = destroy;

	return;
}

/** 
 * @internal
 *
 * @brief Invokes current destroy handler configured to release memory
 * allocate by the fd group created by calling \ref
 * vortex_io_waiting_invoke_create_fd_group.
 *
 * This function is for internal Vortex Library purposes and should
 * not be used by API consumers.
 * 
 * @param fd_group The fd group to destroy. 
 */
void                 vortex_io_waiting_invoke_destroy_fd_group (gpointer fd_group)
{
	/* check that the received fd_group is not null */
	g_return_if_fail (fd_group);

	/* check current destroy IO handler configuration */
	if (__vortex_io_waiting_destroy == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "critical error, found that destroy fd set handler is not properly set, this is critical malfunctions");
		return;
	}
	
	/* return a newly allocated fd set */
	__vortex_io_waiting_destroy (fd_group);

	return;
}


/** 
 * @brief Allows to configure the clear handler to be executed while
 * it is required to clear a descriptor set.
 * 
 * @param clear The handler to executed once Vortex internal process
 * requires to create a socket descriptor set.
 */
void                 vortex_io_waiting_set_clear_fd_group    (VortexIoClearFdGroup clear)
{
	/* check if the clear handler is not a NULL reference */
	g_return_if_fail (clear);

	/* set default clear handler */
	__vortex_io_waiting_clear = clear;

	return;
}

/** 
 * @internal
 *
 * @brief Invokes current clear fd group configure handler on the
 * given reference received.
 *
 * This function is internal to vortex library, and it is used at the
 * right place. This function should be be used by API consumers.
 *
 * @param fd_group The fd_group reference to clear.
 */
void                 vortex_io_waiting_invoke_clear_fd_group (gpointer fd_group)
{
	/* check that the received fd_group is not null */
	g_return_if_fail (fd_group);

	/* check current clear IO handler configuration */
	if (__vortex_io_waiting_clear == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "critical error, found that clear fd set handler is not properly set, this is critical malfunctions");
		return;
	}
	
	/* return a newly allocated fd set */
	return __vortex_io_waiting_clear (fd_group);
}

/** 
 * @brief Allows to configure the add socket to fd set operation.
 *
 * The function will fail if the provided handler is NULL.
 *
 * @param add_to The handler to be invoked when it is required to add
 * a socket descriptor into the fd set.
 */
void                 vortex_io_waiting_set_add_to_fd_group     (VortexIoAddToFdGroup add_to)
{
	/* check for NULL reference handlers */
	g_return_if_fail (add_to);

	/* set the new handler */
	__vortex_io_waiting_add_to = add_to;

	return;
}

/** 
 * @internal
 *
 * @brief Invokes current add to operation for the given socket
 * descriptor into the given fd set.
 * 
 * @param fds The socket descriptor to be added. The socket descriptor
 * should be greater than 0.
 *
 * @param on_reading The fd set where the socket descriptor will be
 * added.
 */
void                 vortex_io_waiting_invoke_add_to_fd_group  (gint fds, gpointer fd_group)
{
	g_return_if_fail (fds > 0);
	g_return_if_fail (fd_group);
	
	/* check for properly add to handler configuration */
	if (__vortex_io_waiting_add_to == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "there is no \"add to\" operation defined, this will cause critical malfunctions");
		return;
	}

	/* invoke add to operation */
	__vortex_io_waiting_add_to (fds, fd_group);

	return;
}

/** 
 * @brief Allows to configure the is set operation for the socket on the fd set.
 *
 * The function will fail if the provided handler is NULL.
 *
 * @param is_set The handler to be invoked when it is required to
 * check if a socket descriptor is set into a given fd set.
 */
void                 vortex_io_waiting_set_is_set_fd_group     (VortexIoIsSetFdGroup is_set)
{
	/* check for NULL reference handlers */
	g_return_if_fail (is_set);

	/* set the new handler */
	__vortex_io_waiting_is_set = is_set;

	return;
}

/** 
 * @internal
 *
 * @brief Invokes current is set to operation for the given socket
 * descriptor into the given fd set.
 * 
 * @param fds The socket descriptor to be checked. The socket
 * descriptor should be greater than 0.
 *
 * @param on_reading The fd set where the socket descriptor will be
 * checked for its activation.
 */
gboolean              vortex_io_waiting_invoke_is_set_fd_group  (gint fds, gpointer fd_group)
{
	g_return_val_if_fail (fds > 0,  FALSE);
	g_return_val_if_fail (fd_group, FALSE);
	
	/* check for properly add to handler configuration */
	if (__vortex_io_waiting_is_set == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "there is no \"is set\" operation defined, this will cause critical malfunctions");
		return FALSE;
	}

	/* invoke add to operation */
	return __vortex_io_waiting_is_set (fds, fd_group);
}





/** 
 * @brief Allows to configure the wait on fd group handler. 
 * 
 * This handler allows to configure how is implemented a non-consuming
 * resource on a given fd group.
 *
 * The invoked function should return a selected error code according
 * to the following values:
 * 
 * - A timeout operation was being received, then <b>-1</b> should
 * be returned.
 *
 * - A error has being received but the call loop should continue
 * without paying attention to it, without checking for the sockets
 * for its status, and to perform as soon as possible the same
 * wait. This is used because current current default implementation,
 * which used the IO "select" call could return because a signal was
 * received. If the IO implemented doesn't require to implement this
 * concept, just ignore it. Returned value should be <b>-2</b>.
 *
 * - A non recoverable error has happened, and vortex reader should
 * exist. In this case <b>-3</b> should be returned. This error code
 * is also used by the internal vortex IO handler invocation to report
 * that a wait handler haven't been defined.
 *
 * - No socket descriptor have change but it is required to perform
 * again the blocking IO wait. Then <b>0</b> should be returned.
 *
 * - For the rest of the cases, the number of sockets could be
 * returned by this function, however, this is not used by the Vortex
 * Library. To report that a socket have changed (or a group of then
 * inside the fd set), and you don't want to return how many sockets
 * have changed, just return a positive value (<b>1</b>).
 * 
 * @param wait_on The handler to be used. The function will fail on
 * setting the handler if it is provided a NULL reference.
 */
void                 vortex_io_waiting_set_wait_on_fd_group    (VortexIoWaitOnFdGroup wait_on)
{
	/* check for NULL reference received */
	g_return_if_fail (wait_on);

	/* set new default handler to be used */
	__vortex_io_waiting_wait_on = wait_on;

	return;
}

/** 
 * @internal
 *
 * @brief Invokes current implementation for IO blocking until the
 * requested operation could be performed of any of the given sockets
 * descriptors inside the descriptor group selected.
 *
 * @param fd_group The socket descriptor group where the request wait
 * operation will be performed.
 *
 * @param wait_to The operation requested.
 * 
 * @return Current error code or the number of sockets that have
 * changed.
 * 
 */
gint                 vortex_io_waiting_invoke_wait             (gpointer           fd_group, 
								gint               max_fds,
								VortexIoWaitingFor wait_to)
{
	/* check for NULL reference */
	if (__vortex_io_waiting_wait_on == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "wait handler is not defined, this is a critical error that will produce a great malfunctions");
		return -3;
	}

	/* perform IO blocking operation */
	return __vortex_io_waiting_wait_on (fd_group, max_fds, wait_to);
}


/* @} */


