/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#include <vortex.h>
#define LOG_DOMAIN "vortex-listener"

GAsyncQueue   * vortex_listener_wait_lock = NULL;
GStaticMutex    __listener_mutex          = G_STATIC_MUTEX_INIT;
GStaticMutex    __listener_unlock         = G_STATIC_MUTEX_INIT;

/* configuration handler for OnAccepted signal */
VortexOnAcceptedConnection __vortex_listener_on_accept      = NULL;
gpointer                   __vortex_listener_on_accept_data = NULL;

typedef struct _VortexListener {
	gchar      * host;
	gchar      * port;
	GHashTable * channels;

}VortexListener;


gint __vortex_listener_get_port (gchar * port)
{
	return strtol (port, NULL, 10);
}

/**
 * \defgroup vortex_listener Vortex Listener: Related function to create Vortex Listener (server for accepting incoming petitions)
 */

/**
 * \addtogroup vortex_listener
 * @{
 */

/** 
 * @brief Common task to be done to accept a connection before
 * greetings message was issued while working as a Listener.
 * 
 * @param connection The connection to  accept.
 */
void vortex_listener_accept_connection    (VortexConnection * connection)
{
	/*
	 * register the connection on vortex reader from here, the
	 * connection get a non-blocking state
	 */
	vortex_reader_watch_connection      (connection);

	/* flag the connection to be on initial step */
	vortex_connection_set_data (connection, "initial_accept", GINT_TO_POINTER (TRUE));

	/*
	 * Because this is a listener, we don't want to pay attention
	 * to free connection on errors. connection already have 2
	 * references (reader and writer), so let's reference counting
	 * to the job of free connection resources
	 */
	vortex_connection_unref (connection, "vortex listener (initial accept)");

	/* close connection and free resources */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "worker ended, connection registered on manager (initial accept)");

	return;
}

/** 
 * @internal
 *
 * Internal vortex library function. This function does the initial
 * accept for a new incoming connection. New connections are accepted
 * through two steps: an initial accept and final negotiation. A
 * connection to be totally accepted must step over two previous
 * steps.
 *
 * The reason to accept connections following this procedure is due to
 * Vortex Library way of reading data from all connections. 
 * 
 * Reading data from remote BEEP peers is done inside the vortex
 * reader which, basicly, is a loop executing a "select" call. 
 *
 * While reading data from those sockets accepted by the "select" call
 * is not a problem, however, it is actually a problem to accept new
 * connections because it implies reading data from remote peer: the
 * initial greeting, and writing data to remote peer: the initial
 * greeting response. 
 *
 * During the previous negotiation a malicious client can make
 * negotiation to be stopped, or sending data in an slow manner,
 * making the select loop to be blocked, even stopped. As a consequence
 * this malicious client have thrown down the reception for all
 * channels inside all connections.
 *
 * However, the vortex reader loop is prepared to avoid this problem
 * with already accepted connections because it doesn't pay attention
 * to those connection which are not sending any data and also support
 * to receive fragmented frames which are stored to be joined with
 * future fragment. Once connection is accepted the vortex reader is
 * strong enough to avoid DOS (denial of service) attacks (well, it
 * should be ;-).
 *
 * That way the mission for the first step: to only accept the new
 * connection and send the initial greeting to remote peer and *DO NOT
 * READING ANYTHING* to avoid DOS. On a second step, the response
 * reading is done and the connection is totally accepted in the
 * context of the vortex reader.
 *
 * A connection initial accepted is flagged to be on that step so
 * vortex reader can recognize it. 
 *
 * @param client_socket A new socket being accepted to be read.
 */
void __vortex_listener_initial_accept (VORTEX_SOCKET client_socket)
{
	VortexConnection     * connection = NULL;

	/* before doing anything, we have to create a connection */
	connection = vortex_connection_new_empty (client_socket, VortexRoleListener);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received connection from: %s:%s", 
	       vortex_connection_get_host (connection),
	       vortex_connection_get_port (connection));

	/*
	 * activate vortex writer for this connection because all
	 * sending process is done through vortex sequencer plus
	 * vortex writer.
	 */
	vortex_writer_watch_connection      (connection);

	/* send greetings, get actual profile installation and report
	 * it to init peer */
	if (! vortex_greetings_send (connection)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to send initial greeting message");
		
		/*
		 * This unref sentence is properly defined. Opposite
		 * ref call done to this unref is actually done by
		 * vortex_connection_new_empty.
		 */ 
		vortex_connection_unref (connection, "vortex listener");
		return;
	}

	/*
	 * Perform an initial accept, flagging the connection to be
	 * into the initial accept stage.
	 */
	vortex_listener_accept_connection (connection);

	return;
}

/** 
 * @internal
 *
 * This function is for internal vortex library purposes. This
 * function actually does the second accept step, that is, to read the
 * greeting response and finally accept the connection is that response
 * is ok.
 *
 * You can also read the doc for __vortex_listener_initial_accept to
 * get an idea about the initial step.
 *
 * Once the greeting response is ok, the function unflag this
 * connection to be "being accepted" so the connection starts to work.
 * 
 * @param frame The frame which should contains the greeting response
 * @param connection the connection being on initial accept step
 */
void __vortex_listener_second_step_accept (VortexFrame * frame, VortexConnection * connection)
{
	gboolean result;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "greetings sent, waiting reply");

	/* process greetings from init peer */
	if (!vortex_greetings_is_reply_ok (frame, connection)) {
		/* previous function already unref frame object
		 * received is something goes wrong */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "wrong greeting rpy from init peer, closing session");
		__vortex_connection_set_not_connected (connection, "wrong greeting rpy from init peer, closing session");
		return;
	}

	/* because the greeting is ok, parse it */
	if (!__vortex_connection_parse_greetings (connection, frame)) {
		/* previous function doesn't perform a frame
		 * deallocation, unref it*/
		vortex_frame_free (frame);
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "wrong greetings received, closing session");
		__vortex_connection_set_not_connected (connection, "wrong greetings received, closing session");
		return;
	}

	/* frame accepted */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "accepting connection on vortex_reader (second accept step)");
	
	/* free the last frame and watch connection on changes */
	vortex_frame_free (frame);
	
	/* flag the connection to be totally accepted. */
	vortex_connection_set_data (connection, "initial_accept", NULL);

	/* notify here for a new connection created */
	if (__vortex_listener_on_accept != NULL) {
		/* call to the handler defined */
		result = __vortex_listener_on_accept (connection, __vortex_listener_on_accept_data);
		if (result == FALSE) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "the server application level have dropped the provided connection");
			__vortex_connection_set_not_connected (connection, "connection filtered by on accept handler");
		}
	}

	return;	
}



void vortex_listener_accept_connections (gint server_socket)
{
	struct sockaddr_in inet_addr;
#ifdef G_OS_WIN32
	int               addrlen;
#else
	socklen_t         addrlen;
#endif
	VORTEX_SOCKET      client_socket;

	addrlen       = sizeof(struct sockaddr_in);
	/* accept the connection new connection */
	client_socket = accept (server_socket, (struct sockaddr *)&inet_addr, &addrlen);
	if (client_socket == VORTEX_SOCKET_ERROR) {
		g_warning ("accept () failed\n");
		return;
	}

	/* instead of negotiate the connection at this point simply
	 * accept it to negotiate it inside vortex_reader loop.  */
	__vortex_listener_initial_accept (client_socket);

	return;
}

typedef struct _VortexListenerData {
	gchar               * host;
	gchar               * port;
	VortexListenerReady   on_ready;
	gpointer              user_data;
	gboolean              threaded;
}VortexListenerData;

gpointer __vortex_listener_new (VortexListenerData * data)
{
	struct hostent     * he;
        struct in_addr     * haddr;
        struct sockaddr_in   saddr;
	struct sockaddr_in   sin;
	VORTEX_SOCKET        fd;
#ifdef G_OS_WIN32
	BOOL                 unit      = TRUE;
	int                  sin_size  = sizeof (sin);
#else    	
	gint                 unit      = 1;
	socklen_t            sin_size  = sizeof (sin);
#endif	
	gchar              * host      = data->host;
	gchar              * port      = data->port;
	gint                 int_port  = __vortex_listener_get_port (port);
	gboolean             threaded  = data->threaded;
	VortexListenerReady  on_ready  = data->on_ready;
	gpointer             user_data = data->user_data;
	gchar              * message   = NULL;
	
	/* free data */
	g_free (data);

	if (vortex_profiles_registered () == 0) {
		message = "no registered profile was found, unable to start";
		goto error;
	}
	
	he = gethostbyname (host);
        if (he == NULL) {
		message = "unable to get hostname by calling gethostbyname";
		goto error;
	}

	haddr = ((struct in_addr *) (he->h_addr_list)[0]);
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		message = "unable to create a new socket";
		goto error;
        }

#ifdef G_OS_WIN32
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (gchar *)&unit, sizeof(BOOL));
#else
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &unit, sizeof (unit));
#endif

	memset(&saddr, 0, sizeof(struct sockaddr_in));
	saddr.sin_family          = AF_INET;
        saddr.sin_port            = htons(int_port);
        memcpy(&saddr.sin_addr, haddr, sizeof(struct in_addr));

        if (bind(fd, (struct sockaddr *)&saddr,  sizeof (struct sockaddr_in)) == VORTEX_SOCKET_ERROR) {
		vortex_close_socket (fd);
		message = "unable to bind address";
		goto error;
        }

	if (listen(fd, 5) == VORTEX_SOCKET_ERROR) {
		message = "an error have occur while executing listen";
		goto error;
        }

	/* notify listener */
	if (getsockname (fd, (struct sockaddr *) &sin, &sin_size) < -1) {
		message = "an error have occur while executing getsockname";
		goto error;
	}

	/* unref the host and port value */
	vortex_support_free (2, host, g_free, port, g_free);

	/* register the listener socket at the Vortex Reader process.  */
	vortex_reader_watch_listener (fd);
	if (threaded) {
		on_ready (inet_ntoa(sin.sin_addr), ntohs (sin.sin_port), VortexOk,
			  "server ready for requests", user_data);
	}
	return NULL;

 error:
	/* unref the host and port */
	vortex_support_free (2, host, g_free, port, g_free);
	if (threaded) {
		on_ready (NULL, 0, VortexError, message, user_data);
	}else {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to start vortex server, error was: %s, unblocking vortex_listener_wait",
		       message);
		/* notify the listener that an error was found
		 * (because the server didn't suply a handler) */
		g_static_mutex_lock (&__listener_unlock);
		g_async_queue_push (vortex_listener_wait_lock, GINT_TO_POINTER (TRUE));
		vortex_listener_wait_lock = NULL;
		g_static_mutex_unlock (&__listener_unlock);
	}
	return NULL;	
}

/** 
 * @brief Creates a new Vortex Listener accepting incoming connections
 * on the given <b>host:port</b> configuration.
 *
 * If user pass in an \ref VortexListenerReady "on_ready" callback,
 * it will be notified on it, in a separated thread, once the process
 * have finished. Check documentation for the \ref VortexListenerReady
 * handler which is the async definition for \ref VortexListenerReady "on_ready" handler.
 * 
 * On that notification will also be passed the host and port actually
 * allocated. Think about using as host 0.0.0.0 and port 0.
 * 
 * These values will cause to \ref vortex_listener_new to allocate the
 * system configured hostname and a randomly free port. See
 * \ref vortex_handlers "this section" for more info about on_ready.
 *
 * Host and port value provided to this function could be unrefered
 * once returning from this function. The function performs a local
 * copy for those values, that are deallocated at the appropriate
 * moment.
 *
 * Keep in mind that you can actually call several times to this
 * function before calling to \ref vortex_listener_wait, to make your
 * process to be able to accept connections from several ports and
 * host names at the same time. 
 *
 * While providing the port information, make sure your process will
 * have enough rights to allocate the port provided. Usually, ports
 * from 1 to 1024 are reserved to listener programms that runs with
 * priviledges.
 *
 * Here is an little example to start a vortex listener server:
 *
 * \code
 * // On this example you'll find:
 * //   - 3 handlers: frame_received, start_channel and
 * //                 close_channel (which handle the event
 * //                 they represent).
 * //   - An entry point (main function) which creates a
 * //     a simple listener server, using previous three
 * //     basic handlers.
 *
 * // a frame handler
 * void frame_received (VortexChannel    * channel,
 *                      VortexConnection * connection,
 *                      VortexFrame      * frame,
 *                      gpointer           user_data)
 * {
 *        // Received a frame from the remote side.
 *        // process it and reply to it if it is a MSG.
 *        return;
 * }
 *
 * gboolean start_channel (gint               channel_num, 
 *                        VortexConnection * connection, 
 *                        gpointer           user_data)
 * {
 *        g_print ("Received an start message=%d!!\n",
 *                 channel_num);
 *        // if the async notifier returns TRUE, the channel
 *        // is implicitly created, if FALSE is returned the
 *        // channel creation is denied and a reply error
 *        // is sent.
 *        return TRUE;
 * }
 *
 * gboolean close_channel (gint               channel_num, 
 *                         VortexConnection * connection, 
 *                         gpointer           user_data)
 * {
 *        g_print ("Got a close message notification!!\n");
 *
 *        // if TRUE is returned, the channel is
 *        // accepted to be closed. Otherwise the channel
 *        // will not be closed and an error reply will be
 *        // sent to the remote peer.
 *        return TRUE;
 * }
 * int main (int argc, char ** argv) {
 *      // enable log to see whats going on 
 *      vortex_log_enable (TRUE);
 *      
 *      // init vortex library (and check its result!)
 *      if (! vortex_init ()) {
 *           g_print ("Unable to initialize vortex library\n");
 *           exit (-1);
 *      }
 * 
 *      // register a profile
 *      vortex_profiles_register ("http://vortex.aspl.es/profiles/example", 
 *                                start_channel, NULL,
 *                                close_channel, NULL,
 *                                frame_received, NULL);
 * 
 *      // now create a vortex server
 *      vortex_listener_new ("0.0.0.0", "3000", NULL, NULL);
 *      
 *      // wait for listener to finish (maybe due to vortex_exit call)
 *      vortex_listener_wait ();
 *  
 *      // end vortex internal subsystem (if no one have done it yet!)
 *      vortex_exit ();
 * 
 *      // that's all to start BEEPing!
 *      return 0;     
 * }  
 * \endcode
 *
 * @param host the host to listen to.
 * @param port the port to listen to.
 * @param on_ready a optional notify callback to get when vortex listener is ready to perform replies.
 * @param user_data a user defined pointer to be passed in to <i>on_ready</i> handler.
 */
void vortex_listener_new (gchar * host, gchar * port, VortexListenerReady on_ready, gpointer user_data)
{

	VortexListenerData * data;
	
	g_return_if_fail (host && (* host));
	g_return_if_fail (port && (* port));

	/* init listener module */
	vortex_listener_init ();
	
	/* prepare function data */
	data             = g_new0 (VortexListenerData, 1);
	data->host       = g_strdup (host);
	data->port       = g_strdup (port);
	data->on_ready   = on_ready;
	data->user_data  = user_data;
	data->threaded   = (on_ready != NULL);
	
	/* make request */
	if (data->threaded) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invoking listener_new threaded mode");
		vortex_thread_pool_new_task ((GThreadFunc) __vortex_listener_new, data);
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invoking listener_new non-threaded mode");
	__vortex_listener_new (data);

	return;
	
}



/** 
 * @brief Blocks a listener (or listeners) launched until vortex finish.
 * 
 * This function should be called after creating a listener (o
 * listeners) calling to \ref vortex_listener_new to block current
 * thread.
 * 
 * This function can be avoided if the program structure can ensure
 * that the programm will not exist after calling \ref
 * vortex_listener_new. This happens when the program is linked to (or
 * implements) and internal event loop.
 *
 * This function will be unblocked when the vortex listener created
 * ends or a failure have occur while creating the listener. To force
 * an unlocking, a call to \ref vortex_listener_unlock must be done.
 * 
 */
void vortex_listener_wait ()
{
	GAsyncQueue * temp;
	/* double locking to ensure waiting */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Locking listener");
	if (vortex_listener_wait_lock != NULL) {
		/* get a local reference to the queue and work with it */
		temp = vortex_listener_wait_lock;

		/* get blocked until the waiting lock is released */
		g_async_queue_pop   (temp);
		
		/* unref the queue */
		g_async_queue_unref (temp);
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "(un)Locked listener");

	return;
}

/**
 * @brief Signals to unblock the listener blocked at the \ref vortex_listener_wait.
 * 
 * Under normal circumstances, this function must not be called
 * directly from user space. This function actually unblock waiting
 * listeners on vortex_listener_wait function.
 *
 * Vortex already call to this function once the \ref vortex_exit
 * function is called. 
 *
 * Because some designs could require to control how memory is
 * released once the listener is unblocked from the \ref
 * vortex_listener_wait, this function could help to control an
 * trigger that unblock operation.
 *
 * Futher calls to this function will cause no operation. This
 * function is thread safe.
 *
 **/
void vortex_listener_unlock ()
{
	/* unlock listener */
	g_static_mutex_lock (&__listener_unlock);
	if (vortex_listener_wait_lock != NULL) {

		/* push to signal listener unblocking */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "(un)Locking listener..");
		g_async_queue_push (vortex_listener_wait_lock, GINT_TO_POINTER (TRUE));

		/* nullify */
		vortex_listener_wait_lock = NULL;
		g_static_mutex_unlock (&__listener_unlock);
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "(un)Locking listener: already unlocked..");
	g_static_mutex_unlock (&__listener_unlock);
	return;
}

/**
 * @internal
 * 
 * Internal vortex function. This function allows vortex_init to
 * initialize vortex listener module.
 **/
void vortex_listener_init ()
{
	/* init lock */
	g_static_mutex_lock (&__listener_mutex);
	
	/* create listener locker */
	if (vortex_listener_wait_lock == NULL) {
		vortex_listener_wait_lock = g_async_queue_new ();
	}

	/* unlock */
	g_static_mutex_unlock (&__listener_mutex);
	return;
}

/** 
 * @brief Allows to configure a handler that is executed once a
 * connection have been accepted.
 *
 * The handler to be configured could be used as a way to get
 * notifications about connections created, but also as a filter for
 * connections that must be dropped.
 * 
 * @param on_accepted The handler to be executed.
 *
 * @param data User space data to be passed in to the handler
 * executed.
 */
void          vortex_listener_set_on_connection_accepted (VortexOnAcceptedConnection on_accepted, 
							  gpointer data)
{
	g_return_if_fail (on_accepted);

	/* store handler configuration */
	__vortex_listener_on_accept      = on_accepted;
	__vortex_listener_on_accept_data = data;

	/* nothing more */
	return;
}


/** 
 * @brief Support function for Vortex Library listeners, that reads a
 * xml file that contains listener information and starts the
 * listener.
 *
 * The file the function will try to find is <b>conf.xml</b>. It will
 * lookup for all paths configured through: 
 * 
 *   - \ref vortex_support_find_data_file
 * 
 * Here is an example for a configuration file:
 * \code
 * <vortex-listener>
 *   <listener>
 *     <hostname>0.0.0.0</hostname>
 *     <port>44000</port>
 *   </listener>
 * </vortex-listener>
 * \endcode
 *
 * If the function returns FALSE, a call to vortex_exit is not
 * required because it is already done by this function.
 * 
 * @return TRUE if the listener was started because the file was read
 * successfully otherwise FALSE is returned.
 */
gboolean          vortex_listener_parse_conf_and_start ()
{
	/* listener xml configuration */
	axlDoc   * doc;
	axlDtd   * dtd;
	axlError * error;

	axlNode  * listener;
	axlNode  * aux;

	/* host and port references */
	gchar    * host;
	gchar    * port;

	/* a full path reference to the file */
	gchar    * full_path_file;

	/* load the document */
	full_path_file = vortex_support_find_data_file ("conf.xml");
	doc            = axl_doc_parse_from_file (full_path_file, &error);
	g_free (full_path_file);

	if (doc == NULL) {
		
		/* drop a log message */
		g_printerr ("Unable to open conf.xml file, for host and port configuration. Error reported: %s\n",
			    axl_error_get (error));

		/* release the error reported */
		axl_error_free (error);
		
		/* terminate the library */
		vortex_exit ();
		return FALSE;
	}
	

	/* load the xml listener conf DTD */
	full_path_file = vortex_support_find_data_file ("vortex-listener-conf.dtd");
	if (full_path_file == NULL) {
		g_printerr ("Unable to open DTD validation file: vortex-listener-conf.dtd file, for host and port configuration.\n");
		
		/* terminate the library */
		vortex_exit ();
		return FALSE;
	}

	dtd            = axl_dtd_parse_from_file (full_path_file, &error);
	g_free (full_path_file);
	
	if (dtd == NULL) {
		/* drop a log message */
		g_printerr ("Unable to open conf.xml file, for host and port configuration. Error reported: %s\n",
			    axl_error_get (error));
		
		/* release the error reported */
		axl_error_free (error);

		/* release the document read */
		axl_doc_free (doc);
		
		/* terminate the library */
		vortex_exit ();
		return -1;
	}

	/* validate the xml listener document config */
	if (!axl_dtd_validate (doc, dtd, &error)) {
		g_printerr ("Unable to validate listener configuration. Check your settings. Error reported: %s\n",
			    axl_error_get (error));

		/* release the error reported */
		axl_error_free (error);

		/* release the document read */
		axl_doc_free (doc);

		/* release the DTD read */
		axl_dtd_free (dtd);
		
		/* terminate the library */
		vortex_exit ();
		return -1;
	}

	/* get a reference to the fist <listener> configuration */
	listener = axl_doc_get (doc, "/vortex-listener/listener");

	do {
		/* get the <host> content */
		aux  = axl_node_get_child_nth (listener, 0);
		host = axl_node_get_content_trim (aux, NULL);

		/* get the <port> content */
		aux  = axl_node_get_child_nth (listener, 1);
		port = axl_node_get_content_trim (aux, NULL);

		/* starts a listener for the given configuration */
		vortex_listener_new (host, port, NULL, NULL);

		/* now go for the next listener configuration */
	}while ((listener = axl_node_get_next (listener)) != NULL);
	
	/* listener configuration already read, release document and DTD */
	axl_doc_free (doc);

	axl_dtd_free (dtd);

	return TRUE;	
}

/* @} */
