/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#include <vortex.h>

static VortexHash * registered_profiles = NULL;
#define LOG_DOMAIN "vortex-profiles"

typedef struct _VortexProfile
{
	gchar                        * profile_name;
	VortexOnStartChannel           start;
	gpointer                       start_user_data;
	VortexOnStartChannelExtended   start_extended;
	gpointer                       start_extended_user_data;
	VortexOnCloseChannel           close;
	gpointer                       close_user_data;
	VortexOnFrameReceived          received;
	gpointer                       received_user_data;
	gchar                        * mime_type;
	gchar                        * transfer_encoding;

} VortexProfile;

/**
 * \defgroup vortex_profiles Vortex Profiles: Related functions to manage profiles inside Vortex Library.
 */

/**
 * \addtogroup vortex_profiles
 * @{
 */

void __destroy_profile_item (VortexProfile * profile)
{
	/* free the profile name if were defined */
	if (profile->profile_name != NULL)
		g_free (profile->profile_name);
	
	/* free the mime type value if were defined */
	if (profile->mime_type != NULL)
		g_free (profile->mime_type);

	/* free the transfer encoding value if were defined */
	if (profile->transfer_encoding != NULL)
		g_free (profile->transfer_encoding);

	/* free the profile node itself */
	g_free (profile);
	
	return;
}

/** 
 * @brief Init vortex profiles module. 
 *
 * Init vortex profiles internal state. This is actually called by
 * \ref vortex_init so this function is not useful for vortex library
 * consumers.
 */
void     vortex_profiles_init ()
{
	/* initialize hash table if weren't */
	if (registered_profiles == NULL)
		registered_profiles = vortex_hash_new_full (g_str_hash, g_str_equal, 
							    NULL, /* no key destruction function */
							    (GDestroyNotify) __destroy_profile_item); /* value destroy function */

	return;
}

/** 
 * @internal
 *
 * @brief Default start handler, that is, accept to create a channel
 * always.
 */
gboolean __vortex_profiles_default_start (gint               channel_num,
					  VortexConnection * connection,
					  gpointer           user_data)
{
	return TRUE;
}

/** 
 * @internal
 *
 * @brief Default close handler, that is, accept to close the channel
 * always.
 */
gboolean __vortex_profiles_default_close (gint               channel_num,
					  VortexConnection * connection,
					  gpointer           user_data)
{
	return TRUE;
}


/** 
 * @brief Allows to register a new profile inside the Vortex Library.
 *
 * Register a profile to be used on channel creation. Profiles are
 * used by Vortex channels to know which message to exchange and the
 * channel semantics. To be able to create vortex channels you must
 * register at least one profile.
 *
 * On vortex session establishment, vortex peer acting as server sends
 * to vortex peer acting as client (or initiator) its registered
 * profiles. This enable both sides to know if they can talk together.
 *
 * In order to get an idea about profile names to use you can see
 * actual reserved profiles name defined by BEEP RFC. Some on them
 * are:
 *
 *  <ul>
 *   <li>The one time password profile: <i>"http://iana.org/beep/SASL/OTP"</i>
 *   <li>The TLS profile: <i>"http://iana.org/beep/TLS"</i>
 *   <li>The profile used by Coyote layer at Af-Arch: <i>"http://fact.aspl.es/profiles/coyote_profile"</i>
 *  </ul>
 *
 * You must not free the <i>uri</i> argument memory. This function
 * does not make a copy so, if you free <i>uri</i> you will get
 * unexpected behaviors. It's recommended to use static values to
 * avoid problems.
 *
 * Associated to the profile being registered are the handlers. There
 * are three handlers to be executed during the profile life.
 * 
 * When a remote peer wants to create a channel sends a start channel
 * message. On that event, when an start message is received, the \ref
 * VortexOnStartChannel "start" handler will be executed. 
 * 
 * When a remote peer wants to close an already created channel it
 * sends the close channel message. The \ref VortexOnCloseChannel "close" 
 * handler is executed on that event.
 * 
 * For all frames received, \ref VortexOnFrameReceived "received" handler will be executed.
 *
 * You can get more info about these handlers \ref vortex_handlers
 * "here". You can also read the following \ref profile_example
 * "document to know more about profiles".
 * 
 * If you don't provide handlers for \ref VortexOnStartChannel "start"
 * and \ref VortexOnCloseChannel "close", a default handler will be used.
 * 
 * These default handlers always return TRUE, so, on channel creation
 * it will accept always and on channel closing it will accept
 * always.
 *
 * If you don't provide a received handler, all data received will be
 * acknowledged but dropped.
 *
 * There are some exception to previous information. First one is that
 * there are two levels of handlers to be executed on events for
 * channels with a profile. The first level is defined by previous
 * ones handlers. But a second level of handlers exists on per-channel
 * basis.
 *
 * This second level of handlers are executed for the same events
 * before the first level is executed.  If a handler for a particular
 * event is not found on second level, then the first handler is
 * executed.
 *
 * If a handler for a particular event is found on second level, the
 * is executed and first level handler not.
 *
 * This allows you to have several levels of handlers to be executed
 * in a general way and handlers to be executed for a particular
 * channel.
 * 
 * Another exceptions comes with the \ref VortexOnStartChannel "start"
 * handler. It only allows you to get notified about the channel
 * number requested to be created and the connection where it was
 * received the petition. On most cases it is not needed more
 * information. 
 * 
 * However, the start message have several optional attributes and
 * content element that are used by profile definitions to implement
 * things such as TLS. On that case it is need to get notified with
 * all data available. You can check \ref
 * vortex_profiles_register_extended_start function to know more about
 * this issue.
 *
 * You can check \ref vortex_manual_dispatch_schema "this section" to know more
 * about how second level, first level and wait reply method 
 * implemented by Vortex Library to receive frames.
 *
 * You can also check \ref profile_example "this tutorial" about
 * creating new profiles for your application. 
 * 
 * @param uri                 a profile name to register
 * @param start               a handler to control channel creation
 * @param start_user_data     user defined data to be passed in to start handler
 * @param close               a handler to control channel termination
 * @param close_user_data     user defined data to be passed in to close handler
 * @param received            a handler to control incoming frames
 * @param received_user_data  user defined data to be passed in to received handler
 */
void vortex_profiles_register (gchar                 * uri,
			       VortexOnStartChannel    start,
			       gpointer                start_user_data,
			       VortexOnCloseChannel    close,
			       gpointer                close_user_data,
			       VortexOnFrameReceived   received,
			       gpointer                received_user_data)
{
	VortexProfile * profile;

	/* find out if we already have registered this profile */
	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		/* create a new profile */
		profile                     = g_new0 (VortexProfile, 1);
		profile->profile_name       = g_strdup (uri);
		profile->start              = start ? start : __vortex_profiles_default_start;
		profile->start_user_data    = start_user_data;
		profile->close              = close ? close : __vortex_profiles_default_close;
		profile->close_user_data    = close_user_data;
		profile->received           = received;
		profile->received_user_data = received_user_data;

		/* register the new profile */
		vortex_hash_replace (registered_profiles, profile->profile_name, profile);

		return;
	}
	
	/* set new data for the given profile */
	profile->start              = start ? start : __vortex_profiles_default_start;
	profile->start_user_data    = start_user_data;
	profile->close              = close ? close : __vortex_profiles_default_close;
	profile->close_user_data    = close_user_data;
	profile->received           = received;
	profile->received_user_data = received_user_data;

	return;
}

/** 
 * @brief Allows to configure which is the default mime type to be set
 * for those channels created under the given profile.
 *
 * Default mime type and content transfer encoding used under beep is:
 *  - Content-Type: "application/octet-stream"
 *  - Content-Transfer-Encoding: binary
 *
 * They are used by default if no other value is set. 
 *
 * The function will perform a local copy of all the values provided
 * on this function, so you could safely unref them once the function
 * had finished.
 * 
 * @param uri The profile which the given values for the mime
 * configuration will be applied.
 *
 * @param mime_type Content-Type configuration value. You can use NULL
 * values for parameter. In the case the value is not, previous value
 * is conserved.
 *
 * @param transfer_encoding Content-Transfer-Encoding configuration
 * value. You can use NULL values for parameter. In the case the value
 * is not, previous value is conserved.
 */
void     vortex_profiles_set_mime_type           (gchar                 * uri,
						  gchar                 * mime_type,
						  gchar                 * transfer_encoding)
{
	VortexProfile * profile;

	/* find out if we already have registered this profile */
	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Attempting to configure mime and content type configuration over a profile not registered (%s)",
		       uri);
		return;
	}

	/* save current mime type if it is defined */
	if (mime_type != NULL) {
		if (profile->mime_type != NULL)
			g_free (profile->mime_type);
		profile->mime_type = g_strdup (mime_type);
	}
	
	/* save current transfer encoding type if it is defined */
	if (transfer_encoding != NULL) {
		if (profile->transfer_encoding != NULL)
			g_free (profile->transfer_encoding);
		profile->transfer_encoding = g_strdup (transfer_encoding);
	}
	
	/* just return */
	return;
}

/** 
 * @brief Allows to get current configuration for the default mime
 * type used for the given uri profile.
 * 
 * @param uri The BEEP unique profile identification to get mime type
 * configuration from.
 * 
 * @return Current mime type configuration or NULL. Value returned
 * must not be unrefered. It is a internal copy. NULL is returned only
 * if an internal error have happen, in this case, the profile
 * requested doesn't exist.
 */
gchar * vortex_profiles_get_mime_type (gchar * uri)
{
	VortexProfile * profile;

	/* find out if we already have registered this profile */
	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Attempting to get mime type configuration over a profile not registered (%s)",
		       uri);
		return NULL;
	}

	return (profile->mime_type != NULL) ? profile->mime_type : "application/octet-stream";
}

/** 
 * @brief Allows to get current configuration for the default content
 * transfer encoding used for the given uri profile.
 *
 * @param uri The BEEP unique profile identification to get content
 * transfer encoding configuration from.
 * 
 * @return Current content transfer encoding configuration or
 * NULL. Value returned must not be unrefered. It is a internal
 * copy. NULL is returned only if an internal error have happen, in
 * this case, the profile requested doesn't exist.
 */
gchar * vortex_profiles_get_transfer_encoding (gchar * uri)
{
	VortexProfile * profile;

	/* find out if we already have registered this profile */
	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Attempting to get mime type configuration over a profile not registered (%s)",
		       uri);
		return NULL;
	}

	return (profile->transfer_encoding != NULL) ? profile->transfer_encoding : "binary";
}

/** 
 * @brief Allows to register an extended version for the start message. 
 * 
 * Extended version supports the same than \ref VortexOnStartChannel
 * handler but also giving support for the profile requested, the
 * serverName attribute, the profile content and the profile content
 * encoding.
 *
 * Most profile definitions doesn't need the extended version for the
 * start message received. If the profile being implemented doesn't
 * need the profile uri, the serverName, the profile content and the
 * encoding you should use the \ref VortexOnStartChannel "start handler" 
 * defined at \ref vortex_profiles_register.
 *
 * Once defined the extended handler, the start handler defined at
 * \ref vortex_profiles_register will be ignored.
 * 
 * @param uri                      The uri profile to register the start channel extended
 * @param extended_start           The handler to be invoked when an start message is received
 * @param extended_start_user_data User defined data to be passed in to the handler.
 */
void     vortex_profiles_register_extended_start (gchar                        * uri,
						  VortexOnStartChannelExtended   extended_start,
						  gpointer                       extended_start_user_data)
{
	VortexProfile * profile;

	/* find out if we already have registered this profile */
	profile     = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		/* create a new profile if it weren't defined */
		profile                           = g_new0 (VortexProfile, 1);
		profile->profile_name             = uri;

		/* register extended start message */
		profile->start_extended           = extended_start;
		profile->start_extended_user_data = extended_start_user_data;

		/* save data */
		vortex_hash_replace (registered_profiles, profile->profile_name, profile);
	}
	
	/* register extended start message */
	profile->start_extended           = extended_start;
	profile->start_extended_user_data = extended_start_user_data;

	return;
}

/** 
 * @internal
 *
 * Invoke start handler for selected profile. As defined for
 * \ref vortex_profiles_register, the \ref VortexOnStartChannel "start handler"
 * must be defined or a default will be provided. Start handler is
 * invoked in order to know if a new channel using selected profile
 * can be created.
 * 
 * First it is check for an extended start handler. If not defined,
 * the start handlers is executed.
 * 
 * @param uri                   The uri profile identifying the channel requested to be created.
 * @param channel_num           The channel num requested to be created.
 * @param connection            The connection where the channel creation request has been received.
 * @param serverName            The serverName start attribute received.
 * @param profile_content       The profile content received.
 * @param profile_content_reply Optional profile content to be set on the start channel reply.
 * @param encoding              The profile content encoding.
 * 
 * @return TRUE if the channel was allowed to be created or FALSE if it fails.
 */
gboolean vortex_profiles_invoke_start (gchar            * uri, 
				       gint               channel_num, 
				       VortexConnection * connection,
				       gchar            * serverName, 
				       gchar            * profile_content, 
				       gchar           ** profile_content_reply, 
				       VortexEncoding     encoding)
{
	VortexProfile * profile;

	g_return_val_if_fail (connection,          FALSE);
	g_return_val_if_fail (channel_num >= 0,    FALSE);
	g_return_val_if_fail (uri,                 FALSE);
	g_return_val_if_fail (registered_profiles, FALSE);

	/* look up for the profile definition */
	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "requiring to invoke start handler on a not registered profile");
		return FALSE;
	}

	/* check for a start extended */
	if (profile->start_extended != NULL)
		return profile->start_extended (uri, channel_num, connection, serverName, 
						profile_content, profile_content_reply, encoding,
						profile->start_extended_user_data);

	/* if no defined, exec start handler (no check is required
	 * because it has, at least, the default start handler
	 * defined) */
	return profile->start (channel_num, connection, profile->start_user_data);
}

/** 
 * @brief Returns if the given profiles have actually defined the start
 * handler.
 * 
 * This function check for the definition of the \ref
 * VortexOnStartChannel "start" handler defined at \ref
 * vortex_profiles_register or an \ref VortexOnStartChannelExtended "extended start" 
 * handler defined at \ref vortex_profiles_register_extended_start.
 *
 * It also returns no start handler defined if <i>uri</i> is not a
 * registered profile.
 * 
 * @param uri The uri to check.
 * 
 * @return TRUE if close handler is defined or FALSE if not
 */
gboolean vortex_profiles_is_defined_start (gchar * uri)
{
	VortexProfile * profile;

	g_return_val_if_fail (uri && (* uri),      FALSE);
	g_return_val_if_fail (registered_profiles, FALSE);

	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "looking up for a start handler on a non-registered profile");
		return FALSE;
	}

	return ((profile->start != NULL) || (profile->start_extended != NULL));
}

/** 
 * @internal
 * 
 * Invokes the first level close handler. This handler is always defined due to a close
 * handler provided by the user or due to the default handler provided by the vortex
 * library. 
 * 
 */
gboolean vortex_profiles_invoke_close (gchar            * uri, 
				       gint               channel_num,
				       VortexConnection * connection)
{
	VortexProfile * profile;

	g_return_val_if_fail (connection,          FALSE);
	g_return_val_if_fail (channel_num >= 0,    FALSE);
	g_return_val_if_fail (uri,                 FALSE);
	g_return_val_if_fail (registered_profiles, FALSE);

	profile = vortex_hash_lookup (registered_profiles, uri);

	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "requiring to invoke close handler on a not registered profile");
		return FALSE;
	}

	return profile->close (channel_num, connection, profile->start_user_data);
}

/** 
 * @brief Returns if the given profiles have actually defined the close handler.
 *
 * It also returns no close handler defined if <i>uri</i> is not a
 * registered profile. 
 * 
 * @param uri the uri profile to check.
 * 
 * @return TRUE if close handler is defined or FALSE if not
 */
gboolean vortex_profiles_is_defined_close (gchar * uri)
{
	VortexProfile * profile;

	g_return_val_if_fail (uri && (* uri),      FALSE);
	g_return_val_if_fail (registered_profiles, FALSE);

	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "looking up for a close handler on a non-registered profile");
		return FALSE;
	}

	return (profile->close != NULL);
}

typedef struct _VortexProfileReceivedData {
	VortexProfile    * profile;
	gint               channel_num;
	VortexConnection * connection;
	VortexFrame      * frame;
}VortexProfileReceivedData;

gpointer __vortex_profiles_invoke_frame_received (gpointer __data)
{
	VortexProfileReceivedData * data         = __data;
	VortexProfile             * profile      = data->profile;
	VortexConnection          * connection   = data->connection;
	VortexFrame               * frame        = data->frame;
	VortexChannel             * channel      = NULL;
	gboolean                    is_connected = FALSE;

	/* get a reference to channel number so we can check after
	 * frame received handler if the channel have been closed.
	 * Once the frame received have finished this will help us to
	 * know if application space have issued a close channel. */
	gint                      channel_num  = data->channel_num;

	/* free paramters data */
	g_free (data);

	/* record actual connection state */
	is_connected = vortex_connection_is_ok (connection, FALSE);

	if (! is_connected) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "invoking frame receive on a non-connected session");
		goto free_resources;
	}

	/* get a reference to the channel to manage and the channel number */
	channel      = vortex_connection_get_channel (connection, channel_num);

	/* sees the connection is not properly running because the
	 * channel reference returned was null. */
	if (channel == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			    "unable to get the channel reference (channel num=%d) over the connection (id=%d), avoiding frame deliver",
			    channel_num, vortex_connection_get_id (connection));
		goto free_resources;
	} /* end if */

	/* invoke frame received on this channel */
	profile->received (channel, connection, frame, profile->received_user_data);

	/* before frame receive handler we have to check if client
	 * have closed its connection */
	is_connected = vortex_connection_is_ok (connection, FALSE);

	if (! is_connected ) {
		/* the connection have been closed inside frame
		 * receive */
		goto free_resources;
	}

	/* The function __vortex_channel_0_frame_received_close_msg
	 * can be blocked awaiting to receive all replies
	 * expected. The following signal tries to wake up a possible
	 * thread blocked until last_reply_expected change. */
	if (vortex_connection_channel_exists (connection, channel_num))
		vortex_channel_signal_on_close_blocked (channel);

	/* log a message */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "invocation frame received handler for channel %d finished (first level: profiles)",
	       channel_num);

 free_resources:
	/* decrease reference counting */
	vortex_connection_unref (connection, "first level handler (frame received)");

	/* before invoke profile received frame free frame */
	vortex_frame_free (frame);

	/* return nothing */
	return NULL;
}

/** 
 * @internal
 * 
 * Invoke the first level frame received handler for the given profile
 * into the selected channel. This function also frees the vortex
 * frame passed in.
 *
 * @param uri 
 * @param channel_num 
 * @param connection 
 * @param frame 
 * 
 * @return TRUE if frame was delivered to a handler or FALSE if
 * frame was not delivered.
 */
gboolean vortex_profiles_invoke_frame_received (gchar            * uri,
						gint               channel_num,
						VortexConnection * connection,
						VortexFrame      * frame)
{
	VortexProfile             * profile;
	VortexProfileReceivedData * data;

	g_return_val_if_fail (uri && (*uri), FALSE);
	g_return_val_if_fail (channel_num >= 0, FALSE);
	g_return_val_if_fail (connection, FALSE);
	g_return_val_if_fail (frame, FALSE);
	g_return_val_if_fail (registered_profiles, FALSE);

	profile = vortex_hash_lookup (registered_profiles, uri);

	if ((profile == NULL) || (profile->received == NULL)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invoking frame received handler on a profile which haven't been defined");
		return FALSE;
	}

	/*
	 * invoke frame received handler. Because frame must be run
	 * inside a separated thread (as defined in user doc) we have
	 * to do that. This contrast with close and start invocation
	 * which have to return immediately is channel can be close or
	 * created.
	 *
	 * Moreover, the close and start channel are invoked in a
	 * separated thread because the channel 0's frame received
	 * handler (the only one channel which can process close and
	 * start message is the 0), is run inside a separated thread. See vortex_frame_
	 
	 * invoke frame received */
	data              = g_new0 (VortexProfileReceivedData, 1);
	data->profile     = profile;
	data->channel_num = channel_num;
	data->connection  = connection;
	data->frame       = frame;

	/* increase connection reference counting to avoid reference
	 * loosing if connection is closed */
	vortex_connection_ref (connection, "first level handler (frame received)");

	vortex_thread_pool_new_task (__vortex_profiles_invoke_frame_received, data);
	
	return TRUE;
}

/** 
 * @brief  Returns if the given profiles have actually defined the received  handler. 
 * 
 * It also returns no received handler defined if <i>uri</i> is not a
 * registered profile.
 * 
 * @param uri 
 * 
 * @return TRUE if received handler is defined or FALSE if not
 */
gboolean vortex_profiles_is_defined_received (gchar * uri)
{
	VortexProfile * profile;

	g_return_val_if_fail (uri && (* uri),      FALSE);
	g_return_val_if_fail (registered_profiles, FALSE);

	profile = vortex_hash_lookup (registered_profiles, uri);
	if (profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "looking up for a received handler on a non-registered profile");
		return FALSE;
	}

	return (profile->close != NULL);
}



void __get_actual_list (gpointer       key,
			gpointer       value,
			gpointer       user_data)
{
	GList ** ptr_list = user_data;
	GList *      list = (GList *) (*ptr_list);

	list              = g_list_append (list, key);
	(*ptr_list)       = list;

	return;
}

/** 
 * @brief Return actual profile registered list.
 * 
 * The GList returned contains all uri that have been registered by
 * using \ref vortex_profiles_register. 
 * 
 * You must free returned after using it. Use g_list_free.  Remember
 * not to remove any value inside returned list.
 * 
 * @return the actual profile registered list.
 */
GList * vortex_profiles_get_actual_list ()
{
	GList * result = NULL;
	g_return_val_if_fail (registered_profiles, NULL);

	vortex_hash_foreach (registered_profiles, __get_actual_list, &result);

	return result;
}

/** 
 * @brief Return the actual number of profiles registered. 
 * 
 * @return number of profiles registered.
 */
gint    vortex_profiles_registered ()
{
	if (registered_profiles == NULL)
		return 0;
	return vortex_hash_size (registered_profiles);
}

/** 
 * @brief Return if a profile identifier is registered. 
 * 
 * @param uri The unique URI identifier used to lookup.
 * 
 * @return TRUE if the profile is registered or FALSE if not. 
 */
gboolean vortex_profiles_is_registered (gchar * uri)
{
	g_return_val_if_fail (registered_profiles, FALSE);

	return (vortex_hash_lookup (registered_profiles, uri) != NULL);
}

/** 
 * @internal
 * @brief Clean up vortex profile module (memory allocated).
 *
 * This is an internal memory maintenance function that is not useful
 * for Vortex Library API consumers. It is called from vortex_exit
 * function at the right time.
 */
void     vortex_profiles_cleanup                 ()
{
	if (registered_profiles == NULL)
		return;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "stopping vortex profiles module");

	/* destroy the hash, including profile items. It is done by
	 * configuring __destroy_profile_item function at hash time
	 * creation. */
	vortex_hash_destroy (registered_profiles);

	return;
}
/* @} */
