/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#include <vortex.h>

#define LOG_DOMAIN "vortex-reader"


/**
 * \defgroup vortex_reader Vortex Reader: The module that reads you frames. 
 */

/**
 * \addtogroup vortex_reader
 * @{
 */

GAsyncQueue        * vortex_reader_queue    = NULL;
GAsyncQueue        * vortex_reader_stopped  = NULL;
GList              * con_list    = NULL;
GList              * srv_list    = NULL;

typedef enum {CONNECTION, LISTENER, TERMINATE} WatchType;
typedef struct _VortexReaderData {
	WatchType          type;
	VortexConnection * connection;
	VORTEX_SOCKET      listener;
}VortexReaderData;

/** 
 * @internal
 *
 * @brief Internal value that allows the vortex to accept incoming
 * messages with msgno=1 rather than the value that should be used 0.
 *
 * See \ref vortex_reader_allow_msgno_starting_from_1.
 */
gboolean __vortex_reader_accept_msgno_startig_from_1 = TRUE;

/**
 * @internal
 * 
 * Join incoming frames into a single one if complete flag is
 * defined. It also check incoming frame received so if some error is
 * found the connection is set to be not connected and a FALSE is returned.
 * 
 * Return value: TRUE if frame joining was successful or FALSE if fails
 **/
gboolean __vortex_reader_join_frames (VortexChannel * channel, VortexFrame ** frame)
{
	VortexFrame * previous_one;
	VortexFrame * new_one;

	/* get a reference to previous frame */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting previous frame for the channel: %d",
	       vortex_channel_get_number (channel));
	previous_one = vortex_channel_get_previous_frame (channel, (* frame));

	/* check if we have the first frame with more flag set */
	if (previous_one == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "there is no previous frame on channel: %d", 
		       vortex_channel_get_number (channel));
		vortex_channel_set_previous_frame (channel, NULL, (* frame));
		return TRUE;
	}

	/* get a new reference to a new allocated frame which have the
	 * content of previous_one and frame */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "joining frames (previous frame was found)..");
	new_one      = vortex_frame_join_extending (previous_one, (* frame));

	/* if new_one is NULL this means frame joining have failed,
	 * broken beep peer */
	if (new_one == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received a frame part of a fragment which have incompatible values (sync lost, closing session)..");
		__vortex_connection_set_not_connected (vortex_channel_get_connection (channel),
						       "received a frame part of a fragment which have incompatible values");
		/* do not free any frame, it is already done by
		 * vortex_channel_free caller for the frame received
		 * as a parameter. */
		return FALSE;
	}

	/* if we have a new frame build which has the more flag
	 * activated, store it  */
	if (vortex_frame_get_more_flag (new_one)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "the frame have the more flag activated, set it as the new previous frame..");
		vortex_channel_set_previous_frame (channel, previous_one, new_one);
	}

	/* now free resources allocated by previous one and actual
	 * frame */
	if (previous_one)
		vortex_frame_free (previous_one);
	
	/* free previous frame and return back the new one */
	vortex_frame_free ((* frame));

	/* update callers reference to the right frame generated by
	 * joining previous frame and the received */
	(* frame) = new_one;

	/* return if the joining process was ok */
	return TRUE;
}



/** 
 * @internal
 *
 * @brief NUL specific frame checkings
 * 
 * @param frame The frame to check
 * @param connection The connection where the frame was received.
 * 
 * @return TRUE if all checks have been passed. Otherwise FALSE is returned.
 */
gboolean __vortex_reader_process_socket_check_nul_frame (VortexFrame      * frame, 
							 VortexConnection * connection)
{
	
	switch (vortex_frame_get_type (frame)) {
	case VORTEX_FRAME_TYPE_NUL:
		/* check zero payload size at NUL frame */
		if (vortex_frame_get_payload_size (frame) != 0) {
			__vortex_connection_set_not_connected (connection, 
							       "Received header starts with \"NUL\", and the payload is non-zero");
			return FALSE;
		}

		/* check more flag at NUL frame reply. */
		if (vortex_frame_get_more_flag (frame)) {
			__vortex_connection_set_not_connected (connection, 
							       "Received header starts with \"NUL\", and the continuator indicator is *");
			return FALSE;
		}
		break;
	default:
		return TRUE;
	}
	return TRUE;
}

/** 
 * @internal
 *
 * Updates current window size to track SEQ frames status. This
 * function not only updates current incoming buffer status but also
 * reports a new SEQ frame to notify remote side that the incoming
 * buffer have changed.
 * 
 * @param channel The channel where the incoming buffer change was detected.
 *
 * @param frame The frame that contains application data that is
 * meaningful to perform a right SEQ frame counting.
 */
void __vortex_reader_update_incoming_buffer_and_notify (VortexConnection  * connection,
							VortexChannel     * channel,
							VortexFrame       * frame)
{

	VortexWriterData * writer;
	VortexChannel    * channel0;
	gint               ackno, window;

	/* now, we have to update current incoming max seq no allowed
	 * for future checkings on this channel and to generate a SEQ
	 * frame signaling this state to the remote peer. */
	switch (vortex_frame_get_type (frame)) {
	case VORTEX_FRAME_TYPE_MSG:
	case VORTEX_FRAME_TYPE_ANS:
	case VORTEX_FRAME_TYPE_RPY:
	case VORTEX_FRAME_TYPE_ERR:
		/* only perform this task for those frames that have
		 * actually payload content to report and the current
		 * window size have changed. */
		if (vortex_channel_update_incoming_buffer (channel, frame, &ackno, &window)) {
			/* It seems there are something to report.
			 * 
			 * Now create a SEQ frame with higher priority to
			 * report remote side that the current max seq no
			 * value have been updated. */
			writer            = g_new0 (VortexWriterData, 1);
			writer->type      = VORTEX_FRAME_TYPE_SEQ;
			writer->msg_no    = 0;
			writer->the_frame = vortex_frame_seq_build_up_from_params (vortex_channel_get_number (channel),
										   ackno,
										   window);
			writer->the_size  = strlen (writer->the_frame);
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "notifying remote side that current buffer status is %s",
			       writer->the_frame);
			/* Queue the vortex writer message to be sent with
			 * higher priority, currently, the following function
			 * will check if the packet to send is a SEQ frame to
			 * apply this priority. */
			channel0 = vortex_connection_get_channel (connection, 0);
			if (!vortex_sequencer_queue_and_signal_sem (channel0, writer)) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to queue a SEQ frame");
				/* deallocate memory on error, because no one
				 * will do it. */
				vortex_frame_free (frame);
				return;
			}
		} /* end if */
		break;
	default:
		/* do nothing */
		break;
	}
	
	return;
}

/** 
 * @internal
 *
 * @brief Allows to check if the given channel is expecting to
 * received the message number that is carrying the frame.
 *
 * This function not only perform the checks required but also allows
 * to be compatible with non-conformance BEEP implementations that
 * uses as starting message number the value 1 rather than 0 for the
 * very first one, even not being the channel 0.
 * 
 * @param channel The channel where the message was received.
 * @param frame   The frame received that has to be checked.
 * 
 * @return TRUE if the reader check is passed. Otherwise FALSE is
 * returned.
 */
gboolean vortex_channel_check_incoming_msgno (VortexChannel * channel, 
					      VortexFrame * frame)
{
	/* get a reference to the connection that holds the channel */
	VortexConnection * connection = vortex_channel_get_connection (channel);

	/* compatibility test, out of the BEEP standard to interop
	 * with BEEP implementations that uses as a starting message
	 * the value 1. */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "about to checking expected message to be received on this channel");
	if (vortex_channel_get_next_expected_msg_no (channel) == 0 &&
	    vortex_frame_get_msgno (frame) == 1) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, 
		       "received an starting message for the channel value that begins with 1 rather than 0");

		if (__vortex_reader_accept_msgno_startig_from_1) {
			/* because the remote peer is missing the very first
			 * message, we have to report this to the channel
			 * values so, next message are in sequence. */
			
			/* next sentence makes the vortex library to assume
			 * that the message msgno=0 have been received. */
			vortex_channel_update_status_received (channel, 0, UPDATE_MSG_NO);
			vortex_channel_update_status          (channel, 0, UPDATE_RPY_NO);
			
			return TRUE;
		}
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "if you want to make vortex reader to accept message with msgno starting from 1, you can use vortex_reader_allow_msgno_starting_from_1");
		
	}

	/* perform the check */
	if (vortex_channel_get_next_expected_msg_no (channel) != vortex_frame_get_msgno (frame)) {
		__vortex_connection_set_not_connected (connection, "expected message number for channel wasn't found");
		
		/* log a critical log */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "expected message number %d for channel %d wasn't found, but received %d",
		       vortex_channel_get_next_expected_msg_no (channel), 
		       vortex_channel_get_number (channel),
		       vortex_frame_get_msgno (frame));
		
		/* free the frame */
		vortex_frame_free (frame);
		return FALSE;
	}
	/* check passed */
	return TRUE;
}

/**
 * @internal
 * 
 * The main purpose of this function is to dispatch received frames
 * into the appropriate channel. It also makes all checks to ensure the
 * frame receive have all indicators (seqno, channel, message number,
 * payload size correctness,..) to ensure the channel receive correct
 * frames and filter those ones which have something wrong.
 *
 * This function also manage frame fragment joining. There are two
 * levels of frame fragment managed by the vortex reader.
 * 
 * We call the first level of fragment, the one described at RFC3080,
 * as the complete frame which belongs to a group of frames which
 * conform a message which was splitted due to channel window size
 * restrictions.
 *
 * The second level of fragment happens when the vortex reader receive
 * a frame header which describes a frame size payload to receive but
 * not all payload was actually received. This can happen because
 * vortex uses non-blocking socket configuration so it can avoid DOS
 * attack. But this behavior introduce the asynchronous problem of
 * reading at the moment where the whole frame was not received.  We
 * call to this internal frame fragmentation. It is also supported
 * without blocking to vortex reader.
 *
 * While reading this function, you have to think about it as a
 * function which is executed for only one frame, received inside only
 * one channel for the given connection.
 *
 * @param connection the connection which have something to be read
 * 
 **/
void __vortex_reader_process_socket (VortexConnection * connection)
{

	VortexFrame      * frame;
	VortexFrameType    type;
	VortexChannel    * channel;
	gboolean           more;
	gchar            * raw_frame;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "something to read");

	/* check if there are pre read handler to be executed on this 
	   connection. */
	if (vortex_connection_is_defined_preread_handler (connection)) {
		/* if defined preread handler invoke it and return. */
		vortex_connection_invoke_preread_handler (connection);
		return;
	}

	/* before doing anything, check if the connection is broken */
	if (! vortex_connection_is_ok (connection, FALSE))
		return;

	/* read all frames received from remote site */
	frame   = vortex_frame_get_next (connection);
	if (frame == NULL) 
		return;

	/* get frame type to avoid calling everytime to
	 * vortex_frame_get_type */
	type    = vortex_frame_get_type (frame);

	/* NOTE: After this point, frame received is
	 * complete. vortex_frame_get_next function takes cares about
	 * joining frame fragments. */

	/* check for debug to throw some debug messages.*/
	if (vortex_log_is_enabled ()) {
		raw_frame = vortex_frame_get_raw_frame (frame);
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame received (before all filters)\n%s",
		       raw_frame);
		g_free (raw_frame);
	}
	
	/* check if this connection is being initially accepted */
	if (GPOINTER_TO_INT (vortex_connection_get_data (connection, "initial_accept"))) {
		/* it doesn't matter to have a connection accepted or
		 * not accepted to the vortex reader mission, so
		 * simply call second step accept and return.  */
		__vortex_listener_second_step_accept (frame, connection);
		return;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "passed initial accept stage");

	/* channel exists, get a channel reference */
	channel = vortex_connection_get_channel (connection,
						 vortex_frame_get_channel (frame));
	if (channel == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received a frame referring to a non-opened channel, closing session");
		__vortex_connection_set_not_connected (connection, 
						       "received a frame referring to a non-opened channel, closing session");
		/* free the frame */
		vortex_frame_free (frame);
		return;		
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "passed connection existence stage");	

	/* check message numbers and reply numbers sequencing */
	switch (type) {
	case VORTEX_FRAME_TYPE_MSG:
		/* MSG frame type: check if msgno is correct */
		if (!vortex_channel_check_incoming_msgno (channel, frame))
			return;
		break;
	case VORTEX_FRAME_TYPE_RPY:
	case VORTEX_FRAME_TYPE_ERR:
	case VORTEX_FRAME_TYPE_ANS:
	case VORTEX_FRAME_TYPE_NUL:
		/* RPY or ERR frame type: check if reply is expected */
		if (vortex_channel_get_next_expected_reply_no (channel) != vortex_frame_get_msgno (frame)) {

			__vortex_connection_set_not_connected (connection, "expected reply message previous to received");

			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "expected reply message %d previous to received %d",
			       vortex_channel_get_next_expected_reply_no (channel),
			       vortex_frame_get_msgno (frame));

			/* free the frame */
			vortex_frame_free (frame);
			return;
		}
		break;
	case VORTEX_FRAME_TYPE_SEQ:
		/* manage incoming SEQ frames, check if the received
		 * ackno value is inside the seqno range for bytes
		 * already sent. */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received a SEQ frame: SEQ %d %d %d",
		       vortex_frame_get_channel (frame), vortex_frame_get_seqno (frame),
		       vortex_frame_get_payload_size (frame));

		if (vortex_frame_get_seqno (frame) > vortex_channel_get_next_seq_no (channel)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "received a SEQ frame specifying a seqno reference value that wasn't used (ackno: %d > max seq no sent: %d)",
			       vortex_frame_get_seqno (frame), vortex_channel_get_next_seq_no (channel));
			
			__vortex_connection_set_not_connected (connection, 
							       "received a SEQ frame specifying a seqno reference value that wasn't used");
			/* unref frame due to errors */
			vortex_frame_free (frame);
			return;
		}
		
		/* check that the window size is not bigger than MAX_BUFFER_SIZE */
		if (MAX_BUFFER_SIZE < vortex_frame_get_payload_size (frame)) {
			__vortex_connection_set_not_connected (connection, "received a SEQ frame specifying a not allowed value for the window size");

			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "received a SEQ frame specifying a not allowed value for the window size");
			
			/* unref frame due to errors */
			vortex_frame_free (frame);
			return;
		} 
		
		/* update information about the buffer state of the
		 * remote peer for the given channel */
		vortex_channel_update_remote_incoming_buffer (channel, frame);

		/* unref the SEQ frame because it is no longer
		 * needed */
		vortex_frame_free (frame);

		/* signal vortex sequencer to start re-sequencing
		 * previous hold messages on the channel updated */
		vortex_sequencer_signal_update (channel);
		return;
	default:
		/* do not make nothing (only checking msgno and
		 * rpyno) */
		break;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "passed message number checking stage");

	/* Check next sequence number, this check is always applied,
	 * for SEQ frames received this case is not reached. */
	if (vortex_channel_get_next_expected_seq_no (channel) !=
	    vortex_frame_get_seqno (frame)) {

		__vortex_connection_set_not_connected (connection, "expected seq no number for channel wan't found");

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "expected seq no %d for channel wasn't found, received %d",
		       vortex_channel_get_next_expected_seq_no (channel),
		       vortex_frame_get_seqno (frame));

		/* free the frame */
		vortex_frame_free (frame);
		return;
	}

	/* check other environment conditions */
	if ((type == VORTEX_FRAME_TYPE_NUL) && 
	    !__vortex_reader_process_socket_check_nul_frame (frame, connection)) {
		/* free the frame */
		vortex_frame_free (frame);	
		return;
	}	

	/* check if remote peer is sending more data over the channel
	 * than expected. The following check is done to ensure that
	 * the received frame is inside the sequence allowed due to
	 * buffer sizes. So, the check sums current sequence number of
	 * the message plus the sequence number added that is result
	 * from the frame size minus one unit. */
	if ((vortex_frame_get_seqno (frame) + vortex_frame_get_content_size (frame) - 1) > 
	    vortex_channel_get_max_seq_no_accepted (channel)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "Protocol violation, received a frame larger than the maximum buffer expected, your session will be closed (seq no: %d + size: %d > max seq no: %d",
		       vortex_frame_get_seqno (frame), vortex_frame_get_content_size (frame),
		       vortex_channel_get_max_seq_no_accepted (channel));
		__vortex_connection_set_not_connected (connection,
						       "Protocol violation, received a frame larger than the maximum buffer expected, your session will be closed");
		/* unref no longer needed frame */
		vortex_frame_free (frame);
		return;
	}

	/* update channel internal status for next incoming messages */
	switch (type) {
	case VORTEX_FRAME_TYPE_MSG:
		/* is a MSG frame type: update msgno and seqno */
		vortex_channel_update_status_received (channel, 
						       vortex_frame_get_content_size (frame), 
						       vortex_frame_get_more_flag (frame) ?
						       UPDATE_SEQ_NO :
						       UPDATE_SEQ_NO | UPDATE_MSG_NO);
		break;
	case VORTEX_FRAME_TYPE_RPY:
	case VORTEX_FRAME_TYPE_ERR:
		/* flag the channel with reply_processed so, close
		 * channel process can avoid to close the
		 * channel. This is necessary because the channel close
		 * process can be waiting to received (and process)
		 * all replies for message sent to be able to close
		 * the channel.  
		 * 
		 * And we have to do this flagging before updating
		 * channel status because close process will compare
		 * messages sent to message replies receive. */
		vortex_channel_flag_reply_processed (channel, FALSE);
		

		/* is a ERR or RPY type: update rpy no and seqno */
		vortex_channel_update_status_received (channel, 
						       vortex_frame_get_content_size (frame), 
						       vortex_frame_get_more_flag (frame) ? 
						       UPDATE_SEQ_NO : 
						       UPDATE_SEQ_NO | UPDATE_RPY_NO);
		break;
	case VORTEX_FRAME_TYPE_NUL:
		/* And we have to do this flagging before updating
		 * channel status because close process will compare
		 * messages sent to message replies receive. */
		vortex_channel_flag_reply_processed (channel, FALSE);

		/* is a ERR or RPY type: update rpy no and seqno */
		vortex_channel_update_status_received (channel, 0, UPDATE_RPY_NO);
		
		break;
	case VORTEX_FRAME_TYPE_ANS:
		/* look above description */
		vortex_channel_flag_reply_processed (channel, FALSE);

		/* is not a MSG frame type: update seqno */
		vortex_channel_update_status_received (channel, 
						       vortex_frame_get_content_size (frame), 
						       UPDATE_SEQ_NO);
		break;
	case VORTEX_FRAME_TYPE_UNKNOWN:
	default:
		/* nothing to do */
		break;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "passed channel update status due to frame received stage");
	/* now update current incoming buffers to track SEQ frames */
	__vortex_reader_update_incoming_buffer_and_notify (connection, channel, frame);

	/* If we have a frame to be joined then threat it instead of
	 * invoke frame received handler. This is done by checking for
	 * more flag. */
	more = vortex_frame_get_more_flag (frame);

	/* check complete frames flags:
	 * If more is enabled, the frame signal that more(*) frames have to come, or
	 * The more(.) flag is not enabled but we have a previously stored frame and the channel
	 * have the complete_flag activated, that is, the last frame of a series of frames [****.]  */
	if (more || (!more && vortex_channel_have_previous_frame (channel, frame))) { 

		/* join frames and return if fail or frame resulting
		 * from the join process is also a incomplete frame */
		if (__vortex_reader_join_frames (channel, &frame)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame fragmented received, joined with previous frame");

			/* get current more flag */
			more  = vortex_frame_get_more_flag (frame);

			/* only return and keep frame joining hide from the App level if
			 * the complete flag is activated and the more flag is FALSE */
			if (more && vortex_channel_have_complete_flag (channel)) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "more flag or completed flag detected, skipping to the next frame");
				return;
			}
		} else {
			/* frame joining have failed, log and unref
			 * it */
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "frame fragment received is not valid, giving up for this session");
			vortex_frame_free (frame);
			__vortex_connection_set_not_connected (vortex_channel_get_connection (channel), 
							       "frame fragment received is not valid, giving up for this session");
			return;
		}
		/* we have a complete frame, deliver it */
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "passed frame checking stage");

	/* invoke frame received handler for second level and, if not
	 * defined, the first level handler */
	if (vortex_channel_invoke_received_handler (connection, channel, frame)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame delivered on second (channel) level handler channel");
		return; /* frame was successfully delivered */
	}
	
	
	/* if not, try to deliver to first level. If second level
	 * invocation was ok, the frame have been freed. If not, the
	 * first level will do this */
	if (vortex_profiles_invoke_frame_received (vortex_channel_get_profile    (channel),
						   vortex_channel_get_number     (channel),
						   vortex_channel_get_connection (channel),
						   frame)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame delivered on first (profile) level handler channel");
		return; /* frame was successfully delivered */
	}
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
	       "unable to deliver incoming frame, no first or second level handler defined, dropping frame");

	/* in case of a frame reply */
	switch (type) {
	case VORTEX_FRAME_TYPE_RPY:
	case VORTEX_FRAME_TYPE_ERR:
	case VORTEX_FRAME_TYPE_ANS:
	case VORTEX_FRAME_TYPE_NUL:
		/* flag this frame over the channel to be delivered because a
		 * dead-lock can happen if we receive a close message at this
		 * point. We have to be able to close a channel having defined
		 * a frame receive handler for any level or not. */
		vortex_channel_flag_reply_processed (channel, TRUE);
		break;
	default:
		/* do nothing */
		break;
	}
	
	/* unable to deliver the frame, free it */
	vortex_frame_free (frame);

	/* that's all I can do */
	return;
}

/** 
 * @internal 
 *
 * @brief Classify vortex reader items to be managed, that is,
 * connections or listeners.
 * 
 * @param data The internal vortex reader data to be managed.
 * 
 * @return TRUE if the item to be managed was clearly read or FALSE if
 * an error on registering the item was produced.
 */
gboolean vortex_reader_register_watch (VortexReaderData * data)
{
	VortexConnection * connection;

	switch (data->type) {
	case CONNECTION:
		connection = data->connection;
		g_free (data);
		if (!vortex_connection_is_ok (connection, FALSE)) {
			/* check if we can free this connection */
			vortex_connection_unref (connection, "vortex reader (process)");
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received a non-valid connection, ignoring it");
			return FALSE;
		}
			
		/* now we have a first connection, we can start to wait */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "new connection to be watched (%d)", 
		       vortex_connection_get_socket (connection));
		con_list = g_list_append (con_list, connection);

		break;
	case LISTENER:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "new listener connection to be watched (%d)",
		       data->listener);
		srv_list = g_list_append (srv_list, GINT_TO_POINTER (data->listener));
		g_free (data);
		break;
	case TERMINATE:
		/* just unref vortex reader data */
		g_free (data);
		break;
	}		
	
	return TRUE;
}

/** 
 * @internal
 * @brief Read the next item on the vortex reader to be processed
 * 
 * Once an item is read, it is check if something went wrong, in such
 * case the loop keeps on going.
 * 
 * The function also checks for terminating vortex reader loop by
 * looking for TERMINATE value into the data->type. In such case FALSE
 * is returned meaning that no further loop should be done by the
 * vortex reader.
 *
 * @return TRUE to keep vortex reader working, FALSE if vortex reader
 * should stop.
 */
gboolean vortex_reader_read_queue ()
{
	VortexReaderData * data;
	gboolean           should_continue;

	do {
		data            = g_async_queue_pop (vortex_reader_queue);
		should_continue = (data->type != TERMINATE);
	}while (!vortex_reader_register_watch (data));

	return should_continue;
}

gboolean vortex_reader_read_pending ()
{
	VortexReaderData * data;
	gint               length;
	gboolean           should_continue = TRUE;

	length = g_async_queue_length (vortex_reader_queue);
	while (length > 0) {
		
		length--;
		data            = g_async_queue_pop (vortex_reader_queue);
		should_continue = (data->type != TERMINATE);
		vortex_reader_register_watch (data);
	}
	return should_continue;
}

VORTEX_SOCKET   __vortex_reader_build_set_to_watch (gpointer on_reading)
{
	GList            * new_list    = NULL;
	GList            * list_aux    = g_list_first (con_list);
	VORTEX_SOCKET      max_fds     = 0;
	VORTEX_SOCKET      fds         = 0;
	gint               actual_size = g_list_length (list_aux);
	VortexConnection * connection;

	/* read connection list */
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		connection = list_aux->data;
		if (!vortex_connection_is_ok (connection, FALSE)) {
			vortex_connection_unref (connection, "vortex reader (process)");
			continue;
		}
		
		/* get the socket to ge added and get its maximum
		 * value */
		fds        = vortex_connection_get_socket (connection);
		max_fds    = MAX (fds, max_fds);

		/* add the socket descriptor into the given on reading
		 * group */
		vortex_io_waiting_invoke_add_to_fd_group (fds, on_reading);

		/* add the connection to the list to be watched */
		new_list   = g_list_append (new_list, connection);
	}
	/* update connection list */
	g_list_free (con_list);
	con_list = new_list;

	/* read listener list */
	list_aux   = g_list_first (srv_list);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		fds        = GPOINTER_TO_INT (list_aux->data);
		max_fds    = MAX (fds, max_fds);

		/* add the socket descriptor */
		vortex_io_waiting_invoke_add_to_fd_group (fds, on_reading);
	}


	/* connection list watch */
	if (actual_size != g_list_length (con_list)) {

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "connection watched by the reader");
		list_aux        = g_list_first (con_list);

		for (;list_aux; list_aux = g_list_next (list_aux)) {
			connection = list_aux->data;
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "   connection to id:%d) %s:%s", 
			       vortex_connection_get_id   (connection),
			       vortex_connection_get_host (connection),
			       vortex_connection_get_port (connection));
		}
	}

	return max_fds;
	
}

void __vortex_reader_check_connection_list (gpointer on_reading)
{
	GList             * list_aux   = g_list_first (con_list);
	GList             * new_list   = NULL;
	VORTEX_SOCKET       fds        = 0;
	VortexConnection  * connection = NULL;

	for (; list_aux; list_aux = g_list_next (list_aux)) {

		/* check if we have to keep on listening on this
		 * connection */
		connection = list_aux->data;
		if (!vortex_connection_is_ok (connection, FALSE)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "free non-ok connection resources");
			vortex_connection_unref (connection, "vortex reader (process)");
			continue;
		}
		
		/* get the connection and socket. */
	        fds = vortex_connection_get_socket (connection);
		
		/* ask if this socket have changed */
		if (vortex_io_waiting_invoke_is_set_fd_group (fds, on_reading)) {
			__vortex_reader_process_socket (connection);
		}

		/* check the connection */
		if (!vortex_connection_is_ok (connection, FALSE)) {
			vortex_connection_unref (connection, "vortex reader (process)");
			continue;
		}

		/* build up new connection list */
		new_list = g_list_append (new_list, connection);

	}


	/* set the new connection list to be watched */
	g_list_free (con_list);
	con_list = g_list_first (new_list);

	return;
}

void __vortex_reader_check_listener_list (gpointer on_reading)
{
	GList * list_aux = g_list_first (srv_list);
	gint    fds      = 0;

	for (; list_aux; list_aux = g_list_next (list_aux)) {
		
		/* get the connection and socket. */
		fds        = GPOINTER_TO_INT (list_aux->data);
		if (vortex_io_waiting_invoke_is_set_fd_group (fds, on_reading)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "listener (%d) have requests, processing..", fds);
			vortex_listener_accept_connections (fds);
		}
	}
	
	return;
}

/** 
 * @internal
 *
 * @brief Internal function called to stop vortex reader and cleanup
 * memory used.
 * 
 */
void __vortex_reader_stop_process (gpointer on_reading)
{

	/* stop vortex reader process unreferring already managed
	 * connections */

	g_async_queue_unref (vortex_reader_queue);

	/* unref listener connections */
	if (g_list_first (srv_list))
		g_list_free (g_list_first (srv_list));

	/* unref initiators connections */
	if (g_list_first (con_list))
		g_list_free (g_list_first (con_list));

	/* unref IO waiting object */
	vortex_io_waiting_invoke_destroy_fd_group (on_reading); 

	/* signal that the vortex reader process is stopped */
	g_async_queue_push (vortex_reader_stopped, GINT_TO_POINTER (1));

	return;
}

gpointer __vortex_reader_run (GAsyncQueue * vortex_reader_queue)
{
	gpointer           on_reading;
	VORTEX_SOCKET      max_fds     = 0;
	VORTEX_SOCKET      result;
	gint               error_tries = 0;

	/* initialize the read set */
	on_reading  = vortex_io_waiting_invoke_create_fd_group ();

	/* first step. Waiting blocked for our first connection to
	 * listen */
 __vortex_reader_run_first_connection:
	if (!vortex_reader_read_queue ()) {
		/* seems that the vortex reader main loop should
		 * stop */
		__vortex_reader_stop_process (on_reading);
		return NULL;
	}

	while (TRUE) {
		/* reset descriptor set */
		vortex_io_waiting_invoke_clear_fd_group (on_reading);

		/* build socket descriptor to be read */
		max_fds = __vortex_reader_build_set_to_watch (on_reading);

		if ((g_list_length (con_list) == 0) && (g_list_length (srv_list) == 0)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "no more connection to watch for, putting into sleep thread");
			goto __vortex_reader_run_first_connection;
		}

		/* perform IO blocking wait for read operation */
		result = vortex_io_waiting_invoke_wait (on_reading, max_fds, READ_OPERATIONS);

		/* check for timeout error */
		if (result == -1 || result == -2)
			continue;

		/* check errors */
		if ((result < 0) && (errno != 0)) {

			error_tries++;
			if (error_tries == 2) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
					    "tries have been reached on reader, error was=(errno=%d): %s exiting..",
					    errno, g_strerror (errno));
				return NULL;
			}
			continue;
		}

		/* check for fatal error */
		if (result == -3) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "fatal error received from io-wait function, exiting from vortex reader process..");
			__vortex_reader_stop_process (on_reading);
			return NULL;
		}


		/* check for each listener */
		__vortex_reader_check_listener_list (on_reading);

		/* check for each connection to be watch is it have check */
		__vortex_reader_check_connection_list (on_reading);

		/* we have finished the connection dispatching, so
		 * read the pending queue elements to be watched */
		
		/* reset error tries */
		error_tries = 0;

		/* read new connections to be managed */
		if (!vortex_reader_read_pending ()) {
			__vortex_reader_stop_process (on_reading);
			return NULL;
		}

		if (result > 0) 
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "watching %d connection", 
			       g_list_length (con_list) + g_list_length (srv_list));
	}
	return NULL;
}

/**
 * @internal
 * 
 * Adds a new connection to be watched on vortex reader process. This
 * function is for internal vortex library use.
 **/
void vortex_reader_watch_connection (VortexConnection * connection)
{
	VortexReaderData * data;
	g_return_if_fail (vortex_connection_is_ok (connection, FALSE));
	g_return_if_fail (vortex_reader_queue);

	if (!vortex_connection_set_nonblocking_socket (connection)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to set non-blocking I/O operation, sking connection registration, closing session");
 		return;
	}

	/* increase reference counting */
	vortex_connection_ref (connection, "vortex reader (process)");
       
	/* prepare data to be queued */
	data             = g_new0 (VortexReaderData, 1);
	data->type       = CONNECTION;
	data->connection = connection;

	/* push data */
	g_async_queue_push (vortex_reader_queue, data);

	return;
}

/**
 * @internal
 *
 * Install a new listener to watch for new incoming connections.
 **/
void vortex_reader_watch_listener   (VORTEX_SOCKET listener)
{
	VortexReaderData * data;
	g_return_if_fail (listener > 0);
	
	/* prepare data to be queued */
	data             = g_new0 (VortexReaderData, 1);
	data->type       = LISTENER;
	data->listener   = listener;

	/* push data */
	g_async_queue_push (vortex_reader_queue, data);

	return;
}

/**
 * @internal
 * 
 * Creates the reader thread process. It will be waiting for any
 * connection that have changed to read its connect and send it
 * appropriate channel reader.
 * 
 * Return value: 
 **/
void vortex_reader_run () 
{
	/* create vortex queue */
	vortex_reader_queue    = g_async_queue_new ();
	vortex_reader_stopped  = g_async_queue_new ();

	/* create the vortex reader main thread */
	g_thread_create ((GThreadFunc) __vortex_reader_run,
			 (gpointer)    vortex_reader_queue, 
			 FALSE, NULL);
}

/** 
 * @internal
 * @brief Cleanup vortex reader process.
 */
void vortex_reader_stop ()
{
	VortexReaderData * data;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "stopping vortex reader ..");

	/* create a bacon to signal vortex reader that it should stop
	 * and unref resources */
	data       = g_new0 (VortexReaderData, 1);
	data->type = TERMINATE;

	/* push data */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "pushing data stop signal..");
	g_async_queue_push (vortex_reader_queue, data);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "signal sent reader ..");

	/* waiting until the reader is stoped */
	g_async_queue_pop (vortex_reader_stopped);
	g_async_queue_unref (vortex_reader_stopped);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "vortex reader process stopped");

	return;
}

/** 
 * @brief Makes the vortex reader to allow accepting the first message
 * over a channel with msgno=1.
 *
 * BEEP RFC3080 defines that every message sent over a channel have a
 * msgno value ranging from 0 up to 2^(31) - 1. This means that the
 * very first message sent over a channel should have a value 0 and
 * steps into the next value by one unit.
 *
 * However, to the previous rule there are one exception: the channel
 * 0. This administrative channel, that allows BEEP to control and
 * manage channels, do not ever send the message 0. This is because it
 * is implicit to the connection. 
 *
 * Once a BEEP peer is connected, the initial greeting is exchanged,
 * without waiting to a first message (with msgno=0) required to do
 * so. But this initial greeting is carried on a RPY message with
 * msgno=0. This creates a great confusion because every RPY message
 * should reply to a MSG message sent. 
 *
 * Thus, the first message exchanged under the channel 0 have a value
 * msgno=1 because the first message MSG with msgno=0 is never sent,
 * then an RPY with msgno=0 is sent (the initial greeting), and the
 * the next message to be exchanged is the msgno=1.
 *
 * Well, previous description is taken by some BEEP implementation to
 * be the normal situation for every channel. This is clearly
 * wrong. However, it is preferred to interop with other libraries
 * rather blame them (to interop is a lovely thing that should be
 * pursued, no one wins when blame occurs).
 *
 * By default, the library is configured to accept first messages
 * starting from msgno=1. You can also be strict and set a FALSE value
 * to this function.
 * 
 * @param value TRUE to allow messages received on a channel to start
 * with 1 rather than 0.
 */
void vortex_reader_allow_msgno_starting_from_1 (gboolean value)
{
	__vortex_reader_accept_msgno_startig_from_1 = value;
	return;
}

/* @} */
