/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>
#include <unistd.h> /* gethostname */

/**
 * \defgroup vortex_sasl Vortex SASL: SASL profile support and related functions
 */

/**
 * \addtogroup vortex_sasl
 * @{
 */

/** 
 * @internal
 * @brief Key definitions which are global to all implementations
 */
#define SASL_AUTHID                     "sasl:authid"
#define SASL_AUTHZID                    "sasl:authzid"
#define SASL_PASSWORD                   "sasl:password"
#define SASL_REALM                      "sasl:realm"
#define SASL_ANONYMOUS_TOKEN            "sasl:anonymous:token"
#define SASL_IS_AUTHENTICATED           "sasl:is:authenticated"
#define SASL_METHOD_USED                "sasl:method:used"
#define SASL_PROFILE_BEGIN_NEGOCIATED   "sasl:profile:being-negociated"

#ifdef ENABLE_SASL_SUPPORT
#include <gsasl.h>
#define SASL_CONTEXT         "gsasl:ctx"
#define SASL_SESSION         "gsasl:session"
#define SASL_CONTEXT_DESTROY  (GDestroyNotify)gsasl_done 
#endif

#define LOG_DOMAIN "vortex-sasl"

/** 
 * @internal
 * @brief ANONYMOUS validation handler. This is actually invoked by
 * the library to notify user space that a ANONYMOUS auth request was
 * received.
 */
VortexSaslAuthAnonymous __vortex_sasl_anonymous_auth_handler = NULL;

/** 
 * @internal
 * @brief PLAIN validation handler. This is actually invoked by
 * the library to notify user space that a PLAIN auth request was
 * received.
 */
VortexSaslAuthPlain     __vortex_sasl_plain_auth_handler = NULL;

/** 
 * @internal
 * @brief CRAM-MD5 validation handler. This is actually invoked by
 * the library to notify user space that a CRAM-MD5 auth request was
 * received.
 */
VortexSaslAuthCramMd5     __vortex_sasl_cram_md5_auth_handler = NULL;

/** 
 * @internal
 * @brief DIGEST-MD5 validation handler. This is actually invoked by
 * the library to notify user space that a DIGEST-MD5 auth request was
 * received.
 */
VortexSaslAuthDigestMd5     __vortex_sasl_digest_md5_auth_handler = NULL;

/** 
 * @internal
 * @brief EXTERNAL validation handler. This is actually invoked by
 * the library to notify user space that a EXTERNAL auth request was
 * received.
 */
VortexSaslAuthExternal __vortex_sasl_external_auth_handler = NULL;

/** 
 * @internal
 * @brief ANONYMOUS validation handler. This is actually invoked by
 * the library to notify user space that a ANONYMOUS auth request was
 * received. Passes the user-defined pointer.
 */
VortexSaslAuthAnonymousFull __vortex_sasl_anonymous_auth_handler_full = NULL;

/** 
 * @internal
 * @brief PLAIN validation handler. This is actually invoked by
 * the library to notify user space that a PLAIN auth request was
 * received. Passes the user-defined pointer.
 */
VortexSaslAuthPlainFull     __vortex_sasl_plain_auth_handler_full = NULL;

/** 
 * @internal
 * @brief CRAM-MD5 validation handler. This is actually invoked by
 * the library to notify user space that a CRAM-MD5 auth request was
 * received. Passes the user-defined pointer.
 */
VortexSaslAuthCramMd5Full     __vortex_sasl_cram_md5_auth_handler_full = NULL;

/** 
 * @internal
 * @brief DIGEST-MD5 validation handler. This is actually invoked by
 * the library to notify user space that a DIGEST-MD5 auth request was
 * received. Passes the user-defined pointer.
 */
VortexSaslAuthDigestMd5Full     __vortex_sasl_digest_md5_auth_handler_full = NULL;

/** 
 * @internal
 * @brief EXTERNAL validation handler. This is actually invoked by
 * the library to notify user space that a EXTERNAL auth request was
 * received. Passes the user-defined pointer.
 */
VortexSaslAuthExternalFull __vortex_sasl_external_auth_handler_full = NULL;


/** 
 * @brief Allows to check and initialize SASL support for Vortex Library.
 * 
 * Because Vortex Library SASL support building could be disabled, you
 * should do a call to \ref vortex_sasl_is_enabled to ensure you
 * application requiring SASL support will work.
 * 
 * You can call this function as many times as your application requires. Here
 * is an example:
 * \code
 * if (!vortex_sasl_is_enabled ()) {
 *     g_print ("Unable to initialize SASL support\n");
 *     return -1;
 * }
 * g_print ("SASL support activated!\n");
 * 
 * // do SASL stuff
 * \endcode
 * 
 * 
 * @return TRUE if current Vortex Library is already compiled with
 * SASL profile support. Otherwise FALSE is returned.
 */
gboolean           vortex_sasl_is_enabled                ()
{
#ifndef ENABLE_SASL_SUPPORT
	/* return false if the library wasn't compiled with support
	 * for sasl */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "current Vortex Library doesn't have SASL support");
	return FALSE;
#else
	/* just return TRUE to start working with SASL */
	return TRUE;
#endif	
}

/** 
 * @brief Allows to configure SASL properties used for the SASL authentication process.
 *
 * While using SASL profiles, several properties are required. These
 * properties are the user to authenticate, the password, etc.
 * 
 * This function allows to set properties using values that are
 * required to be unrefered or not. In the case a propertie provided
 * is required to be deallocated you should define the destroy handler
 * for: <b>value_destroy</b>. Once the session is closed, the value
 * destroy handler is run to deallocate provided value.
 * 
 * Properties that are available are: 
 *  - \ref VORTEX_SASL_AUTH_ID
 *  - \ref VORTEX_SASL_AUTHORIZATION_ID
 *  - \ref VORTEX_SASL_PASSWORD
 *  - \ref VORTEX_SASL_REALM
 *  - \ref VORTEX_SASL_ANONYMOUS_TOKEN
 *
 * Here is an example:
 * \code 
 * // set the user (without providing a value_destroy handler because
 * // bob is statically allocated)
 * vortex_sasl_set_propertie (connection,  VORTEX_SASL_AUTH_ID, 
 *                            "bob", NULL);
 *
 * // set the password (providing a value_destroy handler because the
 * // password was dynamically allocated)
 * vortex_sasl_set_propertie (connection, VORTEX_SASL_PASSWORD, 
 *                            password, g_free); 
 * \endcode
 *
 * 
 * @param connection The connection where the properties will be set.
 *
 * @param prop       The properties to set.
 *
 * @param value      The value associated to the properties.
 *
 * @param value_destroy A optional destroy function allowing to pass
 * in properties values that are required to be unrefered when the
 * session is closed.
 */
gboolean           vortex_sasl_set_propertie             (VortexConnection     * connection,
							  VortexSaslProperties   prop,
							  gchar                * value,
							  GDestroyNotify         value_destroy)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "calling to set_properties with a Vortex Library without SASL support");
	return FALSE;
#else
	/* check the connection status before setting the
	 * properties */
	if (connection == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Setting SASL properties on a null reference connection.");
		return FALSE;
	}

	/* check the properties being set */
	if ((prop <= 0) || (prop >= VORTEX_SASL_PROP_NUM)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Setting SASL properties which is not a registered value.");
		return FALSE;
	}
	
	/* check the value itself */
	if (value == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Setting SASL properties which its value is null.");
		return FALSE;
	}

	/* store the property */
	switch (prop) {
	case VORTEX_SASL_AUTH_ID:
		vortex_connection_set_data_full (connection, SASL_AUTHID, 
						 value, NULL, value_destroy);
		break;
	case VORTEX_SASL_AUTHORIZATION_ID:
		vortex_connection_set_data_full (connection, SASL_AUTHZID, 
						 value, NULL, value_destroy);
		break;
	case VORTEX_SASL_PASSWORD:
		vortex_connection_set_data_full (connection, SASL_PASSWORD, 
						 value, NULL, value_destroy);
		break;
	case VORTEX_SASL_REALM:
		vortex_connection_set_data_full (connection, SASL_REALM, 
						 value, NULL, value_destroy);
		break;
	case VORTEX_SASL_ANONYMOUS_TOKEN:
		/* especial case, anonymous token will be also the
		 * authid, that's why we are setting the same
		 * value. This will cause successful anonymous SASL
		 * negotiation to return the anonymous token while
		 * calling to: vortex_sasl_get_propertie (con,
		 * VORTEX_SASL_AUTHID) */
		vortex_connection_set_data_full (connection, SASL_ANONYMOUS_TOKEN,
						 value, NULL, value_destroy);

		vortex_connection_set_data_full (connection, SASL_AUTHID, 
						 value, NULL, NULL);
	case VORTEX_SASL_PROP_NUM:
		/* nothing to do */
		break;
	}

	return TRUE;
#endif	
}

/** 
 * @brief Allows to get current SASL properties from the given connection.
 *
 * Allows to get already set properties using \ref
 * vortex_sasl_set_propertie. See documentation for \ref
 * vortex_sasl_set_propertie to know more about properties available.
 * 
 * @param connection The connection where the propertie will be searched.
 * @param prop The propertie to search
 * 
 * @return The propertie value requested or NULL if it is not defined
 * or Vortex Library does not support SASL.
 */
gchar            * vortex_sasl_get_propertie             (VortexConnection     * connection,
							  VortexSaslProperties   prop)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "calling to get_properties with a Vortex Library without SASL support");
	return NULL;
#else
	/* check the connection status before setting the propertie */
	if (connection == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Getting SASL properties on a null reference connection.");
		return NULL;
	}

	/* check the propertie being set */
	if ((prop <= 0) || (prop >= VORTEX_SASL_PROP_NUM)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Getting SASL propertie which is not a registered value.");
		return FALSE;
	}
	
	/* store the property */
	switch (prop) {
	case VORTEX_SASL_AUTH_ID:
		return vortex_connection_get_data (connection, SASL_AUTHID);
	case VORTEX_SASL_AUTHORIZATION_ID:
		return vortex_connection_get_data (connection, SASL_AUTHZID);
	case VORTEX_SASL_PASSWORD:
		return vortex_connection_get_data (connection, SASL_PASSWORD);
	case VORTEX_SASL_REALM:
		return vortex_connection_get_data (connection, SASL_REALM);
	case VORTEX_SASL_ANONYMOUS_TOKEN:
		return vortex_connection_get_data (connection, SASL_ANONYMOUS_TOKEN);
	case VORTEX_SASL_PROP_NUM:
		/* nothing to do */
		break;
	}
	/* no propertie found */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to find propertie value, maybe buggie implementation");
	return NULL;
#endif
}

/** 
 * @brief Allows to check if the given connection have been
 * successfully authenticated.
 * 
 * Keep in mind this function will return FALSE either because the
 * connection is not successfully authenticated or because current
 * Vortex Library doesn't have SASL support. 
 * 
 * Another issue is that this function will return TRUE for SASL
 * operations requiring any SASL profile that have successfully
 * ended. This includes ANONYMOUS profile. 
 *
 * Obviously, ANONYMOUS profile <b>*DOES NOT*</b> not provide any
 * security measure. Inside environments that allows \ref VORTEX_SASL_ANONYMOUS "ANONYMOUS" and
 * other profiles such as \ref VORTEX_SASL_PLAIN "PLAIN", \ref VORTEX_SASL_CRAM_MD5 "CRAM-MD5" and \ref VORTEX_SASL_DIGEST_MD5 "DIGEST-MD5" could yield to
 * authorize users that only have provided an anonymous token. 
 * 
 * A good practice is not to mix on the same session ANONYMOUS profile
 * and authentication based profiles. Design you code keeping this
 * issue in mind.
 *
 * Once the connection is authenticated you can use the following
 * function to get the method that was used (selected by either the
 * user or the server side): 
 * 
 *  - \ref vortex_sasl_auth_method_used
 *
 * Additionally, you can use the \ref vortex_sasl_get_propertie
 * function to get auth properties used for the authentication. This
 * function and the previous one, \ref vortex_sasl_auth_method_used,
 * must be used only once called to \ref vortex_sasl_is_authenticated
 * and getting an afirmative result. Otherwise, data returned by \ref
 * vortex_sasl_auth_method_used or \ref vortex_sasl_get_propertie must
 * not be trusted.
 * 
 * @param connection The connection to check for authentication status.
 * 
 * @return TRUE if the connection is authenticated, FALSE if not. 
 */
gboolean           vortex_sasl_is_authenticated          (VortexConnection     * connection)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Current Vortex Library doesn't support SASL profiles. Unable to return current connection authentication status.");
	return FALSE;
#else
	if (connection == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "received a null referenced while trying to authenticate the connection");
		return FALSE;
	}

	return (vortex_connection_get_data (connection, SASL_IS_AUTHENTICATED) != NULL);

#endif	
}

/** 
 * @brief Allows to get SASL method used to authenticate the connection. 
 *
 * This function could be used, once used \ref
 * vortex_sasl_is_authenticated, to get the method that was used to
 * authenticate the connection. In environments where several methods
 * are provided, this function is useful to get the security level the
 * user has negotiated.
 * 
 * @param connection The connection that has been authenticated.
 * 
 * @return The SASL method used, represented by an string or NULL if
 * the connection wasn't authenticated. String reference returned must
 * not be deallocated. Values returned are the one represented by \ref
 * VORTEX_SASL_ANONYMOUS, \ref VORTEX_SASL_EXTERNAL, \ref
 * VORTEX_SASL_PLAIN, \ref VORTEX_SASL_CRAM_MD5, \ref
 * VORTEX_SASL_DIGEST_MD5, \ref VORTEX_SASL_GSSAPI.
 */
gchar            * vortex_sasl_auth_method_used          (VortexConnection     * connection)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Current Vortex Library doesn't have SASL profiles support. Unable to return which method was used to authenticate.");
#else
	if (connection == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Connection provided is a null reference. Unable to method used to authenticate.");
		return NULL;
	}
	
	/* return auth method used */
	return (vortex_connection_get_data (connection, SASL_METHOD_USED));
#endif
}

/** 
 * @internal
 * @brief Support user space notification with SASL activation.
 * 
 * @param connection The connection where the SASL activation is being run.
 * @param status     A process status
 * @param message    The message to report. A textual diagnostic.
 * @param user_data  User space data.
 */
void __vortex_sasl_notify (VortexSaslAuthNotify   process_status, 
			   VortexConnection     * connection, 
			   VortexStatus           status,
			   gchar                * message, 
			   gpointer               user_data)
{
	/* drop to the console a log */
	switch (status) {
	case VortexError:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, (message != NULL) ? message : "no message to report");
		/* flag the connection to be *NOT* authenticated
		 * (remove current authid and authzid and all previous
		 * SASL auth info) */
		vortex_connection_set_data (connection, SASL_IS_AUTHENTICATED, NULL);
		vortex_connection_set_data (connection, SASL_AUTHID,   NULL);
		vortex_connection_set_data (connection, SASL_AUTHZID,  NULL);
		vortex_connection_set_data (connection, SASL_PASSWORD, NULL);
		vortex_connection_set_data (connection, SASL_REALM,    NULL);
		break;
	case VortexOk:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, (message != NULL) ? message : "no message to report");
		/* flag the connection to be authenticated */
		vortex_connection_set_data (connection, SASL_IS_AUTHENTICATED, GINT_TO_POINTER (TRUE));
		break;
	}

	/* notify user space that something to be notify have
	 * happen */
	if (process_status != NULL)
		process_status (connection, status, message, user_data);

	/* that's all man */
	return;
}



#ifdef ENABLE_SASL_SUPPORT
/** 
 * @internal
 *
 * @brief Allows to check and initialize only once the SASL library
 * for the client side.
 * 
 * @return TRUE if the library have been initialized or it is already
 * initialized. Otherwise FALSE is returned.
 */
gboolean __vortex_sasl_create_context (VortexConnection * connection) 
{
	Gsasl * ctx = NULL;
	gint    rc;


	/* initialize the SASL client side */
	rc = gsasl_init (&ctx);
	if (rc != GSASL_OK) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "SASL initialization failure (%d): %s\n",
		       rc, gsasl_strerror (rc));
		return FALSE;
	}

	/* save SASL context for this connection */
	vortex_connection_set_data_full (connection, SASL_CONTEXT, ctx, 
					 NULL, SASL_CONTEXT_DESTROY);

	/* save on the SASL ctx the connection */
	gsasl_callback_hook_set (ctx, connection);

	/* just return TRUE  */
	return TRUE;
}

/** 
 * @internal
 * @brief Configure current Vortex SASL properties into the GNU SASL session.
 * 
 * @param connection 
 */
void vortex_sasl_configure_current_properties (VortexConnection * connection)
{
	gchar hostname[512];

	Gsasl_session * session = vortex_connection_get_data (connection, SASL_SESSION);
	
	gsasl_property_set (session, GSASL_AUTHID,   
			    vortex_sasl_get_propertie (connection, VORTEX_SASL_AUTH_ID));

	gsasl_property_set (session, GSASL_PASSWORD, 
			    vortex_sasl_get_propertie (connection, VORTEX_SASL_PASSWORD));

	gsasl_property_set (session, GSASL_AUTHZID, 
			    vortex_sasl_get_propertie (connection, VORTEX_SASL_AUTHORIZATION_ID));

	gsasl_property_set (session, GSASL_REALM, 
			    vortex_sasl_get_propertie (connection, VORTEX_SASL_REALM));

	gsasl_property_set (session, GSASL_ANONYMOUS_TOKEN, 
			    vortex_sasl_get_propertie (connection, VORTEX_SASL_ANONYMOUS_TOKEN));

	gsasl_property_set (session, GSASL_SERVICE, "beep");
	
	/* get current hostname */
	if (gethostname (hostname, 512) == 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "using as hostname for SASL service: %s", hostname);
		gsasl_property_set (session, GSASL_HOSTNAME, hostname);
	}
	return;
}


typedef struct __VortexSaslStartData {
	VortexConnection     * connection;
	gchar                * profile;
	VortexSaslAuthNotify   process_status;
	gpointer               user_data;
	gboolean               threaded;
}VortexSaslStartData ;

/** 
 * @internal
 *
 * @brief Perform the initial SASL step, returning the initial base64
 * blob to be sent to the server side. 
 * 
 * Support function for __vortex_sasl_start_auth
 *
 * @return the function return an already base64 encoded SASL
 * content. The caller must unref the memory returned once not needed.
 */
gchar * __vortex_sasl_initiator_do_initial_step (gchar                * profile,
						 VortexConnection     * connection,
						 VortexSaslAuthNotify   process_status,
						 gpointer               user_data)
{
	/* variable definitions for the gsasl_client_start function */
	Gsasl_session * session;
	gchar         * base64_chunk;
	gint            rc;

	/* drop a log */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "Starting a SASL session for the selected profile: '%s'",  &(profile[26]));
	
	rc = gsasl_client_start (/* get the SASL context for this connection */
		vortex_connection_get_data (connection, SASL_CONTEXT),
		/* a list of profiles selected (actually only one) */
		&(profile[26]), 
		/* the session object */
		&session);
		
	/* check for sasl result */
	if (rc != GSASL_OK) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Unable to initialize the first SASL step, code=%d, message='%s'",
		       rc, gsasl_strerror (rc));
		return NULL;
	}

	/* store the session */
	vortex_connection_set_data (connection, SASL_SESSION, session);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Setting SASL properties to start the handshake");
	
	/* set properties (username, password, authorization id,
	 * service name ..) */
	vortex_sasl_configure_current_properties (connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Creating initial blob..");

	/* build initial message */
	rc = gsasl_step64 (session, "", &base64_chunk);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unblocked from step64");
	if (rc == GSASL_NEEDS_MORE || rc == GSASL_OK) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
		       "SASL initial start have finished, sending start message to remote server: [size %d] '%s', GSASL message: (%d) %s",
		       strlen (base64_chunk), base64_chunk,
		       rc, gsasl_strerror (rc));
		return base64_chunk;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "initial step have failed: (%d): %s",
	       rc, gsasl_strerror (rc));
	return NULL;
}

/**  
 * @internal
 * @brief Support function to check for an error reply inside the content profile reply.
 * 
 * @return FALSE if the frame received implies that the SASL
 * authentication couldn't be performed. Otherwise TRUE is returned.
 */
gboolean __vortex_sasl_is_error_content (VortexFrame * frame,
					 VortexConnection      * connection, 
					 VortexSaslAuthNotify    process_status, 
					 gpointer                user_data)
{
	if (frame == NULL) {
		__vortex_sasl_notify (process_status, connection, VortexError, 
				      "Received a NULL frame while expenting a SASL server reply", user_data);
		return FALSE;
	}
	
	if (!g_strncasecmp (vortex_frame_get_payload (frame), "<error", 6)) {
		__vortex_sasl_notify (process_status, connection, VortexError, 
				      "Received an negative reply from the server side to authenticate the PEER", user_data);
		return FALSE;
	}
	return TRUE;
}

/** 
 * @internal
 *
 * @brief Support function to extract from a frame received the SASL
 * base64 blob.
 * 
 * @return The blob content reply already decoded or NULL if fails.
 */
gchar   * __vortex_sasl_get_base64_blob (gchar                 * frame_content,
					 VortexConnection      * connection, 
					 VortexSaslAuthNotify    process_status, 
					 gpointer                user_data,
					 gchar                ** status)
{
	/* xml related variable declaration */
	axlDtd       * sasl_dtd;
	axlDoc       * doc;
	axlNode      * blob_node;

	/* SASL related variable declaration */
	gchar         * blob;

	/* first check for a empty blob received */
	if (frame_content == NULL) {
		/* return empty blob and set status to be be null */
		(*status) = NULL;
		return g_strdup ("");
	}

	/* get channel management DTD */
	if ((sasl_dtd = vortex_dtds_get_sasl_dtd ()) == NULL) {
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "unable to load dtd file (sasl.dtd), cannot validate incoming message, returning error frame",
				      user_data);
		return NULL;
	}

	/* parser xml document */
	doc = axl_doc_parse (frame_content, strlen (frame_content), NULL);
	if (!doc) {
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "unable to parse SASL reply, we have a buggy remote peer", user_data);
		return NULL;
	}

	/* validate de document */
	if (! axl_dtd_validate (doc, sasl_dtd, NULL)) {
		/* Validation failed */
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "validation failed for SASL reply, we have a buggy remote peer", user_data);
		/* free document loaded */
		axl_doc_free (doc);
		return NULL;		 
	}
	
	/* get a reference to document root (the blob) */
	blob_node  = axl_doc_get_root (doc);
	if (! NODE_CMP_NAME (blob_node, "blob")) {
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "not found error element on err reply, we have a buggy remote peer",
				      user_data);
		/* free the document */
		axl_doc_free (doc);
		return NULL;
	}
 
	/* get error code returned */
	(* status ) = axl_node_get_attribute_value_copy (blob_node, "status");

	/* get the message value returned */
	blob        = axl_node_get_content_copy (blob_node, NULL);

	/* free not needed data */
	axl_doc_free (doc);

	/* return the decoded blob */
	return blob;
}

/* perform the next additional steps needed to authenticate the
 * peer */
gboolean __vortex_sasl_initiator_do_steps (VortexChannel         * channel, 
					   GAsyncQueue           * queue,
					   VortexConnection      * connection, 
					   VortexSaslAuthNotify    process_status, 
					   gpointer                user_data)
{
	/* variable definitions for the sasl_client_step function */
	gint          rc;

	/* variable definitions for the blob decoding */
	VortexFrame * reply    = NULL;
	gchar       * blob     = NULL;
	gchar       * new_blob = NULL;
	gchar       * status   = NULL;

	do {
		/* now process replies received */
		reply = vortex_channel_get_reply (channel, queue);
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Reply received, checking it");
		
		/* for the first reply, that is a piggyback or a reply
		 * following the channel start response we have to
		 * check if an error code was returned. */
		if (!__vortex_sasl_is_error_content (reply, connection, process_status, user_data)) {
			vortex_support_free (2, reply, vortex_frame_free,  queue, g_async_queue_unref);
			return FALSE;
		}

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Getting blob and status reply");

		/* get the base64 blob reply inside the frame but
		 * already decoded */
		blob = __vortex_sasl_get_base64_blob (vortex_frame_get_payload (reply), 
						      connection, process_status, user_data, &status);
		/* free frame reply */
		vortex_frame_free (reply);

		/* check status value */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "current blob value is: '%s' with status='%s'",
		       (blob != NULL) ? blob : "" , (status != NULL) ? status : "");
		if (status != NULL) {
			if (!g_strcasecmp (status, "complete")) {
				/* undef status value and blob */
				vortex_support_free (3, blob, g_free, status, g_free, queue, g_async_queue_unref);
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "authentication Ok due to status code");
				return TRUE;
			}
			/* unref status value */
			vortex_support_free (1, status, g_free);
		}

		/* check returned value */
		if (blob == NULL) {
			vortex_support_free (1, queue, g_async_queue_unref);
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received empty blob and status code doesn't indicate the SASL negotiation have ended.");
			return FALSE;
		}

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "doing a SASL step..");

		/* feed the SASL engine with the data received from
		 * the remote side */
                rc = gsasl_step64 (vortex_connection_get_data (connection, SASL_SESSION),
				   blob, &new_blob);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "before doing sasl step, status is: (%d) %s, with new_blob=%s",
		       rc, gsasl_strerror (rc), (new_blob != NULL) ? new_blob : "");
		
		/* free blob received */
		vortex_support_free (1, blob, g_free);
		
		/* manage continue case, that is, to keep on sending
		 * base64 pieces of code */
		if (rc == GSASL_NEEDS_MORE || ((new_blob != NULL) && (strlen (new_blob) > 0)) ) {
			/* work around to make gsasl to keep on iterating (REMOVE THIS) once
			 * solved */
			rc = GSASL_NEEDS_MORE;

			/* we need to send more data to the server */
			if (!vortex_channel_send_msgv (channel, NULL, "<blob>%s</blob>", new_blob)) {
				vortex_support_free (2, new_blob, g_free, queue, g_async_queue_unref);
				__vortex_sasl_notify (process_status, connection, VortexError,
						      "Unable to negotiate SASL profile selected, an error have happen while sending data",
						      user_data);
				return FALSE;
			}
		}

		/* unref blob generated and nullify loop variables */
		vortex_support_free (1, new_blob, g_free);
		new_blob = NULL;
		blob     = NULL;
		status   = NULL;

	} while (rc == GSASL_NEEDS_MORE);

	/* unref queue used to read frame from SASL channel */
	vortex_support_free (1, queue, g_async_queue_unref);

	if (rc != GSASL_OK) {
		/* unable to negotiate the SASL auth */
		__vortex_sasl_notify (process_status, connection, VortexError, 
				      "SASL negotiation have failed", user_data);
		return FALSE;
	}

	/* seems we couldn't do more to authenticate the peer:
	 *
	 * Peter: Does this mean the SASL fun have ended Jack?
	 *
	 * Jack: That is, my SASLized friend, I'm afraid user space
	 * code is waiting for our reply! */
	return TRUE;
}

/** 
 * @internal
 * 
 * Support function to __vortex_sasl_start_auth, which filters current
 * SASL auth profiles from the profile list that the provided
 * connection has. Because the auth part of the SASL profiles doesn't
 * include a connection reset, as part of the tuning process, this
 * function removes all profiles from the SASL family to avoid
 * confusion about using them again.
 *
 * In fact this is only a stetic question, because proper SASL BEEP
 * peers only could accept one authentication process for each
 * connection. Receiving such notification, and accepting it twice,
 * would be a protocol violation.
 * 
 * @param connection The connection where the SASL profile list will
 * be filtered.
 */
void __vortex_sasl_filter_profiles (VortexConnection * connection)
{
	GList    * list_aux;
	GList    * profiles = vortex_connection_get_remote_profiles (connection);
	gboolean   changed  = FALSE;
	gchar    * data;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "list status before filtering: %d", g_list_length (profiles));

	do {
		list_aux = g_list_first (profiles);
		changed  = FALSE;
		for (; list_aux ; list_aux = g_list_next (list_aux)) {
			if (!g_strncasecmp (VORTEX_SASL_FAMILY, 
					    (gchar *) list_aux->data, 
					    strlen (VORTEX_SASL_FAMILY))) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "found a SASL profile to remove: %s",
				       (gchar *) list_aux->data);

				/* it seems we found a SASL profile to remove */
				changed  = TRUE;
				
				/* get a reference to the data to be removed */
				data     = list_aux->data;

				/* remove the node from the list */
				profiles = g_list_remove (profiles, data);

				/* remove internal content, in this
				 * case, a profile reference */
				axl_free (data);

				/* break the loop */
				break;
				
			} /* end if */

		} /* end for */
	}while (changed && g_list_length (profiles) > 0);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "list status after filtering: %d", g_list_length (profiles));

	/* set new profile list */
	vortex_connection_set_remote_profiles (connection, profiles);

	/* now the profile list is filtered */
	return;
}

void               __vortex_sasl_start_auth              (VortexSaslStartData * data)
{
	/* local variable declarations received from main function */
	VortexConnection     * connection      = data->connection;
	gchar                * profile         = data->profile;
	VortexSaslAuthNotify   process_status  = data->process_status;
	gpointer               user_data       = data->user_data;
	
	/* local variable declarations for this function */
	VortexChannel        * channel;
	GAsyncQueue          * queue;
	gchar                * base64_blob;

	/* free no longer needed node */
	g_free (data);

	/* do initial client step creating the initial
	 * <blob>base64-content-auth</blob> */
	base64_blob = __vortex_sasl_initiator_do_initial_step (profile,
							       connection, 
							       process_status, 
							       user_data);
	/* check returned blob value */
	if (base64_blob == NULL)
		return;
	

	/* create the channel */
	queue   = g_async_queue_new ();
	channel = vortex_channel_new_fullv (/* the connection where the SASL profile will be negotiated */
		                            connection,   
					    /* use the next free channel to be used */
					    0,
					    /* no serverName value to be used */
					    NULL,         
					    /* the profile to negotiate */
					    profile,      
					    /* no encoding done to the
					     * profile part, actually
					     * enclosed as CDATA */
					    EncodingNone, 
					    /* no close channel specification (it doesn't matter
					     * if the channel negotiated is closed once finished
					     * the process. Auth id get from such process will
					     * remain to the connection. */
					    NULL, NULL,
					    /* make all frames received, including those from the
					     * piggyback, to be received through vortex_channel_get_reply */
					    vortex_channel_queue_reply, queue,
					    /* don't use channel creation notify, we want this
					     * function to block us. */
					    NULL, NULL,
					    /* the profile content to send */
					    (strlen (base64_blob) != 0) ? "<blob>%s</blob>" : "",
					    (strlen (base64_blob) != 0) ? base64_blob       : NULL);
	/* unref no longer needed blob */
	g_free (base64_blob);
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "SASL channel creation reply received");
	if (channel == NULL) {
		/* unref queue */
		g_async_queue_unref (queue);
		
		/* notify user space  */
		__vortex_sasl_notify (process_status, connection, VortexError, 
				      "Unable to create the SASL channel",  user_data);
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Seems the SASL channel have being created");

	/* perform the next additional steps needed to authenticate the peer */
	if (!__vortex_sasl_initiator_do_steps (channel, queue,
					       connection, process_status, user_data)) {
		/* because the SASL profile negotiation have failed,
		 * close the channel */
		vortex_channel_close (channel, NULL);
		return;
	}

	/* SASL operation is OK */

	/* because the SASL profiles could not require to perform a
	 * new greetings message, because the session is not secured
	 * but only authenticated, this function must filter all SASL
	 * mechanisma if the connection detects that SASL is already
	 * running */
	__vortex_sasl_filter_profiles (connection);

	/* configure the profile used by the authentication process */
	vortex_connection_set_data (connection, SASL_METHOD_USED, profile);

	/* notify user space that the SASL profile have successfully
	 * completed */
	__vortex_sasl_notify (process_status, connection, VortexOk,
			      "SASL authentication Ok, let's continue!", user_data);
	return;
}

/** 
 * @internal
 * @brief Ensures all properties needed for the SASL profile used are provided.
 *
 * The function not only checks for properties, it also notify a
 * textual error message in case some properties wasn't set properly.
 * 
 * @param conection      The connection where the SASL operations will be performed
 * @param profile        The SASL mech selected to be checked
 * @param process_status User space SASL profile notify
 * @param user_data      User defined data.
 * 
 * @return TRUE if all properties was provided for the selected mech
 * or FALSE if something is wrong.
 */
gboolean __vortex_sasl_auth_data_sanity_check (VortexConnection     * connection,
					       gchar                * profile, 
					       VortexSaslAuthNotify   process_status,
					       gpointer               user_data)
{
	/* SASL ANONYMOUS checkings */
	if (!g_strcasecmp (profile, VORTEX_SASL_ANONYMOUS)) {
		if (!vortex_sasl_get_propertie (connection, VORTEX_SASL_ANONYMOUS_TOKEN)) {
			__vortex_sasl_notify (process_status, connection, VortexError,
					      "ANONYMOUS profile requires the anonymous token to be set in order to start SASL negotiation",
					      user_data);
			return FALSE;
		}
	} /* end anonymous */

	/* SASL PLAIN and SASL CRAM-MD5 checking */
	if (!g_strcasecmp (profile, VORTEX_SASL_PLAIN) ||
	    !g_strcasecmp (profile, VORTEX_SASL_DIGEST_MD5) ||
	    !g_strcasecmp (profile, VORTEX_SASL_CRAM_MD5)) {
		if (!vortex_sasl_get_propertie (connection, VORTEX_SASL_AUTH_ID)) {
			__vortex_sasl_notify (process_status, connection, VortexError,
					      "selected profile requires the auth id to be set in order to start SASL negotiation",
					      user_data);
			return FALSE;
		}

		if (!vortex_sasl_get_propertie (connection, VORTEX_SASL_PASSWORD)) {
			__vortex_sasl_notify (process_status, connection, VortexError,
					      "selected profile requires the password to be set in order to start SASL negotiation",
					      user_data);
			return FALSE;
		}
	} /* end plain */
	
	/* checks for EXTERNAL mech */
	if (!g_strcasecmp (profile, VORTEX_SASL_EXTERNAL)) {
		if (!vortex_sasl_get_propertie (connection, VORTEX_SASL_AUTHORIZATION_ID)) {
			if (!vortex_sasl_get_propertie (connection, VORTEX_SASL_AUTH_ID)) {
				__vortex_sasl_notify (process_status, connection, VortexError,
						      "selected profile requires the authorization id to be set in order to start SASL negotiation",
						      user_data);
				return FALSE;
			}
		}
	} /* end external */
	    
	return TRUE;
}

#endif

/** 
 * @brief Begin SASL authentication process using the selected profile.
 * 
 * Starts the SASL authentication process using the selected profile
 * for the client side. Available values for <b>profile</b> that are
 * supported are:
 * 
 *  - \ref VORTEX_SASL_ANONYMOUS
 *  - \ref VORTEX_SASL_EXTERNAL
 *  - \ref VORTEX_SASL_PLAIN
 *  - \ref VORTEX_SASL_CRAM_MD5
 *  - \ref VORTEX_SASL_DIGEST_MD5
 * 
 * Each profile requires different properties to be set using \ref
 * vortex_sasl_set_propertie. Check \ref
 * vortex_manual_sasl_for_client_side "using SASL at client side" for
 * details.
 *
 * This function performs all SASL operations in a non-blocking way
 * for the caller. This means that the function returns once called
 * and all notification about SASL process status is done through \ref
 * VortexSaslAuthNotify "process_status".
 *
 * If it is needed a blocking version for this function then \ref
 * vortex_sasl_start_auth_sync should be used.
 *
 * See \ref vortex_manual_sasl_for_client_side "Using SASL at client side" 
 * for a detailed explanation about SASL support inside Vortex
 * for the client side.
 * 
 * @param connection     The connection where the SASL authentication will be performed.
 *
 * @param profile        The SASL profile selected.
 *
 * @param process_status Asynchronous notification for SASL process
 * finish status. This handler <b>is not optional</b>. It has to
 * be defined to get report about SASL status. 
 *
 * @param user_data      User space defined data to be passed to <b>process_status</b>
 */
void               vortex_sasl_start_auth                (VortexConnection     * connection,
							  gchar                * profile, 
							  VortexSaslAuthNotify   process_status,
							  gpointer               user_data)
{
#ifndef ENABLE_SASL_SUPPORT
	/* implement non-SASL supported case */
	__vortex_sasl_notify (process_status, connection, VortexError, 
			      "current Vortex Library doesn't support SASL profiles",
			      user_data);

	return;
#else
	/* SASL enabled case */
	VortexSaslStartData * data;

	/* check connection status */
	if (!vortex_connection_is_ok (connection, FALSE)) {
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "received a not valid connection while trying to start SASL negotiation",
				      user_data);
		return;
	}

	/* check profile received */
	if ((profile == NULL) || g_strncasecmp (VORTEX_SASL_FAMILY, profile, 25)) {
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "profile provided for SASL auth is null or is not a SASL one.",
				      user_data);
		return;
	}

	/* if the given connection is already authenticated */
	if (vortex_sasl_is_authenticated (connection)) {
		__vortex_sasl_notify (process_status, connection, VortexError,
				      "provided connection is already authenticated",
				      user_data);
		return;
	}

	/* perform auth incoming data sanity check */
	if (!__vortex_sasl_auth_data_sanity_check (connection, profile, process_status, user_data))
		return;

	/* try to initizalice the library */
	if (!__vortex_sasl_create_context (connection)) {
		/* notify user space that shit happens while init the
		 * SASL library */
		__vortex_sasl_notify (process_status, connection, VortexError, 
				      "Unable to initialize SASL library",  user_data);
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "SASL initialized..");
	data                 = g_new0 (VortexSaslStartData, 1);
	data->connection     = connection;
	data->profile        = profile;
	data->process_status = process_status;
	data->user_data      = user_data;
	data->threaded       = (process_status != NULL);

	/* invoke the process */
	if (data->threaded) {
		/* threaded mode, the caller will be unblocked */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "begin SASL negotiation under threaded mode");
		vortex_thread_pool_new_task ((GThreadFunc) __vortex_sasl_start_auth, data);
		return;
	}
	/* non-threaded mode, the caller will be blocked until SASL process finish */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "begin SASL negotiation non-threaded mode (blocking the caller)");
	__vortex_sasl_start_auth (data);
	return;
#endif	
}

/** 
 * @internal
 * @brief Support function for \ref vortex_start_auth_sync
 */
void __vortex_sasl_start_auth_sync_process  (VortexConnection * connection,
					     VortexStatus       status,
					     gchar            * status_message,
					     gpointer           user_data)
{
	GAsyncQueue * queue = user_data;

	/* push status and status message and unref man! */
	g_async_queue_push  (queue, GINT_TO_POINTER (status));
	g_async_queue_push  (queue, status_message);
	g_async_queue_unref (queue);

	return;
}


/** 
 * @brief Perform SASL negotiation in a synchronous mode (blocking the caller until process finish).
 * 
 * This function does the same like \ref vortex_sasl_start_auth but in
 * a blocking manner. Check documentation this this function at \ref vortex_sasl_start_auth.
 * 
 * Here is an example:
 * \code
 * // some variable declaration for SASL status
 * VortexStatus   status;
 * gchar        * status_message;
 * 
 * // set required properties
 * vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTH_ID,  
 *                            "bob", NULL);
 * vortex_sasl_set_propertie (connection, VORTEX_SASL_PASSWORD, 
 *                            "secret", NULL);
 *                            
 * // begin SASL negotiation
 * vortex_sasl_start_auth_sync (connection, VORTEX_SASL_CRAM_MD5,
 *                              &status, &status_message);
 * // write out error
 * switch (status) {
 * case VortexOk:
 *	g_print ("OK: message reported: %s\n", status_message);
 *	break;
 * case VortexError:
 *	g_print ("FAIL: message reported: %s\n", status_message);
 *	break;
 * }
 * \endcode
 *
 * See \ref vortex_manual_sasl_for_client_side "Using SASL at client side" 
 * for a detailed explanation about SASL support inside Vortex
 * for the client side.
 * 
 * @param connection The connection where the SASL negotiation will take place.
 * @param profile    The SASL profile selected.
 * @param status     A reference to a \ref VortexStatus variable to notify caller SASL finish status.
 * @param status_message A reference to notify the caller a textual diagnostic about SASL finish status.
 */
void               vortex_sasl_start_auth_sync           (VortexConnection     * connection,
							  gchar                * profile,
							  VortexStatus         * status,
							  gchar               ** status_message)

{
	/* timeout object */
	GTimeVal           timeout;
	gpointer           _status;

	/* create an async queue and increase queue reference so the
	 * function could unref it without worry about race conditions
	 * with sync_process function. */
	GAsyncQueue * queue = g_async_queue_new ();
	g_async_queue_ref (queue);

	/* create timeout */
	g_get_current_time (&timeout);
	g_time_val_add (&timeout, vortex_connection_get_timeout ());

	/* start SASL negotiation */
	vortex_sasl_start_auth (connection,  profile, __vortex_sasl_start_auth_sync_process,
				queue);

	/* get status */
	_status = g_async_queue_timed_pop (queue, &timeout);
	if (_status == NULL) {
		/* seems timeout have happen while waiting for SASL to end */
		(* status)         = VortexError;
		(* status_message) = "Timeout have been reached while waiting for SASL to finish";
		return;
	}

	/* set status value */
	if (status != NULL)
		(* status) = GPOINTER_TO_INT (_status);
	
	/* get message */
	if (status_message != NULL)
		(* status_message) = g_async_queue_pop (queue);
	else
		g_async_queue_pop (queue);

	/* unref the queue */
	g_async_queue_unref (queue);

	return;
}


#ifdef ENABLE_SASL_SUPPORT

/** 
 * @internal
 * 
 * Internal functions which allows to check properties to be passed to
 * each mechanism, ensuring all of them receives the minimum set of
 * required values.
 */
gboolean __validation_handler_check_properties (Gsasl_session * sctx, 
						gchar         * sasl_profile)
{
	/* check for a proper SASL profile name defined */
	if (sasl_profile == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "SASL profile not defined, giving up");
		return FALSE;
	}

	/* check for appropiate values for anonymous profile */
	if (!g_strcasecmp (sasl_profile, "ANONYMOUS")) {
		if (gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN) == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "ANONYMOUS method requires the anonymous token to be configured, giving up");
			return FALSE;
		}	

		/* nothing more to check */
		return TRUE;
	}

	/* check for appropiate values for external profile */
	if (!g_strcasecmp (sasl_profile, "EXTERNAL")) {

		/* check for authorization id value */
		if (gsasl_property_get (sctx, GSASL_AUTHZID) == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, 
				    "EXTERNAL method requires the authorization id to be configured, giving up");
			return FALSE;
		}	

		/* nothing more to check */
		return TRUE;
	}

	/* check for auth attribute */
	if (!g_strcasecmp (sasl_profile, "PLAIN") ||
	    !g_strcasecmp (sasl_profile, "CRAM-MD5") ||
	    !g_strcasecmp (sasl_profile, "DIGEST-MD5")) {

		/* check for auth id */
		if (gsasl_property_get (sctx, GSASL_AUTHID) == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, 
				    "%s method requires the authorization id to be configured, giving up", sasl_profile);
			return FALSE;
		}	

		/* nothing more to check */
		return TRUE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "SASL profile=%s not supported, giving up", sasl_profile);
	return FALSE;
}

/** 
 * @internal
 * @brief Validation support.
 * 
 * @param ctx 
 * @param sctx 
 * @param prop 
 * 
 * @return 
 */
int __vortex_sasl_auth_validation (Gsasl * ctx, Gsasl_session * sctx,
				   Gsasl_property prop)
{
	const gchar      * anonymous_token;
	VortexConnection * connection;
	gchar            * password;
	gchar            * profile;
	gpointer	       user_data;

	/* get the connection reference where the SASL request was received */
	connection      = (VortexConnection *) gsasl_callback_hook_get (ctx);

	/* get profile begin validated */
	profile         = vortex_connection_get_data (connection, SASL_PROFILE_BEGIN_NEGOCIATED);

	/* Get user_data */
	user_data = NULL;
	if (!g_strcasecmp (profile, VORTEX_SASL_ANONYMOUS))
		user_data = vortex_connection_get_data (connection, VORTEX_SASL_ANONYMOUS_USER_DATA);
	else if (!g_strcasecmp (profile, VORTEX_SASL_PLAIN))
		user_data = vortex_connection_get_data (connection, VORTEX_SASL_PLAIN_USER_DATA);
	else if (!g_strcasecmp (profile, VORTEX_SASL_CRAM_MD5))
		user_data = vortex_connection_get_data (connection, VORTEX_SASL_CRAM_MD5_USER_DATA);
	else if (!g_strcasecmp (profile, VORTEX_SASL_DIGEST_MD5))
		user_data = vortex_connection_get_data (connection, VORTEX_SASL_DIGEST_MD5_USER_DATA);
	else if (!g_strcasecmp (profile, VORTEX_SASL_EXTERNAL))
		user_data = vortex_connection_get_data (connection, VORTEX_SASL_EXTERNAL_USER_DATA);	

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received notification to validate an incoming SASL request for=%s over connection id=%d",
	       profile, vortex_connection_get_id (connection));

	/* check for properties configured */
	if (! __validation_handler_check_properties (sctx, profile))
		return GSASL_NO_CALLBACK;

	/* according to properties the SASL listener engine should
	 * validate value received.  ANONYMOUS SASL profile validation
	 * support */
	if (prop == GSASL_VALIDATE_ANONYMOUS) {
		/* check if an anonymous handler was defined */
		if (__vortex_sasl_anonymous_auth_handler == NULL && __vortex_sasl_anonymous_auth_handler_full == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "A SASL anonymous validation handler was not set. Unable to authenticate current request. Please check Vortex documentation to set this handler.");
			return GSASL_NO_CALLBACK;
		}
		
		/* get current anonymous token received from remote
		 * side */
		anonymous_token = gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN);

		/* validate anonymous request */
		if ( (
			__vortex_sasl_anonymous_auth_handler &&
			__vortex_sasl_anonymous_auth_handler (connection, 
							  /* get current anonymous token received from remote side */
							  gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN))
			) || (
			__vortex_sasl_anonymous_auth_handler_full &&
			__vortex_sasl_anonymous_auth_handler_full (connection, 
							  /* get current anonymous token received from remote side */
							  gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN),
							  user_data))
			) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "user space have accepted anonymous login");
			/* configure current connections anonymous
			 * token so the server side could get it
			 * later */
			vortex_sasl_set_propertie (connection, VORTEX_SASL_ANONYMOUS_TOKEN, 
						   (gchar *) gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN), NULL);

			/* configure method used to authenticate */
			vortex_connection_set_data (connection, SASL_METHOD_USED, VORTEX_SASL_ANONYMOUS);
			return GSASL_OK;
		}
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "anonymous validation failed: '%s'", 
		       gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN));
		return GSASL_AUTHENTICATION_ERROR;
	} /* end anonymous profile validation */

	/* PLAIN SASL profile validation support */
	if (prop == GSASL_VALIDATE_SIMPLE) {
		/* check if an plain handler was defined  */
		if (__vortex_sasl_plain_auth_handler == NULL && __vortex_sasl_plain_auth_handler_full == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "A SASL plain validation handler was not set. Unable to authenticate current request. Please check Vortex documentation to set this handler.");
			return GSASL_NO_CALLBACK;
		}
		
		/* validate PLAIN profile */
		if ( (
			__vortex_sasl_plain_auth_handler &&
			__vortex_sasl_plain_auth_handler (connection,
						      /* get current auth_id value received from remote side  */
						      gsasl_property_get (sctx, GSASL_AUTHID),
						      /* get current authorization_id value received from remote side */
						      gsasl_property_get (sctx, GSASL_AUTHZID),
						      /* get current password value received from remote side */
						      gsasl_property_get (sctx, GSASL_PASSWORD))
			) || (
			__vortex_sasl_plain_auth_handler_full &&
			__vortex_sasl_plain_auth_handler_full (connection,
						      /* get current auth_id value received from remote side  */
						      gsasl_property_get (sctx, GSASL_AUTHID),
						      /* get current authorization_id value received from remote side */
						      gsasl_property_get (sctx, GSASL_AUTHZID),
						      /* get current password value received from remote side */
						      gsasl_property_get (sctx, GSASL_PASSWORD),
							  user_data))
			) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "user space have accepted plain login");
			/* configure current plain validation data
			 * used: authid, authorization id and
			 * password */
			vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTH_ID, 
						   (gchar *) gsasl_property_get (sctx, GSASL_AUTHID), NULL);

			vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTHORIZATION_ID, 
						   (gchar *) gsasl_property_get (sctx, GSASL_AUTHZID), NULL);

			vortex_sasl_set_propertie (connection, VORTEX_SASL_PASSWORD, 
						   (gchar *) gsasl_property_get (sctx, GSASL_PASSWORD), NULL);

			/* configure method used to authenticate */
			vortex_connection_set_data (connection, SASL_METHOD_USED, VORTEX_SASL_PLAIN);

			return GSASL_OK;
		}
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "plain profile validation failed for: '%s'", 
		       gsasl_property_get (sctx, GSASL_AUTHID));
		return GSASL_AUTHENTICATION_ERROR;						      
	} /* end plain profile validation */

	if (prop == GSASL_PASSWORD) {
		/* check if we are validating cram-md5 profile */
		if (!g_strcasecmp (profile, "CRAM-MD5")) {
			if (__vortex_sasl_cram_md5_auth_handler == NULL && __vortex_sasl_cram_md5_auth_handler_full == NULL) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
				       "A SASL CRAM-MD5 validation handler was not set. Unable to authenticate current request. Please check Vortex documentation to set this handler.");
				return GSASL_NO_CALLBACK;
			}
			
			/* get password  */
			password = NULL;
			if (__vortex_sasl_cram_md5_auth_handler)
				password = __vortex_sasl_cram_md5_auth_handler (connection, 
									gsasl_property_get (sctx, GSASL_AUTHID));
			else if (__vortex_sasl_cram_md5_auth_handler_full)
				password = __vortex_sasl_cram_md5_auth_handler_full (connection, 
									gsasl_property_get (sctx, GSASL_AUTHID),
									user_data);
			
			
			if (password != NULL) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "cram md5 handler defined and password returned from it");
				gsasl_property_set (sctx, GSASL_PASSWORD, password);
				/* Cram md5 accepted, configure current data for the connection. 
				   Here, at this point there is a problem. Because the connection 
				   isn't still validated, because the password returned must be checked
				   with current HASH received, the following lines could be setting a 
				   validation status for a connection that is finally not authenticated 
				   due a HASH mismatch. This is because the user application must use allways 
				   vortex_sasl_is_authenticated before getting any data about the method used 
				   or its values */
				
				/* configure current cram-md5 validation data
				 * used: authid, and password */
				vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTH_ID, 
							   (gchar *) gsasl_property_get (sctx, GSASL_AUTHID), NULL);

				/* the next instruction store the password duplicating it and
				 * providing a destroy function because the SASL engine already
				 * deallocated this value. */
				vortex_sasl_set_propertie (connection, VORTEX_SASL_PASSWORD, 
							   (gchar *) g_strdup (password), g_free);
				
				/* configure method used to authenticate */
				vortex_connection_set_data (connection, SASL_METHOD_USED, VORTEX_SASL_CRAM_MD5);

				return GSASL_OK;
			}
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "cram md5 handler defined but not password returned from it.");
			return GSASL_AUTHENTICATION_ERROR;
		}

		/* check if we are validating digest-md5 profile */
		if (!g_strcasecmp (profile, "DIGEST-MD5")) {
			if (__vortex_sasl_digest_md5_auth_handler == NULL && __vortex_sasl_digest_md5_auth_handler_full == NULL) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
				       "A SASL DIGEST-MD5 validation handler was not set. Unable to authenticate current request. Please check Vortex documentation to set this handler.");
				return GSASL_NO_CALLBACK;
			}
			
			/* get password  */
			password = NULL;
			if (__vortex_sasl_digest_md5_auth_handler)
				password = __vortex_sasl_digest_md5_auth_handler (connection, 
									  gsasl_property_get (sctx, GSASL_AUTHID),
									  gsasl_property_get (sctx, GSASL_AUTHZID),
									  gsasl_property_get (sctx, GSASL_REALM));
			else if (__vortex_sasl_digest_md5_auth_handler_full)
				password = __vortex_sasl_digest_md5_auth_handler_full (connection, 
									  gsasl_property_get (sctx, GSASL_AUTHID),
									  gsasl_property_get (sctx, GSASL_AUTHZID),
									  gsasl_property_get (sctx, GSASL_REALM),
									  user_data);
			
			if (password != NULL) {
				vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "digest md5 handler defined and password returned from it");
				gsasl_property_set (sctx, GSASL_PASSWORD, password);

				/* Please read the coment done to the
				 * SASL CRAM-MD5 method. The same
				 * applies to this method. */

				/* configure current digest-md5 validation data
				 * used: authid, and password */
				vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTH_ID, 
							   (gchar *) gsasl_property_get (sctx, GSASL_AUTHID), NULL);

				vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTHORIZATION_ID, 
							   (gchar *) gsasl_property_get (sctx, GSASL_AUTHZID), NULL);

				vortex_sasl_set_propertie (connection, VORTEX_SASL_REALM, 
							   (gchar *) gsasl_property_get (sctx, GSASL_REALM), NULL);

				/* the next instruction store the password duplicating it and
				 * providing a destroy function because the SASL engine already
				 * deallocated this value. */
				vortex_sasl_set_propertie (connection, VORTEX_SASL_PASSWORD, 
							   (gchar *) g_strdup (password), g_free);
				
				/* configure method used to authenticate */
				vortex_connection_set_data (connection, SASL_METHOD_USED, VORTEX_SASL_DIGEST_MD5);

				return GSASL_OK;
			}
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "digest md5 handler defined but not password returned from it.");
			return GSASL_AUTHENTICATION_ERROR;
		}
	} /* end cram and digest md5 validation */

	/* EXTERNAL SASL profile validation support */
	if (prop == GSASL_VALIDATE_EXTERNAL) {
		
		/* check if an external handler was defined  */
		if (__vortex_sasl_external_auth_handler == NULL && __vortex_sasl_external_auth_handler_full == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
				    "A SASL external validation handler was not set. Unable to authenticate current request. Please check Vortex documentation to set this handler.");
			return GSASL_NO_CALLBACK;
		}

		/* validate EXTERNAL profile */
		if ((__vortex_sasl_external_auth_handler != NULL && 
		     __vortex_sasl_external_auth_handler (connection, gsasl_property_get (sctx, GSASL_AUTHZID))) 
		    || 
		    (__vortex_sasl_external_auth_handler_full != NULL && 
		     __vortex_sasl_external_auth_handler_full (connection, gsasl_property_get (sctx, GSASL_AUTHZID), user_data))) {
			
			/* configure current external validation data
			 * used: authid */
			vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTH_ID, 
						   (gchar *) gsasl_property_get (sctx, GSASL_AUTHZID), NULL);

			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "user space have accepted external login");
			return GSASL_OK;
		}
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "external profile validation failed for: '%s'", gsasl_property_get (sctx, GSASL_AUTHZID));
		
		return GSASL_AUTHENTICATION_ERROR;
	} /* end external profile validation */

	/* nothing about the propertie received */
	return GSASL_NO_CALLBACK;
}


/** 
 * @internal
 * @brief Configure default callback handler for listener instances
 * 
 * @param connection The listener connection
 */
void __vortex_sasl_set_default_listener_handler (VortexConnection * connection)
{
	gsasl_callback_set (vortex_connection_get_data (connection, SASL_CONTEXT),
			    __vortex_sasl_auth_validation);
	return;
}

/** 
 * @internal
 * @brief Build the SASL reply message.
 * 
 * @param status Current status to be set: "complete", "abort", "continue".
 * @param blob_content Current blob content
 * 
 * @return Return a newly allocated SASL blob reply.
 */
gchar * __build_blob_reply (gchar * status, gchar * blob_content)
{
	gchar * result =  g_strdup_printf ("<blob%s%s%s%s%s%s",
					   (status       != NULL) ? " status='"  : "",
					   (status       != NULL) ? status       : "",
					   (status       != NULL) ? "'"          : "",
					   (blob_content != NULL && (strlen (blob_content) > 0)) ? ">"          : " />",
					   (blob_content != NULL && (strlen (blob_content) > 0)) ? blob_content : "",
					   (blob_content != NULL && (strlen (blob_content) > 0)) ? "</blob>"    : "");
	/* deallocate the blob content */
	if (blob_content != NULL)
		g_free (blob_content);

	/* return built message */
	return result;
}

/** 
 * @internal
 * @brief Function which actually perform sasl server iterations.
 * 
 * @param connection The connection where the sasl server iteration will be performed.
 * @param profile_content profile content 
 * 
 * @return 
 */
gboolean __vortex_sasl_server_iterate (VortexConnection * connection, 
				       gchar            * payload,
				       gchar           ** payload_reply)
{
	gchar         * blob          = NULL;
	gchar         * base64_chunk  = NULL;
	gchar         * status        = NULL;
	Gsasl_session * session = vortex_connection_get_data (connection, SASL_SESSION);
	gint            rc;

	/* get received blob */
	blob = __vortex_sasl_get_base64_blob (payload, NULL, NULL, NULL, &status);
	if (blob == NULL) {
		(* payload_reply) = vortex_frame_get_error_message ("500", "General syntax error: received not properly formated SASL message.", NULL);
		return FALSE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Received blob from client: %s", blob);

	/* free status value (enable checking this value): in the case
	 * of "abort" stop the SASL process. */
	if (status != NULL) {
		if (!g_strcasecmp (status, "abort")) {
			/* free current context and session  */
			vortex_connection_set_data (connection, SASL_CONTEXT, NULL);
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "SASL profile negociation cancelled");
			g_free (status);
			return FALSE;
		}

		/* free status value if not processed */
		g_free (status);
	}

	/* now feed the server SASL context with the blob received from the remote peer */
	rc = gsasl_step64 (session, blob, &base64_chunk);

	/* unref blob */
	g_free (blob);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting error code (%d): %s", rc, gsasl_strerror (rc));

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unblocked from step64 with output=%s", 
	       (base64_chunk != NULL) ? base64_chunk : "");

	switch (rc) {
	case GSASL_NEEDS_MORE:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "more data is required");
		(* payload_reply ) = __build_blob_reply (NULL, base64_chunk);
		/* accept the channel to be created in order to
		 * receive next base64 chunks there is no need to
		 * unref base64_chunck, it is already don at
		 * __build_blob_reply */
		return TRUE;
	case GSASL_AUTHENTICATION_ERROR:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "application level have denied SASL auth");
		(* payload_reply ) = vortex_frame_get_error_message ("535", "Application level have denied to accept SASL auth. Authentication failure.", NULL);
		/* deny on going authentication there is no need to
		 * unref base64_chunck, it is already don at
		 * __build_blob_reply */
		return FALSE;
	case GSASL_OK:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "SASL negotiation have finished");
		(* payload_reply ) = __build_blob_reply ("complete", base64_chunk);
		/* accept the channel to be created and successfully
		 * authenticated */
		vortex_connection_set_data (connection, SASL_IS_AUTHENTICATED, GINT_TO_POINTER (TRUE));
		/* there is no need to unref base64_chunck, it is
		 * already don at __build_blob_reply */
		return TRUE;
	}

	/* in other case, false is returned */
	(* payload_reply) = vortex_frame_get_error_message ("454", "Temporary authentication failure.", NULL);
	return FALSE;
}

/** 
 * @internal 
 * @brief Support for extended channel start received.
 * 
 * @param profile 
 * @param channel_num 
 * @param connection 
 * @param serverName 
 * @param profile_content 
 * @param profile_content_reply 
 * @param encoding 
 * @param user_data 
 * 
 * @return 
 */
gboolean __vortex_sasl_accept_negociation_start (gchar             * profile,
						 gint                channel_num,
						 VortexConnection  * connection,
						 gchar             * serverName,
						 gchar             * profile_content,
						 gchar            ** profile_content_reply,
						 VortexEncoding      encoding,
						 gpointer            user_data)
{

	Gsasl_session * session;
	gint            rc;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Received channel start for mech: %s, profile_content=%s", 
	       profile, profile_content);

	/* create the server context */
	if (!__vortex_sasl_create_context (connection))
		return FALSE;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "setting default auth validation handler for the listener");
	
	/* associate user_data with connection */
	if (!g_strcasecmp (profile, VORTEX_SASL_ANONYMOUS))
    	vortex_connection_set_data (connection, VORTEX_SASL_ANONYMOUS_USER_DATA, user_data);
	else if (!g_strcasecmp (profile, VORTEX_SASL_PLAIN))
    	vortex_connection_set_data (connection, VORTEX_SASL_PLAIN_USER_DATA, user_data);
	else if (!g_strcasecmp (profile, VORTEX_SASL_CRAM_MD5))
    	vortex_connection_set_data (connection, VORTEX_SASL_CRAM_MD5_USER_DATA, user_data);
	else if (!g_strcasecmp (profile, VORTEX_SASL_DIGEST_MD5))
    	vortex_connection_set_data (connection, VORTEX_SASL_DIGEST_MD5_USER_DATA, user_data);
	else if (!g_strcasecmp (profile, VORTEX_SASL_EXTERNAL))
    	vortex_connection_set_data (connection, VORTEX_SASL_EXTERNAL_USER_DATA, user_data);
	else
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING,
			"Unsupported SASL profile used (%s), cannot associate user_data pointer", profile);

	/* configure default server handler for the context created */
	__vortex_sasl_set_default_listener_handler (connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "begin initial server step");

	rc = gsasl_server_start (/* get the SASL context for this connection */
		vortex_connection_get_data (connection, SASL_CONTEXT),
		/* a list of profiles selected (actually only one) */
		&(profile[26]), 
		/* the session object */
		&session);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting error code (%d): %s", rc, gsasl_strerror (rc));

	/* check for sasl result */
	if (rc != GSASL_OK) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Unable to initialize the first SASL step, code=%d, message='%s'",
		       rc, gsasl_strerror (rc));
		return FALSE;
	}

	/* store the session */
	vortex_connection_set_data (connection, SASL_SESSION, session);

	/* store SASL profile being negotiated */
	vortex_connection_set_data_full (connection, SASL_PROFILE_BEGIN_NEGOCIATED, g_strdup (&(profile[26])), NULL, g_free);

	if (!__vortex_sasl_server_iterate (connection, profile_content, profile_content_reply))
		return FALSE;

	/* accept the channel to be created */
	return TRUE;
}

void __vortex_sasl_accept_negociation_frame_receive (VortexChannel    * channel,
						     VortexConnection * connection,
						     VortexFrame      * frame,
						     gpointer user_data)
{
	gchar * payload_reply = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Received SASL frame\n");
	/* perform a SASL iteration */
	__vortex_sasl_server_iterate (connection, vortex_frame_get_payload (frame), &payload_reply);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "SASL reply: %s", payload_reply);

	/* reply SASL iteration result */
	if (! vortex_channel_send_rpy (channel, payload_reply, strlen (payload_reply),
				       vortex_frame_get_msgno (frame))) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to send SASL iteration reply");
	}

	/* free payload reply */
	g_free (payload_reply);
	return;
}

#endif

/** 
 * @brief Allows to set current auth validation handler for the the SASL EXTERNAL profile.
 *
 * @param auth_handler Auth handler to be executed when a query for EXTERNAL authentication is received.
 */
void           vortex_sasl_set_external_validation  (VortexSaslAuthExternal auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting external validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_external_auth_handler = auth_handler;
#endif
}

/** 
 * @brief Allows to set current auth validation handler for the the SASL EXTERNAL profile.
 *
 * @param auth_handler Auth handler to be executed when a query for EXTERNAL authentication is received.
 */
void           vortex_sasl_set_external_validation_full  (VortexSaslAuthExternalFull auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting external validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_external_auth_handler_full = auth_handler;
#endif
}

/** 
 * @brief Allows to set current auth validation handler for the the
 * SASL ANONYMOUS profile.
 *
 * @param auth_handler Auth handler to be executed when a query for
 * ANONYMOUS authentication is received.
 */
void           vortex_sasl_set_anonymous_validation  (VortexSaslAuthAnonymous auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting anonymous validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_anonymous_auth_handler = auth_handler;
#endif
}

/** 
 * @brief Allows to set current auth validation handler for the the
 * SASL ANONYMOUS profile.
 *
 * @param auth_handler Auth handler to be executed when a query for
 * ANONYMOUS authentication is received.
 */
void           vortex_sasl_set_anonymous_validation_full  (VortexSaslAuthAnonymousFull auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting anonymous validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_anonymous_auth_handler_full = auth_handler;
#endif
}

/** 
 * @brief Allows to set the validation handler to be used while
 * authenticating PLAIN SASL profile.
 * 
 * @param auth_handler The handler to be executed.
 */
void               vortex_sasl_set_plain_validation      (VortexSaslAuthPlain     auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting plain validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_plain_auth_handler = auth_handler;
#endif
}

/** 
 * @brief Allows to set the validation handler to be used while
 * authenticating PLAIN SASL profile.
 * 
 * @param auth_handler The handler to be executed.
 */
void               vortex_sasl_set_plain_validation_full      (VortexSaslAuthPlainFull     auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting plain validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_plain_auth_handler_full = auth_handler;
#endif
}

/** 
 * @brief Allows to set the validation handler to be used while authenticating CRAM-MD5 SASL profile.
 * 
 * @param auth_handler The handler to be executed.
 */
void               vortex_sasl_set_cram_md5_validation      (VortexSaslAuthCramMd5     auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting CRAM-MD5 validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_cram_md5_auth_handler = auth_handler;
#endif
}

/** 
 * @brief Allows to set the validation handler to be used while authenticating CRAM-MD5 SASL profile.
 * 
 * @param auth_handler The handler to be executed.
 */
void               vortex_sasl_set_cram_md5_validation_full      (VortexSaslAuthCramMd5Full     auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting CRAM-MD5 validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_cram_md5_auth_handler_full = auth_handler;
#endif
}

/** 
 * @brief Allows to set current auth validation handler for the the SASL DIGEST-MD5 profile.
 *
 * @param auth_handler Auth handler to be executed when a query for DIGEST-MD5 authentication is received.
 */
void           vortex_sasl_set_digest_md5_validation  (VortexSaslAuthDigestMd5 auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting DIGEST-MD5 validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_digest_md5_auth_handler = auth_handler;
#endif
}

/** 
 * @brief Allows to set current auth validation handler for the the SASL DIGEST-MD5 profile.
 *
 * @param auth_handler Auth handler to be executed when a query for DIGEST-MD5 authentication is received.
 */
void           vortex_sasl_set_digest_md5_validation_full  (VortexSaslAuthDigestMd5Full auth_handler)
{
#ifndef ENABLE_SASL_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Setting DIGEST-MD5 validation handler on a Vortex Library instance without SASL support");
	return;
#else
	__vortex_sasl_digest_md5_auth_handler_full = auth_handler;
#endif
}

/** 
 * @brief Allows to configure current Vortex Library process to accept
 * incoming SASL negotiations.
 *
 * This function allows to activate the SASL profile support selected
 * and to provide a handler that will be executed once received an
 * authentication request. There are two functions to activate SASL
 * profile support: this one and the previous API: \ref
 * vortex_sasl_accept_negociation.
 *
 * Both works the same way but this one is able to pass the user data
 * configured to the SASL auth handler once it is executed. Allowed
 * value for <b>mech</b> are:
 *
 *  - \ref VORTEX_SASL_ANONYMOUS
 *  - \ref VORTEX_SASL_EXTERNAL
 *  - \ref VORTEX_SASL_PLAIN
 *  - \ref VORTEX_SASL_CRAM_MD5
 *  - \ref VORTEX_SASL_DIGEST_MD5
 *
 * See \ref vortex_manual_sasl_for_server_side "Using SASL at server side" 
 * for a detailed explanation about SASL support inside Vortex
 * for the server side.
 * 
 * @param mech The SASL mech to be accepted from remote BEEP peers.
 * @param user_data User-defined pointer to be passed to SASL callbacks.
 * 
 * @return TRUE the mechanism was enabled to be accepted.
 */
gboolean           vortex_sasl_accept_negociation_full        (gchar * mech, gpointer user_data)
{
#ifndef ENABLE_SASL_SUPPORT
	/* current Vortex Library doesn't support SASL */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "current Vortex Library doesn't have SASL support");
	return FALSE;
#else
	/* check the mech definition received */
	if (g_strncasecmp (mech, "http://iana.org/beep/SASL/", 26)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "you didn't provide a valid mech string specification.");
		return FALSE;
	}

	/* register the profile received  */
	vortex_profiles_register (mech, 
				  /* do not set the normal channel
				   * start handler because it is need
				   * support to process profile content
				   * received. This is done by the
				   * extended channel start
				   * handler. This is actually set on
				   * the next sentence. */
				  NULL, NULL,
				  /* do not set the close handler,
				   * which implies to accept the
				   * channel to be closed whenever it
				   * is requested. However, negotiated
				   * authid will still remain for the
				   * context of the connection. */
				  NULL, NULL,
				  /* set frame received handlers to be
				   * able to handle and reply SASL
				   * frames received */
				  __vortex_sasl_accept_negociation_frame_receive, NULL);
	
	/* register an extended channel start handler for the mech registered */
	vortex_profiles_register_extended_start	 (mech, 
						  __vortex_sasl_accept_negociation_start, 
						  user_data);

	return TRUE;
#endif	
}



/** 
 * @brief Allows to configure current Vortex Library process to accept
 * incoming SASL negotiations.
 *
 * See also \ref vortex_sasl_accept_negociation_full. Allowed value
 * for <b>mech</b> are:
 *
 *  - \ref VORTEX_SASL_ANONYMOUS
 *  - \ref VORTEX_SASL_EXTERNAL
 *  - \ref VORTEX_SASL_PLAIN
 *  - \ref VORTEX_SASL_CRAM_MD5
 *  - \ref VORTEX_SASL_DIGEST_MD5
 *
 * See \ref vortex_manual_sasl_for_server_side "Using SASL at server side" 
 * for a detailed explanation about SASL support inside Vortex
 * for the server side.
 * 
 * @param mech The SASL mech to be accepted from remote BEEP peers.
 * 
 * @return TRUE the mechanism was enabled to be accepted.
 */
gboolean           vortex_sasl_accept_negociation        (gchar * mech)
{
    return vortex_sasl_accept_negociation_full (mech, NULL);
}

/* @} */
