/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>

GList         * __vortex_support_search_path       = NULL;
GStaticMutex    __vortex_support_search_path_mutex = G_STATIC_MUTEX_INIT;

#define LOG_DOMAIN "vortex-support"

/**
 * \defgroup vortex_support Vortex Support: Support function used across the library
 */

/**
 * \addtogroup vortex_support
 * @{
 */

typedef void (*DeallocFunction) (gpointer data);

/** 
 * @brief Allows to perform several memory dispose operations at the
 * same time. 
 * 
 * This function is mainly by Vortex Library but it could be useful
 * for Vortex Library API consumers. Here is an example:
 * \code
 * // perform a massive dealloc operation
 * vortex_support_free (3,  // three memory disposal operations
 *                      // the first one
 *                      profile, g_free, 
 *                      // the second one
 *                      channel, vortex_channel_free
 *                      // the last one.
 *                      serverName, g_free);
 * \endcode
 * 
 * As you may observed, you have to provide the reference to be
 * deallocated an a destroy function which accept the reference to be
 * unreferenced. 
 * 
 * Any mistake produce calling to this function without providing the
 * right parameters will cause unexpected behaviors and lot of fun
 * debugging memory misfunctions.
 * 
 * @param params The number of disposal operations to be performed.
 */
void   vortex_support_free (gint params, ...)
{
	DeallocFunction        dealloc;
	gpointer               data;
	int                    iterator;
	va_list                args;

	/* open the variable argument stack */
	va_start (args, params);

	/* init and loop over the iterator */
	iterator = 0;
	while (iterator < params) {

		/* get a reference to the deallocator and its resource
		   to deallocate */
		data    = va_arg (args, gpointer);
		dealloc = va_arg (args, DeallocFunction);

		/* dealloc */
		if ((dealloc != NULL) && (data != NULL))
			dealloc (data);
		
		/* update the iterator status */
		iterator++;
	}

	/* end variable argument invocation */
	va_end (args);
	return;
}

/** 
 * @brief Allows to add new search path to be used by \ref vortex_support_find_data_file.
 *
 * This function allows to configure how Vortex Library lookups
 * files. This is important to locate dtd file definitions, SSL
 * certificates and so on. 
 *
 * You can also use this API to install and configure application
 * level files that will be used by your profile implementation.
 * 
 * Once a path is added, \ref vortex_support_find_data_file function
 * will lookup on the provided paths.
 *
 * \code 
 * // add all search paths (maybe from an user interface)
 * vortex_support_add_search_path ("/my/path/to/local/data");
 * vortex_support_add_search_path ("/my/alternative/path");
 *
 * ...
 *
 * // write you file location code in an abstract manner so it could be
 * // used on every platform easily.
 * gchar * my_data_def = vortex_support_find_data_file ("my_data.def");
 *
 * // NOTE that the file to lookup doesn't have any references to
 * // local paths
 * \endcode
 *
 * @param path A new path to be added. The function will perform a copy for the given path
 */
void     vortex_support_add_search_path (gchar * path)
{
	g_return_if_fail (path);
	
	g_static_mutex_lock (&__vortex_support_search_path_mutex);
	__vortex_support_search_path = g_list_append (__vortex_support_search_path,  g_strdup (path));
	g_static_mutex_unlock (&__vortex_support_search_path_mutex);
	return;
}

/** 
 * @brief Adds a new path to be used while looking for files without
 * making a local copy.
 *
 * This function works like \ref vortex_support_add_search_path
 * without making a local copy. This function is useful when it is
 * needed to add an already allocated lookup path.
 *
 * If a static path is needed to be added \ref vortex_support_add_search_path should be used.
 *
 * @param path A new path to be added. Provided path reference mustn't be deallocated.
 */
void     vortex_support_add_search_path_ref (gchar * path)
{
	g_return_if_fail (path);
	
	g_static_mutex_lock (&__vortex_support_search_path_mutex);
	__vortex_support_search_path = g_list_append (__vortex_support_search_path, path);
	g_static_mutex_unlock (&__vortex_support_search_path_mutex);

	return;
}

/** 
 * @brief Allows to lookup for a file into the known vortex data file
 * locations.
 * 
 * While locating data files inside Vortex Library context a set of
 * function are used to allows Vortex Library API consumers to produce
 * easily application level code that is platform/directory-structure
 * independent as possible. 
 * 
 * Because profile definitions needs DTD files, certificates, xml
 * documents, etc, a common problem to face is how to locate this
 * files not only into the development environment but also into the
 * production one.
 * 
 * Vortex Library allows to configure known file locations using \ref
 * vortex_support_add_search_path and \ref vortex_support_add_search_path_ref. 
 *
 * Once configured search paths, application level calls to this
 * function to locate files using basenames such as
 * <b>myCertificate.cert</b>, <b>channel.dtd</b>, etc. This avoid full
 * paths file names that are a problem while moving around the source
 * code not only across different platforms but also on the same.
 *
 * See \ref vortex_support_add_search_path for more information about
 * known data files location.
 *
 * @param name the base file name to lookup.
 * 
 * @return NULL or the file path. Value returned should be unrefered using g_free when no longer needed.
 */
gchar  * vortex_support_find_data_file (gchar * name)
{
	GList * list_aux;
	gchar * file_name;

	g_static_mutex_lock (&__vortex_support_search_path_mutex);	

	/* foreach path installed */
	list_aux = g_list_first (__vortex_support_search_path);
	for (; list_aux ; list_aux = g_list_next (list_aux)) {

		/* build a file path, trying to build ./ or .\ file
		 * path if path "." is used */
		if (g_strcasecmp (list_aux->data, "."))
			file_name = g_build_filename (list_aux->data, name, NULL );
		else
			file_name = g_strdup (name);

		/* check the file to be found */
		if (g_file_test (file_name, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "file found at: %s", file_name);
			g_static_mutex_unlock (&__vortex_support_search_path_mutex);	
			return file_name;
		} /* end if */

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to find file %s on %s",  name, file_name);
		g_free (file_name);

	} /* end for */

	g_static_mutex_unlock (&__vortex_support_search_path_mutex);
	return NULL;
}

/** 
 * @internal
 * @brief Clears internal vortex support module, unrefering all memory allocated.
 */
void     vortex_support_cleanup               ()
{
	GList * list_aux;

	/* check for search path content */
	if (__vortex_support_search_path != NULL) {

		/* free already allocated content */
		list_aux = g_list_first (__vortex_support_search_path);
		for (; list_aux; list_aux = g_list_next (list_aux)) {
			g_free (list_aux->data);
		}
		g_list_free (__vortex_support_search_path);
		__vortex_support_search_path = NULL;
	} 
	
	
	return;
}

/* @} */
