/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>

gboolean __vortex_thread_pool_exclusive = TRUE;

#define LOG_DOMAIN "vortex-thread-pool"

GThreadPool * __vortex_thread_pool;

/* vortex thread pool struct used by vortex library to notify to tasks
 * to be performed to vortex thread pool */
typedef struct _VortexThreadPoolTask {
	GThreadFunc   func;
	gpointer      data;
}VortexThreadPoolTask;

/**
 * @internal
 * 
 * This helper function dispatch the work to the right handler
 **/
void __vortex_thread_pool_dispatcher (gpointer data,
				      gpointer user_data)
{
	VortexThreadPoolTask * task        = data;

	/* at this point we already are executing inside a thread */
	task->func (task->data);
	g_free (task);
		
	/* That's all! */
	return;
}

/**
 * \defgroup vortex_thread_pool Vortex Thread Pool: Pool of threads which runns user defined async notifications.
 */

/**
 * \addtogroup vortex_thread_pool
 * @{
 */

/**
 * @brief Init the Vortex Thread Pool subsystem.
 * 
 * Initializes the vortex thread pool. This pool is mainly used to
 * invoke frame receive handler at any level. Because we have always
 * running threads to send messages (the vortex writer and sequencer)
 * this pool is not needed. 
 *
 * Among the vortex features are listed to create connection, channels
 * and also close channel in a asynchronous way. This means the those
 * process are run inside a separate thread. All those process are
 * also run inside the thread pool.
 *
 * @param max_threads how many threads to start.
 *
 **/
void vortex_thread_pool_init     (gint max_threads)
{
	GError * error = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "creating thread pool threads=%d", max_threads);

	__vortex_thread_pool =   g_thread_pool_new  (__vortex_thread_pool_dispatcher,
						     NULL,
						     max_threads,
						     __vortex_thread_pool_exclusive,
						     &error);
	if (error != NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to create thread pool, error was: %s",
		       error->message);
		exit (-1);
	}
	return;
}

/** 
 * @internal
 *
 * Stops the vortex thread pool, that was initialized by
 * vortex_thread_pool_init.
 * 
 */
void vortex_thread_pool_exit () 
{
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "stopping thread pool..");

	g_thread_pool_free (__vortex_thread_pool, 
			    /* immediatly, but letting current tasks
			     * to finish. */
			    TRUE, 
			    /* wait until all threads are ready */
			    TRUE); 

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "thread pool is stopped..");
	return;
}

/**
 * @brief Queue a new task inside the VortexThreadPool.
 *
 * 
 * Queue a new task to be performed. This function is used by vortex
 * for internal purpose so it should not be useful for vortex library
 * consumers.
 *
 * @param func the function to execute.
 * @param data the data to be passed in to the function.
 *
 * 
 **/
void vortex_thread_pool_new_task (GThreadFunc func, gpointer data)
{
	GError * error = NULL;
	VortexThreadPoolTask * task;


	g_return_if_fail (func);
	g_return_if_fail (data);
	g_return_if_fail (__vortex_thread_pool);

	/* create the task data */
	task       = g_new0 (VortexThreadPoolTask, 1);
	task->func = func;
	task->data = data;

	/* queue the task for the next available thread */
	g_thread_pool_push (__vortex_thread_pool, 
			    task,
			    &error);

	/* check error reported */
	if (error != NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to create new task, error was: %s", 
		       error->message);
	}
	return;
}


/**
 * @brief Returns the running threads the given pool have.
 * 
 * Returns the actual running threads the vortex thread pool have.
 * 
 * @return the thread number or -1 if fails
 **/
gint vortex_thread_pool_get_running_threads ()
{

	g_return_val_if_fail (__vortex_thread_pool, -1);

	return g_thread_pool_get_num_threads (__vortex_thread_pool);
}


/** 
 * @brief Allows to configure the number of threads inside the Vortex
 * Thread Pool module.
 *
 * This function modifies the environment variable
 * <b>"VORTEX_THREADS"</b>, setting the value provided. Later, the
 * value will be used by the \ref vortex_init function to initialize
 * the Vortex Library.
 *
 * This function must be called before \ref vortex_init to take
 * effect. See also:
 *
 *  - \ref vortex_thread_pool_get_num
 *  - \ref vortex_thread_pool_get_running_threads.
 * 
 * @param number The number of working threads that the Vortex Thread
 * Pool will try to start.
 */
void vortex_thread_pool_set_num             (gint number)
{
	gchar * _number;
	g_return_if_fail (number > 0);
	
	/* translate the number into an string representation */
	_number = g_strdup_printf ("%d", number);
	
	/* set the value */
	g_setenv ("VORTEX_THREADS", _number, TRUE);

	/* relese the string */
	g_free (_number);
	
	return;
}

/**
 * @brief Returns how many threads have the given VortexThreadPool
 * 
 * This function is for internal vortex library purpose. This function
 * returns how many thread must be created inside the vortex thread
 * pool which actually is the one which dispatch all frame received
 * callback and other async notifications. 
 *
 * This function pay attention to the environment var <b>"VORTEX_THREADS"</b>
 * so user can control how many thread are created. In case this var
 * is not defined the function will return 5. In case the var is
 * defined but using a wrong value, that is, an non positive integer,
 * the function will abort the vortex execution and consequently the
 * application over it.
 * 
 * @return the number of threads to be created.
 **/
gint vortex_thread_pool_get_num             ()
{
	gint value;

	if (!g_getenv ("VORTEX_THREADS"))
		return 5;

	value = atoi (g_getenv ("VORTEX_THREADS"));
	if (value <= 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "VORTEX_THREAD environment var have a wrong value='%s', exiting", 
		       g_getenv ("VORTEX_THREADS"));
		exit (-1);
	}
	
	return value;
}

/** 
 * @brief Allows to configure current configuration for threads inside
 * the thread pool created.
 *
 * By default, once the thread pool is started, all threads available
 * inside are not usable from outside the pool. However, it could be
 * required to allow the thread pool to share (and use) threads
 * available not only from its thread pool but also from other thread
 * pools created outside the Vortex Library context.
 *
 * This function must be called before \ref vortex_init to take
 * effect.
 * 
 * @param value The new behaviour to configure. By default, internal
 * value is already configured to TRUE. Set FALSE to make the thread
 * pool to behave in a non-exclusive form.
 */
void vortex_thread_pool_set_exclusive_pool  (gboolean value)
{
	/* set the new value */
	__vortex_thread_pool_exclusive = value;

	return;
}
       

/* @} */
