/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

/**
 * \defgroup vortex_tls Vortex TLS: TLS profile support and related functions
 */

/**
 * \addtogroup vortex_tls
 * @{
 */

#define LOG_DOMAIN "vortex-tls"

#include <vortex.h>


/** 
 * @internal
 * @brief Internal default handlers used to define the TLS profile support.
 */
VortexTlsAcceptQuery            vortex_tls_accept_handler      = NULL;
VortexTlsCertificateFileLocator vortex_tls_certificate_handler = NULL;
VortexTlsPrivateKeyFileLocator  vortex_tls_private_key_handler = NULL;

/** 
 * @internal
 *
 * @brief Include headers and default global variables declarations if
 * the library build have built-in support.
 * 
 * What we seek is to support applications that are linked to Vortex
 * Library but in a manner that they do not get undefined function
 * references while using a Vortex Library without TLS support.
 *
 * All function defined at this module have especial handling code
 * done to not perform any operation if the library wasn't built with
 * TLS support.
 */
#ifdef ENABLE_TLS_SUPPORT
#include <openssl/ssl.h>
GStaticMutex init_mutex   = G_STATIC_MUTEX_INIT;
gboolean     already_init = FALSE;
#endif


/** 
 * @brief Initialize and checks if the Vortex Library being used have
 * support for TLS profile.
 * 
 * All client applications using TLS profile must call to this
 * function in order to ensure TLS profile engine has been successfully
 * enabled and to check if the Vortex Library being used have support
 * for TLS profile.
 *
 * @return TRUE if the TLS profile was activated and, implicitly, the
 * Vortex Library being used have TLS support enabled. Otherwise FALSE
 * is returned.
 */
gboolean vortex_tls_is_enabled ()
{

#ifndef ENABLE_TLS_SUPPORT
	/* return false if the library wasn't compiled with support
	 * for tls. */
	return FALSE;
#else
	/* check if the SSL library wasn't initialized */
	g_static_mutex_lock (&init_mutex);
	if (already_init == TRUE) {
		g_static_mutex_unlock (&init_mutex);
		return TRUE;
	}
	already_init = TRUE;
	g_static_mutex_unlock (&init_mutex);


	/* init ssl ciphers and engines */
	SSL_library_init ();

	return TRUE;
#endif	
}

#ifdef ENABLE_TLS_SUPPORT

/** 
 * @internal
 *
 * @brief Default handlers used to actually read from underlying
 * transport while a connection is working under TLS.
 * 
 * @param connection The connection where the read operation will be performed.
 * @param buffer     The buffer where the data read will be returned
 * @param buffer_len Buffer size
 * 
 * @return How many bytes was read.
 */
gint vortex_tls_ssl_read (VortexConnection * connection, gchar * buffer, gint buffer_len)
{
	SSL  * ssl;

	/* get ssl object */
	ssl = vortex_connection_get_data (connection, "ssl-data:ssl");
	if (ssl == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to find ssl object to read data");
		return 0;
	}

	return SSL_read (ssl, buffer, buffer_len);
}

/** 
 * @internal
 *
 * @brief Default handlers used to actually write to the underlying
 * transport while a connection is working under TLS.
 * 
 * @param connection The connection where the write operation will be performed.
 * @param buffer     The buffer containing data to be sent.
 * @param buffer_len The buffer size.
 * 
 * @return How many bytes was written.
 */
gint vortex_tls_ssl_write (VortexConnection * connection, gchar * buffer, gint buffer_len)
{
	SSL * ssl = vortex_connection_get_data (connection, "ssl-data:ssl");
	if (ssl == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to find ssl object to read data");
		return 0;
	}
	return SSL_write (ssl, buffer, buffer_len);
}

/** 
 * @brief Common function which sets needed data for the TLS transport
 * and default callbacks for read and write data.
 * 
 * @param connection The connection to configure
 * @param ssl The ssl object.
 * @param ctx The ssl context
 */
void vortex_tls_set_common_data (VortexConnection * connection, 
				 SSL* ssl, SSL_CTX * ctx)
{

	/* set some necessary ssl data (setting destroy functions for
	 * both) */
	vortex_connection_set_data_full (connection, "ssl-data:ssl", ssl, 
					 NULL, (GDestroyNotify) SSL_free);
	vortex_connection_set_data_full (connection, "ssl-data:ctx", ctx, 
					 NULL, (GDestroyNotify) SSL_CTX_free);

	/* set new handlers for read and write */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "change default handlers to be used for send/recv");
	vortex_connection_set_receive_handler (connection, vortex_tls_ssl_read);
	vortex_connection_set_send_handler    (connection, vortex_tls_ssl_write);

	return;
}

/** 
 * @internal
 * @brief Support data structure for \ref vortex_tls_start_negociation function.
 */
typedef struct __VortexTlsBeginData {
	VortexConnection    * connection;
	gchar               * serverName;
	VortexTlsActivation   process_status;
	gpointer              user_data;
} VortexTlsBeginData;


void __vortex_tls_start_negociation_close_and_notify (VortexTlsActivation   process_status,
						      VortexConnection    * connection,
						      VortexChannel       * channel,
						      gchar               * message,
						      gpointer              user_data)
{
	/* close the channel */
	if (channel != NULL)
		vortex_channel_close (channel, NULL);
	
	/* notify the error */
	if (process_status != NULL) {
		process_status (connection, VortexError, message, user_data);
	}

	/* log the error */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, message, NULL);
	return;
}


/** 
 * @internal
 * @brief Invoke the specific TLS code to perform the handshake.
 * 
 * @param connection The connection where the TLS handshake will be performed.
 * 
 * @return TRUE if the handshake was successfully performed. Otherwise
 * FALSE is returned.
 */
gboolean vortex_tls_invoke_tls_activation (VortexConnection * connection)
{
	SSL_CTX     * ctx;
	SSL_METHOD  * meth;
	SSL         * ssl;
	X509        * server_cert;

	/* create and register the TLS method */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "initializing ciphers method");
	meth = TLSv1_client_method ();

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "initializing TLS context");
	ctx  = SSL_CTX_new (meth); 
	if (ctx == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "error while creating TLS context");
		return FALSE;
	}

	/* create the tls transport */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "initializing TLS transport");
	ssl = SSL_new (ctx);       
	if (ssl == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "error while creating TLS transport");
		return FALSE;
	}

	/* set the file descriptor */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "setting file descriptor");
	SSL_set_fd (ssl, vortex_connection_get_socket (connection));

	/* configure read and write handlers and store default data to
	 * be used while sending and receiving data */
	vortex_tls_set_common_data (connection, ssl, ctx);

	/* do the initial connect connect */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "connecting to remote TLS site");
	while (SSL_connect (ssl) <= 0) {
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "accept function have failed (for initiator side)");
		switch (SSL_get_error (ssl, -1)) {
		case SSL_ERROR_WANT_READ:
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "still not prepared to continue because read wanted");
			break;
		case SSL_ERROR_WANT_WRITE:
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "still not prepared to continue because write wanted");
			break;
		default:
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "there was an error with the TLS negotiation");
			/* now the TLS process have failed, we have to
			 * restore how is read and written the data */
			vortex_connection_set_default_io_handler (connection);
			return FALSE;
		}
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "seems SSL connect call have finished in a proper manner");

	/* check remote certificate (if it is present) */
	server_cert = SSL_get_peer_certificate (ssl);
	if (server_cert == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "server side didn't set a certificate for this session, this is a bad news");
		vortex_support_free (2, ssl, SSL_free, ctx, SSL_CTX_free);
		return FALSE;
	}
	X509_free (server_cert);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS transport negotiation finished");
	
	return TRUE;
}

/** 
 * @internal
 *
 * @brief Support function for \ref vortex_tls_start_negociation which actually do
 * the work for that function without knowing if it is running as a
 * threaded mode or not.
 * 
 * @param data \ref VortexTlsBeginData having all data need to
 * negotiate the TLS process.
 * 
 * @return Always a NULL value.
 */
gpointer __vortex_tls_start_negociation (VortexTlsBeginData * data)
{
	VortexChannel         * channel         = NULL;
	VortexConnection      * connection      = data->connection;
	VortexConnection      * connection_aux  = NULL;
	gchar                 * serverName      = data->serverName;
	VortexTlsActivation     process_status  = data->process_status;
	gpointer                user_data       = data->user_data;
	VortexFrame           * reply           = NULL;
	GAsyncQueue           * queue;
	VORTEX_SOCKET           socket;
	gboolean                status          = FALSE;

	/* free no longer needed data  */
	g_free (data);

	/* Send TLS <ready> message. This will signal the remote peer
	* we need to negotiate a TLS secure layer, once the remote
	* site receive such petition, it should reply using <proceed>
	* or <error>.
	*
	* Actually send a <ready> message means to create a channel
	* using TLS profile piggybacking an initial data, that is,
	* the <ready /> token. 
	* 
	* Piggybacking is just a way to save a round trip based on
	* creating the channel based on TLS profile and then send a
	* MSG with a <ready /> token.  We send with the same message,
	* the channel creation request and the first message sent. If
	* the channel is accepted to be created, the profile content,
	* that is, the piggyback, will be accepted as the first
	* message. */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[1]: create the TLS channel");
	queue   = g_async_queue_new ();
	channel = vortex_channel_new_full (connection, /* the connection */
					   0,          /* the channel vortex chose */
					   serverName, /* the serverName value (no matter if it is NULL) */
					   /* the TLS profile identifier */
					   VORTEX_TLS_PROFILE_URI,
					   /* content encoding */
					   EncodingNone,
					   /* initial content or piggyback and its size */
					   "<ready />", 9,
					   /* close channel notification: we don't set it. */
					   NULL, NULL,
					   /* frame received notification: we don't set it. */
					   vortex_channel_queue_reply, queue,
					   /* on channel crated notification: we don't set it.
					    * It is not needed, we are working on a separated
					    * thread. */
					   NULL, NULL);

	/* As you may observed, previous code doesn't define a close
	 * handler. This is because once the channel is accepted to be
	 * created the TLS negotiation will start and no BEEP message
	 * will be exchanged until the TLS negotiation have finished,
	 * and all channels opened will be closed, including the
	 * channel 0.
	 * 
	 * Once terminated the TLS negotiation, no matter its result,
	 * a new channel 0 will be created, sending again the initial
	 * greeting.
	 *
	 * check channel status */
	if (channel == NULL) {
		/* unref queue */
		g_async_queue_unref (queue);

		/* notify user app level */
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 connection,
								 channel,
								 "Timeout reached while waiting for initial reply to TLS profile query",
								 user_data);
		return NULL;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[2]: TLS channel accepted, waiting for reply (as a piggyback or a plain message)");

	/* 2. wait for receive the proceed or error message:  
	 *
	 * Once the channel is started, we have to receive a reply
	 * message having a <error /> or <proceed />. That reply could
	 * come as a piggyback reply or a simple message. Both
	 * functionality is supported by the \ref
	 * vortex_channel_get_reply function. */

	reply = vortex_channel_get_reply (channel, queue);

	/* unref queue */
	g_async_queue_unref (queue);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[3]: TLS reply received");
	if (reply == NULL) {
		/* no reply received, within the timeout configured or
		 * remote peer have a poorly TLS implementation */
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 connection,
								 channel,
								 "Timeout reached while waiting for initial reply to TLS profile query",
								 user_data);
		return NULL;
	}

	/* In case of a error message is received, notify user space
	 * app and return */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[4]: check TLS reply received ");
	switch (vortex_frame_get_type (reply)) {
	case VORTEX_FRAME_TYPE_RPY:
		if (g_strcasecmp (vortex_frame_get_payload (reply), "<proceed />")) {
			/* content received */
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
			       "Wrong reply received from remote peer for TLS support (<proceed /> != '%s'",
			       vortex_frame_get_payload (reply));

			/* received a unexpected token reply */
			__vortex_tls_start_negociation_close_and_notify (process_status,
									 connection,
									 channel, 
									 "Wrong reply received from remote peer for TLS support",
									 user_data);
			return NULL;
		}

		break;
	case VORTEX_FRAME_TYPE_ERR:
		/* received a unexpected token reply */
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 connection,
								 channel, 
								 "TLS activations was rejected", user_data);
		return NULL;
	default:
		/* unexpected reply received */
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 connection,
								 channel, "Unexpected frame type received", 
								 user_data);
		return NULL;
	}
	/* free reply value */
	vortex_frame_free (reply);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[5]: everything is fine, close all channels..");


	/* prevent vortex library to close the underlying socket once
	 * the connection is closed. */
	vortex_connection_set_close_socket (connection, FALSE);
	

	/* know that it is accepted negotiate the under laying
	 * transport negotiate close all channels but the channel0 */
	if (!vortex_connection_close_all_channels (connection, TRUE)) {
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 connection,
								 channel,
								 "Unable to begin TLS transport negotiation because some channel rejects to be closed",
								 user_data);
		return NULL;
	}

	/* get the socket where the underlying transport negotiation
	 * is taking place. */
	socket = vortex_connection_get_socket (connection);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[6]: channels closed, start tls negotiation on socket=%d..", socket);

	connection_aux = connection;

	/* create a new connection object, set the socket, using the
	 * same connection reference. This is could be seen as not
	 * necessary but it is.  The \ref VortexConnection object have
	 * functions to to define what to do when the connection is
	 * closed, how the data is send and received, or some functions
	 * to store and retrieve data.  */
	connection = vortex_connection_new_empty_from_connection (socket, connection_aux, VortexRoleInitiator);

	/* unref the vortex connection and create a new empty one */
	__vortex_connection_set_not_connected (connection_aux, 
					       "connection instance being closed, without closing session, due to underlying TLS negotiation");
	
	/* the the connection to be blocking during the TLS
	 * negotiation, once called
	 * vortex_connection_parse_greetings_and_enable, the
	 * connection will be again non-blocking. */
	vortex_connection_set_blocking_socket (connection);

	/* do the TLS voodoo */
	status = vortex_tls_invoke_tls_activation (connection);

	vortex_log (LOG_DOMAIN, status ? G_LOG_LEVEL_DEBUG : G_LOG_LEVEL_CRITICAL, 
	       "TLS STAGE[6.1] TLS negociation status is: %s", status ? "OK" : "*** FAIL ***");
	
	/* seems TLS voodoo have finished, register the connection
	 * into the vortex writer: */
	vortex_writer_watch_connection      (connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[7]: TLS voodoo have finished, the connection is created and registered into the vortex writer, sending greetings..");


	/* 5. Issue again initial greetings, greetings just like we
	 * were creating a connection. */
	if (!vortex_greetings_client_send (connection)) {
                vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "failed while sending greetings after TLS negotiation");

		/* because the greeting connection have failed, we
		 * have to close and drop the \ref VortexConnection
		 * object. */
		vortex_connection_unref (connection, "(vortex tls profile)");

		/* notify that the TLS negotiation have failed
		 * because the initial greetings failure */
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 NULL,
								 NULL, 
								 "Initial greetings before a TLS negotiation has failed", user_data);
		return NULL;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[8]: greetings sent, waiting for reply");

	/* 6. Wait to get greetings reply (again issued because the
	 * underlying transport reset the session state) */
	reply = vortex_greetings_client_process (connection);
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[9]: reply received, checking content");

	/* 7. Check received greetings reply, read again supported
	 * profiles from remote site and register the connection into
	 * the vortex reader to start exchanging data. */
	if (!vortex_connection_parse_greetings_and_enable (connection, reply)) {

		/* flag the connection object to be not usable,
		 * unrefering it so eventually resources allocated
		 * will be collected. */
		vortex_connection_unref (connection, "(vortex tls profile)");

		/* notify the user the sad new! */
		__vortex_tls_start_negociation_close_and_notify (process_status,
								 NULL,
								 NULL,
								 "Greetings response parsing have failed before successful TLS negotiation", 
								 user_data);
		return NULL;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS STAGE[10]: greetings reply received is ok, connection is ready and registered into the vortex reader (that's all)");

	/* flag this connection to be already TLS-ficated if the
	 * status is ok */
	if (status) {
		vortex_connection_set_tlsfication_status (connection, TRUE);
	}

	/* 8. finally, report current TLS negotiation status to the
	 * app level.  notify the error */
	if (process_status != NULL) {
		/* notify user space */
		process_status (connection, 
				(status) ? VortexOk : VortexError, 
				(status) ? "TLS negotiation finished, now we can talk using the ciphered way!" :
				"TLS negotiation have failed!",  user_data);
	}
	return NULL;
}
#endif

/** 
 * @brief Starts the TLS underlying transport security negotiation
 * for the given connection.
 *
 * This function implements the client side for the TLS
 * profile. Inside the BEEP protocol definition, the TLS profile
 * provides a way to secure peer to peer data transmission, allowing
 * to take advantage of it for any profile implemented on top of the
 * BEEP stack.  TLS profile allows to secure underlying network
 * connection using the TLS v1.0 protocol.
 *
 * Here is an example on how to activate the TLS profile (details are explained after the example):
 * 
 * \code
 * // simple function with triggers the TLS process
 * void activate_tls (VortexConnection * connection) 
 * {
 *     // initialize and check if current vortex library supports TLS
 *     if (! vortex_tls_is_enabled ()) {
 *         g_print ("Unable to activate TLS, Vortex is not prepared\n");
 *         return;
 *     }
 *
 *     // start the TLS profile negotiation process
 *     vortex_tls_start_negociation (connection, NULL, 
 *                                   process_tls_status, NULL);
 * }
 *
 * // simple function with process the TLS status
 * void process_tls_status (VortexConnection * connection,
 *                          VortexStatus       status,
 *                          gchar            * status_message,
 *                          gpointer           user_data)
 * {
 *      switch (status) {
 *      case VortexOk:
 *           g_print ("TLS negotiation OK! over the new connection %d\n",
 *                    vortex_connection_get_id (connection));
 *
 *           // use the new connection reference provided by 
 *           // this function. Connection provided at 
 *           // vortex_tls_start_negociation have been unrefered.
 *           // In this case, my_server_connection have been 
 *           // already unrefered by the TLS activation
 *           my_server_connection = connection;           
 *
 *           break;
 *      case VortexError: 
 *           g_print ("TLS negotiation have failed, message: %s\n",
 *                    status_message);
 *           // ok, TLS process have failed but, do we have a connection
 *           // still working?. This is checked using
 *           if (vortex_connection_is_ok (connection, FALSE)) {
 *              // well we don't have TLS activated but the connection
 *              // still works
 *              my_server_connection = connection;
 *           } 
 *           break;
 *      }
 *      return;
 * }
 * \endcode
 * 
 * As the previous example shows, the function requires the connection
 * where the TLS process will take place, a handler to \ref VortexTlsActivation "notify the TLS status",
 * an optional serverName string to notify listener side to act as server name and an optional user data.
 *
 * While the TLS negotiation is running, user application could still
 * perform other tasks because this function doesn't block the
 * caller. Once the TLS process have finished, no matter its result,
 * the status will be notified using \ref VortexTlsActivation "process_status" handler.
 *
 * Because the TLS negotiation involves a tuning reset, the
 * connection provided to this function will be unrefered without
 * closing the underlying socket. This means that the connection
 * provided at this function must be no longer used, closed or
 * unrefered. 
 * 
 * The connection provided at \ref VortexTlsActivation
 * "process_status" handler is the one to be used once the TLS process
 * have finished properly.
 *
 * Because the TLS negotiation could fail user code placed at the
 * process function should check if the new connection provided on
 * that function is still working. This is because many things could
 * fail while TLS negotiation takes place. Some of them represents
 * protocol violation involving the connection to be closed, the
 * others just means the TLS have failed but the connection is still
 * valid not only to try again the TLS activation but to use it
 * without secure transmission.
 *
 * Once a connection is TLS-ficated the function just return without
 * doing nothing, if an attempt to secure the connection is made. You
 * could check TLS status for a given connection using \ref
 * vortex_connection_is_tlsficated.
 *
 * Vortex Library TLS design allows to build the library
 * enabling/disabling the TLS support. User space code doesn't need to
 * perform any especial check than calling to \ref vortex_tls_is_enabled function.
 * 
 * If previous function returns FALSE means that the Vortex Library
 * being used have no TLS support and any call to the TLS API will
 * perform no operation. This allows to write the same source code
 * without worry about Vortex Library TLS built-in support.
 *
 * If you need to know more about how to activate TLS support on
 * server side (well, actually for the listener role) check out this
 * function: \ref vortex_tls_accept_negociation.
 * 
 * Additionally, you can also consider to make Vortex Library to
 * automatically handle TLS profile negociation for every connection
 * created. This is done by using \ref vortex_connection_set_auto_tls
 * function. Once activated, every call to \ref vortex_connection_new
 * will automatically negotiate the TLS profile once connected before
 * notifying user space code for the connection creation.  *
 *
 * <b>NOTE:</b> As part of the TLS negotiation, all channels inside
 * the given connection will be closed.
 *
 * <b>NOTE:</b> The connection provided will be unrefered and a new
 * connection will be provided by this function, but using the same
 * underlying transport descriptor. This means the connection passed
 * to this function and the connection created by this function will
 * both link to the same peer but being different instances.
 * 
 * @param connection The connection where the secure underlying transport
 * will be started.
 *
 * @param serverName A server name value to be notified to the
 * remote peer so it could react in a different way depending on this
 * value. Function will perform a copy from the given value if it is
 * not a NULL one. You can free the passed in value just after the
 * function returns.
 *
 * @param process_status A handler to be executed once the process
 * have finished, no matter its result. This handler is required.
 *
 * @param user_data A user defined data to be passed in to the
 * <i>process_status</i> handler.
 *
 *
 */
void vortex_tls_start_negociation (VortexConnection     * connection,
				   gchar                * serverName,
				   VortexTlsActivation    process_status,
				   gpointer               user_data)
{
#ifndef ENABLE_TLS_SUPPORT
	/* write not supported TLS case */
	if (process_status != NULL) {
		process_status (connection, VortexError, 
				"TLS support wasn't build for this Vortex Library instance", user_data);
	}
	return;
#else
	/* TLS supported case */
	VortexTlsBeginData * data;

	/* check environment conditions */
	g_return_if_fail (connection);
	g_return_if_fail (vortex_connection_is_ok (connection, FALSE));
	g_return_if_fail (process_status);


	/* check for already TLS-fication */
	if (vortex_connection_is_tlsficated (connection)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Trying to TLS-ficate a connection already running TLS profile");
		if (process_status != NULL) {
			process_status (connection, VortexOk, 
					"Trying to TLS-ficate a connection already running TLS profile",
					user_data);
			return;
		} 
		/* no handler defined */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "no process status found, you should define this handler.");
		return;
	}


	/* check if remote site support remote profile (this step is
	 * actually not needed because vortex_channel_new already
	 * supports checking for remote profile to be supported. */
	if (!vortex_connection_is_profile_supported (connection, VORTEX_TLS_PROFILE_URI)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "remote peer doesn't support TLS profile, unable to activate TLS support");
		if (process_status != NULL) {
			process_status (connection, VortexError, 
					"Remote peer doesn't support TLS profile. Unable to negotiate secure transport layer",
					user_data);
			return;
		} 

		/* no handler defined */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "no process status found, you should define this handler.");
		return;
	}

	/* create and prepare invocation */
	data                 = g_new0 (VortexTlsBeginData, 1);
	data->connection     = connection;
	data->serverName     = (serverName != NULL) ? g_strdup (serverName) : serverName;
	data->process_status = process_status;
	data->user_data      = user_data;

	vortex_thread_pool_new_task ((GThreadFunc) __vortex_tls_start_negociation, data);
	return;
#endif
}

#ifdef ENABLE_TLS_SUPPORT

/** 
 * @internal
 *
 * @brief Perform final operations to accept the TLS incoming
 * negotiation.
 * 
 * @param connection The connection to be finally accepted.
 */
void vortex_tls_initial_accept (VortexConnection * connection)
{
	SSL     * ssl   = vortex_connection_get_data (connection, "ssl-data:ssl");
	gboolean status = FALSE;

	/* accept the incoming connection */
	if (SSL_accept (ssl) == -1) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "accept function have failed (for listener side)");
		switch (SSL_get_error (ssl, -1)) {
		case SSL_ERROR_WANT_READ:
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "still not prepared to continue because read wanted");
			return;
		case SSL_ERROR_WANT_WRITE:
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "still not prepared to continue because write wanted");
			return;
		default:
			/* TLS-fication process have failed */
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "there was an error while accepting TLS connection");
			
			/* restor default IO handlers  */
			vortex_connection_set_default_io_handler (connection);
			
		}
	}else {
		/* TLS-fication done */
		status = TRUE;
	}
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "TLS-fication status was: %s", status ? "OK" : "*** FAIL ***");

	/* seems TLS voodoo have finished, register the connection
	 * into the vortex writer: */
	vortex_writer_watch_connection      (connection);

	/* Disable preread handler for the connection. We have finished */
	vortex_connection_set_preread_handler (connection, NULL);

	if (status) {
		/* flag this connection to be already TLS-ficated */
		vortex_connection_set_tlsfication_status (connection, TRUE);
	}

	/* 5. Issue again initial greetings, greetings just like we
	 * were creating a connection. */
	if (!vortex_greetings_send (connection)) {
		/* because the greeting connection have failed, we
		 * have to close and drop the \ref VortexConnection
		 * object. */
		vortex_connection_unref (connection, "(vortex tls profile)");
		
	}

	return;
}


/** 
 * @internal
 * @brief Prepares the connection for the TLS profile.
 * 
 * This function is executed when the on close event have happen. This
 * means the all channels are closed and we can start the TLS handshake.
 * 
 * @param connection The connection being closed.
 */
void vortex_tls_prepare_listener (VortexConnection * connection)
{
	SSL_METHOD       * meth;
	SSL_CTX          * ctx;
	SSL              * ssl;
	gchar            * certificate_file;
	gchar            * private_file;
	VortexConnection * new_connection;
	VORTEX_SOCKET   socket;



	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "State[1] session is being closed this means we can start negotiating TLS");

	socket = vortex_connection_get_socket (connection);
	meth   = TLSv1_server_method ();
	ctx    = SSL_CTX_new (meth);
	if (ctx == NULL) {
		/* drop a log and instruct vortex library to close the
		 * given connection including the socket. */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "unable to create SSL context object, unable to start TLS profile");
		vortex_connection_set_close_socket (connection, TRUE);
		return;
	}
	
	/* configure certificate file */
	certificate_file = vortex_connection_get_data (connection, "tls:certificate-file");
	g_print ("Using certificate: %s\n", certificate_file);
	if (SSL_CTX_use_certificate_file (ctx, certificate_file, SSL_FILETYPE_PEM) <= 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "there was an error while setting certificate file into the SSl context, unable to start TLS profile");
		vortex_connection_set_close_socket (connection, TRUE);
		return;
	}

	/* configure private file */
	private_file    = vortex_connection_get_data (connection, "tls:private-file");
	if (SSL_CTX_use_PrivateKey_file (ctx, private_file, SSL_FILETYPE_PEM) <= 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "there was an error while setting private file into the SSl context, unable to start TLS profile");
		vortex_connection_set_close_socket (connection, TRUE);
		return;
	}

	/* check for private key and certificate file to match. */
	if (!SSL_CTX_check_private_key(ctx)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "seems that certificate file and private key doesn't match!, unable to start TLS profile");
		vortex_connection_set_close_socket (connection, TRUE);
		return;
	}
	
	/* create ssl object */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "initializing TLS transport");
	ssl = SSL_new (ctx);       
	if (ssl == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "error while creating TLS transport");
		vortex_connection_set_close_socket (connection, TRUE);
		return;
	}

	/* set the file descriptor */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "setting file descriptor");
	SSL_set_fd (ssl, socket);

	/* prepare the new connection */
	new_connection = vortex_connection_new_empty_from_connection (socket, connection, VortexRoleListener);

	/* release previous objet */
	__vortex_connection_set_not_connected (connection, 
					       "connection instance being closed, without closing session, due to underlaying TLS negoctiation");


	/* set default handlers to write/read and default objects to
	 * be used */
	vortex_tls_set_common_data (new_connection, ssl, ctx);

	/* set the preread handler, where the tls connection will be
	 * finally accepted */
	vortex_connection_set_preread_handler (new_connection, vortex_tls_initial_accept);

	/* accept the connection in an initial step (actually not
	 * executed until the pre-read handler is not executed). This
	 * function flag the connection to be on the initial stage and
	 * register the connection into the vortex reader. */
	vortex_listener_accept_connection (new_connection);
	return;
}

/** 
 * @internal
 *
 * @brief Vortex TLS start handler which process and notify user space
 * all start TLS channel queries.
 * 
 * 
 * 
 * @param profile 
 * @param channel_num 
 * @param connection 
 * @param serverName 
 * @param profile_content 
 * @param profile_content_reply
 * @param encoding 
 */
gboolean vortex_tls_process_start_msg (gchar             * profile,
				       gint                channel_num,
				       VortexConnection  * connection,
				       gchar             * serverName,
				       gchar             * profile_content,
				       gchar            ** profile_content_reply,
				       VortexEncoding      encoding,
				       gpointer            user_data)
{
	gchar * certificate_file;
	gchar * private_key_file;

	/* flag this connection to be already TLS-ficated */
	if (vortex_connection_is_tlsficated (connection)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "received a TLS channel request on an already TLS-ficated connection.");
		return FALSE;
	}

	/* check profile content */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received TLS negotiation request");
	if (profile_content == NULL || g_strcasecmp (profile_content, "<ready />")) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "received a TLS channel request having a missing or wrong profile content");
		return FALSE;
	}

	/* notify user app level that a TLS profile request have being
	 * received and it is required to act as serverName. */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "checking if applications level allows to negotiate the TLS profile");
	if (!vortex_tls_accept_handler (connection, serverName)) {
		(* profile_content_reply)  = vortex_frame_get_error_message ("421", "applications level have reject to negotiate TLS transport layer", NULL);
		
		/* we have to reply TRUE because the expected reply is
		 * a positive one with a proceed or error content. */
		return TRUE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "application level seems to accept negotiate the TLS profile, getting certificate");

	/* get TLS certificate file */
	certificate_file = vortex_tls_certificate_handler (connection, serverName);
	if (certificate_file == NULL) {
		(* profile_content_reply)  = vortex_frame_get_error_message ("421", 
									     "applications level didn't provide a valid location for a certificate file, unable to negotiate TLS transport", 
									     NULL);
		/* we have to reply TRUE because the expected reply is
		 * a positive one with a proceed or error content. */
		return TRUE;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "getting private key file");

	/* get TLS private key file */
	private_key_file = vortex_tls_private_key_handler (connection, serverName);
	if (private_key_file == NULL) {
		(* profile_content_reply)  = vortex_frame_get_error_message ("421", 
									     "applications level didn't provide a valid location for a private key file, unable to negotiate TLS transport", 
									     NULL);
		/* we have to reply TRUE because the expected reply is
		 * a positive one with a proceed or error content. */
		return TRUE;
	}

	/* set the reply to for the TLS channel negotiation. Memory
	 * holding profile content reply should be dynamically
	 * allocated because vortex library will deallocate it. */
	(* profile_content_reply)  = g_strdup ("<proceed />");
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "replying peer that TLS negotiation can start");

	/* Here goes a little tricky-bad-thing-dont-repeat-it-at-home
	 * 
	 * 1) we prepare the socket connection to not be closed even
	 * if the channel 0 is closed on the given connection. */
	vortex_connection_set_close_socket (connection, FALSE);

	/* 2) store certificate and private key files */
	vortex_connection_set_data_full    (connection, "tls:certificate-file", certificate_file, NULL, g_free);
	vortex_connection_set_data_full    (connection, "tls:private-file",     private_key_file, NULL, g_free);

	/* 3) use the on close event to get notify when to start the
	 * TLS handshake */
	vortex_connection_set_on_close     (connection, vortex_tls_prepare_listener);

	return TRUE;
}


/** 
 * @internal
 * This default hander definition is used when no accept handler was
 * defined. It always returns TRUE, which means to always accept the
 * TLS request.
 */
gboolean vortex_tls_default_accept     (VortexConnection * connection,
				        gchar            * serverName)
{
	return TRUE;
}

/** 
 * @internal
 * @brief Default certificate locator function which returns the test
 * certificate used by the Vortex Library.
 */
gchar * vortex_tls_default_certificate (VortexConnection * connection,
				        gchar            * serverName)
{
	return vortex_support_find_data_file ("test-certificate.pem");
}

/** 
 * @internal
 *
 * @brief Default private key locator function which returns the test
 * private key used by the Vortex Library.
 */
gchar * vortex_tls_default_private_key (VortexConnection * connection,
				        gchar            * serverName)
{
	return vortex_support_find_data_file ("test-private-key.pem");
}


/** 
 * @internal
 *
 * @brief Support function for \ref vortex_tls_start_negociation_sync
 * function.
 */
void __vortex_tls_start_negociation_sync_process (VortexConnection * connection,
						  VortexStatus       status,
						  gchar            * status_message,
						  gpointer           user_data)
{
	GAsyncQueue * queue = user_data;

	/* push and unref man! (first status, then status_message and
	 * finally the connection). The following sentences must
	 * follow this order. */
	g_async_queue_push  (queue, GINT_TO_POINTER (status));
	g_async_queue_push  (queue, status_message);
	g_async_queue_push  (queue, connection);
	g_async_queue_unref (queue);
	
	return;
}
#endif


/** 
 * @brief Allows to start a TLS profile negotiation in a synchronous
 * way (blocking the caller).
 *
 * This actually uses \ref vortex_tls_start_negociation. Check out
 * \ref vortex_tls_start_negociation documentation to know more about
 * activating TLS profile.
 * 
 * If TLS profile is not supported the function returns the same
 * connection received. To avoid problems you should use \ref
 * vortex_tls_is_enabled first.
 * 
 * @param connection The connection where the TLS profile negotiation will take place.
 * @param serverName The serverName optional value to request remote peer to act as server name.
 * @param status     An optional reference to a \ref VortexStatus variable so the caller could get a TLS activation status.
 * @param status_message An optional reference to get a textual diagnostic about TLS activation status.
 * 
 * @return The new connection with TLS profile activated. 
 */
VortexConnection * vortex_tls_start_negociation_sync     (VortexConnection * connection,
							  gchar            * serverName,
							  VortexStatus      * status,
							  gchar            ** status_message)
{
#ifndef ENABLE_TLS_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "trying to execute a synchronous tls-fication on a vortex library without TLS support");
	if (status != NULL)
		(* status)         = VortexError;
	if (status_message != NULL)
		(* status_message) = "trying to execute a synchronous tls-fication on a vortex library without TLS support";
	return NULL;
#else	
	/* new connection received */
	VortexConnection * _connection;
	/* timeout object */
	GTimeVal           timeout;
	/* first value recevied from async process handler */
	gpointer           _status;
	
	/* create an async queue and increase queue reference so the
	 * function could unref it without worry about race conditions
	 * with sync_process function. */
	GAsyncQueue * queue = g_async_queue_new ();
	g_async_queue_ref (queue);

	/* wait for the data to be returned, first prepare timeout
	 * value */
	g_get_current_time (&timeout);
	g_time_val_add (&timeout, vortex_connection_get_timeout ());

	/* start TLS negotiation */
	vortex_tls_start_negociation (connection, serverName, 
				      __vortex_tls_start_negociation_sync_process,
				      queue);

	/* get status */
	_status = g_async_queue_timed_pop (queue, &timeout);
	if (_status == NULL) {
		/* seems timeout have happen while waiting for SASL to
		 * end */
		(* status)         = VortexError;
		(* status_message) = "Timeout have been reached while waiting for TLS to finish";
		return NULL;
	}

	/* get status */
	if (status != NULL)
		(* status) = GPOINTER_TO_INT (_status);
	
	/* get message */
	if (status_message != NULL)
		(* status_message) = g_async_queue_pop (queue);
	else
		g_async_queue_pop (queue);

	/* get connection */
	_connection = g_async_queue_pop (queue);

	/* unref the queue */
	g_async_queue_unref (queue);

	/* return received value */
	return _connection;
#endif
}


/** 
 * @brief Allows to configure if current Vortex Library instance will
 * accept TLS incoming connections.
 * 
 * While using TLS under BEEP, as a peer protocol, any instance
 * running could receive a TLS request for its activation. This
 * function allows to configure if such request will be allowed or
 * recognized. 
 * 
 * Default TLS configuration is to not allow receive a TLS request. 
 *
 * This function doesn't disable the possibility to connect to a remote
 * peer and request to negotiate the TLS security transport. 
 *
 * There are two typical scenarios:
 *
 * - 1. Vortex Library client applications doesn't call this function,
 * so any other BEEP peer, that the client instance connects to, can't request to activate TLS transport. But
 * Vortex Library client could issue a call to \ref
 * vortex_tls_start_negociation to enable TLS transport against a BEEP
 * peer accepting TLS requests. In the case the remote peer is running
 * Vortex Library, that peer have already issued a call to \ref vortex_tls_accept_negociation.
 * 
 * - 2. In the other hand, Vortex Library listeners could enable
 * accepting TLS incoming connections so they publish as a possible
 * profile the TLS one. This is done by calling to this function. This
 * action enables the possibility to receive a TLS request to secure
 * the connection.
 *
 * This function allows to define several handlers to configure the
 * TLS support. This handler are defined per profile which means they
 * are global to all TLS profile request received.
 *
 * @param accept_handler A handler executed to notify user app level
 * that a TLS request was received, allowing to accept or deny it
 * according to the value returned by the handler. You can use NULL
 * value for this parameter. This will make Vortex Library to set the
 * default accept handler which always accept every TLS negotiation.
 * 
 * @param certificate_handler A handler executed to know where is
 * located the certificate file to be used to cipher the session. You
 * can use NULL value for this parameter. This will make Vortex
 * Library to set the default certificate handler which returns a path
 * to a test certificate. It is highly recommended to set this handler,
 * however you can use NULL value under development environment. 
 *
 * @param private_key_handler A handler executed to know where is
 * located the private key file to be used to cipher the session. You
 * can use NULL value for this parameter. This will make Vortex Library
 * to set the default private key handler which returns a path to the
 * test private key. It is highly recommended to set this handler,
 * however you can use NULL values under development environment.
 *
 * @return Returns it the current server instance could accept incoming TLS connections.
 */
gboolean    vortex_tls_accept_negociation (VortexTlsAcceptQuery            accept_handler, 
					   VortexTlsCertificateFileLocator certificate_handler,
					   VortexTlsPrivateKeyFileLocator  private_key_handler) 
{
#ifndef ENABLE_TLS_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "current Vortex Library doesn't have TLS support");
	return FALSE;
#else
	if (!vortex_tls_is_enabled ()) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "unable to initialize TLS library, unable to accept incoming requests");
		return FALSE;
	}

	/* set default handlers */
	vortex_tls_accept_handler      = (accept_handler      != NULL) ? accept_handler      : vortex_tls_default_accept;
	vortex_tls_certificate_handler = (certificate_handler != NULL) ? certificate_handler : vortex_tls_default_certificate;
	vortex_tls_private_key_handler = (private_key_handler != NULL) ? private_key_handler : vortex_tls_default_private_key;

	/* register the profile */
	vortex_profiles_register (VORTEX_TLS_PROFILE_URI,
				  /* the start handler */
				  NULL, NULL,
				  /* the close handler */
				  NULL, NULL,
				  /* the frame received */
				  NULL, NULL);

	/* register the extended start message to process all incoming 
	 * start messages received. */
	vortex_profiles_register_extended_start (VORTEX_TLS_PROFILE_URI,
						 vortex_tls_process_start_msg,
						 NULL);

	/* well, we are prepare to tls-ficate! */
	return TRUE;
#endif
}


/* @} */
