/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#include <vortex.h>


#define LOG_DOMAIN "vortex-writer"

GAsyncQueue * vortex_writer_queue    = NULL;
GAsyncQueue * vortex_writer_stopped  = NULL;
GList       * con_writer_list        = NULL;
gpointer      on_writing;

extern VortexSemaphore * vortex_sequencer_sem;


/**
 * @internal
 * 
 * Allows to register a new connection to watch inside the vortex writer process so,
 * frames queued on channels over that connection are sent.
 * This function is not useful for vortex library consumer. It is for internal vortex
 * library purposes.
 **/
void vortex_writer_watch_connection (VortexConnection * connection)
{
	g_return_if_fail (vortex_connection_is_ok (connection, FALSE));
	
	/* increase reference counting */
	vortex_connection_ref (connection, "vortex writer (process)");

	g_async_queue_push (vortex_writer_queue, connection);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "new connection on writer to be watched (%d)",
	       vortex_connection_get_socket (connection));


	return;
			  
}

void __vortex_writer_unref_connection (VortexConnection * connection)
{
	gint pending_msgs = 0;

	/* ensure we are not unrefering a vortex connection with
	 * pending frames to be sent. If we found message pending on
	 * this connection decrease the semaphore. */
	pending_msgs = vortex_connection_get_pending_msgs (connection);
	if (pending_msgs > 0) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, 
		       "vortex writer is unrefering a connection with pending messages=%d, decreasing vortex sequencer semaphore %d units",
		       pending_msgs,
		       pending_msgs);
		vortex_semaphore_down_n (vortex_sequencer_sem, pending_msgs);
	}
	vortex_connection_unref (connection, "vortex writer (process)");	

	return;
}


gboolean __vortex_writer_check_connection (VortexConnection * connection)
{
	/* check the connection and unref it, including its pending
	 * messages if it is not connected. */
	if (!vortex_connection_is_ok (connection, FALSE)) {

		/* unref the connection */
		__vortex_writer_unref_connection (connection);
		
		return FALSE;
	}
	return TRUE;
}

VORTEX_SOCKET __vortex_writer_build_set_to_watch (gpointer on_writing)
{
	GList            * list_aux     = g_list_first (con_writer_list);
	GList            * new_list     = NULL;
	VORTEX_SOCKET      max_fds      = 0;
	VORTEX_SOCKET      fds          = 0;
	VortexConnection * connection;

	/* read connection list */
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		connection = list_aux->data;

		/* check connection sanity */
		if (!__vortex_writer_check_connection (connection))
			continue;
		
		/* get the socket, and get the maximum socket
		 * descriptor watched */
		fds        = vortex_connection_get_socket (connection);
		max_fds    = MAX (fds, max_fds);

		/* add the socket descriptor into the fd set */
		vortex_io_waiting_invoke_add_to_fd_group (fds, on_writing);

 		new_list   = g_list_append (new_list, connection);
	}

	/* update connection list */
	g_list_free (con_writer_list);
	con_writer_list = g_list_first (new_list);

	/* connection list watch */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "connection watched by the writer");
	list_aux        = g_list_first (con_writer_list);
	for (;list_aux; list_aux = g_list_next (list_aux)) {
		connection = list_aux->data;
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "   connection to id:%d) %s:%s (socket=%d, refs=%d)", 
			    vortex_connection_get_id   (connection),
			    vortex_connection_get_host (connection),
			    vortex_connection_get_port (connection),
			    vortex_connection_get_socket (connection),
			    vortex_connection_ref_count (connection));
			    
			    
	}

	return max_fds;
}

/** 
 * @internal
 * 
 * Support function which finally implements the vortex writer
 * stopping process, deallocating all resources.
 */
void __vortex_writer_stop_process ()
{
	VortexConnection * connection;
	GList            * list;

	/* unref the queue */
	g_async_queue_unref (vortex_writer_queue);
	vortex_writer_queue = NULL;
	
	/* unref every connection registered */
	list = g_list_first (con_writer_list);
	for (; list; list = g_list_next (list)) {
		/* get a reference to the connection */
		connection = list->data;
		
		/* unref the connection */
		__vortex_writer_unref_connection (connection);
	}
	
	/* free the list */
	g_list_free (con_writer_list);

	/* destroy the IO abstraction */
	vortex_io_waiting_invoke_destroy_fd_group (on_writing); 
	
	/* signal that the writer is stoped */
	g_async_queue_push (vortex_writer_stopped, GINT_TO_POINTER (1));

	return;
}

gboolean __vortex_writer_watch_first_connection ()
{
	VortexConnection * connection;

	/* wait for the first connection to watch */
	connection      = g_async_queue_pop (vortex_writer_queue);

	/* check if we have being signaled to stop the process */
	if (connection == GINT_TO_POINTER (1)) {
		
		/* stop the writer process */
		__vortex_writer_stop_process ();
		
		/* return that the process can continue */
		return FALSE;
	}


	/* register new connection in list */
	con_writer_list = g_list_append (con_writer_list, connection);

	return TRUE;
}

/** 
 * @internal
 *
 * Internal functions to check more connection to be watched by the
 * vortex writer.
 */
gboolean __vortex_writer_read_pending ()
{
	VortexConnection * connection;
	gint               length;

	con_writer_list = g_list_first (con_writer_list);

	/* reading pending connection */
	length = g_async_queue_length (vortex_writer_queue);
	while (length > 0) {
		
		length--;
		connection = g_async_queue_pop (vortex_writer_queue);

		/* check if we have being signaled to stop the
		 * process */
		if (connection == GINT_TO_POINTER (1)) {
		
			/* stop the writer process */
			__vortex_writer_stop_process ();

			/* return that the process can continue */
			return FALSE;
		}
		
		/* check connection sanity */
		if (!__vortex_writer_check_connection (connection))
			continue;

		/* register the connection */
		con_writer_list = g_list_append (con_writer_list, connection);
	}

	/* return that the process could continue */
	return TRUE;
}

/** 
 * @internal
 * 
 * Stops the vortex writer process.
 */
void vortex_writer_stop ()
{
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "stopping the vortex writer process..");

	/* signal the writer that the process must be stoped and
	 * resources used must be deallocated */
	g_async_queue_push (vortex_writer_queue, GINT_TO_POINTER (1));

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "signalling on the semaphore");
	vortex_semaphore_up (vortex_sequencer_sem);
	
	/* wait until the writer ends */
	g_async_queue_pop (vortex_writer_stopped);
	g_async_queue_unref (vortex_writer_stopped);

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "process stopped");
}

gpointer __vortex_writer_run (gpointer __data)
{
	VortexConnection * connection;
	VORTEX_SOCKET      fds;
	gint               msgs;
	gint               status;
	
	VORTEX_SOCKET      max_fds;
	gint               result;
	GList            * list_aux;

	/* initialize the vortex io waiting */
	on_writing = vortex_io_waiting_invoke_create_fd_group ();

 first_con:

	/*wait for the first connection to watch */
	if (!__vortex_writer_watch_first_connection ())
		return NULL;

	while (TRUE) {
		/* check if we have nothing to watch */
		if (g_list_length (con_writer_list) == 0) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "no more connection to watch on vortex writer");
			goto first_con;
		}

		/* wait until sequencer have queued some data */
		vortex_semaphore_wait (vortex_sequencer_sem);

		/* read the pending queue elements to be watched. Read
		 * new connections to be managed, until the stop
		 * signal is received. */
		if (!__vortex_writer_read_pending ())
			return NULL;

		/* clear set */
		vortex_io_waiting_invoke_clear_fd_group (on_writing);
		
		/* build list connection to watch */
		max_fds = __vortex_writer_build_set_to_watch (on_writing);

		/* wait until something have changed and init
		 * timeout */
		result = vortex_io_waiting_invoke_wait (on_writing, max_fds, WRITE_OPERATIONS);

		if (result == -1 && result == -2 && result == 0) {
			/* waiting timeout have occur or no descriptor
			 * have change or it has happened an error
			 * that is not required to do any more than do
			 * again a wait operation. Its time to check
			 * the queue */
			continue;
		}

		/* we can write, for each connection check if we can
		 * write */
		list_aux       = g_list_first (con_writer_list);
		for (; list_aux; list_aux = g_list_next (list_aux)) {
			/* check if connection is ok */
			connection = list_aux->data;
			if (!vortex_connection_is_ok (connection, FALSE))
				continue;

			fds  = vortex_connection_get_socket (connection);

			/* check if the given socket is prepare to
			 * write */
			if (vortex_io_waiting_invoke_is_set_fd_group (fds, on_writing)) {

				msgs = vortex_connection_do_a_sending_round (connection);
				/* if no message was sent, do not
				 * decrease anything */
				if (msgs == 0) {
					status = vortex_semaphore_get_status (vortex_sequencer_sem);
					if (status == 0) {
						vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
						       "pending frames to be sent ok, writer is going to take a \"siesta\" now");
						break;
					}

					vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
					       "no message sent for connection id=%d but semaphore status signal there are %d message waiting",
					       vortex_connection_get_id (connection),
					       status);
					continue;
				}				
				/* decrease the semaphore for each
				 * message sent */
				vortex_semaphore_down_n (vortex_sequencer_sem, msgs);

			} /* end if */

		} /* end for */

	} /* vortex writer while loop end */

	return NULL;
}

/**
 * vortex_writer_run:
 * 
 * Launches the vortex writer. The vortex writer in conjunction with vortex sequencer are the processes
 * that actually make possible to send frames. Once the vortex sequencer have prepared the frames to
 * be sent and queue them into each channel queue, the vortex writer takes waits for that frames
 * to send them in a round robin manner. 
 **/
void vortex_writer_run ()
{

	/* the vortex writer queue: this queue is used to register new
	 * connection to be watched in order to send frames */
	vortex_writer_queue = g_async_queue_new ();

	/* the vortex writer stop queue, used to stop the vortex
	 * writer process */
	vortex_writer_stopped = g_async_queue_new ();
	
	/* launch the vortex writer thread */
	g_thread_create (__vortex_writer_run, NULL, FALSE, NULL);

	return;
}


/** 
 * @internal
 * @brief frees Vortex Writer Data
 * 
 * @param writer_data 
 */
void vortex_writer_data_free (VortexWriterData * writer_data)
{
	g_return_if_fail (writer_data);
	
	g_free (writer_data->the_frame);
	g_free (writer_data);

	return;
}
