/**
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */
#include <vortex.h>

#define LOG_DOMAIN "vortex-xml-rpc"

#define XML_RPC_BOOT_STATE "vortex-xml-rpc:state"

#define XML_RPC_RESOURCE "vortex-xml-rpc:resource"

VortexXmlRpcValidateResource  __vortex_xml_rpc_validate_resource = NULL;
gpointer                      __vortex_xml_rpc_validate_data     = NULL;
VortexXmlRpcServiceDispatch   __vortex_xml_rpc_service_dispatch  = NULL;
gpointer                      __vortex_xml_rpc_service_data      = NULL;

/**
 * \defgroup vortex_xml_rpc Vortex XML-RPC: XML-RPC profile support and related functions
 */

/* \addtogroup vortex_xml_rpc */
/* @{ */

/** 
 * @brief Allows to get current Vortex Library XML-RPC built-in
 * support.
 *
 * Because XML-RPC profile support could be disabled from be built,
 * this function allows to get current supporting status for the
 * Vortex Library. 
 *
 * Any source code built on top of this API (to get access to XML-RPC
 * features) should call to this function first to ensure the Vortex
 * Library was built with XML-RPC support:
 *
 * \code
 * // check for XML-RPC profile support
 * if (!vortex_xml_rpc_is_enabled ()) {
 *     g_print ("Current vortex library doesn't support XML-RPC\n");
 *     return;
 * }
 *
 * // perform XML-RPC operations.
 * \endcode
 *
 * @return TRUE when the library have built-in XML-RPC support, otherwise FALSE is returned.
 */
gboolean vortex_xml_rpc_is_enabled () 
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Current Vortex Library wasn't built with XML-RPC support");
	return FALSE;
#else
	/* just return TRUE (nothing to initialize) */
	return TRUE;
#endif
}

/** 
 * @internal
 * @brief Support user space notification for XML-RPC boot status notification.
 * 
 * @param channel    The channel created due to the boot process or NULL if fails.
 * @param status     A process status
 * @param message    The message to report. A textual diagnostic.
 * @param user_data  User space data.
 */
void __vortex_xml_rpc_notify (VortexXmlRpcBootNotify    process_status, 
			      VortexChannel           * channel,
			      VortexStatus              status,
			      gchar                   * message, 
			      gpointer                  user_data)
{
	/* drop to the console a log */
	switch (status) {
	case VortexError:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, (message != NULL) ? message : "no message to report");

		/* close the channel if an error have happened while
		 * booting the XML-RPC channel. */
		if (channel != NULL)
			vortex_channel_close (channel, NULL);
		break;
	case VortexOk:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, (message != NULL) ? message : "no message to report");

		/* set the channel to be ready. */
		vortex_channel_set_data (channel, XML_RPC_BOOT_STATE, "ready");
		break;
	}

	/* notify user space that something to be notify have
	 * happen */
	if (process_status != NULL)
		process_status (channel, status, message, user_data);

	/* that's all man */
	return;
}

/** 
 * @internal 
 *
 * @brief Support function to perform user space notification with
 * operation status for xml-rpc invocations.
 *
 * This function is mainly used by the following functions:
 *
 *  - \ref vortex_xml_rpc_invoke
 *  - \ref vortex_xml_rpc_invoke_sync
 *
 * What the function do is to notify to user space code with the
 * result and, according to the results, drop a log.
 *
 * The function also creates a method response object containing the
 * error message if an error is detected.
 * 
 * @param notify The \ref XmlRpcInvokeNotify handler.
 * 
 * @param status The status to report.
 *
 * @param status_message The status message to report.
 *
 * @param channel The channel where the invocation was performed.
 *
 * @param response The response object get from the remote side or
 * NULL if no response object was generated. In this case
 * <i>status</i> variable should be checked to generate a new \ref
 * XmlRpcMethodResponse object reporting the failure.
 *
 * @param user_data User space data.
 */
void __vortex_xml_rpc_notify_response (XmlRpcInvokeNotify     notify,
				       XmlRpcResponseStatus   status,
				       gint                   fault_code,
				       gchar                * fault_message, 
				       VortexChannel        * channel,
				       XmlRpcMethodResponse * response,
				       gpointer               user_data)
{
	switch (status) {
	case XML_RPC_OK:
		/* ok reply received */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invocation reply received!");
		break;
	default:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_WARNING, "invocation failed, message received: (code: %d) %s", 
		       fault_code, fault_message);
		/* bad reply received */
		if (response == NULL) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "building method response stub containing the textual error");
			/* create a XmlRpcMethodResponse object */
			response = method_response_new (status, fault_code, fault_message, NULL);
		}
		break;
		
	}
	/* notify user space about response */
	if (notify != NULL)
		notify (channel, response, user_data);

	return;
}
				       

#ifdef ENABLE_XML_RPC_SUPPORT

/** 
 * @internal
 *
 * Internal function to support parsing xml content received,
 * representing an struct construction. This function is called from
 * __vortex_xml_rpc_parse_value.
 * 
 * @param child The node received represents the xml node reference
 * pointing to the <struct>.
 * 
 * @return A newly allocated method value, containing a representation
 * of a struct or NULL if fails.
 */
XmlRpcMethodValue * __vortex_xml_rpc_parse_struct_value (axlNode * struct_node)
{
	/* the variable which will hold the result */
	gchar              * member_name;
	XmlRpcMethodValue  * member_value;
	XmlRpcStruct       * _struct;
	XmlRpcStructMember * member;
	

	/* some node references */
	axlNode            * name;
	axlNode            * value;
	axlNode            * node_member;


	/* some iterators */
	int                  iterator;
	int                  length;

	/* check the struct node name */
	if (! NODE_CMP_NAME (struct_node, "struct")) {
		/* it seems we have received a node reference which is
		   not an struct */
		return NULL; 
	} 

	/* get the count number of childs inside the struct */
	length   = axl_node_get_child_num (struct_node);

	if (length == 0) {
		/* it seems we have received an struct with no
		 * members */
		return NULL;
	}

	/* create the struct */
	_struct = vortex_xml_rpc_struct_new (length);

	/* iterate over all childs (<member> items) inside the
	 * <struct> received */
	iterator = 0;
	while (iterator < length) {
		/* get a referenced to the member node */
		node_member = axl_node_get_child_nth (struct_node, iterator);
		
		/* get the member name */
		name         = axl_node_get_child_nth (node_member, 0);
		member_name  = axl_node_get_content (name, NULL);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "parsing struct member name=%s",
		       member_name);
		
		/* get the member content */
		value        = axl_node_get_child_nth (node_member, 1);
		member_value = __vortex_xml_rpc_parse_value (value);

		/* create the member and add it to the struct */
		member = vortex_xml_rpc_struct_member_new (member_name, member_value);
		vortex_xml_rpc_struct_add_member (_struct, member);
		
		/* next member to parse */
		iterator++;
	}

	/* return the structure parsed */
	return method_value_new (XML_RPC_STRUCT_VALUE, _struct);
}

/** 
 * @internal
 *
 * Parses a received xml stream representing an xml rpc array.
 *
 * @param array_node The node pointing to the first node <array>.
 * 
 * @return A newly allocated xml-rpc array.
 */
XmlRpcMethodValue * __vortex_xml_rpc_parse_array_value (axlNode * array_node)
{
	/* the variable which will hold the result */
	XmlRpcMethodValue  * method_value;
	XmlRpcArray        * array;
	
	/* some node references */
	axlNode            * value  = NULL;
	axlNode            * data   = NULL;

	/* some iterators */
	int                  iterator;
	int                  length;

	/* check the struct node name */
	if (! NODE_CMP_NAME (array_node, "array")) {
		/* it seems we have received a node reference which is
		   not an struct */
		return NULL; 
	} 
	
	/* get a reference to the child node */
	data = axl_node_get_child_nth (array_node, 0);

	/* check the struct node name */
	if (! NODE_CMP_NAME (data, "data")) {
		/* it seems we have received a node reference which is
		   not an struct */
		return NULL; 
	} 

	/* get the count number of childs inside the array (<data> node) */
	length   = axl_node_get_child_num (data);
	if (length == 0) {
		/* seens we have received and empty array reference */
		return NULL;
	}

	/* create the struct */
	array = vortex_xml_rpc_array_new (length);

	/* iterate over all childs inside the <data> node */
	iterator = 0;
	while (iterator < length) {

		if (iterator == 0) {
			/* get the first value inside the <data> node */
			value = axl_node_get_child_nth (data, iterator);
		}else {
			/* get the next element in the sequence */
			value = axl_node_get_next (value);
		}

		/* parse the value node */
		method_value = __vortex_xml_rpc_parse_value (value);

		/* create the member and add it to the struct */
		vortex_xml_rpc_array_add (array, method_value);

		/* next member to parse */
		iterator++;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "returning an array with (%d) items", 
		    vortex_xml_rpc_array_count (array));

	/* return the structure parsed */
	return method_value_new (XML_RPC_ARRAY_VALUE, array);
}

/** 
 * @internal 
 *
 * @brief Parse the given chunk of xml data, supposing that it points
 * to ta <value /> node.
 *
 * The function returns an allocated \ref XmlRpcMethodValue
 * representing the value that the XML chunk contains.
 * 
 * @param cursor The cursor representing the value to be read.
 * 
 * @return A newly allocated \ref XmlRpcMethodValue.
 */
XmlRpcMethodValue * __vortex_xml_rpc_parse_value (axlNode * value)
{
	gchar             * string_value = NULL;
	XmlRpcMethodValue * result       = NULL;
	axlNode           * child        = NULL;

	g_return_val_if_fail (value, NULL);

	/* check that the received cursor contains a <value/> node */
	if (! NODE_CMP_NAME (value, "value")) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "received a wrong top level node where expected <value />. Received <%s>.", 
		       axl_node_get_name (value));
		return NULL;		
	}
	
	/* get children node */
	child = axl_node_get_child_nth (value, 0);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received a value tag, containing a <%s>", 
	       axl_node_get_name (child));

	/* check for the none value */
	if (NODE_CMP_NAME (child, "none")) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "none value received");
		return method_value_new (XML_RPC_NONE_VALUE, NULL);
	}

	/* in the case of simple top level values, get the reference
	 * to the internal content */
	if (NODE_CMP_NAME (child, "i4") ||
	    NODE_CMP_NAME (child, "int") ||
	    NODE_CMP_NAME (child, "boolean") ||
	    NODE_CMP_NAME (child, "string") ||
	    NODE_CMP_NAME (child, "double") ||
	    NODE_CMP_NAME (child, "dateTime.iso8601") ||
	    NODE_CMP_NAME (child, "base64")) {

		/* reference to the internal content for simple,
		 * non-composite values. */
		string_value = axl_node_get_content (child, NULL);
		g_return_val_if_fail (string_value, NULL);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received simple value='%s'", string_value);

		/* for each value tag, convert into XML-RPC parameter */
		if (NODE_CMP_NAME (child, "i4") || NODE_CMP_NAME (child, "int"))
			result        = method_value_new_from_string (XML_RPC_INT_VALUE, string_value);
		
		/* boolean case */
		if (NODE_CMP_NAME (child, "boolean"))
			result        = method_value_new_from_string (XML_RPC_BOOLEAN_VALUE, string_value);
		
		/* string case */
		if (NODE_CMP_NAME (child, "string"))
			result        = method_value_new_from_string (XML_RPC_STRING_VALUE, string_value);
		       
		/* double case */
		if (NODE_CMP_NAME (child, "double"))
			result        = method_value_new_from_string (XML_RPC_DOUBLE_VALUE, string_value);
		
		/* dateTime.iso8601 */
		if (NODE_CMP_NAME (child, "dateTime.iso8601"))
			result        = method_value_new_from_string (XML_RPC_DATE_VALUE, string_value);
		
		/* base64 case */
		if (NODE_CMP_NAME (child, "base64"))
			result        = method_value_new_from_string (XML_RPC_BASE64_VALUE, string_value);
		
		/* free and return */
		return result;
	}
	
	/* struct case */
	if (NODE_CMP_NAME (child, "struct")) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received a complex value <struct>");
		
		/* parse the struct value */
		result = __vortex_xml_rpc_parse_struct_value (child);
		
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "struct parsed: type %d", 
		       method_value_get_type (result));

	}else if (NODE_CMP_NAME (child, "array")) {
		/* array case */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received a complex value <array>");
		
		/* parse the array value */
		result = __vortex_xml_rpc_parse_array_value (child);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "array parsed: type %d", 
			    method_value_get_type (result));
	}
	
	/* release memory no longer used */
	return result;
}


typedef struct _VortexXmlRpcBootData {
	VortexConnection        * connection;
	gchar                   * serverName;
	gchar                   * resourceName;
	VortexXmlRpcBootNotify    process_status;
	gpointer                  user_data;
}VortexXmlRpcBootData;

/** 
 * @internal
 *
 * @brief Support function for \ref vortex_xml_rpc_boot_channel, which
 * actually performs the work.
 * 
 * @param data An \ref VortexXmlRpcBootData node;
 * 
 * @return NULL, new channel created will be notified through
 * <i>process_status</i> handler.
 */
gpointer __vortex_xml_rpc_boot_channel_process (VortexXmlRpcBootData * data)
{
	VortexConnection        * connection      = data->connection;
	gchar                   * serverName      = data->serverName;
	gchar                   * resourceName    = data->resourceName;
	VortexXmlRpcBootNotify    process_status  = data->process_status;
	VortexChannel           * channel;
	gpointer                  user_data       = data->user_data;
	GAsyncQueue             * queue;
	VortexFrame             * reply;
	gchar                   * msg;
	
	/* unref no longer needed data */
	g_free (data);

	/* create the channel piggybacking initial data (the bootmsg
	 * token) */
	queue   = g_async_queue_new ();
	channel = vortex_channel_new_fullv (connection,   /* the connection where the channel will be created */
					    0,            /* let Vortex Library to choose the next channel num ready */
					    serverName,   /* serverName value to request virtual hosting */
					    VORTEX_XML_RPC_PROFILE, /* the BEEP unique URI profile */
					    EncodingNone, /* no profile content encoding used */
					    NULL, NULL,   /* default close channel handling */
					    vortex_channel_queue_reply, queue,   /* set frame receiving handling using the queue */
					    NULL, NULL,   /* perform channel creation in a blocking fashion */
					    "<bootmsg resource='%s' />",
					    (resourceName != NULL) ? resourceName : "/");
	
	/* check for the channel returned. */
	if (channel == NULL) {
		/* unref queue */
		g_async_queue_unref (queue);

		/* unable to create channel, notify user space */
		__vortex_xml_rpc_notify (process_status, NULL, VortexError,
					 "Unable to create a channel under the profile XML-RPC",  user_data);
		return NULL;
	}

	/* the next frame from listener peer (it doesn't matter to get
	 * blocked this way). The following source code will support
	 * not only piggybacking but also the first frame received on
	 * this channel */
	reply = vortex_channel_get_reply (channel, queue);

	/* unref queue */
	g_async_queue_unref (queue);

	if (reply == NULL) {
		/* unable to create channel, notify user space */
		__vortex_xml_rpc_notify (process_status, NULL, VortexError,
					 "Timeout reached while waiting for the first reply for the profile XML-RPC",  user_data);
		return NULL;
	}

	/* now we have a reply check for errors */
	if (vortex_frame_is_error_message (reply, NULL, &msg)) {
		/* notify error while booting XML-RPC channel */
		__vortex_xml_rpc_notify (process_status, channel, VortexError, msg, user_data);

		/* nullify the channel */
		channel = NULL;

		goto finish_boot_xml_rpc_channel;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "checking resource validation, received: %s",
	       vortex_frame_get_payload (reply));

	/* seems that the remote peer have accepted to create the
	 * XML-RPC channel. Because it is not a good idea to exec the
	 * xml engine to parse only one tag against the DTD, we simply
	 * check for the tag to be received and its content-lenght. */
	if (g_strcasecmp (vortex_frame_get_payload (reply), "<bootrpy />") ||
	    vortex_frame_get_payload_size (reply) != 11) {
		
		/* notify error while booting XML-RPC channel */
		__vortex_xml_rpc_notify (process_status, channel, VortexError, 
					 "Received wrong bootmsg reply while starting a XML-RPC channel", user_data);

		/* nullify the channel */
		channel = NULL;

		goto finish_boot_xml_rpc_channel;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "XML-RPC channel accepted");

	/* notify user space level */
	__vortex_xml_rpc_notify (process_status, channel, VortexOk,
				 "Channel XML-RPC ready and waiting, now let's RPC", user_data);
	    
 finish_boot_xml_rpc_channel:

	/* free message and error code if were defined */
	vortex_support_free (2, msg, g_free, reply, vortex_frame_free);

	return channel;
}
#endif

/** 
 * @brief Perform initial boot step to get confirmation from remote
 * server to accept incoming XML-RPC under the given resource.
 *
 * XML-RPC invocation follows the next diagram:
 *
 * \image html xml-rpc-invocation.png "XML-RPC invocation diagram"
 *
 * From the execution of this function, a new channel (\ref
 * VortexChannel) is get, already initialized under the XML-RPC
 * profile. This channel will be used \ref vortex_xml_rpc_invoke.
 * 
 * @param connection The connection where the XML-RPC invocation will
 * be performed, actually, the connection where the XML-RPC channel
 * will be created.
 *
 * @param serverName For those connection that didn't server
 * serverName virtual host mechanism, this value will be used. Keep in
 * mind that previous channels could be already negotiated the
 * serverName identity making this value to be ignored. The serverName
 * value is negotiated for the first channel requesting it, but once
 * done, its value is global to the session, that is the given
 * connection you are using. This is parameter is optional. If a NULL
 * value is provided, the serverName attribute will be not used.
 *
 * @param resourceName The resource name requested to be accepted.
 * This parameter is not optional. If a NULL value is
 * provided, the resource "/" will be used as default value.
 *
 * @param process_status A mandatory notification handler where the
 * XML-RPC initial boot result will be notified.
 *
 * @param user_data User defined data to be passed in to the
 * <i>process_status</i> notify handler.
 */
void     vortex_xml_rpc_boot_channel (VortexConnection        * connection,
				      gchar                   * serverName,
				      gchar                   * resourceName,
				      VortexXmlRpcBootNotify    process_status,
				      gpointer                  user_data)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Trying to boot a channel under the XML-RPC profile over a vortex library without XML-RPC support.");
	return;
#else
	VortexXmlRpcBootData * data;

	/* perform XML-RPC initial boot operations. */
	if (!vortex_connection_is_ok (connection, FALSE)) {
		__vortex_xml_rpc_notify (process_status, NULL, VortexError,
					 "Provided a non-connected connection while requested to perform XML-RPC boot process",
					 user_data);
		return;
	}
	
	/* check for process status function */
	if (process_status == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "notify function is mandatory, unable to perform XML-RPC boot request");
		return;
	}

	/* prepare data for the process */
	data                 = g_new0 (VortexXmlRpcBootData, 1);
	data->connection     = connection;
	data->serverName     = copy_if_not_null (serverName);
	data->resourceName   = copy_if_not_null (resourceName);
	data->process_status = process_status;
	data->user_data      = user_data;

	/* perform rpc boot process */
	vortex_thread_pool_new_task ((GThreadFunc) __vortex_xml_rpc_boot_channel_process,  data);
	return;
#endif	
}

#ifdef ENABLE_XML_RPC_SUPPORT
/** 
 * @internal
 *
 * @brief Support function for \ref vortex_tls_start_negociation_sync
 * function.
 */
void __vortex_xml_rpc_boot_channel_sync_process  (VortexChannel    * booted_channel,
						  VortexStatus       status,
						  gchar            * message,
						  gpointer           user_data)
{
	GAsyncQueue * queue = user_data;

	/* push and unref man! (first status, then status_message and
	 * finally the connection). This must follow this order. */
	g_async_queue_push  (queue, GINT_TO_POINTER (status));
	g_async_queue_push  (queue, message);
	g_async_queue_push  (queue, booted_channel);
	g_async_queue_unref (queue);
	
	return;
}
#endif

/** 
 * @brief Perform a synchronous (blocking) XML-RPC channel boot.
 *
 * This function perform the same operation like \ref
 * vortex_xml_rpc_boot_channel function but in a blocking
 * manner. This function is ideal for batch process and every system
 * that requires a blocking remote method invocation.
 *
 * In the case that an asynchronous method invocation is required,
 * \ref vortex_xml_rpc_boot_channel should be used.
 *
 * The function allows to provide references to an optional
 * <b>status</b> and an optional <b>message</b> variables, so the
 * caller could perform a better error reporting.
 * 
 * @param connection The connection where the XML-RPC invocation will
 * be performed, actually, the connection where the XML-RPC channel
 * will be created.
 *
 * @param serverName For those connection that didn't server
 * serverName virtual host mechanism, this value will be used. Keep in
 * mind that previous channels could be already negotiated the
 * serverName identity making this value to be ignored. The serverName
 * value is negotiated for the first channel requesting it, but once
 * done, its value is global to the session, that is the given
 * connection you are using. This parameter is optional. If a NULL
 * value is provided, the serverName attribute will be not used.
 *
 * @param resourceName The resource name requested to be
 * accepted. This parameter is not optional. If a NULL value is
 * provided, the resource "/" will be used as default value.
 *
 * @param status An optional status reference that will be filled with
 * the process termination status.
 *
 * @param status_message An optional message reference that will be filled
 * with a textual diagnostic, explaining termination status.
 * 
 * @return A channel already initialized with the XML-RPC profile or
 * NULL if it fails.
 */
VortexChannel     * vortex_xml_rpc_boot_channel_sync       (VortexConnection        * connection,
							    gchar                   * serverName,
							    gchar                   * resourceName,
							    VortexStatus            * status,
							    gchar                  ** status_message)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	/* fill status value if defined */
	if (status != NULL)
		(* status)  = VortexError;

	/* fill message value if defined */
	if (status_message != NULL)
		(* status_message) = "Trying to initialize a XML-RPC channel but, current Vortex Library doesn't provide XML-RPC services";

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
	       "Trying to initialize a XML-RPC channel but, current Vortex Library doesn't provide XML-RPC services");
	return NULL;
#else
	GTimeVal        timeout;
	GAsyncQueue   * queue     = NULL;
	VortexChannel * _channel  = NULL;
	VortexStatus  * _status   = NULL;

	/* check environment values */
	g_return_val_if_fail (vortex_connection_is_ok (connection, FALSE), NULL);
	
	/* creates the async queue */
	queue = g_async_queue_new ();
	g_async_queue_ref (queue);

	/* wait for the data to be returned, first prepare timeout
	 * value */
	g_get_current_time (&timeout);
	g_time_val_add (&timeout, vortex_connection_get_timeout ());
	
	/* begin XML-RPC bootstrapping */
	vortex_xml_rpc_boot_channel (connection, serverName, resourceName,
				     __vortex_xml_rpc_boot_channel_sync_process, 
				     queue);

	/* get status */
	_status = g_async_queue_timed_pop (queue, &timeout);
	if (_status == NULL) {
		/* seems timeout have happen while waiting for SASL to
		 * end */
		(* status)         = VortexError;
		(* status_message) = "Timeout have been reached while waiting for XML-RPC boot to finish";
		return NULL;
	}

	/* get status */
	if (status != NULL)
		(* status) = GPOINTER_TO_INT (_status);
	
	/* get message */
	if (status_message != NULL)
		(* status_message) = g_async_queue_pop (queue);
	else
		g_async_queue_pop (queue);

	/* get connection */
	_channel = g_async_queue_pop (queue);

	/* unref the queue */
	g_async_queue_unref (queue);

	/* return received value */
	return _channel;
#endif	
}

#ifdef ENABLE_XML_RPC_SUPPORT
/** 
 * @internal
 *
 * @brief Internal definition to transport parameters from \ref
 * vortex_xml_rpc_invoke to the threaded function that finally do the
 * work.
 */
typedef struct _VortexXmlRpcInvokeData {
	VortexChannel        * channel;
	XmlRpcMethodCall     * invocator;
	XmlRpcInvokeNotify     reply_notify;
	gpointer               user_data;
}VortexXmlRpcInvokeData;


/** 
 * @internal
 * 
 * Internal function used to get faultCode and faultString values,
 * from the XML-RPC response received. The node received points to the
 * <value> node which contains the <struct> structure with the desired
 * values.
 */
gboolean __vortex_xml_rpc_get_fault_values (axlNode * node, 
					    gint * faultCode, 
					    gchar ** faultString,
					    XmlRpcMethodValue ** _value)
{
	XmlRpcMethodValue * value;
	XmlRpcStruct      * _struct;

	/* get the struct */
	value = __vortex_xml_rpc_parse_value (node);

	/* check that we have received an struct */
	if (method_value_get_type (value) != XML_RPC_STRUCT_VALUE) {
		method_value_free (value);
		return FALSE;
	}

	/* get the struct inside */
	_struct = vortex_xml_rpc_method_value_get_as_struct (value);

	/* get the fault code */
	*faultCode   = vortex_xml_rpc_struct_get_member_value_as_int (_struct, "faultCode");
	
	/* get the fault String */
	*faultString = vortex_xml_rpc_struct_get_member_value_as_string (_struct, "faultString");

	/* set the value reference */
	*_value    = value;
	
	return TRUE;
}

/** 
 * @internal
 *
 * @brief Support function for the \ref __vortex_xml_rpc_invoke
 * function which actually perform the work of unmarshalling data
 * received due to a previous request performed, and executing the
 * proper handler into the application space.
 * 
 * @param channel The channel where the XML-RPC reply is being
 * received.
 *
 * @param connection The connection where the channel lives.
 *
 * @param frame The frame containing the XML-RPC method response data.
 *
 * @param user_data A VortexXmlRpcInvokeData containing, among other
 * data, the method to invoke.
 */
void __vortex_xml_rpc_invoke_process_reply (VortexChannel      * channel, 
					    VortexConnection   * connection, 
					    VortexFrame        * frame, 
					    gpointer             __invocator_data)
{
	/* invocation specific data received inside user_data and
	 * provided once the invocation was initiated at
	 * __vortex_xml_rpc_invoke */
	VortexXmlRpcInvokeData * invocator_data  = __invocator_data;
	XmlRpcInvokeNotify       reply_notify    = invocator_data->reply_notify;
	gpointer                 user_data       = invocator_data->user_data;
	XmlRpcMethodValue      * value;
	XmlRpcMethodResponse   * response        = NULL;

	/* variable declaration for the xml stuff */
	axlDoc                 * doc;
	axlNode                * node;
	axlError               * error;

	/* variables for fault errors */
	gchar                  * faultString;
	gint                     faultCode;
	

	/* release memory hold by user_data, which contains a node of
	   type VortexXmlRpcInvokeData */
	g_free (invocator_data);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reply received, processing..");

	/* flag the channel to have its reply processed */
	vortex_channel_flag_reply_processed (channel, TRUE);
	
	/* parse incoming data */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), &error);
	if (!doc) {
		/* drop a log */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to parse document received, error was: %s", axl_error_get (error));
		axl_error_free (error);

		/* notify user space  */
		__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_BAD_REPLY_RECEIVED,
						  -1, "unable to parse reply received due to a XML-RPC invocation, xml parsing in memory has failed.",
						  channel, NULL, user_data);
		/* nothing more to do */
		return;
	}
	
	/* get the root node */
	node = axl_doc_get_root (doc);
	if (! NODE_CMP_NAME (node, "methodResponse")) {

		/* notify user space  */
		__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_BAD_REPLY_RECEIVED,
						  -1, "received a different top level tag where expected a <methodResponse>",
						  channel, NULL, user_data);
		/* nothing more to do */
		axl_doc_free (doc);
		return;
	}

	/* walk into the first child */
	node = axl_node_get_child_nth (node, 0);

	/* check if we have received a fault reply */
	if (NODE_CMP_NAME (node, "fault")) {

		/* we got a fault reply, now point to the <value />  */
		node = axl_node_get_child_nth (node, 0);

		/* get the fault error code and the fault string */
		if (! __vortex_xml_rpc_get_fault_values (node, &faultCode, &faultString, &value)) {
			/* notify the user space */
			__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_BAD_REPLY_RECEIVED,
							  -1, "Unable to decode the fault structure received, the remote peer is sending a not properly formated fault struct",
							  channel, NULL, user_data);
		}else {
			/* notify the user space */
			__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_FAULT_REPLY,
							  faultCode,  faultString,
							  channel, NULL, user_data);
			/* free the struct received */
			method_value_free (value);
		}

		/* free the document */
		axl_doc_free (doc);
		return;
	}

	
	/* we got something else, check if that is a params as a first
	 * top level node */
	if (NODE_CMP_NAME (node, "params")) {
		/* perform here the code to read data received */
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received a positive response..");

		/* we got a fault reply, now point to the <param />  */
		node = axl_node_get_child_nth (node, 0);

		/* now get the <value> node */
		node = axl_node_get_child_nth (node, 0);

		/* get the value */
		value    = __vortex_xml_rpc_parse_value (node);
		if (value != NULL) {

			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "received method value with type: (%d)", 
				    method_value_get_type (value));
			
			/* create the response object */
			response = vortex_xml_rpc_method_response_new (XML_RPC_OK,
								       -1, NULL, /* fault code especification */
								       value);
			/* notify reply received */
			__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_OK,
							  -1, NULL,
							  channel, response, user_data);
		}else {
			/* notify reply received */
			__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_BAD_REPLY_RECEIVED,
							  -1, "received a positive formated method response but, the content inside, is not properly formated",
							  channel, NULL, user_data);
		}

		/* free the document */
		axl_doc_free (doc);
		return;
	}
	
	
	/* it seems that we have received something that is not the
	 * data expected */
	__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_BAD_REPLY_RECEIVED,
					  -1, "received a bad child top level node to <methodResponse>",
					  channel, NULL, user_data);
	/* nothing more to do */
	axl_doc_free (doc);


	return;
}


/** 
 * @internal
 * @brief Perform the xml-rpc invocation in a blocking manner.
 *
 * Because this function is executed from a separated thread, caller
 * is already unblocked. Rather than using asynchronous callbacks to
 * write this function, which would make the code harder to read and
 * maintain, it is preferred to perform the job in a linear manner.
 *
 * This function converts the invocator object (invocator attribute
 * received) into a XML representation following the definition found
 * for the XML-RPC.
 *
 * Then it waits for the data to be received, until a timeout is
 * reached.
 * 
 * @param data The data needed to perform the invocation.
 * 
 * @return This function always return NULL.
 */
gpointer __vortex_xml_rpc_invoke (VortexXmlRpcInvokeData * data)
{
	VortexChannel        * channel       = data->channel;
	XmlRpcMethodCall     * invocator     = data->invocator;
	XmlRpcInvokeNotify     reply_notify  = data->reply_notify;
	gpointer               user_data     = data->user_data;
	gchar                * message;
	gint                   message_size;

	/* marshall the message */
	message = vortex_xml_rpc_method_call_marshall (invocator, &message_size);
	
	/* set the new frame receive handler and its data */
	vortex_channel_set_received_handler (channel, __vortex_xml_rpc_invoke_process_reply, data);
	
	/* perform the invocation */
	if (!vortex_channel_send_msg (channel, message, message_size, NULL)) {
		/* seems the invocation has failed */
		__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_INVOCATION_FAILURE, 
						  -1, "An error have happened while sending the XML-RPC request message.",
						  channel, NULL, user_data);
		/* release data node */
		g_free (data);

		/* update the reference of the data refernce to NULL */
		vortex_channel_set_received_handler (channel, __vortex_xml_rpc_invoke_process_reply, NULL);
	}
	
	/* release data no longer needed */
	if (vortex_xml_rpc_method_call_must_release (invocator))
		vortex_xml_rpc_method_call_free (invocator);
	g_free (message);
	
	/* nothing more to do */
	return NULL;
}

#endif

/** 
 * @brief Perform an asynchronous invocation using the XML-RPC profile. 
 *
 * This function allows to perform an asynchronous invocation on the
 * given channel, that is already running the XML-RPC profile, using
 * as method invocator the <b>method_call</b> provided.
 *
 * The channel received on this function should be get from: 
 *   - \ref vortex_xml_rpc_boot_channel
 *   - \ref vortex_xml_rpc_boot_channel_sync
 *
 * Once the method is invoked, the reply received is notified on the
 * <b>reply_notify</b> callback. 
 *
 * Because the function will perform the invocation in an asynchronous
 * mode, the caller will not be blocked while calling to this
 * function. If a blocking invocation mode is required check \ref
 * vortex_xml_rpc_invoke_sync.
 * 
 * @param channel The channel, running the XML-RPC profile, where the
 * invocation will take place.
 *
 * @param method_call The method call invocator, representing the
 * method to be invoked plus its parameters.
 *
 * @param reply_notify The notify function where the reply will be
 * reported.
 *
 * @param user_data User defined data that will be passed to the
 * notify function along the reply received.
 */
void                vortex_xml_rpc_invoke                  (VortexChannel           * channel,
							    XmlRpcMethodCall        * method_call,
							    XmlRpcInvokeNotify        reply_notify,
							    gpointer                  user_data)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
	       "performing an asynchronous invocation with a Vortex Library without XML-RPC support");
	return;
#else
	VortexXmlRpcInvokeData * data;

	/* perform some environmental checks for incoming data */
	g_return_if_fail (channel);
	g_return_if_fail (method_call);
	g_return_if_fail (reply_notify);

	/* check for the XML-RPC profile to be running */
	if (!vortex_channel_is_running_profile (channel, VORTEX_XML_RPC_PROFILE)) {
		__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_NOT_XML_RPC_CHANNEL,
						  -1, "An invocation has been detected over a channel that is not running the XML-RPC profile.",
						  channel, NULL, user_data);
		return;
	}

	/* check that the channel provided is at the ready state */
	if (vortex_xml_rpc_channel_status (channel) != XmlRpcStateReady) {
		__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_CHANNEL_NOT_READY,
						  -1, "Detected an XML-RPC invocation over a channel that is not in the ready state",
						  channel, NULL, user_data);
		return;
	}

	/* check if the channel is waiting for a previous reply */
	if (! vortex_channel_is_ready (channel)) {
		__vortex_xml_rpc_notify_response (reply_notify, XML_RPC_WAITING_PREVIOUS,
						  -1, "An XML-RPC invocation was detected over a channel that is actually waiting for a previous invocation.",
						  channel, NULL, user_data);
		return;
	}

	/* prepare data to be passed in to the thread function */
	data               = g_new0 (VortexXmlRpcInvokeData, 1);
	data->channel      = channel;
	data->invocator    = method_call;
	data->reply_notify = reply_notify;
	data->user_data    = user_data;

	/* perform the invocation in a non blocking manner */
	vortex_thread_pool_new_task ((GThreadFunc) __vortex_xml_rpc_invoke, data);
	return;
#endif
}

#ifdef ENABLE_XML_RPC_SUPPORT

/** 
 * @internal
 *
 * @brief Internal support function for \ref
 * vortex_xml_rpc_invoke_sync, that queues the method response
 * received.
 *
 * 
 * @param channel The channel where the method response reply was received.
 * @param response The method response received.
 * @param user_data User space data.
 */
void __vortex_xml_rpc_invoke_sync_process (VortexChannel        * channel,
					   XmlRpcMethodResponse * response,
					   gpointer               user_data)
{
	GAsyncQueue * queue = user_data;

	/* push reply received */
	g_async_queue_push (queue, response);

	/* decrease queue reference and return */
	g_async_queue_unref (queue);
	
	return;
}
#endif

/** 
 * @brief Perform a synchronous XML-RPC invocation.
 * 
 * @param channel The channel where the invocation will be performed.
 * @param method_call The invocator object, representing the method call.
 * 
 * @return A \ref XmlRpcMethodResponse with the reply or NULL if
 * fails.
 */
XmlRpcMethodResponse * vortex_xml_rpc_invoke_sync          (VortexChannel           * channel,
							    XmlRpcMethodCall        * method_call)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "performing a synchronous XML-RPC invocation with a Vortex Library without XML-RPC support");
	return NULL;
#else
	/* timeout object */
	GTimeVal               timeout;
	GAsyncQueue          * queue;
	XmlRpcMethodResponse * response;

	/* perform some additional checks */
	g_return_val_if_fail (channel,     NULL);
	g_return_val_if_fail (method_call, NULL);

	/* creates the asynchronous queue */
	queue = g_async_queue_new ();
	g_async_queue_ref (queue);

	/* wait for the data to be returned, first prepare timeout
	 * value */
	g_get_current_time (&timeout);
	g_time_val_add (&timeout, vortex_connection_get_timeout ());

	/* perform invocation */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "synchronous invocation performed..");
	vortex_xml_rpc_invoke (channel, method_call, __vortex_xml_rpc_invoke_sync_process, queue);

	/* wait until data reach */
	response = g_async_queue_timed_pop (queue, &timeout);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "method response received..");

	/* deallocate queue */
	g_async_queue_unref (queue);

	/* it seems a timeout have happened, stub out a response */
	if (response == NULL)
		response = method_response_new (XML_RPC_TIMEOUT_ERROR, -1, "Invocation timeout, timeout have been reached while waiting for method response.", NULL);
	return response;
#endif
}

/** 
 * @brief Allows to get current XML-RPC channel boot status for the
 * given channel.
 *
 * According to the XML-RPC: there are two states in the BEEP profile
 * for XML-RPC, "boot", the profile's initial state, and "ready".
 *
 * On the "boot" state the channel have been successfully created under
 * the semantic of the XML-RPC profile but not it isn't still booted
 * under some resource. 
 *
 * On the "ready" state the channel is ready to perform XML-RPC
 * requests because both peers agree about resource being accessed.
 *
 * 
 * @param channel The channel to get XML-RPC channel status.
 * 
 * @return Current channel status.
 */
VortexXmlRpcState   vortex_xml_rpc_channel_status   (VortexChannel * channel)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "trying to check current XML-RPC channel status on a Vortex Library which doesn't have support for XML-RPC");
	return XmlRpcStateUnknown;
#else
	gchar * boot_state;

	/* check some environment conditions  */
	if ((channel == NULL) ||
	    !vortex_channel_is_running_profile (channel, VORTEX_XML_RPC_PROFILE)) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "channel provided is null or it is not running the XML-RPC profile, unable to return channel status");
		return XmlRpcStateUnknown;
	}

	boot_state = vortex_channel_get_data (channel, XML_RPC_BOOT_STATE);
	if (!g_strcasecmp (boot_state, "boot"))
		return XmlRpcStateBoot;
	if (!g_strcasecmp (boot_state, "ready"))
		return XmlRpcStateReady;
	
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to get current XML-RPC channel status, returning channel state: unknown");
	return XmlRpcStateUnknown;
#endif
}

#ifdef ENABLE_XML_RPC_SUPPORT

/** 
 * @internal
 *
 * @brief Support function which perform the unmarshalling from an xml
 * representation into the XmlRpcMethodCall representation.
 * 
 * @param frame The frame containing a xml rpc method call object.
 * 
 * @return A newly allocated XmlRpcMethodCall or NULL if fails.
 */
XmlRpcMethodCall * __vortex_xml_rpc_frame_received_parse_method_call (VortexFrame * frame)
{
	axlDoc            * doc;
	axlNode           * node;
	axlNode           * node_aux;

	gchar             * method_name;
	gint                param_count;
	gint                iterator;

	XmlRpcMethodCall  * method_call;
	XmlRpcMethodValue * value;

	/* parse incoming xml data */
	doc = axl_doc_parse (vortex_frame_get_payload (frame), 
			     vortex_frame_get_payload_size (frame), NULL);
	if (!doc) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "unable to parse incoming xml data");
		return NULL;
	}
	
	/* get method name */
	method_name = axl_doc_get_content_at (doc, "/methodCall/methodName", NULL);
	if (method_name == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "unable to get method name value");

		/* free the document */
		axl_doc_free (doc);
		return NULL;
	}

	/* get a reference to the <params> node */
	node        = axl_doc_get (doc, "/methodCall/params");

	/* get how many childs have the <params> tag */
	if (node != NULL)
		param_count = axl_node_get_child_num (node);
	else
		param_count = 0;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "invocation detected for: %s, with %d parameters", method_name, param_count);

	/* build the method call */
	method_call = method_call_new (method_name, param_count);

	/* now get all parameters */
	iterator = 0;
	while ((param_count > 0) && (iterator < axl_node_get_child_num (node))) {
		
		/* get a reference a reference to the <param> tag */
		node_aux = axl_node_get_child_nth (node, iterator);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "node name=<%s>", 
		       axl_node_get_name (node_aux));
		
		/* get a referece to the <value> tag inside */
		node_aux = axl_node_get_child_nth (node_aux, 0);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "node name(2)=<%s>", 
		       axl_node_get_name (node_aux));

		/* parse param value and add it to the method call */
		value = __vortex_xml_rpc_parse_value (node_aux);
		
		/* add the value to the method call */
		method_call_add_value (method_call, value);
		
		/* get a reference to the next element */
		iterator++;
	}

	/* free xml document */
	axl_doc_free (doc);

	/* return method call received */
	return method_call;
}

/** 
 * @internal
 * @brief Frame received handler for XML-RPC channels.
 */
void __vortex_xml_rpc_frame_received (VortexChannel    * channel, 
				      VortexConnection * connection, 
				      VortexFrame      * frame, 
				      gpointer           user_data)
{
	XmlRpcMethodCall     * method_call;
	XmlRpcMethodResponse * method_response = NULL;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "frame received on XML-RPC channel");

	/* get method call received */
	method_call = __vortex_xml_rpc_frame_received_parse_method_call (frame);
	if (method_call == NULL) {
		/* return without doing nothing */
		return;
	}

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "method call name received: %s (params %d), dispatching...",
	       method_call_get_name (method_call), method_call_get_num_params (method_call));

	/* set method call reply data */
	__vortex_xml_rpc_method_call_set_reply_data (method_call, channel, 
						     vortex_frame_get_msgno (frame));

	/* first check for an specific dispatch handler for the given
	 * method call */
	
	/* now see for the general dispatch function is available if
	 * the previous one was not found */
	if (__vortex_xml_rpc_service_dispatch != NULL) {
		
		/* invoke and get the method response */
		method_response = __vortex_xml_rpc_service_dispatch (channel, 
								     method_call, 
								     __vortex_xml_rpc_service_data);
	}

	/* if method response replied is null, the method invocation
	 * reply has been deferred, just return */
	if (method_response == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "XML-RPC reply deferred, leaving frame received handler");
		return;
	}
	
	/* perform reply */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "XML-RPC reply performed, sending it back to the client");
	vortex_xml_rpc_notify_reply (method_call, method_response);
	return;
}

/** 
 * @internal
 *
 * @brief Internal vortex function to parse <bootmsg/> entity, getting
 * resource value inside it. 
 *
 * The function also checks for the message receivced, generating the
 * proper message to be returned in case of error.
 * 
 * @param profile_content The profile content.
 *
 * @param resource A pointer where the resource value will be stored.
 *
 * @param profile_content_reply A pointer where a possible error reply
 * will be stored.
 * 
 * @return TRUE if the message was successfully parsed, or FALSE if
 * some error have happened. In that case, profile_content_reply is
 * filled.
 */
gboolean __vortex_xml_rpc_parse_bootmsg (gchar *  profile_content, 
					 gchar ** resource, 
					 gchar ** profile_content_reply)
{
	axlDtd    * xml_rpc_dtd;
	axlDoc    * doc;
	axlNode   * bootmsg;

	/* get channel management DTD */
	if ((xml_rpc_dtd = vortex_dtds_get_xml_rpc_boot_dtd ()) == NULL) {
		(* profile_content_reply ) = 
			vortex_frame_get_error_message ("451",
							"unable to load dtd file (xml-rpc.dtd), cannot validate incoming message, returning error frame",
							NULL);
		return FALSE;
	}

	/* parser xml document */
	doc =  axl_doc_parse (profile_content, strlen (profile_content), NULL);
	if (!doc) {
		(* profile_content_reply ) = 
			vortex_frame_get_error_message ("501",
							"Invalid xml received while creating a XML-RPC channel.",
							NULL);
		return FALSE;
	}

	/* validate xml received */
	if (! axl_dtd_validate (doc, xml_rpc_dtd, NULL)) {
		/* free the document */
		(* profile_content_reply ) = 
			vortex_frame_get_error_message ("501",
							"xml validation failed while creating a XML-RPC channel.",
							NULL);
		/* free the document */
		axl_doc_free (doc);
		return FALSE;
	}

	/* get a reference to document root (the bootmsg) */
	bootmsg  = axl_doc_get_root (doc);
	if (! NODE_CMP_NAME (bootmsg,  "bootmsg")) {
		/* report the reply */
		(* profile_content_reply ) = 
			vortex_frame_get_error_message ("501",
							"unable to find as document root message the <bootmsg/> entity",
							NULL);
		/* free the document loaded */
		axl_doc_free (doc);
		return FALSE;
	}

	/* get the resource value */
	(* resource ) = axl_node_get_attribute_value_copy (bootmsg, "resource");
	if ((* resource ) == NULL) {
		(* profile_content_reply ) = 
			vortex_frame_get_error_message ("501",
							"resource not especified inside the <bootmsg/> received",
							NULL);
		/* free the document loaded */
		axl_doc_free (doc);
		return FALSE;
	}

	/* free the document loaded */
	axl_doc_free (doc);

	/* return to the upper level function the resource read */
	return TRUE;
}

/** 
 * @internal
 * @brief Start handler for all incoming request wanting to create an
 * XML-RPC channel.
 * 
 * @param profile The profile received, in this case the XML-RPC profile.
 * @param channel_num The channel number received.
 * @param connection The connection where the XML-RPC was received.
 * @param serverName The serverName value to initiate the virtual hosting
 *
 * @param profile_content Profile content, in this case, it could have
 * several probabilities to be the boot resouce message.
 *
 * @param profile_content_reply This variable should contain the value
 * to be replied. (in this case it should be filled with the bootrpy).
 *
 * @param encoding The profile content enconding.
 *
 * @param user_data Application defined data. In this case NULL.
 * 
 * @return The function should return TRUE if the resource requested
 * is accepted or not accepted but so the channel. FALSE should be
 * returned only if the channel and the resource are denied.
 */
gboolean __vortex_xml_rpc_start_msg (gchar             * profile,
				     gint                channel_num,
				     VortexConnection  * connection,
				     gchar             * serverName,
				     gchar             * profile_content,
				     gchar            ** profile_content_reply,
				     VortexEncoding      encoding,
				     gpointer            user_data)
{
	gchar         * resource;
	gboolean        accept = FALSE;
	VortexChannel * channel;

	if (!__vortex_xml_rpc_parse_bootmsg (profile_content, &resource, profile_content_reply))
		return FALSE;

	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "start message received on XML-RPC channel resource=%s",
	       resource);

	/* invoke resource validation */
	if (__vortex_xml_rpc_validate_resource == NULL) {
		accept = TRUE;
	}else {
		/* invoke application level resource validation */
		accept = __vortex_xml_rpc_validate_resource (connection, channel_num, 
							     serverName, resource);
	}
	
	/* reply with a bootrpy or an error message, accepting the channel to be created */
	if (accept) 
		(* profile_content_reply) = g_strdup ("<bootrpy />");
	else
		(* profile_content_reply) = vortex_frame_get_error_message ("550", 
									    "resource requested not accepted by this XML-RPC instance",
									    NULL);

	/* set the resource booted to allow the channel to be classified */
	channel = vortex_connection_get_channel (connection, channel_num);
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "configure channel=%d resource=%s", 
	       channel_num, resource);
	vortex_channel_set_data_full (channel, XML_RPC_RESOURCE, resource, NULL, g_free);

	return TRUE;
}
#endif

/** 
 * @brief Allows to get the channel resource used to boot the provided
 * channel.
 *
 * @param channel The channel to get the resource inside.
 * 
 * @return The resource value or NULL if it fails. The value returned
 * is an internal reference that mustn't be deallocated. To get a
 * properly result, the channel must be already booted (\ref XmlRpcStateReady).
 */
gchar * vortex_xml_rpc_channel_get_resource (VortexChannel * channel)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	return NULL;
#else
	/* get the channel resource */
	return vortex_channel_get_data (channel, XML_RPC_RESOURCE);
#endif
}

/** 
 * @brief Allows to notify a reply that has been generated from the
 * given \ref XmlRpcMethodCall object.
 *
 * Service invocation could be deferred by returning a NULL value at
 * the service dispatch function. Once a reply is generated, it could
 * be send by using this function.
 *
 * This function not only generates the reply but also deallocates the
 * memory reserved by the two incoming parameters.
 *
 * @param method_call The method call that is being replied.
 *
 * @param method_response The method response representing a reply.
 */
void vortex_xml_rpc_notify_reply (XmlRpcMethodCall     * method_call, 
				  XmlRpcMethodResponse * method_response)
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, 
	       "calling to rpc notify reply function inside a Vortex Library without XML-RPC support");
	return;
#else
	VortexChannel * channel      = NULL;
	gint            msg_no       = -1;
	gchar         * reply_string = NULL;
	gint            reply_size   = -1;

	/* check that the reply notification facility is received
	 * right data */
	g_return_if_fail (method_call);
	g_return_if_fail (method_response);
	
	/* get reply data from the given method call */
	__vortex_xml_rpc_method_call_get_reply_data (method_call, &channel, &msg_no);

	/* get the xml representation for the method reply */
	reply_string = vortex_xml_rpc_method_response_marshall (method_response, &reply_size);

	/* check if a marshalled reply was received */
	if (reply_string != NULL && reply_size > 0) {
		/* perform the reply */
		if (! vortex_channel_send_rpy (channel, reply_string, reply_size, msg_no)) 
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "the XML-RPC reply was not possible to be performed");
	}else {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "failed to perform method reply, response message marshalling have failed..");
	}
		
	/* release memory used */
	vortex_support_free (3, 
			     reply_string, g_free,
			     method_call, vortex_xml_rpc_method_call_free,
			     method_response, vortex_xml_rpc_method_response_free);
	
	/* nothing more to do man! */
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "reply performed..");
	return;
#endif	
}

/** 
 * @brief Allow to start receiving incoming XML-RPC request, setting
 * two handlers to validate and process them.
 *
 * This function is provided to allow Vortex BEEP listeners to accept
 * incoming XML-RPC invocations, by providing a resource validation
 * handler and the service dispatch handler.
 *
 * The first handler, the validation resource, is required to notify
 * remote peers that we accept, and support a particular resource. See
 * the \ref VortexXmlRpcValidateResource "validation resource handler"
 * to know more about this. This handler is not required. If NULL is
 * provided, all validation request will be accepted.
 *
 * Once the resource validation process is accepted, the XML-RPC
 * channel is created, and XML-RPC incoming request can start.
 *
 * The second handler is provided to allow the user space to dispath
 * service invocation received. This handler is not optional. Without
 * it, the XML-RPC invocation can't be produced.
 * 
 * The function also register the XML-RPC profile to be advertised on
 * the greetings phase.
 *
 * Here is an example of a listener initializing the vortex engine to
 * accept incoming XML-RPC requests:
 * \code
 * gboolean validate_resource (VortexConnection * connection, 
 *                             gint               channel_number,
 *                             gchar            * serverName,
 *                             gchar            * resource_path)
 * {
 *
 *	g_print ("Resource to validate: '%s' (serverName='%s')\n", 
 *		 resource_path,
 *		 serverName != NULL ? 
 *               serverName : "no serverName value received");
 *	return TRUE;
 * }
 *
 * gint __sum_2_int_int (gint a, gint b, 
 *                       gchar ** fault_error, gint * fault_code)
 * {
 *	if (a == 2 && b == 7) {
 *		// error reply example
 *		REPLY_FAULT ("Unable to sum the 2 and 7 values", -1, 0);
 *	}
 *	// return reply 
 *	return a + b;
 * }
 *
 * XmlRpcMethodResponse * sum_2_int_int (XmlRpcMethodCall * method_call)
 * {
 *	// get method values 
 *	gint                result;
 *	gchar             * fault_error = NULL;
 *	gint                fault_code  = -1;
 *
 *	// get parameters 
 *	gint a = method_call_get_param_value_as_int (method_call, 0);
 *	gint b = method_call_get_param_value_as_int (method_call, 1);
 *
 *	// perform invocation 
 *	result = __sum_2_int_int (a, b, &fault_error, &fault_code);
 *	
 *	// check error reply looking at the fault_error 
 *	if (fault_error != NULL) {
 *		// we have a error reply 
 *		return CREATE_FAULT_REPLY (fault_code, fault_error);
 *	}
 *
 *	// return reply generated 
 *	return CREATE_OK_REPLY (XML_RPC_INT_VALUE, GINT_TO_POINTER (result));
 * }
 *
 * XmlRpcMethodResponse *  service_dispatch (VortexChannel    * channel, 
 *                                           XmlRpcMethodCall * method_call, 
 *                                           gpointer           user_data) 
 * {
 *	
 *	g_print ("*** channel %d, method call received %s (params %d)***\n",
 *		 vortex_channel_get_number (channel),
 *		 method_call_get_name (method_call), 
 *		 method_call_get_num_params (method_call));
 *
 *	// check if the incoming method call is called sum, and has
 *	// two arguments 
 *	if (method_call_is (method_call, "sum", 2, 
 *                          XML_RPC_INT_VALUE, XML_RPC_INT_VALUE, -1)) {
 *		g_print ("*** sum method found... ***\n");
 *		return sum_2_int_int (method_call);
 *		
 *	}
 *	       
 *	// return that the method to be invoked, is not supported 
 *	return CREATE_FAULT_REPLY (-1, "Method call it not supported by this server");
 * }
 *
 * gint main (gint argc, gchar ** argv) 
 * {
 *
 *	// init vortex library 
 *	vortex_init ();
 *
 *	// enable XML-RPC profile 
 *	vortex_xml_rpc_accept_negociation (validate_resource,
 *					   // no user space data for
 *					   // the validation resource
 *					   // function. 
 *					   NULL, 
 *					   service_dispatch,
 *					   // no user space data for
 *					   // the dispatch function. 
 *					   NULL);
 *
 *	// create a vortex server 
 *	vortex_listener_new ("0.0.0.0", "44000", NULL, NULL);
 *
 *	// wait for listeners (until vortex_exit is called) 
 *	vortex_listener_wait ();
 *	
 *	// end vortex function 
 *	vortex_exit ();
 *
 *	return 0;
 * }
 * \endcode
 *
 * @param validate_resource The resource validation handler. This
 * handler is not optional.
 *
 * @param validate_user_data Optional user data to be passed in to the
 * validation method.
 *
 * @param service_dispatch The service dispatch handler. This handler
 * is not optional.
 *
 * @param dispatch_user_data Optional user data to be passed in to the
 * dispatch method.
 */
void                vortex_xml_rpc_accept_negociation      (VortexXmlRpcValidateResource validate_resource,
							    gpointer                     validate_user_data,
							    VortexXmlRpcServiceDispatch  service_dispatch,
							    gpointer                     dispatch_user_data)
							    
{
#ifndef ENABLE_XML_RPC_SUPPORT
	vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Trying to activate incoming requests for XML-RPC for a Vortex Library without support for XML-RPC");
	return;
#else
	/* validate incoming values */
	if (service_dispatch == NULL) {
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "passed in a null value for the service dispatcher");
		return;
	}

	/* store both handlers */
	__vortex_xml_rpc_validate_resource = validate_resource;
	__vortex_xml_rpc_validate_data     = validate_user_data;
	__vortex_xml_rpc_service_dispatch  = service_dispatch;
	__vortex_xml_rpc_service_data      = dispatch_user_data;

	/* enable XML-RPC profile */
	vortex_profiles_register (VORTEX_XML_RPC_PROFILE,
				  /* the start handler */
				  NULL, NULL, 
				  /* close handler */
				  NULL, NULL, 
				  /* frame received handler */
				  __vortex_xml_rpc_frame_received, NULL);
	
	/* now set the extended start message handler to process all
	 * incoming start messages received */
	vortex_profiles_register_extended_start (VORTEX_XML_RPC_PROFILE,
						 __vortex_xml_rpc_start_msg, NULL);

	return;
#endif	
}

/** 
 * @brief Support function for xml-rpc listeners created by the
 * xml-rpc-gen tool, that reads a xml file that contains listener
 * information and starts the listener.
 * 
 * See also \ref vortex_listener_parse_conf_and_start.
 * 
 * @return TRUE if the listener was started because the file was read
 * successfully otherwise FALSE is returned.
 */
gboolean            vortex_xml_rpc_listener_parse_conf_and_start_listeners ()
{
	/* call to legacy implementation */
	return vortex_listener_parse_conf_and_start ();
}

/** 
 * @brief (Un)Marshaller function used by the xml-rpc-gen tool, at the
 * client C stub generated used as \ref XmlRpcInvokeNotify function.
 * 
 * The function expect to receive a int reply, inside an \ref
 * XmlRpcMethodResponse, unmarshall it into a integer value and the
 * call to the process function, defined by \ref XmlRpcProcessInt
 * handler, that is received as an argument at <b>user_data</b>
 * parameter.
 *
 * @param channel The channel where the reply was received.
 *
 * @param response The method response received (\ref
 * XmlRpcMethodResponse).
 *
 * @param user_data The process reply, an user space callback, that
 * must fit the definition for \ref XmlRpcProcessInt.
 */
void vortex_xml_rpc_unmarshall_int    (VortexChannel *channel, 
				       XmlRpcMethodResponse *response, 
				       gpointer user_data)
{
	XmlRpcProcessInt      process = user_data;
	XmlRpcMethodValue   * value;
	XmlRpcResponseStatus  status  = method_response_get_status (response);
	int                   result;
	
	/* positive reply */
	if (status == XML_RPC_OK) {
		/* get the value inside */
		value  = method_response_get_value (response);
		result = method_value_get_as_int (value);
		
		/* notify reply received */
		process (result, status, -1, NULL);
	}else {
		/* notify error reply */
		process (-1, status, 
			 /* fault code */
			 method_response_get_fault_code (response),
			 /* fault string */
			 method_response_get_fault_string (response));
	}

	/* release memory used by the XmlRpcResponse */
	method_response_free (response);

	/* nothing more to do */
	return;
}


/** 
 * @internal
 * 
 * @param response The method response to get the value inside.
 *
 * @param status A reference to the method response status to fill
 * with the status received.
 *
 * @param fault_code A fault code reference to fill with the fault
 * code received or -1 if no error was detected.
 *
 * @param fault_string A fault string reference to fill with the fault
 * string received or NULL if no error was detected.
 * 
 * @return The method value inside or NULL if it fails. In that case,
 * variables received will be filled with error values.
 */
XmlRpcMethodValue * __vortex_xml_rpc_unmarshall_common_sync (XmlRpcMethodResponse  * response,
							     XmlRpcResponseStatus  * status, 
							     gint                  * fault_code, 
							     gchar                ** fault_string)
{
	XmlRpcResponseStatus _status;
	
	/* fill the status received */
	_status = method_response_get_status (response);

	/* report status */
	if (status != NULL)
		*status = _status;

	if (_status == XML_RPC_OK) {
		/* report fault code */
		if (fault_code != NULL)
			*fault_code   = -1;
		
		/* report fault string */
		if (fault_string != NULL)
			*fault_string = "No error reported";
		return method_response_get_value (response);
	}

	/* report fault code */
	if (fault_code != NULL)
		*fault_code   = method_response_get_fault_code (response);

	/* report fault string */
	if (fault_string != NULL)
		*fault_string = method_response_get_fault_string (response);

	/* no method value */
	return NULL;
}

/** 
 * @brief Gets the integer value from the \ref XmlRpcMethodResponse,
 * setting the fault code and the fault string, releasing the \ref
 * XmlRpcMethodResponse passed in.
 *
 * @param response The XmlRpcMethodResponse to unmarshall.
 * 
 * @param status A reference to a XmlRpcResponseStatus to return value
 * stored in the method response.
 *
 * @param fault_code The fault code reference to fill up the reference
 * if an error is found.
 *
 * @param fault_string The fault string reference to fill up if an
 * error is found.
 * 
 * @return Returns the value store inside the method response (\ref
 * XmlRpcMethodResponse).
 */
gint vortex_xml_rpc_unmarshall_int_sync (XmlRpcMethodResponse  * response, 
					 XmlRpcResponseStatus  * status, 
					 gint                  * fault_code,
					 gchar                ** fault_string)
{
	XmlRpcMethodValue * value;
	gint                result = -1;
	
	/* get the value inside and status/fault_code/fault_string */
	value = __vortex_xml_rpc_unmarshall_common_sync (response, status, fault_code, fault_string);
	
	/* if the value returned is not NULL, get the integer value */
	if (value != NULL)
		result = method_value_get_as_int (value);

	/* free response received */
	method_response_free (response);

	/* return result */
	return result;
}


/** 
 * @brief (Un)Marshaller function used by the xml-rpc-gen tool, at the
 * client C stub generated used as \ref XmlRpcInvokeNotify function.
 * 
 * The function expect to receive a string reply, inside an \ref
 * XmlRpcMethodResponse, unmarshall it into a string value and the
 * call to the process function, defined by \ref XmlRpcProcessString
 * handler, that is received as an argument at <b>user_data</b>
 * parameter.
 *
 * @param channel The channel where the reply was received.
 *
 * @param response The method response received (\ref
 * XmlRpcMethodResponse).
 *
 * @param user_data The process reply, an user space callback, that
 * must fit the definition for \ref XmlRpcProcessInt.
 */
void vortex_xml_rpc_unmarshall_string (VortexChannel *channel, 
				       XmlRpcMethodResponse *response, 
				       gpointer user_data)
{
	XmlRpcProcessString   process = user_data;
	XmlRpcMethodValue   * value;
	XmlRpcResponseStatus  status  = method_response_get_status (response);
	gchar               * result;
	
	/* positive reply */
	if (status == XML_RPC_OK) {
		/* get the value inside */
		value  = method_response_get_value (response);
		result = method_value_get_as_string (value);
		
		/* notify reply received */
		process (result, status, -1, NULL);
	}else {
		/* notify error reply */
		process (NULL, status, 
			 /* fault code */
			 method_response_get_fault_code (response),
			 /* fault string */
			 method_response_get_fault_string (response));
	}

	/* release memory used by the XmlRpcResponse */
	method_response_free (response);

	/* nothing more to do */
	return;
}

/** 
 * @brief Gets the string value from the \ref XmlRpcMethodResponse,
 * setting the fault code and the fault string, releasing the \ref
 * XmlRpcMethodResponse passed in.
 *
 * 
 * @param response The XmlRpcMethodResponse to unmarshall.
 * 
 * @param status A reference to a XmlRpcResponseStatus to return value
 * stored in the method response.
 *
 * @param fault_code The fault code reference to fill up the reference
 * if an error is found.
 *
 * @param fault_string The fault string reference to fill up if an
 * error is found.
 * 
 * @return Returns the value store inside the method response (\ref
 * XmlRpcMethodResponse).
 */
gchar * vortex_xml_rpc_unmarshall_string_sync (XmlRpcMethodResponse  * response, 
					    XmlRpcResponseStatus  * status, 
					    gint                  * fault_code,
					    gchar                ** fault_string)
{
	XmlRpcMethodValue * value;
	gchar             * result = NULL;
	
	/* get the value inside and status/fault_code/fault_string */
	value = __vortex_xml_rpc_unmarshall_common_sync (response, status, fault_code, fault_string);
	
	/* if the value returned is not NULL, get the integer value */
	if (value != NULL)
		result = method_value_get_as_string (value);

	
	/* nullify the value inside so the response free function
	 * doesn't free the string returned */
	vortex_xml_rpc_method_value_nullify (value);

	/* free response received */
	method_response_free (response);

	/* return result */
	return result;
}

/** 
 * @brief (Un)Marshaller function used by the xml-rpc-gen tool, at the
 * client C stub generated used as \ref XmlRpcInvokeNotify function.
 * 
 * The function expect to receive a double reply, inside an \ref
 * XmlRpcMethodResponse, unmarshall it into a double value and the
 * call to the process function, defined by \ref XmlRpcProcessDouble
 * handler, that is received as an argument at <b>user_data</b>
 * parameter.
 *
 * @param channel The channel where the reply was received.
 *
 * @param response The method response received (\ref
 * XmlRpcMethodResponse).
 *
 * @param user_data The process reply, an user space callback, that
 * must fit the definition for \ref XmlRpcProcessDouble.
 */
void vortex_xml_rpc_unmarshall_double (VortexChannel *channel, 
				       XmlRpcMethodResponse *response, 
				       gpointer user_data)
{
	XmlRpcProcessDouble   process = user_data;
	XmlRpcMethodValue   * value;
	XmlRpcResponseStatus  status  = method_response_get_status (response);
	double                result;
	
	/* positive reply */
	if (status == XML_RPC_OK) {
		/* get the value inside */
		value  = method_response_get_value (response);
		result = method_value_get_as_double (value);
		
		/* notify reply received */
		process (result, status, -1, NULL);
	}else {
		/* notify error reply */
		process (0, status, 
			 /* fault code */
			 method_response_get_fault_code (response),
			 /* fault string */
			 method_response_get_fault_string (response));
	}

	/* release memory used by the XmlRpcResponse */
	method_response_free (response);

	/* nothing more to do */
	return;	
}

/** 
 * @brief Gets the double value from the \ref XmlRpcMethodResponse,
 * setting the fault code and the fault string, releasing the \ref
 * XmlRpcMethodResponse passed in.
 *
 * 
 * @param response The XmlRpcMethodResponse to unmarshall.
 * 
 * @param status A reference to a XmlRpcResponseStatus to return value
 * stored in the method response.
 *
 * @param fault_code The fault code reference to fill up the reference
 * if an error is found.
 *
 * @param fault_string The fault string reference to fill up if an
 * error is found.
 * 
 * @return Returns the value store inside the method response (\ref
 * XmlRpcMethodResponse).
 */
gdouble vortex_xml_rpc_unmarshall_double_sync (XmlRpcMethodResponse  * response, 
					       XmlRpcResponseStatus  * status, 
					       gint                  * fault_code,
					       gchar                ** fault_string)
{
	XmlRpcMethodValue * value;
	gdouble             result = 0;
	
	/* get the value inside and status/fault_code/fault_string */
	value = __vortex_xml_rpc_unmarshall_common_sync (response, status, fault_code, fault_string);
	
	/* if the value returned is not NULL, get the integer value */
	if (value != NULL)
		result = method_value_get_as_double (value);

	/* free response received */
	method_response_free (response);

	/* return result */
	return result;
}


/** 
 * @brief (Un)Marshaller function used by the xml-rpc-gen tool, at the
 * client C stub generated used as \ref XmlRpcInvokeNotify function.
 * 
 * The function expect to receive a \ref XmlRpcStruct reply, inside an
 * \ref XmlRpcMethodResponse, unmarshall it into a \ref XmlRpcStruct
 * value and the call to the process function, defined by \ref
 * XmlRpcProcessStruct handler, that is received as an argument at
 * <b>user_data</b> parameter.
 *
 * @param channel The channel where the reply was received.
 *
 * @param response The method response received (\ref
 * XmlRpcMethodResponse).
 *
 * @param user_data The process reply, an user space callback, that
 * must fit the definition for \ref XmlRpcProcessStruct.
 */
void vortex_xml_rpc_unmarshall_struct (VortexChannel *channel, 
				       XmlRpcMethodResponse *response, 
				       gpointer user_data)
{
	XmlRpcProcessStruct   process = user_data;
	XmlRpcMethodValue   * value;
	XmlRpcResponseStatus  status  = method_response_get_status (response);
	XmlRpcStruct        * result;
	
	/* positive reply */
	if (status == XML_RPC_OK) {
		/* get the value inside */
		value  = method_response_get_value (response);
		result = method_value_get_as_struct (value);
		
		/* notify reply received */
		process (result, status, -1, NULL);
	}else {
		/* notify error reply */
		process (NULL, status, 
			 /* fault code */
			 method_response_get_fault_code (response),
			 /* fault string */
			 method_response_get_fault_string (response));
	}

	/* release memory used by the XmlRpcResponse */
	method_response_free (response);

	/* nothing more to do */
	return;
}

/** 
 * @brief Gets the struct value from the \ref XmlRpcMethodResponse,
 * setting the fault code and the fault string, releasing the \ref
 * XmlRpcMethodResponse passed in.
 *
 * The function will use the provided unmarshaller function to
 * generate a higher level structure result. This handler is
 * optional. In the case it is not provided, a reference to a \ref
 * XmlRpcStruct will be returned. 
 *
 * This unmarshaller function is mainly used by the xml-rpc-gen tool
 * to create a C client stub that is more suitable for the common use
 * (rather interfacing with \ref XmlRpcStruct).
 *
 * 
 * @param response The XmlRpcMethodResponse to unmarshall.
 *
 * @param unmarshaller User space, higher level unmarshaller function.
 * 
 * @param status A reference to a XmlRpcResponseStatus to return value
 * stored in the method response.
 *
 * @param fault_code The fault code reference to fill up the reference
 * if an error is found.
 *
 * @param fault_string The fault string reference to fill up if an
 * error is found.
 * 
 * @return Returns the value store inside the method response (\ref
 * XmlRpcMethodResponse).
 */
gpointer vortex_xml_rpc_unmarshall_struct_sync (XmlRpcMethodResponse      * response, 
						XmlRpcStructUnMarshaller    unmarshaller,
						XmlRpcResponseStatus      * status, 
						gint                      * fault_code, 
						gchar                    ** fault_string)
{
	XmlRpcMethodValue * value;
	XmlRpcStruct      * _struct = NULL;
	gpointer            result;
	
	/* get the value inside and status/fault_code/fault_string */
	value = __vortex_xml_rpc_unmarshall_common_sync (response, status, fault_code, fault_string);

	/* if the value returned is not NULL, get the integer value */
	if (value == NULL) {
		/* free response received, but after calling value */
		method_response_free (response);

		/* nothing to do, just return NULL */
		return NULL;
	}
	
	/* get the struct inside */
	_struct = method_value_get_as_struct (value);

	/* check if the unmarshaller function is provided */
	if (unmarshaller != NULL) {
		/* unmarshall */
		result = unmarshaller (_struct, FALSE);
	} else {
		/* return the structure */
		result = _struct;
	}

	/* free response received, but after calling value */
	method_response_free (response);

	/* return result */
	return result;
}

/** 
 * @brief (Un)Marshaller function used by the xml-rpc-gen tool, at the
 * client C stub generated used as \ref XmlRpcInvokeNotify function.
 * 
 * The function expect to receive a \ref XmlRpcArray reply, inside an
 * \ref XmlRpcMethodResponse, unmarshall it into a \ref XmlRpcArray
 * value and the call to the process function, defined by \ref
 * XmlRpcProcessArray handler, that is received as an argument at
 * <b>user_data</b> parameter.
 *
 * @param channel The channel where the reply was received.
 *
 * @param response The method response received (\ref
 * XmlRpcMethodResponse).
 *
 * @param user_data The process reply, an user space callback, that
 * must fit the definition for \ref XmlRpcProcessArray.
 */
void vortex_xml_rpc_unmarshall_array (VortexChannel *channel, 
				      XmlRpcMethodResponse *response, 
				      gpointer user_data)
{
	XmlRpcProcessArray    process = user_data;
	XmlRpcMethodValue   * value;
	XmlRpcResponseStatus  status  = method_response_get_status (response);
	XmlRpcArray         * result;
	
	/* positive reply */
	if (status == XML_RPC_OK) {
		/* get the value inside */
		value  = method_response_get_value (response);
		result = method_value_get_as_array (value);
		
		/* notify reply received */
		process (result, status, -1, NULL);
	}else {
		/* notify error reply */
		process (NULL, status, 
			 /* fault code */
			 method_response_get_fault_code (response),
			 /* fault string */
			 method_response_get_fault_string (response));
	}

	/* release memory used by the XmlRpcResponse */
	method_response_free (response);

	/* nothing more to do */
	return;	
}

/** 
 * @brief Gets the array value from the \ref XmlRpcMethodResponse,
 * setting the fault code and the fault string, releasing the \ref
 * XmlRpcMethodResponse passed in.

 * The function will use the provided unmarshaller function to
 * generate a higher level structure result. This handler is
 * optional. In the case it is not provided, a reference to a \ref
 * XmlRpcArray will be returned. 
 *
 * This unmarshaller function is mainly used by the xml-rpc-gen tool
 * to create a C client stub that is more suitable for the common use
 * (rather interfacing with \ref XmlRpcStruct).
 *
 * 
 * @param response The XmlRpcMethodResponse to unmarshall.
 *
 * @param unmarshaller User space, higher level unmarshaller function.
 * 
 * @param status A reference to a XmlRpcResponseStatus to return value
 * stored in the method response.
 *
 * @param fault_code The fault code reference to fill up the reference
 * if an error is found.
 *
 * @param fault_string The fault string reference to fill up if an
 * error is found.
 * 
 * @return Returns the value store inside the method response (\ref
 * XmlRpcMethodResponse).
 */
gpointer vortex_xml_rpc_unmarshall_array_sync (XmlRpcMethodResponse    * response, 
					       XmlRpcArrayUnMarshaller   unmarshaller,
					       XmlRpcResponseStatus    * status, 
					       gint                    * fault_code,
					       gchar                  ** fault_string)
{
	XmlRpcMethodValue * value;
	XmlRpcArray       * result = NULL;
	gpointer            higher_result;
	
	/* get the value inside and status/fault_code/fault_string */
	value = __vortex_xml_rpc_unmarshall_common_sync (response, status, fault_code, fault_string);

	/* check method value received */
	if (value == NULL) {
		/* free response received, but after calling value */
		method_response_free (response);

		/* nothing to do, just return NULL */
		return NULL;
	}

	/* translate the array received */
	result = method_value_get_as_array (value);

	/* check if the unmarshaller function is provided */
	if (unmarshaller != NULL) {
		/* unmarshall */
		higher_result = unmarshaller (result, FALSE);

		/* free response received */
		method_response_free (response);

		/* return the higher result */
		return higher_result;
	}

	/* free response received */
	method_response_free (response);

	/* return result */
	return result;
}


/* @} */
