/*
 *  LibVortex:  A BEEP (RFC3080/RFC3081) implementation.
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are welcome to
 *  develop proprietary applications using this library without any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  For commercial support on build BEEP enabled solutions contact us:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila N 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 */

#include <vortex.h>

#define LOG_DOMAIN "vortex-xml-rpc"

/**
 * \defgroup vortex_xml_rpc_types Vortex XML-RPC types: XML-RPC data abstraction layer, supporting types used.
 */

/* \addtogroup vortex_xml_rpc_types */
/* @{ */

struct _XmlRpcMethodCall {
	/**
	 * @internal
	 * @brief Method name this MethodCall represents.
	 */
	gchar               * methodName;
	/** 
	 * @internal
	 * @brief How many parameters does this method call object have.
	 */
	gint                  count;
	/** 
	 * @internal
	 *
	 * @brief An internal value to check how many parameters are
	 * added used as a sanity check to avoid setting a method
	 * invocator object more values than the ones set after.
	 */
	gint                  added_count;
	/**
	 * @internal
	 * @brief A list of parameters for the method being invoked.
	 */
	XmlRpcMethodValue  ** params;

	/** 
	 * @internal
	 *
	 * @brief Allows to configure if the given object reference to
	 * be deallocated once performed the invocation. By default,
	 * all invocators are released, however, for serveral, and
	 * continuous invocations, this could be configured.
	 */
	gboolean              release_after_invoke;

	/** 
	 * @internal
	 *
	 * Internal channel number where the method call was
	 * received. This is used by the Vortex Library to perform
	 * deferred method invocation.
	 */
	VortexChannel      *  channel;

	/** 
	 * @internal
	 * 
	 * Internal message number this method call must reply to, to
	 * correlate the method reply to be generated.
	 */
	gint                  msg_no;
};

struct _XmlRpcMethodResponse {
	/**
	 * @internal
	 *
	 * @brief Current method response status. In case the response
	 * is not positive the the member response contains a pointer
	 * to the XmlRpcMethodResponseFail. 
	 * 
	 * Otherwise a pointer to a unique XmlRpcMethodValue is stored
	 * on the same member.
	 */
	XmlRpcResponseStatus   status;
	/**
	 * @internal
	 * @brief Stores current method response value.
	 */
	XmlRpcMethodValue    * value;
	/** 
	 * @internal
	 * @brief Current string representation for the value stored.
	 */
	gchar                * string_value;
	
	/**
	 * @internal
	 * @brief Contains the reply error found on the method response.
	 */
	gint                   fault_code;
	/**
	 * @internal
	 * @brief Contains a textual diagnostic for the error found.
	 */
	gchar                * fault_string;

};

struct _XmlRpcMethodResponseFail {

};

struct _XmlRpcMethodValue {
	/**
	 * @internal
	 * @brief Represents which values is actually hold by the this param
	 */
	XmlRpcParamType type;
	/**
	 * @internal
	 *
	 * @brief An union having the right value. Remember to leave
	 * this value as the last value.
	 */
	union {
		/* with this attribute is modeled the int and the
		 * boolean value. */
		gint    int_value; 
		/* with this attribute not only is modeled the string
		 * value but also the the base64 string. */
		gchar * string_value;
		gdouble double_value;
		XmlRpcStruct * rpc_struct;
		XmlRpcArray  * rpc_array;
	} value;
};

struct _XmlRpcStruct {
	/**
	 * @internal
	 * @brief How many member does this struct have.
	 */
	gint                  count;

	/** 
	 * @internal
	 * 
	 * An internal count for elements added at a particular
	 * moment. Used to check when new members are added and to
	 * check that the struct contains all members that are
	 * expected.
	 */
	gint                  added_count;

	/**
	 * @internal
	 * @brief An array of members stored on this struct.
	 */
	XmlRpcStructMember ** members;
};

struct _XmlRpcStructMember {
	/**
	 * @internal
	 * @brief The name of the member stored inside a struct 
	 */
	gchar               * name;
	/**
	 * @internal
	 * @brief The value of the member stored, that is a new
	 * XmlRpcMethodValue. This allows to have not only simple
	 * values but also more struct or arrays.
	 */
	XmlRpcMethodValue   * value;
};

struct _XmlRpcArray {
	/**
	 * @internal
	 * @brief How many array elements this array have.
	 */
	gint count;

	/** 
	 * @internal
	 * How many items were added to the array.
	 */
	gint added_count;
	/**
	 * @internal
	 * @brief An array of XmlRpcMethodValue values stored.
	 */
	XmlRpcMethodValue ** values;
};

/** 
 * @brief Creates a new method call object, representing a remote
 * procedure invocation.
 *
 * The object created by this function could be understood as a object
 * invocator, or a closure, that is, an high level object that
 * represents an method to be invoked. 
 *
 * Once the object is created by this function, it is needed (if
 * necessary) to add parameter values to it. This is done by using the
 * following functions:
 *
 *  - \ref vortex_xml_rpc_method_call_add_value
 *  - \ref vortex_xml_rpc_method_call_set_value
 *  - \ref vortex_xml_rpc_method_call_create_value
 *
 * Once the method object is completely built, then a call to the
 * following function is required to actually perform the invocation:
 * 
 *  - \ref vortex_xml_rpc_invoke
 *  - \ref vortex_xml_rpc_invoke_sync
 * 
 * Once the \ref XmlRpcMethodCall is created and used with the
 * invocation API (\ref vortex_xml_rpc_invoke_sync and \ref
 * vortex_xml_rpc_invoke), it will be deallocated automatically. This
 * is done to avoid the problem caused to deallocate this object mainly
 * for asynchronous invocation that makes really difficult to know
 * when to done this operation. 
 *
 * However, this behaviour could be configured. See \ref
 * vortex_xml_rpc_method_call_must_release. Keep in mind that the
 * invocator created by this function could be reused several times,
 * maybe to implement a kind of cache. But, if configured the
 * automatic release mechanism to not happen, you will responsible for
 * calling to \ref vortex_xml_rpc_method_call_free, when no longer
 * needed the reference.
 *
 * @param methodName The method name to invoke. This value must be not
 * NULL. The function will perform a local copy, so methodName passed
 * in value could be unrefered once this function finish.
 *
 * @param parameters The number of parameters this method invocator
 * object will contain. This value must be greater or equal to 0.
 * 
 * @return A new \ref XmlRpcMethodCall object, or NULL if it
 * fails. The function could only fail if it receive wrong parameters
 * values according to parameter description.
 */
XmlRpcMethodCall * vortex_xml_rpc_method_call_new (gchar * methodName, gint parameters)
{
	XmlRpcMethodCall * result;

	/* first check environment conditions */
	g_return_val_if_fail (methodName,      NULL);
	g_return_val_if_fail (parameters >= 0, NULL);

	result                       = g_new0 (XmlRpcMethodCall, 1);
	result->methodName           = g_strdup (methodName);
	result->release_after_invoke = AXL_TRUE;

	/* set the number of parameters this method name will support
	 * in the case the value is higher than 0 */
	result->count          = parameters;
	if (result->count > 0) {
		/* only allocate memory enough to hold pointers to */
		result->params = g_new0 (XmlRpcMethodValue *, result->count);
	}
	
	/* that's all men */
	return result;
}

/** 
 * @brief Allows to create a new \ref XmlRpcMethodValue object from
 * the given <i>type</i> and the given <i>value</i>.
 * 
 * You can use this function to create parameters to be added to
 * method objects (\ref XmlRpcMethodCall) using:
 *
 *  - \ref vortex_xml_rpc_method_call_add_value
 *  - \ref vortex_xml_rpc_method_call_set_value
 *
 * You can also skip this function and go directly to add the value
 * being created by this function to the method call using:
 * 
 *  - \ref vortex_xml_rpc_method_call_create_value
 *  - \ref vortex_xml_rpc_method_call_create_value_from_string
 *
 * This function returns a new \ref XmlRpcMethodValue reference which
 * should be unrefered using \ref vortex_xml_rpc_method_value_free. 
 *
 * This function will not perform a local copy for the given
 * <i>value</i> unless \ref XML_RPC_STRING_VALUE is provided. In that
 * case a local copy is made to enable the caller to unref the value
 * provided just returing from this function.
 *
 * This particular behavior is explained from the fact the other types
 * like bool or int doesn't required to perform a local copy, and
 * array and struct values are allocated through the provided
 * interface creating a copy.
 * 
 * But, there is no way to figure out if the string value received was
 * allocated on the heap or it is an stack activated value or simple
 * an static string.
 * 
 * Here are some examples:
 * \code
 * XmlRpcMethodValue * value;
 *
 * // Create a method value representing an integer value
 * value = vortex_xml_rpc_method_value_new (XML_RPC_INT_VALUE, 
 *                                          GPOINTER_TO_INT(10));
 * // free the value created 
 * vortex_xml_rpc_method_value_free (value);
 *
 * // Create a method value representing an string 
 * value = vortex_xml_rpc_method_value_new (XML_RPC_STRING_VALUE,
 *                                          "test");
 * // free the value created 
 * vortex_xml_rpc_method_value_free (value);
 * \endcode
 * 
 * @param type The type to be created.
 * @param value The value associated with the type being created
 * 
 * @return A new reference to \ref XmlRpcMethodValue object or NULL if
 * it fails. The function could only fail if the given value is NULL
 * or the type provided is not a valid value found in the enumeration,
 * not including \ref XML_RPC_UNKNOWN_VALUE.
 */
XmlRpcMethodValue * vortex_xml_rpc_method_value_new         (XmlRpcParamType     type,
							     gpointer            value)
{
	XmlRpcMethodValue * __value;

	/* perform some environment conditions */
	g_return_val_if_fail (XML_RPC_UNKNOWN_VALUE < type && type <= XML_RPC_NUM_SUPPORTED_VALUES,
			      NULL);
	/* check received value to not be null for pointer values */
	if (type != XML_RPC_INT_VALUE     &&
	    type != XML_RPC_DOUBLE_VALUE  &&
	    type != XML_RPC_BOOLEAN_VALUE && 
	    type != XML_RPC_NONE_VALUE    &&
	    type != XML_RPC_STRUCT_VALUE  &&
	    type != XML_RPC_ARRAY_VALUE) {
		g_return_val_if_fail (value, NULL);
	}

	__value       = g_new0 (XmlRpcMethodValue, 1);
	__value->type = type;
	
	/* save the value provided checking if we have received and string. */
	switch (type) {
	case XML_RPC_STRING_VALUE:
	case XML_RPC_BASE64_VALUE:
		__value->value.string_value = g_strdup (value);
		break;
	case XML_RPC_STRING_REF_VALUE:
	case XML_RPC_BASE64_REF_VALUE:
		/* own the reference received, do not allocate
		 * again */
		__value->value.string_value = value;

		/* update the enumerator value */
		if (type == XML_RPC_STRING_REF_VALUE)
			__value->type = XML_RPC_STRING_VALUE;
		if (type == XML_RPC_BASE64_REF_VALUE)
			__value->type = XML_RPC_BASE64_VALUE;
		break;
	case XML_RPC_INT_VALUE:
	case XML_RPC_BOOLEAN_VALUE:
		__value->value.int_value = GPOINTER_TO_INT (value);
		break;
	case XML_RPC_DOUBLE_VALUE:
		if (value != NULL) 
			__value->value.double_value = (gdouble) (* (gdouble *)value);
		else
			__value->value.double_value = 0;
		break;
	case XML_RPC_STRUCT_VALUE:
		/* check for null value to produce a none indicator */
		if (value == NULL) {
			/* set node value */
			__value->type  = XML_RPC_NONE_VALUE;

			/* do not set anyting more */
		}else {
			/* received a reference to the XmlRpcStruct
			 * pointer */
			__value->value.rpc_struct = value;
		}
		break;
	case XML_RPC_ARRAY_VALUE:
		/* check for null value to produce a none indicator */
		if (value == NULL) {
			/* set node value */
			__value->type  = XML_RPC_NONE_VALUE;

			/* do not set anyting more */
		}else {
			/* received a reference to the XmlRpcArray
			 * pointer */
			__value->value.rpc_array = value;
		}
		break;
	case XML_RPC_NONE_VALUE:
		/* nothing */
		break;
	default:
		/* nothing to do */
		break;
	}
	/* return the value */
	return __value;
}

/** 
 * @brief Allows to nullify provided method value, in the case an
 * internal reference is contained to avoid deallocating it once the
 * \ref vortex_xml_rpc_method_value_free is called.
 * 
 * @param __value The method value to nullify.
 */
void                vortex_xml_rpc_method_value_nullify         (XmlRpcMethodValue * __value)
{
	/* save the value provided checking if we have received and string. */
	switch (__value->type) {
	case XML_RPC_STRING_VALUE:
	case XML_RPC_BASE64_VALUE:
		/* nullify the value */
		__value->value.string_value = NULL;
		break;
	case XML_RPC_STRING_REF_VALUE:
	case XML_RPC_BASE64_REF_VALUE:
		/* nullify the value */
		__value->value.string_value = NULL;
		break;
	case XML_RPC_INT_VALUE:
	case XML_RPC_BOOLEAN_VALUE:
		/* nothing to do in this case */
		break;
	case XML_RPC_DOUBLE_VALUE:
		/* nothing to do in this case */
		break;
	case XML_RPC_STRUCT_VALUE:
		/* nullify the rpc struct */
		__value->value.rpc_struct = NULL;
		break;
	case XML_RPC_ARRAY_VALUE:
		/* nullify the rpc struct */
		__value->value.rpc_array = NULL;
		break;
	default:
		/* nothing to do */
		break;
	}
	/* return the value */
	return;
}

/** 
 * @brief Creates a new \ref XmlRpcMethodValue object from the given
 * type, using as value the string provided.
 *
 * The string provided must contain a value according to the type
 * provided. See also \ref
 * vortex_xml_rpc_method_value_new_from_string2, which could create a
 * method value by providing all its values as string references.
 * 
 * @param type The type for the new instance.
 * @param string_value A string value, representing the type configured.
 * 
 * @return A newly created \ref XmlRpcMethodValue or NULL if fails.
 */
XmlRpcMethodValue * vortex_xml_rpc_method_value_new_from_string (XmlRpcParamType     type,
								 gchar             * string_value)
{
	gdouble double_value;
	gint    int_value;
	gchar * string_aux = NULL;

	switch (type) {
	case XML_RPC_INT_VALUE:
	case XML_RPC_BOOLEAN_VALUE:
		int_value = (gint) strtol (string_value, &string_aux, 10);
		if ((string_aux != NULL) && (* string_aux))
			return NULL;
		return method_value_new (type, GINT_TO_POINTER (int_value));
	case XML_RPC_DOUBLE_VALUE:
		if ((string_aux != NULL) && (* string_aux))
			return NULL;
		double_value = strtod (string_value, &string_aux);
		return method_value_new (type, (gpointer) &double_value);
	case XML_RPC_STRING_VALUE:
		return method_value_new (type, string_value);
	case XML_RPC_DATE_VALUE:
		/* not supported yet */
		break;
	case XML_RPC_BASE64_VALUE:
		return method_value_new (type, string_value);
	case XML_RPC_ARRAY_VALUE:
		/* not supported yet: to support this case, an string
		 * format must be defined to allow this */
		break;
	case XML_RPC_STRUCT_VALUE:
		/* not supported yet: the same applies to this case as
		 * the array case. */
		break;
	case XML_RPC_NONE_VALUE:
		/* create the none value */
		return method_value_new (type, NULL);
	default:
		/* nothing to do */
		break;
	}

	return NULL;
}

/** 
 * @brief Allows to create a new method value (\ref XmlRpcMethodValue)
 * by providing two strings which are the type and the value
 * representation.
 *
 * The type string which have the following values according to the
 * desired type value:
 *
 * <table>
 *   <tr>
 *     <td>Type </td><td>Enum type value</td><td>Value to be used</td>
 *   </tr>
 *   <tr>
 *     <td>Integer value </td><td>\ref XML_RPC_INT_VALUE</td><td>"int"</td>
 *   </tr>
 *   <tr>
 *     <td>Boolean value </td><td>\ref XML_RPC_BOOLEAN_VALUE</td><td>"boolean"</td>
 *   </tr>
 *   <tr>
 *     <td>String value </td><td>\ref XML_RPC_STRING_VALUE</td><td>"string"</td>
 *   </tr>
 *   <tr>
 *     <td>Base64 string value </td><td>\ref XML_RPC_BASE64_VALUE</td><td>"base64"</td>
 *   </tr>
 *   <tr>
 *     <td>Struct value </td><td>\ref XML_RPC_STRUCT_VALUE</td><td>Not supported</td>
 *   </tr>
 *   <tr>
 *     <td>Array value </td><td>\ref XML_RPC_ARRAY_VALUE</td><td>Not supported</td>
 *   </tr>
 * </table>
 * 
 * @param type An string which specifies the method value type. See the table above.
 *
 * @param string_value The string value associated to the method
 * value, which is a proper string representation of the string value.
 * 
 * @return A newly created method value, that must be deallocated by
 * using \ref vortex_xml_rpc_method_value_free. This is not required
 * if the value is set as a method parameter (for a \ref
 * XmlRpcMethodCall) or a method response (\ref
 * XmlRpcMethodResponse). The function will return NULL if some
 * parameter received is NULL or the type specification is not
 * properly formated.
 */
XmlRpcMethodValue * vortex_xml_rpc_method_value_new_from_string2 (gchar * type,
								  gchar * string_value)
{
	/* perform some environment checks */
	g_return_val_if_fail (type, NULL);
	g_return_val_if_fail (string_value, NULL);

	if (!g_strcasecmp ("int", type))
		return vortex_xml_rpc_method_value_new_from_string (XML_RPC_INT_VALUE, string_value);

	if (!g_strcasecmp ("boolean", type))
		return vortex_xml_rpc_method_value_new_from_string (XML_RPC_BOOLEAN_VALUE, string_value);

	if (!g_strcasecmp ("string", type))
		return vortex_xml_rpc_method_value_new_from_string (XML_RPC_STRING_VALUE, string_value);

	return NULL;
}

/** 
 * @brief Allows to add the given \ref XmlRpcMethodValue in the next
 * parameter position available.
 *
 * This function allows to add created \ref XmlRpcMethodValue into the
 * given \ref XmlRpcMethodCall following the position of the call
 * ordering.
 * 
 * Here is an example where a method with two parameters is created
 * and then these two parameter values are added sequentially:
 *
 * \code
 * XmlRpcMethodCall  * method_call;
 * XmlRpcMethodValue * value;
 * 
 * // create a method call object setting that it will have only 1 
 * // parameter.
 * method_call = method_call_new ("examples.getStateName", 1);
 *
 * // create the value and add it
 * value       = method_value_new (XML_RPC_INT_VALUE, 
 *                                 GPOINTER_TO_INT(41));
 *
 * // add the value into the invocator object
 * method_call_add_value (method_call, value);
 *
 * // create the next value
 * value      = method_value_new (XML_RPC_STRING_VALUE, "test");
 *
 * // add the value into the invocation object
 * method_call_add_value (method_call, value);
 *
 * \endcode
 *
 * As we have observed from the example, the parameters are added as
 * they are created, then this function keeps tracking about the
 * position where the next element added should be stored. Moreover,
 * the function checks you don't add more parameters than the value
 * configured while creating the \ref XmlRpcMethodCall object, at \ref method_call_new.
 *
 * The function will fail if some of the parameters received are NULL
 * or the caller is trying to add more parameter than the \ref
 * XmlRpcMethodCall could accept. The function will not perform any
 * copy from data received.
 * 
 * @param method_call The \ref XmlRpcMethodCall object where the value will be added.
 * @param value       The \ref XmlRpcMethodValue to be added.
 * 
 */
void                vortex_xml_rpc_method_call_add_value    (XmlRpcMethodCall  * method_call,
							     XmlRpcMethodValue * value)
{
	/* perform a sanity check */
	g_return_if_fail (method_call);
	g_return_if_fail (value);
	g_return_if_fail (method_call->count != method_call->added_count);
	
	/* add the object */
	method_call->params[method_call->added_count] = value;

	/* increase added count */
	method_call->added_count++;
	
	return;
}

/** 
 * @brief Allows to set a particular value at a particular position
 * into the parameter invocation order.
 *
 * Because \ref vortex_xml_rpc_method_call_add_value works in a
 * sequential mode, it could be inconvenient under some
 * circumstances. This function allows to skip over the sequential
 * adding feature to implement an index-based parameter configuration.
 *
 * Keep in mind that added in a index-based manner you'll have to pay
 * attention about number of parameters, its order, etc. The function
 * will not modify values used to perform sequential adding. This mean
 * that sequential and index-based adding operation should not be
 * mixed on the same method call object.
 *
 * Position values should be taken to range from 0 up to number of
 * parameters - 1. Function with 2 parameters have as available and
 * valid positions [0,1].
 *
 * Once a value is added, previous stored value, if were defined, is
 * unrefered. 
 *
 * The function will not perform any operation if received values are
 * NULL or the position provided is wrong. 
 *
 * @param method_call The method call where the index-based adding operation will be performed.
 * @param position    The position where operate
 * @param value       The value to be set.
 */
void                vortex_xml_rpc_method_call_set_value    (XmlRpcMethodCall  * method_call,
							     gint                position,
							     XmlRpcMethodValue * value)
{
	XmlRpcMethodValue * __previous;

	/* perform some environment condition checks */
	g_return_if_fail (method_call);
	g_return_if_fail (value);
	g_return_if_fail (0 <= position && position < method_call->count);
	
	/* get previous value to perform a deallocation operation */
	__previous = method_call->params[position];

	/* set new value */
	method_call->params[position] = value;

	/* check and deallocate if necessary */
	if (__previous != NULL)
		vortex_xml_rpc_method_value_free (__previous);

	/* peter: hey jack, this function you have writtn' looks
	 *        pretty simple, you are a great programmer! */
	/* jack: i won't let you play with my PSP! */
	return;
}

/** 
 * @brief Convenience function to avoid XML-RPC API consumers to write to much code.
 *
 * This function will create a \ref XmlRpcMethodValue using the given
 * value and then will add it using the \ref vortex_xml_rpc_method_call_add_value.
 *
 * The function will not perform any operation if the received value
 * are NULL or the type provided doesn't meet the criteria stated at
 * \ref vortex_xml_rpc_method_value_new.
 *
 * @param method_call The method call object where the operation will be performed.
 * @param type        The type value for the method value object to be created.
 * @param value       The value associated to the method value object to be created.
 */
void                vortex_xml_rpc_method_call_create_value (XmlRpcMethodCall  * method_call,
							     XmlRpcParamType     type,
							     gpointer            value)
{
	XmlRpcMethodValue * __value;

	g_return_if_fail (method_call);
	g_return_if_fail (value);
	
	/* create the value to be added and check if it was succesful */
	__value = vortex_xml_rpc_method_value_new (type, value);
	if (__value == NULL)
		return;

	/* add the value created into the method call object */
	vortex_xml_rpc_method_call_add_value (method_call, __value);
	
	return;
}

/** 
 * @brief Convenience function to avoid XML-RPC API consumers to write to much code.
 * 
 * This function will create a \ref XmlRpcMethodValue using the given
 * value (through \ref vortex_xml_rpc_method_value_new_from_string)
 * and then will add it using the \ref vortex_xml_rpc_method_call_add_value.
 *
 * 
 *
 * @param method_call The method call object where the operation will be performed.
 * @param type        The type value for the method value object to be created.
 * @param string_value  The value associated to the method value object to be created, in a string form.
 */
void                vortex_xml_rpc_method_call_create_value_from_string (XmlRpcMethodCall * method_call,
									 XmlRpcParamType    type,
									 gchar            * string_value)
{
	XmlRpcMethodValue * __value;

	g_return_if_fail (method_call);
	g_return_if_fail (string_value);
	
	/* create the value to be added and check if it was succesful */
	__value = vortex_xml_rpc_method_value_new_from_string (type, string_value);
	if (__value == NULL)
		return;

	/* add the value created into the method call object */
	vortex_xml_rpc_method_call_add_value (method_call, __value);

	return;
}


/** 
 * @brief Allows to get current method name for the given \ref XmlRpcMethodCall object.
 *
 * The value returned by this function is the method name that will be
 * invoked. You must not unref the value returned by this function,
 * because it is an internal copy to the \ref XmlRpcMethodCall object.
 * 
 * @param method_call The method call to get its name.
 * 
 * @return An string reference representing the method name or NULL if
 * it fails. The function only fails if the method call object
 * reference provided is NULL.
 */
gchar             * vortex_xml_rpc_method_call_get_name           (XmlRpcMethodCall  * method_call)
{
	g_return_val_if_fail (method_call, NULL);

	return method_call->methodName;
}

/** 
 * @brief Allows to get which is the number of parameters that the
 * given \ref XmlRpcMethodCall object has.
 *
 * The parameter number returned by this function is the one specified
 * while calling to \ref vortex_xml_rpc_method_call_new. It is not the
 * number of parameter that are already added.
 *
 * You should use \ref
 * vortex_xml_rpc_method_call_get_current_num_params to get which is
 * the number of parameter that are added. A properly formated \ref
 * XmlRpcMethodCall should return the same value while calling to both
 * function.
 *
 * @param method_call The method call object to get its parameter numbers.
 * 
 * @return The function return which is the number of parameter that
 * has the object or -1 if it fails. The function only fails if the
 * given method call is NULL.
 */
gint            vortex_xml_rpc_method_call_get_num_params     (XmlRpcMethodCall  * method_call)
{
	g_return_val_if_fail (method_call, -1);

	return method_call->count;
}

/** 
 * @brief Allows to get current number of parameters already added to
 * the given \ref XmlRpcMethodCall object.
 * 
 * @param method_call The method call object to get current parameter
 * number.
 * 
 * @return The number of parameter already added or -1 if fails.
 */
gint            vortex_xml_rpc_method_call_get_current_num_params (XmlRpcMethodCall * method_call)
{
	g_return_val_if_fail (method_call, -1);

	return method_call->added_count;
}

/** 
 * @brief Allows to get the param value (\ref XmlRpcMethodValue) from
 * the given \ref XmlRpcMethodCall at the selected position.
 *
 * 
 *
 * @param method_call The method call object where the param value
 * will be retrieved.
 *
 * @param position The param value to get. Valid positions ranges from
 * 0 up to param number -1.
 * 
 * @return An referece to the \ref XmlRpcMethodValue or NULL if
 * fails. Returned value should not be deallocated. The function will
 * check that the given method call object is not NULL and the
 * position is in the range of {0..param number - 1}. 
 */
XmlRpcMethodValue * vortex_xml_rpc_method_call_get_param_value (XmlRpcMethodCall * method_call, 
								gint position) 
{
	g_return_val_if_fail (method_call, NULL);
	g_return_val_if_fail ((position >= 0) && (position < method_call->count), NULL);
	g_return_val_if_fail (method_call, NULL);
	
	return method_call->params[position];
}

/** 
 * @brief Helper function that allows to get the &lt;integer> value from
 * the given position for the arguments received inside the given \ref
 * XmlRpcMethodCall object.
 *
 * This function is a short-cut to use: 
 *   - \ref vortex_xml_rpc_method_call_get_param_value 
 * and then calling to:
 *   - \ref vortex_xml_rpc_method_value_get_as_int
 *
 *
 * @param method_call The method call object where the value will be
 * extracted.
 *
 * @param position The position to check.
 * 
 * @return The function will return the value associated on the given
 * possition or -1 if fails. Keep in mind that a \ref
 * XmlRpcMethodValue with \ref XML_RPC_INT_VALUE could have as a value
 * a <b>-1</b>. So you have to ensure you are not getting a value that
 * is not an &lt;integer> or, the parameter defined by the given position
 * exists.
 */
gint                vortex_xml_rpc_method_call_get_param_value_as_int    (XmlRpcMethodCall * method_call,
									  gint position)
{
	XmlRpcMethodValue * value;

	
	/* get xml rpc method value */
	value = vortex_xml_rpc_method_call_get_param_value (method_call, position);

	/* check environment conditions */
	g_return_val_if_fail (value, -1);
	g_return_val_if_fail (value->type == XML_RPC_INT_VALUE || value->type == XML_RPC_DOUBLE_VALUE, -1);

	/* return an int value */
	return value->value.int_value;
}

/** 
 * @brief Allow to get current double value associated to the giveh
 * \ref XmlRpcMethodCall, at the given position for the parameter
 * value.
 *
 * If you are looking for a function to get the double inside a
 * particular \ref XmlRpcMethodValue, you can check \ref
 * vortex_xml_rpc_method_value_get_as_double.
 * 
 * @param method_call The method call where the double value will be
 * retrieved.
 *
 * @param position The parameter position to get.
 * 
 * @return A double value or 0.0 if fails. There is no easy way to get
 * that the returned 0.0 value is because an error or due to that
 * indeed, the parameter accessed have that value.
 */
gdouble            vortex_xml_rpc_method_call_get_param_value_as_double (XmlRpcMethodCall * method_call,
									 gint position)
{
	XmlRpcMethodValue * value;
	
	/* get xml rpc method value */
	value = vortex_xml_rpc_method_call_get_param_value (method_call, position);

	/* return the value inside */
	return vortex_xml_rpc_method_value_get_as_double (value);
}

/** 
 * @brief Convenience function to the the string associated to the
 * given parameter, indentified by the position, inside the given \ref
 * XmlRpcMethodCall object.
 *
 * If you are looking for a function that returns the string inside a
 * particular \ref XmlRpcMethodValue, check \ref vortex_xml_rpc_method_value_get_as_string.
 * 
 * @param method_call The method call object.
 *
 * @param position The position of the parameter, being
 * XML_RPC_STRING_VALUE, where the value will be received.
 * 
 * @return A string associated or NULL if fails. The string returned
 * must not be deallocated. 
 */
gchar *             vortex_xml_rpc_method_call_get_param_value_as_string (XmlRpcMethodCall * method_call,
									  gint position)
{
	XmlRpcMethodValue * value;
	
	/* get xml rpc method value */
	value = vortex_xml_rpc_method_call_get_param_value (method_call, position);

	/* return the string value inside, if it is possible */
	return vortex_xml_rpc_method_value_get_as_string (value);
}

/** 
 * @brief Allows to get the struct provided at the given position.
 * 
 * @param method_call The method call where the struct is being retrieved.
 *
 * @param position The position to look up.
 * 
 * @return A reference to the struct or NULL if it fails.
 */
XmlRpcStruct      * vortex_xml_rpc_method_call_get_param_value_as_struct (XmlRpcMethodCall * method_call,
									  gint position)
{
	XmlRpcMethodValue * value;
	
	/* get xml rpc method value */
	value = vortex_xml_rpc_method_call_get_param_value (method_call, position);

	/* return the string value inside, if it is possible */
	return vortex_xml_rpc_method_value_get_as_struct (value);
}

/** 
 * @brief Allows to get the XmlRpcArray inside the provided method call.
 * 
 * @param method_call The method call where the XmlRpcArray is being
 * requested.
 *
 * @param position The position to look up.
 * 
 * @return A reference to the XmlRpcArray or NULL if it fails.
 */
XmlRpcArray       * vortex_xml_rpc_method_call_get_param_value_as_array  (XmlRpcMethodCall * method_call,
									  gint position)
{
	XmlRpcMethodValue * value;
	
	/* get xml rpc method value */
	value = vortex_xml_rpc_method_call_get_param_value (method_call, position);

	/* return the string value inside, if it is possible */
	return vortex_xml_rpc_method_value_get_as_array (value);	
}

/* function prototype */
gchar * vortex_xml_rpc_marshall_method_value (XmlRpcMethodValue * value);

/** 
 * @internal
 * 
 * Marshall the provided struct representation into an xml
 * representation.
 * 
 * @param _struct The struct to marshall
 * 
 * @return A memory allocated representation for the provided
 * structure.
 */
gchar * vortex_xml_rpc_types_marshall_struct (XmlRpcStruct * _struct)
{
	gchar * stream_aux;
	gchar * stream_aux2;
	gchar * stream_result;
	gint    iterator;
	

	/* initial struct header */
	stream_result = g_strdup ("<struct>");
	iterator      = 0;

	while (iterator < _struct->added_count) {

		/* marshall the method member */
		stream_aux2 = vortex_xml_rpc_marshall_method_value (_struct->members[iterator]->value);
		
		/* get the member representation */
		stream_aux  = g_strdup_printf ("%s<member><name>%s</name>%s</member>",
					       stream_result,
					       _struct->members[iterator]->name,
					       stream_aux2);

		/* release the member method value allocated */
		g_free (stream_aux2);
		g_free (stream_result);

		/* update the new stream result */
		stream_result = stream_aux;

		/* update iterator count */
		iterator++;
	}
	
	/* add final representation for the struct */
	stream_aux   = stream_result;
	stream_result = g_strdup_printf ("%s</struct>", stream_result);
	g_free (stream_aux);
	
	return stream_result;
}


/** 
 * @internal
 *
 * Internal function which allows to marshal the provided XmlRpcArray
 * array reference into the Xml RPC presentation.
 * 
 * @param array The array to marshall.
 * 
 * @return Returns an string containing the array marshalled.
 */
gchar * vortex_xml_rpc_types_marshall_array (XmlRpcArray * array)
{
	gchar * stream_aux;
	gchar * stream_aux2;
	gchar * stream_result;
	gint    iterator;

	/* initial struct header */
	stream_result = g_strdup ("<array><data>");
	iterator      = 0;

	while (iterator < array->added_count) {

		/* marshall the method member */
		stream_aux2 = vortex_xml_rpc_marshall_method_value (array->values[iterator]);
		
		/* get the member representation */
		stream_aux  = g_strdup_printf ("%s%s",
					       stream_result,
					       stream_aux2);

		/* release the member method value allocated */
		g_free (stream_aux2);
		g_free (stream_result);

		/* update the new stream result */
		stream_result = stream_aux;

		/* update iterator count */
		iterator++;
	}
	
	/* add final representation for the struct */
	stream_aux   = stream_result;
	stream_result = g_strdup_printf ("%s</data></array>", stream_result);
	g_free (stream_aux);
	
	return stream_result;
}

/** 
 * @internal Marshall the provided method value, returning its string
 * representation.
 * 
 * @param value The method value to marshall.
 * 
 * @return An allocated string representing the method value
 * marshalled.
 */
gchar * vortex_xml_rpc_marshall_method_value (XmlRpcMethodValue * value)
{
	gchar * stream_result;
	gchar * stream_aux;

	switch (method_value_get_type (value)) {
	case XML_RPC_INT_VALUE:
		return g_strdup_printf ("<value><i4>%d</i4></value>", 
					method_value_get_as_int (value));
		break;
	case XML_RPC_BOOLEAN_VALUE:
		return g_strdup_printf ("<value><boolean>%d</boolean></value>", 
					method_value_get_as_int (value));
	case XML_RPC_DOUBLE_VALUE:
		return g_strdup_printf ("<value><double>%g</double></value>", 
					method_value_get_as_double (value));
		break;
	case XML_RPC_STRING_VALUE:
		return g_strdup_printf ("<value><string>%s</string></value>",
					(gchar *) value->value.string_value);
		break;
	case XML_RPC_DATE_VALUE:
		/* not implemented yet */
		break;

	case XML_RPC_BASE64_VALUE:
		/* not implemented yet */
		break;
	case XML_RPC_STRUCT_VALUE:
		/* marshall the value */
		stream_aux    = vortex_xml_rpc_types_marshall_struct (value->value.rpc_struct);
		
		/* use the value marshalled */
		stream_result = g_strdup_printf ("<value>%s</value>",
						 stream_aux);
		/* free the marshalled value */
		g_free (stream_aux);

		return stream_result;
	case XML_RPC_ARRAY_VALUE:
		/* marshall the value */
		stream_aux    = vortex_xml_rpc_types_marshall_array (value->value.rpc_array);
		
		/* use the value marshalled */
		stream_result = g_strdup_printf ("<value>%s</value>",
						 stream_aux);
		/* free the marshalled value */
		g_free (stream_aux);

		return stream_result;
		break;
	case XML_RPC_NONE_VALUE:
		/* return the none value */
		return g_strdup_printf ("<value><none /></value>");
	default:
		/* nothing to do */
		return NULL;
	}
	
	/* this point is never reached */
	return NULL;
}

/** 
 * @brief Perform a marshalling from the invocator provided
 * (<b>method_call</b>) into the appropiate XML-RPC representation.
 *
 * This function allows to convert the given method call into the
 * XML-RPC representation. The function return the xml stream
 * representing the invocator in a dynamically allocated value. It
 * should be deallocated once no longer needed.
 *
 * Additionally, the function allows to set a pointer to a integer
 * (<b>size parameter</b>) so the function could return the xml stream
 * size generated. This method is preferred over getting the xml
 * stream and then using something similar to <b>strlen</b>.
 * 
 * @param method_call The method invocator to marshall into the xml
 * representation.
 *
 * @param size A pointer to a integer where the size for the given
 * stream will be returned. This is an optional parameter.
 * 
 * @return The function return the xml representation or NULL if it
 * fails. The function will the that the method call object received
 * is not NULL and all parameters required were added. This means that
 * the method parameter count and parameters added count should be
 * equal.
 */
gchar             * vortex_xml_rpc_method_call_marshall     (XmlRpcMethodCall  * method_call,
							     gint              * size)
{
	XmlRpcMethodValue * value;
	gchar             * stream_result;
	gchar             * stream_aux;
	gchar             * stream_aux2;
	gint                iterator;

	g_return_val_if_fail (method_call, NULL);
	g_return_val_if_fail (method_call->count == method_call->added_count, NULL);

	/* get initial method header */
	stream_result = g_strdup_printf ("<?xml version=\"1.0\"?><methodCall><methodName>%s</methodName>",
					 method_call->methodName);

	/* if the method have at least one parameter, add them */
	if (method_call->count > 0) {
		/* get a refence to the previous message, add initial
		 * <params> header and release no longer used
		 * memory */
		stream_aux    = stream_result;
		stream_result = g_strdup_printf ("%s<params>", stream_aux);
		g_free (stream_aux);

		/* iterate over all param values */
		for (iterator = 0; iterator < method_call->count; iterator++) {
			/* get a reference to the value */
			value       = method_call_get_param_value (method_call, iterator);
			
			/* marshall the param value according to its type */
			stream_aux  = vortex_xml_rpc_marshall_method_value (value);
			
			/* get the concatenated result */
			stream_aux2 = g_strdup_printf ("%s<param>%s</param>", stream_result, stream_aux);
			
			/* free previous data */
			g_free (stream_aux);
			g_free (stream_result);

			/* set the new stream result */
			stream_result = stream_aux2;

		}

		/* finally close the param values section for the
		 * invocation xml message */
		stream_aux    = stream_result;
		stream_result = g_strdup_printf ("%s</params>", 
						 (stream_aux != NULL) ? stream_aux : NULL);
		g_free (stream_aux);
	}

	/* close the message and return the value */
	stream_aux    = stream_result;
	stream_result = g_strdup_printf ("%s</methodCall>", stream_aux);
	g_free (stream_aux);
	
	/* return current result, setting current size. The following
	 * operation is based on strlen operation. This is ok because
	 * XML-RPC standard doesn't support raw binary formats. To
	 * include binary data, it should be encoded into a base64
	 * format.
	 */
	if (size != NULL) 
		(* size ) = strlen (stream_result);
	return stream_result;
}


/** 
 * @brief Unref all memory allocated by the method call object and its
 * childs, including method call values.
 *
 * This function will perform a deep unref operation on the given
 * method call object. 
 * 
 * @param method_call The method call to be deallocated.
 */
void                vortex_xml_rpc_method_call_free         (XmlRpcMethodCall  * method_call)
{
	gint iterator;
	g_return_if_fail (method_call);

	/* for each method value object, perform an unref operation over it */
	for (iterator = 0; iterator < method_call->added_count; iterator ++) {
		vortex_xml_rpc_method_value_free (method_call->params[iterator]);
	}

	/* unref params container */
	g_free (method_call->params);
	
	/* unref methodName value */
	g_free (method_call->methodName);
	
	/* unref the object itself */
	g_free (method_call);
	
	return;
}

/** 
 * @brief Allows to configure how the XML-RPC invocation mechanism
 * manages invocator objects that have performed its invocation.
 *
 * By default, every \ref XmlRpcMethodCall object created to perform
 * an invocation is deallocated automatically once the invocation is
 * performed. However, if it is planned to perform the same invocation
 * several times or to implement a kind of invocation cache, then this
 * function should be used to notify Vortex Library to not release the
 * object once performed its invocation.
 *
 * @param method_call The \ref XmlRpcMethodCall object to configure.
 *
 * @param release TRUE to release the object after perform the
 * invocation, FALSE to not deallocate it.
 */
void                vortex_xml_rpc_method_call_release_after_invoke (XmlRpcMethodCall * method_call, 
								     gboolean release)
{
	g_return_if_fail (method_call);

	/* store the value */
	method_call->release_after_invoke = release;
	
	return;
}

/** 
 * @brief Returns current status for releasing the given object after
 * performing an invoke.
 * 
 * @param method_call The method call to check for its configuration.
 * 
 * @return TRUE if the given method call object should be deallocated
 * after the invoke is performed. Otherwise FALSE is returned.
 */
gboolean            vortex_xml_rpc_method_call_must_release (XmlRpcMethodCall * method_call)
{
	g_return_val_if_fail (method_call, FALSE);

	return method_call->release_after_invoke;
}

/** 
 * @brief Allows to check if the given method call meets the requirements provided.
 * 
 * Before executing a method call, this function could be used to
 * peform some basic method call recognition, based on the method
 * name, the parameter number, and tye expected type for the incoming
 * data.
 * 
 * Here is an example to validate the method called "sum" with 2
 * integer parameters:
 *
 * \code
 * if (vortex_xml_rpc_method_call_is (method_call, "sum", 2, 
 *                                    XML_RPC_INT_VALUE,
 *                                    XML_RPC_INT_VALUE, -1) {
 *     g_print ("Great, this is the method we where looking for");  
 * }
 * \endcode
 *
 * Argument type validation must always finish with a <b>-1</b>
 * value. If it is only required to perform a method call recognition
 * based on the method name and its argument number without taking
 * into account the argument type validation, then <b>-1</b> must be
 * used as the last value.
 * 
 * @param method_call The method call to check. 
 *
 * @param method_name The method name to check. This value is
 * optional. If it is not provided, the method name will not be
 * checked.
 *
 * @param param_num The method parameter number to check. To disable
 * method parameter number check, -1 must be used.
 *
 * 
 * @return TRUE if method match, otherwise FALSE is returned.
 */
gboolean            vortex_xml_rpc_method_call_is           (XmlRpcMethodCall * method_call, 
							     gchar            * method_name,
							     gint               param_num,
							     ...)
{
	va_list             args;
	XmlRpcParamType     type;
	gint                iterator = 0;
	XmlRpcMethodValue * value;

	g_return_val_if_fail (method_call, FALSE);
	
	/* check method name */
	if (method_name != NULL) {
		if (g_strcasecmp (method_call->methodName, method_name))
			return FALSE;
	}

	/* check method parameter number */
	if (param_num != -1) {
		if (param_num != method_call->count)
			return FALSE;
	}

	/* open std arg list */
	va_start (args, param_num);

	/* now check method call type signature */
	while ((type = va_arg (args, XmlRpcParamType)) != -1) {

		/* get the value from the given position */
		value = method_call_get_param_value (method_call, iterator);
		if (value == NULL) {
			/* uhmm...it seems that it was requested to
			 * check for an argument that thet the
			 * function doesn't have. */
			va_end (args);
			return FALSE;
		}
		
		/* check type */
		if (method_value_get_type (value) != type) {
			/* method value type doesn't match */
			va_end (args);
			return FALSE;
		}
		
		/* update iterator to check */
		iterator++;
	}

	/* close the standard arg */
	va_end (args);

	return TRUE;
}

/** 
 * @internal
 *
 * @brief Allows to configure reply data to the given \ref
 * XmlRpcMethodCall object.
 */
void                __vortex_xml_rpc_method_call_set_reply_data (XmlRpcMethodCall * method_call,
								 VortexChannel    * channel,
								 gint               msg_no)
{
	g_return_if_fail (method_call);
	
	method_call->channel = channel;
	method_call->msg_no  = msg_no;
	
	return;
}

/** 
 * @internal
 *
 * @brief Allows to get reply data configured in the given \ref
 * XmlRpcMethodCall object.
 */
void                __vortex_xml_rpc_method_call_get_reply_data (XmlRpcMethodCall *  method_call,
								 VortexChannel    ** channel,
								 gint              * msg_no)
{
	g_return_if_fail (method_call);

	if (channel != NULL)
		(* channel ) = method_call->channel;
	if (msg_no != NULL)
		(* msg_no ) = method_call->msg_no;
	return;
}

/** 
 * @brief Perform a memory deallocation operation over the given value.
 *
 * This function will perform a deep unref operation over the given
 * value, including string stored, array, structs and all values
 * inside those holder structures.
 * 
 * @param value The method value to be destroyed.
 */
void                vortex_xml_rpc_method_value_free        (XmlRpcMethodValue * value)
{
	g_return_if_fail (value);

	/* unref internal value according to the type. */
	switch (value->type) {
	case XML_RPC_INT_VALUE:
	case XML_RPC_BOOLEAN_VALUE:
		/* do nothing */
		break;
	case XML_RPC_STRING_VALUE:
	case XML_RPC_BASE64_VALUE:
		/* perform a simple deallocation operation */
		if (value->value.string_value != NULL)
			g_free (value->value.string_value);
		break;
	case XML_RPC_STRUCT_VALUE:
		/* struct case */
		if (value->value.rpc_struct != NULL)
			vortex_xml_rpc_struct_free (value->value.rpc_struct);
		break;
	case XML_RPC_ARRAY_VALUE:
		/* array case */
		if (value->value.rpc_array != NULL)
			vortex_xml_rpc_array_free (value->value.rpc_array);
		break;
	case XML_RPC_DATE_VALUE:
		/* date case */
		break;
	default:
		/* do nothing */
		break;
	}
	
	/* unref the node itself */
	g_free (value);

	return;
}

/** 
 * @brief Allows to get current type for the given method value.
 *
 * @param value The \ref XmlRpcMethodValue object to get its type.
 * 
 * @return Current type the given value has or \ref
 * XML_RPC_UNKNOWN_VALUE if it fails.
 */
XmlRpcParamType     vortex_xml_rpc_method_value_get_type    (XmlRpcMethodValue * value)
{
	g_return_val_if_fail (value, XML_RPC_UNKNOWN_VALUE);

	return value->type;
}

/** 
 * @brief Allows to get current integer value from the given method
 * value object (\ref XmlRpcMethodValue).
 *
 * This function not only is used to get param values for the type
 * \ref XML_RPC_INT_VALUE. Because the XML-RPC definition says that a
 * boolean value is represented with the integer values 0 and 1,
 * corresponding to FALSE and TRUE, this function is also used to get
 * the value for a \ref XML_RPC_BOOLEAN_VALUE.
 *
 * @param value The method value to get its value as an integer.
 * 
 * @return The Integer value or -1 if it fails. Keep in mind that, the
 * stored value could be also -1. Check that you are not passing a
 * method value that is NULL or it doesn't hold an integer.
 */
gint                vortex_xml_rpc_method_value_get_as_int  (XmlRpcMethodValue * value)
{
	g_return_val_if_fail (value, -1);
	g_return_val_if_fail (value->type == XML_RPC_INT_VALUE || 
			      value->type == XML_RPC_BOOLEAN_VALUE, -1);

	/* return current internal value as an integer */
	return value->value.int_value;
}

/** 
 * @brief Allows to get the string inside the method value, supposing
 * the method value represents an \ref XML_RPC_STRING_VALUE.
 * 
 * @param value The method value where the value will be returned.
 * 
 * @return An internal reference to the string inside the method value
 * or NULL if it values. The value returned must not be deallocated.
 */
gchar *             vortex_xml_rpc_method_value_get_as_string (XmlRpcMethodValue * value)
{
	/* check environment conditions */
	g_return_val_if_fail (value, NULL);
	g_return_val_if_fail (value->type == XML_RPC_STRING_VALUE, NULL);

	/* return an int value */
	return value->value.string_value;
}

/** 
 * @brief Allows to get the double value inside the provided \ref XmlRpcMethodValue.
 * 
 * @param value The method value where the operation is being
 * requested.
 * 
 * @return The double inside the method value.
 */
gdouble             vortex_xml_rpc_method_value_get_as_double   (XmlRpcMethodValue * value)
{
	/* check environment conditions */
	g_return_val_if_fail (value, 0.0);
	g_return_val_if_fail (value->type == XML_RPC_DOUBLE_VALUE, 0.0);

	/* return an int value */
	return value->value.double_value;
}

/** 
 * @brief Allows to get the struct (\ref XmlRpcStruct)  inside the provided \ref XmlRpcMethodValue.
 * 
 * @param value The method value which is expected to have an struct
 * inside (\ref XmlRpcStruct).
 * 
 * @return An internal reference to the struct inside the given \ref
 * XmlRpcMethodValue or NULL if it fails. The returned reference must
 * not be deallocated.
 */
XmlRpcStruct     * vortex_xml_rpc_method_value_get_as_struct   (XmlRpcMethodValue * value)
{
	/* check environment conditions */
	g_return_val_if_fail (value, NULL);
	g_return_val_if_fail (value->type == XML_RPC_STRUCT_VALUE, NULL);

	/* return an int value */
	return value->value.rpc_struct;
}

/** 
 * @brief Allows to get the array (\ref XmlRpcArray)  inside the provided \ref XmlRpcMethodValue.
 * 
 * @param value The method value which is expected to have an array
 * inside (\ref XmlRpcArray).
 * 
 * @return An internal reference to the array inside the given \ref
 * XmlRpcMethodValue or NULL if it fails. The returned reference must
 * not be deallocated.
 */
XmlRpcArray       * vortex_xml_rpc_method_value_get_as_array    (XmlRpcMethodValue * value)
{
	/* check environment conditions */
	g_return_val_if_fail (value, NULL);
	g_return_val_if_fail (value->type == XML_RPC_ARRAY_VALUE, NULL);

	/* return an int value */
	return value->value.rpc_array;
}

/** 
 * @brief Allows to create a new \ref XmlRpcMethodResponse object,
 * representing a request reply having either a \ref XmlRpcMethodValue
 * or a error code with a fault string.
 * 
 * Usually this function is not called directly by API consumers. This
 * function is used by the Vortex XML-RPC implementation to create
 * \ref XmlRpcMethodResponse object from data received as a reply or
 * due to a failure to report.
 * 
 * @param status The status code to be used for this method
 * response. If the the status value used is \ref XML_RPC_OK, then the
 * object returned will be a positive reply containing the given \ref
 * XmlRpcMethodValue reference received. In the case the status is not
 * \ref XML_RPC_OK, then the object generated will be a negative
 * reply containing an status code with an error message.
 *
 * @param fault_code The status code to be used according to the
 * value of the <i>status</i> value received.
 *
 * @param fault_string The status message to be set according to the
 * value of the <i>status</i> value received. This value, in the case
 * it is provided, will be copied. Caller could deallocate the string
 * provided just after returning this function.
 *
 * @param value The \ref XmlRpcMethodValue to be set if the given
 * <i>status</i> value is \ref XML_RPC_OK. The reference provided to
 * this function will not be duplicated and the \ref
 * XmlRpcMethodResponse object should be the unique owner of the
 * reference received. Keep in mind that the value received will be
 * deallocated once called to \ref
 * vortex_xml_rpc_method_response_free.
 * 
 * @return A new \ref XmlRpcMethodResponse object that must be
 * deallocated by using \ref vortex_xml_rpc_method_response_free. The
 * function could fail, returning NULL if either a \ref XML_RPC_OK
 * status is used and the value reference is null or the status value
 * is not \ref XML_RPC_OK and the fault_string is null.
 */
XmlRpcMethodResponse * vortex_xml_rpc_method_response_new  (XmlRpcResponseStatus  status,
							    gint                  fault_code,
							    gchar               * fault_string,
							    XmlRpcMethodValue   * value)
{
	XmlRpcMethodResponse * response;
	
	if (status == XML_RPC_OK)
		g_return_val_if_fail (value != NULL, NULL);

	if (status != XML_RPC_OK) 
		g_return_val_if_fail (fault_string != NULL, NULL);
	
	response         = g_new0 (XmlRpcMethodResponse, 1);
	response->status = status;

	/* returning a ok object */
	if (status == XML_RPC_OK)
		response->value = value;

	/* returning a failure object */
	if (status != XML_RPC_OK) {
		response->fault_code   = fault_code;
		response->fault_string = g_strdup (fault_string);
	}
	/* return object created */
	return response;
}

/** 
 * @brief Creates a new positive \ref XmlRpcMethodResponse object,
 * inserting a \ref XmlRpcMethodValue instance created by providing
 * the <b>type</b> and <b>value</b> this function receives.
 * 
 * @param type The value type.
 * @param value The pointer type.
 * 
 * @return A newly allocated \ref XmlRpcMethodResponse object,
 * containing a positive reply, with the a \ref XmlRpcMethodValue
 * inside it. NULL is returned if fail.
 */
XmlRpcMethodResponse * vortex_xml_rpc_method_response_create (XmlRpcParamType type,
							      gpointer        value)
{

	XmlRpcMethodValue    * __value;
	XmlRpcMethodResponse * response;

	/* creates a new method value from received values */
	__value    = method_value_new (type, value);
	g_return_val_if_fail (__value, NULL);

	/* returns current method response value */
	response = method_response_new (XML_RPC_OK, -1, NULL, __value);
	if (response == NULL)
		method_value_free (__value);
	
	/* return current status */
	return response;
	
}

/** 
 * @brief Deallocates the given \ref XmlRpcMethodResponse object.
 *
 * @param response The object to deallocate.
 */
void vortex_xml_rpc_method_response_free (XmlRpcMethodResponse * response)
{
	g_return_if_fail (response);

	/* deallocates internal data */
	if (response->value != NULL)
		vortex_xml_rpc_method_value_free (response->value);

	if (response->string_value != NULL)
		g_free (response->string_value);

	if (response->fault_string != NULL)
		g_free (response->fault_string);
	
	/* deallocates the node itself */
	g_free (response);
	
	return;
}

/** 
 * @brief Allows to perform a marshalling operation into the XML
 * representation for the given \ref XmlRpcMethodResponse object.
 * 
 * @param response The object to marshall into the a XML representation.
 * @param size If the param is not NULL, it will be filled with the message size.
 * 
 * @return A newly allocated string containig the XML representation
 * for the given \ref XmlRpcMethodResponse.
 */
gchar                * vortex_xml_rpc_method_response_marshall         (XmlRpcMethodResponse * response,
									gint                 * size)
{
	gchar * result     = NULL;
	gchar * string_aux = NULL;
	
	/* check received method response */
	g_return_val_if_fail (response, NULL);

	switch (response->status) {
	case XML_RPC_OK:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "ok reply to be marshalled, type=%d", response->value->type);

		/* check that the XmlRpcMethodValue is defined */
		g_return_val_if_fail (response->value, NULL);
		switch (response->value->type) {
		case XML_RPC_INT_VALUE:
			/* implement the int value */
			result = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value><i4>%d</i4></value></param></params></methodResponse>",
						  response->value->value.int_value);
			break;
		case XML_RPC_BOOLEAN_VALUE:
			/* implement the boolean value */
			result = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value><boolean>%d</boolean></value></param></params></methodResponse>",
						  response->value->value.int_value);
			
			break;
		case XML_RPC_DOUBLE_VALUE:
			/* implement the double value */
			result = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value><double>%g</double></value></param></params></methodResponse>",
						  response->value->value.double_value);
			break;
		case XML_RPC_STRING_VALUE:
		case XML_RPC_STRING_REF_VALUE:
			/* implement the string value */
			result = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value><string>%s</string></value></param></params></methodResponse>",
						  response->value->value.string_value);
			break;
		case XML_RPC_DATE_VALUE:
			/* the already not implemented */
			break;
		case XML_RPC_BASE64_VALUE:
		case XML_RPC_BASE64_REF_VALUE:
			/* implemented as string, by performing the
			 * translation automatically */
			result = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value><base64>%s</base64></value></param></params></methodResponse>",
						  response->value->value.string_value);
			break;
		case XML_RPC_STRUCT_VALUE:
			/* marshall the struct */
			string_aux = vortex_xml_rpc_types_marshall_struct (response->value->value.rpc_struct);
			result     = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value>%s</value></param></params></methodResponse>",
						      string_aux);
			g_free (string_aux);
			break;
		case XML_RPC_ARRAY_VALUE:
			/* marshall the array received */
			string_aux = vortex_xml_rpc_types_marshall_array (response->value->value.rpc_array);
			result     = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><params><param><value>%s</value></param></params></methodResponse>",
						      string_aux);
			g_free (string_aux);
			break;
		default:
			/* nothing to do, seems an error */
			return NULL;
			   
		}
		/* reply generated */
		break;
	default:
		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "negative reply to be marshalled");
		/* generate a error reply */
		result = g_strdup_printf ("<?xml version=\"1.0\"?><methodResponse><fault><value><struct><member><name>faultCode</name><value><int>%d</int></value></member><member><name>faultString</name><value><string>%s</string></value></member></struct></value></fault></methodResponse>",
					  response->fault_code, response->fault_string);
		/* negative replies */
		break;
	}
	
	/* get size for the xml content generated */
	if (result != NULL && size != NULL)
		(* size ) = strlen (result);
	
	/* return what we have generated */
	return result;
}

/** 
 * @brief Returns current status for the invocation performed which
 * has return the given \ref XmlRpcMethodResponse.
 * 
 * @param response The method response to get its status.
 * 
 * @return Current Status or \ref XML_RPC_UNKNOWN_ERROR if the given
 * response is NULL.
 */
XmlRpcResponseStatus vortex_xml_rpc_method_response_get_status (XmlRpcMethodResponse * response)
{
	g_return_val_if_fail (response, XML_RPC_UNKNOWN_ERROR);

	return response->status;
}

/** 
 * @brief Allows to get a string representation for the \ref
 * XmlRpcMethodValue inside the given \ref XmlRpcMethodResponse.
 *
 * The string returned must be not deallocated. This is already done
 * while calling to \ref vortex_xml_rpc_method_response_free.
 *
 * This function can only be called if the method response object
 * status is \ref XML_RPC_OK.
 * 
 * @param response The method response to get its internal value as an string.
 * 
 * @return The string internal representation or NULL if the incoming
 * \ref XmlRpcMethodResponse status is not \ref XML_RPC_OK.
 */
gchar * vortex_xml_rpc_method_response_stringify (XmlRpcMethodResponse * response)
{
	g_return_val_if_fail (response, NULL);
	g_return_val_if_fail (response->status == XML_RPC_OK, NULL);

	if (response->string_value != NULL)
		return response->string_value;

	switch (method_value_get_type (response->value)) {
	case XML_RPC_INT_VALUE:
	case XML_RPC_BOOLEAN_VALUE:
		response->string_value = g_strdup_printf ("%d", response->value->value.int_value);
		break;
	case XML_RPC_DOUBLE_VALUE:
		response->string_value = g_strdup_printf ("%g", response->value->value.double_value);
		break;
	case XML_RPC_DATE_VALUE:
		/* not supported yet */
		break;
	case XML_RPC_STRING_VALUE:
		response->string_value = g_strdup (response->value->value.string_value);
		break;
	case XML_RPC_BASE64_VALUE:
		response->string_value = g_strdup (response->value->value.string_value);
		break;
	case XML_RPC_STRUCT_VALUE:
		/* not supported yet */
		break;
	case XML_RPC_ARRAY_VALUE:
		/* not supported yet */
		break;
	default:
		/* nothing to do */
		break;
	}

	/* return current internal string value representation */
	return response->string_value;
}

/** 
 * @brief Allows to get the \ref XmlRpcMethodValue inside the provided \ref XmlRpcMethodResponse.
 *
 * @param response The method response where the value will be
 * extracted. The method response must be a positive reply (\ref
 * XML_RPC_OK).
 * 
 * @return An internal reference to the method value inside (\ref
 * XmlRpcMethodValue) or NULL if it fails. The returned reference must
 * not be deallocated.
 */
XmlRpcMethodValue    * vortex_xml_rpc_method_response_get_value        (XmlRpcMethodResponse * response)
{
	g_return_val_if_fail (response, NULL);
	g_return_val_if_fail (response->status == XML_RPC_OK, NULL);

	/* return method value inside */
	return response->value;
}

/** 
 * @brief Get current fault code from the given \ref XmlRpcMethodResponse object.
 *
 * This function must only be used when the \ref XmlRpcMethodResponse
 * object status is not \ref XML_RPC_OK.
 * 
 * @param response The method response object to get fault code from.
 * 
 * @return The fault code or -1 if fail. 
 */
gint    vortex_xml_rpc_method_response_get_fault_code (XmlRpcMethodResponse * response)
{
	g_return_val_if_fail (response, -1);
	g_return_val_if_fail (response->status != XML_RPC_OK, -1);

	return response->fault_code;
}

/** 
 * @brief Get current fault string from the given \ref XmlRpcMethodResponse object.
 *
 * This function must only be used when the \ref XmlRpcMethodResponse
 * object status is not \ref XML_RPC_OK.
 *
 * @param response The method response object to get fault string from.
 * 
 * @return The fault string. String value returned from this function
 * must not be deallocated.
 */
gchar * vortex_xml_rpc_method_response_get_fault_string (XmlRpcMethodResponse * response)
{
	g_return_val_if_fail (response, NULL);
	g_return_val_if_fail (response->status != XML_RPC_OK, NULL);

	return response->fault_string;
}

/** 
 * @brief Allows to create an empty xml rpc struct.
 *
 * Once created the struct object with this function, a call to the
 * next function will be required to fill it with struct members:
 *
 *  - \ref vortex_xml_rpc_struct_add_member
 *
 * The previous function will require a reference to a \ref
 * XmlRpcStructMember, which could be created using:
 *
 *  - \ref vortex_xml_rpc_struct_member_new
 *
 * An XML-RPC struct could support holding members of any type,
 * including nesting new structs or arrays. 
 *
 * @param count The number of members that the struct will
 * contain. This value must range from 1 up to the desired value. 
 * 
 * @return A newly allocated struct reference that must be deallocated
 * once no longer needed using \ref vortex_xml_rpc_struct_free. 
 */
XmlRpcStruct         * vortex_xml_rpc_struct_new                       (gint count)
{
	XmlRpcStruct * value;

	/* check environment conditions */
	g_return_val_if_fail (count > 0, NULL);

	/* create a new reference */
	value          = g_new0 (XmlRpcStruct, 1);
	value->count   = count;
	value->members = g_new0 (XmlRpcStructMember *, count);

	return value;
}

/** 
 * @brief Allows to get the number of members added to the provided \ref XmlRpcStruct.
 * 
 * @param _struct The struct that is required to returns its member
 * count.
 * 
 * @return The number of members the provided struct has or -1 if it
 * fails.
 */
gint                   vortex_xml_rpc_struct_get_member_count          (XmlRpcStruct * _struct)
{
	g_return_val_if_fail (_struct, -1);

	/* return the number of member items */
	return _struct->count;
}

/** 
 * @brief Checks the provided strings, using as upper limit the
 * provided count, as members for the provided struct.
 * 
 * @param _struct The struct reference to check.
 * @param member_count The number of members to check.
 * 
 * @return TRUE if the provided names are equal on its values and its
 * order for the ones found on the provided struct. Otherwise, FALSE
 * if it fails.
 */
gboolean               vortex_xml_rpc_struct_check_member_names        (XmlRpcStruct * _struct,
									gint           member_count,
									...)
{
	gchar * member_name;
	gint    iterator = 0;
	va_list args;

	g_return_val_if_fail (_struct, FALSE);
	g_return_val_if_fail (_struct->count == member_count, FALSE);

	/* open std arg list */
	va_start (args, member_count);

	/* now check method call type signature */
	while (iterator < member_count) {

		/* get the member name */
		member_name = va_arg (args, gchar *);
		
		/* check the member name */
		if (! axl_cmp (_struct->members[iterator]->name, member_name)) {
			/* close the standard arg */
			va_end (args);

			/* seems that the provided member is not equal */
			return FALSE;
		}
		
		/* next member */
		iterator++;
	}
	
	/* close the standard arg */
	va_end (args);

	return TRUE;
}

/** 
 * @brief Allows to check the member types, for the provided
 * structure, including the order.
 *
 * @param _struct The struct to check its members.
 * @param member_count The member count to check.
 * 
 * @return TRUE if all types received matched with members
 * stored. Otherwise FALSE is returned.
 */
gboolean               vortex_xml_rpc_struct_check_member_types        (XmlRpcStruct * _struct,
									gint           member_count,
									...)
{
	gchar     * member_type;
	gint        iterator = 0;
	gboolean    error = FALSE;
	va_list     args;

	g_return_val_if_fail (_struct, FALSE);
	g_return_val_if_fail (_struct->count == member_count, FALSE);

	/* open std arg list */
	va_start (args, member_count);

	/* now check method call type signature */
	while (iterator < member_count) {

		/* get the member name */
		member_type = va_arg (args, gchar *);

		vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "checking member type for=%s (iterator=%d)",
		       member_type, iterator);

		/* check the member type */
		if (axl_cmp (member_type, "int") && _struct->members[iterator]->value->type != XML_RPC_INT_VALUE)
			error = TRUE;

		if (axl_cmp (member_type, "boolean") && _struct->members[iterator]->value->type != XML_RPC_BOOLEAN_VALUE) {
			error = TRUE;
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "type found: %d != %d", 
			       _struct->members[iterator]->value->type, XML_RPC_BOOLEAN_VALUE);
		}

		if (axl_cmp (member_type, "string") && _struct->members[iterator]->value->type != XML_RPC_STRING_VALUE)
			error = TRUE;

		if (axl_cmp (member_type, "double") && _struct->members[iterator]->value->type != XML_RPC_DOUBLE_VALUE)
			error = TRUE;

		if (axl_cmp (member_type, "base64") && _struct->members[iterator]->value->type != XML_RPC_BASE64_VALUE)
			error = TRUE;

		if (axl_cmp (member_type, "date") && _struct->members[iterator]->value->type != XML_RPC_DATE_VALUE)
			error = TRUE;

		if (axl_cmp (member_type, "struct")                                 && 
		    _struct->members[iterator]->value->type != XML_RPC_STRUCT_VALUE &&
		    _struct->members[iterator]->value->type != XML_RPC_NONE_VALUE )
			error = TRUE;

		if (axl_cmp (member_type, "array")                                  && 
		    _struct->members[iterator]->value->type != XML_RPC_ARRAY_VALUE  && 
		    _struct->members[iterator]->value->type != XML_RPC_NONE_VALUE )
			error = TRUE;

		/* check for error found and return */
		if (error) {
			vortex_log (LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "member type checking for=%s failed (iterator=%d)",
			       member_type, iterator);

			/* close the standard arg */
			va_end (args);

			/* seems that the provided member is not equal */
			return FALSE;
		}

		/* next member */
		iterator++;
	}
	
	/* close the standard arg */
	va_end (args);

	return TRUE;
}

/** 
 * @brief Allows to create a new instance that represents an struct member.
 *
 * The function creates a new struct member which is named by the
 * given name, and holds an \ref XmlRpcMethodValue reference, which
 * could contain any method value supported.
 *
 * The reference created could be deallocated by calling to \ref
 * vortex_xml_rpc_struct_free which not only releases memory allocated
 * by the \ref XmlRpcStruct instance but also for all of its members
 * represented by \ref XmlRpcStructMember.
 *
 * Additionally you can also use \ref vortex_xml_rpc_struct_member_free.
 * 
 * @param name The member name. The function will perform a local copy
 * for the name provided.
 *
 * @param value The \ref XmlRpcMethodValue that will contain the meber
 * being created. The function will create a reference to a newly
 * created \ref XmlRpcStructMember that will own the \ref
 * XmlRpcMethodValue reference provided. No local copy is done for
 * this value.
 * 
 * @return A newly allocated reference to a \ref XmlRpcStructMember
 * that must be deallocated by using \ref
 * vortex_xml_rpc_struct_member_free or let it to be deallocated by
 * calling to \ref vortex_xml_rpc_struct_free, once the result is
 * added as a member to the struct released. The function will return NULL if some parameter received is NULL.
 */
XmlRpcStructMember   * vortex_xml_rpc_struct_member_new                (gchar * name, 
									XmlRpcMethodValue * value)
{
	XmlRpcStructMember * result;

	g_return_val_if_fail (name, NULL);
	g_return_val_if_fail (value, NULL);

	/* create a result reference */
	result        = g_new0 (XmlRpcStructMember, 1);
	result->name  = g_strdup (name);
	result->value = value;
	
	/* return the result created */
	return result;
}


/** 
 * @brief Adds a \ref XmlRpcMethodValue created to the provided \ref XmlRpcStruct instance as a new member.
 *
 * Because an struct (\ref XmlRpcStruct) is only a container for named
 * values (\ref XmlRpcStructMember) called members, this function
 * allows to add more values.
 *
 * Keep in mind that the value is added on the next free position, but
 * this is not important since the XML-RPC standard states that there
 * is no order inside a struct, that is, two structs are equivalent if
 * the have the same members no matter how they are ordered.
 *
 * The function will also check if the member added could be supported
 * by the struct. The struct, at time creation (\ref
 * vortex_xml_rpc_struct_new), is configured to have a particular
 * number of members.
 * 
 * @param _struct The struct (\ref XmlRpcStruct) where the member (\ref XmlRpcStructMember) will be added.
 * 
 * @param member The member to be added.
 */
void                   vortex_xml_rpc_struct_add_member                (XmlRpcStruct * _struct, 
									XmlRpcStructMember * member)
{
	/* some environment checks */
	g_return_if_fail (_struct);
	g_return_if_fail (member);
	g_return_if_fail (_struct->added_count < _struct->count);

	/* add the member */
	_struct->members [_struct->added_count] = member;

	/* increase the cound */
	_struct->added_count++;
	
	/* job done */
	return;
}

/** 
 * @internal
 * 
 * Internal function to get the struct member inside, looking by its name.
 */
XmlRpcStructMember * __vortex_xml_rpc_struct_get_member_by_name (XmlRpcStruct * _struct,
								 char         * member_name)
{
	int iterator = 0;
	XmlRpcStructMember * member;

	/* perform some environment checks */
	g_return_val_if_fail (_struct, NULL);
	g_return_val_if_fail (member_name, NULL);

	/* lookup for the member */
	while (iterator < _struct->added_count) {
		
		/* get a reference to the member */
		member = _struct->members[iterator];

		/* check the member name */
		if (! g_strcasecmp (member->name, member_name))
			return member;

		/* update the iterator */
		iterator++;
	}

	/* return that no member wasn't found */
	return NULL;	
}

/** 
 * @brief Allows to get the integer value associated to the provided
 * member name, inside the provided struct.
 *
 * See also:
 * - \ref vortex_xml_rpc_struct_get_member_value_as_string
 * - \ref vortex_xml_rpc_struct_get_member_value_as_double
 * - \ref vortex_xml_rpc_struct_get_member_value_as_struct
 * - \ref vortex_xml_rpc_struct_get_member_value_as_array
 *
 * @param _struct The struct where the member value is being required.
 *
 * @param member_name The member name to get its value.
 * 
 * @return The integer value requested. The function will always
 * return -1 if some parameter provided is NULL. There is no way to
 * check that a -1 returned is due to an error or due to be the value
 * which is actually stored inside the struct at the provided
 * member. Check that all parameters provided are not NULL to get a
 * properly function.
 */
int                    vortex_xml_rpc_struct_get_member_value_as_int   (XmlRpcStruct       * _struct,
									gchar              * member_name)
{
	XmlRpcStructMember * member;

	g_return_val_if_fail (_struct, -1);
	g_return_val_if_fail (member_name, -1);
	
	/* get the member */
	member = __vortex_xml_rpc_struct_get_member_by_name (_struct, member_name);
	g_return_val_if_fail (member, -1);
	
	/* get the integer inside */
	return vortex_xml_rpc_method_value_get_as_int (member->value);
	
}

/** 
 * @brief Allows to get the string value stored in a selected member at the provided structure.
 * 
 * See also:
 *
 * - \ref vortex_xml_rpc_struct_get_member_value_as_double
 * - \ref vortex_xml_rpc_struct_get_member_value_as_struct
 * - \ref vortex_xml_rpc_struct_get_member_value_as_array
 * - \ref vortex_xml_rpc_struct_get_member_value_as_int
 * 
 * @param _struct The struct where the operation will be performed.
 * @param member_name The member name to look up for the string
 * inside.
 * 
 * @return An internal reference to the string stored inside the
 * member selected. Returned result must not be deallocated. The
 * function will return NULL if the provided parameters are NULL or
 * the member doesn't exist.
 */
gchar *                vortex_xml_rpc_struct_get_member_value_as_string   (XmlRpcStruct       * _struct,
									   gchar              * member_name)
{
	XmlRpcStructMember * member;

	g_return_val_if_fail (_struct, NULL);
	g_return_val_if_fail (member_name, NULL);
	
	/* get the member */
	member = __vortex_xml_rpc_struct_get_member_by_name (_struct, member_name);
	g_return_val_if_fail (member, NULL);
	
	/* get the integer inside */
	return vortex_xml_rpc_method_value_get_as_string (member->value);
}

/** 
 * @brief Return the double value inside the provided struct at the selected member.
 *
 * See also:
 *
 * - \ref vortex_xml_rpc_struct_get_member_value_as_string
 * - \ref vortex_xml_rpc_struct_get_member_value_as_struct
 * - \ref vortex_xml_rpc_struct_get_member_value_as_array
 * - \ref vortex_xml_rpc_struct_get_member_value_as_int
 * 
 *
 * @param _struct The struct where the operation will be performed.
 * @param member_name The member name to look up for the double value
 * inside.
 * 
 * @return Returns the double inside the struct or 0.0 if it fails. 
 */
gdouble                    vortex_xml_rpc_struct_get_member_value_as_double   (XmlRpcStruct       * _struct,
									       gchar              * member_name)
{
	XmlRpcStructMember * member;

	g_return_val_if_fail (_struct, -1);
	g_return_val_if_fail (member_name, -1);
	
	/* get the member */
	member = __vortex_xml_rpc_struct_get_member_by_name (_struct, member_name);
	g_return_val_if_fail (member, -1);
	
	/* get the integer inside */
	return vortex_xml_rpc_method_value_get_as_double (member->value);
}

/** 
 * @brief Allows to get the struct inside the provided member on the provided struct.
 * 
 * See also:
 * - \ref vortex_xml_rpc_struct_get_member_value_as_int
 * - \ref vortex_xml_rpc_struct_get_member_value_as_string
 * - \ref vortex_xml_rpc_struct_get_member_value_as_double
 * - \ref vortex_xml_rpc_struct_get_member_value_as_array
 * 
 * @param _struct The struct where the operation will be performed.
 * @param member_name The member name to look up for the struct
 * inside.
 * 
 * @return An internal reference to the struct stored inside the
 * provided member or NULL if fails. 
 */
XmlRpcStruct              * vortex_xml_rpc_struct_get_member_value_as_struct   (XmlRpcStruct       * _struct,
										gchar              * member_name)
{
	XmlRpcStructMember * member;

	g_return_val_if_fail (_struct, NULL);
	g_return_val_if_fail (member_name, NULL);
	
	/* get the member */
	member = __vortex_xml_rpc_struct_get_member_by_name (_struct, member_name);
	g_return_val_if_fail (member, NULL);

	/* check for none value */
	if (method_value_get_type (member->value) == XML_RPC_NONE_VALUE)
		return NULL;
	
	/* get the integer inside */
	return vortex_xml_rpc_method_value_get_as_struct (member->value);
}

/** 
 * @brief Allows to get the array inside the provided structure, at the select member.
 * 
 * See also:
 *
 * - \ref vortex_xml_rpc_struct_get_member_value_as_int
 * - \ref vortex_xml_rpc_struct_get_member_value_as_string
 * - \ref vortex_xml_rpc_struct_get_member_value_as_struct
 * - \ref vortex_xml_rpc_struct_get_member_value_as_double
 * 
 * @param _struct The struct where the operation will be performed.
 * @param member_name The member name to look up for the array
 * inside.
 * 
 * @return Returns an internal reference to the array stored inside
 * the member or NULL if it fails. The result must not be deallocated.
 */
XmlRpcArray              * vortex_xml_rpc_struct_get_member_value_as_array   (XmlRpcStruct       * _struct,
									      gchar               * member_name)
{
	XmlRpcStructMember * member;

	g_return_val_if_fail (_struct, NULL);
	g_return_val_if_fail (member_name, NULL);
	
	/* get the member */
	member = __vortex_xml_rpc_struct_get_member_by_name (_struct, member_name);
	g_return_val_if_fail (member, NULL);
	
	/* get the integer inside */
	return vortex_xml_rpc_method_value_get_as_array (member->value);
}


/** 
 * @brief Allows to deallocate the provided struct reference (\ref
 * XmlRpcStruct), including all of its members.
 * 
 * @param _struct The struct to be deallocated.
 */
void                    vortex_xml_rpc_struct_free                      (XmlRpcStruct * _struct)
{
	int iterator = 0;

	/* do nothing if a NULL reference is received */
	if (_struct == NULL)
		return;

	/* release items added */
	while (iterator < _struct->added_count) {

		/* release the member pointed by the iterator
		 * position */
		vortex_xml_rpc_struct_member_free (_struct->members[iterator]);

		/* upgrade the iterator */
		iterator++;
	}

	/* free the members array */
	g_free (_struct->members);
	
	/* release the memory hold by the struct itself */
	g_free (_struct);
	
	return;
}

/** 
 * @brief Allows to deallocate the member (\ref XmlRpcStructMember)
 * provided.
 * 
 * @param member The member to release. The function will also release
 * the \ref XmlRpcMethodValue inside using \ref
 * vortex_xml_rpc_method_value_free.
 */
void                    vortex_xml_rpc_struct_member_free               (XmlRpcStructMember * member)
{
	/* do nothing if a NULL reference is received */
	if (member == NULL)
		return;

	/* free the member name */
	g_free (member->name);

	/* free the value inside */
	vortex_xml_rpc_method_value_free (member->value);

	/* free the node itself */
	g_free (member);

	return;
}

/** 
 * @brief Allows to create a new \ref XmlRpcArray reference that will
 * contain the provided number of item.
 *
 * The \ref XmlRpcArray is created to support \ref XmlRpcMethodValue
 * items inside. Several functions are provided to get access to the
 * content and configure it:
 *
 *  - \ref vortex_xml_rpc_array_count
 *  - \ref vortex_xml_rpc_array_max_count
 *  - \ref vortex_xml_rpc_array_get
 *  - \ref vortex_xml_rpc_array_set
 *
 * Once the \ref XmlRpcArray is not required, it must be deallocated
 * by using:
 *
 * - \ref vortex_xml_rpc_array_free
 * 
 * @param count The maximum number of items that the XmlRpcArray will
 * contain. This value is not optional and must be greater (or equal
 * to) 0.
 * 
 * @return A newly allocated reference to a \ref XmlRpcArray.
 */
XmlRpcArray          * vortex_xml_rpc_array_new                        (gint count)
{
	XmlRpcArray * result;

	/* check conditions */
	g_return_val_if_fail (count >= 0, NULL);

	/* create the reference */
	result         = g_new0 (XmlRpcArray, 1);
	result->count  = count;

	/* allocate memory to hold method values */
	result->values = g_new0 (XmlRpcMethodValue *, count);

	/* return the result */
	return result;
}

/** 
 * @brief Deallocates the provided \ref XmlRpcArray reference.
 *
 * @param array The \ref XmlRpcArray array to destroy.
 */
void                   vortex_xml_rpc_array_free                       (XmlRpcArray * array)
{
	gint iterator;

	g_return_if_fail (array);

	/* release all method values inside */
	for (iterator = 0; iterator < array->added_count; iterator++) {
		
		/* release the method value */
		vortex_xml_rpc_method_value_free (array->values[iterator]);
	}
	
	/* release the array */
	g_free (array->values);
	g_free (array);
	
	return;
}

/** 
 * @brief Returns the number of items inside the provided \ref XmlRpcArray.
 *
 * This function will return the actual number of items the array
 * have. This is different from the maximum number of items added to
 * the array. To get the maximum number of items check \ref vortex_xml_rpc_array_max_count).
 * 
 * @param array The array where the number of items will be returned.
 * 
 * @return The nuber of items inside the array. The function will
 * return -1 if it fails.
 */
gint                   vortex_xml_rpc_array_count                      (XmlRpcArray * array)
{
	/* check conditions */
	g_return_val_if_fail (array, -1);

	/* return the result required */
	return array->added_count;

}

/** 
 * @brief Returns the maximum number of items inside the provided
 * array, that is, the number of items that could hold the provided
 * array (the value configured at \ref vortex_xml_rpc_array_new).
 * 
 * @param array The array that is being requested for the maximum
 * count.
 * 
 * @return The number of items that the array could hold. The function
 * will return -1 if it fails.
 */
gint                   vortex_xml_rpc_array_max_count (XmlRpcArray * array)
{
	/* check conditions */
	g_return_val_if_fail (array, -1);

	/* return the result required */
	return array->count;
}

/** 
 * @brief Returns the reference of the \ref XmlRpcMethodValue inside
 * the provided \ref XmlRpcArray, at the provided position.
 * 
 * @param array The array where the data inside, at the given position
 * will be returned.
 *
 * @param index The index being requested. It must be from 0 up to N -
 * 1, where N is the result get from \ref vortex_xml_rpc_array_count.
 * 
 * @return A reference to the \ref XmlRpcMethodValue or NULL if it
 * fails. The reference returned must not be deallocated. It will be
 * deallocated once the \ref vortex_xml_rpc_array_free is called.
 */
XmlRpcMethodValue    * vortex_xml_rpc_array_get                        (XmlRpcArray * array,
									gint index)
{
	/* check the index */
	g_return_val_if_fail (index >= 0 && index < array->added_count, NULL);
	
	/* return the method value */
	return array->values[index];
}

/** 
 * @brief Configures the provided \ref XmlRpcMethodValue at the
 * provided position, for the selected \ref XmlRpcArray.
 *
 * If a method value where previously set, it will be replaced by the
 * new reference provided. In the same context, the old reference will
 * be deallocated.
 *
 * @param array The array to be configured.
 * @param index The index where the value will be placed.
 * @param value The method value to set.
 *
 * 
 */
void                   vortex_xml_rpc_array_set                        (XmlRpcArray * array,
									gint index,
									XmlRpcMethodValue * value)
{
	/* check the index */
	g_return_if_fail (index >= 0 && index < array->count);
	
	/* set the method value */
	array->values[index] = value;
	
	/* return */
	return;
}

/** 
 * @brief Addes the next item (\ref XmlRpcMethodValue) to the provided
 * \ref XmlRpcArray.
 *
 * The function could fail if more items that the maximum items
 * (vortex_xml_rpc_array_new) configured are added.
 * 
 * @param array The \ref XmlRpcArray where the item will be added.
 *
 * @param value The \ref XmlRpcMethodValue that will be added. The
 * reference received will be owned by the \ref XmlRpcArray, so it
 * could be deallocated once called \ref vortex_xml_rpc_array_free.
 *
 */
void                   vortex_xml_rpc_array_add                        (XmlRpcArray * array,
									XmlRpcMethodValue * value)
{
	/* check conditions */
	g_return_if_fail (array->added_count < array->count);

	/* add the item and increase the count */
	array->values [array->added_count] = value;
	array->added_count ++;

	/* nothing more */
	return;
}

/* @} */
