/**
 *  LibVortex:  A BEEP implementation for af-arch
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or   
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vortex.h>
#include <glib.h>
#include <readline/readline.h>
#include <readline/history.h>

/** 
 * @brief Connect used across the vortex client.
 */
VortexConnection * connection;

/** 
 * @brief Next auto tls value to be used
 */
gboolean           auto_tls_profile = TRUE;


void vortex_client_show_help () {
	g_print ("Vortex-client cmd help:\n");
	g_print (" -- general commands --\n");
	g_print ("  help              - Show this help.\n");
	g_print ("  quit              - Quit from tool.\n");
	g_print ("  log               - Enable/Disable vortex console log.\n");
	g_print ("  color log         - Enable/Disable vortex console log using ansi terminal colors.\n");
	g_print ("\n -- connect commands --\n");
	g_print ("  connect           - Connects to a remove vortex (BEEP enabled) server\n");
	g_print ("  open              - The same as connect.\n");
	g_print ("  enable tls        - Try to negociate underlaying TLS profile.\n");
	g_print ("  auto tls          - Enable/Disable automatic TLS profile negociation for new connections\n");
	g_print ("  begin auth        - Start SASL authentication process\n");
	g_print ("  close             - Close connection opened.\n");
	g_print ("\n -- debuging peers --\n");
	g_print ("  show profiles     - Show remote peer profiles, features and localize\n");
	g_print ("                      values.\n");
	g_print ("  channel status    - Returns actual channel status.\n");
	g_print ("  connection status - Returns actual connection status.\n");
	g_print ("\n -- raw frame support --\n");
	g_print ("  write frame       - Write and sends a frame.\n");
	g_print ("  write fragment    - Write and sends a frame but splitting the message\n");
	g_print ("                      into several pieces to test remote frame fragment\n");
	g_print ("                      reading support.\n");
	g_print ("  write start msg   - Sends a raw start channel message and returns\n");
	g_print ("                      the remote peer response to stdout.\n");
	g_print ("  write close msg   - Sends a raw close channel message and returns\n");
	g_print ("                      the remote peer response to stdout.\n");
	g_print ("\n -- interacting with remote peers --\n");
	g_print ("  new channel       - Creates a new channel using the vortex API.\n");
	g_print ("  close channel     - Close a channel using the vortex API.\n");
	g_print ("  send message      - Send a message using the vortex API.\n\n");
	g_print ("\n -- xml-rpc (RFC3529) invocation --\n");
	g_print ("  new invoke        - Creates and launch an XML-RPC invocation\n");
	g_print ("                      against the BEEP peer already connected.\n");

}

void vortex_show_initial_greetings () 
{
	g_print ("Vortex-client v.%s: a cmd tool to test vortex (and BEEP-enabled) peers\n", VERSION);
	g_print ("Copyright (c) 2005 Advanced Software Production Line, S.L.\n");
}

gchar *  get_and_check_cancel () {
	gchar * response = NULL;

	g_print ("  0) cancel process\n");
	response = readline ("you chose: ");
	if (!g_strncasecmp (response, "0", 1)) {
		g_free (response);
		response = NULL;
		g_print ("canceling process..\n");
	}
	return response;
}

VortexFrameType check_frame_type (gchar * response) {

	g_return_val_if_fail (response && (* response), VORTEX_FRAME_TYPE_UNKNOWN);

	if (!g_strncasecmp (response, "1", 1)) {
		return VORTEX_FRAME_TYPE_MSG;
	}
	if (!g_strncasecmp (response, "2", 1)) {
		return VORTEX_FRAME_TYPE_RPY;
	}
	if (!g_strncasecmp (response, "3", 1)) {
		return VORTEX_FRAME_TYPE_ERR;
	}
	if (!g_strncasecmp (response, "4", 1)) {
		return VORTEX_FRAME_TYPE_ANS;
	}
	if (!g_strncasecmp (response, "5", 1)) {
		return VORTEX_FRAME_TYPE_NUL;
	}
	return VORTEX_FRAME_TYPE_UNKNOWN;
}
	

gint32  get_number_int (gchar * ans, gint32 limit) {
	gchar * response;
	gint    num;
 label:
	g_print (ans);
        response  = readline ("you chose: ");
        num       = atoi (response);
	if (num < 0 || num > limit ) {
                g_free (response);
		g_print ("You must provide a valid number from 0 up to %d\n", limit);
		goto label;
	}
        g_free (response);
        response = NULL;
	return num;
}

guint32 get_number_long (gchar * ans, gint64 limit) {
	gchar * response;
	guint32 num;

 label:
	g_print (ans);
	response  = readline ("you chose: ");
	num       = g_strtod (response, NULL);
	if (num < 0 || num > limit ) {
                g_free (response);
		g_print ("You must provide a valid number from 0 up to %llu\n", limit);
		goto label;
	}
        g_free (response);
        response = NULL;
	return num;
}

void vortex_client_write_frame (VortexConnection * connection, gboolean fragment)
{
	gchar            * response          = NULL;
	VortexFrameType    type              = VORTEX_FRAME_TYPE_UNKNOWN;
	gint32             channel_num       = 0;
	gint32             message_num       = 0;
	guint64            sequence_num      = 0;
	gint32             payload_size      = 0;
	gint32             ansno_num         = 0;
	gchar            * payload           = NULL;
	gchar            * str               = NULL;
	gchar            * the_whole_payload = NULL;
	gboolean           continuator;
	gchar            * the_frame;
	gint               iterator;
	gint               length;
	
	g_print ("What type of frame do you want to send?\n");
	g_print ("  1) MSG\n");
	g_print ("  2) RPY\n");
	g_print ("  3) ERR\n");
	g_print ("  4) ANS\n");
	g_print ("  5) NUL\n");
	response = get_and_check_cancel ();
	if (response == NULL)
		return;

	/* get frame type */
	type = check_frame_type (response);
	g_free (response);

	channel_num = get_number_int ("What is the channel to be used to send frame?\n",
				      MAX_CHANNEL_NO);
	
	message_num = get_number_int ("What is the message number to set?\n",
				      MAX_CHANNEL_NO);
	
	g_print ("What message continuator to set?\n");
	g_print ("  1) . (no more message indicator)\n");
	g_print ("  2) * (more messages to come indicator)\n");
	response = get_and_check_cancel ();
	if (response == NULL)
		return;

	if (!g_strncasecmp (response, "1", 1))
		continuator = FALSE;
	else
		continuator = TRUE;
	g_free (response);

	sequence_num = get_number_long ("What is the sequence number to be used?\n", MAX_SEQ_NO);

	ansno_num    = get_number_int  ("What is the ansno number value ?\n", 214783647);

	payload_size = get_number_int  ("What is the payload size?\n", 214783647);

	g_print ("Write the frame payload to send (control-D to finish on last line):\n");
	g_print ("NOTE: every line you introduce, is completed with a \\r\\n (aka CR LF) terminator\n");
	while ((payload = readline (NULL)) != NULL) {
		if (the_whole_payload == NULL) {
			the_whole_payload = g_strdup_printf ("%s\r\n", payload);
			g_free (payload);
			continue;
		}
		str = the_whole_payload;
		the_whole_payload = g_strdup_printf ("%s%s\r\n", the_whole_payload, payload);
		g_free (str);
		g_free (payload);
	}

	g_print ("Building frame..\n");
	the_frame = vortex_frame_build_up_from_params (type,
						       channel_num,
						       message_num,
						       continuator,
						       sequence_num,
						       payload_size,
						       ansno_num,
						       the_whole_payload ? the_whole_payload : "");
	if (the_whole_payload != NULL)
		g_free (the_whole_payload);
	g_print ("frame to send:\n%s\nDo you want to send this frame?\n", the_frame);
	g_print ("  1) yes\n");
	response = get_and_check_cancel ();
	if (response == NULL) 
		return;
	
	g_print ("Sending frame..\n");
	if (fragment) {
		/* send a raw frame (splitting it) */
		iterator = 0;
		length   = strlen (the_frame) / 3;
		while (iterator < length) {
			g_print ("sending from %d to %d bytes..\n", iterator, length);
			if (!vortex_frame_send_raw (connection, (the_frame + iterator), 
						    length)) {
				g_print ("unable to send frame in raw mode (fragments): %s\n", 
					 vortex_connection_get_message (connection));
			}
			
			/* update indexes */
			iterator +=length;
			if ((strlen (the_frame + iterator)) < length)
				length = strlen (the_frame + iterator);

		}
		

	} else {
		/* send a raw frame (complete one)  */
		if (!vortex_frame_send_raw (connection, the_frame, strlen (the_frame))) {
			g_print ("unable to send frame in raw mode: %s\n", 
				 vortex_connection_get_message (connection));
		}
	}
	g_free (the_frame);

	return;
}

void vortex_client_get_channel_status (VortexConnection * connection)
{
	gchar         * response;
	gint            channel_num;
	VortexChannel * channel;

	response    = readline ("Channel number to get status: ");
	channel_num = atoi (response);
	g_free (response);

	channel = vortex_connection_get_channel (connection, channel_num);
	if (channel == NULL) {
		g_print ("No channel %d is actually opened, leaving status operation..\n", channel_num);
		return;
	}
	
	g_print ("Channel %d status is: \n", channel_num);
	g_print ("  Profile definition: \n");
	g_print ("     %s\n", vortex_channel_get_profile (channel));
	g_print ("  Synchronization: \n");
	if (vortex_channel_get_next_msg_no (channel) == 0) 
		g_print ("     Last msqno sent:          no message sent yet\n");
	else
		g_print ("     Last msqno sent:          %d\n",  vortex_channel_get_next_msg_no (channel) - 1);
	
	g_print ("     Next msqno to use:        %d\n",  vortex_channel_get_next_msg_no (channel));
	if (vortex_channel_get_next_expected_msg_no (channel) == 0) 
		g_print ("     Last msgno received:      no message received yet\n");
	else
		g_print ("     Last msgno received:      %d\n",  vortex_channel_get_next_expected_msg_no (channel) - 1);
	g_print ("     Next reply to sent:       %d\n",  vortex_channel_get_next_reply_no (channel));
	g_print ("     Next reply exp. to recv:  %d\n",  vortex_channel_get_next_expected_reply_no (channel));
	g_print ("     Next exp. msgno to recv:  %d\n",  vortex_channel_get_next_expected_msg_no (channel));
	g_print ("     Next seqno to sent:       %u\n",  vortex_channel_get_next_seq_no (channel));
	g_print ("     Next seqno to receive:    %u\n",  vortex_channel_get_next_expected_seq_no (channel));
	
	return;
}

gboolean  check_connected (gchar * error, VortexConnection * connection) {
	
	if ((connection == NULL) || (!vortex_connection_is_ok (connection, FALSE))) {
		g_print ("%s, connect first\n", error);
		return FALSE;
	}
	return TRUE;

}

void vortex_client_write_start_msg (VortexConnection * connection)
{
	GList           * list_aux;
	GList           * profiles;
	gchar           * response;
	gchar           * the_frame;
	gchar           * the_msg;
	gint              profile;
	gint              channel_num;
	gint              iterator = 1;
	VortexChannel   * channel0;
	VortexFrame     * frame;
	WaitReplyData   * wait_reply;
	gint              msg_no;

	/* get profile to use */
	g_print ("This procedure will send an start msg to remote peer. This actually\n");
	g_print ("bypass vortex library channel creation API. This is used for debuging\n");
	g_print ("beep enabled peer responses and its behaviour.\n\n");
		 
	g_print ("Select what profile to use to create for the new channel?\n");
	profiles = vortex_connection_get_remote_profiles (connection);
	g_print ("profiles for this peer: %d\n", g_list_length (profiles));
	list_aux = g_list_first (profiles);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		g_print ("  %d) %s\n", iterator, (gchar*) list_aux->data);
		iterator++;
	}
	response = get_and_check_cancel ();
	if (response == NULL) {
		g_print ("canceling process..\n");
		return;
	}
	profile  = atoi (response);
	g_free (response);

	/* get channel num to use */
	channel_num = get_number_int ("What channel number to create: ", 2147483647);
	
	/* get management channel */
	channel0    = vortex_connection_get_channel (connection, 0);

	/* create the start message */
	the_msg     = g_strdup_printf ("Content-Type: application/beep+xml\x0D\x0A\x0D\x0A<start number='%d'>\x0D\x0A   <profile uri='%s' />\x0D\x0A</start>\x0D\x0A",
				       channel_num, 
				       (gchar *) g_list_nth_data (profiles, profile - 1));
	
	the_frame   = vortex_frame_build_up_from_params (VORTEX_FRAME_TYPE_MSG,
							 0,
							 vortex_channel_get_next_msg_no (channel0),
							 FALSE,
							 vortex_channel_get_next_seq_no (channel0),
							 strlen (the_msg),
							 0,
							 the_msg);

	/* send the message */
	wait_reply = vortex_channel_create_wait_reply ();
	if (!vortex_channel_send_msg_and_wait (channel0, the_msg, strlen (the_msg), &msg_no, wait_reply)) {
		vortex_channel_free_wait_reply (wait_reply);
		g_print ("unable to send start message");
		return;
	}
	
	/* free no longer needed data */
	g_free (the_msg);
	g_free (the_frame);

	/* wait for response */
	g_print ("start message sent, waiting reply..");
	frame = vortex_channel_wait_reply (channel0, msg_no, 
					   wait_reply);

	/* check frame replied */
	if (vortex_frame_get_type (frame) == VORTEX_FRAME_TYPE_RPY) {
		g_print ("channel start ok, message (size: %d)\n%s", 
			 vortex_frame_get_payload_size (frame),
			 vortex_frame_get_payload (frame));
	}else {
		g_print ("channel start failed, message (size: %d)\n%s",
			 vortex_frame_get_payload_size (frame),
			 vortex_frame_get_payload (frame));
	}
	vortex_frame_free (frame);
	return;
}

gboolean vortex_client_on_close_received (gint               channel_num, 
					  VortexConnection * connection, 
					  gpointer           user_data) {

	g_print ("\n*** Close notification received: ***\n");
	g_print ("*** A close notification for the channel: %d over the connection %d) from %s:%s\n",
		 channel_num, vortex_connection_get_id (connection),
		 vortex_connection_get_host (connection),
		 vortex_connection_get_port (connection));
	
	return TRUE;
}

void vortex_client_on_frame_received     (VortexChannel    * channel, 
					  VortexConnection * connection, 
					  VortexFrame      * frame, 
					  gpointer user_data)
{
	GTimeVal  time_val;
	g_get_current_time (&time_val);

	g_print ("\n*** Frame received notification (%ld:%ld): ***\n",
		 time_val.tv_sec, time_val.tv_usec);
	g_print ("*** A frame has been received over the channel: %d inside the connection %d) from %s:%s\n",
		 vortex_channel_get_number (channel), vortex_connection_get_id (connection),
		 vortex_connection_get_host (connection),
		 vortex_connection_get_port (connection));
	if (vortex_frame_get_type (frame) == VORTEX_FRAME_TYPE_ANS) {
		g_print ("*** Ansno number received: %d\n", vortex_frame_get_ansno (frame));
	}

	g_print ("*** its content is:    \n%s",
		 vortex_frame_get_payload (frame));

	return;
}
  	




void vortex_client_new_channel (VortexConnection * connection)
{
	GList         * list_aux;
	GList         * profiles;
	gchar         * response;
	gint            profile;
	gint            channel_num;
	gint            iterator = 1;
	VortexChannel * channel;

	/* get profile to use */
	g_print ("This procedure will request to create a new channel using Vortex API.\n");
	g_print ("NOTE: this procedure will create a channel but, without having implemented.\n");
	g_print ("      the profile selected for the channel, you can only be notified about\n");
	g_print ("      the content of the \"close channel\" and \"frame receive\" events.\n\n");

		 
	g_print ("Select what profile to use to create for the new channel?\n");
	profiles = vortex_connection_get_remote_profiles (connection);
	g_print ("profiles for this peer: %d\n", g_list_length (profiles));
	list_aux = g_list_first (profiles);
	for (; list_aux; list_aux = g_list_next (list_aux)) {
		g_print ("  %d) %s\n", iterator, (gchar*) list_aux->data);
		iterator++;
	}
	response = get_and_check_cancel ();
	if (response == NULL) {
		g_print ("canceling process..\n");
		return;
	}
	profile  = atoi (response);
	g_free (response);

	/* get channel num to use */
	channel_num = get_number_int ("What channel number to create: ", 2147483647);

	g_print ("creating new channel..");
	channel     = vortex_channel_new (connection, 
					  channel_num, 
					  (gchar *) g_list_nth_data (profiles, profile - 1),
					  vortex_client_on_close_received, NULL, 
					  vortex_client_on_frame_received, NULL, 
					  NULL, NULL);
	if (channel == NULL) {
		g_print ("failed\n");
		return;
	}
	g_print ("ok, channel created: %d\n", vortex_channel_get_number (channel));

	return;
}

void vortex_client_write_close_msg (VortexConnection * connection)
{
	g_print ("no implemented yet..\n");
}

void vortex_client_close_channel (VortexConnection * connection)
{
	VortexChannel * channel;
	gint            channel_num;

	g_print ("This procedure will close a channel using the vortex API.\n");
	/* get channel num to use */
	channel_num = get_number_int ("What channel number to close: ", 2147483647);
	
	if (channel_num == 0) {
		g_print ("You can't close channel 0. It will be closed when session (or connection) is closed\n");
		return;
	}

	/* get a channel reference and also check if this channel is
	 * registered by the vortex connetion */
	channel = vortex_connection_get_channel (connection, channel_num);
	if (channel == NULL) {
		g_print ("no channel %d is opened, finishing..\n", channel_num);
		return;
	}

	/* close the channel */
	if (vortex_channel_close (channel, NULL)) 
		g_print ("Channel close: ok\n");
	else
		g_print ("Channel close: failed\n");
	
	return;
}

void vortex_client_send_message (VortexConnection * connection)
		
{
	gint              channel_num;
	VortexChannel   * channel;
	VortexFrame     * frame;
	WaitReplyData   * wait_reply   = NULL;
	gchar           * msg_to_send;
	gchar           * wait;
	gboolean          do_a_wait;
	gboolean          result;
	gint              msg_no;

	g_print ("This procedure will send a message using the vortex API.\n");
	
	/* get the channel number to use and check this message exists */
	channel_num = get_number_int  ("What channel do you want to use to send the message? ",
				       MAX_CHANNEL_NO);
	
	if (!vortex_connection_channel_exists (connection, channel_num)) {
		g_print ("This channel doesn't exists..\n");
		return;
	}

	/* get the message to send */
	msg_to_send = readline ("Type the message to send:\n");
	if (msg_to_send == NULL) {
		g_print ("no message to send..\n");
		return;
	}

	/* get if user wants to wait for the message */
	g_print ("Do you want to wait for message reply? (Y/n) ");
	wait  = readline (NULL);
	do_a_wait = ((wait != NULL) && ((strlen (wait) == 0) || !g_strcasecmp ("Y", wait) || !g_strcasecmp ("y", wait)));
	if (wait != NULL)
		g_free (wait);

	/* send the message */
	channel             = vortex_connection_get_channel (connection, channel_num);
	if (do_a_wait) {
		wait_reply  = vortex_channel_create_wait_reply ();
		result      = vortex_channel_send_msg_and_wait (channel, msg_to_send, 
								strlen (msg_to_send), &msg_no, wait_reply);
	}else {
		result      = vortex_channel_send_msg (channel, msg_to_send, strlen (msg_to_send), &msg_no);
	}
	
	/* check result */
	if (!result) {
		g_print ("Unable to send the message\n");
		g_free (msg_to_send);
		if (do_a_wait)
			vortex_channel_free_wait_reply (wait_reply);
		return;
	}
	g_free (msg_to_send);
	g_print ("Message number %d was sent..\n", msg_no);

	/* do a wait if posible */
	if (do_a_wait) {
		g_print ("waiting for reply...");
		frame = vortex_channel_wait_reply (channel, msg_no, wait_reply);
		g_print ("reply received: \n%s\n", vortex_frame_get_payload (frame));
		vortex_frame_free (frame);
		return;
	}
	g_print ("leaving without waiting..\n");
	return;
}

void __vortex_client_connection_status (gpointer key, gpointer value, gpointer user_data)
{
	VortexChannel * channel = value;

	g_print ("  channel: %d, profile: %s\n", 
		 vortex_channel_get_number (channel), 
		 vortex_channel_get_profile (channel));

	return;
}

void vortex_client_connection_status (VortexConnection * connection)
{
	gchar * authid = vortex_sasl_get_propertie (connection, VORTEX_SASL_AUTH_ID);

	g_print ("Connection status:\n   Id=%d\n   TLS status=%s\n   SASL auth status=%s, auth id=%s\n", 
		 vortex_connection_get_id (connection),
		 vortex_connection_is_tlsficated (connection) ? "On" : "Off",
		 vortex_sasl_is_authenticated    (connection) ? "On" : "Off",
		 authid ? authid : "none");

	g_print ("Created channel over this session:\n");

	vortex_connection_foreach_channel (connection, __vortex_client_connection_status, NULL);
	
	return;
}

void vortex_client_begin_auth () {
	gint      profile;
	gchar   * profile_selected = NULL;
	gchar   * response;

	/* SASL status notification */
	VortexStatus       status;
	gchar            * status_message;

	/* check for connection status */
	if (!check_connected ("can't start SASL auth if not connected first", connection))
		return;
	
	/* check for SASL support */
	if (!vortex_sasl_is_enabled ()) {
		g_print ("Unable to begin SASL negociation. Current Vortex Library doesn't support SASL");
		return;
	}
	
	g_print ("Choose a profile to negociate: \n");
	g_print ("1) %s\n", VORTEX_SASL_ANONYMOUS);
	g_print ("2) %s\n", VORTEX_SASL_EXTERNAL);
	g_print ("3) %s\n", VORTEX_SASL_PLAIN);
	g_print ("4) %s\n", VORTEX_SASL_CRAM_MD5);
	g_print ("5) %s\n", VORTEX_SASL_DIGEST_MD5);
	g_print ("6) %s\n", VORTEX_SASL_GSSAPI);
	g_print ("7) %s\n", VORTEX_SASL_KERBEROS_V4);
	g_print ("0) Cancel\n");
	profile = get_number_int ("Which profile should I use to authenticate?\n", 7);
	switch (profile) {
	case 0:
		/* do nothing */
		break;
	case 1:
		/* anonymous mech */
		profile_selected = VORTEX_SASL_ANONYMOUS;
		response = readline ("Which is the anonymous token to be used (example: anonymous@foobar.com): ");
		vortex_sasl_set_propertie (connection, VORTEX_SASL_ANONYMOUS_TOKEN,
					   response, g_free);
		break;
	case 2:
		/* external mech */
		profile_selected = VORTEX_SASL_EXTERNAL;
		response = readline ("Which is the auth id to be used (example: bob): ");
		vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTHORIZATION_ID,
					   response, g_free);
		break;
	case 3:
		/* plain mech */
		profile_selected = VORTEX_SASL_PLAIN;

	get_user_and_password_data:

		/* get auth id to be used */
		response = readline ("Which is the auth id to be used (example: bob): ");
		vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTH_ID,
					   response, g_free);

		/* get proxy auth id to be used */
		if (!g_strcasecmp (profile_selected, VORTEX_SASL_PLAIN) ||
		    !g_strcasecmp (profile_selected, VORTEX_SASL_DIGEST_MD5)) {
			response = readline ("Which is the auth id proxy to be used (example: alice): ");
			vortex_sasl_set_propertie (connection, VORTEX_SASL_AUTHORIZATION_ID,
						   response, g_free);
		}

		/* get password to be used */
		response = readline ("Which password to be used (example: secret): ");
		vortex_sasl_set_propertie (connection,  VORTEX_SASL_PASSWORD,
					   response, g_free);

		/* get realm to be used */
		if (!g_strcasecmp (profile_selected, VORTEX_SASL_DIGEST_MD5)) {
			response = readline ("Which is the realm to be used (example: aspl.es): ");
			vortex_sasl_set_propertie (connection, VORTEX_SASL_REALM, response, g_free);
		}
		
		break;
	case 4:
		/* cram md5 mech (same as plain) */
		profile_selected = VORTEX_SASL_CRAM_MD5;
		goto get_user_and_password_data;
		break;
	case 5:
		/* digest md5 mech */
		profile_selected = VORTEX_SASL_DIGEST_MD5;
		goto get_user_and_password_data;
		break;
	case 6:
		/* gssapi mech */
		profile_selected = VORTEX_SASL_GSSAPI;
		g_print ("No supported yet!\n");
		return;
	case 7:
		/* kerberos v4 mech */
		profile_selected = VORTEX_SASL_KERBEROS_V4;
		g_print ("No supported yet!\n");
		return;
	}

	/* begin SASL negociation */
	vortex_sasl_start_auth_sync (connection, profile_selected, &status, &status_message);

	/* print error */
	switch (status) {
	case VortexOk:
		g_print ("OK: SASL successfully negociated, message reported: %s\n", 
			 status_message);
		break;
	case VortexError:
		g_print ("FAIL: There was an error for SASL negociation, message reported: %s\n",
			 status_message);
		break;
	}

	return;
}

/** 
 * @brief Perform an XML-RPC invocation on the given connection.
 * 
 * @param connection The connection where the XML-RPC invocation will
 * take place.
 */
void vortex_client_begin_xml_rpc_invocation (VortexConnection * connection)
{
	XmlRpcMethodCall     * invocator;
	XmlRpcMethodResponse * response;
	XmlRpcMethodValue    * value;
	VortexChannel        * channel;
	VortexStatus           status;
	gchar                * message;

	/* method data */
	gchar                * method_name;
	gchar                * method_count;
	gchar                * param_type;
	gchar                * param_value;

	gint                   count;
	gint                   iterator;

	/* create the XML-RPC channel in a synchronous way */
	channel = vortex_xml_rpc_boot_channel_sync (connection, NULL, NULL, &status, &message);

	/* show message returned */
	if (status == VortexError) {
		g_print ("Unable to create XML-RPC channel, message was: \n   %s\n", message);
		return;
	}else {
		g_print ("XML-RPC channel created, message was: \n   %s\n", message);
	}


	
	/* create invocator object */
	method_name  = readline ("method name to invoke: ");
	method_count = readline ("method count to invoke: ");
	count        = atoi (method_count);
	invocator    = method_call_new (method_name, count);
	
	g_free (method_name);
	g_free (method_count);

	/* configure method parameters */
	iterator = 0;
	while (iterator < count) {
		
		/* get the method type */
		param_type  = readline ("method parameter type (int, string, boolean): ");
		param_value = readline ("method parameter value: ");
		
		/* create the method value */
		value = vortex_xml_rpc_method_value_new_from_string2 (param_type, param_value);
		
		/* release the values allocated */
		g_free (param_type);
		g_free (param_value);

		if (value == NULL) {
			g_print ("an error was detected while building the parameter\n");
			continue;
		}

		/* add the method value to the invocator */
		method_call_add_value (invocator, value);

		/* updte the iterator */
		iterator++;
	}				       

	/* peform a synchronous method */
	g_print ("   Performing XML-RPC invocation..\n");
	response = vortex_xml_rpc_invoke_sync (channel, invocator);

	switch (method_response_get_status (response)) {
	case XML_RPC_OK:
		g_print ("Reply received ok, result is: %s\n", method_response_stringify (response));
		break;
	default:
		g_print ("RPC invocation have failed: (code: %d) : %s\n",
			 method_response_get_fault_code (response),
			 method_response_get_fault_string (response));
		break;
	}

	/* free the response received */
	method_response_free (response);

	return;
}


/** 
 * @brief Support function to __vortex_connection_has_uri_from that
 * finally extract uri information into the provided values.
 */
gboolean __vortex_connection_has_uri_from_aux (gchar * uri, gchar ** host, gchar ** port, gboolean * activate_tls)
{
	gchar    ** pieces     = NULL;
	gchar    ** host_part  = NULL;
	gboolean    result     = FALSE;

	/* first check if we have an uri from using :// */
	pieces = g_strsplit (uri, "://", 2);
	if (pieces == NULL || pieces [1] == NULL) {
		/* we don't have a :// so, the uri received is the form host[:port] */
		host_part = g_strsplit (uri, ":", 2);
		if (host_part == NULL || host_part [1] == NULL) {
			/* no host part */
			(* host ) = g_strdup (uri);
		}else {
			/* host and port part */
			(* host)  = g_strdup (host_part[0]);
			(* port)  = g_strdup (host_part[1]);
		}

		/* free and flag that the values where read */
                if (host_part != NULL)
		        g_strfreev (host_part);
		result = TRUE;
	}else {
		/* we have a :// specification, use profile
		 * information to figure out which is the port to
		 * use */
	} 

        /* free pieces memory only if allocated */
        if (pieces != NULL)
	       g_strfreev (pieces);

	return result;
}

/** 
 * @brief Helper function to the main loop that allows to check if the
 * given command containis a URI value specifying host and port to
 * connect.
 *
 * URI values allows are the following:
 *  
 *  - host
 *  - host:port
 *  - beep://host:port
 *  - beeps://host:port
 *  - xmlrpc.beep://host:port
 *  - xmlrpc.beeps://host:port
 * 
 * All previous URI specifications are resumed into the following:
 * 
 *  - [profile.] [beeps:// | beep://] host [:port]
 * 
 * That is, an optional profile implementation, an optional beep
 * protocol specification, that is required is the profile part is
 * used, followed by the compulsory host name and ended by the
 * optional port value.
 *  
 * @param line The command line where the URI could be found.
 * 
 * @param host Optional reference where the host information will be
 * stored if a URI is found.
 *
 * @param port Optional reference where the port information will be
 * stored if a URI is found.
 *
 * @param activate_tls Optional reference where the TLS activation is
 * required to be negociated at connection time.
 *
 * 
 * @return TRUE if the URI is activated or FALSE if the data must be
 * required.
 */
gboolean __vortex_connection_has_uri_from (gchar * line, 
					   gchar ** host, 
					   gchar ** port, 
					   gboolean * activate_tls)
{
	gchar ** line_pieces = NULL;
	gint     iterator    = 0;
	gboolean result      = FALSE;

        /* set default values for the host and the port */
        (* host ) = NULL;
        (* port ) = NULL;

	/* get current pieces */
	line_pieces  = g_strsplit (line, " ", 0);
	
	/* show all pieces found */
        if (line_pieces != NULL) {
		for (iterator = 0; line_pieces[iterator]; iterator++) {
			/* skip empty content */
			if ((line_pieces[iterator] == NULL) || (strlen (line_pieces[iterator]) == 0))
				continue;

			/* skip connect keyword */
			if (!g_strcasecmp ("connect", line_pieces[iterator]))
				continue;

			/* use the information found to get URI
			 * information */
			result = __vortex_connection_has_uri_from_aux (line_pieces[iterator], host, port, activate_tls);
		}
        }

	/* release memory */
	g_strfreev (line_pieces);

	/* report that no URI was found */
	return result;
}

int main (int argc, char *argv[])
{
        gchar            * host         = NULL;
        gchar            * port         = NULL;
        gboolean           activate_tls = FALSE;
        GList            * profiles     = NULL;
	gchar            * line         = NULL;
	
	/* TLS status notification */
	VortexStatus       status;
	gchar            * status_message;

	/* init vortex library */
	if (!vortex_init ()) {
		g_print ("ERROR: unable to initialize vortex library");
		exit (-1);
	}

	vortex_show_initial_greetings ();

	while (TRUE) {

		/* unref previous command */
		if (line != NULL)
			g_free (line);
		line = NULL;
		
		if ((vortex_connection_is_ok (connection, FALSE)))
			line = readline ("[===] vortex-client > ");
		else
			line = readline ("[=|=] vortex-client > ");

		if (line == NULL) {
			continue;
		}
		
		/* add the command read to the history */
		add_history (line);

		if (!g_strncasecmp ("help", line, 4)) {
			vortex_client_show_help ();
			continue;
		}

		if (!g_strncasecmp ("log", line, 3)) {
			vortex_log_enable (! vortex_log_is_enabled () );
			g_print (" vortex log is: %s\n", vortex_log_is_enabled () ? "enabled" : "disabled");
			continue;
		}

		if (!g_strncasecmp ("color log", line, 9)) {
			vortex_color_log_enable (! vortex_color_log_is_enabled ());

			g_print (" vortex color log is: %s\n", vortex_color_log_is_enabled () ? "enabled" : "disabled");
			continue;
		}

		if (!g_strncasecmp ("show profiles", line, 13)) {
			if (!check_connected ("can't show remote peer profiles", connection))
				continue;

			/* check for supported profiles */
			g_print ("Supported remote peer profiles:\n");
			profiles = vortex_connection_get_remote_profiles (connection);
			for (profiles = g_list_first (profiles) ;profiles; profiles = g_list_next (profiles)) {
				g_print ("  %s\n", (gchar *) profiles->data);
			}

			/* check for features requested by remote peer */
			if (vortex_connection_get_features (connection)) {
				g_print ("Features:\n  %s\n", vortex_connection_get_features (connection));
			}

			/* check for localize requested by remote peer */
			if (vortex_connection_get_features (connection)) {
				g_print ("Localize:\n  %s\n", vortex_connection_get_localize (connection));
			}
			
			continue;
		}

		if (!g_strncasecmp ("connection status", line, 17)) {
			if (!check_connected ("can't show connection status", connection))
				continue;
			
			vortex_client_connection_status (connection);
			
			continue;
		}

		if (!g_strncasecmp ("auto tls", line, 8)) {
			/* enable auto tls profile negociation not allowing TLS failures */
			vortex_connection_set_auto_tls (auto_tls_profile, FALSE, NULL);
			g_print ("Auto TLS profile negociation is: %s\n", auto_tls_profile ? "ON" : "OFF");
			auto_tls_profile = !auto_tls_profile;
			continue;
		}

		if (!g_strncasecmp ("enable tls", line, 10)) {
			if (!check_connected ("can't enable TLS if not connected first", connection))
				continue;

			/* initialize and check if current vortex library supports TLS */
			if (! vortex_tls_is_enabled ()) {
				g_print ("Unable to activate TLS, Vortex Library is not prepared\n");
				continue;
			}

			/* enable TLS negociation */
			connection = vortex_tls_start_negociation_sync (connection, NULL, 
									&status,
									&status_message);

			g_print ("\nTLS Negociation status: %s, message=%s\n",
				 (status == VortexOk) ? "Ok" : "Error",
				 (status_message != NULL) ? status_message : "none");

			
			continue;
		}

		if (!g_strncasecmp ("begin auth", line, 10)) {
			vortex_client_begin_auth ();
			continue;
		}

		if (!g_strncasecmp ("open", line, 4) ||
		    !g_strncasecmp ("connect", line, 7)) {

			/* close previous  */
			if (vortex_connection_is_ok (connection, FALSE))
				vortex_connection_close (connection);
			
			/* check if the user has specified a host:port
			 * value */
			if (!__vortex_connection_has_uri_from (line, &host, &port, &activate_tls)) {
				
				/* get user data and connect to vortex server */
				host = readline ("server to connect to: ");
				if (host == NULL) {
					g_print ("Wrong value received for host, connection cancelled");
					continue;
				}
                        }

                        /* if the port is not properly set, ask for it */
			if (port == NULL) {
				port = readline ("port to connect to: ");
                                if (port == NULL)
                                        port = g_strdup ("44000");
			}
			
                        g_print ("connecting to %s%s%s..", host, (port != NULL) ? ":" : "", (port != NULL) ? port : "");

			/* create a vortex session */
			connection = vortex_connection_new (host, port, NULL, NULL);

			/* check if connection is ok */
			if (!vortex_connection_is_ok (connection, FALSE)) {
				
				/* check for printing a message */
				if ((host != NULL) && (port != NULL)) {
					g_print ("Unable to connect to vortex server on: %s:%s vortex message was: %s\n",
							 host, port, vortex_connection_get_message (connection));
				} else {
					g_print ("Unable to connect using given host and port data.\n");

				}
				/* not needed resources */
				vortex_support_free (2, host, g_free, port, g_free);
				vortex_connection_close (connection);
				connection = NULL;
				continue;
			}
			/* not needed resources */
			vortex_support_free (2, host, g_free, port, g_free);

			g_print ("ok: vortex message: %s\n", vortex_connection_get_message (connection));
			continue;
		}

		if (!g_strncasecmp ("close channel", line, 13)) {
			if (!check_connected ("can't close channel", connection))
				continue;
			g_print ("closing the channel..\n");
			vortex_client_close_channel (connection);
			continue;
		}

		if (!g_strncasecmp ("quit", line, 4)) {
			if (vortex_connection_is_ok (connection, FALSE)) {
				if (!vortex_connection_close (connection)) {
					g_print ("there was an error while closing the connection..");
				}
			}
			
			/* free the line */
			g_free (line);
			
			/* readline clean up */
			clear_history ();	

			/* finish vortex client */
			vortex_exit ();

			/* exit from this program */
			return 0;
		}

		if (!g_strncasecmp ("close", line, 5)) {
			if (!check_connected ("can't close, not connected", connection))
				continue;
			
			if (!vortex_connection_close (connection)) {
				g_print ("Unable to close the channel");
			}else {
				connection = NULL;
			}
			continue;
		}

		if (!g_strncasecmp ("write frame", line, 11)) {
			if (!check_connected ("can't write and send a frame", connection))
				continue;

			vortex_client_write_frame (connection, FALSE);
			continue;
		}

		if (!g_strncasecmp ("channel status", line, 14)) {
			if (!check_connected ("can't get channel status", connection))
				continue;

			vortex_client_get_channel_status (connection);
			continue;
		}

		if (!g_strncasecmp ("write start msg", line, 15)) {
			if (!check_connected ("can't send start msg", connection))
				continue;

			vortex_client_write_start_msg (connection);
			continue;
		}

		if (!g_strncasecmp ("new channel", line, 11)) {
			if (!check_connected ("can't create new channel", connection))
                                continue;

			vortex_client_new_channel (connection);
			continue;
		}

		if (!g_strncasecmp ("write close msg", line, 15)) {
			if (!check_connected ("can't send close msg", connection))
				continue;

			vortex_client_write_close_msg (connection);
			continue;
		}

		if (!g_strncasecmp ("send message", line, 12)) {
			if (!check_connected ("can't send a message", connection))
				continue;
			
			vortex_client_send_message (connection);
			continue;
		}

		if (!g_strncasecmp ("write fragment", line, 14)) {
			if (!check_connected ("can't send a message", connection))
				continue;
			
			vortex_client_write_frame (connection, TRUE);
			continue;
		}
		if (!g_strncasecmp ("new invoke", line, 10)) {
			if (!check_connected ("can't perform a XML-RPC invocation if not connected first", connection))
				continue;

			vortex_client_begin_xml_rpc_invocation (connection);
			continue;
		}

		/* show help if a command was not recognized */
		if (line != NULL && strlen (line) >= 1) {
			g_print ("Command '%s' not recognized..\n", line);
			vortex_client_show_help ();
		}
	}

	/* finaly free connection */
	if (vortex_connection_is_ok (connection, FALSE))
		vortex_connection_close (connection);
	
	return (0);
}

