#include <vortex.h>

/** 
 * @brief Validation resource function. 
 * 
 * This function is called to validate the resource
 * requested. Resources are a way to organize remote method to be
 * invoked, grouped under a common resource name.
 *
 * Default resource used is "/". 
 * 
 * @param connection The connection where the validation request was
 * received.
 *
 * @param channel_number The channel number that is being requested.
 *
 * @param serverName The serverName requested while validating
 *
 * @param resource_path Resource path received. This could be any
 * value, including path.
 * 
 * @return 
 */
gboolean validate_resource (VortexConnection * connection, 
			    gint               channel_number,
			    gchar            * serverName,
			    gchar            * resource_path)
{

	g_print ("Received request to validate resource: '%s' (virtual host: serverName='%s')\n", 
		 resource_path,
		 serverName != NULL ? serverName : "no serverName value received");
	return TRUE;
}

/** 
 * @brief Final invocation function, that perform the sum operation.
 * 
 * @param a First parameter.
 *
 * @param b Second parameter.
 *
 * @param fault_error Optional fault error to be filled in the case an
 * error is found.
 *
 * @param fault_code Optinonal fault error code to be filled in the
 * case an error is found.
 * 
 * @return Return a int value.
 */
gint sum_2_int_int (gint a, gint b, gchar ** fault_error, gint * fault_code)
{
	if (a == 2 && b == 7) {
		/* the function couldn't perform a sum operation for
		 * the 2 and 7 values. */
		REPLY_FAULT ("Current implementation is not allowed to sum the 2 and 7 values", -1, 0);
	}
	
	/* for the rest of cases, just sum the two incoming values */
	return a + b;
}

/** 
 * @brief Method invocation for the sum method, having 2 parameter
 * that are int values.
 * 
 * @param method_call The method call received, representing the sum
 * method.
 * 
 * @return A newly allocated \ref XmlRpcMethodResponse, having current reply.
 */
XmlRpcMethodResponse * __sum_2_int_int (XmlRpcMethodCall * method_call)
{
	/* reply value */
	gint                result;
	gchar             * fault_error = NULL;
	gint                fault_code  = -1;

	/* get parameters */
	gint a = method_call_get_param_value_as_int (method_call, 0);
	gint b = method_call_get_param_value_as_int (method_call, 1);

	/* perform invocation */
	result = sum_2_int_int (a, b, &fault_error, &fault_code);
	
	/* check error reply looking at the fault_error */
	if (fault_error != NULL) {
		/* we have a error reply */
		return CREATE_FAULT_REPLY (fault_code, fault_error);
	}

	/* return reply generated */
	return CREATE_OK_REPLY (XML_RPC_INT_VALUE, GINT_TO_POINTER (result));
	
}

/** 
 * @brief Process method call invocation, without unrefereing the \ref
 * XmlRpcMethodCall object.
 * 
 * @param channel The channel where the method invocation was received.
 * @param method_call The method call received.
 * @param user_data User space data.
 */
XmlRpcMethodResponse *  service_dispatch (VortexChannel * channel, XmlRpcMethodCall * method_call, gpointer user_data) 
{
	
	g_print ("*** channel %d, method call received %s (params %d)***\n",
		 vortex_channel_get_number (channel),
		 method_call_get_name (method_call), 
		 method_call_get_num_params (method_call));

	/* check if the incoming method call is called sum, and has
	 * two arguments */
	if (method_call_is (method_call, "sum", 2, XML_RPC_INT_VALUE, XML_RPC_INT_VALUE, -1)) {
		g_print ("*** sum method found... ***\n");
		return __sum_2_int_int (method_call);
		
	}
	       
	/* return that the method to be invoked, is not supported */
	return CREATE_FAULT_REPLY (-1, "Method call received couldn't be dispatched because it not supported by this server");
}


gint main (gint argc, gchar ** argv) 
{

	/* init vortex library */
	vortex_init ();

	/* enable XML-RPC profile */
	vortex_xml_rpc_accept_negociation (validate_resource,
					   /* no user space data for
					    * the validation resource
					    * function. */
					   NULL, 
					   service_dispatch,
					   /* no user space data for
					    * the dispatch function. */
					   NULL);

	/* create a vortex server */
	vortex_listener_new ("0.0.0.0", "44000", NULL, NULL);

	/* also listener on the following port */
	vortex_listener_new ("0.0.0.0", "42000", NULL, NULL);

	/* wait for listeners (until vortex_exit is called) */
	vortex_listener_wait ();
	
	/* end vortex function */
	vortex_exit ();

	return 0;
}

