/**
 *  LibExploreArguments: a library to parse command line options
 *  Copyright (C) 2005 Advanced Software Production Line, S.L.
 * 
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 *  
 *  You may find a copy of the license under this software is released
 *  at COPYING file. This is LGPL software: you are wellcome to
 *  develop propietary applications using this library withtout any
 *  royalty or fee but returning back any change, improvement or
 *  addition in the form of source code, project image, documentation
 *  patches, etc. 
 *
 *  Contact us at:
 *          
 *      Postal address:
 *         Advanced Software Production Line, S.L.
 *         C/ Dr. Michavila Nº 14
 *         Coslada 28820 Madrid
 *         Spain
 *
 *      Email address:
 *         info@aspl.es - http://fact.aspl.es
 *
 */
#include <exarg.h>

#define LOG_DOMAIN "exarg"

GHashTable * argument_options          = NULL;
GList      * params                    = NULL;
gchar      * exarg_exec_name           = NULL;
gchar      * __exarg_usage_header      = NULL;
gchar      * __exarg_help_header       = NULL;
gchar      * __exarg_post_usage_header = NULL;
gchar      * __exarg_post_help_header  = NULL;

GStaticMutex __exarg_destroy_mutex     = G_STATIC_MUTEX_INIT;

/**
 * \mainpage LibExArg (Explore Arguments Library)  Manual
 * \section Introduction
 *
 *  LibExArg library is a small, easy to use with a consistent API
 *  which allows you to parse command line argument for your program.
 *
 *  It is build on top of glib which make it really portable and at
 *  this point is running on <b>GNU/Linux</b> platforms and
 *  <b>Microsoft Windows</b> platforms.
 *
 *  For details on how to starting to use LibExarg, you can find the
 *  \ref exarg "module documentation for LibExArg here".
 *
 * \section licence Library License
 *
 *  The library is released under the terms of the GNU LGPL (The GNU
 *  Lesser General Public License) licence. You can get a copy of the
 *  license http://www.gnu.org/licenses/licenses.html#LGPL.
 *
 *  This license allows you to write applications using LibExArg which
 *  are using some GPL compatible license or commercial applications.
 *
 *  You can find \ref exarg_example "here an example about using LibExArg".
 */

/**
 * \page exarg_example A simple example on using LibExarg
 *
 * \section Introduction
 *
 *  You may find interesting the following example which makes use of
 *  LibExArg. This example install 3 arguments to be accepted as valid
 *  command line options and the write down to the terminal the result
 *  of those command used by the user.
 * 
 *  Once the arguments are installed, the user can execute 
 *
 *
 *  \include "test-exarg.c"
 * 
 */

/** \defgroup exarg ExArg module: functions to parse command line argument in your program
 *
 */


/**
 * \addtogroup exarg
 * @{ 
 */




typedef struct _ExArgNodeOption {
	gchar     * arg_name;
	gchar     * arg_short_name;
	ExArgType   type;
	gchar     * description;
	gchar     * string_value;
	gint        int_value;
	gboolean    is_defined;
}ExArgNodeOption;


/**
 * @internal
 * Type definition used at \ref exarg_lookup_node function.
 */
typedef struct _ExArgLookUpNode {
	ExArgNodeOption * result;
	gchar           * string_searched;
}ExArgLookUpNode;


/** 
 * @internal
 * @brief Support function for \ref exarg_lookup_node
 *
 * 
 * @param key 
 * @param value 
 * @param user_data 
 */
void __exarg_lookup_node (gpointer key,
			  gpointer value,
			  gpointer user_data)
{
	ExArgNodeOption * node        = (ExArgNodeOption *) value;
	ExArgLookUpNode * lookup_node = (ExArgLookUpNode *) user_data;
	
	/* avoid lookup up */
	if (lookup_node->result != NULL)
		return;

	/* check value with long argument form */
	if (!g_strcasecmp (lookup_node->string_searched, node->arg_name)) {
		lookup_node->result = node;
		return;
	}

	/* check value with short argument form */
	if ((node->arg_short_name != NULL) && !g_strncasecmp (lookup_node->string_searched, node->arg_short_name, strlen (lookup_node->string_searched))) {
		lookup_node->result = node;
		return;
	}
	return;
}


/**
 * @internal
 * @brief Perfom a look up inside the argument option installed.
 *
 * This function tries to return the ExArgNodeOption using the given
 * value as index key. 
 *
 * The given value can represent the long key option or the short
 * form.
 * 
 * @param arg_name the key to use on lookup process
 *
 * @return the function retuns the argument option if found or NULL if
 * not.
 */
ExArgNodeOption * exarg_lookup_node (gchar * arg_name)
{
	ExArgNodeOption * result      = NULL;
	ExArgLookUpNode * lookup_node = NULL;

	g_return_val_if_fail (arg_name && (* arg_name), NULL);

	/* prepare search */
	lookup_node                  = g_new0 (ExArgLookUpNode, 1);
	lookup_node->string_searched = arg_name;

	g_hash_table_foreach (argument_options, __exarg_lookup_node, lookup_node);

	/* end search */
	result = lookup_node->result;
	g_free (lookup_node);

	/* return the */
	return result;
}




gboolean exarg_is_argument (gchar * argument) 
{

	if (!g_strncasecmp (argument, "--", 2)) {
		return TRUE;
	}
	if ((strlen (argument) == 2) && (argument[0] == '-')) {
		return TRUE;
	}
	return FALSE;
}

gchar * exarg_get_argument (gchar * argument)
{
	if (!g_strncasecmp (argument, "--", 2))
		return &(argument[2]);
	if ((strlen (argument) == 2) && (argument[0] == '-'))
		return &(argument[1]);
	return NULL;
}

gboolean exarg_check_help_argument (gchar * argument)
{
	if (!g_strcasecmp (argument, "help") || !g_strcasecmp (argument, "?"))
		return TRUE;
	return FALSE;
}

void     exarg_wrap_and_print (gchar * text, gint size_to_wrap, gchar * fill_text)
{
	gchar    ** stringv      = NULL;
	gint        i;
	gint        sum          = 0;
	gchar     * working_line = NULL;
	gchar     * aux_string   = NULL;
	gboolean    first_line   = TRUE;
		
	if (strlen (text) <= size_to_wrap) {
		g_print ("%s\n", text);
		return;
	}
	
	stringv = g_strsplit (text, " ", 0);
	for (i = 0; stringv[i]; i++) {
		if (working_line) {
			aux_string   = working_line;
			working_line = g_strdup_printf ("%s %s", working_line, stringv[i]);
			g_free (aux_string);
		}else
			working_line = g_strdup (stringv[i]);
		sum = sum + strlen (stringv[i]) + 1;

		if (sum >= size_to_wrap) {
			sum = 0;
			if (first_line) {
				first_line = FALSE;
				g_print ("%s\n", working_line);
			}else
				g_print ("%s%s\n", fill_text, working_line);
			g_free (working_line);
			working_line = NULL;
		}

	}
	if (sum) {
		g_print ("%s%s\n", fill_text, working_line);
		g_free (working_line);
	}
	g_strfreev (stringv);
	
	
	return;
}

void __exarg_show_usage_foreach (gpointer key, gpointer value, gpointer user_data)
{
	gchar           ** string_aux = (gchar **) user_data;
	gchar            * str        = (* string_aux);
	ExArgNodeOption  * node       = (ExArgNodeOption  *) value;

	if (node->arg_short_name)
		(*string_aux) = g_strdup_printf (" [-%s|--%s]", node->arg_short_name, node->arg_name);
	else
		(*string_aux) = g_strdup_printf (" [--%s]", node->arg_name);

	if (str && (* str)) {
		(*string_aux) = g_strdup_printf ("%s%s", str, (*string_aux));
		g_free (str);
	}
	return;
}

void exarg_show_usage (gboolean show_header)
{
	gchar * string_aux = NULL;
	
	if (show_header && (__exarg_usage_header && (* __exarg_usage_header)))
		g_printerr (__exarg_usage_header);

	g_printerr ("Usage: %s ", exarg_exec_name);
	
	g_hash_table_foreach (argument_options, __exarg_show_usage_foreach, &string_aux);
	exarg_wrap_and_print (string_aux, 55, "               ");
	g_printerr ("\n");

	if (show_header && (__exarg_post_usage_header && (* __exarg_post_usage_header)))
		g_printerr (__exarg_post_usage_header);
	
	return;
}

void __exarg_show_help_foreach (gpointer key, gpointer value, gpointer user_data)
{
	ExArgNodeOption * node = (ExArgNodeOption  *) value;
	if (node->arg_short_name) {
		g_printerr ("  -%s, --%-18s       ", 
			    node->arg_short_name, node->arg_name);
		exarg_wrap_and_print (node->description, 40, "                                 ");
	}else {
		g_printerr ("  --%-22s       ",
			    node->arg_name);
		exarg_wrap_and_print (node->description, 40, "                                 ");
	}
	return;
}

void   exarg_show_help () 
{

	if (__exarg_help_header && (* __exarg_help_header))
		g_printerr (__exarg_help_header);

	exarg_show_usage (FALSE);

	g_printerr ("\nCommand options:\n");
	g_hash_table_foreach (argument_options, __exarg_show_help_foreach, NULL);

	g_printerr ("\nHelp options:\n");
	g_printerr ("  -?, --help                    Show this help message.\n");
	g_printerr ("  --usage                       Display brief usage message.\n");

	if (__exarg_post_help_header && (* __exarg_post_help_header))
		g_printerr (__exarg_post_help_header);

	return;
}

void exarg_check_argument_value (gchar ** argv, gint iterator) 
{

	if (argv[iterator] == NULL) {
		g_printerr ("error: not defined value for argument: %s, exiting..",
			    argv[iterator - 1]);
		exit (-1);
	}

	if (exarg_is_argument (argv[iterator])) {
		g_printerr ("error: not defined value for argument: %s, instead found another argument: %s, exiting..",
			    argv[iterator - 1],
			    argv[iterator]);
		exit (-1);
	}

	return;
}



/** 
 * \brief Makes exarg to start parsing argument options.
 *
 * Makes LibExArg to start command line parsing by using arguments
 * installed. 
 * 
 * Once this functions is called it is posible to call the set of
 * function which returns data obtained: \ref exarg_is_defined,
 * \ref exarg_get_int, \ref exarg_get_string and \ref exarg_get_params.
 *
 * The advantage is that LibExArg allows you to get access to command
 * line data from any point of your program.
 * 
 * @param argc the argc value the current application have received.
 * @param argv the argv value the current application have received.
 */
void       exarg_parse         (gint         argc,
				gchar     ** argv)
{
	gint              iterator = 1;
	ExArgNodeOption * node;
	gchar           * argument;

	/* check how many argument have been defined */
	if (iterator == argc)
		return;

	exarg_exec_name = (gchar *) g_basename (argv[0]);
	
	/* iterate over all arguments */
	while (iterator < argc) {
		if (!exarg_is_argument (argv[iterator])) {
			/* save and increase iterator */
			params = g_list_append (params, argv[iterator]);
			iterator++;
			continue;
		}
		
		/* get clean argument */
		argument = exarg_get_argument (argv[iterator]);

		/* check for help argument */
		if (exarg_check_help_argument (argument)){
			exarg_show_help ();
			exarg_end ();
			exit (0);
		}
		
		/* check for usage argument */
		if (!g_strcasecmp (argument, "usage")) {
			exarg_show_usage (TRUE);
			exarg_end ();
			exit (0);
		}

		/* get node argument */
		node = exarg_lookup_node (argument);
		if (node == NULL) {
			g_printerr ("argument not found: %s, try: %s --help\n",
				    argv[iterator], exarg_exec_name);
			exarg_end ();
			exit (-1);
		}
		
		/* check if node is defined */
		if (node->is_defined) {
			g_printerr ("argument %s already defined, exiting..\n",
				    argv[iterator]);
			exarg_end ();
			exit (-1);
		}

		
		/* set this argument to be defined */
		node->is_defined = TRUE;
		switch (node->type) {
		case EXARG_NONE:
			/* nothing to do */ 
				break;
		case EXARG_INT:
			/* save int value */
			iterator++;

			exarg_check_argument_value (argv, iterator);

			node->int_value    = atoi (argv[iterator]);
			break;
		case EXARG_STRING:
			iterator++;
			
			exarg_check_argument_value (argv, iterator);

			node->string_value = argv[iterator];

			break;
		}
		
		/* update iterator */
		iterator++;
	}
	return;
}

/**
 * \brief Ends exarg library execution.
 * 
 * Terminates the exarg function. This function will free all
 * resources allocated so the library cannot be used any more for the
 * current execution.
 *
 * This function is reatrant. Several threads can actually call this
 * function. It will take care about making only one thread to
 * actually free resources.
 **/
void       exarg_end           ()
{
	g_static_mutex_lock (&__exarg_destroy_mutex);
	if (argument_options == NULL) {
		g_static_mutex_unlock (&__exarg_destroy_mutex);
		return;
	}

	g_hash_table_destroy (argument_options);
	argument_options = NULL;
	g_static_mutex_unlock (&__exarg_destroy_mutex);

	return;
}

/**
 * \brief Disable autohelp.
 * 
 * Allows to disable auto help and --help and -? option
 * recognition. This is useful if you don't want your program to
 * accept --help and --usage options.
 **/
void       exarg_disable_help ()
{
	return;
}

/**
 * \brief Adds user defined header to automatic usage command generated.
 * 
 * Once a program is linked to libexarg is able to produce a usage
 * help by accepting the --usage command line option. But that usage
 * help is showed as is. This function allows you to define a header
 * to be showed preceding the usage info. You can use this, for
 * example, to introduce your copyright.
 *
 * This function doesn't make a copy of the given string so you
 * must not free the header provided or use an static string.
 **/
void       exarg_add_usage_header (gchar * header)
{
	__exarg_usage_header = header;
}

/**
 * \brief Adds user defined header to automatic help command generated.
 * 
 * Once a program is linked to libexarg is able to produce a help by
 * accepting the --help or -? command line option. But that help is
 * showed as is. This function allows you to define a header to be
 * showed preceding the help info. You can use this, for example, to
 * introduce your copyright.
 *
 * This function doesn't make a copy of the given string so you
 * must not free the header provided or use an static string.
 **/
void       exarg_add_help_header  (gchar * header)
{
	__exarg_help_header = header;
}

/**
 * \brief Adds user defined post header for usage command.
 * 
 * This function works pretty much like \ref exarg_add_usage_header but
 * adding the header provided to be appended at the end of the usage
 * automatically generated.
 **/
void       exarg_post_usage_header (gchar * post_header)
{
	__exarg_post_usage_header = post_header;
}

/**
 * \brief Adds user defined post header for help command.
 * 
 * This function works pretty much like \ref exarg_add_help_header but
 * adding the header provided to be appended at the end of the help
 * automatically generated.
 **/
void       exarg_post_help_header (gchar * post_header)
{
	__exarg_post_help_header = post_header;
}

/**
 * \internal
 * Internal ExArg function. This function allows \ref exarg_install_arg to
 * check if there are already installed an short argument as the one
 * been installed.
 **/
void __exarg_check_short_arg (gpointer key, gpointer value, gpointer user_data)
{
	ExArgNodeOption * node         = (ExArgNodeOption  *) value;
	gchar           * arg_to_check = (gchar *) user_data;

	if (node->arg_short_name != NULL) {
		if (!g_strcasecmp (node->arg_short_name, arg_to_check)) {
			g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "installing a short argument which already have been used: %s, exiting..",
			       arg_to_check);
			exit (-1);
		}
	}

	return;
}

/**
 * \brief Installs a new command line to be accepted.
 * 
 * This functions allows you to install new arguments to be
 * accepted. Every argument passed in to the program which is not
 * installed throught this function will no be accepted and will
 * generate a non recognized command line error.
 *
 * Let's see an example on how to use exarg to install the --version
 * argument. You should call exarg as follows:
 * 
 * \code
 *       exarg_install_arg ("version", "v", EXARG_NONE, 
 *                          "show program version");
 *
 * \endcode
 * 
 * Previous line have installed a argument called "version" which will
 * be invoked by the user as --version. The short alias
 * in this case is "v" which as we have see will allow user to invoke
 * your program as -v.
 *
 * Because you could install arguments which may conflict, this
 * function will abort the program execution on that case. This will
 * ensure you, as programer, to have a binary compiled with no
 * problems due to exarg.
 *
 * Later on, you can use \ref exarg_is_defined to check the status of
 * installed arguments. Once you install an argument the user may
 * invoke it and you could check that status using previous
 * function. Because the "version" argument type is \ref EXARG_NONE you can
 * only use \ref exarg_is_defined. But, for other argument types as
 * \ref EXARG_STRING, you can also use \ref exarg_get_string function to get the
 * value defined by user. Let's see an example.
 *
 * \code
 *       exarg_install_arg ("load-library", "l", EXARG_STRING,
 *                          "loads the library defined as argument");
 * \endcode
 * 
 * This previous line will allow user to invoke your program as 
 * --load-library path/to/lib. Then you can use \ref exarg_get_string to get
 * the path/to/lib value by using:
 * 
 * \code
 *       // check user have defined this argument
 *       if (exarg_is_defined ("load-library")) {
 *
 *             arg_value = exarg_get_string ("load-library");
 *
 *             g_print ("The library defined by user was: %s\n", 
 *                      arg_value);
 *       }
 * 
 * \endcode
 *
 * You must install all argument before calling \ref exarg_parse. That function
 * will parse argument options by using installed arguments. Check the info
 * about that function.
 * 
 * ExArg is not thread-safe which doesn't means anything wrong but you
 * must call all exarg function from the same thread to get the right
 * results. 
 *
 * This function will not do a copy from arg_name, arg_short_name or
 * description. This means you should not free that values while
 * using exarg. To end exarg using check \ref exarg_end. It is recomended to use
 * static values and shows on previous examples.
 **/
void       exarg_install_arg  (gchar     * arg_name, 
			       gchar     * arg_short_name, 
			       ExArgType   type,
			       gchar     * description)
{
	ExArgNodeOption * node;

	/* init hash table */
	if (argument_options == NULL) 
		argument_options = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
	else {
		/* check if there are an argument with the same name */
		if (exarg_lookup_node (arg_name)) {
			g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "installing an argument already installed ('%s'), exiting..", arg_name);
			exarg_end ();
			exit (-1);
		}
		
		/* check if there are an shor argument with the same
		 * name */
		if (arg_short_name != NULL)
			g_hash_table_foreach (argument_options, __exarg_check_short_arg, arg_short_name);
	}
		
	/* create node option */
	node                 = g_new0 (ExArgNodeOption, 1);
	node->arg_name       = arg_name;
	node->arg_short_name = arg_short_name;
	node->type           = type;
	node->description    = description;

	/* install it */
	g_hash_table_replace (argument_options, node->arg_name, node);

	return;
}

/**
 * \brief Installs several command lines to be accepted.
 * 
 * This function does the same that \ref exarg_install_arg but making 
 * group of argument to be installed in one step.
 *
 * Think about installing two argument. To install them we have to do
 * the following:
 *
 * \code
 *    exarg_install_arg ("version", "v", 
 *                       EXARG_NONE, "show argument version");
 *    exarg_install_arg ("load-library", "l", 
 *                       EXARG_STRING, "load my library");
 * \endcode
 *
 * Because some people doesn't like to perform several calls to the
 * same funcion the previous code can be done by using one step as:
 *
 * \code
 *    exarg_install_argv (2, "version", "v", EXARG_NONE, 
 *                        "show argument version", 
 *                        "load-library", "l",
 *                        EXARG_STRING, "load my library");
 * \endcode
 *
 * When you install argument this way you have to specify how many
 * argument is going to be installed. In this case that number is
 * 2. This allows exarg to know how many arguments needs to search
 * inside the argv. If something is wrong while specifying the number
 * or argument or the argument information itself you'll get a
 * segmentation fault or something similar.
 *
 * While calling to \ref exarg_install_arg the short_name argument can be
 * optional. This is *NOT* applied to this function. If you don't want
 * to define a short argument name you must use NULL as value.
 * 
 * @param num_arg Must be at least 1 or error will happen and library will abort
 * application execution.
 **/
void       exarg_install_argv (gint num_arg, ...)
{
	va_list      args;
	gint         iterator;
	gchar      * arg_name;
	gchar      * arg_short_name;
	ExArgType    type;
	gchar      * description;

	if (num_arg <= 0) {
		g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "specifying a number of arguments invalid=%d, exiting..",
		       num_arg);
		exarg_end ();
		exit (-1);
	}

	va_start (args, num_arg);

	for (iterator = 0; iterator < num_arg; iterator++) {

		/* get the argument info */
		arg_name       = va_arg (args, gchar *);
		arg_short_name = va_arg (args, gchar *);
		type           = va_arg (args, ExArgType);
		description    = va_arg (args, gchar *);
		
		exarg_install_arg (arg_name, arg_short_name, type, description);
	}
	va_end (args);
	return;
}


/**
 * \brief Allows to check if a user have defined a command.
 * 
 * Once an argument is installed the user may or may not define it. To
 * define it simply means to use this argument as command line
 * option. This function allows to check if an argument was used for
 * the current command line option.
 *
 * This function is expecting to receive the argument name to lookup
 * not the short name. User may be using the short name to invoke the
 * argument but the lookup is allways done by using the argument name.
 * 
 * \return TRUE if argument was defined or FALSE if not.
 **/
gboolean   exarg_is_defined   (gchar     * arg_name)
{
	ExArgNodeOption * node;

	/* perform some environment checks */
	g_return_val_if_fail (arg_name, FALSE);
	g_return_val_if_fail (argument_options, FALSE);

	node = exarg_lookup_node (arg_name);
	if (node == NULL) {
		return FALSE;
	}
	return node->is_defined;
}

/** 
 * @brief Allows to check several values to be defined at the same time.
 *
 * This allows to check if several command line options have been
 * defined at the same time. 
 * Example:
 * \code
 *   if (exarg_is_definedv ("param1", "param2", NULL)) {
 *         //param1 and param2 have been defined by the user
 *   }
 * \endcode
 *
 * This is a short way for actually doing:
 * \code
 *  if (exarg_is_defined ("param1") && exarg_is_defined ("param2")) {
 *      // param1 and para2 have been defined by the user
 *  }
 * \endcode
 * 
 * Do not forget to pass a NULL value for the last item.
 * 
 * @return TRUE if all values are defined, otherwise FALSE.
 */
gboolean   exarg_is_definedv      (gchar * first_value, ...)
{
	gboolean   result = TRUE;
	gchar    * string_value;
	va_list    args;

	g_return_val_if_fail (first_value, FALSE);

	/* check first value */
	if (!exarg_is_defined (first_value))
		return FALSE;

	/* open stdargs */
	va_start        (args, first_value);
	string_value = va_arg (args, gchar *);

	/* check for last NULL value */
	while ((string_value != NULL) && (* string_value)) {
		/* check next value */
		result       = result && exarg_is_defined (string_value);

		/* get next value */
		string_value = va_arg (args, gchar *);
	}

	/* return current value */
	va_end (args);
	return result;
}

/**
 * \brief Allows to get defined string for a given command line.
 * 
 * Returns the value associated with the argument name. The returned
 * value must not be deallocated. If it is needed a string copy use
 * \ref exarg_get_string_alloc.
 * 
 * \return The value associated with the param or NULL if fail.
 **/
gchar    * exarg_get_string   (gchar     * arg_name)
{
	ExArgNodeOption * node;

	g_return_val_if_fail (arg_name,         NULL);
	g_return_val_if_fail (argument_options, NULL);

	node = exarg_lookup_node (arg_name);
	if (node == NULL) {
		g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "trying to get an string value for an argument which haven't been defined, exiting..");
		exit (-1);
	}
	if (node->type != EXARG_STRING) {
		g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
		       "trying to get an string value from an argument which is not defined as string, exiting..");
		exit (-1);
	}
	
	return node->string_value;
}

/**
 * \brief Allows to get defined string for a given command line allocating the result.
 * 
 * Returns the value associated with the argument name. The returned
 * value must be <b>deallocated</b> using g_free. You can also use
 * \ref exarg_get_string to get a value that doesn't need to be
 * unrefered.
 * 
 * \return The value associated with the param or NULL if fail.
 **/
gchar    * exarg_get_string_alloc (gchar * arg_name)
{
	gchar * string_value;

	string_value = exarg_get_string (arg_name);
	if (string_value == NULL)
		return NULL;
	return g_strdup (string_value);
}

/**
 * \brief Allows to get int value for a given command line argument.
 * 
 * Returns the value associated with the argument name. The returned
 * may be 0 which means the user may not defined that argument or may
 * defined it by using the 0 as value. To avoid confusing cases you
 * should use \ref exarg_is_defined to know if the user have used that
 * argument and then call this function.
 * 
 * \return the value defined for this argument. If the argument
 * is null or the exarg is not initialized the function will return -1.
 **/
gint       exarg_get_int      (gchar     * arg_name)
{
	ExArgNodeOption * node;

	g_return_val_if_fail (arg_name,         -1);
	g_return_val_if_fail (argument_options, -1);

	node = exarg_lookup_node (arg_name);
	if (node == NULL) {
		g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "trying to get an int value for an argument which haven't been defined, exiting..");
		exit (-1);
	}

	if (node->type != EXARG_INT) {
		g_log (LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "trying to get an int value from an argument which is not defined as int, exiting..");
		exit (-1);
	}
	
	return node->int_value;
}

/**
 * \brief Returns free params defined at command line.
 * 
 * Every value which is not an argument is considered to be a
 * parameter and store to be returned by this function.
 *
 * An example about using this function could be the next. A program
 * have defined several arguments using the function \ref exarg_install_arg
 * but that program also needs to receive several file path. Once the
 * \ref exarg_parse have detected all argument, all parameters are stored
 * to be returned on this function. 
 *
 * A program which may receive the argument --save-all but also several
 * files:
 *
 * \code
 *    a_program --save-all file1 file2 file3
 * \endcode
 *
 * Will need to call this function to get the list: file1 file2 and
 * file3. 
 * 
 * \return the list of params for the program. The list return
 * should not be freed
 **/
GList    * exarg_get_params   ()
{
	return params;

}


/** 
 * @brief Allows to get the number of parameters that were defined by
 * the user. 
 *
 * Check also the following function \ref exarg_get_params_num.
 * 
 * @return The number of parameters starting from 0.
 */
gint       exarg_get_params_num   ()
{
	if (params == NULL)
		return 0;
	return g_list_length (params);
}





/* @} */

