/* Hey emacs, show me this like 'c': -*- c -*-
 *
 * xml-rpc-gen: a protocol compiler for the XDL definition language
 * Copyright (C) 2006 Advanced Software Production Line, S.L.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

/* global include */
#include <xml-rpc.h>


#define HELP_HEADER "xml-rpc-gen: a protocol compiler for the XDL language\n\
Copyright (C) 2006  Advanced Software Production Line, S.L.\n\n"


#define POST_HEADER "\nIf you have question, bugs to report, patches, you can reach us\n\
at <vortex@lists.aspl.es>."

/** 
 * @internal DTD declaration.
 */
axlDtd * dtd;


void show_tutorial ()
{
	g_print (HELP_HEADER);
	g_print ("To use xml-rpc-gen:\n\n");
	g_print ("1) Define an XDL file (or IDL file), describing your \n");
	g_print ("   XML-RPC interface as defined in the documentation. \n");
	g_print ("2) Comple the example using the following:\n");
	g_print ("   xml-rpc-gen test.xdl\n\n");

	g_print ("Try also xml-rpc-gen --help\n");
}


void xml_rpc_gen_finish ()
{
	/* terminate exarg library */
	exarg_end ();

	/* release the dtd */
	axl_dtd_free (dtd);

	/* terminate axl library */
	axl_end ();
	
	return;
}

/** 
 * @internal
 * Shows current version.
 */
void show_version ()
{
	g_print (VERSION);
}

/** 
 * @internal
 * @brief Compiles the provided file.
 * 
 */
gboolean xml_rpc_gen_compile_selected (gchar * selected)
{
	axlError * error;
	axlDoc   * doc = NULL;

	axlNode  * node;
	axlNode  * service;

	gchar    * out_dir;
	gchar    * comp_name;

	/* load the xml document */
	if (! exarg_is_defined ("idl-format"))
		doc = axl_doc_parse_from_file (selected, &error);

	/* check result */
	if (doc == NULL) {
		/* check if the input format was expected to be xml
		   (XDL) */
		if (exarg_is_defined ("xdl-format")) {
			g_printerr ("error: unable to parse file '%s', expected XDL format. Error was:\n%s",
				    selected, axl_error_get (error));
			axl_error_free (error);
			/* stop from parsing */
			return FALSE;
		}
	    
		/* try to load the document as an IDL definition using
		 * the structured format */
		doc = xml_rpc_gen_translate_idl_to_xdl (selected, &error);

		/* seems that the document provided is either not XDL
		 * and IDL. */
		if (doc == NULL) {
			/* report the error */
			g_printerr ("error: unable to parse file '%s', error was:\n%s",
				    selected, axl_error_get (error));
			axl_error_free (error);
			
			/* return that the document wasn't compiled */
			return FALSE;
		}
	}else {
		xml_rpc_report ("detected XDL format definition..\n");
	}

	xml_rpc_report ("document is well-formed: %s..\n", selected);

	/* validate file received */
	if (! axl_dtd_validate (doc, dtd, &error)) {
		/* report the error */
		g_printerr ("error: validation failed for '%s', error was:\n%s",
			    selected, axl_error_get (error));
		axl_error_free (error);

		/* return that the document wasn't compiled */
		return FALSE;
	}

	xml_rpc_report ("document is valid: %s..\n", selected);

	/* get out dir */
	if (exarg_is_defined ("out-dir")) 
		out_dir = exarg_get_string ("out-dir");
	else
		out_dir = "out";

	/* get the name node */
	comp_name = axl_doc_get_content_at (doc, "/xml-rpc-interface/name", NULL);

	/* trim the name */
	axl_stream_trim (comp_name);
	
	xml_rpc_report ("component name: '%s'..\n", comp_name);

	/* get a service reference */
	service = axl_doc_get (doc, "/xml-rpc-interface/service");

	/* now produce the stub required */
	if (! exarg_is_defined ("only-server"))
		xml_rpc_c_stub_create (doc, out_dir, comp_name);
	else
		xml_rpc_report ("client stub have been disabled..\n");
	
	/* now produce the server required */
	if (! exarg_is_defined ("only-client"))
		xml_rpc_c_server_create (doc, out_dir, comp_name);
	else
		xml_rpc_report ("server stub have been disabled..\n");
	
	/* check if the autoconf generation files is activated */
	if (! exarg_is_defined ("disable-autoconf")) {
		xml_rpc_report ("generating autoconf files..\n");
		xml_rpc_autoconf_c_stub_create (doc, out_dir, comp_name);
		xml_rpc_autoconf_c_server_create (doc, out_dir, comp_name);
	}else {
		xml_rpc_report ("autoconf files have been disabled..\n");
	}

	/* document parsed */
	axl_doc_free (doc);
	
	xml_rpc_report ("compilation ok");
	
	return TRUE;
}

/** 
 * @internal
 *
 * Compiles the provided document.
 */
void xml_rpc_gen_compile ()
{
	GList * files;
	GList * list_aux;

	/* get file list */
	files    = exarg_get_params ();
	list_aux = g_list_first (files);

	/* iterate over all elements */
	for (;list_aux; list_aux = g_list_next (list_aux)) {
		/* check that the file exist and it is regular */
		if (! g_file_test (list_aux->data, G_FILE_TEST_IS_REGULAR)) {
			g_printerr ("error: file '%s' doesn't exists or it is not a regular file.\n",
				    list_aux->data);
			return;
		}

		/* drop a log */
		xml_rpc_report ("compiling: %s..\n", list_aux->data);

		/* compile the selected file */
		if (!xml_rpc_gen_compile_selected (list_aux->data))
			return;
	}
	
	/* nothing more to do */
	return;
}

int main (int argc, char **argv)
{
	axlError * error;
	gchar    * file;

	/* init exarg library */
	exarg_install_arg ("version", "v", EXARG_NONE,
			   "Shows version info");
	exarg_install_arg ("out-dir", "o", EXARG_STRING, 
			   "Configures the output directory to place stub and skeletons generated. Default value, if not provided, is \"out\", starting from the current directory.");
	exarg_install_arg ("disable-autoconf", "a", EXARG_NONE,
			   "Disables autoconf files generation. By default, activated.");
	exarg_install_arg ("only-client", "c", EXARG_NONE,
			   "Only produces the client connector, the stub library used to interface with the XML-RPC component. By default, the client stub is generated.");
	exarg_install_arg ("only-server", "s", EXARG_NONE,
			   "Only produces the server stub, the XML-RPC component that is receiving the invocation. By default, the server stub is generated.");
	exarg_install_arg ("xdl-format", "x", EXARG_NONE,
			   "Instruct the tool to expect and process only XDL formated files. This option is useful because the tool fails into parse IDL format if XDL parsing fails, getting not proper errors if a XDL format was defined. This option is not compatible with --idl-format.");
	exarg_install_arg ("idl-format", "i", EXARG_NONE,
			   "Instruct the tool to expect and process only IDL formated files");

	
	/* configure headers to be showed */
	exarg_add_usage_header  (HELP_HEADER);
	exarg_add_help_header   (HELP_HEADER);
	exarg_post_help_header  (POST_HEADER);
	exarg_post_usage_header (POST_HEADER);	

	/* init and parse arguments */
	exarg_parse (argc, argv);

	if (exarg_is_defined ("xdl-format") && 
	    exarg_is_defined ("idl-format")) {
		g_printerr ("error: --xdl-format and --idl-format options are incompatible.");

		/* do nothing */
		return -1;
	}

	/* install search paths */
	xml_rpc_support_add_search_path_ref (g_build_filename (INSTALL_DIR, NULL ));

	/* do not use the add_search_path_ref version to force
	   xml-rpc-gen library to perform a local copy path */
	xml_rpc_support_add_search_path (".");


	/* init the axl library and load the DTD */
	axl_init ();
	
	/* file the file to load */
	file = xml_rpc_support_find_data_file ("xml-rpc.dtd");
	if (file == NULL) {
		/* show error found */
		g_printerr ("error: unable to find DTD definition: 'xml-rpc.dtd', check your installation\n");

		/* terminate xml-rpc-gen tool */
		xml_rpc_gen_finish ();
		return -1;
	}

	/* load the DTD XML-RPC definition */
	dtd = axl_dtd_parse_from_file (file, &error);
	if (dtd == NULL) {
		/* show error found */
		g_printerr ("error: unable to load dtd file '%s', check your installation, \n%s\n",
			    "xml-rpc.dtd", axl_error_get (error));
		axl_error_free (error);

		/* terminate xml-rpc-gen tool */
		xml_rpc_gen_finish ();
		return -1;
	}
	
	/* release the file path created */
	g_free (file);

	/* check and show program version */
	if (exarg_is_defined ("version")) {
		/* show current version */
		show_version ();

		/* terminate xml-rpc-gen tool */
		xml_rpc_gen_finish ();
		return 0;
	}

	/* check the number of parameters */
	if (exarg_get_params_num () > 0) {
		/* a xdl file was received */
		xml_rpc_gen_compile ();

		/* terminate the xml-rpc-gen tool */
		xml_rpc_gen_finish ();
		return 0;
	}

	/* show mini tutorial */
	show_tutorial ();

	/* terminate xml-rpc-gen tool */
	xml_rpc_gen_finish ();
	return 0;
}
