/* Hey emacs, show me this like 'c': -*- c -*-

 * xml-rpc-gen: a protocol compiler for the XDL definition language
 * Copyright (C) 2006 Advanced Software Production Line, S.L.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

#include <xml-rpc.h>

GList        * __xml_rpc_support_search_path    = NULL;
FILE         * opened_file                      = NULL;
gchar        * opened_file_name                 = NULL;
gchar        * original_user_file_to_merge      = NULL;
gint           xml_rpc_support_num_tabs         = 0;

/** 
 * @internal
 *
 * Allows to create a directory providing a path that is build using a
 * printf-like format, and creating all directories through the path
 * provided.
 * 
 * @param format A printf-like format especifying the directory to be created.
 * 
 * @return A newly allocated string returning the directory
 * created. The result must be unrefered using g_free.
 */
gchar * xml_rpc_support_create_dir      (gchar * format, ...)
{
	gchar   * result_dir       = NULL;
	gchar  ** dir_elements     = NULL;
	gchar   * actual_dir_name  = NULL;
	gchar   * string_ptr       = NULL;
	gint      iterator         = 0;
	gint      build_iterator   = 0;
	va_list   args;
	
	/* open the std args */
	va_start (args, format);
	
	/* create the result dir */
	result_dir = g_strdup_vprintf (format, args);

	/* close the std args */
	va_end (args);

	/* for each element inside the path, create the directory */
	dir_elements    = g_strsplit (result_dir, "/", 0);
	iterator        = 0;


	/* for each piece inside the blocks found for the directory
	 * path */
	while (dir_elements[iterator] != NULL) {
		
		for (build_iterator = iterator; build_iterator <= iterator; build_iterator++) {
			string_ptr = actual_dir_name;
			if (actual_dir_name) {
				/* build a new directory path from
				 * previous pieces */
				actual_dir_name = g_strdup_printf ("%s/%s", actual_dir_name, 
								   dir_elements[build_iterator]);
				g_free (string_ptr);
			}else {
				/* actual directory name, the first
				 * one piece created */
				actual_dir_name = g_strdup_printf ("%s", dir_elements[0]);
			}
		}
		
		iterator++;
		
		if (!g_file_test (actual_dir_name, G_FILE_TEST_IS_DIR)) {
			if (g_mkdir (actual_dir_name, 0770) < 0) {
				g_print ("*** Error: unable to create %s directory\n", actual_dir_name);
				exit (-1);
			}
		}
	}

	/* free path elements */
	g_strfreev (dir_elements);

	/* return current string */
	return result_dir;
}

/** 
 * @internal
 *
 * Allows to open a file, that is located at the path provided for the
 * printf-like format.
 * 
 * @param format A path that identifies the file to be opened.
 */
void    xml_rpc_support_open_file       (gchar * format, ...)
{
	va_list   args;
	va_start (args, format);
	original_user_file_to_merge = g_strdup_vprintf (format, args);
	va_end (args);

	/* because we are using a backslash representation to open
	 * files we need to do some foo to be run-time portable with
	 * windows platforms.  original_user_file_to_merge =
	 * af_gen_support_get_native_file_name
	 * (original_user_file_to_merge); */
	

	opened_file_name = g_strdup_printf ("%s.xml-rpc-gen-version", original_user_file_to_merge);
	if (opened_file) {
		fclose (opened_file);
		opened_file = NULL;
	}


	opened_file = fopen (opened_file_name, "w");
	if (!opened_file) {
		g_print ("*** Error: unable to create %s file\n", opened_file_name);
		exit (-1);
	}
	
	/* reset automatic tab indentation */
	xml_rpc_support_num_tabs = 0;

	return;
}

/** 
 * @internal
 *
 * Generates an error message and then exist according to the values provided.
 * 
 * @param message The message to report.
 *
 * @param must_exit A variable that controls if the function must
 * perform a call to the exit or not.
 */
void    xml_rpc_support_error           (gchar * message, gboolean must_exit, ...)
{
	gchar    * result;
	va_list    args;

	/* open the stdarg to get the message to print, and close
	 * it */
	va_start (args, must_exit);
	result = g_strdup_vprintf (message, args);
	va_end (args);

	printf ("error: %s\n", result);

	/* release memory used */
	g_free (result);

	/* check if a exit is required */
	if (must_exit)
		exit (-1);

	return;
}

/** 
 * @internal
 *
 * Moves a file from the detination provided to the destination provided.
 * 
 * @param from The source destination.
 * @param to The file destination.
 */
void   xml_rpc_support_move_file (gchar * from, gchar * to)
{

#ifdef G_OS_WIN32
	/* first remove the file to be overwrited. On unix platforms
	 * this is not necesary because move (rename) already
	 * overwrite destination file. On windows platform is not
	 * posible to rename a file with a file on the destination. */
	if (g_file_test (to, G_FILE_TEST_IS_REGULAR) && (g_unlink (to) != 0)) {
		xml_rpc_support_error ("unable to remove destination file: %s to be overwrited by: %s at rename operation",
				       FALSE, to, from);
		return;
	}
#endif

	/* now move the file */
	if (g_rename (from, to) != 0) {
		xml_rpc_support_error ("unable to move file: %s to %s", FALSE, from, to);
	}
	return;
}


/** 
 * @internal
 *
 * Close the current opened file.
 */
void    xml_rpc_support_close_file      ()
{
	gchar  * command_line   = NULL;
	gchar  * question       = NULL;
	gchar  * answer         = NULL;
	gchar  * indent_program = NULL;


	/* close the file opened */
	fclose (opened_file);
	opened_file = NULL;
	
	/* move the file to the final destination */
	xml_rpc_support_move_file (opened_file_name, original_user_file_to_merge);

	/* free original file to merge: currently not supported */
	g_free (original_user_file_to_merge);
	original_user_file_to_merge = NULL;

	/* free file name value */
	g_free (opened_file_name);
	opened_file_name = NULL;
	
	return;
}	


/** 
 * @internal
 *
 * Push current indent increasing the tab size.
 */
void    xml_rpc_support_push_indent     ()
{
	xml_rpc_support_num_tabs++;
}

/** 
 * @internal
 * 
 * Pop current indent decreasing the tab size.
 */
void    xml_rpc_support_pop_indent      ()
{
	xml_rpc_support_num_tabs--;
}

/** 
 * @internal
 *
 * Writes the provided content, using a printf-like format, to the
 * current opened file.
 * 
 * @param format 
 */
void xml_rpc_support_write (gchar * format, ...) {
	
	va_list     args;
	int       i;

	/* write tabular according to current indent configuration */
	if (xml_rpc_support_num_tabs > 0) {
		for (i = 0; i < xml_rpc_support_num_tabs; i++) {
			fprintf (opened_file, "\t");
		}
	}

	/* opend the std arg content to write it to the file */
	va_start (args, format);
	vfprintf (opened_file, format, args);
	va_end (args);
	
	/* done */
	return;
}

#define xml_rpc_support_mwrite xml_rpc_support_multiple_write

/** 
 * @internal
 *
 * Allow to write several lines at the same time using one call.
 *
 * Here is an example:
 * \code
 *    xml_rpc_support_multiple_write ("The first line..",
 *                                    "Second line...",
 *                                    "Third line...",
 *                                    NULL);
 * \endcode
 * 
 * Rembember to end every call to this function with a NULL value. 
 * 
 * @param first_line The first line of a series of comma separeted
 * strings ended with a NULL value.
 */
void    xml_rpc_support_multiple_write  (gchar * first_line, ...)
{
	va_list     args;
	gchar     * next_line;

	/* perform some checks */
	g_return_if_fail (first_line);

	/* opend the std args */
	va_start (args, first_line);

	/* write the first line */
	xml_rpc_support_write (first_line);

	/* and each line inside the parameters received */
	while ((next_line = va_arg (args, gchar *)))
		xml_rpc_support_write (next_line);

	/* close std args */
	va_end (args);

	return;
}

#define xml_rpc_support_write_sl  xml_rpc_support_sl_write

/** 
 * @internal
 *
 * Allows to write content to the opened file as defined for \ref
 * af_gen_support_write but without placing tabular content.
 * 
 * @param format A printf-like format especifying the content to write
 * to the file.
 */
void    xml_rpc_support_sl_write        (gchar * format, ...)
{
	va_list   args;

	/* write content defined by the parameters */
	va_start (args, format);
	vfprintf (opened_file, format, args);
	va_end (args);

	/* return */
	return;

}

/** 
 * @internal Drop a log to the console, with the line heading.
 *
 * This function should be used only for dropping messages or logs to
 * the console, not for reporting errors.
 * 
 * @param format A printf-like format containing a log to drop.
 */
void    xml_rpc_report (gchar * format, ...)
{
	gchar * result;

	va_list   args;

	/* write content defined by the parameters */
	va_start (args, format);
	result = g_strdup_vprintf (format, args);
	va_end (args);

	/* drop the log */
	g_print ("[*] %s", result);

	/* release the memory hold */
	g_free (result);

	/* return */
	return;
}


/** 
 * @internal
 * 
 * Makes the provided file, located at the path formed by the
 * printf-like function, to be executable.
 * 
 * @param format The file to be executable, especified by the a
 * printf-like path.
 */
void    xml_rpc_support_make_executable (gchar * format, ...)
{
	gchar * result;
	
	va_list   args;

	/* write content defined by the parameters */
	va_start (args, format);
	result = g_strdup_vprintf (format, args);
	va_end (args);

	/* make a chmod operation */
	if (chmod (result, 0770) < 0) {
		g_printerr ("error: unable to make executable the file: '%s'\n", result);
	}
	
	/* release the memory hold */
	g_free (result);

	/* return */
	return;
	
}

/** 
 * @internal
 * 
 * Allows to configure a new search path to locate files.
 * 
 * @param path The new path to be used as a base dir to locate files.
 */
void     xml_rpc_support_add_search_path     (gchar * path)
{
	g_return_if_fail (path);
	
	/* add search path */
	__xml_rpc_support_search_path = g_list_append (__xml_rpc_support_search_path,  g_strdup (path));

	return;
}

/** 
 * @internal
 * 
 * Allows to configure a new search path to locate files. However, the
 * parameter received is reference that is already allocated. The
 * function will not perform a copy.
 * 
 * @param path The new path to be used as a base dir to locate files.
 */
void     xml_rpc_support_add_search_path_ref (gchar * path)
{
	/* store the search path */
	__xml_rpc_support_search_path = g_list_append (__xml_rpc_support_search_path, path);

	return;
}

/** 
 * @internal
 *
 * Performs an internal lookup for the file requested, using the
 * configured path locations.
 * 
 * @param name The file name to search.
 * 
 * @return The full path for the file located or NULL if it fails.
 */
gchar  * xml_rpc_support_find_data_file      (gchar * name)
{
	GList * list_aux;
	gchar * file_name;

	/* foreach path installed */
	list_aux = g_list_first (__xml_rpc_support_search_path);
	for (; list_aux ; list_aux = g_list_next (list_aux)) {

		/* build a file path, trying to build ./ or .\ file
		 * path if path "." is used */
		if (g_strcasecmp (list_aux->data, "."))
			file_name = g_build_filename (list_aux->data, name, NULL );
		else
			file_name = g_strdup (name);

		/* check the file to be found */
		if (g_file_test (file_name, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
			return file_name;
		} /* end if */

		g_free (file_name);

	} /* end for */

	return NULL;
}

